/***************************************************************\
*	cqi.c							*
*	Invert the constant-Q transformed signals		*
*	Based on cqt.c						*
*	08nov90 dpwe						*
\***************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <strings.h>
/* personal lib */
#include <window.h>
#include <sndf.h>
#include <pvoc.h>

#define PI M_PI

#define SHORTMAX 32768

#define	DFLT_8VES	6	/* number of analysis octaves */
#define	DFLT_NBIN	24	/* number of filter bins per octave */
#define	DFLT_QFCT	64	/* default Q factor per bin */
#define	DFLT_SLOP	1	/* boost HF by 3db/oct? */
#define	DFLT_DC		1	/* remove DC from kernels ? */
#define DFLT_GAIN	1	/* multiply o/p by this */
#define MAXOCTS		12	/* limit to number of octaves */

#define OP_FEXT 	".irc"	/* default output file extension */

extern char * rindex();		/* why do I have to do this ? */

typedef struct {
    float mag, pha;
    } cpxpolar;

int   lf_cut;	/* global flag to use 3db/oct HF boost */
int   dc_out;	/* global flag to cancel dc from cos kernels */

char *programName;

main(argc,argv)
    int argc;
    char **argv;
    {
    char  *inpPath;
    float *kernels;
    long  nBins,nFrms,nOcts,octRatio;
    float qFactor;
    float botFrq;
    float gain;
    long  kernLen,hopSize;
    int   dilate;
    int   i,err = 0;
    long  dbgOct;		/* just one octave resynthed for debug */

    cpxpolar *result, *curRslt;
    SFSTRUCT *snd;
    float    *signal,*pf;
    long     sigLen,l;

    PVSTRUCT *pvh;
    long  frBs;
    long  pvFrams;

    char  savePath[80];

    WINDAT	*wdptr = NULL;
    WINDAT	*octwins[MAXOCTS];
    char	winName[32];

    programName = argv[0];
    savePath[0] = '\0';

    nOcts   = DFLT_8VES;
    nBins   = DFLT_NBIN;
    qFactor = DFLT_QFCT;
    lf_cut  = DFLT_SLOP;
    dc_out  = DFLT_DC;
    gain    = DFLT_GAIN;

    nOcts = -1;		/* to allow -o to specify unique resynth octave */
    /* Check arguments */
    for(i=1; i<argc && err==0; ++i)
	{
	if(argv[i][0]=='-')
	    switch(argv[i][1])
		{
	    case 'q': case 'Q':
		qFactor = atof(argv[++i]);
		if(qFactor<1)
		    {
		    fprintf(stderr,"%s: Q of %s = %.2f must be > 1\n",
			    programName, argv[i], qFactor);
		    err = 1;
		    }
		break;
	    case 'b':
		nBins = atol(argv[++i]);
		if(nBins < 2 || nBins > 300)
		    {
		    fprintf(stderr,"%s: Bins of %s = %d must be 2..300\n",
			    programName, argv[i], nBins);
		    err = 1;
		    }
		break;
	    case 'o':
		nOcts = atol(argv[++i]);
/*		if(nOcts < 1 || nOcts > MAXOCTS)
		    {
		    fprintf(stderr,"%s: Octs of %s = %d must be 1..%d\n",
			    programName, argv[i], nOcts, MAXOCTS);
		    err = 1;
		    }
 */		break;
	    case 's':
		lf_cut = !DFLT_SLOP;
		break;
	    case 'g':
		gain = atof(argv[++i]);
		break;
	    case 'd':
		dc_out = !DFLT_DC;		/* atoi(argv[++i]); */
		break;
	    default:
		fprintf(stderr,"%s: unrecog option: %s \n",
			programName, argv[i]);
		err = 1;
		break;
		}
	else{
	    if(inpPath==NULL)
		inpPath = argv[i];
	    else
		strcpy(savePath,argv[i]);
	    }
	}

    dbgOct = nOcts;

    if(err || inpPath == NULL)
	Usage();		/* never returns */
    if(err = PVReadFile(inpPath, &pvh))
	{
	fprintf(stderr,"pv error %s on %s\n", PVErrMsg(err), inpPath);
	exit(1);
	}
    result  = (cpxpolar *)PVDataLoc(pvh);
    hopSize = pvh->frameIncr;
    /* Assume nBins has been correctly set up (default or -b) */
    frBs = pvh->frameBsize;	/* should be nOcts*nBins */
    octRatio = (long)(.5+(pvh->maxFreq / pvh->minFreq));
    nOcts = 1L;
    while(octRatio>2 && ++nOcts<MAXOCTS)	octRatio>>=1L;
    
    if( (frBs/sizeof(cpxpolar))%nOcts != 0 )
	{
	fprintf(stderr,"%s: wds/frame of %d inconsistent with octs %d\n",
		programName, frBs/sizeof(cpxpolar), nOcts);
	exit(1);
	}
    nBins = frBs/(sizeof(cpxpolar)*nOcts);
/*    while(nOcts>9)
	{	nOcts /= 2;	nBins *= 2;	}
    while(nOcts<4)
	{	nOcts *= 2;	nBins /= 2;	}
 */
    nFrms   = pvh->dataBsize/pvh->frameBsize;
    sigLen  = hopSize * nFrms;

    if(pvh->info[0] == 'Q' && pvh->info[1] == '=')
	qFactor = atof(&pvh->info[2]);
    if(qFactor<1)
	{
	fprintf(stderr,"%s: Q of %s = %.2f must be > 1\n",
		programName, argv[i], qFactor);
	exit(1);
	}
    kernLen = qFactor*8;
    botFrq  = 2*PI/5;		/* slowest = 5 samples per cyc (fastest 2x) */

    kernels = (float *)malloc(kernLen*2*nBins*sizeof(float));
    if(!kernels)
	{
	fprintf(stderr,"malloc fail\n");
	exit(-1);
	}
    printf("Setup kernels..\n");
    SetupRKernels(nBins,qFactor,botFrq,kernLen,kernels,gain);
    /*    printf("Setup display..\n");
	  wdptr = NewWin("cqi",1);
	  printf("Let's see..\n");
	  for(i=0; i<1+0*nBins; ++i)
	  {
	  DoDisp(wdptr,&kernels[kernLen*i], kernLen+1);
	  DoDisp(wdptr,&kernels[kernLen*(i+nBins)], kernLen+1);
	  }
	  
    
      for(i=0; i<kernLen; ++i)	dumstor[i] = 0;
      dumstor[5] = 1;	dumstor[13] = .8; dumstor[14] = 1; dumstor[15] = .8;
      DoDisp(wdptr,dumstor, kernLen/8);
      UpSampleBy2(dumstor, kernLen/8, dumstor);
      DoDisp(wdptr,dumstor, kernLen/4);
      UpSampleBy2(dumstor, kernLen/4, dumstor);
      DoDisp(wdptr,dumstor, kernLen/2);
      UpSampleBy2(dumstor, kernLen/2, dumstor);
      DoDisp(wdptr,dumstor, kernLen);
      
      ExitGraph();
      exit(0);
      /* */
    /*  result  = (cpxpolar *)malloc(nFrms*nOcts*nBins*sizeof(cpxpolar));  */
    if( err = SFAlloc(&snd, sigLen*sizeof(float), SFMT_FLOAT,
		   pvh->samplingRate, pvh->channels, 4) )
	{
	fprintf(stderr,"sf error: %s\n",SFErrMsg(err));
	exit(1);
	}
    signal = (float *)SFDataLoc(snd);
    /* MUST be cleared to zero */
    l = sigLen;	pf = signal;
    while(l--)
	*pf++ = 0.0;
    curRslt = result; /* start at bottom octave, not + (nOcts-1)*nBins; */
    dilate  = 1;
    printf("Frames %d\n",nFrms);
    /* point to the first word of the top octave in the 1st frame */
    sigLen >>= (nOcts-1);
    hopSize >>= (nOcts-1);	/* setup for lowest (first) octave */
    for(i=0; i<nOcts; ++i)
	{
	sprintf(winName,"Oct %d",i);
/*	octwins[i] = NewWin(winName,1); */
	printf("Octave %d  ",i); fflush(stdout);
/*
    l = sigLen;	pf = signal;
    while(l--)
	*pf++ = 0.0;
 */
	if(dbgOct<0 || i == dbgOct)
	    UndoKernels(signal,sigLen,kernels,kernLen,
		     nBins,hopSize,nBins*nOcts,dilate,curRslt);
/*	DoDisp(octwins[i], signal, sigLen); */
	if(i<nOcts-1)	
/*	    if(dbgOct<0 || i < nOcts-2 )	*/
		UpSampleBy2(signal, sigLen, signal);
	sigLen  *= 2;
	hopSize *= 2;
	/*	dilate  /= 2;	*/
	curRslt += nBins;	/* forward a binlength - next 8ve up */
	}
    sigLen /= 2;
    hopSize /= 2;	/* undo last multiply in loop */
    if(snd->format != SFMT_SHORT)
	if( err = ConvertSoundFormat(&snd, SFMT_SHORT))
	    {
	    fprintf(stderr, "snd err: %s\n", SFErrMsg(err));
	    exit(1);
	    }
    if(savePath[0] == '\0')	/* no output file specified - defaul */
	SubstExtn(inpPath,OP_FEXT,savePath);
    /*    printf("pv magic %ld hedsiz %ld datsiz %ld fmt %d SR %.1f \n",
	  pvh->magic, pvh->headBsize, pvh->dataBsize,
	  pvh->format, pvh->samplingRate);	*/
    printf("\nwriting %s\n",savePath);	
    if(err = SFWrite(savePath, snd))
	{
	fprintf(stderr,"saving sf error: %s on %s\n",SFErrMsg(err),savePath);
	exit(1);
	}
    ExitGraph();
    }

Usage()			/* print out usage message and exit */
    {
    fprintf(stderr,
"Usage: %s [-q Qfact][-b bins/oct][-o octs][-s][-d] inFTfile [outSndfile]\n",
	    programName);
    fprintf(stderr,
	    "-q Qfact   is the q-factor of the bins       (default %6.1f)\n",
	    (float)DFLT_QFCT);
/*    fprintf(stderr,
	    "-b bins    is the number of filter per 8ve   (default %6d)\n",
	    DFLT_NBIN);
    fprintf(stderr,
	    "-o octs    is the # of 8ves, starting at top (default %6d)\n",
	    DFLT_8VES);
 */
    fprintf(stderr,
	    "-s         switches %3s 3dB/oct HF boost     (default %3s)\n",
	    (!DFLT_SLOP) ? "in ":"out", DFLT_SLOP ? "in ":"out");
    fprintf(stderr,
	    "-d         switches %3s dc removal frm kerns (default %3s)\n",
	    (!DFLT_DC) ? "in ":"out", DFLT_DC ? "in ":"out");
    fprintf(stderr,
	    "inFTfile   is a PVOC-type analysis file for input\n");
    fprintf(stderr,
	    "outSndfile is a SNDF soundfile, default is inStem+'%s'\n",
	    OP_FEXT);
    exit(1);
    }
