/***************************************************************\
*	cqifoo.c						*
*	One-off to generate some test signals			*
*	20may91, 08mar91 dpwe					*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#else
#include <stdio.h>
#include <malloc.h>
#include <strings.h>
#endif
#include <math.h>
#define PI M_PI		/* for that alone */
/* personal libs */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include <readaqt.h>
#include "cqlib.h"
#include "ivbuf.h"

#define DFLT_GAIN	1	/* multiply o/p by this */
#define DEF_EXT 	".irc"	/* default output file extension */
#define LOBES		16	/* lobes for resampler */

#define	DFLT_SLOP	1	/* boost HF by 3db/oct? */
#define DFLT_DC		1	/* cancel DC from cos kernels? */

/* to satisfy cqlib.c */
int   lf_cut = DFLT_SLOP;	/* global flag to use 3db/oct HF boost */
int   dc_out = DFLT_DC;	/* global flag to cancel dc from cos kernels */

/* static data */
       char	*programName;
static int	verb   = 0;		/* -v flag */
static float	gain   = DFLT_GAIN;

/* static fn prototypes */
#ifdef __STDC__
static long FileBuf(FILE *outFile, SAMPL *buf, long size, int flush);
static void PerformCQInversion(char *inPath, char *outPath);
static void Usage(void);
#else	/* no proto args */
static long FileBuf();
static void PerformCQInversion();
static void Usage();
#endif

main(argc,argv)
    int argc;
    char **argv;
    {
    int 	i = 0;
    int 	err = 0;
    char	*inPath = NULL, *outPath = NULL;

    programName = argv[0];
    /**** %% DPWE STD ARG HANDLER %% ****/
    i = 0;
    while(++i<argc && err == 0)
	{
	char c, *token, *arg, *nextArg;
	int  argUsed;

	token = argv[i];
	if(*token++ == '-')
	    {
	    if(i+1 < argc)	nextArg = argv[i+1];
	    else		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)
		{
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)
		    {
		case 'g':		gain = atof(arg);	argUsed = 1;
		    if(gain == 0.0)
			{  fprintf(stderr,"%s: gain of %s = %4.2f silly\n",
				   programName, arg, gain);
			   err = 1;	}
		    break;
		case 'v': 	verb = 1; break;
		default:	fprintf(stderr,"%s: unrec option %c\n",
					programName, c);
		                err = 1; break;
		    }
		if(argUsed)
		    {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		    }
		}
	    }
	else{
	    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	outPath = argv[i];
	    else{ fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
		  err = 1;  }
	    }
	}
    /* other tests on args HERE ! */

    if(err || inPath == NULL)
	Usage();		/* never returns */

    if(outPath == NULL)
	{
	outPath=(char *)malloc(sizeof(char)*(strlen(inPath)+strlen(DEF_EXT)));
	strcpy(outPath, inPath);
	SubstExtn(inPath, DEF_EXT, outPath);
	}
    /* args all dealt with.  On to the real business */
    PerformCQInversion(inPath, outPath);
    }

static long FileBuf(outFile, buf, size, flush) /* arg to accept done samples */
    FILE  *outFile;
    SAMPL *buf;
    long  size;
    int   flush;	/* irrelevent flag ? */
    {
/*    fprintf(stderr, "FileBuf: given %ld\n", size);	*/
    ScaleSampls(buf, 1, buf, 1, size, (LONGSAMPL)(gain * (float)SAMPLUNITY));
    Sampl2Short(buf, 1, (short *)buf, 1, size);	
    /** NAUGHTY - no mandate to clobber buf */
    fwrite(buf, sizeof(short), size, outFile);
    }

static void PerformCQInversion(inPath, outPath)
    char	*inPath;
    char 	*outPath;
    {
    CQTINFO 	*cqi;
    SFSTRUCT	*outSnd;
    FILE	*inFile, *outFile;
    long 	hopSize, kernLen, red;
    long	frame, trigSize, latency, prefill;
    float	*kernels, *rdBuf, *tmpBuf;
    float	qFactor;
    float	wt, reWeight[MAXBINS];
    int 	nBins, nOcts;
    int 	i, oct, err, done;
    char	desc[DESCLEN];
    RsmpDescType  *rsDptr;
    IvRSBufType   *ivb[MAXOCTS], *lastIv;
#ifdef DEBUGGRAPHS
    WINDAT	*wdptr[MAXOCTS];
#endif

    /* open input PV file */
    if( (inFile = OpenCQTPVfile(inPath, &cqi))==NULL )
	PVDie(PVE_NOPEN, inPath);
    if(cqi->totalFrames != PV_UNK_LEN)
	fprintf(stderr, 
		"Est input frames %d, output sample frames %ld (%.3f secs)\n",
		cqi->totalFrames,cqi->totalFrames*cqi->hopSize,cqi->duration);
    nBins = cqi->nBins;    nOcts = cqi->nOcts;    qFactor = cqi->qFactor;
    kernLen = cqi->kernLen;    hopSize = cqi->hopSize;

    /* create & open output SNDF file */
    if( err = SFAlloc(&outSnd, SF_UNK_LEN, SFMT_SHORT, cqi->pvh->samplingRate,
		      cqi->pvh->channels, SFDFLTBYTS) )
	SFDie(err, NULL);
    if( (outFile = SFOpenWrHdr(outPath, outSnd)) == NULL )
	SFDie(SFE_NOPEN, NULL);

    /* allocate buffers */
    rsDptr  = NewResampKernel(1, 2, LOBES, (float)sqrt(2.0)); /* 1:2 upsamp */
    rdBuf   = (float *)malloc(sizeof(cpxpolar)*nBins);
    tmpBuf  = (float *)malloc(sizeof(float)*kernLen);
    kernels = (float *)malloc(sizeof(float)*kernLen*2*nBins);
/*    SetupRKernels(nBins, qFactor, PI/2*cqi->guardBand, kernLen, kernels,
		  (FLOATARG)1.0/384.0);	*/
    NewSetupKernels(nBins,qFactor,cqi->guardBand*PI/2,kernLen,kernels,
		    1, (float)1.0/476.0);	/* empirical scale fix */

    for(i = 0; i<nBins; ++i)
	{
	ClearFloats(tmpBuf,1,kernLen);
	FileBuf(outFile, tmpBuf, kernLen, 1);
	ClearFloats(tmpBuf,1,kernLen);
	CopyFloats(kernels+2*i*kernLen,1,tmpBuf,1,kernLen, FWD);
	ScaleSampls(tmpBuf, 1, tmpBuf, 1, kernLen, 100.0);
	FileBuf(outFile, tmpBuf, kernLen, 1);
	}
    ClearFloats(tmpBuf,1,kernLen);
    FileBuf(outFile, tmpBuf, kernLen, 1);

    for(i = 0; i<kernLen; ++i)
	{
	printf("%.6f ", 500.0*kernels[i]);
	}

/*    fprintf(stderr, "%d\n",frame);	*/
    SFCloseWrHdr(outFile, outSnd);
    }

static void Usage()		/* print out usage message and exit */
    {
    fprintf(stderr,
"Usage: %s [-g gain][-v] inFTfile [outSndfile]\n",
	    programName);
    fprintf(stderr,
	    "-g gain    scale output by <gain>       (default %6.3f)\n",
	    (float)DFLT_GAIN);
    fprintf(stderr,
	    "-v         enable verbose mode (actually, no effect)\n");
    fprintf(stderr,
	    "inFTfile   is a PVOC-type analysis file for input\n");
    fprintf(stderr,
	    "outSndfile is a SNDF soundfile, default is inStem+'%s'\n",
	    DEF_EXT);
    exit(1);
    }
