/***************************************************************\
*	cqt.c							*
*	Some kind of constant-Q transform analysis		*
*	08nov90 dpwe						*
\***************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <strings.h>
/* personal library */
#include <window.h>
#include <sndf.h>
#include <pvoc.h>

#define PI M_PI

#define SHORTMAX 32768

#define	DFLT_8VES	6	/* number of analysis octaves */
#define	DFLT_NBIN	24	/* number of filter bins per octave */
#define	DFLT_QFCT	64	/* default Q factor per bin */
#define	DFLT_SLOP	1	/* boost HF by 3db/oct? */
#define DFLT_DC		1	/* cancel DC from cos kernels? */
#define MAXOCTS		20	/* most octaves possible */

#define OP_FEXT 	".aqt"	/* default output file extension */

extern char * rindex();		/* why do I have to do this ? */

int   lf_cut;	/* global flag to use 3db/oct HF boost */
int   dc_out;	/* global flag to cancel dc from cos kernels */

char *programName;

main(argc,argv)
    int argc;
    char **argv;
    {
    char  *inpPath;
    float *kernels;
    long  nBins,nFrms,nOcts;
    float qFactor;
    float botFrq, topFrq;	/* diff units */
    long  kernLen,hopSize;
    int   dilate;
    int   i,err = 0;

    cpxpolar *result, *curRslt;
    SFSTRUCT *inSnd;
    float    *signal;
    long     sigLen;

    PVSTRUCT *pvh;
    long  fsBs;
    long  pvFrams;

    char  savePath[80];
    char  qString[32];

    WINDAT	*wdptr = NULL;

    programName = argv[0];
    savePath[0] = '\0';

    nOcts   = DFLT_8VES;
    nBins   = DFLT_NBIN;
    qFactor = DFLT_QFCT;
    lf_cut  = DFLT_SLOP;
    dc_out  = DFLT_DC;

    /* Check arguments */
    for(i=1; i<argc && err==0; ++i)
	{
	if(argv[i][0]=='-')
	    switch(argv[i][1])
		{
	        case 'q': case 'Q':
		    qFactor = atof(argv[++i]);
		    if(qFactor<1)
			{
			fprintf(stderr,"%s: Q of %s = %.2f must be > 1\n",
				programName, argv[i], qFactor);
			err = 1;
			}
		    break;
		case 'b':
		    nBins = atol(argv[++i]);
		    if(nBins < 2 || nBins > 300)
			{
			fprintf(stderr,"%s: Bins of %s = %d must be 2..300\n",
				programName, argv[i], nBins);
			err = 1;
			}
		    break;
		case 'o':
		    nOcts = atol(argv[++i]);
		    if(nOcts < 1 || nOcts > MAXOCTS)
			{
			fprintf(stderr,"%s: Octs of %s = %d must be 1..%d\n",
				programName, argv[i], nOcts, MAXOCTS);
			err = 1;
			}
		    break;
		case 's':
		    lf_cut = !DFLT_SLOP;	/* atoi(argv[++i]); */
		    break;
		case 'd':
		    dc_out = !DFLT_DC;		/* atoi(argv[++i]); */
		    break;
		default:
		    fprintf(stderr,"%s: unrecog option: %s \n",
			    programName, argv[i]);
		    err = 1;
		    break;
		}
	else{
	    if(inpPath==NULL)
		inpPath = argv[i];
	    else
		strcpy(savePath,argv[i]);
	    }
	}

    if(err || inpPath == NULL)
	Usage();		/* never returns */
    kernLen = qFactor*8;
    hopSize = qFactor;		/* best be 2^n since will be halved each 8ve */
    botFrq  = 2*PI/5;		/* slowest = 5 samples per cyc (fastest 2x) */
    if(err = SFRead(inpPath, &inSnd))
	{
	fprintf(stderr,"sound error %s on %s\n", SFErrMsg(err), inpPath);
	exit(-1);
	}
    if(inSnd->format != SFMT_FLOAT)
	if( err = ConvertSoundFormat(&inSnd, SFMT_FLOAT))
	    {
	    fprintf(stderr, "snd err: %s\n", SFErrMsg(err));
	    exit(-1);
	    }
    signal  = (float *)SFDataLoc(inSnd);
    sigLen  = (inSnd->dataBsize)/sizeof(float);

    kernels = (float *)malloc(kernLen*2*nBins*sizeof(float));
    if(!kernels)
	{
	fprintf(stderr,"malloc fail\n");
	exit(-1);
	}
    printf("Setup kernels..\n");
    SetupKernels(nBins,qFactor,botFrq,kernLen,kernels);
/*    printf("Setup display..\n");
    wdptr = NewWin("cqt",1);
    printf("Let's see..\n");
    for(i=0; i<nBins; ++i)
	{
	DoDisp(wdptr,&kernels[kernLen*i], kernLen+1);
	DoDisp(wdptr,&kernels[kernLen*(i+nBins)], kernLen+1);
	}
*/
/*
    for(i=0; i<kernLen; ++i)	dumstor[i] = 0;
    dumstor[5] = 1;
    DoDisp(wdptr,dumstor, kernLen);
    DownSampleBy2(dumstor, kernLen, dumstor);
    DoDisp(wdptr,dumstor, kernLen/2);
    DownSampleBy2(dumstor, kernLen/2, dumstor);
    DoDisp(wdptr,dumstor, kernLen/4);
    DownSampleBy2(dumstor, kernLen/4, dumstor);
    DoDisp(wdptr,dumstor, kernLen/8);
    DownSampleBy2(dumstor, kernLen/8, dumstor);

    ExitGraph();
    exit(0);
/* */
    nFrms   = sigLen / hopSize;
/*  result  = (cpxpolar *)malloc(nFrms*nOcts*nBins*sizeof(cpxpolar));  */
    fsBs    = (nOcts*nBins)*sizeof(cpxpolar);
    pvFrams = nOcts*nBins*nFrms;
    topFrq  = 2* botFrq*(inSnd->samplingRate)/(2*PI);
    if( err = PVAlloc(&pvh, pvFrams*sizeof(cpxpolar), PVFLOAT,
	        inSnd->samplingRate, inSnd->channels, kernLen, hopSize, 
	        fsBs, PVPOLAR, topFrq/(float)(1L<<nOcts), topFrq, PVEXP, 32 ))
	{
	fprintf(stderr,"pv error: %s\n",PVErrMsg(err));
	exit(-1);
	}
    sprintf(pvh->info, "Q=%5.2f", qFactor);
    result  = (cpxpolar *)PVDataLoc(pvh);
    curRslt = result + (nOcts-1)*nBins;
    dilate  = 1;
    printf("Frames %d\n",nFrms);
    /* point to the first word of the top octave in the 1st frame */
    for(i=0; i<nOcts; ++i)
	{
	printf("Octave %d  ",i); fflush(stdout);
	oldAppKernels(signal,sigLen,kernels,kernLen,
		     nBins,hopSize,nBins*nOcts,dilate,curRslt);
	DownSampleBy2(signal, sigLen, signal);
	sigLen  /= 2;
	hopSize /= 2;
/*	dilate  *= 2;	*/
	curRslt -= nBins;	/* start back a binlength - next 8ve */
	}
    if(savePath[0] == '\0')	/* no output file specified - defaul */
	SubstExtn(inpPath,OP_FEXT,savePath);
/*    printf("pv magic %ld hedsiz %ld datsiz %ld fmt %d SR %.1f \n",
	   pvh->magic, pvh->headBsize, pvh->dataBsize,
	   pvh->format, pvh->samplingRate);	*/
    printf("\nwriting %s\n",savePath);	
    if(err = PVWriteFile(savePath, pvh))
	{
	fprintf(stderr,"saving pv error: %s\n",PVErrMsg(err));
	exit(-1);
	}
    ExitGraph();
    }

Usage()			/* print out usage message and exit */
    {
    fprintf(stderr,
"Usage: %s [-q Qfact][-b bins/oct][-o octs][-s][-d] inSndfile [outFTfile]\n",
	    programName);
    fprintf(stderr,
	    "-q Qfact   is the q-factor of the bins       (default %6.1f)\n",
	    (float)DFLT_QFCT);
    fprintf(stderr,
	    "-b bins    is the number of filter per 8ve   (default %6d)\n",
	    DFLT_NBIN);
    fprintf(stderr,
	    "-o octs    is the # of 8ves, starting at top (default %6d)\n",
	    DFLT_8VES);
    fprintf(stderr,
	    "-s         switches %3s 3dB/oct HF boost     (default %3s)\n",
	    (!DFLT_SLOP) ? "in ":"out", DFLT_SLOP ? "in ":"out");
    fprintf(stderr,
	    "-d         switches %3s kernel dc canc'l'n   (default %3s)\n",
	    (!DFLT_DC) ? "in ":"out", DFLT_DC ? "in ":"out");
    fprintf(stderr,
	    "inSndfile  is a SNDF soundfile for input\n");
    fprintf(stderr,
	    "outFTfile  is a PVOC-type file, default is inStem+'%s'\n",
	    OP_FEXT);
    exit(1);
    }
