/***************************************************************\
*	ivbuf.c  (after resambuf.c)				*
* Buffer objects for cq inversion				*
* dpwe 23feb91							*
\***************************************************************/

#ifdef __STDC__
#include <stdlib.h>
#else
#include <unistd.h>
#endif
#include <math.h>
#include <string.h>

#include "ivbuf.h"

/* static function prototypes */
#ifdef __STDC__
#else
#endif

/* implementations */

IvRSBufType *NewIvRSBuf(bufSiz, rsBufSiz, fill, chSiz, chFn, chArg, rdp, 
			status, desc)
    /* Instance a new inversion buffer with two routes in. */
    long 	bufSiz;
    long 	rsBufSiz;
    long 	fill;
    long	chSiz;		/* blocking for passing complete data */
    long	(*chFn)();	/* function to pass complete data to */
    void	*chArg;
    RsmpDescType *rdp;
    long	status;
    char 	*desc;
    {
    int  inPd;
    int  outPd;
    int  khLen;
    IvRSBufType *irbp;

/*    fprintf(stderr, 
	    "newIVB: siz %ld, rsiz %ld, fill %ld, csiz %ld, carg %lx, d %s\n",
	    bufSiz, rsBufSiz, fill, chSiz, chArg, desc);
/* */
    irbp = (IvRSBufType *)malloc(sizeof(IvRSBufType));
    irbp->rsmpBufPtr  = NewResampBuf(rsBufSiz, 0L, rdp, desc);
    irbp->buf = (SAMPL *)malloc(sizeof(SAMPL)*bufSiz);
    irbp->bufSize = bufSiz;
    ClearSampls(irbp->buf, 1, irbp->bufSize);
    irbp->firstUnsent = (fill<0)? -fill : 0;
    irbp->firstUntouched = (fill<0)? -fill : fill;
    irbp->firstUnstraight = (fill<0)? 0 : fill;
    irbp->firstUnresampled = (fill<0)? -fill : 0;
    /* these will sink <-fill> pts before starting to send any on */
    irbp->chainSize = chSiz;
    irbp->chainFn   = chFn;
    irbp->chainArg  = chArg;
    irbp->status    = status;
    if(desc)
	strncpy(irbp->desc, desc, DESCLEN);
    return irbp;
    }
    

void FreeIvRSBuf(irBp)
    IvRSBufType *irBp;
    {
    FreeResampBuf(irBp->rsmpBufPtr);
    free(irBp->buf);
    free(irBp);
    }

long StraightInIvRSBuf(irBp, sptr, siz, hopSize, flush)
    /* Just write data into new buffer */
    IvRSBufType *irBp;
    SAMPL 	*sptr;
    long 	siz;
    long 	hopSize;	/* how far next buf will be ahead */
    int		flush;		/* flush tolerates resamplebuf overflow */
    {
    long	base = irBp->firstUnstraight;
    long	top  = irBp->firstUntouched;

/*    fprintf(stderr,"StrtIn(%s): size %ld\n", irBp->desc, siz);	*/
/*    fprintf(stderr,"[%s] ", irBp->desc);	fflush(stderr);	     	*/
    if(base + siz > irBp->bufSize)
	{  fprintf(stderr, "StrInIVB: overflow %ld on %ld, siz %ld (%s)\n",
		   siz, base, irBp->bufSize, irBp->desc);
	   abort();	}
    OverlaySampls(sptr, 1, irBp->buf+base, 1, siz, (LONGSAMPL)SAMPLUNITY);
    irBp->firstUnstraight = base+hopSize;	/* where to start next one */
    if(base+siz > top)
	irBp->firstUntouched = base+siz;	/* new HWM in buffer */
    UpdateIvRSBuf(irBp,flush);

    return siz;
    }

long ResampleInIvRSBuf(irBp, sptr, siz, flush)
    /* resample new data into buffer, pass to chain fn, empty with emptFn */
    IvRSBufType *irBp;
    SAMPL 	*sptr;
    long 	siz;
    int 	flush;
    {
    long	base = irBp->firstUnresampled;
    long	top  = irBp->firstUntouched;
    long	unsent_bot, unsent_top, RSsiz;
    RsmpBufType *rsBp        = irBp->rsmpBufPtr;

    ScaleSampls(sptr, 1, sptr, 1, siz, (LONGSAMPL)(sqrt(2)*.5*SAMPLUNITY));
    /* hmmm... */
    unsent_bot = RBSize(rsBp,flush);
	/* chainSize*(RBSize(rsBp,flush)/chainSize); /* sent up to here */
    AddToResampBuf(rsBp, sptr, siz, flush);
    unsent_top = RBSize(rsBp,flush);	RSsiz = unsent_top-unsent_bot;
/*    fprintf(stderr,"RsmpIn(%s): size %ld became %ld\n", 
	    irBp->desc, siz, RSsiz);	*/
    if(base + RSsiz > irBp->bufSize)
	{  fprintf(stderr, "StrInIVB(%s): overflow %ld on %ld, siz %ld\n",
		   irBp->desc, RSsiz, base, irBp->bufSize);
	   abort();	}
    OverlaySampls(RBBufPtr(rsBp, unsent_bot, RSsiz), 1,
		  irBp->buf+base, 1, RSsiz, (LONGSAMPL)SAMPLUNITY);
    DropFromResampBuf(rsBp, RSsiz);		/* used them now */
    irBp->firstUnresampled = base+RSsiz;
    if(base+RSsiz > top)
	irBp->firstUntouched = base+RSsiz;	/* new HWM in buffer */
    UpdateIvRSBuf(irBp,flush);

    return siz;		/* number accepted, not number inserted */
    }

void DropFromIvRSBuf(irBp, dropSize)
    /* remove first <size> pts from buffer, update my ptrs */
    IvRSBufType *irBp;
    long 	dropSize;
    {
    long 	sentSize, straightSize, resampledSize;
    long	top   = irBp->firstUntouched;
    long	keep;
    SAMPL	*base = irBp->buf;

/*    fprintf(stderr,"Drop(%s): drop %ld of %ld (sent %ld)\n", 
	    irBp->desc, dropSize, top, irBp->firstUnsent);
 */
    straightSize  = irBp->firstUnstraight;
    resampledSize = irBp->firstUnresampled;
    if(top < dropSize)
	dropSize = top;
    keep = top-dropSize;
    CopySampls(base+dropSize, 1, base, 1, keep, REV);
    ClearSampls(base + keep, 1, dropSize);
    
    sentSize = irBp->firstUnsent;
    sentSize -= dropSize;
    if(sentSize < 0)		sentSize = 0;
    irBp->firstUnsent = sentSize;
    straightSize -= dropSize;
    if(straightSize < 0)	straightSize = 0;
    irBp->firstUnstraight = straightSize;
    resampledSize -= dropSize;
    if(resampledSize < 0)	resampledSize = 0;
    irBp->firstUnresampled = resampledSize;
    top -= dropSize;
    if(top < 0)			top  = 0;
    irBp->firstUntouched = top;
    }

long IVValid(irBp)	/* return size of valid points in IvBuf */
    IvRSBufType *irBp;
    {
    long straightSize, resampledSize;
    long top;

    straightSize  = irBp->firstUnstraight;
    resampledSize = irBp->firstUnresampled;
    top		  = irBp->firstUntouched;

    if((irBp->status&IVST_RSIN) && resampledSize < top)	
	top = resampledSize;		/* we must wait for resamp in */
    if(straightSize < top)
	top = straightSize;

    return top;
    }

void UpdateIvRSBuf(ivBp,flush)
    IvRSBufType *ivBp;
    int flush;
    {
    long top = IVValid(ivBp);
    long chainSize = ivBp->chainSize;
    long unsent    = ivBp->firstUnsent;
    long sent, taken;
    
/*    fprintf(stderr,"Update(%s): top %ld\n", ivBp->desc, top);	*/
    if(ivBp->chainFn == NULL)	return;
    while((top-unsent)>chainSize)
	{
	sent  = chainSize * ((top-unsent)/chainSize);
	taken = (*(ivBp->chainFn))(ivBp->chainArg, ivBp->buf+unsent, 
				   sent, flush);
	DropFromIvRSBuf(ivBp, unsent+taken);
	top = IVValid(ivBp);
	unsent = ivBp->firstUnsent;
	}
    }
