/***************************************************************\
*	ivbuf.h  (after resmp.c)				*
* Buffer object for cq inversion				*
* dpwe 23feb91							*
\***************************************************************/

#include "samplops.h"
#include "resambuf.h"

#define DESCLEN 16

typedef struct
    {
    RsmpBufType	*rsmpBufPtr;	/* target resampling buffer */
    SAMPL	*buf;
    long	bufSize;
    long	firstUntouched;
    long	firstUnstraight;
    long	firstUnresampled;
    long	firstUnsent;
    long	status;		/* word of flags */
    long	chainSize;	/* block size for chaining function */
    long	(*chainFn)();	/* func to call with new data */
    void 	*chainArg;	/* argument for chained function */
    char	desc[DESCLEN];	/* user space for label for object */
    }	IvRSBufType;

#define IVST_RSIN	32	/* this bit indicates wait for resampl in */

/* the empty and chaining functions are called whenever enough pts are held as
   long r = (*fn)(long arg, SAMPL *buf, long size, int flushFlag);
   where r is returned as the number of points actually consumed,
   and size is guaranteed to be a multiple of the relevent block size */

#ifdef __STDC__
IvRSBufType *NewIvRSBuf(long bufSize, long rsBufSize, long preFill,
			long chSiz, long (*chFn)(), void *chArg,
			RsmpDescType *rdp, long status, char *desc);
    /* Instance a new inversion buffer with two routes in. */

void FreeIvRSBuf(IvRSBufType *trBp);

long StraightInIvRSBuf(IvRSBufType *trBp, SAMPL *sptr, long siz, long hop,
		       int flush);
    /* resample new data into buffer, pass to chain fn, empty with emptFn */

long ResampleInIvRSBuf(IvRSBufType *trBp, SAMPL *sptr, long siz, int flush);
    /* resample new data into buffer, pass to chain fn, empty with emptFn */

void DropFromIvRSBuf(IvRSBufType *trBp, long size);
    /* remove first <size> pts from buffer, update my ptrs */

long IVTop(IvRSBufType *irBp);	/* return size of valid points in IvBuf */

void UpdateIvRSBuf(IvRSBufType *irBp, int flush);

#else /* no prototype args */

IvRSBufType	*NewIvRSBuf();
void		FreeIvRSBuf();
long	        StraightInIvRSBuf();
long	        ResampleInIvRSBuf();
void 		DropFromIvRSBuf();
long 		IVTop();
void 		UpdateIvRSBuf();

#endif
