/*******************************************************************\
*   DSPbench.h							    *
*								    *
*   header file for DSPBench					    *
*   After MiniEdit.						    *
\*******************************************************************/

#include <sndf.h>
#include <genutils.h> 	/* Mymalloc() */
#include "macalike.h"

#ifndef THINK_C		/* (no fancy stuff on mac) */
/* #define CQFILE  1	/* CQFILE set means spectral files are multirate */
/* #define TRAX    1	/* TRAX enables spectra-by-track synthesis */
/* set CQFILE and TRAX in makefile if you like */
#endif

/* #undef CQFILE /*!!!!*/
#ifdef CQFILE
#include <readaqt.h>
#else
/* #include <pvoc.h> */
#include <readaft.h>
#endif

#ifdef TRAX
#include <llists.h>
#endif

#ifndef FLOATARG
#define FLOATARG double
#endif	/* for floating-point arguments to fns - ansi is fussy */

extern char *programName;

#ifdef THINK_C

#define TEwindowID 		128
#define	TGwindowID		129
#define ErrorAlert		256
#define	AdviseAlert		257

/* resource IDs of menus */
#define appleID			128
#define fileID			129
#define editID			130
#define viewID			131

/* Edit menu command indices */
#define undoCommand 	1
#define cutCommand		3
#define copyCommand		4
#define pasteCommand	5
#define clearCommand	6

/* Menu indices */
#define appleM			0
#define fileM			1
#define editM			2
#define	viewM			3
#define traxM			4

#define	vNormal			1
#define vWaterf			2
#define	vFourier		3
#define vOneFrm			4
#define vFourFrm		5
#define vPvoc			6
#define vPsyAc			7

#define fmNew		1
#define fmOpen		2
#define fmClose		4
#define fmSave		5
#define fmSaveAs	6
#define fmRevert	7
#define fmPageSetUp	9
#define fmPrint		10
#define fmQuit		12

#define aaSave		1
#define aaDiscard	2
#define aaCancel	3

#define mAbout 		1		/* fake for linking? */

#else	/* not THINK_C */
#define fileM			13
#define editM			9
#define	viewM			3
#define traxM			21
#define groupM			32

#define fmQuit			15
#define mAbout			23
#define	vNormal			1	/* instead of mac */
#define vWaterf			2
#define vPvoc			14
#define vOneFrm			28
#define	vFourier		16
#define vFourFrm		17
#define vPsyAc			29
#define vPhase			41
#ifdef TRAX
#define tTrack			20
#define tReset			19
#define tSynth			18
#define tLoad			25
#define tSave			26
#define tPrint			27
#define tExspc			30
#define tSaveT			31
#define gGrow			33
#define gProm			34
#define gOnset			35
#define gHarm			36
#define gMeta			37
#define gNoise			38
#define gOther			39
#define gNois2			40
#endif	/* TRAX */
#endif	/* THINK_C */

#define SBarWidth   15

#ifndef NULL
#define NULL 0L
#endif

#define true -1
#define false 0

/* colour codes for SetColour() */
#define NUM_COLOURS	8
#define BLACK	0
#define WHITE	1
#define RED	2
#define GREEN	3
#define BLUE	4
#define CYAN	5
#define MAGENTA	6
#define YELLOW	7

typedef struct recRcd **RECRHAN;
typedef struct recRcd			/* for linked lists of rectangles */
    {
    Rect    rec;			/* rectangle of this one */
    int     dirty;			/* do we need to erase the rect ? */
    int     id;				/* function id of this region */
    int     sbits;			/* some flag bits */
    RECRHAN sib;			/* pointer to next sibling rectangle (or NULL) */
    RECRHAN child;			/* pointer to child rectangles (or NULL) */
    } RECRCD;

/* bits for .sbits */
#define SBNONE	0x00
#define SBLAST	0x01	/* NO!! -\->indicate */
#define SBDRAW	0x02	/* Draw command for this node */
#define SBCLIK	0x04	/* Click-in command for this node */
#define SBDNC	0x06	/* both draw and clickable */

enum{					/* codes for id in recRcd */
    Nullid,	/* unused .. */
    Wwinid,	/* the whole window */			/*  1 */
    Infboxid,	/* the info pane */
    Cboxid,	/* control button pane */
    Cplayid,	/* play button in ctl pane */
    Crecid,	/* record button in ctl pane */		/*  5 */
    Cquitid,	/* quit button in ctl pane */
    Lviewid,	/* long view pane */
    Sviewid,	/* short view pane */
    Lvtimlid,	/* timeline in longview */
    Lvsndid,	/* soundgraph in longview */		/* 10 */
    Lvsbarid,	/* scrollbar in longview */
    Lvvscid,	/* Vertical scale in long view */
    Svsndid,	/* soundgraph in shortview */
    Svtimlid,	/* timeline in shortview */
    VscaleUpid,	/* increase the vscale button */	/* 15 */
    VscaleDownid,	/* decrease vscale button */
    Svvscid,	/* Vertical scale in short view */
    MenuBarid,	/* menu bar at top of window */
    Shoutid,	/* shout hint line, at bottom of window */
    Cprintid,	/* new control button for printing */
    Csynthid 	/* and another, to synth tracks */
    };

#define dShort	2
#define dFloat	4		    /* types for tgWdata.dType */

#define unityScale 2	/* the default scale amount */

typedef struct sndBite		/* subsection of a sound */
    {
    SFSTRUCT	*snd;		/* the parent sound */
    long	bStart;		/* byte offset of start of sound */
    long	bEnd;		/* byte offset of end of sound */
    char	*loc;		/* pointer to actual sound in memory */
    } SNDBITE;

typedef struct frqView		/* looking at a part of a spectrum */
    {
#ifdef CQFILE
    CQTINFO	*afile;		/* the analysis file */
#else /* !CQFILE -> FFTFILE */
    PVSTRUCT	*afile;		/* the other analysis file */
#endif /* CQFILE */
    float	minFrq;		/* left end of frq axis */
    float	maxFrq;		/* right end of frq axis */
    float	minMag;		/* bottom edge of mag axis */
    float	maxMag;		/* top end of mag axis */
    float	samplingRate;	/* lose: NOT passed to ShowFFT any other way */
    float	*ftbuffer;	/* store result of fft here */
    long	ftsize;		/* remember its size */
    int 	valid;		/* flag if need to recalc */
    } FRQVIEW;

typedef struct sndWdata
    {
    SFSTRUCT	*snd;		/* the sound file header */
#ifdef CQFILE
    CQTINFO	*cqi;		/* extended header for multirate files */
#else
    PVSTRUCT	*pvs;		/* fft analysis data if avai */
#endif
#ifdef TRAX
    LIST	trax;		/* root of track list */
#endif
    SNDBITE	*view;		/* subsect of sound is view */
    SNDBITE	*slct;		/* subsect of sound is selected region */
    FRQVIEW	*frqV;		/* view in spectral modes */
    RECRHAN	recTreeH;	/* where on the screen this all goes */
    PIXMAP	pixmap;		/* cache for grayscale images */
    int 	pixW;		/* size of pixmap */
    int 	pixH;
    int 	vMode;		/* view mode - 2 channel etc */
    int 	scale;		/* scaling factor */
    int 	wflen;		/* baseline length for waterfall */
    } SNDWDATA;


/* typedef struct tgWdata
/*     {
/*     Handle	soundHan;	    /* handle to sound block */
/*     long	sfBLen;		    /* BYTES in soundHan block */
/*     long	vBstt;
/*     long	vBend;		    /* subview start and end in bytes */
/*     int	vMode;		    /* view mode */
/*     long	sBstt;
/*     long	sBend;		    /* select rgn start and end in bytes */
/*     double	sRate;		    /* sample rate */
/*     int	dType;		    /* sample data type */
/*     int	chans;		    /* number of channels */
/*     int	scale;		    /* vertical scale factor; 2= unity */
/*     int 	wflen;		    /* number of points in wtrfall baseline */
/*     PVSTRUCT	pvs;		    /* pointer to PV analysis stuff */
/*     RECRHAN	recTreeH;	    /* handle to tree of rect areas */
/*     };
 */

enum{		/* for view mode word - mode of zoomed view */
    noVM,
    monoVM,	/* mono sound - one channel  */
    leftVM,	/* left of stereo */
    rightVM,	/* right of stereo */
    oneVM,
    twoVM,
    threeVM,
    fourVM,	/* indiv channels of quad */
    multiVM,	/* each channel has its own trace */
    wfallVM,	/* view as waterfall stacked segments */
    specVM,	/* a spectral view mode */
    ftVM,	/* fourier transform view mode */
    onefrVM,	/* show one frame from the data file */
    spgrmVM,	/* spectrogram view mode */
    psyDispVM	/* psycho acoustic display mode (new style) */
    };

typedef struct winRcd **WINRHAN;

#ifdef THINK_C
typedef struct teWdata
	{
	TEHandle	TEH;
	int 		linesInFolder;
	ControlHandle	vScroll;
	};								/* Data fields for TE window */
#endif

typedef struct winRcd
    {
    WINRHAN	    nextHan;
    WindowPtr	myWindow;
    Str255	    theFileName;
    int		    theVRefNum;
    Rect	    dragRect;
    char	    dirty;
    int		    winType;	    /* type of window: text, graphic etc */
    union
	{
/*	struct	tgWdata     tg;		*/
	struct  sndWdata    snd;
#ifdef THINK_C
	struct	teWdata		te;
#endif
	}	    w;		    /* window specific data */
    } WINRCD;

/* So to access the TEH, use (*winRHan)->w.te.TEH   */

/*  codes for winType  */
enum{
    nullWin,
    TEWin,
    TGraphWin,
    FGraphWin
    };


/***** EXTERNAL VARIABLES DECLARATION ******/
/*  extern  MenuHandle	*myMenus; */
extern	WINRCD	    **winRoot;
extern	WINRCD	    **frntWin;
extern	WINRCD	    **dbugWin;
#ifdef THINK_C
extern  unsigned long colrs[];
#else
extern  unsigned long *colrs;	/* pixel grey scale */
#endif
extern  int         grays;	/* entries in colrs */

#define lSide	1	/* defines for passing sides to gutter (GutterR in db.grutl.c) */
#define rSide	2
#define tSide	4
#define bSide	8

#define WFLENMAX 2047

/* prototypes.. */
#ifdef __STDC__
WINRCD **MyNewWindow(WINRCD **, Str255, int);
WINRCD **ours(WindowPtr);
RECRCD **FindRhan(RECRCD **,int);   /* search tree for pane id */
double	Duration(SFSTRUCT *);	    /* returns float of duration of sound */
double	FloatProp(FLOATARG,long,long);
long	LongProp(long,FLOATARG,FLOATARG,int);
short	MapTtoX(FLOATARG, Rect, FLOATARG, FLOATARG);	/* convert a time to an X co-o */
double	MapXtoT(int, Rect, FLOATARG, FLOATARG);	/* converse of above */
double	MapItoF(int, int, int, FLOATARG, FLOATARG);
int	MapFtoI(FLOATARG, FLOATARG, FLOATARG, int, int);
FRQVIEW *NewFrqView(FLOATARG srate);
#else
WINRCD **MyNewWindow();
WINRCD **ours();
RECRCD **FindRhan();
double	Duration();
double	FloatProp();
long	LongProp();
short	MapTtoX();
double	MapXtoT();
double	MapItoF();
int	MapFtoI();
FRQVIEW *NewFrqView();
#endif




