/*******************************************************************\
*   db.d.ft.c		       					    *
*   Show fourier transform of signal				    *
*   30apr92 dpwe	       				   	    *
\*******************************************************************/

#include "DSPbench.h"
#include <samplops.h>
#include <vecops.h>

#define MAXFFTSIZE 65536

void ShowFT(rec, sndb, frqv)
    Rect	*rec;
    SNDBITE	*sndb;
    FRQVIEW	*frqv;
    {
    if(!frqv->valid || frqv->imgBuf == NULL)
	CalcFFTimg(frqv, sndb);
    ReshowFT(rec, frqV);
    }

void CalcFFTimg(frqv, sndb)
    FRQVIEW	*frqv;
    SNDBITE	*sndb;
    {
    long 	frames;		/* samples in region we were passed */
    long 	t_points;	/* size of the transform (2^n) */
    float	*tmpbuf;	/* where to do all the calc */

    if(frqv->imgBuf != NULL)
	Myfree(frqv->ftbuffer);
    /* need to store float array of whole fft */
    frames = SBframeCount(sndb);
    if(frames > MAXFFTSIZE)	/* don't do *ridiculous* transforms.. */
	frames = MAXFFTSIZE
    t_points = EnclPwrOf2(frames);
    tmpbuf = Mymalloc(2*t_points, sizeof(float), "CalcFFTimg");
    /* 2*t_points because it's a complex thing */
    ClearFloats(tmpbuf, 1, 2*t_points);
    /* copy out in 'packed' format - i.e. floats not complex */
    Short2Float(SBstart(sndb), tmpbuf, SBchans(sndb), 1, frames, FWD);
    LogFT(tmpbuf, frames, t_points);
    frqv->ftbuffer = tmpbuf;
    frqv->ftsize   = t_points;
    frqv->valid = 1;
    }

/* calculate the FFT of a chunk of floats by windowing, rotating, 
   padding in the middle to the next largest power of two, FFT,
   cobnvert result to polar co-os and convert mags to dB */
/* returns newly Mymalloc'd buffer */
float *LogFT(buf, samps, size)	
    float	*buf;		/* real input data - not touched */
    long 	samps;		/* samples */
    long	size;		/* buffer size = fft size */
    {
    float 	*winBuf;
    float 	*rslt;

    winBuf  = MakeHalfWin(samps, (float)1.0/SHORTMAX, 1);
    ApplyHalfWin(buf, winBuf, smps);
    HalfRotate(buf, smps);		/* make ends in middle .. */
    PadInMiddle(buf, smps, size);	/* insert zeros in center */ 
    UnpackReals(buf, size);
    fft((complex *)buf, size);
    Rect2Polar(buf, size);
    /* condition magnitude */
    VLog(buf, 2, buf, 2, 1.0, size);
    VSMul(tmpBuf, 2, 8.6859, frameSize);	/* now in dB */
    /* absolute peak now always at .9/2.0 * 100 dB = +95dB */
    /* Condition phase as group delay (1st order diff of unwrapped phase) */
    UnwrapPhases(tmpBuf+1, 2, tmpBuf+1, 2, frameSize, (FLOATARG)0.0 );	
    VFODiff(tmpBuf+1, 2, tmpBuf+1, 2, frameSize);  /* actually d(phase)/df */
    VSMul(tmpBuf+1 ,2, 1.0/6, frameSize);   /* 1/6 to get +/- pi in middle */
    }


void ReshowFT(rec, frqV)
    Rect	*rec;
    FRQVIEW	*frqv;
    {
    Rect	inRec;
    double	binRange, frqGap, minMag, magRng;
    long 	loBin, hiBin, i;
    Point	pt;
    int 	first;

    inRec = *rec;
    /* offset rects for scales ... */

    /* figure which bins we display given frq range */
    binRange = frqv->samplingRate/2.0;
    frqGap = binRange / (frqv->ftsize/2);
    loBin  = frqv->minFreq/frqGap;
    hiBin  = frqv->maxFreq/frqGap;
    minMag = frqv->minMag;
    magRng = frqv->maxMag - frqv->minMag;
    first  = 1;
    for(i=loBin; i<=hiBin; ++i)
	{
	pt.h = inRec.left + (inRec.right - inRec.left)
	      * (i - loBin)/(hiBin - loBin);
	pt.v = inRec.bottom - (inRec.top - inRec.bottom)
	      * (*(frqv->buffer+2*i) - minMag)/magRng;
	Clip2D(&inRec, &pt);
	if(first)	MoveTo(pt.h, pt.v);
	else	       	LineTo(pt.h, pt.v);
	first = 0;
	}
    }
