/*******************************************************************\
*   db.dsp.c							    *
*   freq domain drawing functions for DSPbench			    *
*   30sep90 dpwe						    *
\*******************************************************************/

#include <stdio.h>
#include <string.h>
#include <dsputil.h>
#include <fft.h>
#include <vecops.h>
#include <genutils.h>
#include <samplops.h>

#include <math.h>
#include "DSPbench.h"

#define MAXFFTSIZE 65536

/* static globals */
static complex *basis;	    /* LUTable for FFT */

#define SHORTMAX 32767

FixAndCopy(sceF, sinc, dstS, dinc, max, pts)
    float *sceF;
    short *dstS;
    long pts;
    FLOATARG max;
    {
    double   factor,f;
    short   s;

    factor = ((double)SHORTMAX)/max;
    while(pts--)
	{
	f = factor*(*sceF++);
	if(f>SHORTMAX)	s = SHORTMAX;
	else if(f<-SHORTMAX)	s = -SHORTMAX;
	else s = (short)f;
	*dstS++ = s;
	}
    }

ShowFFT(rec,inPtr,endPtr,chs,mode,scale,frqv)
    Rect    rec;
    short   *inPtr;
    short   *endPtr;
    int     chs;	/* only 1st channel is converted */
    int     mode;
    int     scale;
    FRQVIEW *frqv;
    {
    float   *tmpBuf,*tmpBuf2,*winBuf;
    double   max;
    short   *tmpSBuf;
    long    frameSize = 1;
    long    smps, i;
    float   *fp;
    long 	six, eix;

    smps = (endPtr-inPtr)/chs;
    while(frameSize < smps && frameSize < MAXFFTSIZE)
	frameSize <<= 1L;
    if(frameSize < smps)	smps = frameSize;
    if(frqv->valid == 0 || frqv->ftbuffer == NULL)
	{
	if(frqv->ftbuffer != NULL)	Myfree((char *)frqv->ftbuffer);
	tmpBuf  = MakeBuf(frameSize * 2);
	frqv->ftbuffer = tmpBuf;
	frqv->ftsize = frameSize * 2;
	basis = AssignBasis(NULL,frameSize); /* set up FFT tables */
    	winBuf  = MakeHalfWin(smps /*frameSize*/,(double)1.0/SHORTMAX,1);
    	Short2Float(inPtr, chs, tmpBuf, 1, smps); /* cast to Mono too */
    	ApplyHalfWin(tmpBuf,winBuf,smps	/*frameSize*/);
    	HalfRotate(tmpBuf,smps	/*frameSize*/);
    	PadInMiddle(tmpBuf,smps, frameSize); 
    	UnpackReals(tmpBuf, frameSize);
    	FFT2((complex *)tmpBuf, frameSize, 1,basis); /* was FFT2real */
    	Rect2Polar(tmpBuf, frameSize);
    	/* condition magnitude */
    	VLog(tmpBuf, 2, tmpBuf, 2, .3/(double)scale, frameSize);
	/*    	VSMul(tmpBuf, 2, ((double)scale)/20.0, frameSize); */
    	VSMul(tmpBuf, 2, 8.6859, frameSize); /* now in dB */
    	VSMul(tmpBuf, 2, 1/(double)50, frameSize); /* now fsd is 100 dB */
    	max = VMax(tmpBuf, 2, frameSize);
    	VSAdd(tmpBuf, 2, -max+.9, frameSize);
	/* absolute peak now always at .9/2.0 * 100 dB = +95dB */
	/* Condition phase as group delay (1st order dif of unwrapped phase) */
	/* first zero phase of bins < 50 dB */
	for(i = 0; i<frameSize; ++i)  {
	    if(tmpBuf[2*i] < 0)		tmpBuf[2*i+1] = 0;
	}
/*    	UnwrapPhases(tmpBuf+1, 2, tmpBuf+1, 2, frameSize, (FLOATARG)0.0 );  */
/*    	VFODiff(tmpBuf+1, 2, tmpBuf+1, 2, frameSize); /* actually d(phs)/df */
    	VSMul(tmpBuf+1 ,2, 1.0/3.5, frameSize); /* 1/3.5 to get +/- pi in middle */
    	if(chs == 2)		/* stereo : show both L&R magnitudes */
	    {
	    tmpBuf2 = MakeBuf(frameSize * 2);
	    Short2Float(inPtr+1, chs, tmpBuf2, 1, smps);
	    ApplyHalfWin(tmpBuf2, winBuf, smps);
	    HalfRotate(tmpBuf2, smps);
	    PadInMiddle(tmpBuf2, smps, frameSize);
	    UnpackReals(tmpBuf2, frameSize);
	    FFT2((complex *)tmpBuf2, frameSize, 1, basis);
	    Rect2Polar(tmpBuf2, frameSize);
	    VLog(tmpBuf2, 2, tmpBuf2, 2, .3/(double)scale, frameSize);
	    VSMul(tmpBuf2, 2, 8.6859, frameSize); /* now in dB */
	    VSMul(tmpBuf2, 2, 1/(double)50, frameSize); /* now fsd is 100 dB */
	    VSAdd(tmpBuf2, 2, -max+.9, frameSize);	
	    CopyFloats(tmpBuf2, 2, tmpBuf+1, 2, frameSize, FWD);
	    free(tmpBuf2);
	    }
	frqv->valid = 1;
	free(winBuf);
	}
    tmpBuf = frqv->ftbuffer;
    tmpSBuf = (short *)Mymalloc(frameSize, sizeof(short), "ShowFFT");
    FixAndCopy(tmpBuf, 1, tmpSBuf, 1, (FLOATARG)1, frameSize);
    six = frqv->minFrq/frqv->samplingRate*frameSize;
    eix = frqv->maxFrq/frqv->samplingRate*frameSize;
    /* ShowSound(rec,tmpSBuf,tmpSBuf+frameSize,2,multiVM,2,SFMT_SHORT,NULL); */
    ShowSound(rec,tmpSBuf+2*six,tmpSBuf+2*eix,2,multiVM,2, SFMT_SHORT,NULL);
    Myfree((char *)tmpSBuf);
    }

void ShowOneFrame(wrcd,rec)
    WINRCD  **wrcd;
    Rect    rec;
    {
#ifdef CQFILE
    CQTINFO	*cqi;
#else
    PVSTRUCT	*pvh;
#endif
    long	sStt, sEnd;	/* sample frame indices */
    long 	sfSiz, frInc;
    Rect    	inrect;
    long    	i, frame, flPerFrame;
    int 	scale,err=0;
    long   	len;
    /*    RgnHandle    dummy;	*/
static float 	*specBuf = NULL;
    int 	framesToDisp;
    double	time;

#ifdef CQFILE
    cqi  = (*wrcd)->w.snd.cqi;
    if(cqi == NULL)
#else
    pvh  = (*wrcd)->w.snd.pvs;
    if(pvh == NULL)
#endif
	{
	MarkX(&rec);
	return;
	}

    if(specBuf == NULL)
#ifdef CQFILE
	if((specBuf = (float *)Mymalloc((long)cqi->nBins*cqi->nOcts, 
				      sizeof(cpxpolar), "Show1fr1"))==NULL)
#else
	if((specBuf = (float *)Mymalloc(pvh->frameBsize, 1L,
							"Show1fr2"))==NULL)
#endif
	    err = 1;
    if(err)
	{
	fprintf(stderr,"%s:Show1Fr: malloc fail\n",programName);
	exit(1);
	}
    scale = (*wrcd)->w.snd.scale;
    sfSiz= Ssize(((*wrcd)->w.snd.view)->snd);
    sStt = ((*wrcd)->w.snd.view)->bStart / (long)sfSiz;
    sEnd = ((*wrcd)->w.snd.view)->bEnd / (long)sfSiz;
#ifdef CQFILE
    len  = cqi->nBins * cqi->nOcts;
    time = sStt / ((*wrcd)->w.snd.view)->snd->samplingRate;
#else
    frInc = pvh->frameIncr;
/*    frSiz = pvh->frameSize;	*/
    flPerFrame = pvh->frameBsize / sizeof(float);
    frame = sStt / frInc;
    len   = pvh->frameBsize/(2*sizeof(float));	/* #of mag points */
#endif
    /*    PushClip(&dummy, rec);	*/
    /*    printf("Frame size %ld incr %ld 1st %ld last %ld\n",
	  frSiz, frInc, firstFrame, lastFrame);	*/
    /* reset static list of past peaks to match */
#ifdef CQFILE
    SpecOfCQTFile(cqi, time, (double)0.001, (cpxpolar *)specBuf);
#else
    CopyFloats(((float *)PVDataLoc(pvh))+frame*flPerFrame, 1, specBuf, 1, 
	       flPerFrame, FWD);
#endif
    ShowMagPhase(specBuf, len, &rec, scale);
    /*    PopClip(&dummy);	*/
    }

ShowMagPhase(specBuf, len, rec, scale)
    float *specBuf;	/* this is overwritten */
    long  len;
    Rect  *rec;
    int   scale;
    {
    double max;
    short *tmpSBuf = (short *)specBuf;

    /* condition magnitude */
    VLog(specBuf, 2, specBuf, 2, .3/(double)scale, len);
/*    VSMul(specBuf, 2, ((double)scale)/20.0, len); */
    VSMul(specBuf, 2, 8.6859, len);		/* now in dB */
    VSMul(specBuf, 2, 1/(double)50, len);	/* now fsd is 2x50 = 100 dB */
    max = VMax(specBuf, 2, len);
    VSAdd(specBuf, 2, -max+.9, len);	
    /* absolute peak now always at .9/2.0 * 100 dB = +95dB */
    /* Condition phase as group delay (1st order diff of unwrapped phase) */
    UnwrapPhases(specBuf+1, 2, specBuf+1, 2, len, (FLOATARG)0.0 );	
/*    VFODiff(specBuf+1, 2, specBuf+1, 2, len);	*/
    VSMul(specBuf+1, 2, 1.0/6, len);	/* 1/6 to get +/- pi in middle */
    FixAndCopy(specBuf, 1, tmpSBuf, 1, (FLOATARG)1, 2*len);
    ShowSound(*rec, tmpSBuf, tmpSBuf+2*len, 2, multiVM, 2, SFMT_SHORT, NULL);
    }

