/*******************************************************************\
*   db.exspec.c							    *
*   calculate expectations of spectra based on past tracks	    *
*   1992oct27 dpwe						    *
\*******************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* types ***********/

/***** static globals ******/
/* these now used by GrowTrax too */
TKLIST theTails = NULL;	/* remember the tails we found */
double tailsTime;		/* ..and the time they matched */

/* to keep track code happy */
       int graphs = 0;

/******** prototypes ********/

/******* implementations *****/

PKLIST exFindPkBeforeT(PKLIST pkl, FLOATARG time)
{	/* find the node in the track imm. prec specified time */
    if(pkl==NULL || pkl->peak->tim > time)
	return NULL;
    while(pkl && pkl->next)  {
	if(pkl->next->peak->tim > time)
	    return pkl;		/* pkl->peak->tim <= time, ..next.. is not */
	else
	    pkl = pkl->next;
    }
    return NULL;		/* got to end - all < time */
}

TKLIST origexTailsForTime(FLOATARG time, TKLIST allTrax)
{	/* make a list of all the track tails crossing that time.
	   Output is a LIST of (new) TRACK structs, which include pointers 
	   to PKLISTs which are the tails starting at or after the time. */
    TKLIST	remTrx;
    TKLIST	tailsRoot;
    TKLIST	*tailsTail;
    PKLIST	pkl;
    TKLIST	trackNode;
    TRACK	*origtrk, *newtrk;

    remTrx = allTrax;		/* search through all tracks */
    tailsRoot = NULL;
    tailsTail = &tailsRoot;

    /* trax in allTrax are sorted in start time order, so we 
       can stop looking when there are no tracks left, or the 
       next track *starts* after our trigger time */
    if(remTrx)  {
	origtrk = (TRACK *)CAR(remTrx);
	while(remTrx && origtrk->tktop->peak->tim <= time)  {
	    if(pkl = exFindPkBeforeT(origtrk->tktop, time))  {
		newtrk = tkNew(pkl);
		trackNode = new_node(NULL, (LIST)newtrk, T_TRACK);
		/* append_ptr is passed the address of a node pointer, 
		   and returns the address of the node pointer into which 
		   it just wrote the new node, which it found by cdring 
		   until it found a next containing NULL */
		tailsTail = append_ptr(tailsTail, trackNode);
		/* force BOTTOM (chased down by NewTRK) to pt to found pos */
		newtrk->tkbot = newtrk->tktop;
		/* fixup flags in case this track is unrendered */
		if(origtrk->flags < 0)
		    newtrk->flags = origtrk->flags;
		else {	/* may still in fact be undisplayable */
		    if(FindNonnegX(newtrk->tktop))
			newtrk->flags = 0;
		    else
			newtrk->flags = -1;	/* this track unrendered */
		}
	    }
	    remTrx = cdr(remTrx);
	    if(remTrx)
		origtrk = (TRACK *)CAR(remTrx);
	}
    }
    return tailsRoot;
}

TKLIST newexTailsForTime(FLOATARG time, TKLIST allTrax, TKLIST *rtnAffected)
{	/* make a list of all the track tails crossing that time.
	   Output is a LIST of (new) TRACK structs, which include pointers 
	   to PKLISTs which are the tails starting at or after the time.
	   Original version instanced new track structures.  This one 
	   also returns a list of the elements from allTrax concerned.
	 */
    TKLIST	remTrx;
    TKLIST	tailsRoot;
    TKLIST	*tailsTail;
    PKLIST	pkl;
    TKLIST	trackNode;
    TRACK	*origtrk, *newtrk;
    TKLIST	fromAll;
    TKLIST	*fromAllTail;

    remTrx = allTrax;		/* search through all tracks */
    tailsRoot = NULL;
    tailsTail = &tailsRoot;
    fromAll = NULL;
    fromAllTail = &fromAll;

    /* trax in allTrax are sorted in start time order, so we 
       can stop looking when there are no tracks left, or the 
       next track *starts* after our trigger time */
    if(remTrx)  {
	origtrk = (TRACK *)CAR(remTrx);
	while(remTrx && origtrk->tktop->peak->tim <= time)  {
	    if(pkl = exFindPkBeforeT(origtrk->tktop, time))  {
		/* this track is being added, so put it into 
		   our list of all the trax */
		fromAllTail = append_ptr(fromAllTail, 
					 new_node(NULL, (LIST)origtrk, 
						  T_TRACK));
		/* Construct a *new* trk, consisting only of the tail */
		newtrk = tkNew(pkl);
		trackNode = new_node(NULL, (LIST)newtrk, T_TRACK);
		/* append_ptr is passed the address of a node pointer, 
		   and returns the address of the node pointer into which 
		   it just wrote the new node, which it found by cdring 
		   until it found a next containing NULL */
		tailsTail = append_ptr(tailsTail, trackNode);
		/* force BOTTOM (chased down by NewTRK) to pt to found pos */
		newtrk->tkbot = newtrk->tktop;
		/* fixup flags in case this track is unrendered */
		if(origtrk->flags < 0)
		    newtrk->flags = origtrk->flags;
		else {	/* may still in fact be undisplayable */
		    if(FindNonnegX(newtrk->tktop))
			newtrk->flags = 0;
		    else
			newtrk->flags = -1;	/* this track unrendered */
		}
	    }
	    remTrx = cdr(remTrx);
	    if(remTrx)
		origtrk = (TRACK *)CAR(remTrx);
	}
    }
    if(rtnAffected != NULL)  {
	*rtnAffected = fromAll; 	/* tell caller which tracks affected */
    } else {	/* not going to use that list */
	free_list(fromAll);
    }
    return tailsRoot;
}

TKLIST exTailsForTime(FLOATARG time, TKLIST allTrax)
{	/* make a list of all the track tails crossing that time.
	   Output is a LIST of (new) TRACK structs, which include pointers 
	   to PKLISTs which are the tails starting at or after the time. */
    return newexTailsForTime(time, allTrax, NULL);
}

void exDrawTails(WINRCD **wrcd)
{	/* highlight the track tails we are considering on spectrogram */
    /* draw all tails in special color */
    RedrawTrax(theTails, TKM_SPECIAL);
}    

static float    *flubber = NULL;
static long 	flubsiz = 0;

void CopyAndDisp(float *src, int sinc, long npts, WINDAT *win, char *str, 
		 int unwrap, FLOATARG gain)
{
    if(gain == 0.0)	gain = 1.0;
    if(flubber == NULL || npts > flubsiz)  {
	if(flubber)	free(flubber);
	flubsiz = npts;
	flubber = TMMALLOC(float, flubsiz, "flubber");
    }
    assert(flubsiz >= npts);

    VCopy(src, sinc, flubber, 1, npts);
    VSMul(flubber, 1, gain, npts);
    if(unwrap)  {
	UnwrapPhases(flubber, 1, flubber, 1, npts, (FLOATARG)0.0);
    }
    if(str && strlen(str) > 0)
	SetLabel(win, str);	
    DoDisp(win, flubber, npts);
}

void DispComplex(cpxpolar *src, int sinc, long npts, 
		 WINDAT *mw, WINDAT *pw, char *str)
{
    CopyAndDisp((float *)src, 2*sinc, npts, mw, str, 0, 1.0);
    CopyAndDisp(((float *)src)+1, 2*sinc, npts, pw, str, 0, 1.0/PI);
}

void tklFree(TKLIST tracklist)
{	/* dispose of a list of track structures */
    free_homo_list(tracklist, (NodeFunc)tkFree);
}

extern PKLIST ExpectSpect(LIST curTrax, FLOATARG t0, cpxpolar *spec, 
			  int bins, int octs, FLOATARG topfrq, FLOATARG Q);

static cpxpolar *actSpec = NULL, *expSpec, *difSpec, *ex2Spec, *df2Spec;
static WINDAT	*wm1 = NULL, *wm2, *wm3, *wm4, *wp1, *wp2, *wp3, *wp4;

/* #define FAKE */	/* flag to run test mode */

void ExspecTrax(wrcd)
    WINRCD **wrcd;
{
    double 	time, dummy_end, dummy_dur, dt;
    CQTINFO 	*cqi = (*wrcd)->w.snd.cqi;
    int 	totBins;
    char 	mystr[32];
    PKLIST	errPks;
    PKLIST	prdPks;
    PeakPos	myPeak;
    PKNODE	myPkNode;

    if(cqi == NULL)	return;		/* no loaded CQI */
    fprintf(stderr, "FindTT:");		fflush(stderr);

    totBins  = cqi->nBins * cqi->nOcts;
    /******** init static stuff *****************/
    if(actSpec == NULL)  {	/* initialize buffers */
	actSpec = TMMALLOC(cpxpolar, totBins, "exspec act");
	expSpec = TMMALLOC(cpxpolar, totBins, "exspec exp");
	difSpec = TMMALLOC(cpxpolar, totBins, "exspec dif");
	ex2Spec = TMMALLOC(cpxpolar, totBins, "exspec ex2");
	df2Spec = TMMALLOC(cpxpolar, totBins, "exspec df2");
    }
    if(wm1 == NULL)	{	/* initialize windows */
	flubsiz = totBins;
	flubber = TMMALLOC(float, flubsiz, "exspec flb");
	wm1 = NewWin("act mag", 0);    wp1 = NewWin("act phs", 0);
	wm2 = NewWin("est mag", 0);    wp2 = NewWin("est phs", 0);
	wm3 = NewWin("dif mag", 0);    wp3 = NewWin("dif phs", 0);
	wm4 = NewWin("df2 mag", 0);    wp4 = NewWin("df2 phs", 0); /* 1 */
	wdLink(wm1, wp1);	wdLink(wp1, wm2);
	wdLink(wm2, wp2);	wdLink(wp2, wm3);
	wdLink(wm3, wp3);	wdLink(wp3, wm4);
	wdLink(wm4, wp4);
    }

    /************** find time of start of current selection ********/
    STimeLims(wrcd,&time,&dummy_end,&dummy_dur);
    fprintf(stderr, "time=%.6f; ", time);	fflush(stderr);

    /* if time has changed from previous seln, dispose of old */
    if(theTails && time != tailsTime)  {
	tklFree(theTails);
	theTails = NULL;
    }
    if(theTails == NULL)  {	/* try and find some new tails */
	theTails = exTailsForTime(time, allTrax);
	fprintf(stderr, "tails=%ld; ", top_size(theTails));
	tailsTime = time;
	exDrawTails(wrcd);	/* draw up tails if appropriate to mode */
    }
    dt = 0.00001; 	/* ??? */
    SpecOfCQTFile(cqi, time, dt, actSpec);
    prdPks = ExpectSpect(theTails, time, expSpec, cqi->nBins, cqi->nOcts, 
			 cqi->pvh->maxFreq, cqi->qFactor);	
#ifdef FAKE
    /* Fake for now */
    myPkNode.peak = &myPeak;	myPkNode.next = NULL;
    myPeak.frq = 1000.0;	myPeak.mag = 0.0 /* dB */;
    myPeak.phs = 0.0;
    PkListToSpec(&myPkNode, actSpec, cqi->nBins, cqi->nOcts, 
		 cqi->pvh->maxFreq, cqi->qFactor);
    prdPks = &myPkNode;
    myPeak.frq = 1001.0;	myPeak.mag = -1.0;
    myPeak.phs = PI/100.0;
#endif /* FAKE */
    PkListToSpec(prdPks, expSpec, cqi->nBins, cqi->nOcts, 
		 cqi->pvh->maxFreq, cqi->qFactor);
    CVCopy(actSpec, 1, difSpec, 1, totBins);
    CVVSub(expSpec, 1, difSpec, 1, totBins);

    errPks = (PKLIST)ListPeaks((float *)difSpec, 2, ((float *)difSpec)+1, 2, 
			       totBins, /* sign */ 1, time, cqi->nBins, 
			       cqi->pvh->minFreq, cqi->pvh->maxFreq, 
			       /* logflag */ 1);

    exMatchPksAndErrs(cqi, prdPks, errPks);
    PkListToSpec(prdPks, ex2Spec,  cqi->nBins, cqi->nOcts, 
			 cqi->pvh->maxFreq, cqi->qFactor);
    CVCopy(actSpec, 1, df2Spec, 1, totBins);
    CVVSub(ex2Spec, 1, df2Spec, 1, totBins);

    pklistFree(errPks);
    errPks = (PKLIST)ListPeaks((float *)difSpec, 2, ((float *)difSpec)+1, 2, 
			       totBins, /* sign */ 1, time, cqi->nBins, 
			       cqi->pvh->minFreq, cqi->pvh->maxFreq, 
			       /* logflag */ 1);

    exMatchPksAndErrs(cqi, prdPks, errPks);

    /* convert to log (?? how is it they are lin) */
    Vlin2DB((float *)actSpec, 2, (float *)actSpec, 2, (double)1, totBins);
    Vlin2DB((float *)expSpec, 2, (float *)expSpec, 2, (double)1, totBins);

    sprintf(mystr, "time %.6f", time);
    DispComplex(actSpec, 1, totBins, wm1, wp1, mystr);
    DispComplex(expSpec, 1, totBins, wm2, wp2, mystr);
    DispComplex(difSpec, 1, totBins, wm3, wp3, mystr);
    DispComplex(df2Spec, 1, totBins, wm4, wp4, mystr);

#ifndef FAKE	/* can't `free' fake peak 'coz it's a local */
    pklistFree(prdPks);
#endif /* !FAKE */
    pklistFree(errPks);

    fprintf(stderr, "\n");
}

void exHumpMagSlope(FLOATARG fc, FLOATARG linmag, FLOATARG Q, FLOATARG f, 
		    double *humpmag, double *humpslope)
{	/* return one value of our convolved filter hump
	   Hump is of mag <linmag>, with governed by <Q>, centre frq <fc>
	   Value of that hump at <f> returned in <humpmag>
	   slope (d(mag)/df) in <humpslope> */
    double	df, k;

    /* after ../trk/exspec.c, empirical width fixup .. */
    Q *= 1.19394;
    df = f-fc;
    k  = Q*Q/2.0/f/f;
    /* divide by frq of MEAS.PT, not frq of PEAK CTR */
    *humpmag = exp(-k*df*df)*linmag;
    /* also want to know d(humpmag)/d(frq) for frq error correction */
    *humpslope = *humpmag * (-2.0*k*df);
    /* when f>fc, df is > 0, humpslope is < 0 (hump returning to zero) */
}

#define LINOFDB(mag)	(exp((mag)/LOGN2DB))
#define DBOFLIN(mag)	(LOGN2DB*log(mag))

void exRecommendUpdates(PeakPos *estPk, FLOATARG Q, PeakPos *errPk,
			double *dfrq, double *magfact, double *dpha)
{	/* recommend changes to frq, mag and pha of est pk 
	   based on chrs of error peak */
    double	humpMag, humpSlope;
    double	errParr, errPerp;	/* parallel and normal cpt of err */
    double	angle, actMag;
    double 	errMag, errMidAngle;

    exHumpMagSlope(estPk->frq, LINOFDB(estPk->mag), Q, errPk->frq,
		   &humpMag, &humpSlope);
    errMag = LINOFDB(errPk->mag);

/*    *dpha = atan(errPerp/humpMag);	*/
    angle = estPk->phs + PI - errPk->phs;
    /* reconstruct actual spectrum at that point by cosine rule */
    actMag = sqrt(humpMag*humpMag + errMag*errMag 
		  - 2*humpMag*errMag*cos(angle));
    /* find angle between est and act by sine rule */
    *dpha = asin(errMag*sin(angle)/actMag);

    errMidAngle = estPk->phs + (*dpha)/2.0;
    errParr = errMag * cos(errPk->phs - errMidAngle);
    errPerp = errMag * sin(errPk->phs - errMidAngle);
    /* this will be wrong for each error peak in a pair, but it 
       should cancel out something like correctly */
    *magfact = errParr/humpMag;

    *dfrq = -errParr/(humpSlope * (1+*magfact));

}

#define ERR_PK_BIN_THRESH	(2.0)
/* error peaks must lie within this many bin-widths of an actual 
   peak to be counted */
#define MAIN_PK_DB_THRESH	(-5.0)
/* parent peak must be this large to even start looking for errors */

void exMatchErrPksToPk(PeakPos *peak, PKLIST errPks, int binsPoct, FLOATARG Q)
{	/* see which of the peaks of the error spectrum appear to 
	   lie close to the specified track */
    double logbinratio, fr;
    int   found = 0;
    double	dfrq, magfact, dpha;
    double	tdfrq = 0.0, tmagfact = 0.0, tdpha = 0.0;
    double	nfound = 0.0;

    logbinratio = log(2.0)/binsPoct;	/* log of frq ratio betwn adj bins */

    while(errPks)  {
	fr = log(errPks->peak->frq / peak->frq);
	if( fabs(fr) < ERR_PK_BIN_THRESH*logbinratio )  {
	    if(found == 0)  {
		fprintf(stderr, "Err pks found:\n");
		PrintPeak(peak, "Ori");
		found = 1;
	    }
	    PrintPeak(errPks->peak, "Err");
	    exRecommendUpdates(peak, Q, errPks->peak,
			       &dfrq, &magfact, &dpha);
/*	    fprintf(stderr, "Rec updates: df %f dm %f dpha %f\n",
		    dfrq, magfact, dpha);   */
	    tdfrq += dfrq;  tmagfact += magfact;  tdpha += dpha;
	    nfound += 1.0;
	}
	errPks = errPks->next;
    }
    if(nfound > 0.0)  {
	tdfrq /= nfound;	tmagfact /= nfound;   tdpha /= nfound;
	fprintf(stderr, "Chg(%.1f): frq %.2f mag %.2f%% pha %.6fpi\n", 
		nfound, tdfrq, 100.0*tmagfact, tdpha/PI);
	tmagfact += 1.0;	/* that is how it is defined; the proportion 
				   of the current peak that must be added */
	peak->mag = DBOFLIN(LINOFDB(peak->mag)*tmagfact); /* reduces */
	peak->frq += tdfrq;
	peak->phs += tdpha;
    }
}

void exMatchErrPksToTrk(TRACK *trk, PKLIST errPks, int binsPoct)
{	/* see which of the peaks of the error spectrum appear to 
	   lie close to the specified track */
    double logbinratio, fr;
    int   found = 0;

    logbinratio = log(2.0)/binsPoct;	/* log of frq ratio betwn adj bins */

    while(errPks)  {
	fr = log(errPks->peak->frq / trk->tkbot->peak->frq);
	if( fabs(fr) < ERR_PK_BIN_THRESH*logbinratio )  {
	    if(found == 0)  {
		fprintf(stderr, "Err pks found:\n");
		PrintPeak(trk->tkbot->peak, "Ori");
		found = 1;
	    }
	    PrintPeak(errPks->peak, "Err");
	}
	errPks = errPks->next;
    }
}

exMatchTrksAndErrs(CQTINFO *cqi, LIST theTails, PKLIST errPks)
{
    TRACK *trk;

    while(theTails)  {
	trk = (TRACK *)(theTails->data);
	if(trk->tkbot->peak->mag > MAIN_PK_DB_THRESH)  {
	    exMatchErrPksToTrk(trk, errPks, cqi->nBins);
	}
	theTails = theTails->next;
    }
}

exMatchPksAndErrs(CQTINFO *cqi, PKLIST prdPks, PKLIST errPks)
{
    while(prdPks)  {
	if(prdPks->peak->mag > MAIN_PK_DB_THRESH)  {
	    exMatchErrPksToPk(prdPks->peak, errPks, cqi->nBins, cqi->qFactor);
	}
	prdPks = prdPks->next;
    }
}

#else /* !CQFILE */   

void ExspecTrax(wrcd)	/* dummy for menu */
{
}

#endif /* CQFILE */   

#endif /* TRAX */   
