/*******************************************************************\
*   db.frqv.c		       					    *
*   handle frequency-view descriptor structure			    *
*   29mar90 dpwe	       				   	    *
\*******************************************************************/

#include <string.h>
#include "DSPbench.h"

FRQVIEW *NewFrqView(srate)
    FLOATARG	srate;
    {
    FRQVIEW	*new;

    new = (FRQVIEW *)Mymalloc(1, sizeof(FRQVIEW), "NewFrqView");
    new->afile = NULL;
    new->minFrq = 0.0;
    new->maxFrq = srate/2.0;
    new->minMag = -60.0;
    new->maxMag = 0.0;
    new->samplingRate = srate;
    new->ftbuffer = NULL;
    new->ftsize = 0;
    new->valid = 0;
    return new;
    }

#ifdef CQFILE
void SetFrqView(frqv, astr)
    FRQVIEW	*frqv;
    CQTINFO 	*astr;
    {
    frqv->afile = astr;
    frqv->minFrq = astr->pvh->minFreq;
    frqv->maxFrq = astr->pvh->maxFreq;
    frqv->minMag = -60.0;
    frqv->maxMag = 0.0;
    }
#else /* !CQFILE */
void SetFrqView(frqv, astr)
    FRQVIEW	*frqv;
    PVSTRUCT	*astr;
    {
    frqv->afile = astr;
    frqv->minFrq = astr->minFreq;	/* presumably zero */
    frqv->maxFrq = astr->maxFreq;	/* presumably samplingRate/2 */
    frqv->minMag = -60.0;
    frqv->maxMag = 0.0;
    }
#endif /* CQFILE */

void BoundFrqvF(FRQVIEW *frqv, double minFreq, double maxFreq)
{ 	/* ensure that frqv range lies within specified range */
    if(frqv->minFrq < minFreq)	frqv->minFrq = minFreq;
    if(frqv->maxFrq > maxFreq)	frqv->maxFrq = maxFreq;
}

void MakeSndBite(pbite,snd,stt,end)
    SNDBITE	**pbite;
    SFSTRUCT	*snd;
    long	stt,end;
    {
    *pbite = (SNDBITE *)Mymalloc(1L,sizeof(SNDBITE),"MakeBite");
    (*pbite)->snd = snd;
    (*pbite)->bStart = stt;
    (*pbite)->bEnd   = end;
    }

int GetSndBitePtrs(bite, pstt, pend)
    SNDBITE	*bite;
    char	**pstt;
    char	**pend;
    {
    char	*sndBase = (char *)SFDataLoc(bite->snd);

    *pstt = sndBase + bite->bStart;
    *pend = sndBase + bite->bEnd;

    return bite->snd->format;
    }

void BiteTimeLims(bite,pstt,pend,pdur)
    SNDBITE	*bite;
    double	*pstt,*pend,*pdur;
    {
    long	stt,end;
    SFSTRUCT *snd = bite->snd;
    long     bLen;

    *pdur = Duration(snd);
    bLen  = snd->dataBsize;
    stt   = bite->bStart;
    end   = bite->bEnd;
    *pstt = FloatProp(*pdur, stt, bLen );
    *pend = FloatProp(*pdur, end, bLen ) ;
    }

int SBchans(sndb)
    SNDBITE	*sndb;
    {
    return sndb->snd->channels;
    }

int SBsampleFrameSize(sndb)
    SNDBITE	*sndb;
    {
    return SFSampleFrameSize(sndb->snd);
    }

long SBframeCount(sndb)
    SNDBITE	*sndb;
    {
    return (sndb->bEnd - sndb->bStart)
	     / SBsampleFrameSize(sndb);
    }

short *SBstart(sndb)
    SNDBITE	*sndb;
    {
    return (short *)(SFDataLoc(sndb->snd)+sndb->bStart);
    }

