/*****************************************************/
/* db.grays.c                                        */
/* Functions to configure gray scale on X display    */
/* dpwe 31mar93 after db.xmain.c                     */
/*****************************************************/

#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include "DSPbench.h"

/* for gray scale definitions */
#define GRAYBITS	6
#define GRAYS   (1<<GRAYBITS)
#define GSCALEUP (1<<(8-GRAYBITS)) /* such that GRAYS*GSCALEUP = 256 */
#define H_RED	256	/* default hue for spectrogram */
#define H_GREEN	256
#define H_BLUE	256

#ifdef __STDC__
static unsigned long *def_gray(Display *disp, int screen, 
			       int red, int green, int blue);
#else
static unsigned long *def_gray();
#endif

/* Globals */
extern	Display			*disp;
extern	int     		scrnNum;
extern	Window  		theWin;
extern	XWindowAttributes	wAttribs;
extern	GC      		gc;

extern	char			*programName;

static int 	graycolors = 0;		/* allocate coloured graymap? */
static int 	bicolore   = 0;		/* only use black and white? */

unsigned long 	*colrs;
int		grays;

typedef struct {	/* one anchor point in RGB space */
    int red;
    int green;
    int blue;
} RGB_TRIPLE;

static int SetColorCells(Display *display, int screen, 
			 unsigned long gray[], int grays, 
			 RGB_TRIPLE *rgb1, RGB_TRIPLE *rgb2)
{	/* fill in a range of color cells linearly between rgb pairs */
    Colormap color_map;
    XColor exact_def,col_def;

    unsigned long plane_masks[1];
    unsigned long i;
    long	su = 256/grays;
    int 	r0, g0, b0;
    int 	dr, dg, db;

    r0 = rgb1->red   << 8;	dr = rgb2->red   - rgb1->red;	
    g0 = rgb1->green << 8;	dg = rgb2->green - rgb1->green;	
    b0 = rgb1->blue  << 8;	db = rgb2->blue  - rgb1->blue;	

    color_map = DefaultColormap(display, screen);
    if(XAllocColorCells(display,color_map,	
			False,plane_masks,0,gray,grays)==0) {
	fprintf(stderr,"SetColorCells fail (XAllocColorCells)\n");
	exit(1);
    }
    else {
	for(i=0; i<grays; ++i)  {
	    col_def.pixel = gray[i];
	    col_def.red =   r0+(dr*i*su);
	    col_def.green = g0+(dg*i*su);
	    col_def.blue =  b0+(db*i*su);
	    col_def.flags = 7;
	    XStoreColor(display,color_map,&col_def);
	}
    }
    XFlush(display);
    return 1;
}

typedef struct {	/* an anchor point plus how many steps before next */
    RGB_TRIPLE 	rgb;
    int		npts;
} RGB_SEGMENT;

#define RGB_POINTS	5
RGB_SEGMENT rgb_table[RGB_POINTS + 1] = 
{  { {   0,   0,   0 }, 16 },
   { {  32,  16, 192 }, 8 }, 
   { { 128,  32, 144 }, 8 }, 
   { { 192,  32,  64 }, 16 }, 
   { { 255, 208,  64 }, 16 },
   { { 255, 255, 255 }, 0 } };


#define BW_POINTS	1
RGB_SEGMENT bw_table[BW_POINTS + 1] = 
{  { { 255, 255, 255 }, 32 }, 
   { {   0,   0,   0 }, 0 } };

/*
#define RGB_POINTS	4	/* how many line segments in our table * /
RGB_SEGMENT rgb_table[RGB_POINTS + 1] = 
{  { {   0,   0,   0 }, 8 }, 
   { {  32,  32, 255 }, 8 }, 
   { {  32, 128,  32 }, 8 }, 
   { { 220, 200,  96 }, 8 },
   { { 255, 255, 192 }, 0 } };	/* N+1 points for N segs * /   
 */

static unsigned long *def_gray(display,screen,red,green,blue)
    Display *display; 
    int screen,red,green,blue;
{
    static unsigned long ucolors[GRAYS];
    int	   depth, i;
    RGB_SEGMENT *table;
    int	   tabsize;

    if(graycolors)  {	/* monochrome (grayscale) colormap */
	table = bw_table;
	tabsize = BW_POINTS;
    } else {
	table = rgb_table;
	tabsize = RGB_POINTS;
    }

    depth = DisplayPlanes(display,screen);
    /* <bicolore> is command line '-b' flag to use just 2 colors */
    if(bicolore || depth < 3)	/* less than eight colours - use B&W */
	{
	grays = 2;
	ucolors[0] = XBlackPixel(display,screen);
	ucolors[1] = XWhitePixel(display,screen);
	return(ucolors);
	}
    else
	{
	grays = 1<<(depth - 1);		/* up to half of the avai grays.. */
	if(grays > GRAYS)	grays = GRAYS;	/* but at most GRAYS */
/*	SetColorCells(display, screen, ucolors, grays, 
		      0, 0, 0, red, green, blue);    */
	/* try a funky color map */
	grays = 0;
	for(i = 0; i<tabsize; ++i)  {
	    SetColorCells(display, screen, &ucolors[grays], table[i].npts, 
			  &table[i].rgb, &table[i+1].rgb);
	    grays += table[i].npts;
	} 
	return ucolors;
    }
}
    
static unsigned long *old_def_gray(display,screen,red,green,blue)
    Display *display; 
    int screen,red,green,blue;
    {
    Colormap color_map;
    XColor exact_def,col_def;

    unsigned long plane_masks[1];
    unsigned long i;
    static unsigned long ucolors[GRAYS];
    int depth;
    long scaleup;

    depth = DisplayPlanes(display,screen);
    if(depth < 3)	/* less than eight colours - use B&W */
	{
	grays = 2;
	ucolors[0] = XBlackPixel(display,screen);
	ucolors[1] = XWhitePixel(display,screen);
	return(ucolors);
	}
    else
	{
	grays = 1<<(depth - 1);		/* up to half of the avai grays.. */
	if(grays > GRAYS)	grays = GRAYS;	/* but at most GRAYS */
	scaleup = 256L/grays;
	color_map = DefaultColormap(display, screen);
	if(XAllocColorCells(display,color_map,	
			False,plane_masks,0,ucolors,grays)==0)	
	    {
	    printf("DEF COLOR FAULT (XAllocColorCells)\n");
	    grays = 2;	/* monochrome fallback */
	    ucolors[0] = XBlackPixel(display,screen);
	    ucolors[1] = XWhitePixel(display,screen);
	    return(ucolors);
	    }
	else
	    {
	    for(i=0; i<grays; i++)
		{
		col_def.pixel = ucolors[i];		
		col_def.red = (red*i)*scaleup;
		col_def.green = (green*i)*scaleup;
		col_def.blue = (blue*i)*scaleup;
		col_def.flags = 7;
		XStoreColor(display,color_map,&col_def);
		}
	    }
	XFlush(display);
/*    	printf("grays allocated OK\n");	*/
	return(ucolors);
	}
    }



