/*******************************************************************\
*   db.growtx.c							    *
*   take a set of tracks and 'grow' shorter ones in constant	    *
*   ratio to the longer ones					    *
*   (for WASPAA track-based reconstruction demo)		    *
*   1993mar11 dpwe						    *
\*******************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/

/******** prototypes ********/

/* from db.exspec.c .. */
extern PKLIST exFindPkBeforeT(PKLIST trk, FLOATARG time);
	/* find the node in the track imm. prec specified time */
TKLIST newexTailsForTime(FLOATARG time, TKLIST allTrax, TKLIST *rtnAffected);
	/* make a list of all the track tails crossing that time.
	   Output is a LIST of TRACK structs, which include pointers 
	   to PKLISTs which are the tails starting at or after the time. 
	   Also, if rtnAffected is non-null, it is pointed to a list 
	   of tracks from allTrax which contain the pklists stolen 
	   for the contents of the return.  (can fix them later). */
/* internal */
TKLIST SelTrxAbvMag(TKLIST trx, double mthresh);
TKLIST tkGetIntersect(TKLIST trx, double tMin, double tMax,
		      double fMin, double fMax);
TRACK *tkFindMaxMag(TKLIST trx);
    /* return the car of the element of <trx> with the highest pk mag */
TKLIST tkFindWoodblocks(TKLIST trx);

/******* implementations *****/

void trxNewSelect(TKLIST trx)
{	/* copy the tracklist as the new track selection */
    /* first, discard old selection */
    if(selTrax)  {
	RedrawTrax(selTrax, 1);	/* unhilite selection */
	free_list(selTrax);
    }
    selTrax = (TKLIST)copy_top((LIST)trx);
    qsort_list((LIST)selTrax, tkCmpTrxByT);
    RedrawTrax(selTrax, 2);	/* re-hilite new selection */
}

LIST remove_adj_dup_cars(LIST l)
{	/* go down the list.  If the CAR of one node is the 
	   same as the CAR of its CDR, unlink & free the CDR. */
    LIST *ptr = &l;

    while(*ptr && CDR(*ptr))  {
	if(CAR(*ptr) == CAR(CDR(*ptr)))  {
	    free_next(*ptr);
	} else {
	    advance_ptr(&ptr);
	}
    }
    return l;
}

void trxAddSelect(TKLIST trx)
{	/* add the tracklist to the current selection.
	   Duplicates are OK. */
    RedrawTrax(trx, TKM_SELECT);	/* hilite added selection */
    selTrax = append(copy_top((LIST)trx), selTrax);
    qsort_list((LIST)selTrax, tkCmpTrxByT);
    selTrax = remove_adj_dup_cars(selTrax);
}

void PromTrax(WINRCD **wrcd)
{
    double time, t_end, t_dur, dt;
    CQTINFO 	*cqi = (*wrcd)->w.snd.cqi;
    TKLIST	tails;
    TRACK	*lowest, *track;
    double	lowfrq, frq, tim, mag;
    PKLIST	lowpkl, pkl;
    int		done = 0;
    
    if(cqi == NULL)	return;		/* no loaded CQI */

    /* hijack for testing track filters - dpwe 1993apr16 */
    if(selTrax)  {
	RedrawTrax(selTrax, 1);	/* unhilite selection */
	tails = tkFindWoodblocks(selTrax);
	free_list(selTrax);
    }
/*    tails = tkGetIntersect(allTrax, 1.0, 2.0, 1000.0, 2000.0); */
    selTrax = tails;
    RedrawTrax(selTrax, 2);	/* re-hilite new selection */

    return;
}

void GrowTrax(WINRCD **wrcd)
{
    double time, t_end, t_dur, dt;
    CQTINFO 	*cqi = (*wrcd)->w.snd.cqi;
    TKLIST	tails, affected;
    TRACK	*lowest, *track;
    double	lowfrq, frq, tim, mag;
    PKLIST	lowpkl, pkl;
    int		done = 0;
    
    if(cqi == NULL)	return;		/* no loaded CQI */

    STimeLims(wrcd,&time,&t_end,&t_dur);
    fprintf(stderr, "Grow: time=%.6f..%.6f\n", time, t_end);

    if(theTails)  {	/* dispose of existing tails */
	tklFree(theTails);
	theTails = NULL;
    }

    if(theTails == NULL)  {	/* try and find some new tails */
	theTails = newexTailsForTime(time, selTrax, &affected);
	if(theTails == NULL)  {
	    fprintf(stderr, "no tails found\n");
	    return;
	}
	fprintf(stderr, "tails=%ld\n", top_size(theTails));
	tailsTime = time;
	exDrawTails(wrcd);	/* draw up tails if appropriate to mode */
    }

    /* find lowest */
    tails = theTails;
    lowest = (TRACK *)car(tails);
    lowfrq = lowest->tkbot->peak->frq;
    while(tails)  {
	track = (TRACK *)car(tails);
	frq = track->tkbot->peak->frq;
	fprintf(stderr, "frq: %.2f \n", frq);
	if(frq < lowfrq) {
	    lowfrq = frq;
	    lowest = track;
	}
	track->st.dfdt = 0;	/* steal these values to make average */
	track->st.favg = 0;
	track->st.mavg = 0;
	tails = cdr(tails);
    }
    lowpkl = lowest->tkbot;	/* get ready to step beyond all tkbots */

    /* build ratios over selected time */
    while(!done && lowpkl && lowpkl->next && lowpkl->next->peak->tim < t_end) {
	lowpkl = lowpkl->next;
	tim = lowpkl->peak->tim;
	frq = lowpkl->peak->frq;
	mag = lowpkl->peak->mag;
	tails = theTails;
	while(tails)  {		/* for each tail at each low samp pt */
	    track = (TRACK *)car(tails);
	    if( (pkl = exFindPkBeforeT(track->tkbot, tim)) != NULL)  {
		track->tkbot = pkl;
		track->st.favg += pkl->peak->frq/frq;
		track->st.mavg += pkl->peak->mag - mag;
		track->st.dfdt += 1.0;	/* weighting count for avgs */
	    } else {
		done = 1;
	    }
	    tails = cdr(tails);
	}
    }
    
    /* finish up calculating averages */
    PrintPeak(lowpkl->peak, "lowpkl");
    tails = theTails;
    while(tails)  {
	track = (TRACK *)car(tails);
	track->st.favg = track->st.favg / track->st.dfdt;
	track->st.mavg = track->st.mavg / track->st.dfdt;
	fprintf(stderr, "tk: frq %.2f  fratio %.3f  magdiff %.3f dB\n", 
		track->tkbot->peak->frq, track->st.favg, track->st.mavg);
	PrintPeak(track->tkbot->peak, "");
	tails = cdr(tails);
    }

    /* now extend based on lowest track */
    /* Current strategy : for each remaining sample in lowest track
       grow each other track by one peak node exactly aligned in 
       time and with frq and mag locked to bottom track by ratios 
       calculated above.

       In general, should continue with as many of the tracks as are 
       still existing, not just lowest, and not assume that lowest is 
       best prototype for the ensemble.  But this will work for the 
       one example I am playing with now (!!).

       Abrupt transition from actual values to locked average values - 
       should perhaps fade in over a few sample points.  Also, suddenly 
       these HF tracks are very undersampled, since the SR is derived 
       from the low track.  But we don't care;  it will synth OK.
       */
    while(lowpkl)  {
	tim = lowpkl->peak->tim;
	frq = lowpkl->peak->frq;
	mag = lowpkl->peak->mag;
	tails = theTails;
	while(tails)  {
	    track = (TRACK *)car(tails);
	    if(track->tkbot->peak->frq != frq)  {   /* NOT lowest itself */
		pkl = track->tkbot;
		pkl->next = (PKNODE *)
		            new_node(NULL, 
				     (NODE *)NewPeak(frq*track->st.favg, 
						     mag+track->st.mavg, 
						     NOMAX, tim), 
				     T_PEAK);
		track->tkbot = pkl->next;
	    } else {  	/* lowest track itself as a track - just step on */
		track->tkbot = track->tkbot->next;
	    }
	    tails = cdr(tails);
	}
	lowpkl = lowpkl->next;
    }
    /* now fix up the track header info (allows it to save OK)
       1993oct14 */
    tails = affected;	/* elements of 'selTrax' which were actually 
			   the 'owners' of the pklists made into 
			   new tracks in theTails */
    while(tails)  {
	track = (TRACK *)car(tails);
	track->tkbot = tkStatsRecalc(track->tktop, &(track->st) );
	tails = cdr(tails);
    }
	
}

int tk_mag_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by mag range */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double mag = FindTrkMaxMag(track);

    if(mag >= rg->min && mag < rg->max)		/* passes filter */
	return 1;
    else
	return 0;
}

int tk_dur_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by *duration* */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double dur = track->st.tmax - track->st.tmin;

    if(dur >= rg->min && dur < rg->max)		/* passes filter */
	return 1;
    else
	return 0;
}

void FindTrkStartEnd(TRACK *track, double *pstart, double *pend)
{	/* return the time support of the track */
    PeakPos 	*pk;
    PKLIST	pkl = track->tktop;

    *pstart = track->st.tmin;
    *pend   = track->st.tmax;

    return;
}

int tk_time_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by duration overlap */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double start, end;

    /* quick hack: check for start of track beyond end */
    if( track->st.tmin > rg->max)	return -1;
    /* if it's beyond the end, should abort search : assume
       input tracks sorted in onset order */

    FindTrkStartEnd(track, &start, &end);

    if(end < rg->min || start > rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

int tk_onset_pred(LIST data, int type, void *arg)
{	/* predicate to allow filtering of track list by duration overlap */
    TRACK *track = (TRACK *)data;
    RANGE_STR 	*rg = (RANGE_STR *)arg;
/*    double onset = track->tktop->peak->tim;	*/
    double onset = track->st.tmin;

    if(onset < rg->min)
	return 0;
    if(onset >= rg->max)
	return -1;	/* no more ?? */
    else
	return 1;
}

void FindTrkFrqRng(TRACK *track, double *pfmin, double *pfmax)
{	/* return the `frequency support' of the track */
    PeakPos 	*pk;
    PKLIST	pkl = track->tktop;
    double 	fmin, fmax, frq;

    *pfmin = track->st.fmin;
    *pfmax = track->st.fmax;

    return;
}	

int tk_frq_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by frq range */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double fmin, fmax;	

    FindTrkFrqRng(track, &fmin, &fmax);

    if(fmax < rg->min || fmin > rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

TKLIST SelTrxAbvMag(TKLIST trx, double mthresh)
{	/* return a new list of tracks taken from <trx> whose peak
	   magnitude exceeds <mthresh> */
    TKLIST 	rslt;
    RANGE_STR	rg;

    rg.min = mthresh;
    rg.max = BIGMAX;	/* 'plus infinity' */
    rslt = (TKLIST)keep_if((LIST)trx, (NodeFunc)tk_mag_pred, &rg);
    return rslt;
}

TKLIST SelTrxByMag(TKLIST trx, double mmin, double mmax)
{	/* return a new list of tracks taken from <trx> whose peak
	   magnitude fall in <mmin> .. <mmax>.
	   <mmax> is useful to avoid grouping to *larger* following 
	   tracks that should group on their own. */
    TKLIST 	rslt;
    RANGE_STR	rg;

    rg.min = mmin;
    rg.max = mmax;
    rslt = (TKLIST)keep_if((LIST)trx, (NodeFunc)tk_mag_pred, &rg);
    return rslt;
}

static int tk_magavg_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by magavg in range */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double mag = track->st.mavg;

    if(mag >= rg->min && mag < rg->max)		/* passes filter */
	return 1;
    else
	return 0;
}

TKLIST SelTrxByAvMag(TKLIST trx, double mmin, double mmax)
{	/* return a new list of tracks taken from <trx> whose *average*
	   magnitudes fall in <mmin> .. <mmax>. */
    TKLIST 	rslt;
    RANGE_STR	rg;

    rg.min = mmin;
    rg.max = mmax;
    rslt = (TKLIST)keep_if((LIST)trx, (NodeFunc)tk_magavg_pred, &rg);
    return rslt;
}

TKLIST SelTrxByDur(TKLIST trx, double dmin, double dmax)
{	/* filter <trx>, returning only those whose duration (tmax-tmin)
	   falls in the range specified */
    TKLIST 	rslt;
    RANGE_STR	rg;

    rg.min = dmin;
    rg.max = dmax;
    rslt = (TKLIST)keep_if((LIST)trx, (NodeFunc)tk_dur_pred, &rg);
    return rslt;
}

TKLIST tkGetIntersect(TKLIST trx, double tMin, double tMax,
		       double fMin, double fMax)
{	/* return tracks from the list with support overlapping 
	   the specified time-frequency rectangle */
    TKLIST 	tkl = NULL;
    RANGE_STR	rg;

    /* first select by time */
    rg.min = tMin;	rg.max = tMax;
    tkl = (TKLIST)keep_if_abort((LIST)trx, (NodeFunc)tk_time_pred, &rg);
    /* then select result by frq */
    rg.min = fMin;	rg.max = fMax;
    tkl = (TKLIST)keep_if((LIST)tkl, (NodeFunc)tk_frq_pred, &rg);
    return tkl;
}

TKLIST tkGetOnset(TKLIST trx, double tMin, double tMax)
{	/* return track subset whose onsets lie in the given range */
    TKLIST 	tkl = NULL;
    RANGE_STR	rg;

    rg.min = tMin;	rg.max = tMax;
    tkl = (TKLIST)keep_if_abort((LIST)trx, (NodeFunc)tk_onset_pred, &rg);
    return tkl;
}

int tk_find_max_mag(TRACK *track, int type, void *arg)
{	/* node function records node with highest peak mag */
    LSRCH_STR *mms = (LSRCH_STR *)arg;	/* threshold and answer */
    double mag = FindTrkMaxMag(track);

    if(mag > mms->range.max)  {
	mms->result = (LIST)track;
	mms->range.max = mag;
    }
    return 1;
}

TRACK *tkFindMaxMag(TKLIST trx)
{	/* return the element in the track list that has highest peak mag */
    LSRCH_STR 	lss;
    TRACK	*rslt;

    lss.result = (LIST)CAR(trx);	/* start with arbitrary item */
    lss.range.max = FindTrkMaxMag((TRACK *)lss.result);
    /* apply to all the tracks.. */
    maparg((LIST)trx, (NodeFunc)tk_find_max_mag, &lss);
    rslt = (TRACK *)lss.result;
    return rslt;
}

int tk_sum_dur(TRACK *track, int type, void *arg)
{	/* node function sums track durations over set */
    LSRCH_STR *mms = (LSRCH_STR *)arg;	/* threshold and answer */

    mms->range.min += track->st.tmax - track->st.tmin;
    return 1;
}

double tkSumDurns(TKLIST trx)
{	/* Return the sum of the durations of all the tracks in the list */
    LSRCH_STR 	lss;

    lss.range.min = 0.0;	/* use this to sum - init to zero */
    /* apply to all the tracks.. */
    maparg((LIST)trx, (NodeFunc)tk_sum_dur, &lss);
    return lss.range.min;
}

TKLIST tkCheckWoodblock(TKLIST trx)
{	/* predicate decides if this tracklist looks like a woodblock, 
	   may return pruned or augmented version, or nothing (?) */
    TKLIST 	rslt = NULL, nbrs, nnbrs, wholeobj;
    double	time, freq, mag1, mag2;
    TRACK	*tk1;
    PeakPos	*pk1;

    tk1  = ((TRACK *)CAR(trx));
    pk1  = tk1->tktop->peak;
    mag1 = FindTrkMaxMag((TRACK *)CAR(trx));
/*    time = pk1->tim;	freq = pk1->frq;	*/
    time = tk1->st.tmin;	freq = tk1->st.favg;
    /* find all unselected tx in wider neighborhood */
    nbrs = tkGetIntersect(allTrax, time - 0.050, time + 0.150, 
			  freq * 2./3., freq * 3./2.);
    /* tracks very near in frq will be part of the same */
    nnbrs = tkGetIntersect(nbrs, time, tk1->st.tmax + 0.030,
			  freq / 1.01, freq * 1.01);
    /* find very close tracks which are not *too* much quieter than this one, 
       but also *are not louder* (louder tracks will have own objects) */
    wholeobj = SelTrxByMag(nnbrs, mag1-20.0, mag1+2.0);
    /* (have to add a little over mag1 for max, else won't find self!) */
    free_list((LIST)nnbrs);

    rslt = (TKLIST)remove_list((LIST)nbrs, (LIST)wholeobj);
    free_list((LIST)nbrs);
    nbrs = rslt;
    mag2 = FindTrkMaxMag(tkFindMaxMag(nbrs));
    free_list((LIST)nbrs);
    fprintf(stderr, "t %.6f f %.1f margin %.2f dB\n", time, freq, mag1 - mag2);

    if(mag1 - mag2 >= 5.0)  {
	RedrawTrax(wholeobj, TKM_SPECIAL);
	return wholeobj;
    }
    else  {
	RedrawTrax(wholeobj, TKM_SELECT);
	free_list((LIST)wholeobj);
	return NULL;
    }
}

static SELPNL_STR *spnl = NULL;

int selWblkCB(void *data, int state, SELENTRY *se, int btn)
{	/* callback from spnl selects a woodblock */
    if(state == SPNL_ACTIVATE) {
	if(ShiftIsDown())
	    AddRemoveSelTrax((TKLIST)data, NULL);
	else 
	    trxNewSelect((TKLIST)data);
    }
    return 1;
}

TKLIST tkFindWoodblocks(TKLIST trx)
{	/* special case : try and filter out woodblock signals */
    TKLIST wb, wb2, wbs2, wbs = NULL, rslt = NULL, *rslt_tail;
    char name[32];
    int  sz;

    if(spnl == NULL)  {
	spnl = spnlNew("Woodblocks");
    } else {
	/* should dispose of all last time's woodblocks here */
	spnlReset(spnl);
    }

    /* fixed absolute threshold: wb must have core above -2 db */
    wbs = SelTrxAbvMag(trx, -2.0);

    rslt_tail = &rslt;
    while(wbs) {
	wb = (TKLIST)new_node(NULL, (LIST)CAR(wbs), T_TRACK);
	wb2 = tkCheckWoodblock(wb);
	wbs2 = remove_list(wbs, wb);
	free_list((LIST)wb);
	free_list((LIST)wbs);
	wbs = remove_list(wbs2, wb2);
	free_list((LIST)wbs2);
	rslt_tail = append_ptr(rslt_tail, wb2);
	if( (sz = top_size(wb2)) > 0)  {
	    TRACK *tk = (TRACK *)CAR(wb2);
	    wb2 = (TKLIST)copy_top(wb2);	/* make priv copy for CB */
	    sprintf(name, "wb@%.6f,%.1fHz(%dtx)", 
		    tk->st.tmin, tk->st.favg, sz);
	    spnlAddEntry(spnl, name, selWblkCB, (void *)wb2);
	}
    }
    /* some long woodblocks may overlap in time, so rslt may need sorting.. */
    return rslt = qsort_list(rslt, tkCmpTrxByT);
}    

#else /* !CQFILE */   

void GrowTrax(WINRCD **wrcd)	/* dummy for menu */
{
}
void PromTrax(WINRCD **wrcd)	/* dummy for menu */
{
}

#endif /* CQFILE */   

#endif /* TRAX */   
