/*******************************************************************\
*   db.grutl.c							    *
*   graphic utility functions for DSPbench			    *
*   13apr90 dpwe						    *
*   1992jul05	Added in fns to draw lines from float arrays	    *
\*******************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <vecops.h>
#include <genutils.h>
#include <samplops.h>

#include "db.grutl.h"

ShowIcon(r,iHan)
    Rect    r;
    Icon   iHan;
    {
    Rect    r2;
    RgnHandle	oldClip;

/*  FillRect(&r , black); */

    PushClip(&oldClip,r);

    r2 = r;
    InsetRect(&r2,1,1);
    ClipRect(&r2);
    r2 = r;
    r2.left   = (r2.left + r2.right - ICONWIDTH) /2;
    r2.right  = r2.left + ICONWIDTH;
    r2.top    = 1+(r2.top + r2.bottom - ICONHEIGHT) /2;
    r2.bottom = r2.top + ICONHEIGHT;

    PlotIcon(&r2,iHan);

    PopClip(&oldClip);
    }

/* #define LINEH	13		/* pixels between lines */
extern int LINEH;	/* set up dynamically in db.tgdrw.c */

#define VVOFFS	-1		/* a different vertical offset */
#define HHOFFS	1		/* horiz offset */
#define CWIDTH	6		/* 'std' chr width for positioning */
#define IOFFS	4		/* extra offset for info text */

InfoString(r,x,y,s,u)		/* write a string to the info box */
    Rect    r;
    int     x,y;		/* Chr. cell co-os */
    char    *s;
    int     u;			/* Underlined ?? */
    {
    WriteString(r.left+IOFFS+CWIDTH*x, r.top + (y-1)*LINEH + VVOFFS, s);
    if(u)
	{
	MoveTo(r.left, r.top + y*LINEH);
	LineTo(r.right, r.top + y*LINEH);
	}
    }

WriteString(x,y,s)	/* just write string on screen */
    int x;
    int	y;
    char *s;
    {
    MoveTo(HHOFFS+x,LINEH+VVOFFS+y);
    CtoPstr(s);
    DrawString((unigned char *)s);
    PtoCstr((unsigned char *)s);
    }


ABox(rec)
    Rect    rec;
    {
    EraseRect(&rec);
    FrameRect(&rec);
/*  MoveTo(rec.left,  rec.top);
    LineTo(rec.right, rec.bottom);
    MoveTo(rec.left,  rec.bottom);
    LineTo(rec.right, rec.top); */
    }

OutlineR(r)
    Rect    r;
    {
    MoveTo(r.left,r.top);
    LineTo(r.right,r.top);
    LineTo(r.right,r.bottom);
    LineTo(r.left,r.bottom);
    LineTo(r.left,r.top);
    }

GutterR(r,sides,lw,rw,th,bh)	/* draw a border line back from sides */
    Rect    r;
    int     sides;
    int     lw,rw,th,bh;	/* 4 gutter sizes */
    {
    if(sides & lSide)
	{
	MoveTo(r.left+lw,r.top);
	LineTo(r.left+lw,r.bottom);
	}
    if(sides & rSide)
	{
	MoveTo(r.right-rw,r.top);
	LineTo(r.right-rw,r.bottom);
	}
    if(sides & tSide)
	{
	MoveTo(r.left,r.top+th);
	LineTo(r.right,r.top+th);
	}
    if(sides & bSide)
	{
	MoveTo(r.left,r.bottom-bh);
	LineTo(r.right,r.bottom-bh);
	}
    }

dbWWriteIn(prec, str, ali)
    Rect *prec;
    char *str;
    int  ali;	/* alignment: -1 = left, 0 = ctr, 1 = right */
    {
    int len;
    int y = prec->top-3;

/*    y = (prec->top+prec->bottom)-LineHeight();	*/
/*    len = strlen(str)*CWIDTH;	/* ! */
    len = myTextWidth(str);
    if(ali == 1)
	WriteString(prec->right - len, y, str);
    else if(ali == -1)
	WriteString(prec->left, y, str);
    else /* centre alignment */
	WriteString((prec->left+prec->right-len)/2, y, str);
    }

/* #define     HTICKSPC    70	/* ideal # pixels between ticks */
/* #define     VTICKSPC    20	/* ideal # pixels between ticks */
#define     TKPX	2	/* pixels for the tick mark */
#define     VOFFS       3	/* Border to offset text from ticks */
/* #define     HOFFS       5	/* Border to offset text from ticks */
/* #define     LINEH	13	/* Guess at character height */

double NearestRoundNum(num, sub)
    double	num;
    double	*sub;	/* return next size down too as sub-step */
    {
static int	possVals = 3;
static double    rndNums[3][2] = {{ .707, 1.0 },{ 1.414, 2.0 },{ 3.163, 5.0}};
static double	oom = 10.0;	/* round number configuration */
    double	scale = 1.0;
    double	ref;
    int 	i;

    ref = rndNums[0][0];	/* lower bound for lowest round number */
    if(num < 0)		{ num *= -1.0; 	scale *= -1.0; }
    while(num < ref)
	{
	num *= oom;	/* next order of magnitude */
	scale /= oom;	/* keep track of order of magnitude */
	}
    while(num >= oom*ref)
	{
	num /= oom;	/* next order of magnitude */
	scale *= oom;	/* keep track of order of magnitude */
	}
    /* new num is in range ref <= num < oom*ref */
    /* quantize.. */
    i = possVals - 1;
    while(num < rndNums[i][0] && i>0)
	--i;
    num = scale * rndNums[i][1];
    if(i < 2)
	*sub = scale * rndNums[i-2+possVals][1] / oom;
    else
	*sub = scale * rndNums[i-2][1];
    return num;
    }

int DecPoints(step)	/* return number of decimal points to code a step */
    double step;
    {
    double ref = 1.0;
    int   pts = 0;

    if(step<0) step = -step;
    if(step==0)	return 0;
    while(ref>step)
	{ ref /= 10.0; ++pts;  }
    return pts;
    }

ReplaceOrAppend(char *s, int c1, int c2)
/*    char *s;		/* target string */
/*    char c1;		/* char to find */
/*    char c2;		/* char to replace it with - or stick on end */
    {
    char *t;

    if( (t = strchr(s, c1)) == NULL)
	{ t = s + strlen(s);	*(t+1) = '\0';	} /* hope for space */
    *t = c2;
    }

void AddTimeLine(pr,t1,t2,vnh,sns,und,tix,pntx) 
    /* write linear scale into a box */
    Rect    *pr;	    /* where to put it (& how long it is) */
    double   t1,t2;	    /* start and end times of box */
    int     vnh;	    /* 0 => horiz timeline, 1 => vert timeline */
    int	    sns;	    /* 1 => low to high; 0 => hi to low */
    int     und;	    /* flag to put ticks : 1= top of box, 0 = bottom */
    int	    tix[];	    /* array of indices which are labelled */
    int	    *pntx;
    {
    double   dt;
    short   dx,xb,xi,tkh,guardx,tkspc;
    double   xPerT, tPerTk, xPerTk;
    int     tkPerIx, tkix;
    long    n0;
    double   nrst, pwr, subdiv;
    double   t0, x0;
    int     dp,nc;
    char    s[32];
    Rect    trec;
    double   rescale = 1.0;
    char    rescCh = '.';

/*    XTextItem txitem[1];
    int     dir, fa, fd;
    XCharStruct	xcs;

    txitem[0].delta  = 1;
    txitem[0].font   = None;	*/

    trec  = *pr;
    *pntx = 0;

    dt = t2-t1;
    if(dt==0.0) return;
    if(dt<0)	dt = -dt;   /* args in wrong order ? */
    if(dt>5000) { rescale = .001; rescCh = 'k'; }
    if(vnh)
	{
	int hoffs = myTextWidth(" ")/2;

	dx = pr->bottom - pr->top;
	xb = pr->top;
	trec.bottom = trec.top;
	trec.right -= hoffs;
	trec.left  += hoffs;
	tkspc = 2*LINEH;
	guardx = LINEH/2;
	}
    else{
	dx = pr->right - pr->left;
	xb = pr->left;
	trec.right  = trec.left;
	trec.bottom -= VOFFS;
	trec.top += VOFFS + (und?0:-(TKPX-1));
	tkspc = myTextWidth("0.0000   ");
	guardx = 6*CWIDTH;
	}
    xPerT = ((double)dx)/dt;
    tPerTk  = ((double)tkspc)/xPerT;	/* ideal time spacing */
    /* find 1, 2 or 5 x 10^n closest to tPerTk */
    tPerTk = NearestRoundNum(tPerTk, &subdiv);
    tkPerIx = (int)(.5+tPerTk/subdiv);	/* had a rounding problem here */
    dp = DecPoints(tPerTk*rescale);
    tPerTk = subdiv;	/* actually refer to period of ticks */
    xPerTk = tPerTk * xPerT;
    n0 = (long)(.99+(t1/tPerTk));    /* earliest time exact multiple of tPerTk */
    t0 = tPerTk*(double)n0;  /* earliest time exact multiple of tPerTk */
    tkix = n0%tkPerIx;

    x0 = (t0-t1)*xPerT;
    while(x0 < dx)
	{
	if(sns)
	    xi = xb+(short)x0;
	else
	    xi = xb + dx - (short)x0;
	tkh = TKPX;
	if(tkix==0 && x0<dx-guardx)
	    {
	    myftoa(t0*rescale,s,dp);
	    if(rescale != 1.0)
		ReplaceOrAppend(s, '.', rescCh); /* replace dp with e.g. 'k' */
/*	    nc = strlen(s);
	    XQueryTextExtents(disp, XGContextFromGC(gc),s,nc,&dir, &fa, &fd,
			      &xcs);
	    txitem[0].chars = s;
	    txitem[0].nchars = nc;	*/
	    if(vnh)
		{  trec.top = xi-LINEH/2; trec.bottom = trec.top;
		   dbWWriteIn(&trec, s, 1);	}
/*		XDrawText(disp, win, gc, y0+pr->right-pr->left-xcs.width, 
			  xi+1+xcs.ascent+2*VOFFS, txitem, 1);	*/
	    else
		{  trec.left = xi;	trec.right = trec.left;
		   dbWWriteIn(&trec, s, -1); }
/*		XDrawText(disp, win, gc, xi+1, y0, txitem, 1);	*/

	    tkh = 2*TKPX;
	    *tix++ = xi;	/* remember where we put these ticks */
	    ++(*pntx);
	    }
	if(vnh)
	    {  MoveTo(und?pr->left:pr->right, xi);
	       LineTo(und?(pr->left+tkh):(pr->right-tkh), xi);  }
/*	    XDrawLine(disp, win, gc, und?pr->left:pr->right, xi, 
		  und?(pr->left+tkh):(pr->right-tkh), xi);	*/
	else
	    {  MoveTo(xi, und?pr->top:pr->bottom);
	       LineTo(xi, und?(pr->top+tkh):(pr->bottom-tkh));  }
/*	    XDrawLine(disp, win, gc, xi, und?pr->top:pr->bottom, 
		  xi, und?(pr->top+tkh):(pr->bottom-tkh));	*/
	t0 += tPerTk;
	x0 += xPerTk;
	tkix = (tkix +1 )%tkPerIx;
	}
    }

void AddLogTimeLine(pr,t1,t2,vnh,sns,und,tix,pntx) 
    /* write logarithmic scale into a box */
    Rect    *pr;	    /* where to put it (& how long it is) */
    double   t1,t2;	    /* start and end times of box */
    int     vnh;	    /* 0 => horiz timeline, 1 => vert timeline */
    int	    sns;	    /* 1 => low to high; 0 => hi to low */
    int     und;	    /* flag to put ticks : 1= top of box, 0 = bottom */
    int	    tix[];	    /* array of indices which are labelled */
    int	    *pntx;
    {
    double   dt, pixScale, idealStep;
    short   dx,xb,xi,tkh;
    double   t0, nextt0, nt0, nextdt, x0, tkdt, nexttkdt;
    int     dp,nc,tkspc;
    char    s[32];

    Rect   trec;

    trec = *pr;
    *pntx = 0;

    if(t1*t2 <= 0.0)	{ fprintf(stderr,"LogLine: invalid %f..%f\n", t1,t2);
			  abort(); }
    dt = t2-t1;
    if(dt==0.0) return;
/*    if(dt<0)	dt = -dt;   /* args in wrong order ? */
    if(vnh)
	{
	int hoffs = myTextWidth(" ")/2;

	dx = pr->bottom - pr->top - 1;
	xb = pr->top+1;
	trec.bottom = trec.top;
	trec.right -= hoffs;
	trec.left  += hoffs;
	tkspc = 2*LINEH;
	}
    else{
	dx = pr->right - pr->left - 1;
	xb = pr->left + 1;
	trec.right  = trec.left;
	trec.bottom -= VOFFS;
	trec.top    += VOFFS + (und?0:-(TKPX-1));
	tkspc = myTextWidth("0.0000   ");
	}
    pixScale = log(t2/t1)/dx;	/* so t(x) = t1 * exp(pixscale*(x-left)) */
    idealStep = exp(pixScale*tkspc);
    if(idealStep>2.0)	idealStep = 2.0;	/* else lose orders of mag */
    /* 2.0 is the smallest ratio between adjacent 'RoundNumbers' - I know */
/*    fprintf(stderr,"idealStep %f\n",idealStep);	*/

    dt = NearestRoundNum(t1 * (idealStep-1.0), &tkdt);
    t0 = dt * MAX(1, (int)(.0 + t1/dt));	/* hacked - blocking 1993apr */
    dt = NearestRoundNum(t0 * (idealStep-1.0), &tkdt);
    t0 = dt * (int)(.5 + t0/dt);

    while( (x0 = log(t0/t1)/pixScale) < dx)
	{
/*	fprintf(stderr, "t0 %.0f dt %.0f tkdt %.0f\n", t0, dt, tkdt);	/* */
	if(x0>=0)
	    {
	    if(sns)
		xi = xb+(short)x0;
	    else
		xi = xb + dx - (short)x0;
	    dp = DecPoints(dt);
	    myftoa(t0,s,dp);
	    if(vnh)
		{
		trec.top = xi - LINEH/2;	trec.bottom = trec.top;
		dbWWriteIn(&trec, s, 1);
		}
	    else{
		trec.left = xi;		trec.right = trec.left;
		dbWWriteIn(&trec, s, -1);
		}

	    tkh = 2*TKPX;
	    *tix++ = xi;	/* remember where we put these ticks */
	    ++(*pntx);
	    if(vnh)
		{
		MoveTo(und?pr->left:pr->right, xi);
		LineTo(und?(pr->left+tkh):(pr->right-tkh), xi);
		}
/*		XDrawLine(disp, win, gc, und?pr->left:pr->right, xi, 
			  und?(pr->left+tkh):(pr->right-tkh), xi); */
	    else
		{
		MoveTo(xi, und?pr->top:pr->bottom);
		LineTo(xi, und?(pr->top+tkh):(pr->bottom-tkh));
		}
/*		XDrawLine(disp, win, gc, xi, und?pr->top:pr->bottom, 
			  xi, und?(pr->top+tkh):(pr->bottom-tkh)); */
	    }
	nextt0 = t0+dt;
	nextdt = NearestRoundNum(nextt0 * (idealStep-1.0), &nexttkdt);
	nextt0 = dt * (int)(.5 + nextt0/dt);
	nextdt = NearestRoundNum(nextt0 * (idealStep-1.0), &nexttkdt);
	nt0    = nextdt * (int)(.51 + nextt0/nextdt);
	/* had to make it .51 so that .1*(.35/.1 + .5) == .4 not .3 */
	if(nt0 > nextt0)	nextt0 = nt0;	/* poke & hope for 1, 2, 5 */
	tkh = TKPX;
	while(((t0 += tkdt) < (.99*nextt0)) && (x0 = log(t0/t1)/pixScale)<dx )
	    if(x0>0) 
		{
		if(sns)
		    xi = xb+(short)x0;
		else
		    xi = xb + dx - (short)x0;
		if(vnh)
		    {
		    MoveTo(und?pr->left:pr->right, xi);
		    LineTo(und?(pr->left+tkh):(pr->right-tkh), xi);
		    }
/*		XDrawLine(disp, win, gc, und?pr->left:pr->right, xi, 
				und?(pr->left+tkh):(pr->right-tkh), xi); */
		else
		    {
		    MoveTo(xi, und?pr->top:pr->bottom);
		    LineTo(xi, und?(pr->top+tkh):(pr->bottom-tkh));
		    }
/*		XDrawLine(disp, win, gc, xi, und?pr->top:pr->bottom, 
				xi, und?(pr->top+tkh):(pr->bottom-tkh)); */
		}
	t0 = nextt0;
	dt = nextdt;
	tkdt = nexttkdt;
	}
    }

myftoa(f,s,dp)				/* convert a double to C string */
	FLOATARG f;
	char     *s;
	int      dp;			/* number of desired decimal places */
	{
	char     format[32];
	
	sprintf(format,"%%.%df",dp);
	sprintf((char *)s,format,f);
	}

FloatToString(f,s,dp)		/* like built in NumToString, but for floats (& Str255) */
	double	f;
/*	Str255	s;	*/
	char 	*s;
	int 	dp;				/* number of decimal places */
	{
	myftoa(f,s,dp);
	CtoPstr((char *)s);	  
	}

void LegendLine(prec, marker, colour, text)
    Rect *prec;
    int marker;
    int colour;
    char *text;
    {
    Point pts[2];
    Rect  r2;
    int   leftgutter;

    pts[0].v = pts[1].v = (prec->top+prec->bottom) >> 1;
    r2 = *prec;
    leftgutter = (r2.right - r2.left) / 4;
    r2.left += leftgutter;
    pts[0].h = r2.left - 0.7*leftgutter;
    pts[1].h = r2.left - 0.2*leftgutter;
    SetColour(colour);
    DrawLines(pts, 2);
    if(marker >= 0)
	DrawMarkPoints(&pts[0], 2, marker);
    dbWWriteIn(&r2, text, -1);
    SetColour(BLACK);
    }

#define MAXSHOWPTS  16384L	/* maximum number of points to display */
#define ELSPERPT	2	/* two display points for each data bar */
static Point *ptsBuf = NULL;

void ShowFloats(prec, data, skip, size, Xmin, Xmax, Ymin, Ymax, colour)
    Rect 	*prec;
    float	*data;
    int 	skip;
    long	size;
    FLOATARG	Xmin, Xmax, Ymin, Ymax;	/* Xmin and Xmax are 0.0 .. 1.0 */
    int 	colour;
    {
    long i;
    double x, x1, y;
    RgnHandle	oldClip;
    long nPts = 0;

    if(size > MAXSHOWPTS)  {
	fprintf(stderr, "ShowFloats: %ld pts > %ld\n",
		size, MAXSHOWPTS);
	abort();
	}
    if(ptsBuf == NULL)
	ptsBuf = (Point *)Mymalloc(MAXSHOWPTS*ELSPERPT, 
				   sizeof(Point), "showfloats");
    for(i = 0; i<size; ++i)  {
	x  = i/(double)size;
	x1 = (i+1)/(double)size;
	y  = data[i*skip];
	if( (Xmin <= x && x < Xmax) || (Xmin <= x1 && x1 < Xmax))  {
	    ptsBuf[nPts].v = 
		ptsBuf[nPts+1].v = MapFtoI(y, Ymin, Ymax, 
					   prec->bottom, prec->top);
	    ptsBuf[nPts].h = MapFtoI(x, Xmin, Xmax, 
				    prec->left, prec->right);
	    ptsBuf[nPts+1].h = MapFtoI(x1, Xmin, Xmax, 
				      prec->left, prec->right);
	    nPts += 2;
	    }
	}
    if(nPts) {
	PushClip(&oldClip, *prec);
	SetColour(colour);
	DrawLines(ptsBuf, nPts);
	SetColour(BLACK);
	FlushDisplay();
	PopClip( &oldClip );
	}
    }

/* by contrast, this function takes interleaved X and Y floats */
void ShowFloatsAsLine(prec, data, size, 
		      Xmin, Xmax, Ymin, Ymax,
		      marker, colour)
    Rect 	*prec;
    float	*data;
    long	size;
    FLOATARG	Xmin, Xmax, Ymin, Ymax;
    int 	marker;
    int 	colour;
    {
    double x,y;
    long i, nPts = 0;
    RgnHandle	oldClip;

    if(size > MAXSHOWPTS)  {
	fprintf(stderr, "ShowFloats: %ld pts > %ld\n",
		size, MAXSHOWPTS);
	abort();
	}
    if(ptsBuf == NULL)
	ptsBuf = (Point *)Mymalloc(MAXSHOWPTS*ELSPERPT, 
				   sizeof(Point), "showfloats");
    for (i=0; i<size; i++)
	{
	x = *data++;
	y = *data++;
	if(x >= Xmin && x < Xmax)  {
	/*    if(y<Ymin)		y = Ymin;
	 *    else if(y>Ymax)	y = Ymax;
	 */
	    ptsBuf[nPts].h = MapFtoI(x, Xmin, Xmax,
				     prec->left, prec->right);
	    ptsBuf[nPts].v = MapFtoI(y, Ymin, Ymax, 
				     prec->bottom, prec->top);
	    ++nPts;
	    }
	}
    if(nPts)  {
	PushClip(&oldClip, *prec);
	SetColour(colour);
	if(marker >= 0)
	    DrawMarkPoints(ptsBuf, nPts, marker);
	DrawLines(ptsBuf, nPts);
	FlushDisplay();
	SetColour(BLACK);
	PopClip( &oldClip );
	}
    }

