/***********************************************************************\
*   db.harmon.c								*
*   Identify harmonic relationships between tracks - not unlike 	*
*   the original `growtx'						*
*   1993apr25 dpwe							*
\***********************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <genutils.h>		/* for Convo ??? (in DSPbench.h anyway) */
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/
static SELPNL_STR *spnl = NULL;		/* maybe offer user choices */
static cpxpolar *spect = NULL;		/* buffer to read back actual spcts */
static CQTINFO 	*cqi = NULL;
static WINDAT *wndw1 = NULL;
static WINDAT *wndw2 = NULL;

/******** prototypes ********/

/******* implementations *****/

void FindHarmsM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    TGLIST	tails, t2;
    long	totBins;

    cqi = (*wrcd)->w.snd.cqi;		/* load up our static global */
    if(cqi == NULL || selTrax == NULL)	return;		/* no spectra loaded */

    totBins  = cqi->nBins * cqi->nOcts;
    if(spect == NULL) {
	spect = TMMALLOC(cpxpolar, totBins, "FindNoise:spect");
    }
    if(wndw1 == NULL) {
	/*  wndw1 = NewWin("noise mag", 0);	
	  wndw2 = NewWin("smoothed", 0);
	  */
    }
    if(spnl == NULL)  {
	spnl = spnlNew("Harmon");
    } else {
	/* should dispose of all last time's stuff here */
	spnlReset(spnl);
	spnlUnhide(spnl);
    }
    if(selTrax)  {
	tails = lmappend(selTrax, (LNodeFunc)tgl_harm_tk, allTrax);
	t2    = tglRemoveNearDuplicates(tails);
	free_list((LIST)tails);	tails = t2;
	spnl = spnlFromTkgList(spnl, tails, wrcd);
    }
    return;
}

void rngStretch(RANGE_STR *rg, double val)
{	/* extend the range structure to include <val> */
    if(rg->min > val)		rg->min = val;
    else if(rg->max < val)	rg->max = val;
}

int rngTest(RANGE_STR *rg, double val)
{	/* compare val with range;  -1 if val < rg->min, 1 if val > rg->max, 
	   0 if val in range */
    if(val < rg->min)		return -1;
    else if(val > rg->max)	return 1;
    else			return 0;
}

int rngIntersect(RANGE_STR *one, RANGE_STR *two, RANGE_STR *rslt)
{	/* test and/or calculate the intersection between two ranges */
    if(one->max < two->min || one->min > two->max)
	return 0;
    else {
	if(rslt)  {
	    rslt->min = MAX(one->min, two->min);
	    rslt->max = MIN(one->max, two->max);
	}
	return 1;
    }
}

int rngUnion(RANGE_STR *one, RANGE_STR *two, RANGE_STR *rslt)
{	/* test and/or calculate the union of two ranges */
    int rc = 1;

    if(one->max < two->min || one->min > two->max)
	rc = 0;	/* no overlap .. */
    if(rslt) {	/* .. but calculate union anyway */
	rslt->min = MIN(one->min, two->min);
	rslt->max = MAX(one->max, two->max);
    }
    return rc;
}

static int tk_extend_ranges(NODE *trk, int type, void *ranges)
{	/* node func for finding ranges in t, f, mag */
    TRACK *tk = (TRACK *)trk;
    RANGE_STR *t_rg = &((RANGE_STR *)ranges)[0];
    RANGE_STR *f_rg = &((RANGE_STR *)ranges)[1];
    RANGE_STR *m_rg = &((RANGE_STR *)ranges)[2];
    RANGE_STR r;

    assert(type == T_TRACK);
    r.min = tk->st.tmin;	r.max = tk->st.tmax;
    rngUnion(&r, t_rg, t_rg);
    r.min = tk->st.fmin;	r.max = tk->st.fmax;
    rngUnion(&r, f_rg, f_rg);
    /* by contrast, mag range is JUST max mag in range */
    rngStretch(m_rg, tk->st.mmax);
    return 1;
}

#define LARGE_POS (1e30)

int tklFindRanges(TKLIST trx, RANGE_STR *time_rg, RANGE_STR *frq_rg, 
		  RANGE_STR *mag_rg)
{	/* For each track in list, extend ranges for whole range of track.
	   i.e. <time_rg> gets MIN(tk->tmin)..MAX(tk->tmax) over all trax etc.
	   However, mag_rg only records range of mmax (not mmin). */
    RANGE_STR rs[3];

    rs[0].min = LARGE_POS;	rs[0].max = -LARGE_POS;
    rs[1].min = LARGE_POS;	rs[1].max = -LARGE_POS;
    rs[2].min = LARGE_POS;	rs[2].max = -LARGE_POS;
    maparg(trx, tk_extend_ranges, rs);
    if(time_rg)	{ time_rg->min = rs[0].min; time_rg->max = rs[0].max;  }
    if(frq_rg)	{ frq_rg->min = rs[1].min;  frq_rg->max = rs[1].max;  }
    if(mag_rg)	{ mag_rg->min = rs[2].min;  mag_rg->max = rs[2].max;  }
    return 1;
}

int tkFrqMagRange(TRACK *trk, RANGE_STR *trg, RANGE_STR *frg, RANGE_STR *mrg)
{	/* Return frq and magnitude ranges spanned by the specified track 
	   *over the specified time range* */
    PKLIST pkl;
    double	t1, t2, f1, f2, m1, m2, frac;

/*    fprintf(stderr, "tkFMR: tk%d (t %.6f..%.6f) over t %.6f..%.6f\n", 
           trk->id, trk->st.tmin, trk->st.tmax, trg->min, trg->max); */

    if(!trk || trg->max < trk->st.tmin || trg->min >=trk->st.tmax)  {
	/* out of range - don't touch args, return 0 */
	return 0;
    } else {
	pkl = exFindPkBeforeT(trk->tktop, trg->min);
	if(pkl == NULL || pkl == trk->tktop)
	    pkl = trk->tktop->next;	/* tmin preceded this track, or 
					   hit first (fake) node - skip it,
					   always start at 2nd node */
	frg->max = frg->min = pkl->peak->frq;
	mrg->max = mrg->min = pkl->peak->mag;
	pkl = pkl->next;
	/* DON'T test the peak with no 'next' - it's the dummy last one */
	while(pkl && pkl->next && rngTest(trg, pkl->peak->tim) <= 0)  {
	    rngStretch(frg, pkl->peak->frq);
	    rngStretch(mrg, pkl->peak->mag);
	    pkl = pkl->next;
	}
	return 1;
    }
}

int tkFrqMagForT(TRACK *trk, double time, double *pfrq, double *pmag)
{	/* return the magnitude and frequency of a track at the 
	   specified time */
    PKLIST pkl;
    double	t1, t2, f1, f2, m1, m2, frac;

    if(!trk || time < trk->st.tmin || time >= trk->st.tmax)  {
	/* out of range - clear args & return 0 */
	*pfrq = *pmag = 0.0;
	return 0;
    } else {
	pkl = exFindPkBeforeT(trk->tktop, time);
	assert(pkl->peak->tim <= time && time < pkl->next->peak->tim);
	t1 = pkl->peak->tim;	t2 = pkl->next->peak->tim;
	f1 = pkl->peak->frq;	f2 = pkl->next->peak->frq;
	m1 = pkl->peak->mag;	m2 = pkl->next->peak->mag;
	/* linearly interpolate the frq and (dB) mag */
	frac = (time - t1)/(t2 - t1);
	*pfrq = f1 + frac * (f2-f1);
	*pmag = m1 + frac * (m2-m1);
	return 1;
    }
}

int tIntersect(double tl1, double tu1, double tl2, double tu2, 
	       double *prl, double *pru)
{      /* return 1 if the two time intervals <tl1..tu1> and <tl2..tu2>
	  have an overlap - return it as <*prl..*pru>.  Else 0 */
    *prl = MAX(tl1, tl2);
    *pru = MIN(tu1, tu2);
    if(*pru < *prl)  {	/* no overlap */
	*prl = *pru = 0.0;
	return 0;
    } else {
	return 1;
    }
}

double tksTestRatio(TRACK *tk1, TRACK *tk2, double *pratio, double *prsd)
{	/* test the frequency ratio between two tracks, over common 
	   time extent.  Return common duration */
    double t1, t2, dur;
    double frq, mag, ratio, linmag, magsum, ratiosum, ratsumsq;
    PKLIST	pkl;
    int 	i = 0;	/* debug only */

    tIntersect(tk1->st.tmin, tk1->st.tmax, tk2->st.tmin, tk2->st.tmax, 
	       &t1, &t2);
    dur = t2 - t1;
    if(dur > 0.0)  {	/* there *is* some overlap */
	/* drive comparison from first track (fewer points if lower) ?? */
	pkl = exFindPkBeforeT(tk2->tktop, t1);
	/* ensure pkl->tim >= t1 so both tracks active */
	if(pkl->peak->tim < t1)
	    pkl = CDR(pkl);
	magsum = 0.0;
	ratiosum = 0.0;
	ratsumsq = 0.0;
	while(pkl && pkl->peak->tim < t2)  {
	    tkFrqMagForT(tk1, pkl->peak->tim, &frq, &mag);
	    linmag = LINOFDB(MIN(mag,pkl->peak->mag));
	    magsum   += linmag;
	    ratio  = pkl->peak->frq/frq;
	    ratiosum += linmag*ratio;
	    ratsumsq += linmag*ratio*ratio;
	    pkl = CDR(pkl);
	    ++i;		/* just for debug - how many pts? */
	}
	/* fprintf(stderr, "pts: %d\n", i); */
	ratio = ratiosum / magsum;
	*pratio = ratio;
	*prsd   = sqrt(ratsumsq / magsum - ratio*ratio);
    }
    return dur;
}

int tkCmpFAvg(NODE **pn1, NODE **pn2)
{   /* qsort comparison function:
       Just sort into ascending frq order - ignore times. */
    TRACK *tk1, *tk2;
    double f1, f2;

    tk1 = (TRACK *)(*pn1);
    tk2 = (TRACK *)(*pn2);
    f1 = tk1->st.favg;
    f2 = tk2->st.favg;

    if(f1 > f2)  return 1;
    else if(f1 == f2)  return 0;
    else return -1;
}

int tk_as_harm_pred(TRACK *cand, int type, TRACK *ref)
{	/* list predicate tests <cand> as a harmonic of <ref> 
	   i.e. their frequency ratio is approx const over time, 
	   and close to an integer.  <ref> may be a harmonic itself;
	   all freq ratios are scaled by <position> */
    double cdur, rdur;
    double fr, rsd, ov, rratio, avovpr;
    double position = ref->st.dfdt;	/* (hidden in here) */

    cdur = cand->st.tmax - cand->st.tmin;
    rdur = ref->st.tmax - ref->st.tmin;
    ov = tksTestRatio(ref, cand, &fr, &rsd);
    fr *= position;		/* frq ratio to fundamental, not this tk */
    rratio = rint(fr);
/*    fprintf(stderr, "::ratio %6.3f, rsd/fr %.6f, ovt %.6f(%.1f%%)\n", 
	    fr, rsd/fr, ov, 100.0*ov/rdur);	*/

    /* time overlap criteria: 
       The PROPORTION OF OVERLAP TIME IN THE ENTIRE TRACK DURATION
         AVERAGED over the two tracks in question
	 MUST exceed 55% : 
	 If the time suport of one track is a subset of that of the 
	 other, its proportion of overlap time will be 100%; in 
	 this case, it must constitute at least 10% of the duration 
	 of the first track to pass */
    avovpr = .5*(ov/cdur + ov/rdur);
    if( (rsd/fr) < 0.05 && avovpr > .55 && fabs(rratio - fr)/fr < .02) {
	return 1;
    } else {
	return 0;
    }
}

#define MAXHARMONIC	(10.0)	/* search for harmonics in wdw this wide */

TKLIST tkFindHarmonics(TRACK *trk, double position, TKLIST focus)
{	/* Find tracks that serve well as harmonics of <trk>, where 
	   the fundamental frequency is obtained as the frequency 
	   of <trk> divided by <position> - <position> typically close 
	   to an integer */
    TRACK *tk2;
    TKLIST src, rslt, *rsTail;
    double fr, rsd, ov, avovpr, dur1, dur2, rratio;
    double tons;

    if(!trk)
	return NULL;

    if(focus == NULL)
	focus = allTrax;

    tons = tkOnsetT(trk);
    dur1 = trk->st.tmax - trk->st.tmin;
/*    tk_print_onset(trk, 0, NULL);	*/
/*    src = (TKLIST)CDR(trx);  */
    src = tkGetIntersect(focus, trk->st.tmin, trk->st.tmax, 
			 trk->st.fmin/position, 
			 MAXHARMONIC * trk->st.fmax/position);
/*    fprintf(stderr, "Harmonics: considering %ld trx\n", top_size(src)); */

    trk->st.dfdt = position;	/* tuck it in here */

    rslt = keep_if(src, (NodeFunc)tk_as_harm_pred, trk);
    free_list(src);

    return rslt; /*  = qsort_list(rslt, tkCmpTrxByT); */
}

static TKGROUP *tg_harmonic_gen(LIST node, int *ptype, void *args[])
{	/* LNodeFunc to generate harmonics */
    double 	position = *(float *)&node;
    TRACK 	*base = (TRACK *)args[0];
    TKLIST 	focus = (TKLIST)args[1];
    TKGROUP 	*rslt;
    TKLIST	parts;

    parts = tkFindHarmonics(base, position, focus);
    rslt = tkgrNew(parts, TG_HARM);
    rslt->v.harm.fund = (TKLIST)new_node(NULL, (LIST)base, T_TRACK);
    rslt->v.harm.ratio = position;
    *ptype = T_TKGROUP;
    return rslt;
}

#define HARM_TOP_ROOT 	4
/* test track as root of harmonic rule in positions 1..4 */

TGLIST tgHarmons(TRACK *trk, TKLIST focus)
{	/* generate a set of harmonic track groups based on trk */
    TGLIST rslt, rs2;
    LIST   poses = NULL;
    int    i;
    float  f;
    void   *args[2];		/* for passing args to LNodeFunc */

    for(i = 1; i <= HARM_TOP_ROOT; ++i)  {
	f = (float)i;
        poses = new_node(poses, *(LIST *)&f, T_1);	/* T_1 -> float */
    }
    /* generate a harmonic set for each entry in <poses> (!) */
    args[0] = (void *)trk;
    args[1] = (void *)focus;
    rslt = lmaparg(poses, (LNodeFunc)tg_harmonic_gen, args);
    /* remove near-subsets */
    rs2 = tglRemoveNearDuplicates(rslt);
    free_list(rslt);
    rslt = rs2;

    return rslt;
}

#define HARM_DUR_THRSH	(0.100)		/* can't have harmonic seeds shorter */

TGLIST tgl_harm_tk(TRACK *trk, int *ptype, TKLIST focus)
{	/* LNodeFunc to generate harmonic sets from tracks */
    double dur = trk->st.tmax - trk->st.tmin;

    assert(*ptype == T_TRACK);
    *ptype = T_LIST;	/* returning a TGLIST */

    if(dur < HARM_DUR_THRSH)
	return NULL;
    else
	return tgHarmons(trk, focus);
}

void tkgrOutputForMatlab(TKGROUP *tkg)
{	/* print some stuff out to paste into a matlab file
	   to register this track group */
    fprintf(stderr, "%8.6f %8.6f %7.1f %7.1f %d\n", 
	    tkg->st.tmin, tkg->st.tmax, tkg->st.fmin, tkg->st.fmax, 
	    tkg->type);
}

void tkgrSelect(TKGROUP *tkg, int merge)
{	/* make the tracks in the <tkg> be the current selection.
	   If <merge> is nonzero, add them to the existing selection, 
	   otherwise replace it */
    if(merge == 0)  {
	/* should test for TKG_META (nested group) here !! */
	trxNewSelect(tkg->parts);
    }
    else {
	AddRemoveSelTrax(tkg->parts, NULL);
    }
}

void selTkgrSaveCB(SELPNL_STR *spnl)
{	/* somehow we're going to save the groups in this selpnl */
    TGLIST tgl = NULL;
    TGLIST *runner = &tgl;
    TKGROUP *tkg;
    int i, n;
    SELENTRY *se;
    char filename[256], *title;

    fprintf(stderr, "now spnl = %lx\n", spnl);
    n = spnlGetNEntries(spnl);
    for(i = 0; i<n; ++i)  {
	se = spnlGetNth(spnl, i);
	tkg = (TKGROUP *)se->cbData;
	*runner = new_node(NULL, (LIST)tkg, T_TKGROUP);
	advance_ptr(&runner);
    }
    title = spnlGetTitle(spnl);
    SubstExtn(title, ".tgl", filename);
    tglSave(tgl, filename, title);
}

void selTkgrLoadCB(SELPNL_STR *spnl)
{	/* somehow we're going to load the groups in this selpnl */
    TGLIST tgl;
    char filename[256], *title;

    title = spnlGetTitle(spnl);
    SubstExtn(title, ".tgl", filename);
    if(tgl = tglLoad(filename, &title, allTrax))  {
	spnl = spnlFromTkgList(spnl, tgl, NULL);
	spnlSetTitle(spnl, title);
    }
}

static int selTkgrCB(void *data, int state, SELENTRY *se, int btn)
{	/* callback from spnl selects a tkgr */
    TKGROUP *tkg = (TKGROUP *)data;

    if(state == SPNL_ACTIVATE)  {
	if(ShiftIsDown())  {
	    tkgrSelect(tkg, 1);		/* exor tracks w/ sel */
	} else if(btn == 1)  {
	    tkgrSelect(tkg, 0);		/* replace sel w/ trx */
	} else if(btn == 2) {
	    trxAddSelect(tkg->parts);	/* merge trx into sel */
	} else {
	    tkgrOutputForMatlab(tkg);	/* report text */
	}
    }
    return 1;
}

static int tg_add_spnl(LIST node, int type, void *arg)
{	/* node function adds track groups to a selpanel */
    TKGROUP *tkg = (TKGROUP *)node;
    SELPNL_STR *spnl = (SELPNL_STR *)arg;
    char str[128];

    assert(type == T_TKGROUP);
    tkgrTitle(tkg, str);
    spnlAddEntry(spnl, str, selTkgrCB, (void *)tkg);
    if(tkg->label == NULL)
	tkg->label = strdup(str);
    return 1;
}

int tg_time_cmp(TKGROUP **ptg1, TKGROUP **ptg2)
{	/* for qsort of TKGROUPS into ascending time, return 1 if *ptg2 
	   points to a TKGROUPS of later minimum time than *ptg1 */
    double w1 = (*ptg1)->st.tmin;
    double w2 = (*ptg2)->st.tmin;
    
    if(w2 < w1)		return 1;
    else if (w1 < w2)	return -1;
    else		return 0;
}

extern void SynthFromTracks(WINRCD **wrcd);

SELPNL_STR *spnlFromTkgList(SELPNL_STR *spnl, TGLIST tgl, WINRCD **wrcd)
{	/* present a list of track groups as entries in a selpanel */
    if(spnl == NULL)  {
	spnl = spnlNew("TrackGroups");
    } else {
	/* should dispose of all last time's groups here */
	spnlReset(spnl);
    }
    /* add load/save functions */
    if(wrcd)  {
	spnlSetHackFn(spnl, "Play/Save/Load", 
		      SynthFromTracks, wrcd, 
		      selTkgrSaveCB, spnl,
		      selTkgrLoadCB, spnl);
    }
    fprintf(stderr, "spnl = %lx\n", spnl);
    /* sort groups by start time */
    tgl = copy_top(tgl);	/* make our own copy we can mess with */
    tgl = (TGLIST)qsort_list((LIST)tgl, (NodeCmpFn)tg_time_cmp);
    maparg((LIST)tgl, tg_add_spnl, (void *)spnl);
    free_list(tgl);
    return spnl;
}

#else /* !CQFILE */   

void FindHarmsM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    /* dummy */
}

#endif /* CQFILE */   

#endif /* TRAX */   

