/*******************************************************************\
*   db.macplay.c                                                    *
*   Mac sound manager interface for dspBench                        *
*   07aug90 dpwe                                                    *
\*******************************************************************/

#include "DSPbench.h"
#include <IEEE80.h>
/* THINK_C 4
   #include <SoundMgr.h>
 */
#include <Sound.h>
#define rate44khz ((Fixed)(44100.0*65536))

/*********************** P L A Y **************************************/

#define BLKSPERPLAY 4   /* merge this many blocks into each call */

#define NPLAYBUFS   8
#define STARTUPBUFS 1   /* fill up this many before we start writing them */

#define TIMEOUTLIM 1000

#define STEREO
#define USEEXTSH

#ifdef USEEXTSH
#define SOUNDHEADER ExtSoundHeader
#else /* !USEEXTSH i.e. use stdSH */
#undef STEREO   /* cannot have stereo with stdSH */
#define SOUNDHEADER SoundHeader
#endif /* USEEXTSH */

typedef unsigned char uchar;
static int playBufIx = -1;     /* index of buffer currently setting up */
static int sentBufIx = -1;     /* index of buffer last sent to sound mgr */
static int doneBufIx = -1;     /* index of buffer last completed by snd mgr */
static int sendStarted = 0;    /* if we haven't send *any* blocks yet, wait */
static int sendCallIx = -1;    /* count calls to assemble into play blks */
static SOUNDHEADER   *playBufs[NPLAYBUFS];      /* all our buffers */
static SndChannelPtr schn;
static SndCommand    scmd;

extern void double_to_ieee_80(double val, uchar *p);      /* (for extSH) */

static void playCallback(void)
{	/* function called back by sound manager indicates buffer free */
    ++doneBufIx;
}

static int playInit(int chans, double srate, long framesPerBlock)  
{               /* setup the sound manager */
    long            frames = framesPerBlock;
    long            i;
    SOUNDHEADER     *tmpH;
    OSErr               oerr;
    Fixed               fxSr;

    playBufIx = -1;         /* index of buffer currently setting up */
    sentBufIx = -1;         /* index of buffer last sent to sound mgr */    
    doneBufIx = -1;         /* index of buffer last completed by snd mgr */
    sendStarted = 0;        /* if we haven't send *any* blocks yet, wait */
    sendCallIx = -1;        /* count calls to assemble into play blks */
#ifndef STEREO
    chans = 1;              /* force mono - only way it works for now */
#endif /* !STEREO */
    /* setup sound output hardware */
    schn = NULL;
    if(chans == 1)  
        oerr = SndNewChannel(&schn, sampledSynth, initMono, 
                             (SndCallBackProcPtr)playCallback);     
    else /* chans == 2 */
        oerr = SndNewChannel(&schn, sampledSynth, initStereo, 
                             (SndCallBackProcPtr)playCallback);     

    /* allocate all the play buffers */
    for(i = 0; i < NPLAYBUFS; ++i) {
        playBufs[i] = (SOUNDHEADER *)NewPtr(sizeof(SOUNDHEADER)
                                            + BLKSPERPLAY * chans 
                                            * frames * sizeof(uchar));
        if(playBufs[i] == NULL)  {
            D_prints("Mem alloc fail in playInit\n");
            return 0;
        }
    }
    fxSr = (Fixed)(srate * (1L<<16));
    if(42000 < srate && srate < 46000)       fxSr = rate44khz;
    else if(21000 < srate && srate < 23000)  fxSr = rate22khz;
    else if(10500 < srate && srate < 11500)  fxSr = rate11khz;  

    tmpH = playBufs[0];
    tmpH->samplePtr   = tmpH->sampleArea;
    tmpH->sampleRate  = fxSr;
    tmpH->loopStart   = 0L;   
    tmpH->loopEnd     = 0L;
    tmpH->baseFrequency = 60; /* encode & middle C */
#ifdef USEEXTSH
    tmpH->encode      = extSH;      /* was stdSH */
    tmpH->numChannels = chans;      /* */
    tmpH->numFrames   = frames;
    double_to_ieee_80(srate, (uchar *)&(tmpH->AIFFSampleRate));
    tmpH->markerChunk = NULL;
    tmpH->instrumentChunks = NULL;
    tmpH->AESRecording = NULL;
    tmpH->sampleSize   = 8;
#else /* !USEEXTSH - stdSH instead */
    tmpH->encode      = stdSH;
    tmpH->length      = BLKSPERPLAY * chans * frames * sizeof(uchar);
#endif /* USEEXTSH */
        
    for(i = 1; i < NPLAYBUFS; ++i) {
        *(playBufs[i]) = *tmpH;         /* copy config to other bufs */
        (playBufs[i])->samplePtr = (playBufs[i])->sampleArea;
    }
}

static void playSendBuf(int idx)
{       /* actually issue Snd commands for indexed buffer */
    scmd.cmd    = bufferCmd;
    scmd.param1 = 0;
    scmd.param2 = (long)playBufs[idx];
    SndDoCommand(schn, &scmd, FALSE);
    scmd.cmd    = callBackCmd;           /* so that buffer is released */
    SndDoCommand(schn, &scmd, FALSE);
    ++sentBufIx;
}

static void playFree(void)
{       /* unsetup everything */
    long i;
    int  timeout;

    while(sentBufIx < playBufIx)  {
        playSendBuf( (sentBufIx+1) % NPLAYBUFS);
    }
    scmd.cmd = nullCmd;
    timeout = 0;
    while(doneBufIx < playBufIx && ++timeout < TIMEOUTLIM)  {
        SndDoCommand(schn,&scmd,FALSE);
    }
    SndDisposeChannel(schn, TRUE); /* waits for commands to end */
    for(i = 0; i < NPLAYBUFS; ++i)  {
        DisposPtr(playBufs[i]);
    }
}

static void playSend(short *sptr, int chans, long frames)
{       /* actually play some samples .. must agree with init */
    uchar *uptr;
    long  i, samples;
    int       pbIdx;

    sendCallIx = (sendCallIx+1) % BLKSPERPLAY;      /* static groups blocks */

    /* On entry, playBufIx is the index of the last buffer we sent.
       We check that against the number that have completed;  if 
       there are more than NPLAYBUFS `in progress', we block until 
       another completes, thereby freeing one of our buffers */
    if(sendCallIx == 0)      {              /* about to start a new frame */
        ++playBufIx;
        scmd.cmd = nullCmd;
        while((playBufIx - doneBufIx) > NPLAYBUFS)  {
            SndDoCommand(schn, &scmd, FALSE); /* null command allw callback */
        }
    }

    pbIdx = (playBufIx) % NPLAYBUFS;                /* line up the next one */
    if(sendCallIx == 0)  {  /* first frame in this Str */
#ifdef USEEXTSH
        (playBufs[pbIdx])->numFrames = 0;
#else /* !USEEXTSH */
        (playBufs[pbIdx])->length = 0 * sizeof(uchar);
#endif /* USEEXTSH */
    }

    /* convert 16 bit signed to 8 bit unsigned */
#ifdef STEREO   /* implies USEEXTSH */
    uptr = (uchar *)((playBufs[pbIdx])->sampleArea)
                       +(playBufs[pbIdx])->numFrames*(chans*sizeof(uchar));
    samples = frames*chans;
    for(i = 0; i<samples; ++i)  {  
        *uptr++ = 0x80 + ((*sptr)>>8);
        sptr += 1;
    }
    (playBufs[pbIdx])->numFrames += frames;
#else /* !STEREO */
#ifdef USEEXTSH
    uptr = (uchar *)((playBufs[pbIdx])->sampleArea)
                      +(playBufs[pbIdx])->numFrames*sizeof(uchar);
#else /* !USEEXTSH */
    uptr = (uchar *)((playBufs[pbIdx])->sampleArea)+(playBufs[pbIdx])->length;
#endif /* USEEXTSH */
    for(i = 0; i<frames; ++i)  {
        *uptr++ = 0x80 + ((*sptr)>>8);
        sptr += chans;
    }
#ifdef USEEXTSH
    (playBufs[pbIdx])->numFrames += frames;
#else /* !USEEXTSH */
    (playBufs[pbIdx])->length += frames * sizeof(uchar);
#endif /* USEEXTSH */
#endif /* STEREO */

    if(sendCallIx == BLKSPERPLAY - 1)  {    /* just FINISHED a block to play */
        if(sendStarted)  {              /* we are sending frames .. */
            playSendBuf(pbIdx);
        } else {
            if(playBufIx == STARTUPBUFS - 1)  {  /* compltd this many frames */
                for(i = 0; i <= pbIdx; ++i)  {  /* send all the bufs at 1ce */
                    playSendBuf(i);
                }
                sendStarted = 1;
            }       /* else not started - just accumulate frames */
        }
    }
}


#define CHUNKLEN    1024     /* chop sound up into pieces this big to send */
/* globals */
static int  finished = 0;    /* flags how many blocks have finished */

/* prototypes */

/* implementations */
play(stt,end,chans,dsize,srate,scale)   /* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;      /* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;
{
    long    i;
    long    nFrames, framesLeft, framesPerBlock;
    short   *sound;

    nFrames = (end-stt)/chans;
    framesPerBlock = CHUNKLEN;
    if( playInit(chans, srate, framesPerBlock) )  {
        framesLeft = nFrames;
        sound = stt;
        while(framesLeft)  {
            framesPerBlock = MIN(framesPerBlock, framesLeft);
            SystemTask();
            playSend(sound, chans, framesPerBlock);
            sound += chans * framesPerBlock;
            framesLeft -= framesPerBlock;
        }
    playFree();
    }
}               

SndCallback()       /* call back from snd channel */
    /* run at interrupt time, so cannot do memory ops.  Just semaphore */
    {
    ++finished;
    }

oldplay(stt,end,chans,dsize,srate,scale)        /* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;      /* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;
    {
    short           *nxtbit;
    SoundHeader     *tmpH;
    unsigned char   *bPtr;
    long            slen,i,clen;
    SndChannelPtr   chan;
    OSErr           oerr;
    SndCommand      scmd;
    int             nchunk,cix;
    SoundHeader     **hantab;
    int             freed = 0;      /* how many blocks have been freed */
    int                 timeout;

/*      if(srate > 22200)  {
                chans *= 2;
                srate /= 2;             
        }                                       /* arrange to skip every other sample */


    slen = (end - stt); /* number of SAMPLES */
    nchunk = (slen / (chans*CHUNKLEN)) +1;
    hantab = (SoundHeader **)NewPtr(sizeof(Handle *)*nchunk);

    nxtbit = stt;
    chan = NULL;
    oerr = SndNewChannel(&chan, sampledSynth, initMono, (SndCallBackProcPtr)SndCallback);       
    if(oerr != 0)
        {
        D_prints("Play: OS error :");   D_printn((int)oerr);
        }
    else{
        finished = 0;
        for(cix = 0; cix<nchunk; ++cix)     /* chunking loop */
            {
                scmd.cmd = nullCmd;
                timeout = 0;
                while(freed < cix - 5 && ++timeout < TIMEOUTLIM)  {             
                        /* only allow it to get 5 frames ahead */
                        if(finished > freed)  { 
                        /*      D_printn(-freed);       D_prints(" "); /* */
                                DisposPtr(hantab[freed++]);     /* semaphore that some sound used up */
                        }
                        SndDoCommand(chan,&scmd,FALSE);
                }

            clen = (end-nxtbit)/chans;
            if(clen > CHUNKLEN) clen = CHUNKLEN;
            tmpH = (SoundHeader *)NewPtr(sizeof(SoundHeader)
                                               + clen*sizeof(unsigned char));
            if(tmpH == NULL)    /* we probably ran out of memory */
                {
                scmd.cmd = quietCmd;    scmd.param1 = 0;    scmd.param2 = 0;
                SndDoImmediate(chan, &scmd);
                scmd.cmd = flushCmd;    scmd.param1 = 0;    scmd.param2 = 0;
                SndDoImmediate(chan, &scmd);
                break;
                }
            hantab[cix] = tmpH;     /* remember this handle */
            tmpH->samplePtr = NULL;
            tmpH->length = clen*sizeof(unsigned char);
            tmpH->sampleRate = (Fixed)((long)(srate * 65536));
            tmpH->loopStart = 0L;   tmpH->loopEnd = clen; /* debug */
            tmpH->encode = stdSH;       /* new */
            tmpH->baseFrequency /* was Note */ = /* (stdSH<<8) */ +60; /* encode & middle C */
            bPtr = (unsigned char *)(tmpH->sampleArea);
            i = clen;
            while(i--)
                {
                *bPtr++ = 0x80+(unsigned char)((scale * ((*nxtbit)>>1))>>8);
                nxtbit += chans;    /* just first channel */
                }
            scmd.cmd = bufferCmd;   scmd.param1 = 0;
            scmd.param2 = (long)tmpH;
            SndDoCommand(chan,&scmd,FALSE);
            scmd.cmd = callBackCmd;         /* so that pointers are freed */
            SndDoCommand(chan,&scmd,FALSE);
        /*    D_printn(cix);    D_prints(" ");  /* */
        /*    SystemTask();     */
            }
        scmd.cmd = nullCmd;
        timeout = 0;
        while(freed < nchunk && ++timeout < TIMEOUTLIM)  {
                if(finished > freed)  { 
                /*      D_printn(-freed);       D_prints(" ");  /* */
                        DisposPtr(hantab[freed++]);     /* semaphore that some sound used up */
                }
                SndDoCommand(chan,&scmd,FALSE);
        }

        SndDisposeChannel(chan, FALSE); /* waits for commands to end */
        if(cix < nchunk)    /* for loop did not complete */
            {   D_prints("Only "); D_printn(cix); D_prints("chunks of "); D_printn(nchunk); }
        for(--cix; cix >= freed; --cix)
            DisposPtr(hantab[cix]); /* get rid of all the handles at the end */
        }
    DisposPtr(hantab);
    }

/* The extSH would appear to allow buffer commands that pass stereo 
   16 bit samples to the sound manager.  But it does not work.  dpwe 1993feb11 */

newplay(stt,end,chans,dsize,srate,scale)        /* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;      /* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;
    {
    short           *nxtbit;
    ExtSoundHeader     *tmpH;
    ExtSoundHeader     **hantab;
    unsigned char   *bPtr;
    long            slen,i,clen;
    SndChannelPtr   chan;
    OSErr           oerr;
    SndCommand      scmd;
    int             nchunk,cix;
    int             freed = 0;      /* how many blocks have been freed */
    int                 timeout;

    slen = (end - stt); /* number of SAMPLES */
    nchunk = (slen / (chans*CHUNKLEN)) +1;
    hantab = (ExtSoundHeader **)NewPtr(sizeof(Handle *)*nchunk);

    nxtbit = stt;
    chan = NULL;
    oerr = SndNewChannel(&chan, sampledSynth, initMono, (SndCallBackProcPtr)NULL);      
    if(oerr != 0)
        {
        D_prints("Play: OS error :");   D_printn((int)oerr);
        }
    else{
        finished = 0;
        for(cix = 0; cix<nchunk; ++cix)     /* chunking loop */
            {
                scmd.cmd = nullCmd;
                timeout = 0;
                while(freed < cix - 5 && ++timeout < TIMEOUTLIM)  {             
                        /* only allow it to get 5 frames ahead */
                        if(finished > freed)  { 
                        /*      D_printn(-freed);       D_prints(" "); /* */
                                DisposPtr(hantab[freed++]);     /* semaphore that some sound used up */
                        }
                        SndDoCommand(chan,&scmd,FALSE);
                }

            clen = (end-nxtbit)/chans;
            if(clen > CHUNKLEN) clen = CHUNKLEN;
            tmpH = (ExtSoundHeader *)NewPtr(sizeof(ExtSoundHeader));
            if(tmpH == NULL)    /* we probably ran out of memory */
                        {
                        scmd.cmd = quietCmd;    scmd.param1 = 0;    scmd.param2 = 0;
                        SndDoImmediate(chan, &scmd);
                        scmd.cmd = flushCmd;    scmd.param1 = 0;    scmd.param2 = 0;
                        SndDoImmediate(chan, &scmd);
                        break;
                        }
            hantab[cix] = tmpH;     /* remember this handle */
                tmpH->samplePtr = (Ptr)nxtbit;
                tmpH->numChannels = chans;
                tmpH->sampleRate = (Fixed)((long)(srate * 65536));
                    tmpH->loopStart = 0L;   tmpH->loopEnd = 0L;
                    tmpH->encode = extSH;       /* was stdSH */
                    tmpH->baseFrequency = 60; /* encode & middle C */
                tmpH->numFrames = clen;
                double_to_ieee_80(srate, (unsigned char *)&(tmpH->AIFFSampleRate));
                tmpH->markerChunk = NULL;
                tmpH->instrumentChunks = NULL;
                tmpH->AESRecording = NULL;
                tmpH->sampleSize = 16;  /* bits/sample */
            scmd.cmd = bufferCmd;   scmd.param1 = 0;
            scmd.param2 = (long)tmpH;
            SndDoCommand(chan,&scmd,FALSE);
            scmd.cmd = callBackCmd;         /* so that pointers are freed */
            SndDoCommand(chan,&scmd,FALSE);
                nxtbit += clen*chans;
        /*    D_printn(cix);    D_prints(" ");  /* */
        /*    SystemTask();     */
            }
        scmd.cmd = nullCmd;
        timeout = 0;
        while(freed < nchunk && ++timeout < TIMEOUTLIM)  {
                if(finished > freed)  { 
                /*      D_printn(-freed);       D_prints(" ");  /* */
                        DisposPtr(hantab[freed++]);     /* semaphore that some sound used up */
                }
                SndDoCommand(chan,&scmd,FALSE);
        }

        SndDisposeChannel(chan, FALSE); /* waits for commands to end */
        if(cix < nchunk)    /* for loop did not complete */
            {   D_prints("Only "); D_printn(cix); D_prints("chunks of "); D_printn(nchunk); }
        for(--cix; cix >= freed; --cix)
            DisposPtr(hantab[cix]); /* get rid of all the handles at the end */
        }
    DisposPtr(hantab);
    }

record(stt,end,chans,dsize,srate,scale) /* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;    /* ignored - assumed = sizeof(short) in what follows */
    double   srate;    /***** NO RESAMPLING if != 8000 Hz - just wrong pitch */
    int     scale;    /* crude integer scale factor; 2 => no scaling */
    {
    D_prints("No record on Mac yet - sorry\n");
    }

#define  TMPSND "tmp.snd"

xtract(stt,end,chans,dsize,srate,scale) /* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;    /* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;    /* crude integer scale factor; 2 => no scaling */
    {
    long    outBSize,outBRem;
    int     format;
    int     c,err;
    long    d;
    short   *ptr1;
    short   *ptr2;
    SFSTRUCT    *tmpSound;
    long    nblocks;
    int     thisBlSiz;
    FILE    *op;

    outBSize = (((long)(end-stt))*sizeof(short));
    /* assu short *stt etc, o/p short */
    format = SFMT_SHORT;
    /*  printf("Allocate %ld bytes, format %d, srat %ld\n",outBSize,
    format, (long)srate);   /*  */
    if(err = SFAlloc(&tmpSound, outBSize, format, srate, chans, 4))
    {
    fprintf(stderr,"%s: sound error %s\n", programName,
        SFErrMsg(err));
    exit(1);
    }
    ptr1 = stt;
    ptr2 = (short *)SFDataLoc(tmpSound);
    while(ptr1<end)     /* copy whole to a new byte sound */
    {
    *ptr2++ = *ptr1++;
    }
    if(err = SFWrite(TMPSND,tmpSound))
    {
    fprintf(stderr,"%s: sound error %s\n", programName,
        SFErrMsg(err));
    exit(1);
    }
    printf("region extracted to %s (%ld bytes)\n",
       TMPSND, outBSize);
    SFFree(tmpSound);
    }


