/***********************************************************************\
*   db.metat.c								*
*   Connect tracks head to tail (by `proximity') to construct 		*
*   `meta-tracks' that look a lot like tracks but are made of several.  *
*   Uses TKGROUP and EPS-CLOSE technology to produce ambiguous results. *
*   1993may06 dpwe							*
\***********************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <genutils.h>		/* for Convo ??? (in DSPbench.h anyway) */
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/
static SELPNL_STR *spnl = NULL;		/* maybe offer user choices */
static cpxpolar *spect = NULL;		/* buffer to read back actual spcts */
static WINDAT *wndw1 = NULL;
static WINDAT *wndw2 = NULL;

/******** prototypes ********/
TKLIST oldtxMakeMeta(TKLIST trx);
	/* Grow into a metatrack */
TGLIST tgMakeMetatrax(TRACK *tk, TKLIST focus);
    /* generate all metatracks starting at tk */

/******* implementations *****/

void DoMetaM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    TGLIST	tails, t2;

    if(selTrax == NULL)	return;		/* no selection - can't go */

    if(wndw1 == NULL) {
	/*  wndw1 = NewWin("noise mag", 0);	
	  wndw2 = NewWin("smoothed", 0);
	  */
    }
    if(spnl == NULL)  {
	spnl = spnlNew("Metatrax");
    } else {
	/* should dispose of all last time's stuff here */
	spnlReset(spnl);
	spnlUnhide(spnl);
    }
/*    tails = tgMakeMetatrax((TRACK *)CAR(selTrax), allTrax);	*/
    tails = lmappend(selTrax, (LNodeFunc)tgl_cont_tk, allTrax);
    t2    = tglRemoveNearDuplicates(tails);
    free_list((LIST)tails);	tails = t2;

    spnl = spnlFromTkgList(spnl, tails, wrcd);
    return;
}

#define FAVG_PTS	6	/* number of pts to avg to get tail frq */

double tkAvgTailFrq(TRACK *trk)
{   /* figure out frequency at the end of this track, but weight it over 
       a few points to give it some stability */
    PKLIST pkl;
    PeakPos *pk;
    long n;
    double linmag, fsum, msum;
    int i = 0;

    /* find node close to end of pklist */
    n = MAX(1, trk->st.npts - FAVG_PTS - 1);	/* -1 skips last -100dB pt */
    pkl = (PKLIST)nth((LIST)trk->tktop, n);
    /* average frq over all remaining pks */
    fsum = msum = 0.0;
    while(pkl)  {
	pk = pkl->peak;
	pkl = CDR(pkl);
	linmag = LINOFDB(pk->mag);
	fsum += pk->frq * linmag;
	msum += linmag;
	++i;	/* just debug */
    }
/*    fprintf(stderr, "txAvFT: %d pts, f = %.1f (~%.1f)\n", 
	    i, fsum/msum, trk->st.favg); */
    return fsum/msum;
}

double tkAvgHeadFrq(TRACK *trk)
{   /* figure out frequency at the *start* of this track, but weight it over 
       a few points to give it some stability */
    PKLIST pkl;
    PeakPos *pk;
    long n;
    double linmag, fsum, msum;
    int i = 0;

    /* start with first node in pklist - except skip -100dB pt at front */
    pkl = (PKLIST)CDR(trk->tktop);
    /* average frq over next FAVG_PTS pks */
    fsum = msum = 0.0;
    while(pkl && i < FAVG_PTS)  {
	pk = pkl->peak;
	pkl = CDR(pkl);
	linmag = LINOFDB(pk->mag);
	fsum += pk->frq * linmag;
	msum += linmag;
	++i;	/* just debug */
    }
/*    fprintf(stderr, "txAvFH: %d pts, f = %.1f (~%.1f)\n", 
	    i, fsum/msum, trk->st.favg); */
    return fsum/msum;
}

double tfDistance(double t1, double f1, double t2, double f2, double fscale)
{	/* metric of distance btwn 2 pts in t-f -- hard!
	   <fscale> controls tradeoff between two dimensions. 
	   = 1 -> one semitone is the same distance as one period 
	   = 1/12 -> a period counts a whole octave (onset tests)  
	   = 4 -> a quarter of a semitone will make up for an extra period */
    /* how to trade time for frequency? */
    double tchr = 1.0/sqrt(f1*f2);	/* chr time is one cycle of avg frq */
    /* trade one cycle in time for one-quarter semitone ??? */
    double semitDiff = fscale * 12.0 * ONEONLOGTWO*log(f2/f1);
    double timeDiff, dist;

    timeDiff = (t2-t1)/(2.0*tchr);	/* (3 tchr is 'added tail') */
    if(timeDiff < 0.0)	/* 2nd pt before 1st is less bad??? */
	timeDiff /= 2.0;
    dist = sqrt(timeDiff*timeDiff + semitDiff*semitDiff);

    return dist;
}

/* have to set parameters for nearness sort as statics, since no 
   way to pass add'l arguments through qsort */
static double tkCNtime = 0.0;
static double tkCNfreq = 0.0;

void txSetSortTargetTF(double time, double freq)
{	/* entry point to setup statics for sorts */
    tkCNtime = time;
    tkCNfreq = freq;
}

/* different `fscale' arguments to tfDist for general dist, or onset dist */
#define FSCALE_NEAR	4.0
#define FSCALE_ONS	(1.0/6.0)

int tkCmpNrstFHead(NODE **pn1, NODE **pn2)
{   /* qsort comparison function:
       Sorts in descending `closeness' to specified t-f point.  
       Calculates frq by looking at head of track. */
    TRACK *tk1, *tk2;
    double d1, d2;

    tk1 = (TRACK *)(*pn1);
    tk2 = (TRACK *)(*pn2);
    /* order of terms in distance metric is signif (hmmm) */
    d1 = tfDistance(tkCNtime, tkCNfreq, tk1->st.tmin, tkAvgHeadFrq(tk1), 
		    FSCALE_NEAR);
    d2 = tfDistance(tkCNtime, tkCNfreq, tk2->st.tmin, tkAvgHeadFrq(tk2), 
		    FSCALE_NEAR);

    if(d1 > d2)  return 1;
    else if(d1 == d2)  return 0;
    else return -1;
}

int tkCmpNrstOnset(NODE **pn1, NODE **pn2)
{   /* qsort comparison function:
       Sorts in descending `onset closeness' - 
       mainly time skew between judged onsets, but frequency distance 
       matters too */
    TRACK *tk1, *tk2;
    double t1, t2, d1, d2, m1, m2, dm;

    tk1 = (TRACK *)(*pn1);
    tk2 = (TRACK *)(*pn2);
/*    t1 = tkOnsetT(tk1);
      t2 = tkOnsetT(tk2);
 */
    t1 = tk1->st.tavg;
    t2 = tk2->st.tavg;
    d1 = tfDistance(tkCNtime, tkCNfreq, t1, tk1->st.favg, FSCALE_ONS);
    d2 = tfDistance(tkCNtime, tkCNfreq, t2, tk2->st.favg, FSCALE_ONS);
    /* weight distances by magnitude so louder track also favoured */
/*    m1 = tk1->st.mavg;
    m2 = tk2->st.mavg;
    dm = sqrt( pow(((m1-m2)/10.0),2) );	/* 10dB diff is worth 1 step ?? * /
    if(m1 > m2)
	d2 = sqrt(d2*d2 + dm);
    else
	d1 = sqrt(d1*d1 + dm);
 */
/*    fprintf(stderr, "CmpOns: t1=%8.6f f1=%6.1f m1=%5.1f d1=%6.4f\n", 
	    t1, tk1->st.favg, m1, d1);
    fprintf(stderr, "      : t2=%8.6f f1=%6.1f m1=%5.1f d1=%6.4f\n", 
	    t2, tk2->st.favg, m2, d2);
 */
    if(d1 > d2)  return 1;
    else if(d1 == d2)  return 0;
    else return -1;
}

static int tk_ons_pred(TRACK *tk, int type, void *arg)
{	/* predicate to allow filtering of track list by onset score.
	   Currently used from db.harmon,c: tkOnsetters */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double d1;

    d1 = tfDistance(tkCNtime, tkCNfreq, tk->st.tavg, tk->st.favg,FSCALE_ONS);
    if(d1 < rg->min || d1 > rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

static int tk_nearhead_pred(TRACK *tk, int type, void *arg)
{	/* predicate to allow filtering of track list by 
	   proximity to preset t,f point of (tmin, *AvHeadF*) */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double d1, tim, frq;

    tim = tk->st.tmin;
    frq = tkAvgHeadFrq(tk);
    d1 = tfDistance(tkCNtime, tkCNfreq, tim, frq, FSCALE_NEAR);
    if(d1 < rg->min || d1 > rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

static int tk_nearavg_pred(TRACK *tk, int type, void *arg)
{	/* predicate to allow filtering of track list by 
	   proximity to preset t,f point of (tmin, *favg*) */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double d1, tim, frq;

    tim = tk->st.tmin;
    frq = tk->st.favg;
    d1 = tfDistance(tkCNtime, tkCNfreq, tim, frq, FSCALE_NEAR);
    if(d1 < rg->min || d1 > rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

static int tk_print_dist(TRACK *trk, int type, void *arg)
{	/* print out track and its onset distance */
    fprintf(stderr, "trk%6ld: ot=%9.6f af=%7.1f am=%6.2f dist=%6.3f\n", 
	    trk->id, trk->st.tavg, trk->st.favg, trk->st.mavg, 
	    tfDistance(tkCNtime, tkCNfreq, trk->st.tavg, trk->st.favg, 
		       FSCALE_ONS));
}

void tklPrintOnsDists(TKLIST trx, char *title)
{	/* print out a list of trax with their onset distances */
    fprintf(stderr, "tklPrintOnsDists: '%s':\n", title);
    map((LIST)trx, (NodeFunc)tk_print_dist);
}

int tkCmpNrstFAvg(NODE **pn1, NODE **pn2)
{   /* qsort comparison function:
       Sorts in descending `closeness' to specified t-f point.  
       Uses 'average frq' from header. */
    TRACK *tk1, *tk2;
    double d1, d2;

    tk1 = (TRACK *)(*pn1);
    tk2 = (TRACK *)(*pn2);
    d1 = tfDistance(tkCNtime, tkCNfreq, tk1->st.tmin,tk1->st.favg,FSCALE_NEAR);
    d2 = tfDistance(tkCNtime, tkCNfreq, tk2->st.tmin,tk2->st.favg,FSCALE_NEAR);

    if(d1 > d2)  return 1;
    else if(d1 == d2)  return 0;
    else return -1;
}

LIST strip_entry_ptr(LIST *plist, LIST unwantedCar)
{	/* go through <*plist> looking for the unwantedCar;
	   remove it wherever it occurs.  Will remove multiple entries.
	   Modifies actual list and returns it too. */
    LIST doomed;
    LIST *entryHead = plist;	/* to read back answer */

    while(*plist) {
	if( CAR(*plist) == unwantedCar )  {
	    free_next_ptr(plist);
	} else {
	    /* no need to advance if we skipped a node */
	    advance_ptr(&plist);
	}
    }
    return *entryHead;
}
	    

#define TK_SUCC_THRSH	5.0	/* max distance (in tfDist units) for succs */

#define SUCC_TAIL_TWIN	(20.0)	/* cycles of tail inspected in track */
#define SUCC_TSRCH_BACK	(3.0)	/* search for overlap this far back in time */
#define SUCC_TSRCH_FWD	(10.0)	/* .. and this far fwd (in cycles) */
#define SUCC_FRQ_FACT	(2.0)	/* scale frq window by this */
#define SUCC_MIN_FRATIO	(1.2)	/* Make frq win at least *, / this */

TKLIST tkSuccessors(TRACK *trk, TKLIST focus, TKLIST withouts)
{	/* return a list of tracks that pass the threshold to serve as 
	   the continuation of the passed-in track.  Probably have 
	   to test for proximity both of HeadFrq and favg */
    TKLIST cands, cands2;
    double time, freq, fsd, fmin, fmax, tchr, tmin, tmax;
    RANGE_STR	range, mrg, frg;

    if(focus == NULL)	focus = allTrax;

    fsd  = trk->st.fsd;
    time = trk->tkbot->peak->tim;
    freq = tkAvgTailFrq(trk);
    tchr = 1.0/freq;
    
    range.max = time;	range.min = time - SUCC_TAIL_TWIN*tchr;
    tkFrqMagRange(trk, &range, &frg, &mrg);
    fmin = frg.min;	fmax = frg.max;
    freq = sqrt(fmin * fmax);	/* geom mean */
    fsd = fmax/freq;		/* ratio */
    fsd = pow(fsd, SUCC_FRQ_FACT);	/* widen ratio a little */
    fsd = MAX(SUCC_MIN_FRATIO, fsd);	/* not narrower than .. */
/*    fprintf(stderr, "tkSuccs: frg %.1f..%.1f ->", fmin, fmax);	*/
    fmin = freq/fsd;	fmax = freq*fsd;
    tmin = time - SUCC_TSRCH_BACK * tchr;
    tmax = time + SUCC_TSRCH_FWD  * tchr;
    cands= tkGetIntersect(focus, tmin, tmax, fmin, fmax);
    cands2 = tkGetNarrowIntersect(cands, tmin, tmax, fmin, fmax);
    free_list(cands);
    cands = cands2;
    /* remove the source track from the list */
    strip_entry_ptr(&cands, (LIST)trk);
    /* remove unavailable tracks */
    cands2 = remove_list(cands, withouts);
    free_list(cands);
    cands = cands2;

/*     fprintf(stderr, " %.1f..%.1f (%d cands)\n", fmin, fmax, 
	    top_size((LIST)cands));	*/
/*    RedrawTrax(cands, TKM_SPECIAL);
    FlushDisplay();	/* make sure RedrawnTrax show */

/*    txSetSortTargetTF(time, freq);
    range.min = -1;
    range.max = TK_SUCC_THRSH;
    cands2 = keep_if(cands, (NodeFunc)tk_nearhead_pred, &range);
    free_list(cands);
    cands = cands2;
*/
    return cands;
}

void CheckSuccsM(WINRCD **wrcd)
{   /* just allow us to run the track-successor generator interactively */
    TRACK *tk;
    TKLIST newTrax;

    if(selTrax)  {
	tk = (TRACK *)CAR((LIST)selTrax);
	newTrax = tkSuccessors(tk, NULL, NULL);
	RedrawTrax(newTrax, TKM_SPECIAL);
	free_list(newTrax);
    }
}

static TRACK *oldtkSuccessor(TRACK *trk)
{	/* find a track (if any) suitable to serve as the continuation 
	   of the one passed in a meta-track 
	   (should maybe have way to pass in meta-track params?) */
    TRACK *succ1 = NULL, *succ2;
    TKLIST cands;
    double time, freq, fsd, tchr, dist1 = BIGMAX, dist2 = BIGMAX;

    fsd  = trk->st.fsd;
    time = trk->tkbot->peak->tim;

/*    freq = trk->st.favg;	*/
    freq = tkAvgTailFrq(trk);
    tchr = 3.0 / freq;	/* characteristic time is three cycles */
    cands = tkGetIntersect(allTrax, time - 5.*tchr, time + 5.*tchr, 
			   freq-fsd, freq+fsd);
    
    if(cands)  {
	RedrawTrax(cands, TKM_SPECIAL);
	tkCNtime = time;
	tkCNfreq = freq;
	cands = qsort_list((LIST)cands, tkCmpNrstFHead);
	fprintf(stderr, "tkSucc1: end at %.6f, %.0f\n", time, freq);
	tklPrintOnsets(cands);

	succ1 = (TRACK *)CAR(cands);
	if(succ1 == trk)  {	/* found self!  take next choice */
	    if(CDR(cands))  {
		succ1 = (TRACK *)CAR(CDR(cands));
	    } else {
		succ1 = NULL;	/* no others - fail */
	    }
	}
	free_list((LIST)cands);	/* finished with list of cands */

	if(succ1)  {
	    dist1 = tfDistance(time, freq, 
			       succ1->st.tmin,tkAvgHeadFrq(succ1),FSCALE_NEAR);
	    if(dist1 > TK_SUCC_THRSH)
		succ1 = NULL;	/* don't return it */
	}
    }
 
   /*** Now try whole match again based on favg, not fHead */
    freq = trk->st.favg;
    tchr = 3.0 / freq;	/* characteristic time is three cycles */
    cands = tkGetIntersect(allTrax, time - 5.*tchr, time + 5.*tchr, 
			   freq-fsd, freq+fsd);
    if(cands)  {
	RedrawTrax(cands, TKM_SPECIAL);
	tkCNtime = time;
	tkCNfreq = freq;
	cands = qsort_list((LIST)cands, tkCmpNrstFAvg);
	fprintf(stderr, "tkSucc2: end at %.6f, %.0f\n", time, freq);
	tklPrintOnsets(cands);

	succ2 = (TRACK *)CAR(cands);
	if(succ2 == trk)  {	/* found self!  take next choice */
	    if(CDR(cands))  {
		succ2 = (TRACK *)CAR(CDR(cands));
	    } else {
		succ2 = NULL;	/* no others - fail */
	    }
	}
	free_list((LIST)cands);		/* finished with list of cands */

	if(succ2)  {
	    dist2 = tfDistance(time, freq, 
			       succ2->st.tmin, succ2->st.favg,FSCALE_NEAR);
	    if(dist2 > TK_SUCC_THRSH)
		succ2 = NULL;	/* don't return it */
	}
    }
    if(dist2 < dist1)  {
	succ1 = succ2;
	fprintf(stderr, "favg better match than fhead (%.3f<%.3f)\n", 
		dist2, dist1);
    } else {
	fprintf(stderr, "fhead kept as better (%.3f<%.3f)\n", 
		dist1, dist2);
    }
    return succ1;
}

static TKLIST oldtxMakeMeta(TKLIST trx)
{	/* Grow into a metatrack */
    TRACK *trk;
    TKLIST rslt, *rsTail;

    if(!trx || !(trk = ((TRACK *)CAR(trx))))
	return NULL;

    rsTail = &rslt;
    while(trk)  {
	*rsTail = (TKLIST)new_node(NULL, (LIST)trk, T_TRACK);
	advance_ptr((LIST **)&rsTail);
	trk = oldtkSuccessor(trk);
    }
    return rslt;
}

static TKGROUP *tg_extend_cont(TRACK *tk, int *ptype, TKGROUP *tg)
{	/* LNodeFunc spawns a new TKGRP by extending an existing one */
    TKGROUP *rslt = tkgrDup(tg);
    double  dur = tk->st.tmax - tk->st.tmin;
    double  twt;

    assert(*ptype == T_TRACK);
    tkgrAddTrack(rslt, tk);
    rslt->v.cont.latest = tk;
    twt = LINOFDB(tg->v.cont.magav) * tg->v.cont.ttot 
	   + dur * LINOFDB(tk->st.mavg);
    rslt->v.cont.ttot  = tg->v.cont.ttot + dur;
    rslt->v.cont.magav = DBOFLIN(twt / rslt->v.cont.ttot);
    *ptype = T_TKGROUP;	/* type for return */
    return rslt;
}

#define CONT_MAG_LIM	(-10.0)	
/* don't extend with tracks much quieter than average mag so far */

TGLIST tgMetaExtend(TKGROUP *tg, TKLIST focus, TKLIST *pwithouts)
{	/* find all the candidate extensions, make them all into 
	   new tgs */
    TRACK *last = tg->v.cont.latest;
    TKLIST nexts, n2;
    TGLIST rslt;
    int    npre, npost;
/*    RANGE_STR range;	*/

    assert(tg->type == TG_CONT);
    nexts = tkSuccessors(last, focus, *pwithouts);
    npre = top_size(nexts);
/*
    range.min = tg->v.cont.magav + CONT_MAG_LIM;
    range.max = BIGMAX;
    n2 = keep_if(nexts, (NodeFunc)tk_magavg_pred, &range);
 */
    n2 = SelTrxByAvMag(nexts, tg->v.cont.magav + CONT_MAG_LIM, BIGMAX);
    free_list(nexts);
    nexts = n2;
    if( (npost = top_size(nexts)) != npre)  {
/*	fprintf(stderr, "tg magav %.2f cut from %d to %d\n", 
		tg->v.cont.magav, npre, npost);   */
    }
    *(find_last_ptr(pwithouts)) = nexts;	/* prevent reuse */
    rslt = (TGLIST)lmaparg((LIST)nexts, (LNodeFunc)tg_extend_cont, tg);
    return rslt;
}

static TGLIST tgContClose(TGLIST tgls, TKLIST focus, TKLIST *pwithouts)
{	/* basically epsilon closure of the TGLIST under the 
	   tkSuccessors function - based on Onset's breadth_close */
    TGLIST 	this, new, rslt = NULL, *rsPtr;

    rsPtr  = &rslt;
    while(tgls)  {
	this = extract_next_ptr(&tgls);
	*rsPtr = this;
	new = tgMetaExtend((TKGROUP *)CAR((LIST)this), focus, pwithouts);
	if(new == NULL)  {	/* no progeny - keep parent */
	    advance_ptr(&rsPtr);
	} else {		/* progeny replace parent */
	    free_list(*rsPtr);
	    *rsPtr = NULL;
	    /* prune ?  Probably want to remove items from `new' 
	       if they have added just tracks we already have - 
	       in results?  in unopened nodes? */
	    /* add new nodes to bottom of list (breadth-first) */
	    *(find_last_ptr((LIST *)&tgls)) = new;
	}
    }
    return rslt;
}

TGLIST tgMakeMetatrax(TRACK *tk, TKLIST focus)
{	/* Grow a list of `continuity' metatracks from a single seed track */
    TKLIST	tklst, withouts;
    TKGROUP	*tg;
    TGLIST	start, rslt, rs2;

    tklst = (TKLIST)new_node(NULL, (LIST)tk, T_TRACK);
    tg    = tkgrNew(tklst, TG_CONT);	/* new continuity group */
    tg->v.cont.latest = tk;		/* set `latest' field */
    tg->v.cont.magav  = tk->st.mavg;
    tg->v.cont.ttot   = tk->st.tmax - tk->st.tmin;
    start = (TGLIST)new_node(NULL, (LIST)tg, T_TKGROUP);
    withouts = copy_top(tklst);
    rslt  = tgContClose(start, focus, &withouts);	/* do closure */
    rs2   = tglRemoveNearDuplicates(rslt);
    free_list((LIST)rslt);
    rslt  = rs2;
    free_list(withouts);

    return rslt;
}

TGLIST tgl_cont_tk(TRACK *trk, int *ptype, TKLIST focus)
{	/* LNodeFunc to generate metatracks from tracks */
    assert(*ptype == T_TRACK);
    *ptype = T_LIST;	/* returning a TGLIST */
    return tgMakeMetatrax(trk, focus);
}

#else /* !CQFILE */   

void DoMetaM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    /* dummy */
}

#endif /* CQFILE */   

#endif /* TRAX */   
