/********************************************************/
/* db.muplay.c						*/
/* play routines for SPARC-style 8k mulaw speaker	*/
/* dpwe 27aug91 after smplay.c				*/
/********************************************************/

#include <stdio.h>
#include "DSPbench.h"

/* #include <stdlib.h>	*/
/* #include <sndf.h>	in DSPbench.h */
#include <math.h>	/* for fabs() prototype */

#define PLAYBSIZE 4096		/* (samples) loaded off disk at once */
#define CHUNKSIZE 1024		/* pass to play at once */
#define XTRA 1024
#define OUTPUT_SRATE	8000	/* sampling rate of output */
#define OUTPUT_DEVICE   "/dev/audio"

#ifndef ANSIFILES
#include <sys/types.h>
#include <sys/file.h>
#include <unistd.h>
#define STDIN  0
#define STDOUT 1
#endif

/* static function prototype */
#ifdef __STDC__
static void ConvertFedSound(int inChs, FLOATARG inSR, int scale,
			    long (*grabFn)(),
			    SFSTRUCT *outsnd, FILE *outfile
#ifndef ANSIFILES
			     , int oFd
#endif
			                        );
static void ResetGetSamples(short *ptr, long len);
static long GetSamples(short *where, int size, long num);
static void MyConvertToMono(short *sce, size_t samples, short *dst, int chs);
void FindBestRatio(FLOATARG ratio, int *pi, int *po);
static void ResampInteg(short *inSnd, short *outSnd, 
			int inFrq,int outFrq,long inLen);
static short *SetupResInt(int inFrq, int outFrq, int scale);
static long DoPreResInt(short *inSnd, int inInc, long inLen,
			short *outSnd, int outInc, long outLen, 
			long skew, short *sncTab);
#else /* no prototypes */
static void ConvertFedSound();
static void ResetGetSamples();
static long GetSamples();
static void MyConvertToMono();
void FindBestRatio();
static void ResampInteg();
static short *SetupResInt();
static long DoPreResInt();
#endif /* __STDC__ */

/* static data */

static     SFSTRUCT	*inSnd, *outSnd;
static	   short 	*playbuf;
static     FILE 	*outFile;
#ifndef ANSIFILES
static     int	 	oFd;
#endif
static	   double	output_srate = OUTPUT_SRATE;

muplay(stt,end,chans,dsize,srate,scale)
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;    /* ignored - assumed = sizeof(short) in what follows */
    FLOATARG   srate;
    int     scale;    /* crude integer scale factor; 2 => no scaling */
    {
    int 	i;
    int 	err;

    if(err=SFAlloc(&outSnd, 0L, SFMT_ULAW, 
		   output_srate, SF_MONO, SFDFLTBYTS))
	SFDie(err, NULL);
    outSnd->dataBsize = SF_UNK_LEN;	/* strictly, not 0 */

#ifdef ANSIFILES
	outFile = fopen(OUTPUT_DEVICE, "w");
#else
	oFd = open(OUTPUT_DEVICE, O_WRONLY, 0);
	outFile = fdopen(oFd, "w");
#endif
    if( outFile == NULL )
	{ fprintf(stderr, "%s: couldn't open %s for output\n",
		  programName, OUTPUT_DEVICE);
	  exit(1); }

    /*** debug - actually spout the header too */
/*    if(err=SFWriteHdr(outFile, outSnd, 0))
	{ fprintf(stderr, "%s: snd err %s\n",
		  programName, SFErrMsg(err));
	  exit(1);  }
 */
    ResetGetSamples(stt, end-stt);
    ConvertFedSound(chans, srate, scale, GetSamples, outSnd, outFile
#ifndef ANSIFILES
		     , oFd
#endif
		     );
    fclose(outFile);
    }

#define SINC_PDS	1	/* could go no higher on SPARCstation 2 */

static void ConvertFedSound(inChs, inSR, scale, grabFn, outsnd, outfile
#ifndef ANSIFILES
			    , oFd
#endif ANSIFILES
			    )
/*    SFSTRUCT *insnd;	*/
    int inChs;		/* number of input channels */
    FLOATARG inSR;	/* input sampling rate */
    long (*grabFn)();
    SFSTRUCT *outsnd;
    FILE *outfile;
#ifndef ANSIFILES
    int oFd;
#endif
    {
    int 	not_done = 1;
    size_t	read_in;
    size_t	bufsize, keep_old, num_to_read, new_out, src_smps, used_up;
    size_t	if_klen;
    long	if_index;
    short	*buffer, *where_to_read, *dst_buf, *r_tab;
    int 	r_i, r_o;	/* ratio in terms of frq : expect r_i bigger */
    int 	inPrd, outPrd;
/*    int 	inChs = insnd->channels;	*/
/*    FLOATARG  inSR  = insnd->samplingRate;	*/
    r_i = 100;		/* pass maximum denominator to search over */
    FindBestRatio(outsnd->samplingRate/inSR, &r_i, &r_o);
    fprintf(stderr, "%s: ratio is %d:%d \n", programName, r_i, r_o);
    inPrd = r_o;	outPrd = r_i;

    bufsize = PLAYBSIZE;
    buffer = (short *)malloc(bufsize*sizeof(short)*inChs);
    dst_buf = (short *)malloc(sizeof(short)*
			      (size_t)((bufsize*(long)r_o)/(long)r_i));
    r_tab = SetupResInt(r_i,r_o,scale);
    if_klen = SINC_PDS * ((r_i > r_o)?r_i:r_o);

    if_index = 0;		/* initially no skew between I & O */
    keep_old = 0;		/* initially no samples in buffer */
    while(not_done)
	{
	CopyShorts(buffer + bufsize - keep_old, 1, buffer, 1, keep_old, 0);
        num_to_read = (bufsize - keep_old);
	where_to_read = buffer + keep_old;
	read_in = (*grabFn)(where_to_read, sizeof(short)*inChs, num_to_read);
	not_done = (read_in == num_to_read);
	if(inChs != 1)
	    MyConvertToMono(where_to_read, num_to_read, where_to_read, inChs);
	new_out = ((read_in + keep_old)*inPrd 
		   - if_index - if_klen + outPrd - 1 )/outPrd;
	src_smps = keep_old + read_in;

	if_index = DoPreResInt(buffer, 1, src_smps, 
			       dst_buf, 1, new_out, 
			       if_index, r_tab);
	used_up = (if_index - if_klen + inPrd - 1 )/inPrd;
	if_index = if_index - used_up * inPrd;
	keep_old = src_smps - used_up;

	Short2Ulaw(dst_buf, (char *)dst_buf, 1, 1, new_out);

#ifdef ANSIFILES
	fwrite(dst_buf, sizeof(char), new_out, outfile);
#else
	write(oFd, dst_buf, new_out);
#endif
	}
    free(buffer);
    free(dst_buf);
    }

static short *sndPtr = NULL;
static long  sndIdx  = 0;
static long  sndLen  = 0;

static void ResetGetSamples(ptr, len)
    short *ptr;
    long  len;
    {
    sndIdx = 0;
    sndPtr = ptr;
    sndLen = len;
    }

static long GetSamples(where, size, num)
    short *where;
    int size;	
    long num;
    {
    short *shsrc;

    shsrc = sndPtr + sndIdx;
    num = (num*size)/sizeof(short);	/* how many shorts to transfer */

    if(num > (sndLen-sndIdx))	num = sndLen-sndIdx;	/* hit end */
    if(num < 0)		        num = 0;
    CopyShorts(shsrc, 1, where, 1, num, 0);
    sndIdx += num;
    return((num*sizeof(short))/size);
    }

static void MyConvertToMono(sce, samples, dst, chs)
    short *sce;
    size_t samples;
    short *dst;
    int chs;
    {
    int a;
    long l;
    int p = (chs>2)?2:1;

    while(--samples)
	{
	l = 0; a = chs;
	while(a--)
	    l += *sce++;
	*dst++ = l>>p;
	}
    }

void FindBestRatio(ratio, p_i, p_o)
    FLOATARG ratio;
    int *p_i;
    int *p_o;
    /* find an integer ratio *p_o/*p_i closest to 'ratio', with *p_i
       smaller than its entry value.  Returns immediately if error
       is less than .1%		*/
    {
    int i,max, besti = 0;
    double fo,err, besterr = 1;

    max = *p_i;		/* entry value of i is maximum */
    i = 0;
    while(besterr > .001 && i < max)	/* try every denominator */
	{
	++i;
	fo = ratio * (double)i;
	err = fabs( (fo - (double)((int)(fo+.5)) )/fo );
	if(err < besterr)
	    {
	    besterr = err;
	    besti = i;		/* remember this one, it's good */
	    }
	}
    *p_i = besti;
    *p_o = (int)( (ratio * (double)besti)+.5);
    }

/******************************************************/
/*  downsamp.c                                        */
/*  Performs sample rate conversion by FIR LPF approx */
/*  1620 06dec89 started dpwe                         */
/******************************************************/

/* #include <stdio.h>
   #include <math.h>  */
#define		UNIT	(0x7FFF)    /* Conversion factor for 16 bit fxd pt */

#define		SPDS	SINC_PDS    /* How many sinc lobes to go out */

static void 
ResampInteg(inSnd, outSnd, inFrq, outFrq, inLen)
    short *inSnd;
    short *outSnd;
    int inFrq;
    int outFrq;
    long inLen;
  /* Perform the sample rate conversion: 
		inSnd   is the existing sample to be converted
		outSnd  is a pointer to the (pre-allocated) new soundspace
		inFrq   is the lowest common factor of the input's sample rate
		outFrq  is the factor in the same terms of desired output srate
		inLen	is the number of points in the input sequence
	i.e. if the input sample is at 12kHz and we want to produce an
	8kHz sample, inFrq = 3 and outFrq = 2.  inFrq = 12 and outFrq = 8
	will give identical results more slowly.  inFrq and outFrq are in
	proportion to the bytelength of the respective samples.
!! Should really allocate new soundspace ourselves as we know the length!
   */
/*
	!  !  !  !  !  !  !  !  !  	    input samples at inFrq (5kHz)
	|||||||||||||||||||||||||| time->   interpolation frq (GCM) ("I.F.")
	!    !    !    !    !    !	    output samples at outFrq (3kHz)

	We use inFrq and outFrq, but really they are equivalent to
	outPrd and inPrd equivalently, as the above diagram shows
	*/
    {
    short	*sncTab;	/* A table of sin(x)/x coefficients */

    sncTab = SetupResInt(inFrq,outFrq,2);
    DoPreResInt(inSnd,1,inLen,outSnd,1,0L,0L,sncTab);
    free(sncTab);
    }

static    int 	inPrd, outPrd, in2out;

static short *SetupResInt(inFrq, outFrq, scale)
    int inFrq;
    int outFrq;
    int scale;	/* sick dspB scale : 2 == unity */
    {
    short	*sncTab;	/* A table of sin(x)/x coefficients */
    int 	stLen;
    int 	lpfPrd;		/* LPfilter is sinc(x) over this period */
    int 	i;
    double	theta,dtheta; 

    inPrd = outFrq;	outPrd = inFrq;	/* in units of GCM periods .. . */
	
    lpfPrd = (inPrd > outPrd)? inPrd : outPrd;	
    /* LPF has max frq as min of two frqs concerned i.e. max period */
    /* Length of sncTab is dictated by number of sinc cycles & frq ratio */
    stLen = SPDS * lpfPrd;	/* sinc table is SPDS/2 periods of sinc */
    sncTab = (short *)malloc(sizeof(int)*(stLen+1) );
    theta = 0.0; dtheta = ((double)M_PI)/((double)lpfPrd);
    sncTab[0] = (short)(((long)UNIT)*scale*((long)inPrd)/((long)lpfPrd)/2L);
    /* A unit impulse on the input lasts inPrd, output wiggle lasts lpfPrd */
    /* Integral conservation scaling factor */
    for(i=1; i<=stLen; ++i)	/* build table of sin x / x */
	{
	theta += dtheta;
	sncTab[i] = (int)( (((double)sncTab[0]) * sin(theta)/theta) 
	      * (.54+.46*cos(((double)M_PI)*((double)i)/((double)stLen))) );
	/*	printf("<%d>", sncTab[i]);	/* */
	/* hamming window on top of sinc */
	}
    in2out = stLen/inPrd;
    /* the number of input points sampled per output point (stLen mask) */
    return sncTab;
    }

static long DoPreResInt(inSnd, inInc, inLen, outSnd, outInc, outLen,
			skew, sncTab)
    short *inSnd;
    int inInc;
    long inLen;
    short *outSnd;
    int outInc;
    long outLen;
    long skew;
    short *sncTab;
    {
    long	i;
    long 	phase,nrstIp,x;
    int 	ifIndx;		/* current point's index in I.F. clix */
    int 	j;
    long	a;

    ifIndx = skew;		/* in units of IF, extra pre-ring in inSnd */
    if(outLen == 0L)		/* null input means as long as you can */
	outLen = (inLen*(long)inPrd - skew + outPrd - 1)/(long)outPrd;

    for(i = 0; i<outLen; ++i)	/* output sample loop */
	{
	phase  = ifIndx % inPrd; /* phase difference from imm.prec */
	nrstIp = ifIndx/inPrd;	/* direct index of immed'ly preceding */
	a = ((long)sncTab[phase])*((long)inSnd[inInc*nrstIp]);
	for(j=1; j<in2out; ++j)	/* inner FIR convolution loop */
	    {
	    if( (x = nrstIp-j)>=0 )
		a += ((long)sncTab[j*inPrd+phase]) * ((long)inSnd[inInc*x]);
	    if( (x = nrstIp+j)<inLen )
		a += ((long)sncTab[j*inPrd-phase]) * ((long)inSnd[inInc*x]);
	    }
	a /= (long)UNIT;
	
	if(a>0)	{  if(a>32767)	  a=32767;	} /* flag clip? */
	else	{  if(a<-32767)   a= -32767;	}
	outSnd[outInc * i] = (short)a;
	ifIndx += outPrd;	/* step the high-frq index by inFrq clx */
	}
    return ifIndx;		/* pass back phase perhaps for next call */
    }
