/***********************************************************************\
*   db.noisy.c								*
*   Trying to organize `noise' regions (as a preliminary grouping)	*
*    - take a set of tracks and some seed instant			*
*    - for each octave (some fixed wide f window)			*
*      + are there more than a certain number of tracks? DENSITY check	*
*      (DENSITY could be TOTAL DURATION of tracks WITHIN t/f wdw)	*
*    - if so, calc avg (median) magnitude for them			*
*    - select all tracks ABOVE - 10 dB (or..) of avg HOMOGEN check	*
*    - if still passes DENSITY check, have a noise SLICE		*
*    - first, grow in FRQ - stretch by 1/2 octs?			*
*			  - find pts where actual spect falls below avg *
*    - then grow in TIME: re-find noise some small step away		*
*      apply frq/mag continuity constraints 				*
*   1993apr25 dpwe							*
\***********************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <genutils.h>		/* for Convo ??? (in DSPbench.h anyway) */
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/
static SELPNL_STR *spnl = NULL;		/* maybe offer user choices */
static cpxpolar *spect = NULL;		/* buffer to read back actual spcts */
static cpxpolar *spec2 = NULL;
static cpxpolar *spec3 = NULL;
static cpxpolar *spec4 = NULL;
static double   spectime = BIGMIN;	/* time of spectrum held in spect
                                           - allows it to be reused for eff */
static CQTINFO 	*cqi = NULL;
static WINDAT *wndw1 = NULL;
static WINDAT *wndw2 = NULL;

static int stDispQ = 0;			/* HACK - static flag makes 
					   nsNoiseAt do X display */

/******** prototypes ********/
TKLIST txNoisyObjects(TKLIST trx);
long FrqToBin(double frq);

/******* implementations *****/

int nsGetMinMaxFrq(double *minfrq, double *maxfrq)
{	/* peek into the static CQI header to get the min and 
	   max frequencies */

    if(cqi)  {
	*minfrq = cqi->pvh->minFreq;
	*maxfrq = cqi->pvh->maxFreq;
	return 1;
    }
    else
	return 0;
}

int noisyInitStatics(WINRCD **wrcd, int spnlsq)
{	/* initialize the static data used in this file (esp cqi).
	   Optionally create the selpanels too.  */
    long totBins;

    cqi = (*wrcd)->w.snd.cqi;		/* load up our static global */
    if(cqi == NULL)	return 0;		/* no spectra loaded */

    totBins  = cqi->nBins * cqi->nOcts;
    if(spect == NULL) {
	spect = TMMALLOC(cpxpolar, totBins, "FindNoise:spect");
	spec2 = TMMALLOC(cpxpolar, totBins, "FindNoise:spec2");
	spec3 = TMMALLOC(cpxpolar, totBins, "FindNoise:spec3");
	spec4 = TMMALLOC(cpxpolar, totBins, "FindNoise:spec4");
    }
    if(spnlsq == 1)  {
	if(wndw1 == NULL) {
	    wndw1 = NewWin("noise mag", 0);	
	}
	if(wndw2 == NULL) {
	    wndw2 = NewWin("smoothed", 0);
	}
	if(spnl == NULL)  {
	    spnl = spnlNew("Noisy");	
	}
    } else if(spnlsq == 2)  {
	if(wndw1 == NULL) {
	    wndw1 = NewWin("noise mag", 0);
	}
    }
    return 1;	/* ok */
}

void FindNoiseM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    TGLIST	tails, t2;

    noisyInitStatics(wrcd, 1);
    if(spnl != NULL)  {
	/* should dispose of all last time's stuff here */
	spnlReset(spnl);
	spnlUnhide(spnl);
    }
/*    tails = lmappend(selTrax, (LNodeFunc)tgl_noise_tk, allTrax);
    t2 = tglRemoveNearDuplicates(tails);
    fprintf(stderr, "noise tgs: %d->%d\n", top_size(tails), top_size(t2));
    free_list(tails);	tails = t2;
*/
    if(top_size(selTrax) == 1)
	stDispQ = 1;	/* display if only 1 track */

    tails = tklNoiseGroups(selTrax, allTrax);

    stDispQ = 0;

    spnl = spnlFromTkgList(spnl, tails, wrcd);
    return;
}

#define PROM_LOOKBACK_CYCS	(5.0)	/* sample spec 5 cycs before onset */
#define PROM_OCTAVES		(0.7)	/* look over 2/3 of an octave */

double tkProminence(TRACK *trk, CQTINFO *mycqi)
{	/* calculate the 'prominence' (onset headroom) of passed 
	   track by looking at raw spectrum in vicinity */
    double	dt = 0.00001;	 /* ? */
    double	onsTime, sampTime, tchr;
    double	frq, mag, premag;
    long	totBins;
    long	ctrbin, srchwidth, minbin, maxbin;
    char 	timestr[32];
    
    if(mycqi == NULL)
	mycqi = cqi;	/* hope static is set */
    assert(mycqi);
    totBins = mycqi->nBins * mycqi->nOcts;
    onsTime = tkOnsetTFM(trk, &frq, &mag);
    tchr = 1.0/frq;
    sampTime = onsTime - tchr * PROM_LOOKBACK_CYCS;

    if(spectime < sampTime - tchr || spectime > sampTime + tchr)  {
	/* exisiting spectrum is for a significantly diff. time */
	/* read back the original analysis spectrum for this time */
	SimpleSpecOfCQTFile(mycqi, sampTime, dt, spect);
	spectime = sampTime;	/* record origin of spectrum */
	Vlin2DB((float *)spect, 2, (float *)spect, 2, (double)1, totBins);
    }

/*    sprintf(timestr, "%.6f", onsTime);
    CopyAndDisp((float *)spect, 2, totBins, wndw1, timestr, 0, 1.0);
    SetXLimits(wndw1, mycqi->pvh->minFreq , mycqi->pvh->maxFreq);
    SetLogScale(wndw1, TRUE);
 */
    ctrbin = FrqToBin(frq);
    srchwidth = PROM_OCTAVES * mycqi->nBins;	/* total # bins to inspect */
    minbin = MAX(0, ctrbin - srchwidth/2);
    maxbin = MIN(totBins, 1+ ctrbin + srchwidth/2);
    if(minbin>=maxbin)	return -0.0999;	/*? trk outside spect? */
    premag = VMax((float *)&(spect[minbin].mag), 2, maxbin-minbin);

/*    fprintf(stderr, "tkProm:tk%d onset %.6f/%.1f/%.2f vs %.1f..%.1f/%.2f\n",
	    trk->id, onsTime, frq, mag, BinToFrq(minbin), BinToFrq(maxbin), 
	    premag);
 */
    return mag - premag;
}

#define PROM_THRESH	(3.0)	/* prominence >= this to be selected */

static int tk_prom_pred(TRACK *tk, int type, void *arg)
{	/* predicate to allow filtering of track list by (newstyle) prom'ce */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double prom = tkProminence(tk, NULL);

/*    fprintf(stderr, "tk%d (%.6f/%.1f) prom = %.2f\n", 
	    tk->id, tk->st.tmin, tk->st.favg, prom); 	*/
    if(prom < rg->min || prom >= rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

TKLIST tklSelAbvProm(TKLIST src, double promthresh)
{	/* return only those tracks exceeding the prominence threshold */
    TKLIST newTx;
    RANGE_STR rg;

    rg.min = promthresh;	rg.max = BIGMAX;
    newTx = (TKLIST)keep_if((LIST)src, (NodeFunc)tk_prom_pred, &rg);
    return newTx;
}

void CalcPromM(WINRCD **wrcd)
{	/* check out the new prominence calculation */
    double prom;
    TRACK *tk;
    TKLIST newTx;

    if(selTrax)  {
	noisyInitStatics(wrcd, 0);
/*	tk = (TRACK *)CAR(selTrax);
	prom = tkProminence(tk, cqi);
	fprintf(stderr, "tk%d (%.6f/%.1f) prom = %.2f\n", 
		tk->id, tk->st.tmin, tk->st.favg, prom);
 */
	newTx = tklSelAbvProm(selTrax, PROM_THRESH);
	trxNewSelect(newTx);
    }
}

int extendWhileWithin(float *data, long step, double min, double max, long lm, 
		      long *extent)
{	/* return largest n s.t. min <= data[step*m] < max 
	   for all m = 0..n-1.  Return 0 if !(min <= data[0] < max).
	   <step> can be +ve or -ve.  returned <n> always >= 0 */
    long n = 0;

    assert(step != 0);
    assert(min < max);
    while(min <= data[step*n] && data[step*n] < max && n < lm) 
	++n;
    *extent = n;
    if(n == lm)		return 0;	/* hit end before leaving rg */
    else {
	if(data[step*n] < min)		return -1;	/* left bottom of rg */
	else if(data[step*n] >= max)	return 1;
	else 	abort();	/* huh? */
    }
}

long FrqToBin(double frq)
{	/* convert a frequency value into a spectral bin in a CQI context */
    long bin;
    long   binsPOct = cqi->nBins;
    double minFrq = cqi->pvh->minFreq;

    bin = (long)( 0.5 + log(frq/minFrq)*ONEONLOGTWO*(double)binsPOct);
    return bin;
}

double BinToFrq(long bin)
{	/* convert a spectral bin into a frequency value in a CQI context */
    double frq;
    long   binsPOct = cqi->nBins;
    double minFrq = cqi->pvh->minFreq;

    frq = exp(LOGTWO * (double)bin/(double)binsPOct)*minFrq;
    return frq;
}

void SetHamming(float *dst, long dinc, long len, double scale)
{	/* write a full hamming window of length <len> into 
	   pre-allocated buffer <dst>, stepping by <dinc>.  Peak 
	   value will be <scale>.  Single maximum if <len> is odd */
    long i;
    double tscale = 2.0*PI/(len-1);
    double a0 = 0.54 * scale;
    double a1 = scale - a0;

    for(i = 0; i<len; ++i)  {
	*dst = a0 - a1 * cos(tscale * i);
	dst += dinc;
    }
}

typedef struct {	/* describe one slice of noise spectrum */
    float	time;
    float	fLow;
    float	fHi;
    float	magAv;
    float	trxPoct;	/* average track density */
    float	avDur;		/* average duration of a track */
    float	avFsd;		/* average frqSD/frqAV over tracks */
} NOISE_SPCT;

#define MAXKERNLEN  32

void nsSmoothSpec(float *inspec, long instep, float *outspec, long outstep, 
		   long npts, long kernlen)
{	/* smooth the spectrum with a hamming kernel */
static long lastklen = -1;
static float *mykern = NULL;

    if(mykern == NULL)  {
	mykern = TMMALLOC(float, MAXKERNLEN, "nsSmoothSpec");
    }
    assert(kernlen < MAXKERNLEN);
    if(kernlen != lastklen)  {
	SetHamming(mykern, 1, kernlen, 2.0/(double)kernlen);
	lastklen = kernlen;
    }
    if(inspec != outspec)  {
	VCopy(inspec, instep, outspec, outstep, npts);
    } else {
	/* if inspec == outspec, must be in place i.e. same stepsize */
	assert(instep == outstep);
    }
    Convo(outspec, outstep, npts, mykern, 1, kernlen);
}

typedef struct {	/* we should put sampled spectra in a str like this */
    cpxpolar	*bins;
    long	nbins;
    double	basefrq;
    int 	binsPerOct;
} CQSPECT;

int nsSpecLocMax(cpxpolar *spec, long speclen, long startbin)
{	/* test for the specified point as a local spectral max */
    if(startbin > 0 && startbin < speclen - 1
       && spec[startbin].mag > spec[startbin-1].mag 
       && spec[startbin].mag > spec[startbin+1].mag) {
	return 1;
    } else {
	return 0;
    }
}

int nsNoiseRng(cpxpolar *spec, long speclen, long startbin, 
	       double below, double above, long *plobin, long *phibin)
{	/* find a range of bins that meet our criteria .. not clear */
    double	mag1;
    long	exU, exL;

    mag1 = spec[startbin].mag;
    if(extendWhileWithin((float *)&(spec[startbin]), -2, 
			 mag1+below,mag1+above, startbin, &exL) != 1
       && extendWhileWithin((float *)&(spec[startbin]), 2, 
			    mag1+below,mag1+above, speclen-1-startbin, &exU) 
            != 1)  {  /* fell out bottom or hit end of spec on both sides */
	*plobin = startbin - exL;
	*phibin = startbin + exU;
	return 1;
    } else {
	return 0;
    }
}

typedef struct {	/* a noise range in an instantaneous spectrum */
    float	time;
    float	mag;
    RANGE_STR	frq;
} NOISE_SLICE;

NOISE_SLICE *nslcNew(double time, double fmin, double fmax, double mag)
{	/* allocate a new noise-slice struct, fill in as directed */
    NOISE_SLICE *rslt;

    rslt = TMMALLOC(NOISE_SLICE, 1, "nslcNew");
    rslt->time = time;
    rslt->mag  = mag;
    rslt->frq.min = fmin;
    rslt->frq.max = fmax;
    return rslt;
}

void nslcFree(NOISE_SLICE *nslc)
{	/* destroy noise-slice structure */
    free(nslc);
}

#define OCTAVES(frqratio)	(log((frqratio))/LOGTWO)

#define MIN_NS_OVLP_FCT		0.6	/* noise frq bands match criteria */
#define NS_SKIRT		-10.0	/* noise band is down 10dB ard peak */

NOISE_SLICE *nsGetNrstNoise(double time, NOISE_SLICE *oldNoise)
{	/* search the noise regions at <time> for one that overlaps 
	   with the passed noise (the most? above some thresh?) */
    NOISE_SLICE *rslt;
    double	dt = 0.001;
    long	totBins = cqi->nBins * cqi->nOcts;
    long	klen = cqi->nBins * 1 + 1;	/* kernel is one octave */
static cpxpolar *myspect = NULL;
    int		found = 0;
    long	i, minbin, maxbin;
    double	minfrq, maxfrq, mag;
    RANGE_STR	frg, ovrg;
    double	oocts, nocts, vocts, ovfact;

    if(myspect == NULL)  {
	myspect = TMMALLOC(cpxpolar, totBins, "nsGetNrstNoise");
    }
    /* read back the original analysis spectrum for this time */
    SpecOfCQTFile(cqi, time, dt, myspect);
    Vlin2DB((float *)myspect, 2, (float *)myspect, 2, (double)1, totBins);
    /* smooth the spectrum */
    nsSmoothSpec((float *)myspect, 2, (float *)myspect, 2, totBins, klen);

    oocts = OCTAVES(oldNoise->frq.max/oldNoise->frq.min);
    i = 1;
    while(!found && i<totBins-1)  {
	if( nsSpecLocMax(myspect, totBins, i) )  {
	    if( nsNoiseRng(myspect, totBins, i, NS_SKIRT, 1.0, 
			   &minbin, &maxbin) ) {
		minfrq = BinToFrq(minbin);
		maxfrq = BinToFrq(maxbin);
		frg.min = minfrq;
		frg.max = maxfrq;
		nocts = OCTAVES(maxfrq/minfrq);
		if(rngIntersect(&(oldNoise->frq), &frg, &ovrg))  {
		    vocts = OCTAVES(ovrg.max/ovrg.min);
		    /* overlap factor is average overlap proportion (8ves) */
		    ovfact = 0.5*(vocts/oocts + vocts/nocts);
		    if(ovfact>MIN_NS_OVLP_FCT)	{
			found = 1;
			mag = VAvg((float *)&myspect[minbin], 2, 
				   maxbin-minbin);
		/*	fprintf(stderr, "range@%.6f %6.1f..%6.1f = %6.2f\n", 
				time, minfrq, maxfrq, mag);
		 */
		    }
		}
	    }
	}
	++i;
    }
    if(!found)
	return NULL;
    else {
	rslt = nslcNew(time, minfrq, maxfrq, mag);
	return rslt;
    }
}

TKLIST txSelTFM(TKLIST src, double tmin, double tmax, 
	      double fmin, double fmax, 
	      double mmin, double mmax)
{	/* does a t/f intersect, then applies mag filter too */
    TKLIST rslt = tkGetIntersect(src, tmin, tmax, fmin, fmax);
    TKLIST r2 = SelTrxByMag(rslt, mmin, mmax);

    free_list(rslt);
    return r2;
}

#define NS_MAX_MAG_DIFF	(6.0)	/* biggest allowed step btwn avmgs 
				   in temporally adjacent noiseslices */
#define NS_MIN_OCTVS	(1.0)	/* noise band must be at least 
				   1 octave wide to be recognized */

int ns_rmv_tk(TRACK *tk, int type, void *time_and_spect)
{	/* for each track, zero the energy in the spectrum suggested
	   by that track */
    double time = *(((double **)time_and_spect)[0]);
    double clrv = *(((double **)time_and_spect)[2]);
    float  *spect= *(((float **)(time_and_spect))+1);
    long	totBins = cqi->nBins * cqi->nOcts;
    double	frq, mag;
    long i,bin;

    assert(type == T_TRACK);
    if(tkFrqMagForT(tk, time, &frq, &mag))  { /* there is overlap */
    /*	fprintf(stderr, "nsRm: tk%4d, f=%6.1fHz\n", tk->id, frq);  */
	bin = FrqToBin(frq);
	for(i = bin-2; i<=bin+2; ++i)
	    spect[2*i] = clrv;	/* clear a 5 pt range around max */
    }
    return 1;
}

NOISE_SLICE *nsNoTkNoise(double now, double frq, TKLIST trx, int dispQ)
{	/* Find a noise slice around the specified <t,f> point 
	   *after* removing all the energy from the tracks in <trx> */
    double 	dt = 0.001, clearval = -40.0, mag, minFrq, maxFrq;
    long	totBins = cqi->nBins * cqi->nOcts, bin, lobin, hibin;
    char 	timestr[32];
    long	klen = cqi->nBins * 1 + 1;	/* kernel is one octave */
    double	*data[3];
    NOISE_SLICE *rslt = NULL;

    /* read back the original analysis spectrum for this time */
    SpecOfCQTFile(cqi, now, dt, spect);
    Vlin2DB((float *)spect, 2, (float *)spect, 2, (double)1, totBins);
    /* smooth the spectrum */
    nsSmoothSpec((float *)spect, 2, (float *)spec2, 2, totBins, klen);

    /* for each track at this time, zero that bin */
    VCopy((float *)spect, 2, (float *)spec2, 2, totBins);
    data[0] = &now;
    data[1] = (double *)(spec2);
    data[2] = &clearval;
    /* map the fn that will remove the energy */
    maparg((LIST)trx, (NodeFunc)ns_rmv_tk, data);
    /* smooth the energy-removed spectrum */
    nsSmoothSpec((float *)spec2, 2, (float *)spec4, 2, totBins, klen);

    /* find the noise range containing the start seed */
    bin = FrqToBin(frq);
    mag = spec4[bin].mag;
/*    fprintf(stderr, "nsNoise: tim %.6f s, cfq %6.1f, mag %5.1f\n", 
	    now, frq, mag);
 */
    if(nsNoiseRng(spec4, totBins, FrqToBin(frq), -5.0, +20.0, 
		 &lobin, &hibin))  {
	minFrq = BinToFrq(lobin);
	maxFrq = BinToFrq(hibin);
	mag    = VAvg((float *)&spec4[lobin], 2, hibin-lobin);
	rslt = nslcNew(now, minFrq, maxFrq, mag);
/*	fprintf(stderr, "nsNs@%.6f s: %6.1f :%7.1f @ %5.1f dB\n", 
		now, minFrq, maxFrq, mag);
 */
    }

    if(dispQ)  {
	/* display */
	sprintf(timestr, "%.6f", now);
	CopyAndDisp((float *)spec4, 2, totBins, wndw1, timestr, 0, 1.0);
	SetXLimits(wndw1, cqi->pvh->minFreq , cqi->pvh->maxFreq);
	SetLogScale(wndw1, TRUE);
	if(rslt)  {
	    SelectX(wndw1, minFrq); /* move xhair to lower frq region */
	}
    }
    return rslt;
}

TKLIST nsNewGrowNoise(double now, double frq, TKLIST trx, TKLIST focus)
{	/* find noise region around seed excluding energy from trx.
	   Grow forward as long as it lasts. */
    double 	time = now, tstep = 10.0/frq, mag;
    NOISE_SLICE	*ns1 = NULL, *ns2 = NULL;
    TKLIST 	rslt = NULL, r2;
    double	sttTime, endTime, loFrq, hiFrq;
    double	magSlope, maxMagSlope = -1000.0, minMagSlope = 1000.0;
    double	minMag = 1000.0, maxMag = -1000.0;
    double	minTime = time, maxTime = time;
    long	sliceCount = 0;
    int 	display, case1, case2, case3, case4;
    double	upslope, downslope;

    sttTime = now;
    do  {
	if(ns2)	nslcFree(ns2);
	ns2 = ns1;
	ns1 = nsNoTkNoise(time, frq, trx, display = 0);
	time += tstep;
	if(ns1)  {
	    mag = ns1->mag;
	    ++sliceCount;
	    hiFrq += ns1->frq.max;
	    loFrq += ns1->frq.min;
	    if(minMag > mag)   {
		minMag = mag;
		minTime = time;
	    }
	    if(maxMag < mag)   {
		maxMag = mag;
		maxTime = time;
	    }
	    if(ns2) {
		magSlope = (mag - ns2->mag)/tstep;
		if(minMagSlope > magSlope)	minMagSlope = magSlope;
		if(maxMagSlope < magSlope)	maxMagSlope = magSlope;
	    }
	}
	case1 = (minMagSlope > 0) 
	    || (magSlope < -4*minMagSlope);
	upslope   = (minTime==time)? 0:(mag - 6.0 - minMag)/(time - minTime);
	downslope = (maxTime==time)? 0:(mag + 6.0 - maxMag)/(time - maxTime);
	case2 = (time <= maxTime) 
	    || (downslope > -100.0);
	case3 = (time <= minTime)
	    || (upslope < 100.0);
	/* adding a *constant* to the numerator reduces the influence
	   of noise when divided by v.small dt's (i.e. precompensates 
	   6 dB of ripple in mag measurement level) */
	fprintf(stderr, "nsr@%9.6f/%5.1f: up(%9.6f):%6.0f down(%9.6f):%6.0f\n",
		time, mag, minTime, upslope, maxTime, downslope);
    } while(ns1 && case1 && case2 && case3);

    /* (ns1->mag - minMag) < maxMag-minMag < 10.0) ); */

    if(!ns1)  {
	fprintf(stderr, "nsGrw: end by no spec\n");
    } else if (!case1)  {
	fprintf(stderr, "nsGrw: end mgSl %f > -4*min %f\n", 
		magSlope, minMagSlope);
    } else if (!case2)  {
	fprintf(stderr, "nsGrw: end mgDrop %f..%f in %f = %f dB/sec\n", 
		maxMag, mag, time - maxTime, (mag+6.0-maxMag)/(time-maxTime));
    } else if (!case3)  {
	fprintf(stderr, "nsGrw: end mgRise %f..%f in %f = %f dB/sec\n", 
		minMag, mag, time - minTime, (mag-6.0-minMag)/(time-minTime));
    }
    endTime = time;
    if(sliceCount > 1 && (endTime - sttTime) > 0.015)  {
	hiFrq /= (double)sliceCount;
	loFrq /= (double)sliceCount;
	rslt = txSelTFM(focus, sttTime, endTime, loFrq, hiFrq, 
			minMag, maxMag+10.0);
	/* remove tracks in our excluded set */
	r2 = (TKLIST)remove_list((LIST)rslt, (LIST)trx);
	free_list((LIST)rslt);  rslt = r2;
	if(top_size(rslt)==0)
	    rslt = NULL;
    }
    return rslt;
}
 
/* NOISE_SPCT */
TKLIST nsFindAt(double time, double frq, TKLIST focus, int dispQ)
{	/* find a noise range centred at time passed.
	   Do display if <dispQ> */
    TKLIST 	rslt = NULL, nbrs, nnbrs, wholeobj;
    double	mag1, mag2;
    TRACK	*tk1;
    double	dt, tdur, tchr, tend;
    long	ctrbin, minbin, maxbin, ntrk, ntk2;
    double	minfrq, maxfrq, tkdens, octvs;
    long	totBins = cqi->nBins * cqi->nOcts;
    char 	timestr[32];
    long	klen = cqi->nBins * 1 + 1;	/* kernel is one octave */
    long 	exStep;
    long	i;
    NOISE_SLICE *nslc, *ns2;
    int		found = 0;

    /* read back the original analysis spectrum for this time */
    dt = 0.001; 	/* ??? */
    SpecOfCQTFile(cqi, time, dt, spect);
    Vlin2DB((float *)spect, 2, (float *)spect, 2, (double)1, totBins);
    /* smooth the spectrum */
    nsSmoothSpec((float *)spect, 2, (float *)spec2, 2, totBins, klen);

    if(dispQ)  {
	sprintf(timestr, "%.6f", time);
	CopyAndDisp((float *)spect, 2, totBins, wndw1, timestr, 0, 1.0);
	SetXLimits(wndw1, cqi->pvh->minFreq , cqi->pvh->maxFreq);
	SetLogScale(wndw1, TRUE);

	CopyAndDisp((float *)spec2, 2, totBins, wndw2, timestr, 0, 1.0);
	SetXLimits(wndw2, cqi->pvh->minFreq , cqi->pvh->maxFreq);
	SetLogScale(wndw2, TRUE);
    }
    for(i = 1; i < totBins - 1; ++i)  {
	if( nsSpecLocMax(spec2, totBins, i) )  {
	    if( nsNoiseRng(spec2, totBins, i, -10.0, 1.0, &minbin, &maxbin) ) {
		minfrq = BinToFrq(minbin);
		maxfrq = BinToFrq(maxbin);
		if(minfrq <= frq && frq < maxfrq)  {
		    found = 1;
		    fprintf(stderr, "range %.1f..%.1f\n", minfrq, maxfrq);
		    mag1 = VAvg((float *)&spect[minbin], 2, maxbin-minbin);
		    nslc = nslcNew(time, minfrq, maxfrq, mag1);
		    octvs  = OCTAVES(maxfrq/minfrq);
		    tchr = 1.0 / sqrt(minfrq * maxfrq);
		}
	    }
	}
    }

    if(found == 0 || octvs < NS_MIN_OCTVS)	return NULL;

    /* find 'duration' of noise by searching forward range */
    tend = time;
    mag2 = nslc->mag;	/* finishing avmag */
    while(nslc)  {
	tend += 5*tchr;
	if( fabs(mag2 - nslc->mag) < NS_MAX_MAG_DIFF)  {
	    mag2 = nslc->mag;	/* will hold final avmag */
	    ns2 = nsGetNrstNoise(tend, nslc);
	} else {
	    fprintf(stderr, "mag fail %.2f->%.2f\n", 
		    mag2, nslc->mag);
	    ns2 = NULL;
	}
	nslcFree(nslc);
	nslc = ns2;
    }

    /* find all unselected tx in wider neighborhood (one octave, an instant) */
    nbrs = txSelTFM(focus, time, tend, minfrq, maxfrq, 
		    MIN(mag1, mag2), MAX(mag1, mag2) + 6.0);
    RedrawTrax(nbrs, TKM_SPECIAL);
    /* calculate avg ntrk/octave NOT!! */
    ntrk = top_size((LIST)nbrs);
    tkdens = (float)ntrk/octvs;
    /* remove those significantly long in duration */
    nnbrs = SelTrxByDur(nbrs, 0.0, tchr*20.0);
    free_list((LIST)nbrs);  nbrs = nnbrs;	
    /* find avg durn of remains */
    ntk2 = top_size((LIST)nbrs);
    tdur = tkSumDurns(nbrs) / (double)ntk2;
    
    fprintf(stderr, 
"Ns: f:%.0f-%.0f=%.1fov m:%.2f(%.2f) tx:%d(%d) dns:%.1f avd:%.4f(%.4f)\n", 
    minfrq, maxfrq, octvs, mag2, mag1, ntrk, ntk2, tkdens, tdur, tchr);

    return nbrs;
}

TKLIST tkGrowNoise(TRACK *tk, TKLIST focus)
{	/* See if you can make a good set of tracks that look like 
	   a noiseburst based on these ones. */
    double	tim,frq;
    /* Strategy: have some seed time=t1,freq=f1 point.  Track not important.
       Calculate smoothed spectrum.  Take *smoothed* mag at seed pt = m1.
       Search through bins both ways to (+?/)-10dB of mag 
       -> frq interval fl1..fu1
       Avg smoothed mag in that interval = ma1
       Get all tracks in frq window.  Count -> trx/8ve (=dens1)
       Filter those tracks to +/- 5dB of avg mag (?)
       Find avg duration of those tracks -> ta1
       (Find avg fsd/favg of tracks?)

       Test on wide interval, high density, short avg time.

       Advance by chr time (e.g. length thresh), repeat.
       If frq intervals overlap > 80% AND delta(avmag) < some lim
       can extend.

       -> Noise region represented as a list of NoiseSpects - 
       a structure describing the inst. spectrum allocated to noise
     */
    if(tk == NULL)	return NULL;
    tim = tk->st.tavg;
    frq = tk->st.favg;
    return nsFindAt(tim,frq, focus, stDispQ);
}

TKLIST txNoisyObjects(TKLIST trx)
{	/* go through all of trx, grow some noise objects */
    if(trx)
	return tkGrowNoise((TRACK *)CAR(trx), allTrax);
    else
	return NULL;
}

TGLIST tklNoiseGroups(TKLIST tks, TKLIST focus)
{	/* yr basic NOISE groups - from -track lists, but tries 
	   to up efficiency by working in increasing time, and 
	   NOT growing noise around tracks already part of noise 
	   regions returned so far */
    TKLIST s2, seeds = copy_top(tks);
    TKLIST ns;
    TGLIST r2, rslt = NULL, *rsPtr;

    rsPtr = &rslt;
    seeds = qsort_list(seeds, tkCmpTrxByT);	/* sort in asc. time */
    while(seeds)  {
	ns = tkGrowNoise((TRACK *)CAR(seeds), focus);
	/* used up this track - remove it from seeds */
	free_next_ptr(&seeds);
	if(ns) {
	    /* also remove elements of this noise region from seeds */
	    s2 = remove_list(seeds, ns);
	    free_list(seeds);	seeds = s2;
	    *rsPtr = new_node(NULL, (LIST)tkgrNew(ns, TG_NOISE), T_TKGROUP);
	    advance_ptr(&rsPtr);
	}
    }
    r2 = tglRemoveNearDuplicates(rslt);  
    fprintf(stderr, "tklNoise: %d->%d->%d\n", 
	    top_size(tks), top_size(rslt), top_size(r2));
    free_list(rslt);	rslt = r2;

    return rslt;
}

TGLIST tgl_noise_tg(TKGROUP *tg, int *ptype, TKLIST focus)
{	/* LNodeFunc to generate noise regions from a track *group* -
	   handles the parts all at once with tklNoiseG */
    TGLIST rslt;
    assert(*ptype == T_TKGROUP);
    *ptype = T_LIST;
    rslt = tklNoiseGroups(tg->parts, focus);
/*    fprintf(stderr, "tgl_n_t: returning %d\n", top_size(rslt));  */
    return rslt;
}

TGLIST tgl_noise_tk(TRACK *trk, int *ptype, TKLIST focus)
{	/* LNodeFunc to generate noise regions from tracks. 
	   At present, at most one noise region per track. */
    TKLIST ns;
    TKGROUP *tg;

    assert(*ptype == T_TRACK);
    *ptype = T_LIST;	/* lmappend *dermands* this */
    ns = tkGrowNoise(trk, focus);
    if(ns)  {
	tg = tkgrNew(ns, TG_NOISE);
	return (TGLIST)new_node(NULL, (LIST)tg, T_TKGROUP);
    } else {
	return NULL;
    }
}

#else /* !CQFILE */   

void FindNoiseM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    /* dummy */
}

#endif /* CQFILE */   

#endif /* TRAX */   
