/***********************************************************************\
*   db.onset.c								*
*   Build (grow) `TKGROUP's based on common onset times			*
*   1993apr30 dpwe							*
\***********************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <genutils.h>		/* for Convo ??? (in DSPbench.h anyway) */
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/
static SELPNL_STR *spnl = NULL;		/* maybe offer user choices */
static WINDAT *wndw1 = NULL;

/******** prototypes ********/
TGLIST tkNewOnsets(TRACK *tk, TKLIST focus);
	/* Make onset groups by new method */
TGLIST tgl_NewOnsets_tk(TRACK *tk, int *ptype, TKLIST focus);
	/* Make onset groups by new method */
TGLIST tglNewOnsets(TKLIST trx, TKLIST focus);

/******* implementations *****/

void FindOnsetsM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    TKLIST focus;

    noisyInitStatics(wrcd, 2);

    if(top_size(selTrax) == 1)  {	/* display onset time for lone tk */
	TRACK *trk = (TRACK *)CAR(selTrax);
	double frq, mag;
	double otime = tkOnsetTFM(trk, &frq, &mag);

	if(wndw1 == NULL)  {
	    wndw1 = NewWin("Onset time", 0);
	}

	DoTrkDisp(wndw1, trk, 0);	/* draw mag contour */
	SelectX(wndw1, otime);	/* move xhair to found onset */
    }

    if(spnl == NULL)  {
	spnl = spnlNew("Onsets");
    } else {
	/* should dispose of all last time's stuff here */
	spnlReset(spnl);
	spnlUnhide(spnl);
    }
    if(selTrax)  {	
	TGLIST tails, t2;
	TKLIST myAll;
	RANGE_STR t_rg;

	tklFindRanges(selTrax, &t_rg, NULL, NULL);
	myAll = tkGetIntersect(allTrax, t_rg.min-0.5, t_rg.max, 0, 44100);
	focus = tklOnsetsPrepare(myAll);

	RedrawTrax(focus, TKM_SPECIAL);
 
    	tails = lmappend(selTrax, (LNodeFunc)tgl_onset_tk, allTrax); /* */

	t2 = tglRemoveNearDuplicates(tails);
	fprintf(stderr, "ons: %d->%d->%d\n", 
		top_size(selTrax), top_size(tails), top_size(t2));
	free_list(tails);	tails = t2;
	
	spnl = spnlFromTkgList(spnl, tails, wrcd);

	/* tglSave(tails, "noiseb.tgl", "NoiseBursts"); */
    }
    return;
}

void ThreshSel(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    TKLIST losers, s2;
    RANGE_STR m_rg;

    if(selTrax)  {
	tklFindRanges(selTrax, NULL, NULL, &m_rg);
	fprintf(stderr, "thresh: %d tracks, mag=%.1f..%.1f\n", 
		top_size(selTrax), m_rg.min, m_rg.max);
	/* find tracks in bottom 5 dB of mag range */
	losers = SelTrxByMag(selTrax, m_rg.min-5.0, m_rg.min+5.0);
	RedrawTrax(losers, TKM_NORM);
	s2 = remove_list(selTrax, losers);
	free_list(losers);
	free_list(selTrax);	selTrax = s2;
    }
}

void NewNoiseM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    TKLIST focus;

    noisyInitStatics(wrcd, 2);

    if(top_size(selTrax) == 1)  {	/* display onset time for lone tk */
	TRACK *trk = (TRACK *)CAR(selTrax);
	double frq, mag;
	double otime = tkOnsetTFM(trk, &frq, &mag);

	if(wndw1 == NULL)  {
	    wndw1 = NewWin("Onset time", 0);
	}

	DoTrkDisp(wndw1, trk, 0);	/* draw mag contour */
	SelectX(wndw1, otime);	/* move xhair to found onset */
    }

    if(spnl == NULL)  {
	spnl = spnlNew("Onsets");
    } else {
	/* should dispose of all last time's stuff here */
	spnlReset(spnl);
	spnlUnhide(spnl);
    }
    if(selTrax)  {
	TGLIST tails, t2;

/*	focus = tklOnsetsPrepare(allTrax);

	RedrawTrax(focus, TKM_SPECIAL);
 */
    /*	tails = lmappend(selTrax, (LNodeFunc)tgl_NewOnsets_tk, allTrax); /* */
    	tails = tglNewOnsets(selTrax, allTrax);		/* */

	t2 = tglRemoveNearDuplicates(tails);
	fprintf(stderr, "newons: %d->%d->%d\n", 
		top_size(selTrax), top_size(tails), top_size(t2));
	free_list(tails);	tails = t2;
	
	spnl = spnlFromTkgList(spnl, tails, wrcd);

	tglSave(tails, "noiseb.tgl", "NoiseBursts");
    } else {	/* no seeds selected - load from disk */
	TGLIST tails;
	char   *title;

	tails = tglLoad("noiseb.tgl", &title, allTrax);
	spnl = spnlFromTkgList(spnl, tails, wrcd);
    }
    return;
}

double tkOnsetTFM(TRACK *trk, double *pfrq, double *pmag)
{	/* calculate the onset time of a track based on its mag slope.  
	   Also return the frequency and magnitude at that instant. */
    double time;
    PKLIST pkl;
    double mag0, tim0, mag1, tim1, tchr, magslope, maxslope = 0.0;
    double magthresh;
    int onset = 0;

    /* ONSET TIME is defined as the moment when the MAGNITUDE SLOPE 
       falls below HALF of its peak to date - AS LONG as we are already
       within some threshold of PK MAG for this track */

    tchr = 1.0/trk->st.favg;
    magthresh = trk->st.mmax - 10.0;	/* *must* be above pk-10dB to pass */
    pkl = trk->tktop->next;	/* skip first 'coz -100dB */
    mag0 = pkl->peak->mag;
    tim0 = pkl->peak->tim;
    pkl = CDR(pkl);
    while(!onset && pkl->next)  {	/* DON'T calc slope for last pt */
	mag1 = pkl->peak->mag;
	tim1 = pkl->peak->tim;
	magslope = (mag1 - mag0)/(tim1 - tim0);
	if(magslope > maxslope)	
	    maxslope = magslope;
	else {
	    if(magslope < maxslope / 4.0 && mag1 > magthresh) {
	     /* fprintf(stderr, "onset: slope=%f max=%f time=%.6f\n", 
			magslope, maxslope, tim1);  */
		onset = 1;
		*pfrq = pkl->peak->frq;
		*pmag = mag1;
	    }
	}
	pkl = pkl->next;
	mag0 = mag1;
	tim0 = tim1;
    }
    if(!onset)  {	/* never found a real onset */
	tim1 = trk->st.tavg;	/* return some lame guess */
	*pfrq = trk->st.favg;
	*pmag = trk->st.mavg;
    }
    /* just fly it for now */
/*    DoTrkDisp(wndw1, trk, 0);	/* draw mag contour */
/*    SelectX(wndw1, tim1);	/* move xhair to found onset */
    return tim1;
}

double tkOnsetT(TRACK *trk)
{	/* calculate the onset time of a track based on its mag slope */
    double dummyf, dummym;

    return tkOnsetTFM(trk, &dummyf, &dummym);
}

static int tk_set_onset(TRACK *trk, int type, void *arg)
{	/* MAPpable function calculates and stores onset time of track */
    trk->st.tavg = tkOnsetT(trk);
    return 1;
}

#define TK_ONS_THRESH	4.0	/* max distance (in tfDist units) for succs */

TKLIST tkOnsetters(TRACK *trk, TKLIST focus, TKLIST withouts)
{	/* find tracks above current with onset close to track passed.
	   Return as many as exceed some preset threshold.
	   DO NOT return any tracks from <withouts> */
    TRACK *succ = NULL;
    TKLIST cands, cands2, rslt = NULL;
    double time, freq, fsd, tchr, dist = NOMIN;
    RANGE_STR rg;	/* for filtering by distance */

    if(focus == NULL)	focus = allTrax;

    freq  = trk->st.favg;
    time = trk->st.tavg = tkOnsetT(trk);

    /* find candidates for onset neighbors - only look UP in frequency */
    tchr = 1.0 / freq;	/* characteristic time is one cycle */
    cands = tkGetIntersect(focus, time - 0.5 * tchr, time + 0.5*tchr, 
			   freq, 2.2 * freq);	/* just over an octave */
    /* remove self from candidates */
/*    root = (TKLIST)new_node(NULL, (LIST)trk, T_TRACK);
      cands2 = (TKLIST)remove_list((LIST)cands, (LIST)root);
      free_list((LIST)root);
 */ /* removing self presumably handled by `withouts' */
    cands2 = (TKLIST)remove_list((LIST)cands, (LIST)withouts);
    free_list((LIST)cands);
    cands = cands2;
    
    if(cands)  {
	RedrawTrax(cands, TKM_SPECIAL);
	/* make sure all tracks have right onset stored */
	map((LIST)cands, (NodeFunc)tk_set_onset);
	/* sort in descending `onset closeness' */
	txSetSortTargetTF(time, freq);
	cands = qsort_list((LIST)cands, tkCmpNrstOnset);
	rg.min = -1.0;		rg.max = TK_ONS_THRESH;
	rslt = (TKLIST)keep_if((LIST)cands, tk_onset_pred, &rg);
	free_list((LIST)cands);
     /* fprintf(stderr, "tkOnss: end at %.6f, %.0f\n", time, freq);
	tklPrintOnsDists(rslt, "cands");
      */
    }
    return rslt;
}

static int tk_find_max_frq(TRACK *track, int type, void *arg)
{	/* node function records node with highest head frq */
    LSRCH_STR *mms = (LSRCH_STR *)arg;	/* threshold and answer */
    double frq = tkAvgHeadFrq(track);

    assert(type == T_TRACK);
    if(frq > mms->range.max)  {
	mms->result = (LIST)track;
	mms->range.max = frq;
    }
    return 1;
}

TRACK *tkFindMaxFrq(TKLIST trx)
{	/* return the element in the track list that has highest head frq */
    LSRCH_STR 	lss;
    TRACK	*rslt;

    lss.result = (LIST)CAR(trx);	/* start with arbitrary item */
    lss.range.max = tkAvgHeadFrq((TRACK *)lss.result);
    /* apply to all the tracks.. */
    maparg((LIST)trx, (NodeFunc)tk_find_max_frq, &lss);
/*  fprintf(stderr, "FMF:max = %.1f\n", lss.range.max);  */
    rslt = (TRACK *)lss.result;
    return rslt;
}

TGLIST tgExtendOnset(TKGROUP *tgr, TKLIST focus, TKLIST *pwithouts)
{	/* Grow into an onset group by chasing upwards
	   But *don't* extend by any of the tracks in <*pwithouts>, 
	   and extend <*pwithouts> with the tracks we return. */
    TRACK *trk, *tk2;
    TKGROUP *newtg;
    TKLIST tx, txt, thisTkAsAList;
    TGLIST rslt = NULL, *rsTail;

    if(tgr == NULL)  {
	fprintf(stderr, "tkExOns passed NULL\n");
	return NULL;
    }

    assert(tgr->type == TG_ONSET);	/* going to use v.onset.latest */

/*    fprintf(stderr, "tgExO: tgr->st.npts = %d\n", tgr->st.npts);  */
/*    tkgrPrint(tgr, "tgExO");	*/
/*    trk = tkFindMaxFrq(tgr->parts);	/* find highest frq track to date */
    assert(trk = tgr->v.onset.latest);

/*    DoTrkDisp(wndw1, trk, 0);	/* draw mag contour */
/*    SelectX(wndw1, tkOnsetT(trk));	/* move xhair to found onset */

/*    if(find_el((LIST)*pwithouts, (LIST)trk, T_TRACK)==NULL)  {
/	fprintf(stderr, "!");	fflush(stderr);
/	*pwithouts = new_node(*pwithouts, (LIST)trk, T_TRACK);
/    }
/    thisTkAsAList = (TKLIST)new_node(NULL, (LIST)trk, T_TRACK);
/
/    tx = tkOnsetters(trk, thisTkAsAList);
/    free_list(thisTkAsAList);
 */
    tx = tkOnsetters(trk, focus, tgr->parts);	
    /* CANNOT repeat tracks in a group! */

/*    tx = tkOnsetters(trk, focus, *pwithouts);	*/
    txt = tx;
    rsTail = &rslt;
    while(txt)  {
	newtg = tkgrDup(tgr);
	tk2 = (TRACK *)CAR(txt);
	tkgrAddTrack(newtg, tk2);
	newtg->v.onset.latest = tk2;
	txt = (TKLIST)CDR(txt);
	*rsTail = new_node(NULL, (LIST)newtg, T_TKGROUP);
	advance_ptr(&rsTail);
    }
    *(find_last_ptr((LIST *)&tx)) = *pwithouts;
    *pwithouts = tx;	/* add all these tracks into withouts */
    return rslt;
}

int tg_test_trk(LIST data, int type, void *arg)
{	/* NodeFunc tests for the presence of a given track in 
	   the `parts' of trkgroup */
    TKGROUP	*tkg = (TKGROUP *)data;
    TRACK	*trk = (TRACK *)arg;

    assert(type == T_TKGROUP);
    if(find_car((LIST)tkg->parts, (LIST)trk))
	return 1;	/* track occurs in tkgroup */
    else 
	return 0;	/* track does not occur in tkgroup */
}

int tg_print_fn(LIST data, int type, void *arg)
{	/* NodeFunc prints elements of list of TKGROUPs briefly */
    TKGROUP *tkg = (TKGROUP *)data;
    char str[128];

    assert(type == T_TKGROUP);
    tkgrTitle(tkg, str);
    fprintf(stderr, "%s\n", str);
    return 1;
}

void tglPrint(TGLIST tgls, char *name)
{	/* print brief summary of tg list */
    fprintf(stderr, "tglPrint '%s':\n", name);
    map((LIST)tgls, tg_print_fn);
}

LIST eps_close(LIST listin, LNodeFunc fn, void *arg)
{	/* apply <fn> to all top-level elements of <listin>, 
	   */
    LIST new, rslt = NULL;

    while(listin)  {
	rslt = append(rslt, listin);
	fprintf(stderr, "eps_loop:\n");
	map(listin, tg_print_fn);
	listin = lmappend(listin, fn, arg);
    }
    return rslt;
}

static int tg_filter_onset_nodes(LIST data, int type, void *arg)
{	/* MAPpable func expects a TKGROUP as data and a TGLIST* as arg;
	   runs through the TGLIST* destroying entries that contain 
	   the track indicated by data's ->v.onset.latest field 
	   somewhere in their ->parts;
	   This is the pruning mode of choice for the onset growing:  
	   any lately-opened node that reaches a certain track is 
	   probably preferable to some shorter path to that node */
    TKGROUP *tkg = (TKGROUP *)data;
    TGLIST  *ptgl = (TGLIST *)arg;
    TGLIST  tgl = *ptgl;

    assert(type == T_TKGROUP);
    *ptgl = remove_if((LIST)tgl, tg_test_trk, tkg->v.onset.latest);
    free_list((LIST)tgl);
    return 1;
}

static LIST depth_close(LIST list, LNodeFunc fn, void *arg)
{	/* like eps_close, applies fn to all top level elements, 
	   recursively, until no new elements produced, but operates 
	   depth-first instead of breadth-first */
    LIST this, prod, rslt = NULL, *rsPtr;
    int  type;

    rsPtr = &rslt;
    while(list)  {
	this = CAR(list);
	type = CARTYPE(list);
	*rsPtr = list;
	list = CDR(list);
	set_next(*rsPtr, NULL);	/* unlink this element */
	advance_ptr(&rsPtr);	/* .. and transfer to result list */
	prod = (*fn)(this, &type, arg);
	/* insert new nodes at *top* of list of nodes to `open' */
	*(find_last_ptr(&prod)) = list;
	list = prod;
    }
    return rslt;	/* all the nodes we processed */
}

static LIST tgl_ons_fn(LIST inlist, int *ptype, TKLIST *data[])
{	/* lmappend-able fn does onset extension of TKGROUPS. */
    TKGROUP *tkg = (TKGROUP *)inlist;
    TKLIST  *pwithouts = data[0];
    TKLIST  focus = *(data[1]);

    assert(*ptype == T_TKGROUP);
    *ptype = T_LIST;	/* return is a TGLIST, a kind of list */
    return tgExtendOnset(tkg, focus, pwithouts);
}

static TGLIST tgOnsetClose(LIST list, TKLIST focus, TKLIST *pwithouts)
{	/* like eps_close, applies fn to all top level elements, 
	   recursively, until no new elements produced.
	   Proceeds depth-first, but without using lmappend(), to 
	   allow pruning of waiting nodes after each production. */
    TKLIST this, prod, rslt = NULL, *rsPtr, *botPtr;
    TKLIST *args[2];
    int  type;

    args[0] = pwithouts;
    args[1] = &focus;	/* actually need focus, not &focus; deref later */
    rsPtr = &rslt;
    botPtr = find_last_ptr(&list);
    while(list)  {
	this = CAR(list);
	type = CARTYPE(list);
	*rsPtr = list;
	list = CDR(list);
	set_next(*rsPtr, NULL);	/* unlink this element */
	/* open the node */
	prod = tgl_ons_fn(this, &type, args);
	if(prod == NULL)  {	/* this track had no progeny - keep it */
	    advance_ptr(&rsPtr);
	} else {	/* track had progeny; no need to keep parent */
	    free_list(*rsPtr);
	    *rsPtr = NULL;
	    /* prune nodes remaining to be opened based on this result */
	    maparg(prod, tg_filter_onset_nodes, &list);
	    /* insert new nodes at *bottom* of list of nodes to `open' */
	    *(find_last_ptr(&list)) = prod;
	}
    }
    return rslt;	/* all the nodes we processed */
}

static int tk_find_mag_at(TRACK *track, int type, void *arg)
{	/* node function track with greatest mag at given time */
    LSRCH_STR *mms = (LSRCH_STR *)arg;	/* threshold and answer */
    double time = mms->range.min;	/* (abuse of this field) */
    double frq, mag;

    tkFrqMagForT(track, time, &frq, &mag);
    if(mag > mms->range.max)  {
	mms->result = (LIST)track;
	mms->range.max = mag;
    }
    return 1;
}

TRACK *tkFindMaxMagAtT(TKLIST trx, double time, double *pmag)
{	/* return the element in the track list that has highest mag at T */
    LSRCH_STR 	lss;
    TRACK	*rslt;

    lss.result = NULL;
    lss.range.max = NOMAX;
    lss.range.min = time;	/* abuse of field - use to pass in time arg */
    /* apply to all the tracks.. */
    maparg((LIST)trx, (NodeFunc)tk_find_mag_at, &lss);
    rslt = (TRACK *)lss.result;
    *pmag = lss.range.max;
    return rslt;
}

static int tk_find_mag_over(TRACK *track, int type, void *arg)
{	/* node function track with greatest mag over given time range */
    LSRCH_STR *mms = (LSRCH_STR *)arg;	/* time range and answer */
    RANGE_STR *mrg = (RANGE_STR *)(mms+1);	/* yuk! */
    RANGE_STR frqrg, magrg;

    tkFrqMagRange(track, &(mms->range), &frqrg, &magrg);
/*    fprintf(stderr, "mag_ov: best %.2f this %.2f..%.2f\n", 
	    mrg->max, magrg.min, magrg.max);  */
    if(magrg.max < mrg->max)  {
	mms->result = (LIST)track;
    }
    rngUnion(&magrg, mrg, mrg);
    return 1;
}

TRACK *tkFindMaxMagOverT(TKLIST trx, double tmin, double tmax, double *pmag)
{	/* return the element in the track list that has highest peak mag
	   over the specified time range, and the value of that peak */
    LSRCH_STR 	lss[2];
    RANGE_STR	*mrg = (RANGE_STR *)&(lss[1]);	/* necess for tk_fin_mag_o */
    TRACK	*rslt;

    lss[0].result = NULL;
    lss[0].range.min = tmin;	/* range of LSS passes in trange */
    lss[0].range.max = tmax;
    mrg->min = NOMIN;
    mrg->max = NOMAX;		/* RANGE_STR suivant returns mag range */
    /* apply to all the tracks.. */
    maparg((LIST)trx, (NodeFunc)tk_find_mag_over, lss);
    rslt = (TRACK *)lss[0].result;
    *pmag = mrg->max;
    return rslt;
}

static int tk_ons_time_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by onset intersect */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double onsT = track->st.tavg;

    /* quick hack: check for start of track beyond end */
    if( track->st.tmin > rg->max)	return -1;
    /* if it's beyond the end, should abort search : assume
       input tracks sorted in onset order */

    if(onsT < rg->min || onsT >= rg->max)	/* wholly out of range */
	return 0;
    else
	return 1;
}

static int tk_ons_frq_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by onset intersect */
    RANGE_STR 	*rg = (RANGE_STR *)arg;
    double onsT = track->st.tavg;
    double mag, frq;

    tkFrqMagForT(track, onsT, &frq, &mag);

    if(frq < rg->min || frq >= rg->max)
	return 0;
    else
	return 1;
}

TKLIST tkGetOnsetIntersect(TKLIST trx, double tMin, double tMax,
			   double fMin, double fMax)
{	/* return tracks from the list whose ONSET times 
	   (as read from st.tavg) and ONSET FREQUENCIES
	   (as obtained by tkFrqMagForT(onset)) lie strictly within 
	   the specified range */
    TKLIST 	tkl = NULL;
    RANGE_STR	rg;

    /* first select by time */
    rg.min = tMin;	rg.max = tMax;
    tkl = (TKLIST)keep_if_abort((LIST)trx, (NodeFunc)tk_ons_time_pred, &rg);
    /* then select result by frq */
    trx = tkl;
    rg.min = fMin;	rg.max = fMax;
    tkl = (TKLIST)keep_if((LIST)trx, (NodeFunc)tk_ons_frq_pred, &rg);
    free_list(trx);	/* free the list allocated in last search */
    return tkl;
}

static int tk_narrow_tf_pred(TRACK *track, int type, void *arg)
{	/* predicate to allow filtering of track list by 
	   narrow frq intersect over specified range.
	   <arg> points to *two* range strs, first the time support, 
	   then the required frequency support */
    RANGE_STR 	*trg = (RANGE_STR *)arg;
    RANGE_STR 	*frg = trg+1;
    RANGE_STR	act_frg, act_mrg;
    int		rc = 0;

/*    fprintf(stderr, " +nar_tf: tk%d t %.6f..%.6f, f %.1f..%.1f, \n", 
	    track->id,
	    track->st.tmin, track->st.tmax, track->st.fmin, track->st.fmax);
 */
    /* tkFrqMagRange returns zero if no time overlap, in which case fail */
    if(tkFrqMagRange(track, trg, &act_frg, &act_mrg)) {
    /*  fprintf(stderr, "      +fovl: %.1f..%.1f (m %.2f..%.2f)\n", 
		act_frg.min, act_frg.max, act_mrg.min, act_mrg.max);
     */
	/* we got a frq range for the time range - does it overlap ? */
	if(rngIntersect(&act_frg, frg, NULL))  {
	    rc = 1;
	}
    }
    return rc;
}

TKLIST tkGetNarrowIntersect(TKLIST trx, double tMin, double tMax,
			   double fMin, double fMax)
{	/* Perform slower and more restrictive search for tracks: 
	   for each, calculate the actual frequency span 
	   *over the time range specified only* and pass the 
	   track only if it overlaps with spec range */
    TKLIST 	tkl = NULL;
    RANGE_STR	rgs[2];

    /* tk_narrow_tf_pred wants two adjacent range strs - first is time, 
       second is frq */
    rgs[0].min = tMin;	rgs[0].max = tMax;
    rgs[1].min = fMin;	rgs[1].max = fMax;
/*    fprintf(stderr, "tkGetNar: t %.6f..%.6f, f %.1f..%.1f:\n", 
	    tMin, tMax, fMin, fMax);  */
    tkl = (TKLIST)keep_if((LIST)trx, (NodeFunc)tk_narrow_tf_pred, rgs);
    return tkl;
}

#define ONS_LOOKBACK_FRQF	(1.4)	/* 1.4 makes 1 octave srch rg */
#define ONS_LOOKBACK_CYCS	(5.0)	/* look 5 cycles into past to 
					   find energy to guage onsetness */

static double tkCalcOnsScore(TRACK *tk, TKLIST focus)
{	/* does this track look like an onset i.e. is it louder than 
	   its near predecessors?  Returns +ve dB prominence, zero if 
	   no near predecessors found. */
    double score;
    double onsT = tkOnsetT(tk);
    double preT;
    double frq, mag, premag, tchr;
    double fmin, fmax;
    TKLIST pred, this, pred2;

    if(focus == NULL)	focus = allTrax;

    tkFrqMagForT(tk, onsT, &frq, &mag);
    tchr = 1.0/frq;
    fmin = frq/ONS_LOOKBACK_FRQF;	
    fmax = frq*ONS_LOOKBACK_FRQF;	 /* 1 octave range */
    preT = onsT- ONS_LOOKBACK_CYCS*tchr; /* cmp w/ tracks in preceding wdw */
    pred = tkGetIntersect(focus, preT, onsT, fmin, fmax);
    /* make sure all have onset time set */
    map((LIST)pred, (NodeFunc)tk_set_onset);
    /* now really filter for onset time */
/*  pred2 = tkGetOnsetIntersect(pred, preT, onsT, fmin, fmax); */
    /* no - do narrow filter for frq intersect */
    pred2 = tkGetNarrowIntersect(pred, preT, onsT, fmin, fmax);
    free_list(pred);
    pred = pred2;
    this = (TKLIST)new_node(NULL, (LIST)tk, T_TRACK);
    pred2= remove_list(pred, this);
    free_list(pred);
    free_list(this);
/*    RedrawTrax(pred2, TKM_SPECIAL);	*/
    /* find peak magnitude amongst predecessor tracks directly 
       before onsT */
    tkFindMaxMagOverT(pred2, preT, onsT-tchr, &premag); 
    if(premag == NOMAX)
	score = 0.0;
    else
	score = mag-premag;
/*    fprintf(stderr, "tkOnsScore(tk%d): margin %.1f\n", tk->id, score);  */
    return score;
}

double tkOnsScore(TRACK *tk, TKLIST focus)
{	/* Read back precalculated onset score */
    if(tk->st.dfdt == BIGMIN)	/* not set */
       tk->st.dfdt = tkCalcOnsScore(tk, focus);	/* so set it */
    return tk->st.dfdt;	/* held here temporarily */
}

static int tk_clear_onset_placeholder(TRACK *tk, int type, void *arg)
{	/* set all the dfdt fields (used to hold onset prominence 
	   when it is calculated) to a special value so we can 
	   recognize that it has not been yet calculated */
    tk->st.dfdt = BIGMIN;
    return 1;
}

/* (copied from db.noisy.c) */
#define PROM_THRESH	(0.0)	/* prominence >= this to be selected */

TKLIST tklOnsetsPrepare(TKLIST focus)
{	/* prepare a set of tracks for onset score calculation 
	   by clearing the stolen field.
	   THEN strip down to valid-onset candidates only, 
	   which are returned.  */
    TKLIST ofocus;

    fprintf(stderr, "tklOnsPre: enter..");	fflush(stderr);
    map((LIST)focus, (NodeFunc)tk_clear_onset_placeholder);
    fprintf(stderr, " cleared..");	fflush(stderr);
    /* return only those tracks exceeding the prominence threshold */
    ofocus = tklSelAbvProm(focus, PROM_THRESH);
    fprintf(stderr, " prom sel\n");
    return ofocus;
}

static int tk_onscore_pred(TRACK *tk, int type, LSRCH_STR *lss)
{	/* NodeFunc tests the tkOnsScore of the track in some range */
    TKLIST focus = lss->result;		/* stolen, abused field */
    double score = tkOnsScore(tk, focus);

    if(score < lss->range.min || score >= lss->range.max)
	return 0;
    else
	return 1;
}

static int tk_onscore_weight(TRACK *tk, int type, double *d)
{	/* NodeFunc to calculate weighted onset `scores' 
	   for range of tracks. <d> points to *two* doubles - 
	   the first is SUM(linavmag * score), second 
	   is SUM(linavmag).  After them, a pointer to the FOCUS tklist */
    TKLIST focus  = *(TKLIST *)(d+2);
    double linmag = LINOFDB(tk->st.mavg);
    double score  = tkOnsScore(tk, focus);
    double *sumlin = (d+1);

    DrawTrack(tk, T_TRACK, NULL);

    (*d) += linmag * score;
    (*sumlin) += linmag;
    return 1;
}

void CheckOnsetM(WINRCD **wrcd)
{   /* just allow us to run the track-onset checker interactively */
    TRACK *tk;
    TKLIST newTrax, focus;
    LSRCH_STR	lss;
    double d[3];

    noisyInitStatics(wrcd, 0);

    d[0] = 0.0;
    d[1] = 0.0;
    focus = tklOnsetsPrepare(allTrax);
    *(TKLIST *)&(d[2]) = focus;
    if(selTrax)  {
	RedrawTrax(selTrax, TKM_SPECIAL);
	SetTraxColor(TKM_NORM);	/* tk_onscore_weight currently overdraws */
	maparg((LIST)selTrax, (NodeFunc)tk_onscore_weight, d);
	d[0] /= d[1];
	fprintf(stderr, "Av score = %.2f\n", d[0]);
	lss.range.min = 0.0;	lss.range.max = BIGMAX;	/* select score > 0 */
	lss.result = focus;
	newTrax = keep_if((LIST)selTrax, (NodeFunc)tk_onscore_pred, &lss);
	trxNewSelect(newTrax);
    }
}

double tgOnsScore(TKGROUP *tg, TKLIST focus)
{	/* evaluate the `quality' score for an onset group.
	   This is defined as the weighted average `prominence' in 
	   dB between each onset point and other tracks in 
	   a small region near and just before it i.e. is it, in 
	   fact, an onset? */
    double score = 0.0;
    double d[3];

    if(focus == NULL)	focus = allTrax;

    d[0] = 0.0;
    d[1] = 0.0;
    *(TKLIST *)&(d[2]) = focus;

    maparg((LIST)tg->parts, (NodeFunc)tk_onscore_weight, d);
    score = d[0] / d[1];

/*    fprintf(stderr, "tgOnsSc: lasttk %.6f/%.1f tav %.6f tsd %.6f sc %.3f\n", 
	    tg->v.onset.latest->st.tmin, tg->v.onset.latest->st.favg, 
	    tg->st.tavg, tg->st.tsd, score);
  */
    /* for now, store score where it will appear in selpnl */
    tg->st.mmin = score;

    return score;
}

/* #define ONS_TSD_THRESH    (0.005)	/* spread of onsets below 5 ms */
#define ONS_TSD_THRESH	(0.050)		/* .. maybe 50ms  */
#define ONS_SCO_THRESH	(-0.5)		/* avg prominence not below -1/2dB */

static int tg_prune_bad_ons(TKGROUP *tg, int type, TKLIST focus)
{	/* filter predicate rejects onset groups with too great an 
	   s.d. amongst their onset times, or too low an average onset 
	   margin (dodgy this - had to hack to make it work for spch stop) */
    double score = tgOnsScore(tg, focus);

    if(score < ONS_SCO_THRESH || tg->st.tsd > ONS_TSD_THRESH)
	return 0;
    else
	return 1;
}

TGLIST tkMakeTglOnsets(TRACK *tk, TKLIST focus)
{	/* repeatedly apply tkMakeOnset, gather all results */
    TGLIST rslt, rs2, start;
    TKLIST tklst, withouts;
    TKGROUP *tg;

/*    fprintf(stderr, "tkMakeTglOnsets: based on tk%d (%.6f/%.1f)\n", 
	    tk->id, tk->st.tmin, tk->st.favg);	*/
    withouts = (TKLIST)new_node(NULL, (LIST)tk, T_TRACK);	/* 1st excl */
    tk_set_onset(tk, T_TRACK, NULL);
    tklst = (TKLIST)new_node(NULL, (LIST)tk, T_TRACK);
    tg = tkgrNew(tklst, TG_ONSET);
    tg->v.onset.latest = tk;
    start = (TGLIST)new_node(NULL, (LIST)tg, T_TKGROUP);
    rslt = tgOnsetClose(start, focus, &withouts);
/*    rslt = (TGLIST)eps_close(start, tgl_ons_fn, &withouts);	*/
/*    rslt = (TGLIST)depth_close(start, tgl_ons_fn, &withouts); */
    
    rs2 = keep_if(rslt, (NodeFunc)tg_prune_bad_ons, focus);
/*    free_list(rslt);
    rslt = rs2;
 */
    rs2 = tglRemoveNearDuplicates(rslt);
/*    rs2 = tglRemoveSubsets(rslt);	 */
    free_list(rslt);
    rslt = rs2;

    return rslt;
}

TGLIST tgl_onset_tk(TRACK *trk, int *ptype, TKLIST focus)
{	/* LNodeFunc to generate onset groups from tracks. 
	   Had best call tklOnsetsPrepare(focus) before 1st call.  */
    assert(*ptype == T_TRACK);
    *ptype = T_LIST;	/* returning a TGLIST */
    return tkMakeTglOnsets(trk, focus);
}

static int tk_report_frqmag(TRACK *tk, int type, void *time_range_arg)
{	/* mappable fn reports tracks */
    RANGE_STR	*time_rg = (RANGE_STR *)time_range_arg;
    RANGE_STR	frqRg, magRg;

    assert(type == T_TRACK);
    tkFrqMagRange(tk, time_rg, &frqRg, &magRg);
    fprintf(stderr, "RptRg: tk%4d(%.6f/%.1f): %6.1f:%6.1f Hz, %5.1f:%5.1f\n",
	    tk->id, tk->st.tmin, tk->st.favg, 
	    frqRg.min, frqRg.max, magRg.min, magRg.max);
    return 1;
}

extern TKLIST nsNewGrowNoise(double now, double frq, TKLIST trx, TKLIST focus);

TGLIST tgl_NewOnsets_tk(TRACK *tk, int *ptype, TKLIST focus)
{	/* Make onset groups by new method */
    TGLIST rslt, rs2, start;
    TKLIST tklst, tkl2, withouts;
    TKGROUP 	*tg = NULL;
    TGLIST    	tgl = NULL;
    double	tons, twin, frq, mag, minf, maxf;
    RANGE_STR	timeRg;

    fprintf(stderr, "tkNewOnsets: based on tk%d (%.6f/%.1f)\n", 
	    tk->id, tk->st.tmin, tk->st.favg);

    assert(*ptype == T_TRACK);
    *ptype = T_LIST;	/* mappend *dermands* this */
    
    nsGetMinMaxFrq(&minf, &maxf);
    tons = tkOnsetTFM(tk, &frq, &mag);
    twin = 1.0/frq;
/*    minf = frq/sqrt(2.0);
    maxf = frq*sqrt(2.0);	/* 1 octave range */
    /* find predecessors */
    timeRg.min = tons-2*twin;
    timeRg.max = tons-twin;
    timeRg.max = tons+ /* 100*twin */ 0.5;
    tklst = tkGetIntersect(focus, timeRg.min, timeRg.max, 
			   minf, maxf);
/*    maparg((LIST)tklst, (NodeFunc)tk_report_frqmag, &timeRg);	*/

    /* remove those significantly long in duration */
    tkl2 = SelTrxByDur(tklst, twin*200.0, 9999.9);
    RedrawTrax(tkl2, TKM_SPECIAL);
    free_list((LIST)tklst);  tklst = tkl2;

    tkl2 = nsNewGrowNoise(tons, frq, tklst, allTrax);
    free_list((LIST)tklst);  tklst = tkl2;

    if(tkl2)  {
	tg = tkgrNew(tklst, TG_ONSET);
	tgl = (TGLIST)new_node(NULL, (LIST)tg, T_TKGROUP);
    }
    return tgl;
}

int tg_pr_fn(TKGROUP *tkgroup, int type, void *data)
{
    tkgrPrint(tkgroup, ":");
    return 1;
}

TKLIST tkgrParts(TKGROUP *tg)
{
    return tg->parts;
}

TGLIST tglNewOnsets(TKLIST trx, TKLIST focus)
{
    TGLIST 	rslt = NULL, tgl;
    TGLIST	*ptail = &rslt;
    int 	type;
    TRACK	*tk;
    TKLIST	tl2;

    trx = copy_top(trx);
    while(trx) {
	tk = (TRACK *)CAR((LIST)trx);
	type = CARTYPE(trx);
	tgl = tgl_NewOnsets_tk(tk, &type, focus);
	*ptail = tgl;
	ptail = (TGLIST *)find_last_ptr((LIST *)ptail);
	trx = (TKLIST)cdr((LIST)trx);
	if(tgl)  {
	    tl2 = remove_list(trx, tkgrParts((TKGROUP *)(CAR(tgl))));
	    free_list((LIST)trx);	trx = tl2;
	}
    }
    free_list(trx);
    return rslt;
}

#else /* !CQFILE */   

void FindOnsetsM(WINRCD **wrcd)
{   /* menu entry - handle track selections & dispatch */
    /* dummy */
}

#endif /* CQFILE */   

#endif /* TRAX */   

