/*******************************************************************\
*   db.peaks.c							    *
*   pick & match spectral peaks for DSPbench	    		    *
*   02apr91 dpwe						    *
\*******************************************************************/

/********* switches ********/
/* #define DEBUG 1	/* */
#define CUBPHAS  1	/* comment out this flag for linear frq interp */
#define TPERFRAME .0004 /* 2.5 frames per millisecond? */
/*#define QUIETESTTRACK	-50.0	/* min. dB peak in track to display whole */
/* quietest track is now set up in db.xmain.c as minTrackDB - cl option */
#define DEBUGGRAPHS 1	/* display tracks in little windows */
/* #define PSOUTPUT    1   /* generate postscript polyline output */
/* define PSOUTPUT in Makefile if you like */

/********* includes ********/
#include <dpwelib.h>
#include <dsputil.h>
#include <fft.h>
#include <math.h>

#include <wndw.h>	/* different window interface for PS output */

#include "DSPbench.h"

void ClikZoomOverTrax(wrcd, pwhere, prec, btn)
    /* allow zooming over cq spectrograms (for now) */
    /* hacked version of zoom in/out from psyac stuff */
    /* click in spectrum drags out a marked area */
    WINRCD  **wrcd;	/* basic context */
    Rect    *prec;	/* outer rec of view */
    Point   *pwhere;	/* where clicked */
    int	    btn;	/* which button */
    {
static double rvT1, rvT2, rvMinF, rvMaxF;
static int rvVal = 0;	/* statics allow one level of undo in zoom */
    Rect r2;
    Rect orect;
#ifdef CQFILE
    CQTINFO	*cqi;
#else		
    AFTINFO	*cqi;
#endif		
    PVSTRUCT	*pvh = NULL;
    double	t1, t2, dur, oMinF, oMaxF;
    FRQVIEW	*frqv = (*wrcd)->w.snd.frqV;

    VTimeLims(wrcd, &t1, &t2, &dur);
#ifdef CQFILE
    cqi  = (*wrcd)->w.snd.cqi;
    pvh  = cqi->pvh;
#else /* !CQFILE */
    pvh  = (*wrcd)->w.snd.pvs; 
#endif /* CQFILE */
    oMinF = pvh->minFreq;
    oMaxF = pvh->maxFreq;
    orect = *prec;
    InsetRect(&orect, SBarWidth, 0);
    DragRects(&orect, &r2, pwhere, 
	      1, 3, 0 /* repNexT, axes, visType */); 
    if(r2.left == r2.right)  {
	if(btn==1 || rvVal == 0)  {	/* left btn restores full frq rg */
	    frqv->minFrq = oMinF;
	    frqv->maxFrq = oMaxF;
	    frqv->valid = 0;	/* necessary to force PixMap redraw */
	    RedrawWinBit(wrcd,Sviewid);
	} else {	/* right button reverts zoom view */
	    double tmp;
	    tmp = frqv->minFrq;	frqv->minFrq = rvMinF;	rvMinF = tmp;
	    tmp = frqv->maxFrq; frqv->maxFrq = rvMaxF;	rvMaxF = tmp;
	    tmp = t1; t1 = rvT1; rvT1 = tmp;
	    tmp = t2; t2 = rvT2; rvT2 = tmp;
	    UnmarkSelects(wrcd);
	    SetSTimeLims(wrcd, t1, t2, dur);
	    MarkSelects(wrcd);
	    SelToView(wrcd);
	}
    } else {
	/* save current zoom window in revert statics */
	if(frqv->valid)  {
	    rvT1 = t1;	rvT2 = t2;
	    rvMinF = frqv->minFrq;
	    rvMaxF = frqv->maxFrq;
	    rvVal  = 1;
	}
	if(btn==1) {		/* zoom in */
#ifdef CQFILE
	    double frqScale = log(frqv->maxFrq/frqv->minFrq)
		/(orect.top - orect.bottom);
#else
	    double frqScale = (frqv->maxFrq - frqv->minFrq)
		/(orect.top - orect.bottom);
#endif
	    double minF     = frqv->minFrq;
	    double timScale = (t2 - t1)
		/(orect.right - orect.left);
	    double minT     = t1;

#ifdef CQFILE
	    frqv->minFrq = minF * exp(frqScale * (r2.bottom - orect.bottom));
	    frqv->maxFrq = minF * exp(frqScale * (r2.top - orect.bottom));
#else
	    frqv->minFrq = minF + frqScale * (r2.bottom - orect.bottom);
	    frqv->maxFrq = minF + frqScale * (r2.top - orect.bottom);
#endif
	    t1 = minT + timScale * (r2.left - orect.left);
	    t2 = minT + timScale * (r2.right- orect.left);
	    if(t1 < 0)	t1 = 0;
	    if(t1 > dur)	t1 = dur;
	    if(t2 < 0)	t2 = 0;
	    if(t2 > dur)	t2 = dur;
	    frqv->valid = 0;	/* necessary to force PixMap redraw */
	    UnmarkSelects(wrcd);
	    SetSTimeLims(wrcd, t1, t2, dur);
	    MarkSelects(wrcd);
	    SelToView(wrcd);
	}
	else{			/* other button is inverse zoom out */
	    double minFrq, maxFrq, minTim, maxTim;
	    double pL, pU, fL, fU, mL, mU, pT, pB;
	    /* zoom out makes *current frq range* occupy *range under zoom* */

	    pL = ((double)r2.left - orect.left)/((double)(orect.right-orect.left));
	    pU = ((double)r2.right - orect.left)/((double)(orect.right-orect.left));
	    mL = t1;
	    mU = t2;
	    t1 = (pU*mL - pL*mU)/(pU-pL);
	    t2 = t1 + (mL-t1)/pL;
/* fprintf(stderr, "zm: t1 %.3f t2 %.3f t1' %.3f t2' %.3f\n",  
	mL, mU, t1, t2); */
	    if(t1 < 0.0)		t1 = 0.0;
	    if(t2 > dur)		t2 = dur;

	    pT = ((double)r2.top - orect.top)/((double)(orect.bottom-orect.top));
	    pB = ((double)r2.bottom - orect.top)/((double)(orect.bottom-orect.top));
#ifdef CQFILE
	    fL = log(frqv->minFrq/oMinF);
	    fU = log(frqv->maxFrq/oMinF);
#else
	    fL = frqv->minFrq - oMinF;
	    fU = frqv->maxFrq - oMinF;
#endif
	    maxFrq = (pB*fU - pT*fL)/(pB-pT); /* did algebra on paper */
	    minFrq = maxFrq + (fU-maxFrq)/pT; /* .. trust me! */
#ifdef CQFILE
	    minFrq = oMinF*exp(minFrq);
	    maxFrq = oMinF*exp(maxFrq);
#else
	    minFrq = oMinF + minFrq;
	    maxFrq = oMinF + maxFrq;
#endif
	    if(minFrq < oMinF)	minFrq = oMinF;
	    if(maxFrq > oMaxF)	maxFrq = oMaxF;
	    if(maxFrq < oMinF)	maxFrq = oMinF;
	    if(minFrq > oMaxF)	minFrq = oMaxF;
	    if(minFrq > maxFrq)	{ double tmp = minFrq; minFrq = maxFrq; maxFrq = tmp;  }
	    frqv->minFrq = minFrq;
	    frqv->maxFrq = maxFrq;

	    frqv->valid = 0;	/* necessary to force PixMap redraw */
	    UnmarkSelects(wrcd);
	    SetSTimeLims(wrcd, t1, t2, dur);
	    MarkSelects(wrcd);
	    SelToView(wrcd);
	}
    }
    return;
    }


#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/
#define PEAKWIN	 5	/* how many points to search for maxima over */
#define PEAKWIN2 (PEAKWIN/2)
#define BINTHRESH	2.0	/* arbitrary max (bin) separation for match */

/********* types ***********/

/***** static globals ******/

static TKLIST	lastTracks = NULL;
       TKLIST	allTrax    = NULL;
       TKLIST	selTrax    = NULL;

/********* needed to let trk stuff link *********/
/* print out data on trks in this window */
double dbgFrq = -2000.0;
double dbgFrqWin = 5.0;

/******** prototypes ********/
static int   PickPeaks PARG((float *frame, long len, PeakPos *peaks, 
			     int npeaks, int sign));
static double PeakDistance PARG((PeakPos *pk1, PeakPos *pk2));
static void  FitQuadtoPeak PARG((float *mags, int minc, float *phas, int pinc, 
				 int pkIndex, float *pFrq, float *pMag, 
				 float *pPha, int phaQ));

/******* implementations *****/

void DrawRestOfTrack(PKLIST node)
{
    PeakPos *peak;
    int     x = 0, lastx = 0;

    while(node && x>=0)
	{
	peak = node->peak;
	if( (x = peak->x) >= 0) 
	    {
	    LineTo(x, peak->y);
	    if((x - lastx) > 2)
		DrawX(x, peak->y, 1);
	    lastx = x;
	    }
	node = CDR(node);
	}
    }

PKLIST FindNonnegX(PKLIST pkl)
{	/* 'out-of-range' peaks are marked with their .x field < 0 */
    int 	x = -1;

    while(pkl && x<0)
	{
	if( (x = pkl->peak->x)<0)
	    pkl = CDR(pkl);
	}
    return pkl;
    }

double FindTrkMaxMag(TRACK *trk)
{	/* find largest mag in peaklist */
    double	p, largestPeak = -9999.0;
    PKLIST	pkl = trk->tktop;

    return trk->st.mmax;

    /** extra ignored **/

#ifdef NOT_USED
    while(pkl)
	{
	assert(CARTYPE(pkl)==T_PEAK);
	if( (p=pkl->peak->mag)>largestPeak)	largestPeak = p;
	pkl = CDR(pkl);
	}
    return largestPeak;
#endif /* NOT_USED */
    }

DrawX(x,y,s)
    int 	x;
    int 	y;
    int 	s;
    {
    MoveTo(x-s, y-s);
    LineTo(x+s, y+s);
    MoveTo(x-s, y+s);
    LineTo(x+s, y-s);	/* */
    MoveTo(x,y);	/* just to leave pen in nice place */
    }

int DrawTrack(TRACK *trk, int type, void *arg)
{	/* mapped over a TKLIST, this redraws one track whose peaks 
	   already have their x, y fields set to appropriate screen co-os */
    /* map_with_abort is a useless concept now that overly-quiet tracks 
       that are not displayed may be interleaved with the real thing.  So 
       we return 1 - don't abort */
    PKLIST	pkl;
    PKLIST	node;

    assert(trk && type == T_TRACK);
    pkl = trk->tktop;
    if( trk->flags == -1)	/* ThisTkUnrendered - but not final */
	return 0;
    else if( trk->flags == -2) {  /* ThisTkUnrendered - and no more follow */
/*	fprintf(stderr, "DrawTrack:finished\n");	*/
	return -1;
    }
/*    assert(node = FindNonnegX(pkl));  */
    /* despite fixing the flags in exspectSpec, they will still be wrong 
       after a redraw.  Quick fix is to deal here */
    if(node = FindNonnegX(pkl))  {
	MoveTo(node->peak->x, node->peak->y);
	if(CDR(node))
	    DrawRestOfTrack(CDR(node));
	else
	    DrawX(node->peak->x, node->peak->y, 1);
	return 1;
    } else {
	/* trk->flags = -1; /* should have been! */
	/* fprintf(stderr, "DrawTrack: invisible fixed up\n"); */
	/* DON'T fix up for now, since expec tracks will NEVER 
	   get unfixed */
	return 0;
    }
}

TotalResetTracks(WINRCD **wrcd)
{	/* reload the damn track file then redraw the shortview */
    if(selTrax)  {
	free_list(selTrax);
	selTrax = NULL;
    }
    if(allTrax)  {
	free_homo_list((LIST)allTrax, (NodeFunc)tkFreeIncPeaks);
	allTrax = NULL;
    }
    LoadTrackFile(wrcd);
    RedrawWinBit(wrcd, Sviewid);
}

void SetTraxColor(int mode)
{	/* sets colour for drawing tracks according to the nominal modes - 
	   TKM_NORM, TKM_SELECT, TKM_SPECIAL */
    if(mode == TKM_NORM)
	SetBlue();
    else if(mode == TKM_SELECT)
	SetGreen();
    else if(mode == TKM_SPECIAL)
	SetRed();
    /* else don't touch */
}

void RedrawTrax(TKLIST trax, int color)
{	/* draws all the tracks in the list (in prevailing color) */
    SetTraxColor(color);
    if(trax) 
	map_with_abort((LIST)trax, (NodeFunc)DrawTrack);	
    if(color != TKM_NONE)	/* we messed with color, so reset it */
	SetBlack();
}

ResetTracks(wrcd)
    WINRCD **wrcd;
    {	/* just resets track selection as opposed to reloading file */
#if DEBUG
    fprintf(stderr, "ResetTrax:");	fflush(stderr);
#endif
    if(selTrax != NULL)	/* some tracks currently selected - lose them all */
	{
	RedrawTrax(selTrax, 1);
	free_list(selTrax);
	selTrax = NULL;
	}
    else{		/* nothing to reset, so select all instead */
	selTrax = copy_top(allTrax);
	RedrawTrax(selTrax, 2);
	fprintf(stderr, "All top_size %d, sel top_size %d\n",
		top_size(allTrax), top_size(selTrax));
	}
    }

int TestTrackByPos(NODE *nd, int type, void *data)
{   /* test for track lying near a point on scrn */
    /* (<type> must be T_TRACK or somesuch) */
    TRACK	*trk = (TRACK *)nd;
    Point	*ppt = (Point *)data;
    PKLIST	pkl, node;
    Point	pt, opt;
    Rect	rec;
    int 	found = 0, done = 0;

    assert(type == T_TRACK);
    pkl = trk->tktop;
    if( trk->flags == -1)	/* ThisTkUnrendered, others may follow */
	return 0;
    else if( trk->flags == -2) {  /* ThisTkUnrendered - and no more follow */
/*	fprintf(stderr, "TestTrackByPos:finished\n");	*/
	return -1;
    }
    assert(node = FindNonnegX(pkl));
    opt.h = node->peak->x;	opt.v = node->peak->y;
    /* if this track *starts* to the right of pt in Q, forget search */
    if(opt.h > ppt->h)	return -1;
    node = CDR(node);
    while(node && !found && opt.h >= 0 && !done)  {
	pt.h = node->peak->x; pt.v = node->peak->y;
	if(pt.h >= 0) {
	    PtsToRect(&pt, &opt, &rec);
	    if(PtInRect(*ppt, &rec))
		found = 1;
	    else
		node = CDR(node);
	    if(pt.h > ppt->h)	
		done = 1;	/* gone past this pt - forget it. */
	}
	opt = pt;
    }
    return found;	/* indicate /where/ we found?? */
    /* return code of -1 means `abort the search' (not valid!) */
}

int TestTrackInRect(trk, type, prec)  
    /* test for track lying in a rect on scrn */
    TRACK	*trk;
    int		type;	/* (must be T_TRACK or somesuch) */
    Rect	*prec;
    {	/* Only checks for an endpoint within rect - doesn't do intersection */
    PKLIST	node, pkl;
    Point	pt;
    int 	found = 0, done = 0;

    assert(type == T_TRACK);
    pkl = trk->tktop;
    if( trk->flags == -1)	/* ThisTkUnrendered, may be more */
	return 0;
    else if( trk->flags == -2) {  /* ThisTkUnrendered - and no more follow */
/*	fprintf(stderr, "TestTrackInRect:finished\n");	*/
	return -1;
    }
    assert(node = FindNonnegX(pkl));
    /* don't search if first point of track is to right of search box */
    if(node->peak->x > prec->right)		return -1;
    pt.h = 1;
    while(node && !found && pt.h >= 0 && !done)
	{
	pt.h = node->peak->x;		pt.v = node->peak->y;
	if(pt.h < 0 			/* track has become invisible again */
	   || pt.h > prec->right)	/* gone past end of rect */
	    done = 1;	
	else if( PtInRect(pt, prec) )
	    found = 1;
	node = CDR(node);
	}
    return found;	/* indicate /where/ we found?? */
    } /* returns -1 if 1st h in tail was beyond rect but on screen -> missed */

int TimeFreqComp(NODE **pnode1, NODE **pnode2)
{	/* For insert_by_cmp.
	   Compare two TRACK nodes.  
	   Node 1 will be inserted before node2 if returns -1 */
    TRACK *trk1, *trk2;

    assert(CARTYPE(*pnode1) == T_TRACK && CARTYPE(*pnode2) == T_TRACK);
    trk1 = (TRACK *)CAR(*pnode1);
    trk2 = (TRACK *)CAR(*pnode2);
    return tkCmpTrxByT((NODE **)&trk1, (NODE **)&trk2);
}

TKLIST *AddRemoveSelTrax(TKLIST trax, TKLIST *srchFrom)
{	/* add or remove all the selected trax from the selTrax list.
	   i.e. toggle their membership in selTrax.
	   srchFrom is a point in selTrax which is the earliest we 
	   need to look at (or NULL) */
    TRACK	*theTrack;
    TKLIST	selPt, *ppar, txcpy, txnode;
    int 	type;
    int		colour = TKM_NONE;

    txcpy = copy_top(trax);
    qsort_list((LIST)txcpy, tkCmpTrxByT);
    txnode = txcpy;
    while(txnode)  {
	assert((type = cartype((LIST)txnode)) == T_TRACK);
	theTrack = (TRACK *)CAR(txnode);
	txnode = (TKLIST)CDR(txnode);
	if(srchFrom == NULL)	srchFrom = &selTrax;
	if(selPt
	   =(TKLIST)search_by_fn((LIST)*srchFrom,
				 car_match, (char *)theTrack)) {
	    /* already selected - unselect it */
	    if(colour != TKM_NORM) { 
		colour = TKM_NORM;  SetTraxColor(colour);  
	    }
	    DrawTrack(theTrack, type, NULL);
	    ppar = (TKLIST *)find_par_ptr((LIST *)srchFrom, (LIST)selPt);
	    free_next_ptr((LIST *)ppar);
	} else {
	    if(colour != TKM_SELECT) { 
		colour = TKM_SELECT; SetTraxColor(colour); 
	    }
	    DrawTrack(theTrack, type, NULL);
	    ppar = (TKLIST *)next_ptr(insert_by_comp((LIST *)srchFrom, 
						     new_node(NULL, 
							      (LIST)theTrack, 
							      type),
						     TimeFreqComp));
	}
    }
    free_list(txcpy);
    return ppar;
}


int tk_add_select(LIST data, int type, void *arg)
{	/* function for maparg, adds new tracknodes in right place in 
	   tracklist */
    LIST **ppsel = (LIST **)arg;	/* actually &(TKLIST *) */
    TRACK *theTrack = (TRACK *)data;

    assert(type == T_TRACK);
    *ppsel = next_ptr(insert_by_comp(*ppsel, new_node(NULL, (LIST)theTrack, 
						      T_TRACK), 
				     TimeFreqComp));
    return 1;
}

void MergeSelection(TKLIST new)
{	/* insert the new nodes into the selection list */
    TKLIST *tkptr = &selTrax;	/* running index into list used by tk_add.. */

    maparg(new, tk_add_select, &tkptr);
}

void ClikOverTrax(wrcd, pwhere, prect, btn) /* btn down in trk'd sgm region */
    WINRCD	**wrcd;
    Point	*pwhere;
    Rect	*prect;
    int 	btn;
    {
    TKLIST	trkl, strkl, newSel;
    TRACK	*trk;
    Rect	rect;
    int  	colour = 0;	/* efficiency hack in AddRemove.. */

    if(ShiftIsDown())  {
	ClikZoomOverTrax(wrcd, pwhere, prect, btn);
	return;
    }

    DragRects(prect, &rect, pwhere, 1, 3, 0);
    if(rect.left == rect.right && rect.top == rect.bottom)  {
	/* just a click - no drag :  search by point */
	trkl = search_with_abort(allTrax, TestTrackByPos, pwhere);
	if(trkl)  {	/* found a track */
	    trk = (TRACK *)CAR(trkl);
	    trkl = (TKLIST)new_node(NULL, (LIST)trk, T_TRACK);
	    AddRemoveSelTrax(trkl, NULL);
	    free_list(trkl);
	}
    } else {
	LIST *srchFrom = &selTrax;	/* quicken search in AddRemove */

	/* find ALL the tracks involved */
	trkl = (TKLIST)keep_if_abort((LIST)allTrax, TestTrackInRect, &rect);
/*	fprintf(stderr, "Found %ld tracks total\n", top_size((LIST)trkl)); */
	/* find which of those are currently selected */
	strkl = (TKLIST)keep_list((LIST)selTrax, trkl);
/*	fprintf(stderr,"%ld were already selected\n",top_size((LIST)strkl)); */

	/* unselect the selected */
	newSel = (TKLIST)remove_list((LIST)selTrax, strkl);
	RedrawTrax(strkl, TKM_NORM);
	free_list((LIST)selTrax);
	selTrax = newSel;
/*	fprintf(stderr, "Removing them left %ld\n",top_size((LIST)selTrax)); */

	/* .. and select the unselected (i.e. trkl - strkl) */
	newSel = (TKLIST)remove_list(trkl, strkl);
	free_list(strkl);
	free_list(trkl);
	MergeSelection(newSel);
	RedrawTrax(newSel, TKM_SELECT);
	free_list(newSel);
/*	fprintf(stderr, "Adding new left %ld\n", top_size((LIST)selTrax)); */
    }
    SetBlack();
/*    PrintPeakList(selTrax,"Selected");	 */
    }

void SynthFromTracks(wrcd)
    WINRCD **wrcd;
    {
    CQTINFO	*cqi;

    cqi  = (*wrcd)->w.snd.cqi;
    if(cqi == NULL)
	return;
    if(selTrax == NULL)
	return;
    tkWriteSynth(selTrax, (double)cqi->pvh->samplingRate, 1 /* playq */);
    /* tkWriteSynth is in ../trk/tkplay.c */
    }

void RenderGoodTracks(TKLIST trx, Rect *rec, 
		      double sttTime, double endTime, 
		      double lowFrq, double hiFrq, WIND *wndw) 
{	/* applies the fitness test to each track, and does the 
	   calculation of where it lives in the window, or marks it 
	   as undrawn */
    TKLIST	remTrax = trx;
    TRACK	*thisTk;
extern double	minTrackDB;	/* cl option set up in db.xmain.c */

    while(remTrax)  {
	thisTk  = (TRACK *)CAR(remTrax);
	remTrax = CDR(remTrax);
	if( thisTk->tktop->peak->tim < endTime)
	    if(FindTrkMaxMag(thisTk)> minTrackDB)  {
		RenderTrackOnSgm(thisTk, rec, sttTime, endTime,
				 lowFrq, hiFrq, wndw);
	    } else {
		thisTk->flags = -1;	/* just missed rendition */
	    }
	else
	    MarkThisTrackUnrendered(thisTk);
    }
}

DrawTraxOnSgm(wrcd, rec, frqv, wndw)
    WINRCD **wrcd;
    Rect *rec;
    FRQVIEW *frqv;
    WIND	*wndw;
    {
    TKLIST	remTrax;
    TRACK 	*thisTk;
    double	lowFrq, hiFrq;
    CQTINFO	*cqi;
    double  	sttTime, endTime, duration;
    int 	done = 0;

#if DEBUG
    fprintf(stderr, "DrwTxOnSgm:");	fflush(stderr);
#endif
    cqi  = (*wrcd)->w.snd.cqi;
    lowFrq = cqi->pvh->minFreq;
    hiFrq  = cqi->pvh->maxFreq; 	
    if(frqv->maxFrq > frqv->minFrq)  {	/* new: allow frq zooms */
	lowFrq = frqv->minFrq;
	hiFrq = frqv->maxFrq;
    }
    VTimeLims(wrcd, &sttTime, &endTime, &duration);
#if DEBUG
    fprintf(stderr, "st %.4f, ed %.4f, du %.4f", sttTime, endTime, duration);
    fflush(stderr);
#endif
    SetTraxColor(TKM_NORM);
    if(wndw && selTrax)  {
	/* if going to draw both selected and unselected tracks to 
	   postscript, draw *non* selected in gray (blue) */
	WSetGMode(wndw, WGMD_BLUE, WGMD_BLACK, WGMD_COPY, 1, WGMD_SOLID);
    }
    RenderGoodTracks(allTrax, rec, sttTime, endTime, lowFrq, hiFrq, wndw);
    if(wndw)  {
	/* output to pseudo-ps: rerun the render routine (which includes 
	   driving the ps output) but on the selected tracks only, 
	   and set the pencolour to white */
	WSetGMode(wndw, WGMD_BLACK, WGMD_WHITE, WGMD_COPY, 1, WGMD_SOLID);
	RenderGoodTracks(selTrax, rec, sttTime, endTime, lowFrq, hiFrq, wndw);
    } else {
	RedrawTrax(selTrax, TKM_SELECT);
    }
    SetBlack();

    /* Add exspec tails, if any (see db.exspec.c) 1992oct28 */
    exDrawTails(wrcd);
}

int pkl_mark_unrendered(LIST data, int type, void *arg)
{	/* map-able function does the 'mark unrendered' job - 
	   i.e. sets all x-values to -1 */
    PeakPos *pk;

    assert(type == T_PEAK);
    pk = (PeakPos *)data;
    pk->x = pk->y = -1;
    return 1;
}

MarkThisTrackUnrendered(TRACK *trk)
{
    trk->flags = -2;
/*    map((LIST)trk->tktop, pkl_mark_unrendered);	*/
}

#define MAXSEGS          2000    /* num segs for ps output */

RenderTrackOnSgm(trk, rec, start, end, lowFrq, hiFrq, wind)
    TRACK	*trk;
    Rect 	*rec;
    FLOATARG 	start;
    FLOATARG 	end;
    FLOATARG 	lowFrq;
    FLOATARG 	hiFrq;
    WIND	*wind;
{
    PKLIST	remPeaks;
    PeakPos	*thisPk;
    int 	width, height, x, y, lastx, lasty;
    double	duration, frqRange, thisTime = 0.0;
    double	frqScale, thisFrq;
    /* for new wndw.h postscript output - dpwe 1993mar13 */
static	int	psiOp[MAXSEGS][2];
    int 	psiIx = 0;

    duration = end-start;
    frqRange = hiFrq-lowFrq;
    width = rec->right - rec->left;
    height = rec->bottom - rec->top;
    frqScale = height/log(hiFrq/lowFrq);
    remPeaks = trk->tktop;
    if(remPeaks)	thisTime = ((PeakPos *)CAR((LIST)remPeaks))->tim;
#if DEBUG
    fprintf(stderr, "R(%.4f)", thisTime);	fflush(stderr);
#endif
    lastx = -1;		lasty = -1;
    trk->flags = -1;	/* if lasts, track narrowly missed rendition */
    while(remPeaks && thisTime < end)  {
	thisPk = (PeakPos *)car((LIST)remPeaks);
	remPeaks = (PKLIST)cdr((LIST)remPeaks);
	thisTime = thisPk->tim;
	thisFrq = thisPk->frq;
	if(thisTime >= start && thisTime < end 
	   && thisFrq >= lowFrq && thisFrq < hiFrq)  {
	    trk->flags = 0;	/* MarkThisTrackRendered */
	    x = rec->left+((double)width*(thisTime-start)/duration);
	    y = rec->bottom-(log(thisPk->frq/lowFrq)*frqScale);
	    if(x != lastx)  {	/* a new-ish point? */
		psiOp[psiIx][0] = x;
		psiOp[psiIx][1] = y;
		++psiIx;
	    }
	    if(wind == NULL)  {	/* we will write to screen */
		if(lastx < 0)  {	/* first point in line */
		    DrawX(x,y,1);	
		} else {
		    LineTo(x,y);
		    if( (x-lastx)>2)
			DrawX(x,y,1);	
		}
	    }
	    lastx = x;	lasty = y;
	}
	if(wind==NULL)	{  /* drawing to screen, save co-os for redraw */
	    thisPk->x = lastx; 	thisPk->y = lasty;
	}
    }
    if(wind && psiIx > 2)
	WPolyLine(wind, psiOp, psiIx);

    /* mark rest of track as not on screen */
    map((LIST)remPeaks, pkl_mark_unrendered);
}

#ifdef DEBUGGRAPHS
#define MAXWINS	6
static WINDAT *magWin[MAXWINS];
static WINDAT *frqWin[MAXWINS];
static float  *ddata = NULL;
static int allocWin = 0;
#define DBGASIZE 1000
#endif

int DoTrkDisp(WINDAT *wndw, TRACK *track, int FnotM)
{	/* display the frequency <FnotM==1> or magnitude of a track in 
	   a `window' window <wndw> */
    PKLIST pnode;
    PeakPos *pk, *lpk;
    double t0, t1;
    long n;
    TrackStats	ts;
    int rc = 0;

    if(ddata == NULL)	{
	ddata = (float *)Mymalloc((long)DBGASIZE, sizeof(float), 
						"DoTrkDisp");
    }
    if(track)  {
	/* chop off first and last nodes since they are artificial 
	   fade in/out additions */
	pnode = track->tktop;
	if(top_size((LIST)pnode) < 3)	return 0;  /* too short - can't disp */
	pnode = CDR(pnode);	/* trim first */
	pk = pnode->peak;
	tkStReset(&ts, pk);
	t0 = pk->tim;		/* start time */
	n = 0;
	while(CDR(pnode) && n<DBGASIZE)	/* i.e. until one before last */
	{
#ifdef MATLABSTATSOUT
	    tkStUpdate(&ts, pk, lpk);
	    lpk = pk;
	    tkStBoildown(&ts);
	    /* printout unadorned numbers for MATLAB input */
	    printf("%8.6f %6.1f %6.1f %6.1f %6.1f %6.2f %6.2f %6.2f\n", 
		   pk->tim, pk->frq, ts.favg, ts.fsd, ts.dfdt, 
		   pk->mag, ts.mavg, ts.msd);
	    tkStBoilup(&ts);
#endif /* MATLABSTATSOUT */
	    if(FnotM)
		ddata[n] = pk->frq;
	    else
		ddata[n] = pk->mag;
	    t1       = pk->tim;   /* end time */
	    ++n;	
	 /*   assert(n<DBGASIZE);	/* avoid overrunning display */
	    pnode = CDR(pnode);
	    pk = pnode->peak;	/* cdr checked nonnull at top */
	}
	if(n==DBGASIZE)  {
	    fprintf(stderr,"WARNING: display limited to %d pts\n", n);
	}
	DoDisp(wndw, ddata, n);
	SetXLimits(wndw, t0, t1);
	rc = 1;
    }
    return rc;
}

void ShowCompleteTrax(TKLIST trax)
{
    long n;
    PKLIST pnode;
    double t0, t1;
    int i, ntrax;
    char winname[32];
    TRACK *track;
    TrackStats	ts;
    PeakPos	*pk, *lpk = NULL;

#ifdef DEBUGGRAPHS
    if( (ntrax = top_size((LIST)trax)) == 0)	return;
    if(ddata == NULL)	{
	ddata = (float *)Mymalloc((long)DBGASIZE, sizeof(float), 
						"ShowComplTk");
    }
    while(allocWin < ntrax && allocWin < MAXWINS)  {
	/* only instance the subwindows we need */
	sprintf(winname, "Magnitude %d", allocWin);
	magWin[allocWin] = NewWin(winname, 0);
	if(allocWin)	wdLink(frqWin[allocWin-1], magWin[allocWin]);
	sprintf(winname, "Frequency %d", allocWin);
	frqWin[allocWin] = NewWin(winname, 0);
	wdLink(magWin[allocWin], frqWin[allocWin]);
	++allocWin;
    }

    i = 0;
    track = (TRACK *)CAR(trax);
    while(track && i < allocWin)  {
	DoTrkDisp(magWin[i], track, 0);
	DoTrkDisp(frqWin[i], track, 1);
	trax = (TKLIST)cdr((LIST)trax);
	if(trax) track = (TRACK *)CAR(trax);	
	else 	 track = NULL;
	++i;
    }
#endif /* DEBUGGRAPHS */
}

PrintSelTrax(wrcd)		/* detail selected tracks */
    WINRCD	**wrcd;
    {
#ifdef DEBUGGRAPHS
    if(selTrax)
	ShowCompleteTrax(selTrax);
#else
#endif /* DEBUGGRAPHS */
    tklPrint(selTrax,"Selected tracks:");	 /* */
    }

#else /* CQFILE */

ClikOverTrax(wrcd, pwhere, prect, btn)
    WINRCD	**wrcd;
    Point	*pwhere;
    Rect	*prect;
    int 	btn;
    {
    }
ResetTracks(wrcd)
    WINRCD **wrcd;
    {
    }
SynthFromTracks(wrcd)
    WINRCD **wrcd;
    {
    }
LoadTrackFile(wrcd)
    WINRCD	**wrcd;
    {
    }
SaveTrackFile(wrcd)
    WINRCD	**wrcd;
    {
    }
PrintSelTrax(wrcd)
    WINRCD	**wrcd;
    {
    }
DrawTraxOnSgm(wrcd, rec, frqv, outFile)
    WINRCD **wrcd;
    Rect *rec;
    FRQVIEW *frqv;
    FILE *outFile;
    {
    }

#endif /* CQFILE */

#endif /* TRAX */
