/*******************************************************************\
*   db.psout.c							    *
*   Finally, an attempt to do postscript output correctly	    *
*   Provides a function, bound to a 'print' button, that generates  *
*   a complete postscript image based on the current dspB display.  *
*   Well, we'll see.						    *
*   1993mar12 dpwe						    *
\*******************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <vecops.h>
#include <arrayops.h>	/* don't see that one often .. */
#include <math.h>
#include <wndw.h>
#include <idrawop.h>	/* for graymap end-run and font defs */

#include "DSPbench.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/

/******** prototypes ********/

/******* implementations *****/

void Recontrast(ByArray *ar, int zero, int full)
{	/* rescale bytes in array so that <= zero is 0, and >= full
	   is BYTEMAX etc */
    long npts = ar->rows*ar->cols;
    MyByte *data = ar->data;
    int  v;
    double slope, intercept;

    slope = BYTEMAX/(full-zero);
    intercept = -slope*zero;

    while(npts--)  {
	v = *data;
	v = intercept + slope*v;
	if(v < 0)	v = 0;
	if(v > BYTEMAX)	v = BYTEMAX;
	*data++ = v;
    }
}

#define TIMEWAVEFORM  /* draw the time waveform on the top of the pic */

/* sizes of postscript spectrogram .. ? */
/* #define PSSGM_W	600	
   #define PSSGB_H	300  */
#define PSSGM_W	300
#define PSSGB_H 200
#define TWAVFMH	100
#ifdef TIMEWAVEFORM
#define PSSGM_H (PSSGB_H+TWAVFMH)
#else
#define PSSGM_H PSSGB_H
#endif

extern Display *disp;
extern int     scrnNum;

void Rect2RECT(Rect *rect, RECT *rECT)
{	/* mac and unix style rectangles - oh no! */
    rECT->x = rect->left;
    rECT->y = rect->top;
    rECT->width = rect->right - rect->left;
    rECT->height = rect->bottom - rect->top;
}

extern int traxRevealed;	/* defined in db.sgram.c */

void PSoutput(WINRCD **wrcd)
{	/* generate a nice ps file */
    char 	psFileName[80], tpltFileName[80], *tmps;
    FILE	*psFile, *tpltFile;
    Rect	psRect, r2, r3;
    RECT	psRECT;
    char	*getenv();
static    WIND	*wind = NULL;
    int		x = 50;		/* initial window parameters */
    int 	y = 50;
    int 	w = PSSGM_W+100;
    int 	h = PSSGM_H+50;
    int		sgw = PSSGM_W;
    int		sgh = PSSGM_H;
static ByArray	*ba = NULL;
    double	t1, t2, dur;
    int		tix[1000], ntix;
    FRQVIEW	*frqv = (*wrcd)->w.snd.frqV;
    /* for ShowSound.. */
    short	*stt, *end;
    int	    chans = ((*wrcd)->w.snd.snd)->channels;
    int     scale = (*wrcd)->w.snd.scale;
    int	    mode  = (*wrcd)->w.snd.vMode;
    int 	datatype;
    int		mx, my, mw, mh;
    int 	fix;

    /* now create the new file */
    SubstExtn((*wrcd)->theFileName, ".ps", psFileName);

    if(wind!=NULL)  {
	FreeWind(wind);
	wind==NULL;
    }
    wind = NewXWind(x, y, w, h, psFileName, NULL);
    /* PS file is opened later, along with BB info */

    /* calc bounding box & generate spectrogram graymap */
    psRect.left = 40;	psRect.right = psRect.left + sgw;
    psRect.top = 30;	psRect.bottom = psRect.top + sgh;
    r3 = psRect;
#ifdef TIMEWAVEFORM
    r3.top += TWAVFMH;	/* r3 is where to put actual spectrogram */
    fix = 0;
#else /* TIMEWAVEFORM */
    fix = 1.5 * WTextHeight(wind, "filename title");
#endif /* TIMEWAVEFORM */
    Rect2RECT(&r3, &psRECT);
    mx = psRect.left - WTextWidth(wind, " 10000");
    my = psRect.top - fix;
    mh = -my + psRect.bottom + WTextHeight(wind, "0.12");
    mw = -mx + psRect.right + WTextWidth(wind, "time/s")/2;
    /* now open file with bounding box */
    idOpen(wind, psFileName, mx, my, mw, mh);
	    
    if(ba == NULL)
	ba = NewByteA(sgh, sgw);
    if( ba )  {
	ActualShowSgram(wrcd, frqv, NULL, NULL, ba);
	Recontrast(ba, 104, 224);	/* up the contrast .. ? */
	WColorArray(wind, &psRECT, ba);	/* 1993sep29 */
	WSetBlack(wind);
    }
    WFrameRect(wind, &psRECT);

#ifdef TRAX
    if(traxRevealed)  {
	DrawTraxOnSgm(wrcd, &r3, frqv, wind);
    }
#endif /* def TRAX */

    /* add time lines */
    idSetFont(wind, IDF_HELV_MEDIUM);
    r2 = r3;
    r2.top = r2.bottom;
    r2.bottom = r2.top + 1.5*WTextHeight(wind, "0.00");
    Rect2RECT(&r2, &psRECT);
    VTimeLims(wrcd, &t1, &t2, &dur);
    WAddTimeLine(wind, &psRECT, t1, t2, 0, 1, 1, tix, &ntix);
    idSetFont(wind, IDF_HELV_SMALL);
    r2.left = r2.right;	/* very narrow */
    Rect2RECT(&r2, &psRECT);
    WWriteIn(wind, &psRECT, "time/s", CALIGN);
    
    idSetFont(wind, IDF_HELV_MEDIUM);
    r2 = r3;
    r2.right = r2.left;
    r2.left = r2.right - WTextWidth(wind, "10000");
    Rect2RECT(&r2, &psRECT);
#ifdef CQFILE
    WAddLogTimeLine(wind, &psRECT, frqv->minFrq, frqv->maxFrq, 
		    1, 0, 0, tix, &ntix);
#else /* AFT file */
    WAddTimeLine(wind, &psRECT, frqv->minFrq, frqv->maxFrq, 
		 1, 0, 0, tix, &ntix);
#endif /* CQFILE */
    r2.bottom = r2.top;
    idSetFont(wind, IDF_HELV_SMALL);
/*    r2.top = r2.bottom - WTextHeight(wind, "frq/Hz");  */
/*    r2.bottom = r2.top + WTextHeight(wind, "frq/Hz");  */
    Rect2RECT(&r2, &psRECT);
    WWriteIn(wind, &psRECT, "frq/Hz", LALIGN);

#ifdef TIMEWAVEFORM
    /* draw time waveform above */
    r2 = psRect;
    r2.bottom = r3.top;	
    datatype = GetSndBitePtrs((*wrcd)->w.snd.view, &stt, &end);
    SetLineWidth(wind, 0.5);
    ShowSound(r2,stt,end,chans,mode,scale,datatype,wind);
    SetLineWidth(wind, 1.0);
    Rect2RECT(&r2, &psRECT);
    WFrameRect(wind, &psRECT);
#endif /* TIMEWAVEFORM */

    /* add a title */
    r2 = psRect; 
    r2.top -= fix;
    r2.bottom = r2.top + 1.5*WTextHeight(wind, "Title");
    r2.left += WTextWidth(wind, "  ");
    Rect2RECT(&r2, &psRECT);
    idSetFont(wind, IDF_HELV_MEDIUM);
    WWriteIn(wind, &psRECT, (*wrcd)->theFileName, LALIGN);

    /* copy rest of template file */
 theend:
    idClose(wind);
    fprintf(stderr, "ps written to '%s'\n", psFileName);
    return;
}

void MatlabOutput(WINRCD **wrcd)
{	/* generate 2D ascii array suitable for reading into matlab */
    char 	mFileName[80];
    FILE	*mFile;
static ByArray	*ba = NULL;
    FRQVIEW	*frqv = (*wrcd)->w.snd.frqV;
    int		sgw = 1000;
    int		sgh = 96;
    int		x,y;

    /* now create the new file */
    SubstExtn((*wrcd)->theFileName, ".mat", mFileName);
    if( (mFile = fopen(mFileName, "w")) == NULL )  {
	fprintf(stderr, "MatOut: couldn't create '%s' for output\n", 
		mFileName);
	return;
    }
    /* generate spectrogram graymap */
    if(ba == NULL)
	ba = NewByteA(sgh, sgw);
    if( ba )  {
	ActualShowSgram(wrcd, frqv, NULL, NULL, ba);
	/* write array as decimal out to an ascii file,
	   with each *line* being one column of the image 
	   (i.e. time is outer dimension) */
	for(x = 0; x<sgw; ++x)  {
	    for(y = 0; y<sgh; ++y)  {
		fprintf(mFile, "%3d ", ba->data[x+sgw*y]);
	    }
	    fprintf(mFile, "\n");
	}
    }
    fprintf(stderr, "matlab file '%s' written (%ld bytes)\n", 
	    mFileName, ftell(mFile));
 theend:
    fclose(mFile);
    return;
}

void PrintBtn(WINRCD **wrcd, int btn)
{	/* dispatch click on print button */
    if(btn==1)  {
	PSoutput(wrcd);
    } else if(btn==2)  {
	MatlabOutput(wrcd);
    }
}

