/*******************************************************************\
*   db.psyac.c							    *
*   Psychoacoustic masking display for dspB			    *
*   1992jul03  dpwe						    *
\*******************************************************************/

#include <stdio.h>
#include <string.h>
#include <dsputil.h>
#include <fft.h>
#include <vecops.h>
#include <genutils.h>
#include <samplops.h>

#include <math.h>
#include "db.grutl.h"
#include "db.psyac.h"

#ifdef PSYAC	/* compile-time option in makefile */

extern int LINEH;
extern int VAXLABWIDTH;

/* sparse tables referenced to get peak frequencies */
extern ststrct *sparsel1[], *sparsel2[];
extern DOUBLE *samprate;

#define LEG_WIDTH 100	/* pixel width of legend box */
#define LEG_LINES 6	/* lines in legend, including title */

void psyLegend(x,y)	/* draw up a legend for the psyc display */
    int x;
    int y;		/* top left co-os */
    {
    RgnHandle	oldClip;
    Rect	r1, r2;
    int		ht;

    ht = LINEH*1.5;
    r1.top = y; 			r1.left = x;
    r1.bottom = y + LEG_LINES*ht;	r1.right = x + LEG_WIDTH;

    PushClip(&oldClip, r1);

    EraseRect(&r1);
    FrameRect(&r1);

    r2 = r1;
    r2.top += LINEH/2;
    r2.bottom = r2.top + LINEH;

    dbWWriteIn(&r2, "Psycho 1", 0);		OffsetRect(&r2, 0, ht);
    LegendLine(&r2, -1, BLACK, "signal");	OffsetRect(&r2, 0, ht);
    LegendLine(&r2, -1, RED, "mask");		OffsetRect(&r2, 0, ht);
    LegendLine(&r2, -1, BLUE, "sb ftmax");	OffsetRect(&r2, 0, ht);
    LegendLine(&r2, 0, GREEN, "tonal");		OffsetRect(&r2, 0, ht);
    LegendLine(&r2, 1, MAGENTA, "ntone");	OffsetRect(&r2, 0, ht);

    PopClip( &oldClip );
    }

/* psy disp has whole thing itself */
void psyDisp(prec, stt, end, chans, frqv)
    Rect *prec;
    short *stt;
    short *end;
    int   chans;
    FRQVIEW *frqv;
    {
    Rect	r2;
    int		iar[30], ic;
    DOUBLE 	*tmpBuf;
    DOUBLE 	*smr;
    DOUBLE 	*powsf;
    long  	smps, frameSize, nsbs;
    DOUBLE	srate;
    DOUBLE	minFrq, maxFrq, minDB, maxDB;
    tntstrct 	*ttab;		/* array of tonal and non-tonal components */
    INT	     	ntn, nxt;
    DOUBLE  	*fft_peak;
    ststrct	*dbg_sts;	/* read sparse table for debug only */
    int 	np;
    int  	layer = 2;
    int		srate_ix;	/* problem .. */
    int 	i;
static int	init = 0;

    if(init == 0
       || (frqv->samplingRate != 44100.0 && frqv->samplingRate != 32000.0 
	   && frqv->samplingRate != 48000.0 ))  {
	frqv->samplingRate = 44100.0;
	frqv->minFrq = 0;	frqv->maxFrq = frqv->samplingRate / 2.0;
	frqv->minMag = -30.0;	frqv->maxMag = 70.0;
	init = 1;
	}
    srate = frqv->samplingRate;
    if(srate == 32000.0)	srate_ix = 2;
    else if(srate == 48000.0)	srate_ix = 1;
    else /* if(srate == 44100.0) */	srate_ix = 0;
    minDB = frqv->minMag;    maxDB = frqv->maxMag;
    minFrq = frqv->minFrq;   maxFrq = frqv->maxFrq;

    /* drw lins set in from edge */
    GutterR(*prec,lSide|rSide|bSide, VAXLABWIDTH, SBarWidth, SBarWidth, LINEH);

    r2 = *prec;
    r2.top = r2.bottom-SBarWidth;
    r2.left += VAXLABWIDTH;    		r2.right -= SBarWidth; 
    AddTimeLine(&r2,minFrq,maxFrq,0,1,1 /*vnh,sns,und*/,iar,&ic);

    r2 = *prec;
    r2.right = r2.left+VAXLABWIDTH;	r2.bottom -= SBarWidth;
    AddTimeLine(&r2,minDB,maxDB,1,0,0,iar,&ic);

    smps = (end-stt)/chans;
    np   = (layer == 2)? NP:NP/2;
    /* have to take this many samples actually */
    smps = np;			/* !!!! bad news near end of sample */
    tmpBuf = MakeBuf(smps * chans);
    smr    = MakeBuf(N_SUBBANDS*2);

    Short2Float(stt, chans, tmpBuf, 1, smps);	/* cast to Mono too */
    psy1_detail(np, tmpBuf, srate, smr, 
		&powsf, &ttab, &ntn, &nxt, &fft_peak);	/* detail return */

    /* set r2 to the actual plot region */
    r2 = *prec;
    r2.left += VAXLABWIDTH;		r2.right -= SBarWidth;
    r2.bottom -= LINEH;

    /* draw FFT power output spectrum */
    frameSize = np/2;
    /* copy powsf into tmpBuf in the following manner:
       {0, 1, 1, 2, 2, 3, 3,..... n-2, n-2, n-1, n-1, n}
       This achieves the half-width bars at each end */
    CopyFloats(powsf, 1, tmpBuf, 2, frameSize, FWD);
    CopyFloats(powsf+1, 1, tmpBuf+1, 2, frameSize, FWD);
    VLog(tmpBuf, 1, tmpBuf, 1, 1.0, 2*frameSize);
    VSMul(tmpBuf, 1, 4.3429, 2*frameSize); 	/* power->dB */
    ShowFloats(&r2, tmpBuf, 1, 2*frameSize, 
	       minFrq/(srate/2.0), maxFrq/(srate/2.0), minDB, maxDB, BLACK); 

    nsbs = N_SUBBANDS;
    ShowFloats(&r2, fft_peak, 1, nsbs, 
	       minFrq/(srate/2.0), maxFrq/(srate/2.0), minDB, maxDB, BLUE);
    /* subtract smrs from fftpeak to get masking levels */
    VSMul(smr, 1, -1.0, nsbs);
    VVAdd(fft_peak, 1, smr, 1, nsbs);
    ShowFloats(&r2, smr, 1, nsbs, 
	       minFrq/(srate/2.0), maxFrq/(srate/2.0), minDB, maxDB, RED);


    /* tonal and nontonal components */
    dbg_sts = (layer == 1)? sparsel1[srate_ix]: sparsel2[srate_ix];
    /* tonal components */
    for(i=0; i<ntn; ++i)  {
	tmpBuf[2*i] = srate*dbg_sts[ttab[i].tindx].fft_indx / np;
	tmpBuf[2*i+1] = ttab[i].spl;
	}
    ShowFloatsAsLine(&r2, tmpBuf, ntn, minFrq, maxFrq,
		     minDB, maxDB, 0, GREEN);
    /* nontonal components */
    for(i=0; i<nxt; ++i)  {
	tmpBuf[2*i] = srate*dbg_sts[ttab[ntn+i].tindx].fft_indx / np;
	tmpBuf[2*i+1] = ttab[ntn+i].spl;
	}
    ShowFloatsAsLine(&r2, tmpBuf, nxt, minFrq, maxFrq,
		     minDB, maxDB, 1, MAGENTA);

    /* add legend near top right */
    psyLegend((int)(prec->right - SBarWidth*1.4 - LEG_WIDTH), 
	      (int)(prec->top + SBarWidth*.4));

    Myfree((char *)tmpBuf);
    Myfree((char *)smr);
    }

#else /* null psyDisp if no PSYAC */

void psyDisp(prec, stt, end, chans, frqv)
    Rect *prec;
    short *stt;
    short *end;
    int   chans;
    FRQVIEW *frqv;
    {
    }

#endif /* def PSYAC */
