/********************************************************/
/*	db.sfile.c 											*/
/*														*/
/* Macintosh soundfile interface - based on Csound/Mac	*/
/* dpwe  02apr90										*/
/********************************************************/

#include <MacTypes.h>
#include <FileMgr.h>
#include <TextEdit.h>
#include <WindowMgr.h>
#include <StdFilePkg.h>
#include <ResourceMgr.h>
#include <MemoryMgr.h>

#ifndef NULL
#define NULL 0L
#endif

static	char *dSizeStr = "\psample-size";
static	char *sRateStr = "\psample-rate";
static	char *chansStr = "\pchannels";

void AddMacHeader(fName,chans,srate,dsize)
	char	*fName;	/* Name of soundfile - already created */
	int 	chans;		/* number of channels */
	double	srate;		/* sampling frequency */
	int 	dsize;		/* bytes per sample */
	{
	char	**handle1,**handle2,**handle3;	/* temporary strings */
	char	*str,s2;
	int 	rID;
	long	strType;
	int 	newr = 0;						/* flag that resource is added */
	
	strType = (long)'STR ';
	

	CtoPstr(fName);
	rID = OpenResFile(fName);
	if(rID == -1)
		{									/* error - no resource file */
		CreateResFile(fName);			/* add a resource fork to the soundfile */
		rID = OpenResFile(fName);		/* open it.. */
		}
	PtoCstr(fName);

	handle1 = GetNamedResource('STR ',dSizeStr);
	if(handle1 == NULL)								/* make new resource if there wasn't one */
		{
		handle1 = NewHandle(4);
		newr = 1;
		}
	HLock(handle1);
	str = *handle1;
	NumToString((long)dsize,str);
	HUnlock(handle1);
	if(newr)
		{
		newr = 0;
		AddResource(handle1,strType,1000 /* UniqueID(strType) */,dSizeStr);
		}				/* Digidesign uses specific numbers */
	else
		ChangedResource(handle1);

	handle2 = GetNamedResource('STR ',sRateStr);
	if(handle2 == NULL)								/* make new resource if there wasn't one */
		{
		handle2 = NewHandle(16);
		newr = 1;
		}
	HLock(handle2);
	str = *handle2;
	FloatToString(srate,str,4 /* d.p. */ );
/*	NumToString((long)srate,str);		/*!!!! ignores frac part */
/*	s2 = str + 1 + (int)(str[0]);
 *	NumToString((long)(1000.0*(srate - (double)((int)srate) ) ), s2);
 *	str[0] = (char)((int)s2[0])+((int)str[0]);
 *	s2[0] = '.';				/* phew : 'fixed' float -> ascii using int -> ascii, pasc strs */
	HUnlock(handle2);
	if(newr)
		{
		newr = 0;
		AddResource(handle2,strType,1001 /* UniqueID(strType) */,sRateStr);
		}
	else
		ChangedResource(handle2);

	handle3 = GetNamedResource('STR ',chansStr);
	if(handle3 == NULL)								/* make new resource if there wasn't one */
		{
		handle3 = NewHandle(4);
		newr = 1;
		}
	HLock(handle3);
	str = *handle3;
	NumToString( (long)chans, str);
	HUnlock(handle3);
	if(newr)
		{
		newr = 0;
		AddResource(handle3,strType,1002 /* UniqueID(strType) */,chansStr);
		}
	else
		ChangedResource(handle3);

	CloseResFile(rID);		/* includes releasing all the handles */
	}

int	ReadMacHeader(fName,pchans,psrate,pdsize,plen)
/* Gets standard stuff from Digidesign-type resource header. Returns error code. */
	Str255	*fName;			/* PASCAL Name of file to get it from */
	int		*pchans;		/* returns number of channels */
	double	*psrate;		/* returns sampling-rate */
	int 	*pdsize;		/* returns bytes / sample */
	long	*plen;			/* returns number of bytes in data (sample) */
	{
	int 	rc = 0;			/* return code - couldn't find some resources */
	int 	rfd,dfd,vRefNum,err=0;
	char	**handle,volName[32];
	long	rslt, *prslt;

	GetVol(volName,&vRefNum);
	
	if(FSOpen(fName,vRefNum,&dfd))	/* open (another path to read-only) data fork */
		*plen = -1;				/* error - couldn't open data file to find length */
	else{
		GetEOF(dfd, plen);		/* Find data length */
		FSClose(dfd);			/* release data fork */
		}

	rfd = OpenResFile(fName);
	if(rfd == -1)
		return(-1);			/* error - no resource file */

	prslt = &rslt;
	
	handle = GetNamedResource('STR ',dSizeStr);
	if(handle == NULL)
		rc += 1;				/* error - no sample-size resource */
	else{
		HLock(handle);
		StringToNum(*handle,prslt);
		*pdsize = (int)rslt;
		HUnlock(handle);
		}

	handle = GetNamedResource('STR ',sRateStr);
	if(handle == NULL)
		rc += 2;				/* error - no sample-rate resource */
	else{
		HLock(handle);
		StringToFloat(*handle,psrate);
		HUnlock(handle);
		}
 
	handle = GetNamedResource('STR ',chansStr);
	if(handle == NULL)
		rc += 4;				/* error - no channels resource */
	else{
		HLock(handle);
		StringToNum(*handle,prslt);
		*pchans = (int)rslt;
		HUnlock(handle);
		}

	CloseResFile(rfd);

	return(rc);					/* set bits of rc indicate unfound resource */
	}

	
void MacSetCreator(fName)		/* set the creator & type of this file for Digidesign SF */
	Str255	*fName;				/* PASC Name of soundfile - already created */
	{
	char	volName[32];
	int 	vRefNum;
	FInfo	fInfo;
	
	GetVol(volName,&vRefNum);
	GetFInfo(fName,vRefNum,&fInfo);				/* get creator & type */
	fInfo.fdType = 'Sd2f';
	fInfo.fdCreator = 'Sd2a';
	SetFInfo(fName,vRefNum,&fInfo);				/* rewrite modified */
	}

#define noError 0

/*
 * LoadFile - given a  reference number to an open file, load the file
 * into a relocatable block in the heap and return a handle it and its size.
 *	Returns zero if no error occurs, otherwise returns the OS error code.
 */
short LoadFile(pathnum,bufHan,bufSize)
	int pathnum;
	Handle *bufHan;
	long *bufSize;
	{
	OSErr err = noError;

	err=GetEOF(pathnum,bufSize);
	if(err)
		{ /* cannot determine size of file */
		FSClose(pathnum);		
		return(err);
		}
	*bufHan=NewHandle(*bufSize);
	err=MemError();
	if(err)
		{ /* cannot allocate buffer space */
		FSClose(pathnum);
		return(err); 
		}
	HLock(*bufHan);
	err=FSRead(pathnum,bufSize,**bufHan);
	HUnlock(*bufHan);
	if(err)
		{ /* cannot read file */
		FSClose(pathnum);
		DisposHandle(*bufHan);
		*bufHan=NULL;
		*bufSize=0L;
		return(err);
		}
	err=FSClose(pathnum);
	return(noError);
	}

StringToFloat(s,pf)			/* like built-in StringToNum - takes a Str255 & &float */
	Str255	s;
	double	*pf;
	{
	int 	i,j;
	long	l,rslt;
	
	i = (int)(s[0]);
	for(j=1; j<=i && s[j] != '.'; ++j)
		;
	s[0] = (char)j-1;
	StringToNum(s,&rslt);		/* !!!!! Treats as INT string !!! */
	*pf = (double)rslt;
	if(s[j]=='.')
		{
		s[j] = '1';
		s[j-1] = i-j+1;
		StringToNum(s+j-1,&rslt);		/* .xxx becomes (long)1xxx */
		for(l=1;l<=rslt; l *= 10)
			;						/* find next highest pwr of 10 */
		*pf += (10.0*(double)rslt)/((double)l)-1.0;	/* voila - frac part */
		}
	}

/* FloatToString moved into db.grutl.c */
