/*******************************************************************\
*   db.sgram.c							    *
*   real grey-scale spectrograms for DSPbench			    *
*   03nov90 dpwe						    *
\*******************************************************************/

#define DBOFFSET	20	/* make bigger (0-80) for blacker sgrams */
#define DBRANGE	       120	/* range from black to white */
#include <stdio.h>
#include <string.h>
#include <dsputil.h>
#include <fft.h>
#include <genutils.h>
#include <vecops.h>
#include <arrayops.h>	/* for drawing into an arrays */

#include <math.h>
#include "DSPbench.h"

#ifdef TRAX
#include "db.tracks.h"	/* for RedrawTrax(allTrax, TKM_NORM); */
#endif /* TRAX */

#define SHORTMAX 32767
#define MAXXPTS 2048
#define MEGAMIN -100	/* dB of 0.0 */

/* static globals */

static  Point	*xpts = NULL; /* [MAXXPTS]; */
static  short   *maxs = NULL; /* [MAXXPTS]; */
static  float   *tmp = NULL;  /* [MAXXPTS]; */

/* prototypes */
#ifdef __STDC__
long intLog(long);
#else
long intLog();
#endif

static DrawSGFrame(pvData, len, rec, scale, time, binsPoct, minFreq, 
		   pixm, baser, bytear, col, bwid, showphase)
    /* draw a single PV trace */
    float  *pvData;
    long   len;
    Rect	rec;
    int 	scale;
    FLOATARG	time;		/* current time - just to put in peakList */
    FLOATARG	binsPoct;	/* only for CQfile frq -> y index */
    FLOATARG	minFreq;	/* only for CQfile frq -> y index */
    PIXMAP	pixm;		/* here too! */
    Rect   	*baser;		/* context rect for pixmap */
    ByArray	*bytear;	/* byte array for that kind of output */
    int		col;		/* current column in bytearray */
    int		bwid;		/* width of this bytarray column */
    int		showphase;	/* flag to draw phases not magnitudes */
    {
    long   wPerFrame;
    long   i;
    double  d,t,max = -1e30, min = 1e30;

    /* assume data is floats, format is mag,pha, etc.. */
/*    wPerFrame = pvh->frameBsize/sizeof(float);	
    pvData = ((float *)PVDataLoc(pvh))+wPerFrame*frame;
    len    = pvh->frameBsize/(2*sizeof(float));	/* number of mag points */

    if(showphase)  {
	++pvData;	/* look at phases instead */
    }
    for(i=0; i<len; ++i)
	{
	d = *pvData++;
	if(showphase)  {
	/*    t = (d+PI)/(2*PI);	*/
		if(d>=0)		t = d/(2*PI);
		else			t = (d+2*PI)/(2*PI);
	} else {
	    d *= 10.0*scale;
	    if(d>0.00001)
		t = 8.863*log(d);
	    else
		t = MEGAMIN;
	    t = (t+DBOFFSET)/DBRANGE;  /* -DBOFFSET is black; DBRANGE to wht */
	}
	if(t>max) max = t;
	if(t<min) min = t;
	tmp[i] = t; 
	++pvData;		/* skip frq bin */
	}
    DrawGrayChunk(&rec, scale, max, min, tmp, len, pixm, baser, bytear, 
		  col, bwid);

    }

DrawGrayChunk(prec, scale, max, min, tmp, len, pixm, baser, bytear, 
	      col, bwid)
    /* I like data to be in range 0(wh)..1(bl) */
    Rect	*prec;
    int 	scale;
    double	max, min;
    float	*tmp;
    long	len;
    PIXMAP	pixm;	/* here too! */
    Rect	*baser;	/* base rectangle for offset into pixm */
    ByArray	*bytear;
    int		col;	/* which column to write in byte array */
    int		bwid;	/* how many columns to write ?? */
{
    Rect	g_r, p_r;
    int 	g_w, g_h;
    int		s_x, s_y;
    long	i, j;
    int 	c, lastc = 0;
    int 	b, lastb = 0;
    int		byrows, bycols;
    MyByte	*bdata;
    double	v;

    if(bytear) {
	byrows = bytear->rows;
	bycols = bytear->cols;
	bdata  = &(bytear->data[col]);
    }
    if(baser == NULL)	/* flag *not* to do output */
	prec = NULL;
    if(prec)  {
	g_r.left = prec->left;
	g_r.right = prec->right;
	s_y = prec->top;
    g_h = 0;
    }
    for(i=0; i<len; ++i)
	{
	v = tmp[len-1-i];
	c = (int)(v*(double)grays);
	if(c<0) 	c = 0;
	if(c>=grays)	c = grays-1;
	if(prec) {
	    s_y += g_h;
	    g_h = prec->top - s_y + (int)( ((double)i+1)/(double)len 
					* (double)(prec->bottom - prec->top));
	    g_r.top = s_y;
	    g_r.bottom = s_y + g_h;
	    ColorRect(&g_r,colrs[c]);
	    if(pixm)  {
		p_r = g_r;
		OffsetRect(&p_r, -baser->left, -baser->top);
		pxmPaintRect(pixm, &p_r);   /*** GUESS color is still set!! */
	    }
	}
	if(bytear)  {	/* write values into output byte array */
	    int l, u, d, e, k, pc;

	    b = (int)(v * (double) BYTEMAX);
	    if(b<0)	b = 0;
	    else if(b>BYTEMAX)	b = BYTEMAX;
	    l = (i-0.5)*byrows/len;
	    u = (i+0.5)*byrows/len;
	    d = u-l;
	    j = l;
	    if(j<0)	j = 0;
	    for(; j < u; ++j)  {	/* linearly interpolate */
		e = lastb + ((b - lastb) * (j-l+1))/d;
		*(bdata + bycols*j) = e;
		if(col > 0 && bwid > 1)  {	/* lin interp in x too */
		    pc = *(bdata + bycols*j - bwid);	/* prev column */
		    for(k = 1; k < bwid; ++k)  {
			*(bdata + bycols*j - k) = e + (pc - e)*k/bwid;
		    }
		}
	    }
	    lastb = b;
	}
	lastc = c;
	}
    if(bytear)  {	/* finish off ramp out for output array */
	int l, u, d, e, k, pc;

	b = 0;
	l = (len-0.5)*byrows/len;
	u = (len+0.5)*byrows/len;
	d = u-l;
	if(u>byrows-1)	u = byrows-1;
	for(j = l; j < u; ++j)  {	/* linearly interpolate */
	    e = lastb + ((b - lastb) * (j-l+1))/d;
	    *(bdata + bycols*j) = e;
	    if(col > 0 && bwid > 1)  {	/* lin interp in x too */
		pc = *(bdata + bycols*j - bwid);	/* prev column */
		for(k = 1; k < bwid; ++k)  {
		    *(bdata + bycols*j - k) = e + (pc - e)*k/bwid;
		}
	    }
	}
    }
    
    SetBlack();
    }

#ifdef TRAX
int traxRevealed = 0;

extern RECRCD **FindRhan(RECRCD **rrh, int id);
        /* search tree for pane id */

int RevealTrax(WINRCD **wrcd) 
{	/* menu item hides/shows tracks */
    int id = Svsndid;

    traxRevealed = !traxRevealed;
    if((*wrcd)->w.snd.vMode == (int)spgrmVM)  {
	if(traxRevealed)  {
	    FRQVIEW   *frqv = (*wrcd)->w.snd.frqV;
	    RECRCD    **rh = FindRhan( (*wrcd)->w.snd.recTreeH ,id);
	    Rect      rec = (*rh)->rec;
	    Rect      inrect;

	    inrect = rec;
	    InsetRect(&inrect,SBarWidth,0);
	    inrect.left += 1;		/*!*/
	    DrawTraxOnSgm(wrcd, &inrect, frqv, NULL);
	} else {
	    RedrawWinBit(wrcd, id);
	}
    }
    return traxRevealed;
}
#endif /* TRAX */

ShowSgram(WINRCD **wrcd, Rect rec, FRQVIEW *frqv)
{	/* new entry point does PIXMAP stuff */
    PIXMAP	pixmap;
    int 	pmW, pmH;
    Rect	inrect;

    /* find *actual* rect we will fill .. */
    inrect = rec;
    InsetRect(&inrect,SBarWidth,0);
    inrect.left += 1;		/*!*/
#ifdef CQFILE
    if((*wrcd)->w.snd.cqi == NULL)
#else
    if((*wrcd)->w.snd.pvs == NULL)    
#endif
	{
	MarkX(&inrect);
	return;
	}

    pixmap = (*wrcd)->w.snd.pixmap;
    pmW = (*wrcd)->w.snd.pixW;
    pmH = (*wrcd)->w.snd.pixH;
    if(frqv->valid && pixmap && pmW == RWIDTH(&inrect) 
       && pmH == RHEIGHT(&inrect)) {
	pxmShow(pixmap, &inrect);
#ifdef TRAX
	if(traxRevealed)  {
	    /* tracks assumed already placed - just need refreshing */
	    RedrawTrax(allTrax, TKM_NORM);
	    RedrawTrax(selTrax, TKM_SELECT);
	}
#endif /* def TRAX */
    }
    else {
	if(pmW != RWIDTH(&inrect) || pmH != RHEIGHT(&inrect))  {
	    if(pixmap)
		pxmFree(pixmap);
	    pmW = (*wrcd)->w.snd.pixW = RWIDTH(&inrect);
	    pmH = (*wrcd)->w.snd.pixH = RHEIGHT(&inrect);
	    pixmap = (*wrcd)->w.snd.pixmap = pxmNew(pmW, pmH);
	}
	ActualShowSgram(wrcd, frqv, &inrect, pixmap, NULL);
#ifdef TRAX
	if(traxRevealed)  {
	    DrawTraxOnSgm(wrcd, &inrect, frqv, NULL);
	}
#endif /* def TRAX */
	frqv->valid = 1;
    }
}

static int showPhase = 0;	/* flag changed by menu item? */

int ShowPhases(WINRCD **wrcd)
{  /* showPhase item hit in menu bar */
    if(showPhase)  {
	showPhase = 0;
    } else {
	showPhase = 1;
    } 
    (*wrcd)->w.snd.frqV->valid = 0;
    RedrawWinBit(wrcd, Sviewid);
    return showPhase;
}

ActualShowSgram(wrcd, frqv, inrect, pixm, bytar)
    WINRCD  **wrcd;
    FRQVIEW *frqv;
    Rect    *inrect;
    PIXMAP  pixm;
    ByArray *bytar;
    {
#ifdef CQFILE
    CQTINFO	*cqi;
#else		
    AFTINFO	*cqi;
    PVSTRUCT	*pvh;
#endif		
    long	sStt, sEnd;	/* sample frame indices */
    int 	sfSiz;
    double	fFrInc;
    long	frInc, frSiz;
    long	firstFrame, lastFrame, frame, frames;
    Rect    subrect;
    long    i;
    int 	scale;
    double   overlapFraction = 0.0; /* horiz ovlp of panes */
    int     totalWidth;
    double   left, right, widthPerTrace, hoffs;
    /*    RgnHandle    dummy;	*/
    int		err = 0;
    long   flPerFrame;
    double  tPerFrame;
    long   len;
    int    binSkew = 0, vlen;
static float *specBuf;
static float *prvBuf;
    int 	framesToDisp;
    double  time, binsPoct, minFreq;
    int		tBins;
extern double	minTrackDB;	/* for DIAGHACK */

    int		baCol = 0;
    int		baWidth;

#ifdef CQFILE
#define CFILE 1	/* like CQFILE, but more */
#endif

#ifdef CQFILE
    cqi  = (*wrcd)->w.snd.cqi;
    if(cqi == NULL)
#else
    pvh  = (*wrcd)->w.snd.pvs; 
    if(pvh == NULL)    
#endif
	{
	if(inrect)
	    MarkX(inrect);
	return;
	}
#ifdef CQFILE
    tBins = cqi->nBins * cqi->nOcts;
#else
    tBins = pvh->frameBsize / sizeof(cpxpolar);
#endif
    if(xpts == NULL)  {
	if( (xpts=(Point *)Mymalloc((long)MAXXPTS, sizeof(Point), 
			"ShowSgm:pts"))==NULL)			err = 1;
	else if((maxs=(short *)Mymalloc((size_t)MAXXPTS, sizeof(short), 
			"ShowSgm:maxs"))==NULL)			err = 1;
	else if((tmp=(float *)Mymalloc((size_t)MAXXPTS, sizeof(float), 
			"ShowSgm:tmp"))==NULL)			err = 1;
#ifdef CFILE
	else if((specBuf = (float *)Mymalloc((long)tBins, 
	       sizeof(cpxpolar), "ShowSgm:specBuf"))==NULL)	err = 1;
	else if((prvBuf = (float *)Mymalloc((size_t)tBins,
	       sizeof(cpxpolar), "ShowSgm:prvBuf"))==NULL)	err = 1;
#endif
	if(err)  {
	    fprintf(stderr,"%s:ShowSgram: malloc fail\n",programName);
	    abort();
	}
    }

    if(inrect)
	totalWidth = RWIDTH(inrect);
    else
	totalWidth = bytar->cols;
    scale = (*wrcd)->w.snd.scale;
    sfSiz= Ssize(((*wrcd)->w.snd.view)->snd);
    sStt = ((*wrcd)->w.snd.view)->bStart / (long)sfSiz;
    sEnd = ((*wrcd)->w.snd.view)->bEnd / (long)sfSiz;
#ifdef CFILE
    flPerFrame = len = tBins;
    if(bytar)
	framesToDisp = bytar->cols;
    else
	framesToDisp = totalWidth/2; /* scale;	/* hack (was 16) */
/*    frInc = (sEnd - sStt)/framesToDisp;	/* trick the code */
    tPerFrame = (sEnd-sStt)/
	(framesToDisp*((*wrcd)->w.snd.view)->snd->samplingRate);
    firstFrame = sStt*framesToDisp / (sEnd-sStt);
    lastFrame  = sEnd*framesToDisp / (sEnd-sStt);
#ifdef CQFILE
    binsPoct = (double)cqi->nBins;
#endif
    minFreq  = cqi->pvh->minFreq;
/* #if DIAGHACK */
    vlen = len;
    if(minTrackDB == -51)	/* special flag */
	{
	double hiFrq;

	minFreq = 1000;
	hiFrq  = 6000;	/* special diags one-off */
	vlen = (double)binsPoct*log(hiFrq/minFreq)/log(2.0); /* fake # bins */
	}
    if(frqv->maxFrq > frqv->minFrq)  {	/* new: allow frq zooms */
	double hiFrq;

	minFreq = frqv->minFrq;
	hiFrq  = frqv->maxFrq; 
	vlen = (double)binsPoct*log(hiFrq/minFreq)/log(2.0); /* fake # bins */
    }
/* #endif /* DIAGHACK */
    binSkew =  0.5 + log(minFreq/cqi->pvh->minFreq)/log(2.0)*(double)binsPoct;
    frqv->minFrq = cqi->pvh->minFreq * 
	exp(log(2.0)*((-0.5+(double)binSkew)/(double)binsPoct));
    frqv->maxFrq = cqi->pvh->minFreq * 
	exp(log(2.0)*((-0.5+(double)(binSkew+vlen))/(double)binsPoct));

#else
    frInc = pvh->frameIncr;
    fFrInc = (double)frInc;
    frSiz = pvh->frameSize;
    /* assume data is floats, format is mag,pha, etc.. */
    flPerFrame = pvh->frameBsize / sizeof(float);
    len        = pvh->frameBsize/(2*sizeof(float));	/* #of mag points */
    vlen = len;
/* fprintf(stderr, "frqv: min %.1f max %.1f\n", frqv->minFrq, frqv->maxFrq); */
    if(frqv->maxFrq > frqv->minFrq)  {	/* new: allow frq zooms */
	double hiFrq;

	minFreq = frqv->minFrq;
	hiFrq  = frqv->maxFrq; 
	binSkew = minFreq * len / (pvh->maxFreq - pvh->minFreq);
	vlen = hiFrq * len / (pvh->maxFreq - pvh->minFreq) - binSkew;
	frqv->minFrq = pvh->minFreq + binSkew * (pvh->maxFreq - pvh->minFreq)
	                                           / len;
	frqv->maxFrq = pvh->minFreq + (binSkew + vlen) * 
	                                (pvh->maxFreq - pvh->minFreq) / len;
    }
/*    fprintf(stderr, "kjhsdf: len = %d\n", len);	*/
    firstFrame = sStt / frInc;
    lastFrame  = sEnd / frInc;
    tPerFrame  = frInc/pvh->samplingRate;	/* for what ? */
#endif
    frames = lastFrame - firstFrame;
    if(frames <= 0)
	return;

    /*    PushClip(&dummy, *inrect);	*/
    widthPerTrace = (((double)totalWidth) / 
		  ((double)1.0 + ((double)frames - 1.0)*(1.0-overlapFraction)));
    hoffs  = ((1.0-overlapFraction) * widthPerTrace);
    if(inrect)
	left = (double)RLEFT(inrect);

    /*    printf("Frame size %ld incr %ld 1st %ld last %ld\n",
	  frSiz, frInc, firstFrame, lastFrame);	*/
    if(inrect)
	subrect = *inrect;
#ifdef CFILE
    ClearFloats(prvBuf,1,2*len);
#endif
    baCol = 0;
    for(frame = firstFrame; frame < lastFrame; ++frame)  {
	if(inrect)  {
	    subrect.left = (int)left;
	    subrect.right = (int)(left + widthPerTrace);
	    left += hoffs;
	}
	time = frame*tPerFrame;
	if(bytar)  {
	    /* use last time's baWidth to fixup width by one frame */
	    baWidth = ((frame - firstFrame)*(bytar->cols - 1)
		        /(lastFrame-firstFrame - 1)) - baCol;
	    baCol += baWidth;
	}
#ifdef CQFILE
	/* SimpleSpecOfCQTFile(cqi, time, tPerFrame, (cpxpolar *)specBuf); */
	SpecOfCQTFile(cqi, time, tPerFrame, (cpxpolar *)specBuf);
/*	VVSub(prvBuf, 2, specBuf, 2, len);	/* specBuf -= prvBuf */
/*	VVAdd(specBuf, 2, prvBuf, 2, len);	/* prvBuf += specBuf */
#else
/*	SpecOfAFTFile(cqi, time, tPerFrame, (cpxpolar *)specBuf, 0 );	*/
	specBuf = ((float *)PVDataLoc(pvh))+frame*flPerFrame;
#endif
/*	DrawSGFrame(pvh, frame, subrect, scale);		*/
	DrawSGFrame(specBuf + (sizeof(cpxpolar)/sizeof(float))*binSkew, 
		    vlen, subrect, 64*scale /* 8, or scale */ , 
		    time, binsPoct, minFreq, pixm, inrect, bytar, baCol, 
		    baWidth, showPhase);
	}
    /*    PopClip(&dummy);	*/
    }

void DrawGrayRange(Rect *rec, int vnh)
{	/* draw a thermometer of all 'gray' levels in the rect.
	   If vnh == 1, then gradation is vertical, else horiz */
    int		x,y,w,h,i;
    Rect	myRect;

    FrameRect(rec);
    MarkX(rec);
    x = rec->left;
    y = rec->bottom;
    w = rec->right - rec->left;
    h = rec->top - rec->bottom;	/* start at bottom, work up */
    myRect = *rec;
    for(i = 0; i<grays; ++i)  {
	if(vnh)  {
	    myRect.bottom = y + (i*h)/grays;
	    myRect.top    = y + ((i + 1) * h)/grays;
	} else {
	    myRect.left   = x + (i * w)/grays;
	    myRect.right  = x + ((i + 1) * w)/grays;
	}
	ColorRect(&myRect, colrs[i]);
    }
    SetBlack();
}

