/*******************************************************************\
*   db.spex.c							    *
*   plotting PVOC spectra for DSPbench		    		    *
*   03nov90 dpwe						    *
\*******************************************************************/

/******* stripped out OLDTRACKCODE stuff - 1993apr20 */

#include <stdio.h>
#include <string.h>
#include <dsputil.h>
#include <fft.h>
#include <genutils.h>

#include <math.h>
#include "DSPbench.h"

#define SHORTMAX 32767
#define MAXXPTS 2048

#define KLEN	16	/* length of smoothing kernel */

/* types */

/* static globals */

static  Point	*xpts = NULL; /* [MAXXPTS]; */
static  short   *maxs = NULL; /* [MAXXPTS]; */
static  short   *stmp = NULL;  /* [MAXXPTS]; */
static  float   *tmp = NULL;  /* [MAXXPTS]; */

static  float   kernel[KLEN];
static  long    klen = 0;	/* a flag that kernel not set up */

/* prototypes */
#ifdef __STDC__
long intLog(long);
#else
long intLog();
#endif

static DrawSpecFrame(data, len, rec, scale, time, binsPoct, minFreq)  
    float	*data;		/* mag, pha spectrum */
    long 	len;		/* number of point innit */
    Rect	rec;
    int 	scale;
    FLOATARG	time;		/* current time - just to put in peakList */
    FLOATARG	binsPoct;
    FLOATARG	minFreq;
    {		/* draw a single PV trace */
    long   i;
    int     ntmp;
    float   *mag = data;

    for(i=0; i<len; ++i)
	{
/*	tmp[i] = intLog((long)*mag++);	*/
#ifdef CQFILE
	tmp[i] = 2*2*8.863*log(10.0* *mag++);	/* diff scale for cq files */
#else
	tmp[i] = 2*8.863*log(.001 * *mag++);
#endif
	++mag;		/* skip frq bin */
	}
    DrawBLine(rec, scale, tmp, len);	
/*    DrawALine(rec, scale, tmp, len);		*/
    }

SetupKernel()
    {
    long i;

    for(i=0; i<KLEN; ++i)
	{
	kernel[i] = (2/(double)KLEN)*(.54 - .46 * cos(2*PI*((double)i)/KLEN));
	}
    klen = KLEN;
    }

DrawBLine(rec, scale, data, len)
    Rect	rec;
    int 	scale;
    float	*data;
    long	len;
    {
    long 	i;
    short	*shdat = stmp;

    if(len > MAXXPTS)
	{
	fprintf(stderr, "DrawBLine len %d > %d\n", len, MAXXPTS);
	abort();
	}
    for(i=0; i<len; ++i)
	{
	shdat[i] = (short)(64*data[i]);
	}
    DrawTrace(shdat, len, rec, scale, 0);
    }	/* call waterfall with hidden line removal */

DrawALine(rec, scale, data, len)
    Rect	rec;
    int 	scale;
    float	*data;
    long	len;
    {
    Point *ptptr = xpts;
    long   i;
    short  x_0, y_0;
    long   x_scale, y_scale;
    RgnHandle dummy;

    x_scale = (((long)(rec.right-rec.left))<<16L)/len;
    y_scale = ((((long)(scale * (rec.bottom-rec.top)))<<16L)/SHORTMAX); 
    x_0     = rec.left;
    y_0     = rec.bottom - ((2*(rec.bottom-rec.top))>>2);
    /* y-axis is 2/4s way to top of box - mostly -ve */
    for(i=0; i<len; ++i)
	{
	ptptr->h = x_0 + (short)(((long)i * x_scale)>>16L);
	ptptr->v = y_0 - (short)(((long)(*data++ * y_scale))>>12L);
	++ptptr;
	}
    PushClip(&dummy,rec);
    DrawLines(xpts,(int)len);
/*    MoveTo(rec.left,y_0);	
    LineTo(rec.right,y_0);	*/
    PopClip(&dummy);
    }

#define NFRAMES 6

void ShowSpex(wrcd,rec)
    WINRCD  **wrcd;
    Rect    rec;
    {
#ifdef CQFILE
    CQTINFO	*cqi;
#else
    PVSTRUCT	*pvh;
#endif
    long	sStt, sEnd;	/* sample frame indices */
    int 	sfSiz;
    double	fFrInc;
    long	frInc, frSiz;
    long	firstFrame, lastFrame, frame, frames;
    Rect    inrect;
    long    i;
    int 	scale,err=0;
    double   overlapFraction = 0.9; /* vert ovlp of panes */
    int     totalHeight;
    double   top, bottom, heightPerTrace,voffs;
    long   flPerFrame;
    double  tPerFrame, t0, t1, dur;
    long   len;
    /*    RgnHandle    dummy;	*/
static float *specBuf;
    int 	framesToDisp;
    double	time;
    double	binsPoct, minFreq;	/* only meaningful for cqt */

#ifdef CQFILE
    cqi  = (*wrcd)->w.snd.cqi;
    if(cqi == NULL)
#else
    pvh  = (*wrcd)->w.snd.pvs;
    if(pvh == NULL)
#endif
	{
	MarkX(&rec);
	return;
	}

    if(xpts == NULL)
	{
	if( (xpts=(Point *)Mymalloc((size_t)MAXXPTS, sizeof(Point), 
							"ShowSpx:xpts"))==NULL)
	    err = 1;
	else if((maxs=(short *)Mymalloc((size_t)MAXXPTS, sizeof(short),
							"ShowSpx:maxs"))==NULL)
	    err = 1;
	else if((tmp=(float *)Mymalloc((size_t)MAXXPTS, sizeof(float), 
							"ShowSpx:tmp"))==NULL)
	    err = 1;
	else if((stmp=(short *)Mymalloc((size_t)MAXXPTS, sizeof(short), 
							"ShowSpx:stmp"))==NULL)
	    err = 1;
#ifdef CQFILE
	else if((specBuf = (float *)Mymalloc((size_t)cqi->nBins*cqi->nOcts, 
					      sizeof(cpxpolar), "ShowSpx:spec"))==NULL)
	    err = 1;
#endif
	if(err)
	    {
	    fprintf(stderr,"%s:ShowSgram: malloc fail\n",programName);
	    exit(1);
	    }
	}
    if(klen == 0)
	SetupKernel();
    scale = 4*(*wrcd)->w.snd.scale;
    sfSiz= Ssize(((*wrcd)->w.snd.view)->snd);
    sStt = ((*wrcd)->w.snd.view)->bStart / (long)sfSiz;
    sEnd = ((*wrcd)->w.snd.view)->bEnd / (long)sfSiz;
#ifdef CQFILE
    len  = cqi->nBins * cqi->nOcts;
/*    framesToDisp = scale;	/* hack */
/*    frInc = (sEnd - sStt)/framesToDisp;	/* trick the code */
/*    tPerFrame = frInc/(((*wrcd)->w.snd.view)->snd->samplingRate);
 */
    /* don't know what the hell is going on here - set frame rate 
       to be twenty cycles of sampling rate (used to be 1ms) */
/*    tPerFrame = 20.0/((*wrcd)->w.snd.view)->snd->samplingRate;   */
    /* nonono!  This is waterfall display of CQ file (neat!  didn't 
       know we did that!).  Set (arbitrary) frame size for CQ files 
       to give some fixed number of traces in the display - say 50 */
    STimeLims(wrcd,&t0,&t1,&dur);
    if(t0==t1) {
	t0 = 0;	t1 = dur;
    }
    tPerFrame = (t1-t0)/50;	/* / N_FRAMES_PER_DISP */

    fFrInc = ((*wrcd)->w.snd.view)->snd->samplingRate * tPerFrame;
    frInc = (long)fFrInc;
    if(frInc == 0) {  /* increment was too small - show whole sound */
	t0 = 0;	t1 = dur;
	tPerFrame = (t1-t0)/50;	/* / N_FRAMES_PER_DISP */
	fFrInc = ((*wrcd)->w.snd.view)->snd->samplingRate * tPerFrame;
	frInc = (long)fFrInc;
    }
    framesToDisp = (sEnd-sStt)/frInc;
    binsPoct = (double)cqi->nBins;
    minFreq = cqi->pvh->minFreq;
#else
    frInc = pvh->frameIncr;
    fFrInc = (double)frInc;
    frSiz = pvh->frameSize;
    /* assume data is floats, format is mag,pha, etc.. */
    flPerFrame = pvh->frameBsize / sizeof(float);
    len        = pvh->frameBsize/(2*sizeof(float));	/* #of mag points */
    tPerFrame  = frInc/pvh->samplingRate;
#endif
    firstFrame = sStt / fFrInc;
    lastFrame  = sEnd / fFrInc;
    frames = lastFrame - firstFrame;
    if(frames <= 0)
	return;

    /*    PushClip(&dummy, rec);	*/
    totalHeight    = rec.bottom - rec.top;
    heightPerTrace = (((double)totalHeight) / 
	      ((double)1.0 + ((double)frames - 1.0)*(1.0-overlapFraction)));
    voffs  = ((1.0-overlapFraction) * heightPerTrace);
    inrect = rec;
    InsetRect(&inrect,SBarWidth,0);
    /*    bottom = (double)inrect.bottom;	*/
    top = (double)inrect.top;

    /*    printf("Frame size %ld incr %ld 1st %ld last %ld\n",
	  frSiz, frInc, firstFrame, lastFrame);	*/
    /* reset static list of past peaks to match */
    InitTrace();	/* in db.wfall.c : init hidden line removal */
    for(frame = firstFrame; frame < lastFrame; ++frame)
	{	/* draw frontmost first for HLR */
	/*	inrect.bottom = (int)bottom;
		inrect.top = inrect.bottom - heightPerTrace;
		bottom -= voffs;	*/
	inrect.top = (int)top + (lastFrame - 1 - frame)*voffs;
	inrect.bottom = (int)(inrect.top + heightPerTrace);
	time = frame*tPerFrame;
#ifdef CQFILE
	SpecOfCQTFile(cqi, time, tPerFrame, (cpxpolar *)specBuf);
#else
	specBuf = ((float *)PVDataLoc(pvh))+frame*flPerFrame;
#endif
	DrawSpecFrame(specBuf, len, inrect, scale, time, binsPoct, minFreq);
	}
/*    PrintPeakList(allTrax);		*/
    /*    PopClip(&dummy);	*/
    }
