/*******************************************************************\
*   db.tgdrw.c							    *
*   time graph window drawing functions for DSPbench		    *
*   13apr90 dpwe						    *
\*******************************************************************/

#include <string.h>

#include "DSPbench.h"
#include <samplops.h>		/* for CopyChars */
#include <wndw.h>		/* new PS output model 1993mar13 */

/* #define PSOUTPUT    1 	/* generates idraw-style lineseg output */  
/* define PSOUTPUT in Makefile if you like */
#define PSSOUNDBOTTOM	437     /* 319	/* bottom of ps window for sounds */

/* #define IBOXW (myTextWidth("44.10 kHz  "));	/* width of info box on left of window */
/* #define IBOXW	64	/* width of info box on left of window */
/* #define IBOXH	(10*LineHeight())	/* height of info box */
/* #define IBOXH	130 */
#define LVIEWH	125	/* height of 'long view' at top of window */
/* #define BUTHGT	(2*LineHeight())       /* height of buttons in control panel */
/* #define BUTHGT  25 */
#ifdef THINK_C
/* #define MBARH	0    /* vanishingly small menu bar - don't need it */
#else
/* #define MBARH   (1.5*LineHeight())      /* Height of menu bar area */
/* #define MBARH   16 */
#endif

/*#define VAXLABWIDTH	36	/* width of vertical axis space */
/*#define VAXLABWIDTH	(myTextWidth("-20k0"))	/* width of vertical axis space */

/* externs */
extern	Icon	leftArrowIcon, rightArrowIcon, upArrowIcon, downArrowIcon;
extern  Cursor waitCursor, arrow;

/* globals */
static int IBOXH, IBOXW, BUTHGT, MBARH;
       int VAXLABWIDTH, LINEH;	/* used by other files */

/* prototypes */

static void InitTgdrwGlobals()
    {
    static inited = 0;

    if(inited == 0)
	{
	IBOXH = 10*LineHeight();
	IBOXW = myTextWidth("44.10 kHz  ");
	BUTHGT = 2*LineHeight();
#ifdef THINK_C
	MBARH = 0;	/* no menu bar on macintosh */
#else
	MBARH = 1.5*LineHeight();
#endif
	VAXLABWIDTH = myTextWidth(" -20k0");
	LINEH = LineHeight();
	inited = 1;
	}
    }

RECRCD **MakeRecNode(id, sib, child, sbits)
    int     id;
    RECRCD  **sib, **child;
    int     sbits;
    {
    RECRCD  **rrhan;

    rrhan = (RECRCD **)NewHandle(sizeof(RECRCD));
    (*rrhan)->id = id;
    (*rrhan)->sib   = sib;
    (*rrhan)->child = child;
    (*rrhan)->sbits = sbits;
    return(rrhan);
    }


BuildTGRecTree(wrcd)			/* build a family of TG rects */
    WINRCD  **wrcd;
    {
    RECRCD  **n1,**n2;

    n1 = MakeRecNode(Svtimlid,NULL,NULL,SBNONE);    /* start at the bottom */
    n2 = MakeRecNode(VscaleDownid,NULL,NULL,SBNONE);
    n2 = MakeRecNode(VscaleUpid,n2,NULL,SBNONE);
    n1 = MakeRecNode(Svsndid,n1,n2,SBNONE);
    n1 = MakeRecNode(Svvscid,n1,NULL,SBNONE);
    n1 = MakeRecNode(Sviewid,NULL,n1,SBNONE);
    n2 = MakeRecNode(Lvsbarid,NULL,NULL,SBNONE);
    n2 = MakeRecNode(Lvsndid,n2,NULL,SBNONE);
    n2 = MakeRecNode(Lvtimlid,n2,NULL,SBNONE);
    n2 = MakeRecNode(Lvvscid,n2,NULL,SBNONE);
    n1 = MakeRecNode(Lviewid,n1,n2,SBNONE);
    n2 = MakeRecNode(Cquitid,NULL,NULL,SBNONE);
    n2 = MakeRecNode(Cprintid,n2,NULL,SBNONE);
    n2 = MakeRecNode(Crecid,n2,NULL,SBNONE);
#ifdef TRAX
    n2 = MakeRecNode(Csynthid,n2,NULL,SBNONE);
#endif /* TRAX */
    n2 = MakeRecNode(Cplayid,n2,NULL,SBNONE);
    n1 = MakeRecNode(Cboxid,n1,n2,SBNONE);
    n1 = MakeRecNode(Infboxid,n1,NULL,SBNONE);
    n1 = MakeRecNode(Shoutid,n1,NULL,SBNONE);	    /* newer! */
    n1 = MakeRecNode(MenuBarid,n1,NULL,SBNONE);     /* new! */
    n1 = MakeRecNode(Wwinid,NULL,n1,SBNONE);	    /* Make whole tree */
    (*wrcd)->w.snd.recTreeH = n1;
    }

SetTGRects(wrcd)
    WINRCD  **wrcd;
    {	
    Rect    r1,r2,r3;
    RECRCD  **n1,**n2;

    InitTgdrwGlobals();

    n1 = (*wrcd)->w.snd.recTreeH;
    /* Now fill in rec fields .. */
/*    r1 = (*wrcd)->myWindow->portRect; */
    RectOfWin( (*wrcd)->myWindow, &r1);
    (*n1)->rec = r1;

    n1 = (*n1)->child;	    /* MenuBar */
    r2 = r1; r2.bottom = r2.top + MBARH; r1.top = r2.bottom;
    (*n1)->rec = r2;

    n1 = (*n1)->sib;	    /* Shout box */
    r2 = r1; r2.top = r2.bottom - MBARH; r1.bottom = r2.top;
    (*n1)->rec = r2;

    n1 = (*n1)->sib;  	    /* Infboxid */     
    r2 = r1;	r2.right = r2.left + IBOXW; r2.bottom = r2.top + IBOXH;
    (*n1)->rec = r2;

    n1 = (*n1)->sib;	    /* Cboxid */
    r2.top = r2.bottom; r2.bottom = r1.bottom;
    (*n1)->rec = r2;

    n2 = (*n1)->child;	    /* Cplayid */
    r3 = r2;	r3.bottom = r3.top + BUTHGT;
    (*n2)->rec = r3;

#ifdef TRAX
    n2 = (*n2)->sib;	    /* Csynthid */
    r3.top = r3.bottom;	r3.bottom = r3.top + BUTHGT;
    (*n2)->rec = r3;
#endif /* TRAX */

    n2 = (*n2)->sib;	    /* Crecid */
    r3.top = r3.bottom;	r3.bottom = r3.top + BUTHGT;
    (*n2)->rec = r3;

    n2 = (*n2)->sib;	    /* Cprintid */
    r3.top = r3.bottom;	r3.bottom = r3.top + BUTHGT;
    (*n2)->rec = r3;

    n2 = (*n2)->sib;	    /* Cquitid */
    r3.top = r3.bottom+0.25*BUTHGT;	r3.bottom = r3.top + BUTHGT;
    (*n2)->rec = r3;

    n1 = (*n1)->sib;	    /* Lviewid */
    r2.left = r2.right;
    r2.top = r1.top;	r2.right = r1.right;	r2.bottom = r2.top + LVIEWH;
    (*n1)->rec = r2;

    n2 = (*n1)->child;	    /* Lvvscid */
    r3 = r2;	r3.top += LINEH;	r3.bottom -= SBarWidth;
    r3.right = r3.left + VAXLABWIDTH;
    (*n2)->rec = r3;

    n2 = (*n2)->sib;	    /* Lvtimlid */	
    r3.bottom = r3.top;	    r3.top = r2.top;	
    r3.left = r3.right-SBarWidth;	r3.right = r2.right;
    (*n2)->rec = r3;

    n2 = (*n2)->sib;	    /* Lvsndid */
    r3.top = r3.bottom + 1; r3.bottom = r2.bottom - SBarWidth;
    (*n2)->rec = r3;

    n2 = (*n2)->sib;	    /* Lvsbarid */
    r3.top = r3.bottom; r3.bottom = r2.bottom;
    (*n2)->rec = r3;

    n1 = (*n1)->sib;	    /* Sviewid */
    r2.top = r2.bottom;
    r2.bottom = r1.bottom;
    (*n1)->rec = r2;

    n2 = (*n1)->child;	   /* Svvscid */
    r3 = r2;	r3.right = r3.left + VAXLABWIDTH;
    r3.top += 1;	r3.bottom -= LINEH;
    (*n2)->rec = r3;

    n2 = (*n2)->sib;	   /* Svsndid */
    r3.left = r3.right - SBarWidth;	r3.right = r2.right;
    (*n2)->rec = r3;

    n1 = (*n2)->child;	    /* VscaleUpid - right hand edge */
    r1 = r3; 	
/*    r1.right = r1.left + SBarWidth;		*/
    r1.left = r1.right - SBarWidth;	
    r1.bottom = r1.top+SBarWidth;
    (*n1)->rec = r1;

    n1 = (*n1)->sib;	    /* VscaleDownid */
    r1.bottom = r3.bottom;	
    r1.top = r3.bottom-SBarWidth;
    (*n1)->rec = r1;

    n2 = (*n2)->sib;	    /* Svtimlid */
    r3.top = r3.bottom; r3.bottom = r2.bottom;
    (*n2)->rec = r3;
    }	

int RgeOverlap(int l1, int u1, int l2, int u2)
{
    return !((u1<l2) || (u2<l1));
}

int OverlapRects(Rect *pr1, Rect *pr2)
{	/* rects do not overlap if *either* of their axes do not */
    if(RgeOverlap(pr1->left, pr1->right, pr2->left, pr2->right)
       && RgeOverlap(pr1->top, pr1->bottom, pr2->top, pr2->bottom))
	return(1);
    else
	return(0);
}

DrawRecTreeInside(rrhan,wrcd,sibs,prec) /* sibs=0 => only draws self & children */
    RECRCD  **rrhan;
    WINRCD  **wrcd;
    int     sibs;
    Rect    *prec;
    {
    RECRCD **chdhan;
    int erased;

    while(rrhan!=NULL)
	{
	if(OverlapRects(&((*rrhan)->rec), prec))  {
	    erased = DrawTGpane((*rrhan)->id, (*rrhan)->rec, wrcd );
	    if(chdhan = (*rrhan)->child)  {
		if(erased)  {  /* parent rect erased - must redraw */
		    DrawRecTree(chdhan,wrcd,1);
		}
		else  {   /* only redraw if intersect prec */
		    DrawRecTreeInside(chdhan,wrcd,1, prec);
		}
	    }
	}
	if(sibs)
	    rrhan = (*rrhan)->sib;
	else
	    rrhan = NULL;
    }
    return(1);	    /* needed for RedrawBit search */
}

DrawRecTree(rrhan,wrcd,sibs)	 /* sibs=0 => only draws self & children */
    RECRCD  **rrhan;
    WINRCD  **wrcd;
    int     sibs;
    {
    RECRCD **chdhan;

    while(rrhan!=NULL)
	{
	DrawTGpane((*rrhan)->id, (*rrhan)->rec, wrcd );
	if(chdhan = (*rrhan)->child)
	    DrawRecTree(chdhan,wrcd,1);
	if(sibs)
	    rrhan = (*rrhan)->sib;
	else
	    rrhan = NULL;
	}
    return(1);	    /* needed for RedrawBit search */
    }

TGupdate(wrcd, prec)			    /* update the graph window */
    WINRCD  **wrcd;
    Rect *prec;		/* actually, just this bit */
    {
    Rect anotherRect;

    if(prec == NULL)  {		/* largely for MAC update call */
	prec = &anotherRect;
	RectOfWin( (*wrcd)->myWindow, prec);
    }

    SetPort( (*wrcd)->myWindow );
    SetCursor( &waitCursor );
    SetTGRects(wrcd);
/*  	  DrawRecTree( (*wrcd)->w.snd.recTreeH, wrcd, 0);	*/
    RedrawWinBitInside(wrcd, Wwinid, prec);
    SetCursor( &arrow );
    }

RedrawWinBitInside(wrcd,id,prec)
    WINRCD  **wrcd;
    int     id;
    Rect    *prec;
    {
    RgnHandle	oldClip;
    RECRCD  **rrh;
    Rect	portR;

    SetPort( (*wrcd)->myWindow );

    RectOfWin( (*wrcd)->myWindow, &portR);
    PushClip( &oldClip, portR );
    if( rrh = FindRhan( (*wrcd)->w.snd.recTreeH , id )  )
	DrawRecTreeInside(rrh, wrcd, 0, prec);
    PopClip( &oldClip );
    }	

RedrawWinBit(wrcd,id)
    WINRCD  **wrcd;
    int     id;
    {
    RgnHandle	oldClip;
    RECRCD  **rrh;
    Rect	portR;

    SetPort( (*wrcd)->myWindow );

    RectOfWin( (*wrcd)->myWindow, &portR);
    PushClip( &oldClip, portR );
    if( rrh = FindRhan( (*wrcd)->w.snd.recTreeH , id )  )
	DrawRecTree(rrh, wrcd, 0);
    PopClip( &oldClip );
    }	

RECRCD **FindRhan(rrh,id)	    /* search tree for pane id */
    RECRCD  **rrh;
    int     id;
    {
    RECRCD  **rch;

    while(rrh != NULL)
	{
	if( (*rrh)->id == id)
	    return(rrh);
	if( (rch = (*rrh)->child)!=NULL)
	    if( rch = FindRhan(rch, id) )
		return(rch);
	rrh = (*rrh)->sib;
	}
    return(rrh);
    }

#include <time.h>

static Rect   timeRect;
static time_t lastTime = 0;

char timestr[26];

AddTimeLabel(rec)	/* print time right-aligned in rect */
    Rect	rec;
    {
/*    char	timestr[26]; 	*/
    char	*n, *m;
    time_t	clock;
    struct tm 	*tm;
    Rect	r2;

    r2 = rec;
    r2.top += 3;	r2.right -= 5;	/* alignment fixup */
    time(&clock);
    tm = localtime(&clock);
    strcpy(timestr, asctime(tm));
    if(n = strrchr(timestr, ':'))
       if(m = strchr(n, ' '))
	   CopyChars(m, 1, n, 1, 1+strlen(m), REV);
    if(n = strchr(timestr, '\n'))
	*n = '\0';	/* hammer terminal NL */
    dbWWriteIn(&r2, timestr,1 /* right alignment*/);
    timeRect = rec;
    timeRect.left = timeRect.right - myTextWidth(timestr) - 4;
    lastTime = clock;
    }

UpdateTime()		/* maybe update timestamp */
    {
    time_t	clock;

    if(lastTime)	/* we have valid lastTime and timeRect */
	{
	time(&clock);
	if(clock-lastTime >= 60)	
	    {
	    /* over a minute since we wrote time */
	    EraseRect(&timeRect);
	    AddTimeLabel(timeRect);	/* rewrite time */
	    }
	}
    }

#define BTHOFF 8	/* horiz offset of button labels */

DrawTGpane( id, rec, wrcd )
    int     id;
    Rect    rec;
    WINRCD  **wrcd;
    {
    double   t1,t2,dur,max;
    short   *stt,*end;
    int	    chans = ((*wrcd)->w.snd.snd)->channels;
    int     scale = (*wrcd)->w.snd.scale;
    int	    mode  = (*wrcd)->w.snd.vMode;
    int     i;
    int     datatype;
static    int	    iar[1000], ic;	/* for tikmark return */
extern int deferErase;
    int     didErase = 0;


/*	D_prints("DrawTGpane ");	D_printn(id);	*/

    switch(id)
	{
	case Wwinid:
	    TextFont(4);		/* Monaco */
	    TextSize(9);		/* 9 pt   */
	    break;
	case Shoutid:
	    SetShout(rec);
	    break;
	case MenuBarid:
	    DrawMBar(wrcd, rec);
	    AddTimeLabel(rec);
	    break;
	case Infboxid:
	    InfoBox(wrcd, rec);
	    break;
	case Cboxid:
	    EraseRect(&rec);
	    didErase = 1;	/* must redraw all children */
	    OutlineR(rec);
	    break;
	case Cplayid:
	    InsetRect(&rec,2,2);
	    OutlineR(rec);
	    WriteString(rec.left+BTHOFF, rec.top, "Play");
	    break;
#ifdef TRAX
	case Csynthid:
	    InsetRect(&rec,2,2);
	    OutlineR(rec);
	    WriteString(rec.left+BTHOFF, rec.top, "Synth");
	    break;
#endif /* TRAX */
	case Crecid:
	    InsetRect(&rec,2,2);
	    OutlineR(rec);
	    WriteString(rec.left+BTHOFF, rec.top, "Record");
	    break;
	case Cquitid:
	    InsetRect(&rec,2,2);
	    OutlineR(rec);
	    WriteString(rec.left+BTHOFF, rec.top, "Quit");
	    break;
	case Cprintid:
	    InsetRect(&rec,2,2);
	    OutlineR(rec);
	    WriteString(rec.left+BTHOFF, rec.top, "Print");
	    break;
	case Lviewid:
	    EraseRect(&rec);
	    didErase = 1;	/* must redraw all children */
	    OutlineR(rec);			    /* Frame the whole area */
	    GutterR(rec,lSide|rSide|bSide|tSide,
		    VAXLABWIDTH, SBarWidth, LINEH, SBarWidth);
	    /* draw lines set in from edge */
	    break;
	case Lvvscid:
	    AddTimeLine(&rec,(FLOATARG)-32767.0,(FLOATARG)32767.0,1,0,0,iar,&ic);
	    break;
	case Lvtimlid:
	    VTimeLims(wrcd,&t1,&t2,&dur);	/* get real times for V-regn */
/*	    AddTimeLine(rec,(double)0.0, dur,0);		*/
	    InsetRect(&rec,SBarWidth,0);
	    AddTimeLine(&rec,(double)0.0,dur,0,1,0,iar,&ic);
	    break;
	case Lvsndid:
	    datatype = GetWholeSndPtrs((*wrcd)->w.snd.snd, &stt, &end);
	    ShowSound(rec,stt,end,chans,(int)monoVM,unityScale,datatype,NULL); 
	    /* draw a soundfile */
	    MarkSelect(wrcd,rec);
	    break;
	case Lvsbarid:
	    VTimeLims(wrcd,&t1,&t2,&dur);   /* get real times for V-regn */
	    DrawSBar(rec,t1/dur,t2/dur);
	    break;
	case Sviewid:
	    rec.bottom+=1;
	    if(deferErase==0)  {
		EraseRect(&rec);
		didErase = 1;	/* must redraw all children */
	    }
	    else{
		deferErase -= 1;
		if(deferErase!=0)	SetRed();	else SetBlack();
		}
	    rec.bottom -= 1;
	    OutlineR(rec);			/* Frame the whole area */
	    if(mode == psyDispVM)
		{
		/* psy disp has whole thing itself */
		GetSndBitePtrs((*wrcd)->w.snd.view, &stt, &end);
		psyDisp(&rec, stt, end, chans, (*wrcd)->w.snd.frqV);
		}
	    else{
		/* drw lins set in from edge */
		GutterR(rec,lSide|rSide|bSide, VAXLABWIDTH, SBarWidth, 
			SBarWidth, LINEH);
		}
	    if(mode == (int)spgrmVM)  {
		rec.left = rec.right - SBarWidth;
		rec.top += SBarWidth;  
		rec.bottom -= SBarWidth + LINEH;
		DrawGrayRange(&rec, 1);
	    }
	    break;
	case Svvscid:
	    if(mode != psyDispVM)  {
	    if(TimeMode(wrcd))
		{
		max = (double)32768*2/scale;
		for(i=0; i<chans; ++i)
		    {
		    Rect r2;
		    r2 = rec;
		    r2.top = rec.top + ((rec.bottom-rec.top)*i)/chans;
		    r2.bottom = rec.top + ((rec.bottom-rec.top)*(i+1))/chans;
		    AddTimeLine(&r2,-max,max,1,0,0,iar,&ic);
		    }
		}
	    else if(mode == (int)spgrmVM)
		{
		PVSTRUCT *pvh=NULL;
#ifdef CQFILE
		if( (*wrcd)->w.snd.cqi )
		    pvh = (*wrcd)->w.snd.cqi->pvh;
		if(pvh)
		    {
		    FRQVIEW *frqv = (*wrcd)->w.snd.frqV;
		    /* #if DIAGHACK */
extern double	minTrackDB;	/* cl option set up in db.xmain.c */
#define HMINFRQ 1000
#define HMAXFRQ 6000
		    if(minTrackDB == -51)	/* special flag */	
		    AddLogTimeLine(&rec, (double)HMINFRQ, (double)HMAXFRQ,
				   1, 0, 0, iar, &ic);
		    /* #endif /* DIAGHACK */
		    else {
			if(frqv->maxFrq > frqv->minFrq)  {
			    BoundFrqvF(frqv, pvh->minFreq, pvh->maxFreq);
			    AddLogTimeLine(&rec, frqv->minFrq, frqv->maxFrq,
					   1, 0, 0, iar, &ic);
			} else {
			    AddLogTimeLine(&rec, pvh->minFreq, pvh->maxFreq,
					   1, 0, 0, iar, &ic);
			}
		    }
		    }
#else	/* ft file */
		pvh  = (*wrcd)->w.snd.pvs;
		if(pvh)  {
		    FRQVIEW *frqv = (*wrcd)->w.snd.frqV;
		    if(frqv->maxFrq > frqv->minFrq)  {
			AddTimeLine(&rec, frqv->minFrq, frqv->maxFrq,
				    1, 0, 0, iar, &ic);
		    } else {	
			AddTimeLine(&rec, pvh->minFreq, pvh->maxFreq,
				    1, 0, 0, iar, &ic);
		    }
		}
#endif /* CQFILE / PVFILE */
		}
	    else if(mode == (int)ftVM || mode == (int)onefrVM)
		{
		Rect r2;
		r2 = rec;
		r2.bottom = (rec.top+rec.bottom)/2;
		AddTimeLine(&r2, (double)-100, (double)0,
			    1, 0, 0, iar, &ic);
		r2.top = r2.bottom;	r2.bottom = rec.bottom;
		if(chans == 2)	/* stereo mag spec */
		    AddTimeLine(&r2, (double)-100, (double)0,
				1, 0, 0, iar, &ic);
		else		/* mono phase spec */
		    AddTimeLine(&r2, (double)-3.5, (double)3.5,
				1, 0, 0, iar, &ic);
		}
	    }
	    break;
	case Svsndid:
	    if(mode != psyDispVM)  {
		int datatype;
	    datatype = GetSndBitePtrs((*wrcd)->w.snd.view, &stt, &end);
	    if( mode == (int)wfallVM )
		ShowWfall(rec,stt,end,chans,mode,scale,(*wrcd)->w.snd.wflen);
	    else if ( mode == (int)specVM )
		ShowSpex(wrcd,rec);	
	    else if ( mode == (int)spgrmVM )
		ShowSgram(wrcd,rec, (*wrcd)->w.snd.frqV);
	    else if ( mode == (int)ftVM )
		ShowFFT(rec,stt,end,chans,mode,scale, (*wrcd)->w.snd.frqV);
	    else if ( mode == (int)onefrVM )
		ShowOneFrame(wrcd, rec);
	    else	/* some kind of time-base view */
		{
		ShowSound(rec,stt,end,chans,mode,scale,datatype,NULL);  /* draw sound */
		MarkSelInShort(wrcd,rec);	/* overlay hilite */
		}
	    }
	    break;
	case Svtimlid:
	    if(mode != psyDispVM)  {
	    if( TimeMode(wrcd) || mode == (int)spgrmVM)
		{
		VTimeLims(wrcd,&t1,&t2,&dur);   /* get real times for V-regn */
		InsetRect(&rec,SBarWidth,0);
		AddTimeLine(&rec,t1,t2,0,1,1,iar,&ic);
		}
	    else if (mode == (int)wfallVM )
		DrawSBar(rec, ((double)(*wrcd)->w.snd.wflen)/WFLENMAX,
		          ((double)(*wrcd)->w.snd.wflen)/WFLENMAX);
#ifdef CQFILE
	    else if (mode == (int)onefrVM || mode == (int)specVM)
		{
		PVSTRUCT *pvh=NULL;
		if( (*wrcd)->w.snd.cqi )
		    pvh = (*wrcd)->w.snd.cqi->pvh;
		if(pvh)
		    AddLogTimeLine(&rec, pvh->minFreq, pvh->maxFreq,
				   0, 1, 1, iar, &ic);
		}
#endif /* CQFILE */
	    else {	/* spectra frq scale from FRQVIEW */
		InsetRect(&rec,SBarWidth,0);
/*		AddTimeLine(&rec,(double)0.0,
			    .5*((*wrcd)->w.snd.snd)->samplingRate,
			    0,1,1,iar,&ic);	*/
		AddTimeLine(&rec,(*wrcd)->w.snd.frqV->minFrq,
			    (*wrcd)->w.snd.frqV->maxFrq,
			    0,1,1,iar,&ic);
		}
	    }
	    break;
	case VscaleUpid:
	 /*   FillRect(&rec, ltGray);	*/
	    ShowIcon(rec, upArrowIcon);
	    FrameRect(&rec);
	    break;
	case VscaleDownid:
	 /*   FillRect(&rec, ltGray);  */
	    ShowIcon(rec, downArrowIcon);
	    FrameRect(&rec);
	    break;
	default:
	    break;
	}
    return didErase;	/* MUST we redraw children? */
    }
	

DrawSBar(rec,p1,p2)
    Rect    rec;
    FLOATARG   p1,p2;	    /* thumb limits */
    {
    short   l;
    Rect    r2;

    r2 = rec;
    r2.right = r2.left + SBarWidth;	/* or bottom-top for sq. ends */
    ShowIcon(r2, leftArrowIcon);  
    r2.right = rec.right;
    r2.left = r2.right - SBarWidth;
    ShowIcon(r2, rightArrowIcon);
    rec.bottom += 1;
    InsetRect(&rec, SBarWidth, 0);
    rec.right += 1;
    FrameRect(&rec);

    InsetRect(&rec,1,1);
    FillRect(&rec, ltGray);  
    r2.top = rec.top;	r2.bottom = rec.bottom;
    r2.left  = MapTtoX(p1, rec, (double)0.0, (double)1.0);
    r2.right = MapTtoX(p2, rec, (double)0.0, (double)1.0);
    if(r2.right < r2.left + SBarWidth)	r2.right = r2.left + SBarWidth; /**/
    EraseRect(&r2);
    FrameRect(&r2);
    }

ShowSound(r,s,e,c,m,scale,datatype,wndw)
    Rect    r;		/* where to draw */
    short   *s;		/* memory to draw - start */
    short   *e;		/*                -end */
    int     c;		/* number of channels of (short) data */
    int     m;		/* multi-channel drawing mode */
    int     scale;	/* vertical scaling factor */
    WIND	*wndw;	/* alternate graphout for ps */
    {
    Rect    r2;
    short   height;
    int	    i;
    int     (*showProc)(), ShowSoundShorts(), ShowSoundFloats();

    if(datatype == SFMT_SHORT)
	showProc = ShowSoundShorts;
    else if (datatype == SFMT_FLOAT)
	showProc = ShowSoundFloats;
    else  {  fprintf(stderr, "ShowSound: datatype %d is not short or float\n",
		     datatype);	abort();  }

    if (c==1)
	(*showProc)(r,s,e,1,scale,wndw);	/* mono - no probs */
    else if(m==(int)monoVM)		/* single line - treat as mono */
	(*showProc)(r,s,e,1,scale,wndw);
    else{
	r2 = r;
	height = r.bottom - r.top;
	for(i=0; i<c; ++i)
	    {
	    r2.bottom = r.top+(height*(i+1))/c;
	    MoveTo(r2.left, r2.bottom);
	    LineTo(r2.right, r2.bottom);
	    (*showProc)(r2,s+(long)i,e+(long)i,c,scale,wndw);
	    r2.top = r2.bottom;
	    }
	}
    }

#define POLYLSIZE	1000	/* max line segments */
#define PSWIDTH		600
#define PSHEIGHT	170	/* 400 */

ShowSoundShorts(r,s,e,k,vScale,wndw)
    Rect    r;		/* where to draw */
    short   *s;		/* memory to draw - start */
    short   *e;		/*                - end */
    register int     k;          /* how many bytes to skip */
    int     vScale;	/* vertical scaling factor : 2==unity */
    WIND	*wndw;	/* alternative graphics model for ps output 193mar13 */
    {
    register short *t, *limp;
    register int  most;
    register short d,mn,mx;
    long    len;
    long   hpix,vpix,px,lx;
    long   o,p;
    long   mx2,mn2,tmp;
    RgnHandle	oldClip;
    int	 	mval = 1000;	/* at most this many samples per pixel */
    long   lenapprox, hpixapprox;
#ifdef PSOUTPUT
    float	topPolyLine[POLYLSIZE][2];
    float	botPolyLine[POLYLSIZE][2];
    FILE	*psOutFile;

    psOutFile = fopen("sound.idw","w");
    SetPSbottom(PSSOUNDBOTTOM);
    SetPSLineWidth(0.5);
#endif /* PSOUTPUT */
    /* new ps output */
static int plyTop[POLYLSIZE][2];
static int plyBot[POLYLSIZE][2];
    int 	topSize = 0, botSize = 0;

    PushClip(&oldClip,r);

    if(wndw == NULL)  {	/* fixup for non-wndw mode */
	InsetRect(&r,SBarWidth+1,1);
	r.left -= 0;
	r.right -= 1;	/* */
    }
    len = e - s;
    vpix = (r.bottom - r.top);
    hpix = r.right - r.left + 1;
    px = 0; lx = r.left;
    o = r.top+vpix/2;
    vpix = (vpix * vScale)/2;

    if(len>100000)	/* say - danger of overflow */
	{ lenapprox = len / hpix;	hpixapprox = 1;		}
    else
	{ lenapprox = len;		hpixapprox = hpix;	}
    t = s;
    mx = *t;
    mn = *t;
/*    t += k;		/* pre increment for dec. in loop */
    while(t<e)
	{
	limp = s+k*((++px)*lenapprox / (k*(long)hpixapprox));
/*	t -= (long)k; */
	most = mval;
	do  {
	    d = *t;
	    t += (long)k;
	    if(d>mx) mx = d;
	    if(d<mn) mn = d;
	    } while(t<limp && --most);
	t = limp;		/* in case 'most' cut us off */
	mx2 = (short) ((mx * (long)vpix)>>16);
	mn2 = (short) ((mn * (long)vpix)>>16);
#ifdef PSOUTPUT
	botPolyLine[botSize][0] = 
	    topPolyLine[topSize][0] = (px*PSWIDTH)/hpix;
	topPolyLine[topSize++][1] = ((mx * PSHEIGHT)>>16)+PSHEIGHT/2;
	botPolyLine[botSize++][1] = ((mn * PSHEIGHT)>>16)+PSHEIGHT/2;
	if(topSize >= POLYLSIZE)	--topSize;
	if(botSize >= POLYLSIZE)	--botSize;
#endif /* PSOUTPUT */
	if(wndw)  {
	    plyTop[topSize][0] = plyBot[botSize][0] = lx++;
	    plyTop[topSize++][1] = o-mx2;
	    plyBot[botSize++][1] = o-mn2;
	} else { /* regular output */
	    MoveTo(lx,o-mn2);
	    LineTo(lx++,o-mx2);
	}
	tmp = mx; mx = mn; mn = tmp;
	}
/*    MoveTo(lx,o-mn2);
    LineTo(lx++,o-mx2);
 */
#ifdef PSOUTPUT
    if(psOutFile)
	{
	IdrawPolyLineOutput(topPolyLine, topSize, psOutFile);
	IdrawPolyLineOutput(botPolyLine, botSize, psOutFile);
	}
#endif /* PSOUTPUT */
    if(wndw)  {
	WPolyLine(wndw, plyTop, topSize);
	WPolyLine(wndw, plyBot, topSize);
    }
    PopClip( &oldClip );
    }

ShowSoundFloats(r,s,e,k,vScale)
    Rect    r;		/* where to draw */
    float   *s;		/* memory to draw - start */
    float   *e;		/*                - end */
    register int     k;          /* how many bytes to skip */
    int     vScale;	/* vertical scaling factor : 2==unity */
    {
    register float *t, *limp;
    register int  most;
    register double d,mn,mx;
    long    len;
    long   hpix,vpix,px,lx;
    long   o,p;
    long   mx2,mn2,tmp;
    RgnHandle	oldClip;
    int	 	mval = 1000;	/* at most this many samples per pixel */
    long   lenapprox, hpixapprox;
#ifdef PSOUTPUT
    float	topPolyLine[POLYLSIZE][2];
    float	botPolyLine[POLYLSIZE][2];
    int 	topSize = 0, botSize = 0;
    FILE	*psOutFile;

    psOutFile = fopen("sound.idw","w");
    SetPSbottom(PSSOUNDBOTTOM);
    SetPSLineWidth(0.5);
#endif /* PSOUTPUT */

    PushClip(&oldClip,r);

    InsetRect(&r,SBarWidth+1,1);
    r.left -= 0;
    r.right -= 1;	/* */
    len = e - s;
    vpix = (r.bottom - r.top);
    hpix = r.right - r.left + 1;
    px = 0; lx = r.left;
    o = r.top+vpix/2;
    vpix = (vpix * vScale)/2;

    if(len>100000)	/* say - danger of overflow */
	{ lenapprox = len / hpix;	hpixapprox = 1;		}
    else
	{ lenapprox = len;		hpixapprox = hpix;	}
    t = s;
    mx = *t;
    mn = *t;
/*    t += k;		/* pre increment for dec. in loop */
    while(t<e)
	{
	limp = s+k*((++px)*lenapprox / (k*(long)hpixapprox));
/*	t -= (long)k; */
	most = mval;
	do  {
	    d = *t;
	    t += (long)k;
	    if(d>mx) mx = d;
	    if(d<mn) mn = d;
	    } while(t<limp && --most);
	t = limp;		/* in case 'most' cut us off */
	mx2 = (short) ((mx * (long)vpix)/2);
	mn2 = (short) ((mn * (long)vpix)/2);
#ifdef PSOUTPUT
	botPolyLine[botSize][0] = 
	    topPolyLine[topSize][0] = (px*PSWIDTH)/hpix;
	topPolyLine[topSize++][1] = ((mx * PSHEIGHT)/2)+PSHEIGHT/2;
	botPolyLine[botSize++][1] = ((mn * PSHEIGHT)/2)+PSHEIGHT/2;
	if(topSize >= POLYLSIZE)	--topSize;
	if(botSize >= POLYLSIZE)	--botSize;
#endif /* PSOUTPUT */
	MoveTo(lx,o-mn2);
	LineTo(lx++,o-mx2);
	tmp = mx; mx = mn; mn = tmp;
	}
/*    MoveTo(lx,o-mn2);
    LineTo(lx++,o-mx2);
 */
#ifdef PSOUTPUT
    if(psOutFile)
	{
	IdrawPolyLineOutput(topPolyLine, topSize, psOutFile);
	IdrawPolyLineOutput(botPolyLine, botSize, psOutFile);
	}
#endif /* PSOUTPUT */
    PopClip( &oldClip );
    }

InfoBox(wrcd, r)		/* Write up the info panel for window wrcd in rect r */
    WINRCD  **wrcd;
    Rect    r;
    {
    RgnHandle	oldClip;
    Str255	s,t;
    char	*s2;
    double	dur,t1,t2;
    int		dSize;

    PushClip(&oldClip,r);

    EraseRect(&r);
    FrameRect(&r);
    TextFont(4);		/* Monaco */
    TextSize(9);		/* 9 pt   */

    if( ((*wrcd)->w.snd.snd)->channels==1)
	mystrcpy((char *)s,"Mono");
    else
	mystrcpy((char *)s,"Stereo");
    InfoString(r,0,1,(char *)s,1);

    if( ((*wrcd)->w.snd.snd->format)==SFMT_SHORT)
	mystrcpy((char *)s,"16 bit");
    else if( ((*wrcd)->w.snd.snd->format)==SFMT_FLOAT)
	mystrcpy((char *)s,"float");
    else
	mystrcpy((char *)s,"unrec fmt");

    InfoString(r,0,2,(char *)s,1);

    myftoa( ((*wrcd)->w.snd.snd)->samplingRate/1000.0, (char *)s, 2);
    mystrcat((char *)s," kHz");
    InfoString(r,0,3,(char *)s,1);
	
/*    VTimeLims(wrcd,&t1,&t2,&dur);
    myftoa(dur, (char *)s, 3);
    mystrcat((char *)s," s");
    InfoString(r,0,4,"Length",0);
    InfoString(r,0,5,(char *)s,1);

    mystrcpy((char *)s,"View:");
    InfoString(r,0,6,(char *)s,0);
    myftoa(t1,(char *)s ,3);
    mystrcat((char *)s," to");
    InfoString(r,0,7,(char *)s,0);
    myftoa(t2,(char *)s ,3);
    mystrcat((char *)s," s");
    InfoString(r,0,8,(char *)s,1);

    STimeLims(wrcd,&t1,&t2,&dur);
    mystrcpy((char *)s,"Select:");
    InfoString(r,0,9,(char *)s,0);
    myftoa(t1,(char *)s ,3);
    mystrcat((char *)s," to");
    InfoString(r,0,10,(char *)s,0);
    myftoa(t2,(char *)s ,3);
    mystrcat((char *)s," s");
    InfoString(r,0,11,(char *)s,1);
 */
    STimeLims(wrcd,&t1,&t2,&dur);

    myftoa(dur,(char *)s, 3);
    mystrcat((char *)s," s");
    InfoString(r,0,4,"Length",0);
    InfoString(r,0,5,(char *)s,1);

    r.top += 5*LINEH + 1;
    r.left += 3;
    SelInfo(wrcd, &r);

    if(0)  {
    mystrcpy((char *)s,"Select:");
    InfoString(r,0,6,(char *)s,0);
    myftoa(t1,(char *)s ,3);
    mystrcat((char *)s," to");
    InfoString(r,0,7,(char *)s,0);
    myftoa(t2,(char *)s ,3);
    mystrcat((char *)s," s");
    InfoString(r,0,8,(char *)s,0);
    mystrcpy((char *)s,"( #");
    s2 = (char *)s+3;	/* end of above string */
    NumToString(((*wrcd)->w.snd.slct)->bStart/sizeof(short),
				(unsigned char *)s2);
    PtoCstr((unsigned char *)s2);
    InfoString(r,0,9,(char *)s,0);
    mystrcpy((char *)s,"= ");
    s2 = (char *)s+2;	/* end of above string */
    NumToString(*((short *)SFDataLoc((*wrcd)->w.snd.snd)
		  +(((*wrcd)->w.snd.slct)->bStart/sizeof(short))),
				(unsigned char *)s2);
    PtoCstr((unsigned char *)s2);
    mystrcat((char *)s,")");
    InfoString(r,0,10,(char *)s,1);
}

    PopClip( &oldClip );
    }

/* dpwe 1993jan25 : interactively update select description during sel */
SelInfo(WINRCD **wrcd, Rect *rec)
{	/* write up select info, store reference in statics */
    int 	lineh = LINEH;
    Rect	r2;
    char	str[32];
    static Rect sRec;
    double	t1, t2, dur;

    STimeLims(wrcd,&t1,&t2,&dur);	
    if(rec == NULL)  {
	rec = &sRec;
    }
    else {
	sRec = *rec;
    }
    EraseRect(rec);
    r2 = *rec;
    r2.bottom = r2.top + lineh;
    dbWWriteIn(&r2, "Select:", -1);

    r2.top += lineh;
    r2.bottom = r2.top + lineh;
    sprintf(str, "%.6f :", t1);
    dbWWriteIn(&r2, str, -1);
    
    r2.top += lineh;
    r2.bottom = r2.top + lineh;
    sprintf(str, "%.6f s", t2);
    dbWWriteIn(&r2, str, -1);
    
    r2.top += lineh;
    r2.bottom = r2.top + lineh;
    sprintf(str, "(#%ld", ((*wrcd)->w.snd.slct)->bStart/sizeof(short));
    dbWWriteIn(&r2, str, -1);
    
    r2.top += lineh;
    r2.bottom = r2.top + lineh;
    sprintf(str, "= %d)", *((short *)SFDataLoc((*wrcd)->w.snd.snd)
			    +(((*wrcd)->w.snd.slct)->bStart/sizeof(short))));
    dbWWriteIn(&r2, str, -1);
}
    

double Duration(snd)	    /* returns double of duration of timegraph */
    SFSTRUCT 	*snd;
    {
    long    	bLen;
    double   	dur;

    bLen = snd->dataBsize;
    dur  = ((double)bLen/(Ssize(snd)))/snd->samplingRate;
    return(dur);
    }

Ssize(snd)		    /* bytes per sample frame */
    SFSTRUCT  	*snd;
    {
    return(SFSampleFrameSize(snd));
    }

double	FloatProp(whole,num,denom)
    FLOATARG   whole;
    long    num,denom;
    {
    if(num == 0L)
	return(0.0);
    else
	return( whole*( /* .5 + */ (double)num)/(double)denom );
    }

extern double FloatPropX(FLOATARG whole, long num, long denom);

/*  double	FloatPropX(whole,num,denom)
/*      FLOATARG   whole;
/*      long    num,denom;
/*      {
/*      if(num == 0L)
/*  	return(0.0);
/*      else
/*  	return( whole*( .0  + (double)num)/(double)denom );
/*      }
 */     /*  FloatPropX now in lib in winwndw.c */

long	LongProp(whole,num,denom,gran)
    long    whole;
    FLOATARG   num,denom;
    int     gran;	/* granularity i.e. minimum incremement of 'whole' */
    {
    return(((long)gran) * (long)( .5 + /*-.5+*/ (num/denom) * (double)(whole/(long)gran) ) );
    }

#ifdef THINK_C	/* LongPropY, MapFtoI now in lib/winwndw.c */
long	LongPropY(whole,num,denom,gran)
    long    whole;
    FLOATARG   num,denom;
    int     gran;	/* granularity i.e. minimum incremement of 'whole' */
    {
    return(((long)gran) * (long)( .5 + (num/denom) * (double)(whole/(long)gran) ) );
    }

int MapFtoI(f, fl, fu, il, iu)
    FLOATARG f, fl, fu;		/* value in range */
    int	il, iu;			/* fl and fu map to these ints */
    {
    return il + (int)LongPropY( (long)(iu - il), (f-fl), (fu-fl), 1);
    }
#endif /* THINK_C */

/*  double MapItoF(i, il, iu, fl, fu)
/*      int	i, il, iu;
/*      FLOATARG fl, fu;
/*      {
/*      return fl + FloatPropX( fu-fl, (long)(i-il), (long)(iu-il));
/*      }
 */     /*  MapItoF now in lib in winwndw.c */

short	MapTtoX(t, rec, t1, t2)     /* convert a time to an X co-o */
    FLOATARG   t;
    Rect    rec;
    FLOATARG   t1,t2;		    /* time limits corresponding to rec */
    {				/* was Y */
    return( rec.left + (short)LongProp( (long)(rec.right-rec.left), 
				       (t-t1), (t2-t1), 1) );
    }

double	MapXtoT(x, rec, t1, t2)     /* converse of above */
    int     x;
    Rect    rec;
    FLOATARG   t1,t2;
    {
    return( t1 + /* was X */
	   FloatProp( (t2-t1), (long)(x-rec.left) , 
		                (long)(rec.right-rec.left) ) );
    }

VTimeLims(wrcd,pstt,pend,pdur)	/* calcu real times of stt & end of v-regn */
    WINRCD  **wrcd;
    double   *pstt,*pend,*pdur;	/* where to put answer */
    {
    BiteTimeLims( (*wrcd)->w.snd.view, pstt, pend, pdur);
    }

SetVTimeLims(wrcd,t1,t2,dur)	/* set up V mrkrs for new times as % of dur */
    WINRCD  **wrcd;		/* ignrs any discrep btwn dur passed  & true */
    FLOATARG   t1,t2,dur;
    {
    short   sbytes;
    long    bLen = ((*wrcd)->w.snd.snd)->dataBsize;

    sbytes = Ssize((*wrcd)->w.snd.snd);
    ((*wrcd)->w.snd.view)->bStart = LongProp(bLen, t1, dur, sbytes);
    ((*wrcd)->w.snd.view)->bEnd   = LongProp(bLen, t2, dur, sbytes);
    }

STimeLims(wrcd,pstt,pend,pdur)	/* real times for selected region */
    WINRCD  **wrcd;
    double   *pstt,*pend,*pdur;	/* where to put answer */
    {
    BiteTimeLims( (*wrcd)->w.snd.slct, pstt, pend, pdur );
    }

SetSTimeLims(wrcd,t1,t2,dur)	/* set up S mrkrs for new times as % of dur */
    WINRCD  **wrcd;		/* ignrs any discrep btwn dur passed & true */
    FLOATARG   t1,t2,dur;
    {
    short   sbytes;
    long    bLen = ((*wrcd)->w.snd.snd)->dataBsize;

    sbytes = Ssize((*wrcd)->w.snd.snd);
    ((*wrcd)->w.snd.slct)->bStart = LongProp(bLen, t1, dur, sbytes);
    ((*wrcd)->w.snd.slct)->bEnd   = LongProp(bLen, t2, dur, sbytes);
    }

mystrcat(s1,s2)     /* sticks s2 on the end of s1 - if you have enough room */
    char    *s1,*s2;
    {
    while(*s1++)
	;	    /* scan ptr to end of s1 */
    --s1;	    /* *s1 = '\0' */
    while(*s1++ = *s2++)
	;	    /* copy over s2 until you copy a '\0' */
    }

mystrcpy(s1,s2)     /* copy s2 into s1 */
    char    *s1,*s2;
    {
    while(*s1++ = *s2++)
	;
    }

