/*******************************************************************\
*   db.tgwin.c		       					    *
*   time graph window general functions for DSPbench		    *
*   29mar90 dpwe	       				   	    *
\*******************************************************************/

#include <string.h>
#include "DSPbench.h"
#include <genutils.h>

/* externs */
extern	Handle	leftArrowIcon, rightArrowIcon;
extern  int     done;
extern	Cursor	arrow, waitCursor, xCursor, pauseCursor;
extern  int LINEH;

extern 	int directopen;

/* globals */
int deferErase = 0;
int deferShout = 0;

/* prototypes */
void ClikOverSpectra PARG((WINRCD **wrcd, Point *pwhere, Rect *prec, 
			   int btn, int vmode));

GrowTGwin(wrcd, p)
    WINRCD  **wrcd;
    Point p;
    {
    SetTGRects(wrcd);			/* refill the family of TG rects */
    InvalWinRect(&(*wrcd)->myWindow);
    }

NewTGwin(wrcd)
    WINRCD  **wrcd;
    {
    (*wrcd)->w.snd.snd = NULL;		/* no soundfile attached */
#ifdef CQFILE
    (*wrcd)->w.snd.cqi = NULL;
#else
    (*wrcd)->w.snd.pvs = NULL;		/* no pvoc attached */
#endif
    (*wrcd)->w.snd.frqV = NULL;		/* no frqview created */
    (*wrcd)->w.snd.pixmap = NULL;	/* no pixmap yet */
    (*wrcd)->w.snd.pixW = 0;
    (*wrcd)->w.snd.pixH = 0;
    
    BuildTGRecTree(wrcd);		/* build a family of TG rects */
    }

/* what will our faked-out raw soundfile header be? */
#define FAKE_FMT	SFMT_SHORT
#define FAKE_SRATE	(44100.0)
#define FAKE_CHANS	1		/* arguable */

int RawSFread(name, psnd)	    	/* like SFread, but fakes it all */
    char	*name;			/* file to read from */
    SFSTRUCT	**psnd;			/* where to put result */
    {
    int 	err = SFE_NOPEN;	/* sndf error code */
    FILE	*fp;
    long 	blen, frmlen, frmsize, red;

    if( (fp = fopen(name, "rb")) != NULL )	/* can open at least */
	{
	blen = fSize(fp);
	frmsize = FAKE_CHANS*SFSampleSizeOf(FAKE_FMT);
	frmlen  = blen / frmsize;	/* express byte len as frames */
	blen = frmlen * frmsize;	/* fixup # bytes to read */
	if((err = SFAlloc(psnd, blen, FAKE_FMT, FAKE_SRATE, FAKE_CHANS, 0))==0)
	    if((red = fread(SFDataLoc(*psnd), frmsize, frmlen, fp))<frmlen)
		err = SFE_RDERR;
	fclose(fp);
	if(err != 0)
	    SFFree(*psnd);
	}
    return err;
    }

int ReadTGfile(wrcd, path)	    /* read in a sound file */
    WINRCD  **wrcd;
    char    *path;
    {
    int	err = 0;
    SFSTRUCT **psnd = &((*wrcd)->w.snd.snd);
    char sfName[128],pvName[128];
    char *s,*rindex();
extern int forceChans;

#ifdef THINK_C
    PtoCstr((unsigned char *)path);
    getPath(sfName,path,(*wrcd)->theVRefNum);
    CtoPstr(path);
#else
    strcpy(sfName, path);
#endif
    D_prints("sound file '");
    D_prints(sfName);	D_prints("' loading...\n");
    err = SFRead(sfName, psnd);
#ifdef CQFILE
    SubstExtn(sfName, ".aqt", pvName);
#else
    SubstExtn(sfName, ".aft", pvName);
#endif
    if(err)
    	{
    	D_prints("err in SFRead:");
    	D_prints(SFErrMsg(err));
    	D_prints(" - attempt raw read\n");
	if(err = RawSFread(sfName, psnd))
	    return(err);
	}
    /* have to set up this field before opening anal file */
    (*wrcd)->w.snd.frqV = NewFrqView((*psnd)->samplingRate);
#ifdef CQFILE
    if(OpenCQTPVfile(pvName, &((*wrcd)->w.snd.cqi))==NULL)
	{
	(*wrcd)->w.snd.cqi = NULL;
	D_prints("CQTPV file '");
	D_prints(pvName);
	D_prints("' not found.\n");
	}
    else{
	D_prints("CQTPV file '");
	D_prints(pvName);
	D_prints("' loaded.\n");
	SetFrqView((*wrcd)->w.snd.frqV, (*wrcd)->w.snd.cqi);
	LoadTrackFile(wrcd);
	}
#else
/*    if(OpenAFTPVfile(pvName, &((*wrcd)->w.snd.pvs))==NULL)
	(*wrcd)->w.snd.pvs = NULL;	*/
    if(err = PVReadFile(pvName, &((*wrcd)->w.snd.pvs) ) )
    {
	D_prints("AFTPV file '");
	D_prints(pvName);
	D_prints("' not found.\n");
	}
    /* couldn't read .. leave ptr as NULL */
    else{
	D_prints("AFTPV file '");
	D_prints(pvName);
	D_prints("' loaded.\n");
	SetFrqView((*wrcd)->w.snd.frqV, (*wrcd)->w.snd.pvs);
	}
#endif
    MakeSndBite(&((*wrcd)->w.snd.view),*psnd,0L,(*psnd)->dataBsize);
    MakeSndBite(&((*wrcd)->w.snd.slct),*psnd,0L,0L);
    (*wrcd)->w.snd.vMode = (int)multiVM;
    (*wrcd)->w.snd.scale = 2;
    (*wrcd)->w.snd.wflen = 1024;
    /* now add-in open hacks before rendering window */
    if(forceChans > 0)
	(*psnd)->channels = forceChans;
    if(directopen)  {
	(*wrcd)->w.snd.vMode = (int)spgrmVM;
	((*wrcd)->w.snd.view)->bStart = 
	    SFSampleFrameSize((*wrcd)->w.snd.snd)
	    * (long)(((*wrcd)->w.snd.snd)->samplingRate* 0.820);
	((*wrcd)->w.snd.view)->bEnd   = 
	    SFSampleFrameSize((*wrcd)->w.snd.snd)
	    * (long)(((*wrcd)->w.snd.snd)->samplingRate* 1.670);
	(*wrcd)->w.snd.frqV->minFrq = 350;
	(*wrcd)->w.snd.frqV->maxFrq = 9999;
    }
	
    return(0);
    }

TGdispose( wrcd )
    WINRCD  **wrcd;
    {
    if( ( (*wrcd)->w.snd.snd ) != NULL)
	SFFree( (*wrcd)->w.snd.snd );     /* soundfile space */
#ifdef CQFILE
    if( ( (*wrcd)->w.snd.cqi ) != NULL)
	FreeCQTInfo( (*wrcd)->w.snd.cqi);
#else
    if( ( (*wrcd)->w.snd.pvs ) != NULL)
	PVFree( (*wrcd)->w.snd.pvs );     /* ftfile space */
#endif
    }

static Rect shoutRec;

SetShout(rec)			/* called to set domain for shout */
    Rect rec;
    {
    shoutRec = rec;
    rec.right += 1;	rec.bottom += 1;
    FrameRect(&rec);
    }

void shout(a,b,c,d)			/* show mouse button hints */
    char *a, *b, *c, *d;
    {
/*    fprintf(stderr,"shout: %s / %s / %s  (%s)\n",a,b,c,d);  */
    Rect rec;
    int  len;
    char	msg[64];
    if(deferShout)	/* skip writing up new shout (other messages) */
	{
	--deferShout;
	return;
	}

    rec = shoutRec;
    rec.top += 1;
    rec.bottom += 1;
    EraseRect(&rec);
    len = rec.right - rec.left;
    rec.right = rec.left + (len*3)/10;
    MoveTo(rec.left,rec.bottom);
    mystrcpy(msg,"LBut: ");
    mystrcat(msg,a);
    WriteString(rec.left,rec.top,msg);
    mystrcpy(msg,"MBut: ");
    mystrcat(msg,b);
    WriteString(rec.left+(len*3)/10,rec.top,msg);
    mystrcpy(msg,"RBut: ");
    mystrcat(msg,c);
    WriteString(rec.left+(len*6)/10,rec.top,msg);
    mystrcpy(msg,"(");
    mystrcat(msg,d);
    mystrcat(msg,")");
    WriteString(rec.left+(len*9)/10,rec.top,msg);
    }

MoveInTG(wrcd, theEvent)
    WINRCD	**wrcd;
    EventRecord *theEvent;
    {
    Point	where;
    int		ShoutTGPane();

    WhereOfMotionEvent(theEvent, &where);
    ScanRecTree( where, (*wrcd)->w.snd.recTreeH, wrcd, ShoutTGPane, 0L);
    }

ClickInTG(wrcd, theEvent)
    WINRCD	**wrcd;
    EventRecord *theEvent;
    {
    Point	where;
    int		ClikTGpane();
    int		btn = ButtonOfEvent(theEvent);

    WhereOfEvent(theEvent, &where);
    ScanRecTree( where, (*wrcd)->w.snd.recTreeH, wrcd, ClikTGpane, (long)btn);
    }

ScanRecTree(where,rrhan,wrcd, pfn, data)
    Point   where;
    RECRCD  **rrhan;
    WINRCD  **wrcd;
    int	    (*pfn)();
    long    data;
    {
    RECRCD  **chdhan;
    int     rc = 0;

    while(rrhan!=NULL)
	{
	if( PtInRect(where, &(*rrhan)->rec) )
	    {
	    chdhan = (*rrhan)->child;
	    if(chdhan!=NULL)
		rc = ScanRecTree(where,chdhan,wrcd,pfn,data);
	    if(rc == 0)
		(*pfn)( (*rrhan)->id, (*rrhan)->rec, where, wrcd, data);
	    return(1);
	    }
	rrhan = (*rrhan)->sib;
	}
    return(0);
    }


ClikTGpane( id, rec, where, wrcd, btn )
    int     id;
    Rect    rec;
    Point   where;
    WINRCD  **wrcd;
    long    btn;
    {
    double   t1,t2,dur;

    switch(id)
	{
	case Wwinid:
	    break;
	case MenuBarid:
	    DoMenuItem(wrcd,MClick(wrcd,rec,where));
	    break;
	case Infboxid:
	    XtractSelSnd(wrcd);
	    break;
	case Cboxid:
	    /* done = 1;	/* cbox for exit for now */
	    break;
	case Cplayid:
	    PlaySelSnd(wrcd, (int)btn);
	    break;
#ifdef TRAX
	case Csynthid:
	    SynthFromTracks(wrcd);
	    break;
#endif /* TRAX */
	case Crecid:
	    RecordSelSnd(wrcd, (int)btn);
	    if(btn == 1)
		{
		RedrawWinBit(wrcd,Lviewid);
		RedrawWinBit(wrcd,Sviewid);
		}
	    break;
	case Cquitid:
	    done = 1;		/* flag to cease */
	    break;
	case Cprintid:
	    PrintBtn(wrcd, (int)btn);
	    break;
	case Lviewid:
	    break;
	case Lvtimlid:
	    SelToView(wrcd);	/* clik in time line zooms to select */
	    break;
	case Lvvscid:
	case Lvsndid:
	    if(btn == 2)	/* mid button zooms */
		SelToView(wrcd);
	    else
		{
		dur = Duration( (*wrcd)->w.snd.snd );
		LVSndClik(wrcd, rec,(double)0.0, dur, dur, where, (int)btn);
		}
	    break;
	case Lvsbarid:
	    LVSBClik(wrcd, rec, where);
	    break;
	case Sviewid:
	    break;
	case Svvscid:
	case Svsndid:
	    if( TimeMode(wrcd) )
		{
		if(btn == 2)
		    SelToView(wrcd);
		else
		    {
		    VTimeLims(wrcd, &t1, &t2, &dur);
		    LVSndClik(wrcd, rec, t1, t2, dur, where, (int)btn);
		    }
		}
	    else if( btn != 2 && (*wrcd)->w.snd.vMode == (int)spgrmVM)  {
		ClikZoomOverTrax(wrcd, &where, &rec, (int)btn);
	    }
#ifdef TRAX
	    else if( btn == 2 &&(*wrcd)->w.snd.vMode == (int)spgrmVM)  {
		ClikOverTrax(wrcd, &where, &rec, (int)btn);
	    }
#endif
	    else if( /* (*wrcd)->w.snd.vMode == (int)spgrmVM || /* btn 1,3 */
		    (*wrcd)->w.snd.vMode == (int)specVM || /* many frms */
		    (*wrcd)->w.snd.vMode == (int)ftVM ||    /* FFT */
		    (*wrcd)->w.snd.vMode == (int)onefrVM ||  /* 1frame */
		    (*wrcd)->w.snd.vMode == (int)psyDispVM )  /* psycho */
		ClikOverSpectra(wrcd, &where, &rec, (int)btn, 
				(*wrcd)->w.snd.vMode);


	    break;
	case Svtimlid:
	    if( TimeMode(wrcd) )
		SelToView(wrcd);   /* clik in time line zooms to select */
	    else if( (*wrcd)->w.snd.vMode == (int)wfallVM )
		WFSBClik(wrcd,rec,where);	/* waterfall line length */
	    break;
	case VscaleUpid:
	    ChangeVscale(wrcd,1);
	    break;
	case VscaleDownid:
	    ChangeVscale(wrcd,-1);
	    break;
	default:
	    D_prints("Click in ????\n");
	    break;
	}
    }

int IsButton(id)
    int id;
    {
    return (id==Cplayid || id ==Crecid || id == Cquitid || id == Cprintid
#ifdef TRAX
	    || id == Csynthid
#endif /* TRAX */
	    );
    }

static Rect HiRect;
void Hilite(id, prec)
    int id;
    Rect *prec;
    {
    PenState  dummy;

    SetDottedXorPen(&dummy);
    HiRect = *prec;
    InsetRect(&HiRect, 1, 1);
/*    FrameRect(&HiRect);	*/
    OutlineR(HiRect);
    RestoreSolidCopyPen(&dummy);
    }
void UnHilite(id)
    int id;
    {    /* if id != id stored in last Hilite, panic! */
    PenState  dummy;

    SetDottedXorPen(&dummy);
/*    FrameRect(&HiRect);	*/
    OutlineR(HiRect);
    RestoreSolidCopyPen(&dummy);
    }

#define ARROW	0
#define CROSS   1	/* for cursor state static */

ShoutTGPane( id, rec, where, wrcd, dmy ) /* dispatch 'shout' msgs for rgns */
    int     id;
    Rect    rec;
    Point   where;
    WINRCD  **wrcd;
    long    dmy;		/* dummy as part of (*pfn)() stuff */
    {
    static  int lastShout = (int)Nullid;	/* last shout box */
    static  int curs = ARROW;
    char    strg[32];

    if(id != lastShout )
	{
	/* mouse form hack */
	if(curs != ARROW)
	    {
	    SetCursor(&arrow);
	    curs = ARROW;
	    }
	if(IsButton(lastShout))
	    UnHilite(lastShout);
	if(IsButton(id))
	    Hilite(id, &rec);	
	switch(id)
	    {
	case Wwinid:
	    shout("?","?","?","Wwin");
	    break;
	case MenuBarid:
	    shout("Select menu item","","","MenuBar");
	    break;
	case Infboxid:
	    shout("Extract marked","","","InfBox");
	    break;
	case Cboxid:
	    shout("?","?","?","Cbox");
	    break;
	case Cplayid:
	    shout("Play marked","pause/Reverse","Loop marked","Cplay");
	    break;
#ifdef TRAX
	case Csynthid:
	    shout("Synth sel trx","","","Csynth");
	    break;
#endif /* TRAX */
	case Crecid:
	    sprintf(strg, "defer erase (%d)", deferErase);
	    shout("Record to mark",strg,"release lofi","Crec");
	    break;
	case Cquitid:
	    shout("Quit dspB","","","Cquit");
	    break;
	case Cprintid:
	    shout("Generate PS","Make Matlab Matrix","","Cprint");
	    break;
	case Lviewid:
	    shout("?","?","?","Lview");
	    break;
	case Lvvscid:
	    shout("?","?","?","Lvvsc");
	    break;
	case Lvtimlid:
	    shout("Mark to view","","","Lvtiml");
	    break;
	case Lvsndid:
	    shout("Drag to mark","View mark","Extend mark","Lvsnd");
	    break;
	case Lvsbarid:
	    shout("Drag thumb","","","Lvsbar");
	    break;
	case Sviewid:
	    shout("?","?","?","Sview");
	    break;
	case Svvscid:
	    shout("?","?","?","Svvsc");
	    break;
	case Svsndid:
	    if( TimeMode(wrcd) )
		shout("Drag to mark","View mark","Extend mark","Svsnd");
	    else
		{
		shout("Zoom in", "?", "Zoom out","Svsnd");
		SetCursor(&xCursor);
		curs = CROSS;
		}
	    break;
	case Svtimlid:
	    if( TimeMode(wrcd) )
		shout("?","?","?","Svtim");
	    else if( (*wrcd)->w.snd.vMode == (int)wfallVM )
		shout("?","?","?","Svtim");
	    break;
	case VscaleUpid:
	    shout("?","?","?","Vsu");
	    break;
	case VscaleDownid:
	    shout("?","?","?","Vsd");
	    break;
	default:
/*	    D_prints("Shout in ????\n");	*/
	    break;
	    }
	lastShout = id;
	}
    }

int TimeMode(wrcd)	/* return true for time-like short view mode */
    WINRCD **wrcd;
    {
    int rc = 0;
    int mode = (*wrcd)->w.snd.vMode;

    if( mode == (int)monoVM || mode == (int)leftVM || mode == (int)rightVM
       || mode == (int)oneVM || mode == (int)twoVM || mode == (int)threeVM
       || mode == (int)fourVM || mode == (int)multiVM)
	rc = 1;
    return(rc);
    }

SelToView(wrcd)
    WINRCD  **wrcd;
    {
    long stt,end;

    stt = ((*wrcd)->w.snd.slct)->bStart;
    end = ((*wrcd)->w.snd.slct)->bEnd;
    if( !(stt<end) )
	{
	stt = 0L;
	end = ((*wrcd)->w.snd.snd)->dataBsize;
	}
    ((*wrcd)->w.snd.view)->bStart = stt;
    ((*wrcd)->w.snd.view)->bEnd   = end;
    ((*wrcd)->w.snd.frqV)->valid = 0;
    RedrawWinBit(wrcd,Lvsbarid);
    RedrawWinBit(wrcd,Sviewid);
    RedrawWinBit(wrcd,Infboxid);
    }

ChangeVscale(wrcd,sense)
    WINRCD **wrcd;
    int	   sense;	/* 1 => scale up one click, else down */
    {
    int  scale = (*wrcd)->w.snd.scale;
    int  pwr   = -1;

    if(sense > 0) scale = (scale*2) + ((scale == 1)?1:0); 

    if(scale < 2) scale = 2;

    while(scale > 3)
	{
	++pwr;
	scale /= 2;
	}
    if(scale == 3) scale = 4;
    else if(scale == 2) scale = 3;
    if(pwr>0) 		scale <<= pwr;
    else if(pwr<0)	scale >>= -pwr;
/*    D_prints("New scale:");  D_printn(scale);	*/
    (*wrcd)->w.snd.scale = scale;
/*    RedrawWinBit(wrcd,Lviewid);  make Lview immune to scaling */
    RedrawWinBit(wrcd,Sviewid);
    }

LVSBClik(wrcd, rec, where)
    WINRCD  **wrcd;	/* basic context */
    Rect    rec;	/* outer rec of scroll bar */
    Point   where;	/* where clicked */
    {
    double   f1,f2,dur;

    VTimeLims(wrcd,&f1,&f2,&dur);	    /* get real times for V-regn */
    f1 /= dur; f2 /= dur;
    SBarClik(rec, where, &f1, &f2);	    /* f1 & f2 modified per click */
    f1 *= dur; f2 *= dur;
    ((*wrcd)->w.snd.frqV)->valid = 0;
    SetVTimeLims(wrcd,f1,f2,dur);
    RedrawWinBit(wrcd,Lvsbarid);
    RedrawWinBit(wrcd,Sviewid);
    RedrawWinBit(wrcd,Infboxid);
    }

WFSBClik(wrcd, rec, where)	/* click in the waterfall scrollbar */
    WINRCD  **wrcd;	/* basic context */
    Rect    rec;	/* outer rec of scroll bar */
    Point   where;	/* where clicked */
    {
    double   f1,f2,dur;

    f1 = ((double)(*wrcd)->w.snd.wflen) / WFLENMAX;
    f2 = f1+.01;
    SBarClik(rec, where, &f1, &f2);
    (*wrcd)->w.snd.wflen = f1 * WFLENMAX;
    RedrawWinBit(wrcd,Sviewid);
    }

SBarClik(rec,where,pf1,pf2)	/* low-level : no knowledge of what it's for */
    Rect    rec;	/* outer rec of scroll bar */
    Point   where;	/* where clicked */
    double   *pf1, *pf2; /* thumb limits, scaled 0..1, modified */
    {
    short   mx, lft, rgt, len, thml, thmr;
    double   thumwidth;

    thumwidth = *pf2 - *pf1;
    mx = where.h;
    lft = rec.left + SBarWidth+1;
    rgt = rec.right - SBarWidth;
    len = rgt - lft;
    thml = lft + *pf1 * (double)len;
    thmr = lft + *pf2 * (double)len;
    if(thmr < thml + SBarWidth) thmr = thml + SBarWidth;

    if(mx < lft)	/* left arrow icon */	*pf1 -= thumwidth / 10;
    else if(mx < thml)	/* left grey bit  */	*pf1 -= 9 * thumwidth / 10;
    else if(mx < thmr)	/* in thumb	  */	TrakThum(rec,where,pf1,pf2);
    else if(mx < rgt )	/* right grey bit  */	*pf1 += 9 * thumwidth / 10;
    else		/* right button   */	*pf1 += thumwidth / 10;

    if (*pf1 < 0.0 ) *pf1 = 0.0;
    if (*pf1 > 1.0 - thumwidth) *pf1 = 1.0 - thumwidth;
    *pf2 = *pf1 + thumwidth;
    }

TrakThum(rec,where,pf1,pf2)
    Rect    rec;	/* outer rec of scroll bar */
    Point   where;	/* where clicked */
    double   *pf1, *pf2; /* thumb limits, scaled 0..1, modified */
    {
    short   lft, rgt, len, thml, thmr, thmw;
    Rect    thumRec;
    PenState	oldPenState;
    Point   mousePt,oldMsPt;

    lft = rec.left + SBarWidth+1;
    rgt = rec.right - SBarWidth;
    len = rgt - lft;
    thml = lft + *pf1 * (double)len;
    thmr = lft + *pf2 * (double)len;
    thmw = thmr - thml;
    rec.left += SBarWidth + (where.h-thml) + 1;
    rec.right -= SBarWidth /* + 1 */ ;
    rec.right -= thmw;
    rec.right += (where.h-thml);
    if(thmw < SBarWidth)    thmr = thml + SBarWidth;
    thumRec = rec;  /* InsetRect(&thumRec, 1, 1); */
    thumRec.top += 1;
    thumRec.left = thml;    thumRec.right = thmr;

    SetDottedXorPen(&oldPenState);
    mousePt = where;
    if( StillDown() )
	{
	FrameRect(&thumRec);	    /* first instance of dotbox */
	while( WaitMouseUp() )
	    {
	    oldMsPt = mousePt;
	    GetMouse( &mousePt );
	    if(mousePt.h < rec.left) mousePt.h = rec.left;
	    if(mousePt.h > rec.right)	mousePt.h = rec.right;
	    if(mousePt.h != oldMsPt.h)	    /* lateral movment */
		{
		FrameRect(&thumRec);
		OffsetRect(&thumRec, mousePt.h-oldMsPt.h, 0);
		FrameRect(&thumRec);
		}
	    }
	FrameRect(&thumRec);	    /* finally remove dotbox */
	}
    RestoreSolidCopyPen(&oldPenState);
    *pf2 -= *pf1;
    *pf1 = (0.1 + (double)thumRec.left - lft)/((double)len);
    *pf2 += *pf1;
    }

void ClikOverSpectra(wrcd, pwhere, prec, btn, vmode)
    /* click in spectrum drags out a marked area */
    WINRCD  **wrcd;	/* basic context */
    Rect    *prec;	/* outer rec of view */
    Point   *pwhere;	/* where clicked */
    int	    btn;	/* which button */
    int     vmode;	/* which view mode */
    {
    Rect r2;
    Rect orect;

    orect = *prec;
    InsetRect(&orect, SBarWidth, 0);
    DragRects(&orect, &r2, pwhere, 
	      1, 3, 0 /* repNexT, axes, visType */); 
    if(r2.left == r2.right)
	{
	(*wrcd)->w.snd.frqV->minFrq = 0.0;
	(*wrcd)->w.snd.frqV->maxFrq = (*wrcd)->w.snd.frqV->samplingRate/2.0;
	(*wrcd)->w.snd.frqV->minMag = -30.0;
	(*wrcd)->w.snd.frqV->maxMag = 70.0;
	}
    else if(btn==1)	/* zoom in */
	{	
	double frqScale = ((*wrcd)->w.snd.frqV->maxFrq 
	                       - (*wrcd)->w.snd.frqV->minFrq)
	                 /(orect.right - orect.left);
	double minF = (*wrcd)->w.snd.frqV->minFrq;
	double magScale = ((*wrcd)->w.snd.frqV->minMag 
	                       - (*wrcd)->w.snd.frqV->maxMag)
	                 /(orect.bottom - orect.top);
	double minM = (*wrcd)->w.snd.frqV->maxMag;

	(*wrcd)->w.snd.frqV->minFrq = minF + frqScale * (r2.left - orect.left);
	(*wrcd)->w.snd.frqV->maxFrq = minF + frqScale * (r2.right- orect.left);
	(*wrcd)->w.snd.frqV->minMag = minM + magScale * (r2.bottom -orect.top);
	(*wrcd)->w.snd.frqV->maxMag = minM + magScale * (r2.top - orect.top);
	}
    else{  /* other button is inverse zoom out */
	double minFrq, maxFrq, minMag, maxMag;
	double pL, pU, fL, fU, mL, mU, pT, pB;
	/* zoom out makes *current frq range* occupy *range under zoom* */

	pL = ((double)r2.left - orect.left)/((double)(orect.right-orect.left));
	pU = ((double)r2.right - orect.left)/((double)(orect.right-orect.left));
	fL = (*wrcd)->w.snd.frqV->minFrq;
	fU = (*wrcd)->w.snd.frqV->maxFrq;
	minFrq = (pU*fL - pL*fU)/(pU-pL);	/* did algebra on paper */
	maxFrq = minFrq + (fU-minFrq)/pU;	/* .. trust me! */
	if(minFrq < 0.0)	minFrq = 0.0;
	if(maxFrq > (*wrcd)->w.snd.frqV->samplingRate/2.0)
	    maxFrq = (*wrcd)->w.snd.frqV->samplingRate/2.0;
	(*wrcd)->w.snd.frqV->minFrq = minFrq;
	(*wrcd)->w.snd.frqV->maxFrq = maxFrq;

	pT = ((double)r2.top - orect.top)/((double)(orect.bottom-orect.top));
	pB = ((double)r2.bottom - orect.top)/((double)(orect.bottom-orect.top));
	mL = (*wrcd)->w.snd.frqV->maxMag;
	mU = (*wrcd)->w.snd.frqV->minMag;
	maxMag = (pB*mL - pT*mU)/(pB-pT);
	minMag = maxMag + (mU-maxMag)/pB;
	if(minMag < -100.0)	minMag = -100.0;
	if(maxMag > 100.0)	maxMag = 100.0;
	(*wrcd)->w.snd.frqV->minMag = minMag;
	(*wrcd)->w.snd.frqV->maxMag = maxMag;
	}
    RedrawWinBit(wrcd,Sviewid);
    return;
    }

void ClikOverSpctra(wrcd, pwhere, prec, btn)
    /* click in spectrum drags out a marked area */
    WINRCD  **wrcd;	/* basic context */
    Point   *pwhere;	/* where clicked */
    Rect    *prec;	/* outer rec of view */
    int	    btn;	/* which button */
    {
static Rect r2;
static oldRec = 0;
    Rect orect;

    orect = *prec;
    InsetRect(&orect, SBarWidth, 0);
    DragRects(&orect, &r2, pwhere, 
	      1-oldRec, 1, 1 /* repNexT, axes, visType */); 
/*    oldRec = 1;	*/ 	/* no: new rect every time */
    if(r2.left == r2.right)
	{
	(*wrcd)->w.snd.frqV->minFrq = 0.0;
	(*wrcd)->w.snd.frqV->maxFrq = (*wrcd)->w.snd.frqV->samplingRate/2.0;
	}
    else if(btn==1)	/* zoom in */
	{	
	double frqScale = ((*wrcd)->w.snd.frqV->maxFrq 
	                       - (*wrcd)->w.snd.frqV->minFrq)
	                 /(orect.right - orect.left);
	double minF = (*wrcd)->w.snd.frqV->minFrq;

	(*wrcd)->w.snd.frqV->minFrq = minF + frqScale * (r2.left - orect.left);
	(*wrcd)->w.snd.frqV->maxFrq = minF + frqScale * (r2.right- orect.left);
	}
    else{  /* other button is inverse zoom out */
	double minFrq, maxFrq, frqRg;
	double pL, pU, fL, fU;
	/* zoom out makes *current frq range* occupy *range under zoom* */

	pL = ((double)r2.left - orect.left)/((double)(orect.right-orect.left));
	pU = ((double)r2.right - orect.left)/((double)(orect.right-orect.left));
	fL = (*wrcd)->w.snd.frqV->minFrq;
	fU = (*wrcd)->w.snd.frqV->maxFrq;
	minFrq = (pU*fL - pL*fU)/(pU-pL);	/* did algebra on paper */
	maxFrq = minFrq + (fU-minFrq)/pU;	/* .. trust me! */
	if(minFrq < 0.0)	minFrq = 0.0;
	if(maxFrq > (*wrcd)->w.snd.frqV->samplingRate/2.0)
	    maxFrq = (*wrcd)->w.snd.frqV->samplingRate/2.0;
	(*wrcd)->w.snd.frqV->minFrq = minFrq;
	(*wrcd)->w.snd.frqV->maxFrq = maxFrq;
	}
    RedrawWinBit(wrcd,Sviewid);
    return;
    }

#ifndef MAX
#define MAX(a,b)	(a>b)?(a):(b)
#define MIN(a,b)	(a>b)?(b):(a)
#endif /* ndef MAX */

PtsToRect(pstart, pend, prec)
    Point	*pstart;
    Point 	*pend;
    Rect	*prec;
    {
    prec->top = MIN(pstart->v, pend->v);
    prec->bottom = MAX(pstart->v, pend->v);
    prec->left = MIN(pstart->h, pend->h);
    prec->right  = MAX(pstart->h, pend->h);
    }

int Clip1D(lo, up, px)		/* force a number into a range */
    int 	lo, up;		/* lower and upper bounds */
    int 	*px;		/* number to clip */
    {
    int 	clipd = 0;	/* return if modified number */
    if(lo > up)			/* args in 'wrong' order - swap */
	{  int t = lo;  lo = up;  up = t;  }
    if(*px < lo)	{  *px = lo;  clipd = 1;  }
    else if(*px > up)	{  *px = up;  clipd = 1;  }
    return(clipd);
    }

int Clip2D(limRec, ppt)		/* force a point into a rect */
    Rect	*limRec;
    Point	*ppt;		/* input and output */
    {
    int 	h = ppt->h;
    int 	v = ppt->v;
    int 	rslt = 0;

    rslt = Clip1D(limRec->left, limRec->right, &h);
    rslt = rslt | Clip1D(limRec->top, limRec->bottom, &v);
    /* LOGICAL 'or' prevented second term being executed if first was true! */
    ppt->h = h;		ppt->v = v;
    return rslt;    /* return 1 if we moved the point */
    }

int myAbs(i)
    int 	i;
    {
    return( (i>0)?i:-i );
    }

ChooseCorner(prec, ppt, rslt, nearNfar)	/* return corner of rect nrst pt */
    Rect	*prec;		/* rect in question */
    Point	*ppt;		/* point to start from */
    Point	*rslt;		/* returned point */
    int 	nearNfar;	/* 1 = return nrst; 0 = furthest */
    {
    Point 	nearPt, farPt;

    if(myAbs(ppt->h - prec->left) > myAbs(ppt->h - prec->right))
	{  nearPt.h = prec->right;  farPt.h = prec->left;  }
    else
	{  nearPt.h = prec->left;  farPt.h = prec->right;  }
    if(myAbs(ppt->v - prec->top) > myAbs(ppt->v - prec->bottom))
	{  nearPt.v = prec->bottom;  farPt.v = prec->top;  }
    else
	{  nearPt.v = prec->top;  farPt.v = prec->bottom;  }
    if(nearNfar)	/* decide which point to return */
	*rslt = nearPt;
    else
	*rslt = farPt;
    }

PaintRecFrPts(ppt1, ppt2)	/* paint a rectangle between 2 pts */
    Point *ppt1;
    Point *ppt2;
    {
    Rect	r;

    PtsToRect(ppt1, ppt2, &r);
    PaintRect(&r);
    }

EmphRect(prec, type)
    Rect	*prec;
    int 	type;	/* 1 = invert, 0 = outline */
    {
    if(type == 0)
	FrameRect(prec);
    else
	PaintRect(prec);
    }

DragRects(limRec, selRec, startPt, repNext, axes, visType)
    Rect	*limRec;	/* limits to mouse excursion */
    Rect	*selRec;	/* pass in previous limits, return result */
    Point	*startPt;	/* initial mouse point */
    int 	repNext;	/* 1 = new region, 0 = extend old */
    int 	axes;		/* 1 = L/R, 2 = U/D, 3 = both */
    int 	visType;	/* 1 = invert whole, 0 = outline rect */
    {
    PenState	oldPenState;
    Point	srcPt, dstPt, frmPt, newPt;
    int 	moved;

    SetDottedXorPen(&oldPenState);
    dstPt = *startPt;
    Clip2D(limRec, &dstPt);	/* make sure initial extension is in limit */

    srcPt = dstPt;
    if(repNext == 0)	/* extending region passed in, so find opps crnr */
	ChooseCorner(selRec, &dstPt, &srcPt, 0);

    /* setup fixed L/R or T/B if only varying one axis */
    if( (axes & 1) == 0)	/* left/right from limit rectangle */
	{  srcPt.h = limRec->left;  dstPt.h = limRec->right;	}
    if( (axes & 2) == 0)	/* up/down from limit rectangle */
	{  srcPt.v = limRec->top;   dstPt.v = limRec->bottom;   }
    PtsToRect(&srcPt, &dstPt, selRec);

    if( StillDown() )
	{
	EmphRect(selRec, visType);	/* frame or paint */
	while( WaitMouseUp() )
	    {
	    moved = 0;
	    GetMouse( &newPt );	    Clip2D(limRec, &newPt);
	    frmPt = srcPt;
	    if((axes & 1)  	/* allow L/R movement */
	       && (newPt.h != dstPt.h))		/* and was movement */
		{
		frmPt.h = dstPt.h;	dstPt.h = newPt.h;
		if(visType)	/* if painting, just paint the extra */
		    PaintRecFrPts(&frmPt, &dstPt);
		moved = 1;
		}
	    frmPt = srcPt;	/* in case we have a second dimension */
	    if((axes & 2)	/* allow U/D movement */
	       && (newPt.v != dstPt.v))		/* and was movement */
		{
		frmPt.v = dstPt.v;	dstPt.v = newPt.v;
		if(visType)	/* paint extra if painting */
		    PaintRecFrPts(&frmPt, &dstPt);
		moved = 1;
		}
	    if(moved)
		{	/* undraw, redraw frame */
		if(visType == 0)	FrameRect(selRec);
		PtsToRect(&srcPt, &dstPt, selRec);	/* new selection */
		if(visType == 0)	FrameRect(selRec);
		}
	    }
	EmphRect(selRec, visType);	/* remove last version (xor) */
	}
    RestoreSolidCopyPen(&oldPenState);
    /* return is via selRec */
    }

void NewSelTimes(WINRCD **wrcd, Rect *prec, FLOATARG lt1, FLOATARG lt2, 
	    int baseh, FLOATARG ot1, Rect *pselRec, int btn, FLOATARG dur)
{	/* calculate new selection limits during drag */
    double t1, t2;

    t1 = MapXtoT(pselRec->left, *prec, lt1, lt2);
    t2 = MapXtoT(pselRec->right, *prec, lt1, lt2);
    if(btn != 1)	/* only reset start if left button drag */
    {
	if(pselRec->left == baseh)
	    t1 = ot1;
	else
	    t2 = ot1;
    }
    
    SetSTimeLims(wrcd, t1, t2, dur);
}

UnmarkSelects(WINRCD **wrcd)
{	/* remove any selection highlighting */
    Rect    svrec,lvrec;
    RECRCD  **rh;

    rh = FindRhan( (*wrcd)->w.snd.recTreeH ,(int)Svsndid);
    svrec = (*rh)->rec;
    rh = FindRhan( (*wrcd)->w.snd.recTreeH ,(int)Lvsndid);
    lvrec = (*rh)->rec;
    if( TimeMode(wrcd) )
	UnmarkSelInShort(wrcd,svrec);
    UnmarkSelect(wrcd,lvrec);	/* clear existing marks */
    FlushDisplay();
}

MarkSelects(WINRCD **wrcd)
{	/* perfomr any selection highlighting */
    Rect    svrec,lvrec;
    RECRCD  **rh;

    rh = FindRhan( (*wrcd)->w.snd.recTreeH ,(int)Svsndid);
    svrec = (*rh)->rec;
    rh = FindRhan( (*wrcd)->w.snd.recTreeH ,(int)Lvsndid);
    lvrec = (*rh)->rec;
    if( TimeMode(wrcd) )
	MarkSelInShort(wrcd,svrec);
    MarkSelect(wrcd,lvrec);	/* set the appropriate marks */
    FlushDisplay();
}

LVSndClik(wrcd, rec, lt1, lt2, dur, where, btn)	
    /* click in sound drags out a marked area */
    WINRCD  **wrcd;	/* basic context */
    Rect    rec;	/* outer rec of view */
    FLOATARG   lt1,lt2,dur;    /* times referred by ends of rect */
    Point   where;	/* where clicked */
    int	    btn;	/* which button */
    {
    Rect    selRec,r2,svrec,lvrec;
    Point   mousePt, oldMsPt;
    PenState	oldPenState;
    long    baseh,oh1,oh2;
    RECRCD  **rh;
    double   ot1,t1,t2,ddur,MapXtoT();
    double  fabs();

    rh = FindRhan( (*wrcd)->w.snd.recTreeH ,(int)Svsndid);
    svrec = (*rh)->rec;
    rh = FindRhan( (*wrcd)->w.snd.recTreeH ,(int)Lvsndid);
    lvrec = (*rh)->rec;
/*    if( ((*wrcd)->w.snd.slct)->bStart != ((*wrcd)->w.snd.slct)->bEnd ) */
	{
	if( TimeMode(wrcd) )
	    UnmarkSelInShort(wrcd,svrec);
	UnmarkSelect(wrcd,lvrec);	/* clear existing marks */
	FlushDisplay();
	}

    InsetRect(&rec,SBarWidth+1,0);
    SetDottedXorPen(&oldPenState);
    mousePt = where;
    if(mousePt.h < rec.left)	mousePt.h = rec.left;
    if(mousePt.h > rec.right)	mousePt.h = rec.right;
    baseh   = mousePt.h;
    if(btn != 1)	/* not L button -> extend current mark */
	{
	STimeLims(wrcd,&t1,&t2,&ddur);
	ot1 = MapXtoT(baseh, rec, lt1, lt2);
	if( fabs(t1-ot1) < fabs(t2-ot1) /* t1 > ot1 */ )
	    { baseh = MapTtoX( (t2>lt2)?lt2:t2 
				  , rec, lt1, lt2); ot1 = t2; }
	else
	    { baseh = MapTtoX( (t1<lt1)?lt1:t1
			      , rec, lt1, lt2); ot1 = t1; }
	}
    if(mousePt.h < rec.left)	mousePt.h = rec.left;
    if(mousePt.h > rec.right)	mousePt.h = rec.right;
    /* baseh should already be clipped */
/*    if(baseh < rec.left)	baseh = rec.left;
    if(baseh > rec.right)	baseh = rec.right;
 */
    selRec = rec;
    selRec.left = baseh;
    selRec.right = mousePt.h;
    if(selRec.left > selRec.right)
	{
	selRec.right = baseh;   selRec.left = mousePt.h;
	}
    r2 = selRec;
    if( StillDown() )
	{
	PaintRect(&selRec);	
	while( WaitMouseUp() )
	    {
	    oldMsPt = mousePt;
	    GetMouse( &mousePt );
	    if(mousePt.h < rec.left)	mousePt.h = rec.left;
	    if(mousePt.h > rec.right)	mousePt.h = rec.right;
	    if(mousePt.h != oldMsPt.h)	    /* lateral movment */
		{
		selRec.left = baseh;	selRec.right = mousePt.h;
		if(selRec.left > selRec.right)
		    {
		    selRec.right = baseh;   selRec.left = mousePt.h;
		    }
		r2.left = mousePt.h;	r2.right = oldMsPt.h;
		if(r2.left > r2.right)
		    {
		    r2.right = mousePt.h;   r2.left = oldMsPt.h;
		    }
		PaintRect(&r2);

		NewSelTimes(wrcd, &rec, lt1, lt2, baseh, ot1, &selRec, 
			    btn, dur);
		SelInfo(wrcd, NULL);
		}
	    }
	PaintRect(&selRec);	
	}
    RestoreSolidCopyPen(&oldPenState);
    NewSelTimes(wrcd, &rec, lt1, lt2, baseh, ot1, &selRec, btn, dur);
    SelInfo(wrcd, NULL);
    MarkSelect(wrcd,lvrec);
    ((*wrcd)->w.snd.frqV)->valid = 0;
    if( TimeMode(wrcd) )
	MarkSelInShort(wrcd,svrec);
    RedrawWinBit(wrcd,Infboxid);
    }

UnmarkSelect(wrcd,rec)
    WINRCD  **wrcd;
    Rect    rec;
    {
    double   t1,t2,dur;
    int     len,l,r;
    PenState	oldPenState;

    SetDottedXorPen(&oldPenState);
    InsetRect(&rec,SBarWidth+1,0);
    len = rec.right - rec.left;
    STimeLims(wrcd,&t1,&t2,&dur);
    l = MapTtoX(t1, rec, (double)0, dur);
    r = MapTtoX(t2, rec, (double)0, dur);
    if(r==l)			/* at least one pixel wide */
	++r;
    if(r==rec.right)
	--r;			/* don't invert right frame */
    rec.left = l; rec.right = r;
    PaintRect(&rec);	
    RestoreSolidCopyPen(&oldPenState);
    }

MarkSelect(wrcd,rec)
    WINRCD  **wrcd;
    Rect    rec;
    {
    UnmarkSelect(wrcd,rec);
    }

UnmarkSelInShort(wrcd,rec)
    WINRCD  **wrcd;
    Rect    rec;
    {
    MarkSelInShort(wrcd,rec);
    }

MarkSelInShort(wrcd,rec)
    WINRCD  **wrcd;
    Rect    rec;
    {
    double   vt1,vt2,dur,st1,st2;
    short   l,r;
    PenState	oldPenState;

    STimeLims(wrcd, &st1, &st2, &dur);
    VTimeLims(wrcd, &vt1, &vt2, &dur);
    if(st1<vt2 && st2>vt1)
	{
	if(st1 < vt1) st1 = vt1;
	if(st2 > vt2) st2 = vt2;
	InsetRect(&rec,SBarWidth+1,0);
	l = MapTtoX(st1, rec, vt1, vt2);
	r = MapTtoX(st2, rec, vt1, vt2);
	if(r==l)	/* at least one pixel wide */
	    ++r;
	if(r==rec.right)
	    --r;	/* don't invert right frame */
	rec.left = l; rec.right = r;
	SetDottedXorPen(&oldPenState);
	PaintRect(&rec);	
	RestoreSolidCopyPen(&oldPenState);
	}
    }

int GetWholeSndPtrs(snd, pstt, pend)
    SFSTRUCT	*snd;
    void	**pstt;
    void	**pend;
    {
    *pstt = SFDataLoc(snd);
    *pend = ((char *)*pstt) + snd->dataBsize;

    return snd->format;
    }

ProcessSelection(wrcd,selproc)		/* generic do-something-to-slct */
    WINRCD	**wrcd;
    void	(*selproc)();
    {
    char   		*stt, *end;
    int			dsize;
    int     		scale = (*wrcd)->w.snd.scale;
    SFSTRUCT		*snd  = (*wrcd)->w.snd.snd;
    double		srate;
    int			chans;

    srate = snd->samplingRate;
    chans = snd->channels;
    dsize = SFSampleSizeOf(snd->format);
    GetSndBitePtrs( (*wrcd)->w.snd.slct, &stt, &end );
    if(end <= stt)		/* i.e. if no selection made */
	GetWholeSndPtrs(snd, (void **)&stt, (void **)&end);
    if(end > stt)		/* really is some sound */
	{
	SetCursor( &waitCursor );
	(*selproc)(stt,end,chans,dsize,srate,scale);
	SetCursor( &arrow );
	}
    }

#include <audIO.h>
static AUSTRUCT *auport = NULL;

#ifdef LOFI		/* no play_with_stop on macintosh or sun */
static int lofiGrabbed = 0;
static short *scratchBuffer = NULL;
#ifdef __sgi
#define PLAYBSZ 8192
#else /* !__sgi, maybe DEC */
#define PLAYBSZ	1024
#endif /* __sgi */
#define MAXCHANS 2

void play_with_stop(stt,end,chans,dsize,srate,scale)
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;	/* crude gain - ignored here */
    {
    short 	*src;
    int		chunkSamps;
    int		bestSize;
    int 	done = 0;

    if(scratchBuffer == NULL)
	{
	if( (scratchBuffer = (short *)Mymalloc((long)PLAYBSZ*MAXCHANS, 
	     sizeof(short), "play_ws"))
	   ==NULL)
	    { fprintf(stderr, "play_ws: malloc fail\n"); abort(); }
	}

    if(lofiGrabbed == 0)  {
	/* lofiGrabbed = play_set(chans,dsize,srate,scale); */
	auport = AUOpen(SFMT_SHORT, srate, chans, PLAYBSZ, AIO_OUTPUT);
	lofiGrabbed = (auport != NULL);
    }
    if(lofiGrabbed == 0)
	return;	/* lose but don't crash */
    src = stt;
    bestSize = 1024*chans; /* hblck/2 is frames per blk */
    while(src < end && !done)
	/* samples still to send */
	{
	chunkSamps = end-src;
	if(chunkSamps>bestSize) chunkSamps=bestSize;
	/* play_on(src,chunkSamps/chans); */
	AUWrite(auport, src, chunkSamps/chans);
	src += (chunkSamps);	/* cs/2 is num of frames */
	done= ScanMouseEvent();
	if(done == 2)	/* middle button for pause */
	    {
	    SetCursor( &pauseCursor );
	    if(WaitMouseEvent()==2) /* release pause */
		done = 0;	/* new lease of life */
	    SetCursor( &waitCursor );
	    }
	}
    if(chunkSamps < bestSize)
	{
	chunkSamps = bestSize - chunkSamps;
	ClearShorts(scratchBuffer, 1, chunkSamps);
	/* play_on(scratchBuffer,chunkSamps/chans);  */
	AUWrite(auport, scratchBuffer, chunkSamps/chans);
	}
/*    play_rls();		/* release for parallel code */
    /* AUClose(auport);	lofiGrabbed = 0; */
    }

void ShortRvrs(short *src, long sinc, short *dst, long dinc, long len)
{   /* copy src to dst, reversing the order */
    long  i = len/2;
    short *sl = src;
    short *su = src+sinc*(len-1);
    short *dl = dst;
    short *du = dst+dinc*(len-1);
    short sh;

    while(i--)	{
	sh  = *su;
	*du = *sl;
	*dl = sh;
	su -= sinc;
	du -= dinc;
	sl += sinc;
	dl += dinc;
    }
    if(src != dst)  {	/* better not overlap - do other half */
	i = len - (len/2);
	while(i--)	{
	    sh  = *su;
	    *du = *sl;
	    *dl = sh;
	    su -= sinc;
	    du -= dinc;
	    sl += sinc;
	    dl += dinc;
	}
    }
}


void play_rvs_with_stop(stt,end,chans,dsize,srate,scale)
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;	/* crude gain - ignored here */
    {
    short 	*src;
    int		chunkSamps;
    int		bestSize;
    int 	done = 0;
    int 	ch;

    if(scratchBuffer == NULL)
	{
	if( (scratchBuffer = (short *)Mymalloc((long)PLAYBSZ*MAXCHANS, 
	     sizeof(short), "play_ws"))
	   ==NULL)
	    { fprintf(stderr, "play_ws: malloc fail\n"); abort(); }
	ClearShorts(scratchBuffer, 1, PLAYBSZ*MAXCHANS);
	}

    if(lofiGrabbed == 0)  {
	/* lofiGrabbed = play_set(chans,dsize,srate,scale); */
	auport = AUOpen(SFMT_SHORT, srate, chans, PLAYBSZ, AIO_OUTPUT);
	lofiGrabbed = (auport != NULL);
    }
    if(lofiGrabbed == 0)
	return;	/* lose but don't crash */
    src = end;
    bestSize = 1024*chans; /* hblck/2 is frames per blk */
    while(src > stt && !done)
	/* samples still to send */
	{
	chunkSamps = src-stt;
	if(chunkSamps>bestSize) chunkSamps=bestSize;
	for(ch = 0; ch < chans; ++ch)  {
	    ShortRvrs(src - chunkSamps + ch, chans, 
		      scratchBuffer + ch, chans, chunkSamps/chans);
	}
	/* play_on(scratchBuffer, chunkSamps/chans); */
	AUWrite(auport, scratchBuffer, chunkSamps/chans);
	src -= (chunkSamps);	/* cs/2 is num of frames */
	done= ScanMouseEvent();
	if(done == 2)	/* middle button for pause */
	    {
	    SetCursor( &pauseCursor );
	    if(WaitMouseEvent()==2) /* release pause */
		done = 0;	/* new lease of life */
	    SetCursor( &waitCursor );
	    }
	}
    if(chunkSamps < bestSize)
	{
	chunkSamps = bestSize - chunkSamps;
	ClearShorts(scratchBuffer, 1, chunkSamps);
	/* play_on(scratchBuffer,chunkSamps/chans); */
	AUWrite(auport, scratchBuffer, chunkSamps/chans);
	}
/*    play_rls();		/* release for parallel code */
    }

void loop_with_stop(stt,end,chans,dsize,srate,scale)
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;	/* crude gain - ignored here */
    {
    short 	*src;
    int		chunkSamps;
    int		bestSize;
    int 	done = 0;

    if(scratchBuffer == NULL)
	{
	if( (scratchBuffer = (short *)Mymalloc((long)PLAYBSZ*MAXCHANS, 
								sizeof(short), "loop_ws"))
	   ==NULL)
	    { fprintf(stderr, "loop_ws: malloc fail\n"); abort(); }
	ClearShorts(scratchBuffer, 1, PLAYBSZ*MAXCHANS);
	}
    if(lofiGrabbed == 0)  {
	/* lofiGrabbed = play_set(chans,dsize,srate,scale); */
	auport = AUOpen(SFMT_SHORT, srate, chans, PLAYBSZ, AIO_OUTPUT);
	lofiGrabbed = (auport != NULL);
    }
    if(lofiGrabbed == 0)
	return;	/* lose but don't crash */

    while(!done)
	{
	src = stt;
	bestSize = 1024*chans;	/* hblck/2 is frames per blk */
	while(src < end && !done)	    /* samples still to send */
	    {
	    chunkSamps = end-src;
	    if(chunkSamps>bestSize) chunkSamps=bestSize;
	    /* play_on(src,chunkSamps/chans); */
	    AUWrite(auport, src, chunkSamps/chans);
	    src += (chunkSamps); /* cs/2 is num of frames */
	    done = ScanMouseEvent();
	    if(done == 2)	/* middle button for pause */
		{
		SetCursor( &pauseCursor );
		if(WaitMouseEvent()==2) /* release pause */
		    done = 0;	/* new lease of life */
		SetCursor( &waitCursor );
		}
	    }
/* this code appears to rewrite some of the end of the sample 
   so that the total amount written is an exact multiple of 
   'best size'.  But that's a lose..   maybe important for LOFI */
#ifdef DEC
	if(chunkSamps < bestSize)
	    {
	    chunkSamps = bestSize - chunkSamps;
	    /* play_on(scratchBuffer,chunkSamps/chans); */
	    AUWrite(auport, scratchBuffer, chunkSamps/chans);
	    }
#endif /* DEC */
	}
/*    play_rls();		/* release for parallel code */
    
    }
#else /* !LOFI : covers for *_with_stop on sun and mac */
void play_with_stop(stt,end,chans,dsize,srate,scale)
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;	/* crude gain - ignored here */
    {
	play(stt,end,chans,dsize,srate,scale);
    }
void loop_with_stop(stt,end,chans,dsize,srate,scale)
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    double   srate;
    int     scale;	/* crude gain - ignored here */
    {
	play(stt,end,chans,dsize,srate,scale);
    }
#endif /* !LOFI play special cases */

PlaySelSnd(wrcd, btn)	/* play (selected range) of this time graph */
    WINRCD **wrcd;
    int 	btn;	/* left plays; right loops */
    {
    if(btn==1)
	ProcessSelection(wrcd, play_with_stop);
    else if(btn==2)
	ProcessSelection(wrcd, play_rvs_with_stop);
    else if(btn==3)
	ProcessSelection(wrcd, loop_with_stop);
    }

record(short *stt, short *end, int chans, int dsize, double srate, int scale)
{
    AUSTRUCT *au;

    au = AUOpen(SFMT_SHORT, srate, chans, PLAYBSZ, AIO_INPUT);
    if(au)  {
	AURead(au, stt, (end-stt)/chans);
    }
}

RecordSelSnd(wrcd, btn)	/* Sample an.in (selected range) of this time graph */
    WINRCD **wrcd;
    int btn;
    {
    void record();

    if(btn==1)
	{
    	ProcessSelection(wrcd,record);	/*	*/
	}
    else if(btn==2)		/* middle button defers erases */
	{
	deferErase += 1+1;	/* 1 further exposes before erase .. HACK */
	}
#ifdef LOFI /* not available on macintosh or sun */
    else if(lofiGrabbed)
	{
	/* play_rls(); */
	AUClose(auport, 0);
	lofiGrabbed = 0;
	}
#endif /* LOFI */
    }

XtractSelSnd(wrcd)	/* save to tmp sel'd range of this time graph */
    WINRCD **wrcd;
    {
    void xtract();

    ProcessSelection(wrcd,xtract);
    }


