/*******************************************************************\
*   db.tracks.h							    *
*   Header file for dspBench functions that `do' tracks 	    *
*   dpwe 1993apr20						    *
\*******************************************************************/

#include <dpwelib.h>
#include <math.h>
extern double   rint(double);	/* not in math.h on DEC */
#include <assert.h>

#include "../trk/tkplay.h"	/* pulls in all track stuff */
#include "../trk/tkgroups.h"	/* ..except this */

#include <selwndw.h>		/* define little selection window thingie */

#ifdef DEBUGGRAPHS
#include "window.h"
#endif

/**** constants ****/
#define LOGTWO		0.69314718
#define ONEONLOGTWO	1.442695
#define NOMAX		(-99900.0)
#define NOMIN		(99900.0)		/* seedvalues for max/min */
#define BIGMIN		NOMAX
#define BIGMAX		NOMIN
#define MINUSALOT	(-100.0)		/* dB indicates 0 */

/* `modes' for colouring of overdrawn tracks */
enum { 
    TKM_NONE = 0,	/* `don't touch the color' */
    TKM_NORM,
    TKM_SELECT, 
    TKM_SPECIAL
};

/********* types ***********/
typedef struct  {	/* generic threshold structure */
    float	min;
    float	max;
} RANGE_STR;

typedef struct {	/* structure for arg to search nodefn */
    NODE *result;	/* where to store result (updated) */
    RANGE_STR	range;	/* the values to fit (updated on the fly) */
} LSRCH_STR;

/**** universal static data ****/
/* defined in db.peaks.c */
extern	TKLIST	allTrax;
extern	TKLIST	selTrax;
/* defined in db.exspec.c */
extern  TKLIST	theTails;		/* remember the tails we found */
extern 	double	tailsTime;		/* ..and the time they matched */

/**** important prototypes ****/
int TimeFreqComp(NODE **pnode1, NODE **pnode2);
	/* For insert_by_cmp.
	   Compare two TRACK nodes.  
	   Node 1 will be inserted before node2 if returns -1 */
double FindTrkMaxMag(TRACK *trk);	/* find largest mag in track's pks */
void RedrawTrax(TKLIST trax, int mode);	/* redraw placed trax in color */
int DrawTrack(TRACK *trk, int type, void *arg);	/* redraw one track */
PKLIST FindNonnegX(PKLIST pkl);		/* (def in peaks, used in exspec) */
	/* 'out-of-range' peaks are marked with their .x field < 0 */
TKLIST *AddRemoveSelTrax(TKLIST trax, TKLIST *srchFrom);
	/* add or remove all the selected trax from the selTrax list.
	   i.e. toggle their membership in selTrax.
	   srchFrom is a point in selTrax which is the earliest we 
	   need to look at (or NULL) */

/* from db.exspec.c .. */
extern PKLIST exFindPkBeforeT(PKLIST trk, FLOATARG time);
	/* find the node in the track imm. prec specified time */
TKLIST exTailsForTime(FLOATARG time, TKLIST allTrax);
	/* make a list of all the track tails crossing that time.
	   Output is a LIST of TRACK structs, which include pointers 
	   to PKLISTs which are the tails starting at or after the time. */
void CopyAndDisp(float *src, int sinc, long npts, WINDAT *win, char *str, 
		 int unwrap, FLOATARG gain);
     /* display some floats in a WINDAT window;  optionally unwrap their ph */
void DispComplex(cpxpolar *src, int sinc, long npts, 
		 WINDAT *mw, WINDAT *pw, char *str);
     /* display a complex vector in two WINDATS */

/* from db.growtx.c .. */
TRACK *tkFindMaxMag(TKLIST trx);
    /* return the car of the element of <trx> with the highest pk mag */
TKLIST tkFindWoodblocks(TKLIST trx);
void trxNewSelect(TKLIST trx);
    /* copy the tracklist as the new track selection */
void FindTrkStartEnd(TRACK *track, double *pstart, double *pend);
    /* return the time support of the track */
void FindTrkFrqRng(TRACK *track, double *pfmin, double *pfmax);
    /* return the `frequency support' of the track */
TKLIST SelTrxAbvMag(TKLIST trx, double mthresh);
	/* return a new list of tracks taken from <trx> whose peak
	   magnitude exceeds <mthresh> */
TKLIST SelTrxByMag(TKLIST trx, double mmin, double mmax);
	/* return a new list of tracks taken from <trx> whose peak
	   magnitude fall in <mmin> .. <mmax>.
	   <mmax> is useful to avoid grouping to *larger* following 
	   tracks that should group on their own. */
TKLIST SelTrxByAvMag(TKLIST trx, double mmin, double mmax);
	/* return a new list of tracks taken from <trx> whose *average*
	   magnitude fall in <mmin> .. <mmax>.
	   <mmax> is useful to avoid grouping to *larger* following 
	   tracks that should group on their own. */
TKLIST tkGetIntersect(TKLIST trx, double tMin, double tMax,
		       double fMin, double fMax);
	/* return tracks from the list with support overlapping 
	   the specified time-frequency rectangle */
TKLIST tkGetOnset(TKLIST trx, double tMin, double tMax);
	/* return track subset whose onsets lie in the given range */
TRACK *tkFindMaxMag(TKLIST trx);
	/* return the element in the track list that has highest peak mag */
int selTraxCB(void *data, int state, SELENTRY *se);
	/* SELPNL callback changes the track selection to TKLIST in <data> */
double tkSumDurns(TKLIST trx);
	/* Return the sum of the durations of all the tracks in the list */
TKLIST SelTrxByDur(TKLIST trx, double dmin, double dmax);
	/* filter <trx>, returning only those whose duration (tmax-tmin)
	   falls in the range specified */
int tk_onset_pred(LIST track, int type, void *arg);
	/* predicate to allow filtering of track list by duration overlap */

/* from db.noisy.c */
int noisyInitStatics(WINRCD **wrcd, int spnlq);
	/* initialize the static data used in this file (esp cqi).
	   Optionally init WNDWs and SELPNL too. */
double tkProminence(TRACK *trk, CQTINFO *cqi);
	/* calculate the 'prominence' (onset headroom) of passed 
	   track by looking at raw spectrum in vicinity */
TKLIST tklSelAbvProm(TKLIST src, double promthresh);
	/* return only those tracks exceeding the prominence threshold *
TKLIST txSelTFM(TKLIST src, double tmin, double tmax, 
	      double fmin, double fmax, 
	      double mmin, double mmax);
	/* does a t/f intersect, then applies magmax filter too */
TGLIST tklNoiseGroups(TKLIST tks, TKLIST focus);
	/* yr basic NOISE groups - from -track lists, but tries 
	   to up efficiency by working in increasing time, and 
	   NOT growing noise around tracks already part of noise 
	   regions returned so far */
TGLIST tgl_noise_tg(TKGROUP *tg, int *ptype, TKLIST focus);
	/* LNodeFunc to generate noise regions from a track *group* -
	   handles the parts all at once with tklNoiseG */
TGLIST tgl_noise_tk(TRACK *trk, int *ptype, TKLIST focus);
	/* LNodeFunc to generate noise regions from tracks. 
	   At present, at most one noise region per track. */


/* from db.harmon.c */
void rngStretch(RANGE_STR *rg, double val);
	/* extend the range structure to include <val> */
int rngTest(RANGE_STR *rg, double val);
	/* compare val with range;  -1 if val < rg->min, 1 if val > rg->max, 
	   0 if val in range */
int rngIntersect(RANGE_STR *one, RANGE_STR *two, RANGE_STR *rslt);
	/* test and/or calculate the intersection between two ranges */
int rngUnion(RANGE_STR *one, RANGE_STR *two, RANGE_STR *rslt);
	/* test and/or calculate the union of two ranges */
int tkFrqMagRange(TRACK *trk, RANGE_STR *trg, RANGE_STR *frg, RANGE_STR *mrg);
	/* Return frq and magnitude ranges spanned by the specified track 
	   *over the specified time range* */
int tkFrqMagForT(TRACK *trk, double time, double *pfrq, double *pmag);
	/* return the magnitude and frequency of a track at the 
	   specified time */
int tIntersect(double tl1, double tu1, double tl2, double tu2, 
	       double *prl, double *pru);
       /* return 1 if the two time intervals <tl1..tu1> and <tl2..tu2>
	  have an overlap - return it as <*prl..*pru>.  Else 0 */
double tksTestRatio(TRACK *tk1, TRACK *tk2, double *pratio, double *prsd);
	/* test the frequency ratio between two tracks, over common 
	   time extent.  Return common duration */
int tkCmpFAvg(NODE **pn1, NODE **pn2);
    /* qsort comparison function:
       Just sort into ascending frq order - ignore times. */
TKLIST tkFindHarmonics(TRACK *trk, double position, TKLIST focus);
	/* Find tracks that serve well as harmonics of <trk>, where 
	   the fundamental frequency is obtained as the frequency 
	   of <trk> divided by <position> - <position> typically close 
	   to an integer */
TGLIST tgHarmons(TRACK *trk, TKLIST focus);
	/* generate a set of harmonic track groups based on trk */
TGLIST tgl_harm_tk(TRACK *trk, int *ptype, TKLIST focus);
	/* LNodeFunc to generate harmonic sets from tracks */

/* some extra selpnl funcs here */
void tkgrSelect(TKGROUP *tkg, int merge);
	/* make the tracks in the <tkg> be the current selection.
	   If <merge> is nonzero, add them to the existing selection, 
	   otherwise replace it */
SELPNL_STR *spnlFromTkgList(SELPNL_STR *spnl, TGLIST tgl, WINRCD **wrcd);
	/* present a list of track groups as entries in a selpanel */

/* from db.onset.c */
double tkOnsetTFM(TRACK *trk, double *pfrq, double *pmag);
	/* calculate the onset time of a track based on its mag slope.
	   Also return the frequency and magnitude at that instant. */
double tkOnsetT(TRACK *trk);
    /* calculate the onset time of a track based on its mag slope */
TKLIST tkOnsetters(TRACK *trk, TKLIST focus, TKLIST withouts);
	/* find tracks above current with onset close to track passed.
	   Return as many as exceed some preset threshold.
	   DO NOT return any tracks from <withouts> */
TRACK *tkFindMaxFrq(TKLIST trx);
	/* return the element in the track list that has highest head frq */
TGLIST tgExtendOnset(TKGROUP *tgr, TKLIST focus, TKLIST *pwithouts);
	/* Grow into an onset group by chasing upwards
	   But *don't* extend by any of the tracks in <*pwithouts>, 
	   and extend <*pwithouts> with the tracks we return. */
int tg_test_trk(LIST data, int type, void *arg);
	/* NodeFunc tests for the presence of a given track in 
	   the `parts' of trkgroup */
int tg_print_fn(LIST data, int type, void *arg);
	/* NodeFunc prints elements of list of TKGROUPs briefly */
void tglPrint(TGLIST tgls, char *name);
	/* print brief summary of tg list */
LIST eps_close(LIST listin, LNodeFunc fn, void *arg);
	/* apply <fn> to all top-level elements of <listin> */
TRACK *tkFindMaxMagAtT(TKLIST trx, double time, double *pmag);
	/* return the element in the track list that has highest mag at T */
TRACK *tkFindMaxMagOverT(TKLIST trx, double tmin, double tmax, double *pmag);
	/* return the element in the track list that has highest peak mag
	   over the specified time range, and the value of that peak */
TKLIST tkGetOnsetIntersect(TKLIST trx, double tMin, double tMax,
			   double fMin, double fMax);
	/* return tracks from the list whose ONSET times 
	   (as read from st.tavg) and ONSET FREQUENCIES
	   (as obtained by tkFrqMagForT(onset)) lie strictly within 
	   the specified range */
TKLIST tkGetNarrowIntersect(TKLIST trx, double tMin, double tMax,
			   double fMin, double fMax);
	/* Perform slower and more restrictive search for tracks: 
	   for each, calculate the actual frequency span 
	   *over the time range specified only* and pass the 
	   track only if it overlaps with spec range */
double tkOnsScore(TRACK *tk, TKLIST focus);
	/* does this track look like an onset i.e. is it louder than 
	   its near predecessors?  Returns +ve dB prominence, zero if 
	   no near predecessors found. */
TKLIST tklOnsetsPrepare(TKLIST focus);
	/* prepare a set of tracks for onset score calculation 
	   by clearing the stolen `onset prominence' field.
	   THEN strip down to valid-onset candidates only, 
	   which are returned.  */
double tgOnsScore(TKGROUP *tg, TKLIST focus);
	/* evaluate the `quality' score for an onset group.
	   This is defined as the weighted average `prominence' in 
	   dB between each onset point and other tracks in 
	   a small region near and just before it i.e. is it, in 
	   fact, an onset? */
TGLIST tkMakeTglOnsets(TRACK *tk, TKLIST focus);
	/* repeatedly apply tkMakeOnset, gather all results */
TGLIST tgl_onset_tk(TRACK *trk, int *ptype, TKLIST focus);
	/* LNodeFunc to generate onset groups from tracks */

/* from db.metat.c */
double tkAvgHeadFrq(TRACK *trk);
double tkAvgTailFrq(TRACK *trk);
    /* figure out frequency at the end of this track, but weight it over 
       a few points to give it some stability */
double tfDistance(double t1, double f1, double t2, double f2, double fscale);
	/* metric of distance btwn 2 pts in t-f -- hard!
	   <fscale> controls tradeoff between two dimensions. 
	   = 1 -> one semitone is the same distance as one period 
	   = 1/12 -> a period counts a whole octave (onset tests)  
	   = 4 -> a quarter of a semitone will make up for an extra period */
void txSetSortTargetTF(double time, double freq);
    /* entry point to setup statics for sorts */
int tkCmpNrstFHead(NODE **pn1, NODE **pn2);
    /* qsort comparison function:
       Sorts in descending `closeness' to specified t-f point.  
       Calculates frq by looking at head of track. */
int tkCmpNrstFAvg(NODE **pn1, NODE **pn2);
   /* qsort comparison function:
       Sorts in descending `closeness' to specified t-f point.  
       Uses 'average frq' from header. */
int tkCmpNrstOnset(NODE **pn1, NODE **pn2);
    /* qsort comparison function:
       Sorts in descending `onset closeness' - 
       mainly time skew between judged onsets, but frequency distance 
       matters too */
void tklPrintOnsDists(TKLIST trx, char *title);
	/* print out a list of trax with their onset distances */
LIST strip_entry_ptr(LIST *plist, LIST unwantedCar);
	/* go through <*plist> looking for the unwantedCar;
	   remove it wherever it occurs.  Will remove multiple entries.
	   Modifies actual list and returns it too. */
TKLIST tkSuccessors(TRACK *trk, TKLIST focus, TKLIST withouts);
	/* return a list of tracks that pass the threshold to serve as 
	   the continuation of the passed-in track.  Probably have 
	   to test for proximity both of HeadFrq and favg */
TGLIST tgMetaExtend(TKGROUP *tg, TKLIST focus, TKLIST *pwithouts);
	/* find all the candidate extensions, make them all into 
	   new tgs */
TGLIST tgMakeMetatrax(TRACK *tk, TKLIST focus);
	/* Grow a list of `continuity' metatracks from a single seed track */
TGLIST tgl_cont_tk(TRACK *trk, int *ptype, TKLIST focus);
	/* LNodeFunc to generate metatracks from tracks */


/* from db.groups.c */
LIST list_sub(LIST l1, LIST l2);
	/* returns a list of the CARs of l1 that are not in l2 */
LIST list_diff(LIST l1, LIST l2);
	/* returns a list of the elements from l1 and l2 that do 
int is_subset(LIST *little, int type, LIST *bigger);
	/* NodeFunc tests the first argument as a subset of the arg */
int tg_is_subset(TKGROUP *tg, int type, TKGROUP *bigtg);
	/* returns 1 if <tg->parts> is a subset of <bigtg->parts> */
TGLIST tglRemoveSubsets(TGLIST tgl);
	/* go through a list of tracks removing subsets */
double tklWeight(TKLIST tkl);
	/* calculate the `weight' of a list of tracks */
double tgWeight(TKGROUP *tg);
	/* calculate the weight of an actual TKGROUP */
int tg_weight_cmp(TKGROUP **ptg1, TKGROUP **ptg2);
	/* for qsort into descending weight, return 1 if *ptg2 
	   points to a TKGROUP of lesser weight than *ptg1 */
TGLIST tglRemoveNearDuplicates(TGLIST tgl);
 	/* given a list of track groups, will attempt to prune away 
	   subsets and near-subsets, which are defined as 
	   track groups whose `difference weight' is a small proportion 
	   of their `total weight'.  Return list is sorted in descending 
	   weight. */

