/*******************************************************************\
*   db.trkio.c							    *
*   load and save track files and track set descriptors 	    *
*   1993jan14 dpwe						    *
\*******************************************************************/

/********* switches ********/

/********* includes ********/
#include <dpwelib.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"	/* pulls in all track stuff */

/********* defines *********/
/********* types ***********/
/***** static globals ******/

/* track read frq fix */

static char trackFile[80] = "";		/* name of track file */

/******** prototypes ********/
void SaveTrackSelFile(TKLIST selTrx, TKLIST allTrx);
void LoadSeln(void);

/******* implementations *****/

void TrackReadSel(wrcd)
    WINRCD	**wrcd;
{
    LoadSeln();
    return;
}

void TrackWriteSel(wrcd)
    WINRCD	**wrcd;
{
    SaveTrackSelFile(selTrax, allTrax);
    return;
}

void TrackSaveSel(wrcd)
    WINRCD	**wrcd;
{
    SaveTrackFile(wrcd);
    return;
}

SaveTrackFile(wrcd)
    WINRCD	**wrcd;
    {
    FILE	*file;
    char	pkFileName[255];
    TKHDSTRUCT	*trkh;

    SubstExtn((*wrcd)->theFileName, ".sv.trx", pkFileName);
/*    strcpy(pkFileName, "tmp.trx");	/* don't overwrite! */
    TKAlloc(&trkh, TKFMT_BINARY, ((*wrcd)->w.snd.view)->snd->samplingRate,
	    ((*wrcd)->w.snd.cqi)->duration, 
	    ((*wrcd)->w.snd.view)->snd->channels,
	    TKHDDFLTBYTS);
    if(tkSave(trkh, selTrax, pkFileName))  {
	fprintf(stderr, "SaveTrk: wrote %s\n", pkFileName);
    }
    else {
	fprintf(stderr, "SaveTrk: could not write %s\n", pkFileName);
    }
}

LoadTrackFile(wrcd)
    WINRCD	**wrcd;
    {
    FILE	*file;
    char	pkFileName[255];
    TKHDSTRUCT	*trkh;

    if( (*wrcd)->w.snd.cqi )  {
	SubstExtn((*wrcd)->theFileName, ".trx", pkFileName);
	if( (trkh = tkLoad(&allTrax, pkFileName)))  {
	    strcpy(trackFile, pkFileName);	/* save for posterity */
	    selTrax = NULL; 	/* you better believe it */
	    fprintf(stderr, "track file '%s' loaded.\n", pkFileName);
	}
	else
	    fprintf(stderr, "LoadTrk: cannot open trk file '%s'\n", 
		    pkFileName);
    }
}
    
#define TKSELFNAME   "trksel.txt";

void SaveTrackSelFile(TKLIST theTrax, TKLIST allTrx)
{	/* write out a text file describing current selection */
    FILE *file;
    char *fname;
    long index = 0, step;
    TKLIST allTrRun;

    fname = TKSELFNAME;
    if((file = fopen(fname, "w"))==NULL)  {
	fprintf(stderr, "SvTkSel: couldn't create '%s'\n", fname);
	return;
    }
    /* first line is name of source file */
    fprintf(file, "SourceFile: %s\n", trackFile);
    /* now just list the indices of the theTrax */
    allTrRun = allTrx;
    while(theTrax)  {
	index = ((TRACK *)CAR(theTrax))->id;
	fprintf(file, "%ld\n", index);
	fprintf(stderr, "%ld ", index);
	theTrax = (TKLIST)CDR(theTrax);
    }
    fclose(file);
    fprintf(stderr, "\ntksel file '%s' written\n", fname);
}

extern int tk_check_id(LIST car, int type, void *arg);
/* defined in tkgroups.c */

TKLIST LoadTrackSelFile(TKLIST allTrx)
{	/* read back a file as a selection */
    FILE *file;
    char *fname, srcFile[32];
    LIST trx = NULL, *runner;
    TRACK *trk;
    long index, lastIndex;

    fname = TKSELFNAME;
    if((file = fopen(fname, "r"))==NULL)  {
	fprintf(stderr, "LdTkSel: couldn't open '%s'\n", fname);
	return NULL;
    }
    fprintf(stderr, "tksel file '%s' open\n", fname);
    /* first line is name of source file */
    fscanf(file, "SourceFile: %s\n", srcFile);
    fprintf(stderr, "source file: '%s'\n", srcFile);
    if(strcmp(srcFile, trackFile) != 0) {
	fprintf(stderr, "LdTkSel: selection from '%s' not '%s'\n", 
		srcFile, trackFile);
    }	/* continue anyway */
    /* now just list the indices of the selTrax */
    runner = &trx;
    while(!feof(file))  {
	fscanf(file, "%ld\n", &index);
	fprintf(stderr, "%ld ", index);
	trk = (TRACK *)CAR(search_by_fn((LIST)allTrx, tk_check_id, 
					(void *)index));
	runner = append_ptr(runner, new_node(NULL, (LIST)trk, T_TRACK));
    }
    fclose(file);
    return trx;
}

void LoadSeln(void)
{
    if(selTrax != NULL)	/* some tracks currently selected - lose them all */
	{
	RedrawTrax(selTrax, 1);
	free_list((LIST)selTrax);
	selTrax = NULL;
	}
    selTrax = LoadTrackSelFile(allTrax);
    RedrawTrax(selTrax, 2);
    fprintf(stderr, "All top_size %d, sel top_size %d\n",
	    top_size(allTrax), top_size(selTrax));
}

/********** lisp track summary output ***********/

#define DB2LIN(a)	(exp(a*0.11512925))
#define LIN2DB(a)	(8.6858896*log(a))

void GetTrackStats(PKLIST pklist, double *onset, double *offset, 
		   double *avgmag, double *avgfrq, long *npts)
{	/* evaluate summary stats for a track */
    double	linmag, totalWeight = 0, weightedFrq = 0;
    long	nPks = 0;
    PeakPos	*pk;

    *npts = *onset = *offset = *avgmag = *avgfrq = 0.0;
    if(pklist)  {
	*onset = ((PeakPos *)car((LIST)pklist))->tim;	/* 1st peak */
    }
    while(pklist)  {
	pk = (PeakPos *)car((LIST)pklist);
	pklist = (PKLIST)cdr((LIST)pklist);
	linmag = DB2LIN(pk->mag);
	totalWeight += linmag;
	weightedFrq += linmag * pk->frq;
	++nPks;
    }
    if(nPks)  {
	*offset = pk->tim;	/* last peak left over from loop */
	*avgmag = LIN2DB(totalWeight/(double)nPks);
	*avgfrq = weightedFrq/totalWeight;
	*npts = nPks;
    }
}
	
#define NUMPTSTHRESH 5
#define AVGMAGTHRESH -20.0

void WriteTrackSumm(FILE *stream, PKLIST pklist, int index)
{	/* write a line to the lisp-style track summary file 
	   summarizing details of the track indicated by the <pklist>.  */
    double onset, offset, avgmag, avgfrq;
    long  npts;

    GetTrackStats(pklist, &onset, &offset, &avgmag, &avgfrq, &npts);
    if(npts > NUMPTSTHRESH && avgmag > AVGMAGTHRESH)   {
	fprintf(stream, "(+t %.6f %.6f %.2f %.2f %d)\n", 
		onset, offset-onset, avgfrq, avgmag, index);
    }
}

/*  WriteTrxSumm(wrcd)
    Writes out a text file containing summary information for 
    the tracks in <selTrax>, suitable for reading into my lisp 
    code that attempts to apply grouping rules.
    1993jan21
 */
void WriteTrxSumm(WINRCD **wrcd)
{
    char filename[32];
    FILE *file;
    TKLIST trax;
    TKLIST allTrRun;	/* to find track indices for reference */
    int  index;
    TRACK *trk;

    strcpy(filename, "lisptrx.txt");
    if( (file = fopen(filename, "w")) == NULL)  {
	fprintf(stderr, "Couldn't open '%s' for lisptrx output\n", filename);
	return;
    }
    trax = selTrax;
    allTrRun = allTrax;
    index = 0;
    while(trax)  {
/*	index += CountToElement(&allTrRun, car(trax)); 
	WriteTrackSumm(file, (PKLIST)car(trax), index);	*/
	trk = (TRACK *)CAR(trax);
	WriteTrackSumm(file, trk->tktop, trk->id);
	trax = cdr(trax);
    }
    fprintf(stderr, "Lisptrx file '%s' successfully written (%ld bytes)\n",
	    filename, fSize(file));
    fclose(file);
}

#endif /* CQFILE */
#endif /* TRAX */

