/*******************************************************************\
*   db.wfall.c							    *
*   waterfall plot drawing functions for DSPbench		    *
*   17oct90 dpwe						    *
\*******************************************************************/

#include <stdio.h>
#include <string.h>

#include <dsputil.h>
#include <genutils.h>
#include <fft.h>

#include <math.h>
#include "DSPbench.h"

#define SHORTMAX 32767
#define MAXXPTS 2048

static  Point	*xpts = NULL; /* [MAXXPTS]; */
static  short   *maxs = NULL; /* [MAXXPTS]; */

/* static globals */

/* static */ InitTrace()	/* called before each set of DrawTraces to reset */
    {
    int i, err = 0;

    if(xpts == NULL)
	{
	if( (xpts=(Point *)Mymalloc((long)MAXXPTS, sizeof(Point),
								"InitTrace:xpts"))==NULL)
	    err = 1;
	else if((maxs=(short *)Mymalloc((long)MAXXPTS, sizeof(short), 
								"InitTrace:maxs"))==NULL)
	    err = 1;
	if(err)
	    {
	    fprintf(stderr,"%s:InitTrace: malloc fail\n",programName);
	    exit(1);
	    }
	}
    for(i=0; i<MAXXPTS; ++i)
	maxs[i] = SHORTMAX;	/* set up HLR array */
    }

/* static */ DrawTrace(start,len,rec,scale,hoffs) /* draw part-trace of waterfall */
    short *start;
    int   len;
    Rect  rec;
    int   scale;	/* proportional gain for data, 2 == unity */
    int	  hoffs;	/* 'pixels' to offset <maxs> array */
    {
    Point *ptptr = xpts;
    int    i,j;
    short  x_0, y_0, y_min, y_max;
    long   x_scale, y_scale;
    short  *sh_p = maxs;
    int    poffs;
    int    clip = 0, clipState = 0, segBase = 0;
    
    if(len>MAXXPTS)	len = MAXXPTS;	/* clamp to array size */

    y_min   = rec.top;	y_max = rec.bottom;
    x_scale = (((long)(rec.right-rec.left))<<16L)/len;
    y_scale = (long)scale * ((((long)(rec.bottom-rec.top))<<14L)/SHORTMAX);
    x_0     = rec.left;
    y_0     = (rec.top + rec.bottom)/2;	/* midpoint for bipolar plots */
    for(i=0; i<len; ++i)
	{
	ptptr->h = x_0 + (short)(((long)i * x_scale)>>16L);
	ptptr->v = y_0 - (short)(((long)*start++ * y_scale)>>16L);
	clip = 0;
	if(ptptr->v < y_min)	{ ptptr->v = y_min; clip = 1; }
	if(ptptr->v > y_max)	{ ptptr->v = y_max; clip = 1; }
	if(*sh_p > ptptr->v)
	    *sh_p++ = ptptr->v;
	else
	    ptptr->v = *sh_p++;		/* update HLR table */
	++ptptr;
	if(clip == 1 && clipState == 0)
	    DrawLines(xpts+segBase, i-segBase+1);
	if(clip)
	    segBase = i;
	clipState = clip;
	}
    DrawLines(xpts+segBase,len-segBase);

    poffs = (((long)hoffs)<<16L)/x_scale;
    for(i=0; i<MAXXPTS-poffs; ++i)
	maxs[i] = maxs[i+poffs];
    for(i = MAXXPTS-poffs; i<MAXXPTS; ++i)
	maxs[i] = SHORTMAX;
    }

ShowWfall(rec,start,end,chs,mode,scale,len)
    Rect    rec;
    short   *start;
    short   *end;
    int     chs;	/* assumed mono!! */
    int     mode;
    int     scale;	/* 2 == unity, proportional */
    int     len;
    {
    int     tracelen;
    int     tracehop;
    Rect    inrect;
    int     voffs, hoffs;
    int     frames;
    int     i,err=0;

    InitTrace();
    hoffs = 0			/* was 5 for diag offs */ ;  voffs = 5;
    if(len > 16 && len < WFLENMAX)
	tracelen = len;
    else
	tracelen = 1024;
    tracehop = tracelen;
    frames = (end-start+(tracelen-1))/tracelen;
    inrect = rec;
    InsetRect(&inrect,SBarWidth,0);
    inrect.right -= frames*hoffs;
    inrect.top   += frames*voffs;

    while(start + tracelen <= end)
	{
	DrawTrace(start, tracelen, inrect, scale, hoffs);
	OffsetRect(&inrect, hoffs, -voffs);
	start += tracehop;
	}
    }

