/***************************************************************\
*   tkdbase.c							*
*   Encapsulated database for track representations		*
*   A new layer on an old problem.. 				*
*   .. rather like the old lisp Haaseproj			*
*   dpwe 1993apr15						*
\***************************************************************/

#include <dpwelib.h>

#include "tracks.h"		/* includes llist.h and peaks.h too */

typedef struct _tkdb_index_  {	/* element of tk dbase index table */
    float	time;		/* instant that index describes */
    TKLIST	*current;	/* a separate list of tracks that start before 
				   <time> but are still active at <time> */
    TKLIST	*remains;	/* pointer into the main database tracklist to 
				   the 1st track that starts at/after <time> */
} TKDB_INDEX;
#define T_TKIX	103		/* car-type to indicate an index node */
typedef LIST TKIX_LIST;		/* .. a list of track indices */

typedef struct _tkdb_  {	/* parent structure for a database */
    char	*filename;	/* the source file */
    TKLIST	*trax;		/* the main list of tracks */
    long	ntrax;		/* how many tracks in main list */
    TKIX_LIST	idx;		/* list of TKDB_INDEX index nodes */
    double	idxPeriod;	/* time between index entries */
    long	idxCount;	/* number of index entries in list */
} TKDB;

#define _TKDB_DEFINED_		/* (needed to avoid redefining in tkdbase.h) */
#include "tkdbase.h"

/* ------------- Track Index List management ---------------------------- */

TKDB_INDEX *tkdbixNew(double time, TKLIST *current, TKLIST *remains)
{	/* allocate a new tk db index node */
    TKDB_INDEX *tdix;

    tdix = TMMALLOC(TKDB_INDEX, 1, "tkdbixNew");
    tdix->time = time;
    tdix->current = current;
    tdix->remains = remains;
    return tdix;
}

void tkdbixFree(TKDB_INDEX *tdix)
{	/* free a single index node */
    free(tdix);
}

void tkdbixFreeList(TKIX_LIST *indices)
{	/* free a whole chain of td index entries */
    TKDB_INDEX *tkix;

    while(indices)  {
	tkix = (TKDB_INDEX *)CAR(indices);
	indices = CDR(indices);
	tkdbixFree(tkix);
    }
}

/* ------------- Track Database management ---------------------------- */

TKDB *tkdbNew(void)
{	/* initialize a new track database structure */
    TKDB *tkdb;

    tkdb = TMMALLOC(TKDB, 1, "tkdbNew");
    tkdb->filename = NULL;
    tkdb->trax = NULL;
    tkdb->ntrax = 0;
    /* always have 'trivial' index entry pointing to start of list */
    tkdb->idx = new_node(NULL, (LIST)tkdbixNew(0.0, NULL, tkdb->trax),
			 T_TKIX);
    tkdb->idxPeriod = 0.0;
    tkdb->idxCount  = 1;
    return tkdb;
}

TKDB *tkdbLoad(char *filename)
{	/* load a track file into a database */
    TKDB *tkdb = NULL;



    return tkdb;
}

TKLIST *tkdbGetIntersect(TKDB *tkdb, double tMin, double tMax,
			 double fMin, double fMax)
{	/* return tracks from the database with support overlapping 
	   the specified time-frequency rectangle */
    TKLIST *tkl = NULL;


    return tkl;
}
