/***************************************************************\
*	arrayops.c						*
*  Some stuff to work on arrays					*
*  Snatched out of stripk.c					*
*  dpwe 06mar91							*
\***************************************************************/

#include <dpwelib.h>
#include <math.h>
#include <genutils.h>	/* for GCF() in ResampArray() .. */
/*#include <resample.h>	/* for NewResampKernel() in ResampArray() .. */
#include <pvoc.h>

#include <arrayops.h>	/* includes <resample.h> */

/* Static function prototypes */
SampArray *NewSampA(rows, cols)
    long	rows;
    long	cols;
    {
    SampArray *ret;

    ret = TMMALLOC(SampArray, 1, "NewSampA:ret");
    ret->data = TMMALLOC(SAMPL, rows*cols, "NewSampA:data");
    if(ret->data == NULL)
	{
	fprintf(stderr, "NewSampA: malloc fail on %ld x %ld\n",
		rows, cols);
	abort();
	}
    ret->rows = rows;
    ret->cols = cols;

    return ret;
    }

ByArray   *NewByteA(rows, cols)
    long	rows;
    long	cols;
    {
    ByArray *ret;

    ret = TMMALLOC(ByArray, 1, "NewByteA:ret");
    ret->data = TMMALLOC(MyByte, rows*cols, "NewByteA:data");
    if(ret->data == NULL)
	{
	fprintf(stderr, "NewByteArray: malloc fail on %ld x %ld\n",
		rows, cols);
	abort();
	}
    ret->rows = rows;
    ret->cols = cols;

    return ret;
    }

void BAFree(ba)
    ByArray *ba;
    {
    free(ba->data);
    free(ba);
    }

void SAFree(sa)
    SampArray *sa;
    {
    free(sa->data);
    free(sa);
    }

SAMPL FindArrayMax(sar)
    SampArray *sar;
    {
    long r,c,cols;
    SAMPL *sPtr, f, max;

    r = sar->rows;	cols = sar->cols;
    sPtr = sar->data;
    max = *sPtr;	/* preset to value from the array */
    while(r--)
	{
	c = cols;
	while(c--)
	    if( (f = *sPtr++)>max )	max = f;
	}
    return max;
    }

#define BYTEMAX	255

void SampAToByteA(sa_a, pby_a, factor, logflag, dbFact, offset, min, max)
    SampArray	*sa_a;
    ByArray	**pby_a;
    FLOATARG	factor;		/* linear post-normalise scaling */
    int 	logflag;	/* flag to convert to logs */
    FLOATARG	dbFact;		/* scaling factor in log domain */
    int         offset;		/* flag to offset to fsd/2 */
    int 	min, max;	/* limits to byte values */
    {
    size_t	i;
    MyByte 	*b_ptr;
    SAMPL	*sa_ptr;
    float	f,scalefactor;
    long	b,maxbyte = 0,opSize;

    *pby_a = NewByteA(sa_a->rows, sa_a->cols);
    sa_ptr = sa_a->data;
    b_ptr  = (*pby_a)->data;
    scalefactor = factor/FindArrayMax(sa_a);
    /* <factor> is overload spec'd in command line */
    opSize = (*pby_a)->rows * (*pby_a)->cols;
    for(i=0; i<opSize; ++i)
	{
        f = scalefactor * *sa_ptr++;
	if(logflag)
	    {
	    if(f>0)
		b = (long)((float)BYTEMAX*(1.0+dbFact*.13*log(f)));
	    else
		b = 0;
	    }
	else
	    b = (long)((float)BYTEMAX*f);	/* *4 was in here */
	
	if(!offset)
	    {
	    if(b<0)	  b = 0;   /* we're going to put into an unsigned */
	    if(b>BYTEMAX) b = BYTEMAX;
	    if(b>maxbyte) maxbyte = b;
	    }
	else
	    {
	    if(b>maxbyte)	maxbyte = b;
	    if((-b)>maxbyte)	maxbyte = -b;
	    b >> 1;				/* halve the range */
	    if(b>BYTEMAX/2) b = BYTEMAX/2;	/* bipolar allowed here */
	    if(b<((BYTEMAX/2)-BYTEMAX)) b = (BYTEMAX/2)-BYTEMAX;
	    if(b<0) b += BYTEMAX+1;		/* wrap at full extent */
	    }
	b = BYTEMAX-b; /*!!*/	/* flip blacks for whites */
	if(b>max)	b = max;
	if(b<min)	b = min;
	*b_ptr++ = b;
	}
    fprintf(stderr, "ShortAtoByteA: max byte = %ld\n", maxbyte);
    }

void PickColPeaks(ar)		/* zero-out all but local maxs */
    SampArray	*ar;
    {
    long	r,c,rows,cols;
    SAMPL	s,t,u,v,w, *ps, *ps2;

    ps = ar->data;	rows = ar->rows;	cols = ar->cols;
    for(c = 0; c < cols; ++c)
	{
	ps2 = ps;
	t = *ps2;	ps2 += cols;
	u = *ps2;	ps2 += cols;
	v = *ps2;
	w = *(ps2+cols);
	for(r = 2; r < rows-2; ++r)	/* in one from each end */
	    {
	    s = t; t = u; u = v; v = w; w = *(ps2+2*cols); 
	    if( !(u>s && u>t && u>v && u>w))	/* 5pt local max */
		*ps2 = 0;
	    ps2 += cols;
	    }
	++ps;				/* point to top of next column */
	}
    }

void PickRowPeaks(ar)		/* zero-out all but local maxs */
    SampArray	*ar;
    {
    long	r,c,rows,cols;
    SAMPL	s,t,u,v,w, *ps, *ps2;

    ps = ar->data;	rows = ar->rows;	cols = ar->cols;
    for(r = 0; r < rows; ++r)
	{
	ps2 = ps;
	t = *ps2;	ps2 += 1;
	u = *ps2;	ps2 += 1;
	v = *ps2;
	w = *(ps2+1);
	for(c = 2; c < cols-2; ++c)	/* in one from each end */
	    {
	    s = t; t = u; u = v; v = w; w = *(ps2+2); 
	    if( !(u>s && u>t && u>v && u>w))	/* 5pt local max */
		*ps2 = 0;
	    ps2 += 1;
	    }
	ps += cols;			/* point to top of next column */
	}
    }

void DifferenceArray(ar)		/* replace rows with 1st ord diff */
    SampArray	*ar;
    {
    long	r,c,rows,cols;
    SAMPL 	s, *ps = ar->data;

    rows = ar->rows;	cols = ar->cols;
    for(r = 0; r < rows; ++r)
	{				/* doesn't touch last el in each r */
	for(c = 1; c < cols; ++c)
	    {
	    s = *(ps+1);
	    *ps -= s;
	    *ps++ *= -4;		/* scale up to reflect gain loss */
	    }
	++ps;	/* skip over last in row */
	}
    }

void DifferenceArrayC(ar)	/* replace cols with 1st ord diff */
    SampArray	*ar;
    {
    long	r,c,rows,cols;
    SAMPL 	s, *ps = ar->data, *ps2;

    rows = ar->rows;	cols = ar->cols;
    for(c = 0; c < cols; ++c)
	{				/* doesn't touch last el in each r */
	ps2 = ps;
	for(r = 1; r < rows; ++r)
	    {
	    s = *(ps2+cols);
	    *ps2 -= s;
	    *ps2 *= -4;			/* compensate loss in extrema */
	    ps2 += cols;
	    }
	ps += 1;	/* top of next column */
	}
    }

void DilateArray(par,p,q)
    SampArray	**par;
    long 	p,q;			/* new size for each pixel */
    {
    SampArray	*a2,*ar;
    long	newBsize;

    ar = *par;
    printf("Dilate array: was %d x %d, increase by %d x %d \n",
	   ar->rows, ar->cols, p, q);
    a2 = NewSampA(ar->rows, ar->cols*q);
    DilateVector(ar->data, a2->data, q, ar->cols*ar->rows);
    SAFree(ar);
    Transpose(a2);
    ar = NewSampA(a2->rows, a2->cols*p);
    DilateVector(a2->data, ar->data, p, a2->cols*a2->rows);
    SAFree(a2);
    Transpose(ar);
    *par = ar;
    }

void DilateVector(src, dst, dil, len)
    SAMPL	*src;
    SAMPL	*dst;
    long 	dil;
    long	len;
    {
    long 	i;

    while(len--)
	{
	i = dil;
	do  *dst++ = *src; while(--i);
	++src;
	}
    }
    
void DilArray(par,p,q)
    SampArray	**par;
    long	p,q;			/* new size for array */
    {
    SampArray	*a2,*ar;
    long	newBsize;

    ar = *par;
    printf("Dil array: was %d x %d, increase to %d x %d \n",
	   ar->rows, ar->cols, p, q);
    a2 = NewSampA(ar->rows, q);
    DilVector(ar->data, a2->data, ar->cols, q, ar->rows);
    SAFree(ar);
    Transpose(a2);
    ar = NewSampA(a2->rows, p);
    DilVector(a2->data, ar->data, a2->cols, p, a2->rows);
    SAFree(a2);
    Transpose(ar);
    *par = ar;
    }

void DilVector(src, dst, len, dil, rws)
    SAMPL	*src;
    SAMPL	*dst;
    long	len;
    long 	dil;
    long 	rws;
    {
    long 	i,j;
    long	n,r,d;

    d = (((long)len)<<16L)/((long)dil);
    n = 0;
    while(rws--)
	{
	r = 0L;
	for(i=0; i<dil; ++i)
	    {
	    *dst++ = *src;
	    r += d;
	    if(r>(1L<<16))
		{
		++src;
		r -= 1L<<16;
		}
	    }
	++src;
	}
    }

/* Algorithm for taking arbitrary nxm array and making it pxq */
/* (a) It starts as n rows of m columns.  
       Use the integer ratio resampler to rescale the whole
       vector q/m i.e. to become n rows of q columns
   (b) Transpose the matrix to q rows of n columns
   (c) Resample by p/n to be qxp
   (d) Transpose back to be pxq             */

#define RSLOBES	4	/* half the multiplies per output point */

void ResampArray(par,p,q)
    SampArray	**par;
    long	p,q;	   		/* new size for array */
    {
    SampArray	*a2, *ar;
    long	opSize;
    long 	i,j;
    SAMPL	*b1, *b2;
    RsmpDescType *rdp;
    long 	gcf;

    ar = *par;
    printf("ResampArr: was r %d c %d to be r %d c %d \n",
	   ar->rows, ar->cols, p, q);
    a2 = NewSampA(ar->rows, q);
    b1 = ar->data;	b2 = a2->data;
    opSize = a2->rows * a2->cols;
    ClearSampls(b2, 1, opSize);
    gcf = GCF(q, ar->cols);	/* remove common factors */
    if(gcf != 1)	fprintf(stderr,
				"ResampArr: gcf of %ld and %ld is %ld\n",
				q,ar->cols,gcf);
    rdp = NewResampKernel(ar->cols/gcf, q/gcf, RSLOBES, (float)1);
    for(i = 0; i<ar->rows; ++i)
	{
	ResampleIntoBuffer(b2, a2->cols, b1, ar->cols, 0L, rdp, 0);
	b1 += ar->cols;
	b2 += a2->cols;
	}
    FreeResampKernel(rdp);
    SAFree(ar);
    *par = a2;

    Transpose(a2);
    ar = NewSampA(a2->rows, p);
    b1 = a2->data;	b2 = ar->data;
    opSize = ar->rows * ar->cols;
    ClearSampls(b2, 1, opSize);
    gcf = GCF(p, a2->cols);
    if(gcf != 1)	fprintf(stderr,
				"ResampArr: gcf of %ld and %ld is %ld\n",
				p,a2->cols,gcf);
    rdp = NewResampKernel(a2->cols/gcf, p/gcf, RSLOBES, (float)1);
    for(i=0; i < a2->rows; ++i)
	{
	ResampleIntoBuffer(b2, ar->cols, b1, a2->cols, 0L, rdp, 0);
	b1 += a2->cols;
	b2 += ar->cols;
	}
    FreeResampKernel(rdp);
    SAFree(a2);
    Transpose(ar);
    *par = ar;
    }

void Transpose(ar)	/* swap rows for columns in place */
    /* Lightweight sledgehammer method : tries every point as a seed for
       a swap-ring;  Keeps a flag array with a bit for EVERY ELEMENT IN ARRAY
       which lets it know when it's done */
    SampArray	*ar;
    {
    long  remain = ar->rows * ar->cols;	 /* writes left to do */
    char *flags;
    long flagix,sfx;
    int  flagmask,sfk;
    long i,r,c,sr,sc,rows,cols,ix;
    SAMPL t,u,*data;
    long hits = 0, tries = 0;

    rows = ar->rows; cols = ar->cols; data = ar->data;
    flags = TMMALLOC(char, (rows*cols+7)/8L, "Transpose"); /* 1 bit per cell */
    for(i = 0; i<(rows*cols+7)/8L;++i)	flags[i] = '\0';
    sr = 0; sc = 0;
    flags[0] |= 1<<0; flags[(rows*cols-1)/8] |= 1<<((rows*cols-1)%8);
    remain -= 2;
    /* hand mark top left and bottom right as done, deduct */
    for(sr = 0; remain >0 && sr<rows; ++sr)
	for(sc = 0; remain >0 && sc<cols; ++sc)
	    {
	    r = sr; c = sc;
	    ix = r*cols+c;	/* source */
	    sfx = ix/8;	sfk = 1<<(ix%8);
	    u = *(data+ix);
	    if( (flags[sfx]&sfk) == 0)
		hits += 1;
	    tries += 1;
	    while( (flags[sfx]&sfk) == 0)
		{
		ix = c*rows+r;	/* dest ix, trnspsed pos, new colsz */
		t  = *(data+ix); *(data+ix) = u; u = t; 
		--remain;
		r  = ix/cols;    c  = ix - r*cols; /* current coos of u */
		flagix = ix/8; flagmask = 1<<(ix%8);
		if(flags[flagix]&flagmask)
		    {
		    fprintf(stderr,
		"Transp: crossed rings ar %dx%d, ring %d,%d, pt %d,%d\n",
			    rows,cols,sr,sc,r,c);
		    abort();
		    }
		else
		    flags[flagix] |= flagmask;
		}
	    }
    if(remain != 0)
	{ fprintf(stderr, "Trans: fail remain %d, ar %dx%d\n",
		  remain,rows,cols);
	  abort();	}
    fprintf(stderr,"Trans: finished %dx%d at %d,%d ; hits %d of %d\n",
	    rows, cols, sr, sc, hits, tries);
    free(flags);
    ar->rows = cols; ar->cols = rows;
    }

void oldTranspose(ar)	/* swap rows for columns */
    SampArray	*ar;
    {
    SampArray   *a2;
    SAMPL  *buf, *b, *d;
    size_t newBsize;
    long	  x,y,r,c;

    x = ar->cols; y = ar->rows;
    d = ar->data;
    a2 = NewSampA(x,y);
    b = a2->data;
    for(c = 0; c<x; ++c)
	for(r = 0; r<y; ++r)
	    *b++ = d[c + x*r];
    buf = ar->data;
    ar->data = a2->data;
    a2->data = buf;
    SAFree(a2);
    ar->rows = x;    ar->cols = y; 
    }

void ReverseRows(ar)		/* reverse each row i.e. columnwise */
    SampArray	*ar;
    {
    long 	r,c,cols;
    SAMPL	*sa_ptr,t;

    sa_ptr = ar->data;
    cols = ar->cols;
    for(r = 0; r < ar->rows; ++r)
	{
	for(c = 0; c < ((ar->cols)>>1); ++c)
	    {
	    t = sa_ptr[c];
	    sa_ptr[c] = sa_ptr[cols-c-1];
	    sa_ptr[cols-c-1] = t;
	    }
	sa_ptr += ar->cols;	/* shift base to next row */
	}
    }

void PrintArr(ar)
    SampArray *ar;
    {
    int	i,j;
    printf("Array: rows %d, cols %d\n",ar->rows,ar->cols);
    for(i = 0; i< 10; ++i)
	{
        for(j=0; j<9; ++j)
	    printf("%8.2f",(float)ar->data[(long)i*(long)ar->cols+j]);
	printf("\n");
	}
    }

