/***********************************************************************\
*	audIO.c								*
*	New hardware-indep audio IO interface intended to be portable	*
*	based on lofiplay etc						*
*	dpwe 1993nov17, 13mar92						*
\***********************************************************************/

typedef struct {		/* descriptor for audio IO channel */
    int	mode;			/* AU_INPUT or AU_OUTPUT */
    int chans;			
    int format;
    void *ptr;			/* some kind of handle */
} AUSTRUCT;

#define AUSTRUCT_DEFINED	/* stop redefinition in audIO.h */

#include "audIO.h"	/* for stdio, PARG etc */
#include "play.h"	/* for now we are going to implement this 
			   using the scummy old play level */
#include <assert.h>

AUSTRUCT *AUOpen(int format, double srate, int chans, long desbuf, int mode)
{  /* open a new audio IO channel.  Define data format, #channels and sampling
      rate. Request that output buffer be desbuf frames long.  
      mode is AIO_{INPUT,OUTPUT} */
    AUSTRUCT *au = NULL;
    int	     err = 0;

#ifdef DEBUG 
    fprintf(stderr, "AUOpen(fmt %d, sr %.0lf, ch %d, desb %ld, mo %d)\n",
	    format, srate, chans, desbuf, mode);
#endif /* DEBUG */
    /* check arguments */
    if(format != SFMT_SHORT /* || ... */)  {
	fprintf(stderr, "AUOpen: format %d not supported\n", format);
	++err;
    }
    if(srate < 8000 || srate > 48000 /* || ... */)  {
	fprintf(stderr, "AUOpen: srate %.0f not supported\n", srate);
	++err;
    }
    if(chans < 1 || chans > 2 /* || ... */)  {
	fprintf(stderr, "AUOpen: chans %d not supported\n", chans);
	++err;
    }
    if(mode != AIO_INPUT && mode != AIO_OUTPUT)  {
	fprintf(stderr, "AUOpen: mode %d not supported\n", mode);
	++err;
    }
    /* not sure what to do with desbuf .. */

    if(err == 0)  {
	/* allocate the descriptor */
	au = TMMALLOC(AUSTRUCT, 1, "AUOpen");
	au->mode = mode;	au->chans = chans;
	au->format = format;	au->ptr = NULL;

	/* open the audio */
	if(play_set(chans, SFSampleSizeOf(format), srate, 2) == 0)  {
	    /* error initializing hardware */
	    free(au);
	    return NULL;
	}
	/* (ignores desbuf) */
    }
    return au;
}

long	AUWrite(AUSTRUCT *auch, void *buf, long frames)
{   /* write some sound to the open audio IO channel, per description */
    assert((auch->mode) == AIO_OUTPUT);
    play_on((short *)buf, frames);
    return frames;	/* presumably */
}

long	AURead(AUSTRUCT *auch, void *buf, long frames)
{   /* read some sound from the open audio IO channel, per description */
    assert(auch->mode == AIO_INPUT);
    fprintf(stderr, "AURead: sound input not implemented!\n");
    return 0;
}

void	AUClose(AUSTRUCT *auch)
{  /* close an audio IO channel */
#ifndef LOFI  /* for now, no close on LoFi, since can only re-open once */
    play_rls();
#endif /* LOFI */ 
    free(auch);
}

long AUBufConts(AUSTRUCT *auch)
{   /* see how many sample frames are currently in audio buffer.
       (can be used for synchronization, timing etc). */
    return play_bufremains();
}
