/***************************************************************\
*	byteswap.c		  				*
*   swap bytes e.g. for files between different machines 	*
*   dpwe 11nov91 after sndcvt 20feb91				*
\***************************************************************/

#include <math.h>	/* for fabs() prototype */
#include <ctype.h>	/* for NumericQ() */
#include "dpwelib.h"
#include "byteswap.h"

int MatchByteMode(long red, long target)
{	/* return the bytemode to apply to 'red' to make 'target', or -1 */
    int mode = -1;	/* error return */

    if(lshuffle(red, INORDER) == target)	mode = INORDER;
    else if(lshuffle(red, BYTEREV) == target)	mode = BYTEREV;
    else if(lshuffle(red, WORDREV) == target)	mode = WORDREV;
    else if(lshuffle(red, BYWDREV) == target)	mode = BYWDREV;
    return mode;
}

int HostByteMode()	/* relative to 68000-style (MSB byte first) */
    {
    /* work it out */
    char tchrs[4];
    long l, m = 0x01020304;
    int mode;

    tchrs[0] = '\1';	tchrs[1] = '\2';
    tchrs[2] = '\3';	tchrs[3] = '\4';
    l = *(long *)tchrs;
    if( (mode = MatchByteMode(l, m)) < 0)  {
	fprintf(stderr, "byteswap: unrecognised longword order %lx\n", l);
	abort();
    }
    return mode;
    }

char *BytemodeName(bytemode)
    int bytemode;
    {
    char *boStr;

    switch(bytemode)
	{
    case INORDER:       boStr = "in order";		break;
    case BYTEREV:	boStr = "byte reversed";	break;
    case WORDREV:	boStr = "word reversed";	break;
    case BYWDREV:       boStr = "byte & word reversed";	break;
    default:		boStr = "? byte order";
	}
    return(boStr);
    }

/*** from dec2mac.c ***/
short wshuffle(w,mode)
    int   w;
    int   mode;
    {
    short lb, hb;   /* actually bytes */

    lb = 0xFF & w;
    hb = 0xFF & (w>>8);
    if(mode==BYTEREV || mode == BYWDREV)
	return( (lb<<8) + hb );
    else
	return( (hb<<8) + lb );
    }

long lshuffle(l,mode)
    long l;
    int  mode;
    {
    short lw, hw;

    hw = wshuffle( (short) (l>>16L) , mode );
    lw = wshuffle( (short) l , mode );
    if(mode == INORDER || mode == BYTEREV)
	return( ((0xFFFFL&(long)hw)<<16L) + (0xFFFFL&(long)lw ) );
    else
	return( ((0xFFFFL&(long)lw)<<16L) + (0xFFFFL&(long)hw ) );
    }

long FloatAsLong(f, mode)
    FLOATARG f;
    int   mode;
    {
    long l, *pl;

    pl = (long *)&f;
/*    if(mode == WORDREV || mode == BYWDREV)  /* detect DEC float formats */
/*	l = ToUFloat(f);
/*    else
 */
	l = *pl;
    return(lshuffle(l, mode));
    }

float FloatFromLong(l, mode, fixup)
    long  l;
    int   mode;
    long  fixup;	/* poke and hope to fix DEC float formats */
    {
    long  *pl;
    float f, *pf;

    pf = &f;
    pl = (long *)pf;
    *pl = lshuffle(l,mode);
    *pl += (long)fixup;
    return(f);
/*    if(mode == WORDREV || mode == BYWDREV)  /* detect DEC float formats */
/*	l = ToUFloat(f);	*/
    }

void ConvertBufferBytes(buf, bytes, wordsz, mode)
    char	*buf;
    long 	bytes;
    int 	wordsz;		/* 1, 2 or 4 : maybe should be FLOAT etc */
    int 	mode;
    {
    long 	i;
    short	*sp;
    long	*lp;
    /* no changes needed if buffer is chars, or no swapping implied, or 
       word swapping for width-2 data */

/*    fprintf(stderr, "buf %lx bytes %ld wsz %d mode %d\n",
	    buf, bytes, wordsz, mode);	*/
    if(wordsz < 2 
       || mode == INORDER
       || (mode == WORDREV && wordsz < 4) )	return;
    if(wordsz == 4 && mode == BYTEREV)  {
	wordsz = 2;	/* swapping bytes in a word same for long/short */
    }
    if(wordsz == 2)  {		/* only actual operation on words is swab */
	sp = (short *)buf;
	i = bytes/wordsz;
	while(i > 0)  {
	    *sp = WSWAPB(*sp);
	    ++sp;	--i;
	}
    } else {				/* presume wordsz == 4 */
	lp = (long *)buf;
	i = bytes/wordsz;
	if(mode == BYWDREV) {
	    while(i > 0)  {
		*lp = LSWAPB(*lp);
		++lp; --i;
	    }
	} else {			/* mode == WORDREV */
	    while(i > 0)  {
		*lp = LSWAPW(*lp);
		++lp; --i;
	    }
	}	
    }
}

#ifdef MAIN

main()	/* for test */
    {
    printf("hostmode is %s\n", BytemodeName(HostByteMode()));
    }
	   

#endif /* MAIN */
