/* fakeAF.c

   Emulate the SGI AF calls but redirect to SNDF
   hmmm... might work!
   dpwe 1992sep05
 */

#include "fakeAF.h"
#include <stdlib.h>		/* malloc() and free() */
#include <string.h>  	/* more memcpy() prototype */
#ifdef THINK_C
#include <unix.h>		/* for fdopen() prototype */
#endif /* THINK_C */
#if (MIPSEL && ultrix)   /* MIPS compiler can't handle ANSI varargs */
#ifndef __GNUC__
#undef __STDC__
#endif /* __GNUC__ */
#endif /* (MIPSEL && ultrix) */
#include <varargs.h>		/* for aferror handler varargs */

#define MYINFOMAX	1024	/* makes irc simple */

/* for AFfile.op field */
#define FKAF_READ	1
#define FKAF_WRITE	2

AFfilesetup AFnewfilesetup(void)
    {
    SFSTRUCT *snd;
    int err;

    if(err = SFAlloc(&snd, SF_UNK_LEN, SFMT_SHORT, 44100.0, 2, MYINFOMAX))
	SFDie(err, "AFnewsetup");
    return snd;
    }

AFfilehandle AFopenfile(name, mode, setup)
    char *name;
    char *mode;
    AFfilesetup setup;
{
    AFfile	aff;
    AFfilehandle rslt;
    int 	op;

    if(*mode == 'r')  {
		aff.file = SFOpenRdHdr(name, setup, MYINFOMAX);
		op = FKAF_READ;
	}
    else if(*mode == 'w')  {
		aff.file = SFOpenWrHdr(name, setup);
		op = FKAF_WRITE;
	}
    else{
		fprintf(stderr, "AFopenfile: bad mode %s\n", mode);
		abort();
	}
    if(aff.file == NULL)
		SFDie(SFE_NOPEN, name);
    rslt = (AFfilehandle)malloc(sizeof(AFfile)+setup->headBsize);
    memcpy(rslt, &aff, sizeof(AFfile));
    memcpy(rslt+1, setup, setup->headBsize);	/* copy the SFSTRUCT */
    rslt->snd = (SFSTRUCT *)(rslt+1);
    rslt->op = op;
    return rslt;
}

AFfilehandle  AFopenfd(int fd, char *mode, AFfilesetup setup)
{
    AFfile	aff;
    AFfilehandle rslt;
    int 	op;

	if(*mode == 'r')  {
		aff.file = fdopen(fd, mode);
		if(aff.file)
		    SFerror = SFReadHdr(aff.file, setup, MYINFOMAX);
		else
			SFerror = SFE_NOPEN;	
		op = FKAF_READ;
	}
    else if(*mode == 'w')  {
		aff.file = fdopen(fd, mode);
		if(aff.file)
			SFerror = SFWriteHdr(aff.file, setup, 0);
		else
			SFerror = SFE_NOPEN;
		op = FKAF_WRITE;
	}
	if(SFerror)
		SFDie(SFerror, "AFopenfd");
    rslt = (AFfilehandle)malloc(sizeof(AFfile)+setup->headBsize);
    memcpy(rslt, &aff, sizeof(AFfile));
    memcpy(rslt+1, setup, setup->headBsize);	/* copy the SFSTRUCT */
    rslt->snd = (SFSTRUCT *)(rslt+1);
    rslt->op = op;
    return rslt;
}
	
int AFgetchannels(AFfilehandle handle, int track)
    {
    return handle->snd->channels;
    }

long AFgetsampcnt(AFfilehandle handle, int track)  /* NOT /chans! */
    {
    return handle->snd->dataBsize / SFSampleSizeOf(handle->snd->format);
    }

double AFgetrate(AFfilehandle handle, int track)
    {
    return (double)handle->snd->samplingRate;
    }

void AFclosefile(AFfilehandle handle)
    {
    if(handle->op == FKAF_READ)  {
	fclose(handle->file);
	free(handle);
	}
    else if(handle->op == FKAF_WRITE)  {
	SFCloseWrHdr(handle->file, handle->snd);
	free(handle);
	}
    else {
	fprintf(stderr, "AFclosefile: bad op in handle\n");
	abort();
	}
    }

void AFfreefilesetup(AFfilesetup setup)
    {
    free(setup);
    }

int AFinitchannels(AFfilesetup setup, int track, int chans)
    {
    setup->channels = chans;
    return 0;	/* success */
    }

int AFinitrate(AFfilesetup setup, int track, double rate)
    {
    setup->samplingRate = rate;
    return 0;	/* success */
    }

long AFreadsamps(AFfilehandle handle, int track, void *buf, long nsamps)
    {    /* samples, not sframes */
    long red;
    int  ssiz;

    ssiz = SFSampleSizeOf(handle->snd->format);
    red = fread(buf, ssiz, nsamps, handle->file);
    return red;
    }

long AFwritesamps(AFfilehandle handle, int track, void *buf, long nsamps)
    {	/* samples, not sframes */
    long red;
    int  ssiz;

    ssiz = SFSampleSizeOf(handle->snd->format);
    red = fwrite(buf, ssiz, nsamps, handle->file);
    return red;
    }

long AFreadframes(AFfilehandle handle, int track, void *buf, long count)
    {    /* samples, not sframes */
    long red;
    int  ssiz;

    ssiz = SFSampleFrameSize(handle->snd);
    red = fread(buf, ssiz, count, handle->file);
    return red;
    }

long AFwriteframes(AFfilehandle handle, int track, void *buf, long count)
    {	/* samples, not sframes */
    long red;
    int  ssiz;

    ssiz = SFSampleFrameSize(handle->snd);
    red = fwrite(buf, ssiz, count, handle->file);
    return red;
    }

void _AFdefaulterrorhandler(err, format, va_alist)
	long err;
	const char *format;
	va_dcl
{
    va_list args;

    if(err || (format && *format))  {
	fprintf(stderr, "audiofile error code %d\n", err);
    }
    if((format && *format))  {
	va_start(args);
	vfprintf(stderr, format, args);
	va_end(args);
	fputc('\n', stderr);
    }
}

static AFerrfunc aferrfunc = (AFerrfunc)_AFdefaulterrorhandler;

AFerrfunc AFseterrorhandler(AFerrfunc efunc)	/* what to do ? */
{
    AFerrfunc	oldefunc;

    oldefunc = aferrfunc;
    aferrfunc = efunc;
    return oldefunc;
}

#define BITS_PER_BYTE 8

void AFgetsampfmt(AFfilehandle file, long track, 
				  long *sampfmt, long *sampwidth)
{
	*sampfmt = AF_SAMPFMT_TWOSCOMP;
	*sampwidth = BITS_PER_BYTE * SFSampleSizeOf(file->snd->format);
}
		
void AFinitsampfmt(AFfilesetup setup, long track,
				   long sampfmt, long sampwidth)
{
	if(sampfmt != AF_SAMPFMT_TWOSCOMP)
		SFDie(SFE_UNK_ERR, "AFinitsampfmt: only TWOSCOMP supported");
	if(sampwidth == 8)
		setup->format = SFMT_CHAR;
	else if(sampwidth == 16)
		setup->format = SFMT_SHORT;
	else if(sampwidth == 32)
		setup->format = SFMT_LONG;
	else 
		SFDie(SFE_UNK_ERR, "AFinitsampfmt: width not 8/16/32");
}

void AFinitcompression(AFfilesetup setup, long track, 
					   long compression)
{
	if(compression != AF_COMPRESSION_NONE)
		SFDie(SFE_UNK_ERR, "AF compression not supported");
}
void AFinitfilefmt(AFfilesetup setup, long fmt)
{
}
void AFinitinstids(AFfilesetup setup, long instids[], long ninsts)
{
}
void AFinitmiscids(AFfilesetup setup, long miscids[], long nmisc)
{
}
void AFinitmarkids(AFfilesetup setup, long trackid, 
				   long markids[], long nmarks)
{
}


