/***************************************************************\
*   hexdump.c													*
*	Print hex data formatted nicely (like od -h or bdump)		*
*   dpwe 1993mar19							                    *
\***************************************************************/

#include <dpwelib.h>
#include <assert.h>
#include "hexdump.h"

#define MAXHEXLINE	16	/* bytes displayed each line */
#define ASCIICOL    62  /* where to start ascii */

void hexdumpLine(FILE *stream, BYTE *data, long len, long base)
{	/* format some hex out to the console */
	long i, this;
	int  j, jmax;
	long pos;
	BYTE	*dptr;
	char	c;

 top:
	this = MIN(MAXHEXLINE, len);
	pos = 0;
	fprintf(stream, "%08lx ", base);	
	pos += 9;
	dptr = data;
	for(i = 0; i<this; i += 4)  {
		jmax = MIN(4, this - i);
		for(j = 0; j<jmax; ++j)  {
			fprintf(stream, "%02x ", *dptr++);
			pos += 3;
		}
		fprintf(stream, " ");
		pos += 1;
	}
	while(pos < ASCIICOL)  {
		fprintf(stream, " ");
		++pos;
	}
	dptr = data;
	for(i = 0; i<this; ++i)  {
		c = (char)*dptr++;
		if(c>= ' ' && c < 0x7F)
			fprintf(stream, "%c", c);
		else 
			fprintf(stream, ".");
	}
	fprintf(stream, "\n");
	len -= this;
	base += this;
	data += this;
	if(len)	goto top;
}

void hexdumpLines(FILE *stream, BYTE *data, long len, long base)
{	/* format some hex out to the console.
	   Be intelligent about repeated data. */
	long done = 0;
	long thistime, lasttime = 0;
	int  skippedlast = 0;

	while(done < len)  {
		thistime = MIN(len-done, MAXHEXLINE);
		/* is this line the same as the last one? */
		if(thistime == lasttime 
		   && memcmp(data + done, data + done - lasttime, thistime)==0
		   && (done + thistime) != len /* NOT on last line */ )  {
			/* data matches last chunk */
			if(!skippedlast) {	/* last line was actually printed */
				fprintf(stream, "*\n");		/* indicate some data skipped */
				skippedlast = 1;			/* only do it once */
			}
		} else {
			/* all new data - print it out */
			skippedlast = 0;
			hexdumpLine(stream, data + done, thistime, done);
		}
		done += thistime;
		lasttime = thistime;
	}
}

int hexdumpFile(FILE *file, long len, long base)
{
	BYTE *buf;

	buf = TMMALLOC(BYTE, len, "DumpHexFile");
	if(fread(buf, 1, len, file) < len)
		return 0;
	hexdumpLines(stdout, buf, len, base);
	free(buf);
	return 1;
}	

#ifdef MAIN		/* test code is minimal version of bdump */

#include <genutils.h>	/* for fSize() */

char *programName;

void main(int argc, char **argv)
{
	FILE *file;
	char *filename;

	programName = argv[0];

	if(argc != 2)  {
		fprintf(stderr, "usage: %s filename     prints out file as hex\n", 
				programName);
		exit(1);
	}
	filename = argv[1];
	if( (file = fopen(filename, "rb")) == NULL)  {
		fprintf(stderr, "%s: couldn't open '%s' for read\n", 
				programName, filename);
		exit(1);
	}
	hexdumpFile(file, fSize(file), 0L);
	fclose(file);
	exit(0);
}
	
#endif /* MAIN */
