/*******************************************************\
*	play_cover.c					*
*	accept samples for playing, no buffer size 	*
*	constraints					*
*	dpwe 03mar92					*
\*******************************************************/

#include "play_cover.h"
#include <samplops.h>

#define PERF_SIZE 256		/* blocks to pass to play_on */
#define BUF_SIZE  (4*PERF_SIZE)	/* BUF_SIZE must be exact multiple of PERF_ */

/* #define DEBUG 1	*/

static short 	mybuf[BUF_SIZE];
static int	head = 0;
static int	tail = 0;

play_common(buf, siz, sil)
    short	*buf;
    long	siz;
    int 	sil;	/* if 1, play silence, don't look at buf */
    {
    int 	cop_size, qlen;

#ifdef DEBUG
    fprintf(stderr, "play_c: buf %lx siz %ld sil %d\n", buf, siz, sil);
    fprintf(stderr, "play_c: mbf %lx hed %d tal %d\n", mybuf, head, tail);
#endif /* DEBUG */
    while(siz > 0)
	{
	/* copy bytes into buf up to end of linear buf */
	/* because of alignment, that much is guaranteed to trigger play_on */
	if(siz > (BUF_SIZE - tail))
	    cop_size = BUF_SIZE - tail;
	else
	    cop_size = siz;
	if(sil)
	    ClearShorts(mybuf+tail, 1, cop_size);
	else{
	    CopyShorts(buf, 1, mybuf+tail, 1, cop_size, FWD);
	    buf += cop_size; 
	    }
	tail = (tail+cop_size)%BUF_SIZE;
	siz -= cop_size;
	qlen = 1+((tail - head + BUF_SIZE - 1)%BUF_SIZE);
	/* qlen must be in range 1..BUF_SIZE, not 0..BUF_SIZE-1 */
#ifdef DEBUG
	fprintf(stderr, "play_c: cop'd %d, qlen now %ld\n", cop_size, qlen);
	fprintf(stderr, "play_c: mbf %lx hed %d tal %d\n", mybuf, head, tail);
#endif /* DEBUG */
	while(qlen >= PERF_SIZE)
	    {
	    play_on(mybuf+head, PERF_SIZE); /* won't align wrap unless 
					     BUF_S = n * PERF_S */
	    head = (head+PERF_SIZE)%BUF_SIZE;
	    qlen -= PERF_SIZE;
#ifdef DEBUG
	    fprintf(stderr, "play_c: pld %d, qlen now %ld\n", PERF_SIZE, qlen);
	    fprintf(stderr, "play_c: mbf %lx hed %d tal %d\n", mybuf, head, tail);
#endif /* DEBUG */
	    }
	}
    return;
    }

void play_some(buf, siz)
    short *buf;
    long  siz;
    {
    play_common(buf, siz, 0);
    }

void play_silence(siz)
    long siz;
    {
    play_common(NULL, siz, 1);
    }

void play_some_flush()
    {	/* just play out what is left in buffer, tack on some silence */
    int 	rem_buf;

    rem_buf = BUF_SIZE - tail;
    play_silence(rem_buf);
    }
    
