/***********************************************************************\
*	readaft.h							*
*  functions to extract useful values from aft files (STFT anal)	*
*  dpwe 02sep91 after readaqt.c, 11mar91				*
\***********************************************************************/

#include <stdio.h>	/* for FILE, stderr */
#include <pvoc.h>	/* PVSTRUCTs here, cpxpolar */

#define MAXFCPX	1025	/* maximum complex points per frame */

#ifndef DESCLEN
#define DESCLEN	16
#endif

typedef struct aftinfo	/* pertinent info for AFT files */
    {
    long	totalFrames;	/* in FILE, if known */
    float	duration;
    long	flags;		/* various format info */
    PVSTRUCT	*pvh;
    FILE	*pvFile;	/* might be handy */
    char	desc[DESCLEN];	/* tag for debug */
    } AFTINFO;

extern char *programName;

#ifdef __STDC__
AFTINFO	*NewAFTInfo(char *desc); /* allocate the space */
void 	 FreeAFTInfo(AFTINFO *afi);
FILE 	*OpenAFTPVfile(char *path, AFTINFO **pafi);
    /* attempt to open <path>; check type; allocate & fill new hdrs if OK */
void 	 CloseAFTPVfile(AFTINFO *afi);
AFTINFO	*DecodeAFTPVH(PVSTRUCT *pvh, FILE *file, char *desc);
    /* allocate & fill a new AFTINFO struct from the passed PVSTRUCT */
    /* doesn't access <file>, just puts it in the struct */
void	 SpecOfAFTFile(AFTINFO *afi, FLOATARG time, FLOATARG prd, 
		       cpxpolar *ans, int dphasq);
    /* using the AFT file described by <pvh>
       interpolate to obtain a complete spectrum at the time
       specified by <time>.  Put this complex, polar vector into the space
       pointed to by <ans>, assumed large enough.
       Differentiate phases if dphasq != 0 */
#else
AFTINFO	*NewAFTInfo();
void	 FreeAFTInfo();
FILE	*OpenAFTPVfile();
void	 CloseAFTPVfile();
AFTINFO	*DecodeAFTPVH();
void	 SpecOfAFTFile();
#endif /* __STDC__ */
