/***********************************************************************\
*	readaqt.c							*
*  functions to extract useful values from aqt files (new CQT)		*
*  dpwe 11mar91								*
\***********************************************************************/

#include <dpwelib.h>	/* for FILE, stderr */
#include <genutils.h>	/* for Mymalloc() */
#include <ctype.h>
#include <math.h>
#ifdef __STDC__
double rint(double);	/* is it being missed? */
double fabs(double);
#else
double rint();
double fabs();
#endif /* __STDC__ */
#include <readaqt.h>
#include <samplops.h>   /* e.g. CopyFloats */
#ifdef DEBUGGRAPHS
#include <window.h>	/*for test code only */
#endif /* DEBUGGRAPHS */
#include <arrayops.h>	/* ditto   */

/************** switches *****************/
#define SINC 1
#define GAUSS 0
#define DDIFF -1	/* discrete differentiator also an option */
#define INTERPTYPE GAUSS
/* if INTERPTYPE == 1, then SpecOfCQFile uses sinc interpolator, else gauss */

/* stat fn protos */
#ifdef __STDC__
static long ReadFrames(CQTINFO *cqi, cpxpolar *dstBase, long dstFrm, 
		long rdFrame, long framesToRead);
static void CopyFrames(cpxpolar *srcBase, long srcFrm, cpxpolar *dstBase, 
		       long dstFrm, long nFrms, int dir, int nBins);
static int ReadCQTFileFrame(CQTINFO *cqi, long frame, cpxpolar *dest);
/* static void print_vec(float *vec, long num, char *tit);	*/
void UnwrapPhases(float *src, int sinc, float *dst, int dinc, long len,
		  FLOATARG expDPhas);
#else
static long ReadFrames();
static void CopyFrames();
static int ReadCQTFileFrame();
/* static void print_vec(); */
void UnwrapPhases();
#endif /* __STDC__ */

void PutInfoArg(str, title, arg)
    char *str;		/* character block in which to leave arg */
    char *title;	/* title of arg */
    FLOATARG	arg;	/* (float) value to write */
    {
    sprintf(str+strlen(str),"%s=%.2f;",title,arg);
    }

int GrabInfoArg(str, title, ans)
    char *str;		/* info string to search */
    char *title;	/* name of info to find */
    float *ans;		/* where to put result  */
    {	/* return code is 1 if info found */
    int 	rc = 0, found = 0;
    char	c;

    while(!found && str && *str)
	{
	c = *str;
	while( isspace(c) )	/* skip leading space */
	    c = *++str;
	if(strncmp(str, title, strlen(title)) == 0)
	    found = 1;
	else{
	    str = strchr(str, ';');    	/* skip to next semicolon */
	    if(str)	++str;		/* if found, look at next */
	    }
	}
    if(!found)
	return 0;
    else{
	str = strchr(str, '=');		/* skip to = sign */
	if(str)		++str;		/* and next */
	else		return 0;	/* no = -> fail */
	sscanf(str,"%f",ans);		/* read answer */
	return 1;
	}
    }

int OctOfFileFrame(frame) /* return which 8ve the <frame>th frame is */
    long frame;	
    {
    int val = 0;
    int mask = 1;
    int lev = 0;

    while( (frame&mask) != val && lev < MAXOCTS)
	{ val = (val<<1)+1; mask = (mask<<1)+1; ++lev; }
    return lev;
    }

CQTINFO *NewCQTInfo(desc)
    char *desc;
    /* allocate a new struct, copy in label */
    {
    CQTINFO *rslt;

    rslt = (CQTINFO *)Mymalloc(1L, sizeof(CQTINFO), "NewCQTi");
    if(rslt == NULL)
	{  fprintf(stderr, "NewCQTInfo: malloc fail\n");
	   abort();	}
    /* maybe fill in with defaults? */
    rslt->nBins = 0;	rslt->nOcts = 0;	rslt->qFactor = 0.0;
    rslt->kernLen = 0;	rslt->hopSize = 0;	rslt->guardBand = 0.0;
    rslt->totalFrames = PV_UNK_LEN;	rslt->duration = 0.0;
    rslt->flags = 0;	rslt->pvh = NULL;	rslt->pvFile = NULL;
    if(desc != NULL)
	strncpy(rslt->desc, desc, DESCLEN);
    else
	rslt->desc[0] = '\0';
    return rslt;
    }

void FreeCQTInfo(cqi)
    CQTINFO *cqi;
    /* destroy CQT & all it stands for */
    {
    if(cqi->pvFile != NULL)
	fclose(cqi->pvFile);
    if(cqi->pvh != NULL)
	PVFree(cqi->pvh);
    free(cqi);
    }

FILE *OpenCQTPVfile(path, pcqi)
    char *path;
    CQTINFO **pcqi;
    /* attempt to open the file; check it's type; allocate & fill new
       headers if it works */
    {
    PVSTRUCT	*pvh;
    FILE	*file;

    if( (file = PVOpenAllocRdHdr(path, &pvh))==NULL )
	{
	/* PVDie(PVE_NOPEN, path);	*/
	return NULL;
	}
    *pcqi = DecodeCQTPVH(pvh, file, "noname");
    if(*pcqi == NULL)	/* e.g. not right type of file */
	{
	fclose(file);
	PVFree(pvh);
	return NULL;
	}
    else
	return file;
    }

CQTINFO *DecodeCQTPVH(pvh, file, desc)
    PVSTRUCT *pvh;
    FILE *file;
    char *desc;
    /* allocate & fill a new CQTINFO struct from the passed PVSTRUCT */
    /* doesn't access <file>, just puts it in the struct */
    {
    CQTINFO	*rslt;
    char	*info;
    float	tmpflt;
    long	fbSize;
    int 	chans;

    if( pvh->dataFormat != PVFLOAT || pvh->frameFormat != (PVCQ|PVPOLAR) )
	{ fprintf(stderr, "DecodeCQT: PV file is not float,cq,polr(%ld,%ld)\n",
		  pvh->dataFormat, pvh->frameFormat);
	  return NULL;	}
    chans = pvh->channels;
    if(chans != 1)
	fprintf(stderr,"DecodeCQTPV: warning: chans = %d != 1\n", chans);
    /* haven't even thought about how to deal with such a situation */
    rslt = NewCQTInfo(desc);
    info = pvh->info;
    if(GrabInfoArg(info, "Q", &tmpflt))
	rslt->qFactor = tmpflt;
    else	/* no Qfactor info found */
	rslt->qFactor = DFLT_QFCT;
    if(GrabInfoArg(info, "octs", &tmpflt))
	rslt->nOcts   = (int)tmpflt;
    else
	rslt->nOcts   = DFLT_8VES;
    rslt->kernLen   = pvh->frameSize;		/* qFactor*8; */
    rslt->hopSize   = pvh->frameIncr;		/* kernLen/32; */
    rslt->guardBand = pvh->maxFreq / ( pvh->samplingRate / 2.0 );
    if( (int)(.5+pvh->maxFreq/pvh->minFreq) != (1L<<rslt->nOcts) )
	{  fprintf(stderr,"DecodeCQT: octs=%d inconsistent w/ fRatio %5.2f\n",
		   rslt->nOcts, pvh->maxFreq/pvh->minFreq);
	   abort();	}
    fbSize = pvh->frameBsize;
    rslt->nBins = fbSize/sizeof(cpxpolar);
    if(pvh->dataBsize != PV_UNK_LEN)
	{
	rslt->totalFrames = (pvh->dataBsize/fbSize/* /chans */);
	rslt->duration = ((rslt->totalFrames+1)/2 - 1)*
				rslt->hopSize / pvh->samplingRate;
	/* every other frame is top octave; assume 1/2 lost at each end */
	}
    else{
	rslt->totalFrames = PV_UNK_LEN;
	rslt->duration = 0.0;
	}
    rslt->pvh    = pvh;
    rslt->pvFile = file;
    return rslt;
    }

void EncodeCQTPV(cqi, pvh)
    CQTINFO *cqi;
    PVSTRUCT *pvh;
    /* translate details in <cqi> struct to relevent slots in <pvh> */
    /* DOES NOT allocate new pvh, relies on p->chans, sRate already full */
    {
    char *info;

    pvh->dataFormat = PVFLOAT;
    pvh->frameSize  = cqi->kernLen;
    pvh->frameIncr  = cqi->hopSize;
    pvh->maxFreq    = cqi->guardBand * pvh->samplingRate / 2;
    pvh->minFreq    = pvh->maxFreq / (float)(1L<<cqi->nOcts);
    pvh->frameBsize = cqi->nBins * sizeof(cpxpolar);
    pvh->frameFormat= PVCQ|PVPOLAR;
    pvh->freqFormat = PVEXP;
    if(cqi->totalFrames != PV_UNK_LEN)
	pvh->dataBsize = cqi->totalFrames * cqi->nBins * sizeof(cpxpolar);
    else
	pvh->dataBsize = PV_UNK_LEN;
    info = pvh->info;	info[0] = '\0';
    PutInfoArg(info, "Q", cqi->qFactor);	/* it had better be big enuf */
    PutInfoArg(info, "octs", (FLOATARG)cqi->nOcts);
    }

static long ReadFrames(cqi, dstBase, dstFrm, rdFrame, framesToRead)
    CQTINFO *cqi;
    cpxpolar *dstBase;
    long dstFrm; 
    long rdFrame;
    long framesToRead;
    {
    cpxpolar *rdto;
    long     frameChars;
    int      nBins = cqi->nBins;

    frameChars = sizeof(cpxpolar)*nBins;
    rdto = dstBase + dstFrm*nBins;
    fseek( cqi->pvFile, cqi->pvh->headBsize + frameChars*rdFrame, SEEK_SET);
    return fread(rdto, frameChars, framesToRead, cqi->pvFile);
    }

static void CopyFrames(srcBase, srcFrm, dstBase, dstFrm, nFrms, dir, nBins)
    cpxpolar *srcBase;
    long srcFrm;
    cpxpolar *dstBase;
    long dstFrm;
    long nFrms;
    int dir;
    int nBins;
    {
    CopyFloats((float *)(srcBase + nBins*srcFrm), 1, 
	       (float *)(dstBase + nBins*dstFrm), 1, 2*nBins*nFrms, dir);
    }

#define CACHEFRAMES	2560	/* how many frames to cache (maybe) */
    /* CACHEFRAMES too small -> disk thrash on interp. Too big->slow shifts */
#define FRAMESKIP	512	/* how far to hop the cache fwd/back */

static int ReadCQTFileFrame(cqi, frame, dest)
    CQTINFO *cqi;
    long frame;
    cpxpolar *dest;
    /* read a frame of coefs from an aqt file.  May do caching */
    {
    static FILE 	*cacheFile = NULL;
    static cpxpolar 	*cacheBuf  = NULL;
    static long		cacheBufSize = 0L;
    static long 	baseFrame  = -1;	/* cache invalid */
    static long	 	nFrames    = 0;
    static int		cacheAtEof = 0;

    if(cacheBuf == NULL || frame < 0 || cqi->pvFile != cacheFile)
	{	/* negative frame resets the cache */
    	baseFrame = -1;
	nFrames = 0;	cacheAtEof = 0;
	}
    if(cacheBuf == NULL)
	{
	cacheBufSize = cqi->nBins * (long)CACHEFRAMES;
	cacheBuf = (cpxpolar *)Mymalloc(cacheBufSize, sizeof(cpxpolar), 
								"ReadCQT:cacheBuf");
	if(cacheBuf == NULL)
	    {  fprintf(stderr, 
		       "ReadCQTFileFrame: cannot alloc buffer of %ld cpx\n",
		       cacheBufSize);
	       abort(); }
	}
    if(frame < 0)	return 0;		/* reset cache : all done */
    if(cqi->pvFile == cacheFile && baseFrame >= 0)	/* cache is valid */
	{
	long cacheFrames, newBase, skipFrames;

	if(frame>=baseFrame)
	    {
	    if(frame < baseFrame+nFrames)
		{	/* desired frame is in cache */
		CopyFrames(cacheBuf,frame-baseFrame,dest,0,1,FWD,cqi->nBins);
		return 1;
		}
	    else if(cacheAtEof)	/* attempt to read past EOF */
		{
		ClearFloats((float *)dest,1,2*cqi->nBins);
		return 0;
		}
	    }
	/* here if ( frame < baseFrame ) or ( frame >= baseFrame + nFrames ) */
	/* i.e. cache miss */
	cacheFrames = cacheBufSize / cqi->nBins;
	if(frame < baseFrame)
	    {
	    newBase = FRAMESKIP*(int)(frame/FRAMESKIP);
	    /* always in steps of FRAMESKIP */
/*	    newBase = frame;	/* mebbe in steps of FRAMESKIP */
	    skipFrames = baseFrame - newBase;
	    if(skipFrames < cacheFrames /* nFrames */) /* some overlap */
		{
/*		fprintf(stderr, "CACHE: prepend %ld, start %ld, len %ld\n",
			skipFrames, newBase, nFrames);	/* */
		CopyFrames(cacheBuf, 0, cacheBuf, skipFrames,
			   MIN(cacheFrames - skipFrames, nFrames), 
			   FWD, cqi->nBins);
		ReadFrames(cqi, cacheBuf, 0, newBase, skipFrames);
		baseFrame = newBase;
		if(nFrames + skipFrames < cacheFrames)
		    nFrames += skipFrames;
		else{
		    nFrames = cacheFrames;
		    cacheAtEof = 0;	/* definitely moved back from EOF */
		    }
		return ReadCQTFileFrame(cqi, frame, dest);
		}
	    else{	/* fill whole cache */
		ReadCQTFileFrame(cqi, -1, NULL);	/* reset cache */
		return ReadCQTFileFrame(cqi, frame, dest);
		}
	    }
	else{	/* frame past last in buffer, not at EOF */
/*	    newBase = frame - cacheFrames + 1; 	/* frame will be last */
	    newBase = FRAMESKIP*
		(int)((frame - cacheFrames + FRAMESKIP)/FRAMESKIP);
	    /* frame will be last, modulo alignment */
	    if(newBase<0)	newBase = 0;	/* "impossible" */
	    skipFrames = newBase - baseFrame;
	    if(skipFrames < nFrames)
		{
/*		fprintf(stderr, "CACHE: append %ld, start %ld, len %ld\n",
			skipFrames, newBase, nFrames);	/* */
		CopyFrames(cacheBuf, skipFrames, cacheBuf, 0,
			   nFrames-skipFrames, REV, cqi->nBins);
		nFrames -= skipFrames;
		nFrames += ReadFrames(cqi, cacheBuf, nFrames,
				newBase+nFrames, cacheFrames - nFrames);
		cacheAtEof = (nFrames < cacheFrames);
		baseFrame = newBase;
		return ReadCQTFileFrame(cqi, frame, dest);
		}
	    else{ 	/* fill whole cache */
		ReadCQTFileFrame(cqi, -1, NULL);	/* reset cache */
		return ReadCQTFileFrame(cqi, frame, dest);
		}
	    }
	}
    else{	/* read whole new cache */
	long cacheFrames = cacheBufSize / cqi->nBins;
	long newBase = FRAMESKIP*(int)(frame/FRAMESKIP);
/*	long newBase = frame;	*/

	cacheFile = cqi->pvFile;
	nFrames = ReadFrames(cqi, cacheBuf, 0, newBase, cacheFrames);
	cacheAtEof = (nFrames < cacheFrames);
	baseFrame = newBase;
/*	fprintf(stderr,"CACHE: reread %ld from %ld : read %ld, EOF = %d\n", 
		cacheFrames,newBase,nFrames,cacheAtEof);	/* */
	return ReadCQTFileFrame(cqi, frame, dest);
	}
    }

#define LOBOS	8	/* how many points to read for each interpolation */
#define MAXUNDERSAMPL 4 /* undersampling factor allowed for <prd> */

void SpecOfCQTFile(cqi, time, prd, ans)
    CQTINFO *cqi;
    FLOATARG time;
    FLOATARG prd;
    cpxpolar *ans;
    /* using the CQT file described by <cqi>
       interpolate each octave to obtain a complete spectrum at the time
       specified by <time>.  Low pass filter the result where necessary to
       avoid aliasing if the sampling increment (in seconds) is <prd>.
       Put this complex, polar vector into the space
       pointed to by <ans>, assumed large enough */
    {
    /* the idea is that for each octave we will need to read several
       successive frames around the target point to allow successful
       sinc interpolation of the exact value.
       Procedure: * Figure out which frames we need to read for each oct.
                  * Read and store the frames (one oct at a time?)
		  * Interpolate data sets for answer
     */
    /* This implementation is going to read each octave separately.
       This means that the same portion of file is re-read <nOcts>
       times, but reduces the buffer space required for
       frames in RAM */
    int 	oct;
    long	zeroFrameIndex, frmIndexPerFrame, firstFrame, lastFrame;
    long	numFramesThisOct, frame, i;
    float	frameLocation;
    float	timePerFrame;
    cpxpolar	*rsltPtr;
    cpxpolar	*rdPtr;
static cpxpolar	*rdBuffer = NULL;
static long      rdBufPts = 0;
    int 	nBins = cqi->nBins;
    int 	lobes, maxlobes;
    float	lpfRatio;
    float	octFactor, binFactor, lowestFrq, frqInHz, twoPiDt, edp;

#ifdef DEBUGGRAPHS
static    WINDAT	*wdptr[MAXOCTS];
static 	  WINDAT	*reswin;
    char	dummy[16];
    float	dumbuf[MAXBINS];
#endif

    if(rdBuffer == NULL)
	{
	rdBufPts = MAXBINS * LOBOS * MAXUNDERSAMPL;
	rdBuffer = (cpxpolar *)Mymalloc(rdBufPts, sizeof(cpxpolar), 
								"SpecOfCQT");
	
#ifdef DEBUGGRAPHS
	for(i = 0; i<cqi->nOcts; ++i)
	    wdptr[i] = NULL;
#endif
	}

    timePerFrame = cqi->hopSize/cqi->pvh->samplingRate;	/* doubles each 8ve */
    frmIndexPerFrame = 2;				/* doubles each 8ve */
    rsltPtr = ans + (cqi->nOcts-1)*nBins;		/* start w/ top 8ve */
    binFactor = exp((1.0/((float)nBins))*log(2.0));
    lowestFrq = cqi->pvh->minFreq;
    for(oct = 0; oct<cqi->nOcts; ++oct)
	{
	octFactor = (float)(1L<<(cqi->nOcts-1-oct));
	twoPiDt   = 2.0*PI*timePerFrame;
#ifdef DEBUGGRAPHS
	sprintf(dummy, "Oct %d", oct);
	if(wdptr[oct] == NULL)
	    wdptr[oct] = NewWin(dummy, 0);	
#endif
	zeroFrameIndex = (1L<<oct) - 1;
	numFramesThisOct = (cqi->totalFrames+(1L<<oct))/(2L<<oct);
	maxlobes = rdBufPts / nBins;
	if(timePerFrame < prd)	/* undersampled : need to LPF frames on file */
	    {
	    lobes = (LOBOS*prd)/timePerFrame;
	    lpfRatio = prd/(float)timePerFrame;
	    }
	else{
	    lobes = LOBOS;	/* oversampled : use min. num of frames */
	    lpfRatio = (float)1;
	    }
	if(lobes > maxlobes)	lobes = maxlobes;	/* clamp to buf size */

	frameLocation = time/timePerFrame;
	firstFrame = (int)frameLocation - (lobes/2-1);
	lastFrame  = firstFrame + lobes;	/* one after that read */
	
/*	fprintf(stderr,
    "SCQT(%d): zFI %ld, tPF %6.5f, fIpf %d, ftO %ld, 1st %ld, lst %ld\n",
		oct, zeroFrameIndex, timePerFrame, frmIndexPerFrame,
		numFramesThisOct, firstFrame, lastFrame); /*	*/
	if(firstFrame < 0)	firstFrame = 0;
	else if(firstFrame>=numFramesThisOct)  firstFrame = numFramesThisOct-1;
	if(lastFrame <= 0)	lastFrame = 1;
	else if(lastFrame>numFramesThisOct)    lastFrame = numFramesThisOct;
	rdPtr = rdBuffer;
	for(frame = firstFrame; frame < lastFrame; ++frame)
	    {
	    if(!ReadCQTFileFrame(cqi, zeroFrameIndex+frmIndexPerFrame*frame, 
				rdPtr))
		lastFrame = frame;
	    else{
#ifdef DEBUGGRAPHS
		CopyFloats((float *)rdPtr, 2, dumbuf, 1, nBins, 1);
		DoDisp(wdptr[oct], dumbuf, nBins);	
#endif
		rdPtr += nBins;
		}
	    }
	rdPtr = rdBuffer;
#ifdef DEBUGGRAPHS
	if(reswin==NULL)
	    reswin = NewWin("result",1);	
#endif
/*	if(oct == 0)
	    fprintf(stderr, "SpQTF: frames %ld, pos %6.3f, ratio %6.3f\n",
		    lastFrame-firstFrame, frameLocation-firstFrame,lpfRatio);
 */
	SetupInterpVal(lastFrame-firstFrame, frameLocation-firstFrame,
		       lpfRatio, timePerFrame, INTERPTYPE);
	               /* last arg: 1=sinc, 0=noneg (gauss) */
	for(i = 0; i < nBins; ++i)
	    {
	    DoInterpVal((float *)(rdPtr+i), 2*nBins, (float *)(rsltPtr+i)); 
	    }
	SetupInterpVal(lastFrame-firstFrame, frameLocation-firstFrame,
	       (float)1.0 * lpfRatio, timePerFrame, INTERPTYPE); 
	       /* ? lpf for phs */
	frqInHz = octFactor * lowestFrq;
	for(i = 0; i < nBins; ++i)
	    {
	    edp = twoPiDt*frqInHz;
	    UnwrapPhases(((float *)(rdPtr+i))+1, 2*nBins, 
			 ((float *)(rdPtr+i))+1, 2*nBins,
			 lastFrame - firstFrame, edp);
	    DoInterpVal((float *)(rdPtr+i)+1, 2*nBins,(float *)(rsltPtr+i)+1); 
/*	    if(oct==2)
		{
		float	dphi, afrq;
		dphi = rdPtr[i+nBins*(int)(1+frameLocation-firstFrame)].pha 
		       - rdPtr[i+nBins*(int)(frameLocation-firstFrame)].pha;
		afrq = (dphi/timePerFrame) / (2*PI);
		fprintf(stderr,"frq %d = (%.2f) %.2f (%.2f)\n", 
			i,frqInHz,afrq,edp);
		}
 */
	    frqInHz *= binFactor;
	    }
	MaskPhases(((float *)rsltPtr)+1, 2, ((float *)rsltPtr)+1, 2, nBins);

#ifdef DEBUGGRAPHS
	CopyFloats((float *)rsltPtr, 2, dumbuf, 1, nBins, 1);
	DoDisp(reswin, dumbuf, nBins);	
#endif

	rsltPtr -= nBins;		/* move to next lowest octave */
	frmIndexPerFrame <<= 1;
	timePerFrame *= 2;
	zeroFrameIndex = ((zeroFrameIndex+1)<<1)-1;
	}
    }

void MaskPhases(src, sinc, dst, dinc, len)
    float *src;
    int sinc;
    float *dst;
    int dinc;
    long len;
    {
    float ph, pi = PI, twopi = 2*PI, oneonpi = 1.0/PI;
    long  p;

    while(len--)
	{
	ph = *src;
	p = (long)(oneonpi*ph);
	ph -= pi*(float)((int)((p+((p>=0)?(p&1):-(p&1) ))));	
	/* phase masking after dsputil.c's MMmaskPhs() */
	*dst = ph;
	src += sinc;
	dst += dinc;
	}
    }

void UnwrapPhases(src, sinc, dst, dinc, len, expDPhas)
    float *src;
    int sinc;
    float *dst;
    int dinc;
    long len;
    FLOATARG expDPhas;
    /* copy a sequence of phase angles, add/sub 2PI as appro for contin */
    /* src, dst coincident is OK */
    {
    float oldph,ph,eph,offs,delta,fixup,pi = PI, twopi = 2*PI, oneonpi = 1/PI;
    int  p;

/*    fprintf(stderr, "UP:edp %.3f ", expDPhas);	fflush(stderr);	*/
    oldph = *src - expDPhas;	offs = 0.0;
    while(len--)
	{
	ph = offs + *src;
	eph = oldph + expDPhas;
/*	delta = ph - eph;	*/
	delta = eph - ph;	/* so ph + (quantised delta) ~= eph */
/*	if( delta > pi )   { offs -= twopi; ph -= twopi; }	*/
/*	if( delta < -pi )  { offs += twopi; ph += twopi; } 	*/
	/* quantise delta to 2nPI */
	p = (int)(oneonpi*delta);
	fixup = pi*(float)((int)((p+((p>=0)?(p&1):-(p&1) ))));	
	*dst = oldph = ph + fixup ;	offs += fixup;
	/* offs doesn't really matter now .. */
	src += sinc;
	dst += dinc;
	}
    }

#define MAXINTERPVECLEN		550

void InterpolateValue(vals, vinc, nFrames, pos, res, lpfRatio, sincSmooth)
    float *vals;
    int vinc;
    int nFrames;
    FLOATARG pos;
    float *res;
    FLOATARG lpfRatio;
    int sincSmooth;
    /* Sinc interpolate a single value in a vector */
    /* lpfRatio > 1 indicates excess smoothing factor */
    {
    float x, y, ratio, hamwidth, rslt = 0.0;
    int i;

    if(lpfRatio < 1.0)		lpfRatio = 1.0;
    ratio = 1.0 / lpfRatio;

    if(pos > (float)(nFrames-1)/2.0)
	hamwidth = 1.0/(pos);
    else
	hamwidth = 1.0/((float)nFrames - 1.0 - pos);
    /* make the half-width of the hamming window reach the _further_ end */ 
    if(sincSmooth)
	{
	if( pos == (float)((int)pos) && (pos >= 0 && pos < nFrames) 
	   && ratio==1.0)
	    {			/* exact value : just return it */
	    *res = vals[vinc*(int)pos];
	    return;
	    }
	for(i = 0; i<nFrames; ++i)
	    {
	    x = PI*(((float)i) - pos);	y = ratio * x;
	    if(x == 0)		/* can't deal with sin(y)/0 */
		rslt += ratio * (.54 + .46) * vals[i*vinc];
	    else
		rslt += sin(y)/x * (.54+.46*cos(hamwidth*x)) * vals[i*vinc];
	    }
	}
    else	/* gaussSmooth */
	{
	for(i = 0; i<nFrames; ++i)
	    {
	    x = ((float)i)-pos;	y = ratio * x;	/* y is +/-1 at 1/e pts */
	    rslt += .56419 * ratio * exp(-y*y) * vals[i*vinc];
	    /* .56419 = 1/sqrt(pi) */
	    }
	}
    *res = rslt;
    }

/* For display, a faster method is to linearly interpolate rather than 
   all that suffering */

void SimpleSpecOfCQTFile(cqi, time, prd, ans)
    CQTINFO *cqi;
    FLOATARG time;
    FLOATARG prd;
    cpxpolar *ans;
{
    int 	oct;
    long	zeroFrameIndex, frmIndexPerFrame, firstFrame, lastFrame;
    long	numFramesThisOct, frame, i;
    float	frameLocation, frameFraction;
    float	timePerFrame;
    cpxpolar	*rsltPtr;
    cpxpolar	*rdPtr;
static cpxpolar	*rdBuffer = NULL;
static long      rdBufPts = 0;
    int 	nBins = cqi->nBins;
    int 	lobes, maxlobes;
    float	lpfRatio;
    float	octFactor, binFactor, lowestFrq, frqInHz, twoPiDt, edp;

    if(rdBuffer == NULL)  {
	rdBufPts = MAXBINS * 2;
	rdBuffer = (cpxpolar *)Mymalloc(rdBufPts, sizeof(cpxpolar), 
					"SimpleSpecOfCQT");
    }
    timePerFrame = cqi->hopSize/cqi->pvh->samplingRate;	/* doubles each 8ve */
    frmIndexPerFrame = 2;				/* doubles each 8ve */
    rsltPtr = ans + (cqi->nOcts-1)*nBins;		/* start w/ top 8ve */
    binFactor = exp((1.0/((float)nBins))*log(2.0));
    lowestFrq = cqi->pvh->minFreq;
    for(oct = 0; oct<cqi->nOcts; ++oct)  {
	octFactor = (float)(1L<<(cqi->nOcts-1-oct));
	twoPiDt   = 2.0*PI*timePerFrame;
	zeroFrameIndex = (1L<<oct) - 1;
	numFramesThisOct = (cqi->totalFrames+(1L<<oct))/(2L<<oct);
	maxlobes = rdBufPts / nBins;
	frameLocation = time/timePerFrame;
	frame = (long)frameLocation;
	/* fixed (added div timePerFrame) 1993sep29 */
	frameFraction = (time - (float)(frame*timePerFrame))/timePerFrame;
	rdPtr = rdBuffer;
	ReadCQTFileFrame(cqi, zeroFrameIndex+frmIndexPerFrame*frame, rdPtr);
	ReadCQTFileFrame(cqi, zeroFrameIndex+frmIndexPerFrame*(frame+1),
			 rdPtr+nBins);
	for(i = 0; i < nBins; ++i)  {
	    (rsltPtr+i)->mag = (rdPtr+i)->mag 
		               + frameFraction * ((rdPtr+nBins+i)->mag
						  - (rdPtr+i)->mag);
	}
	rsltPtr -= nBins;		/* move to next lowest octave */
	frmIndexPerFrame <<= 1;
	timePerFrame *= 2;
	zeroFrameIndex = ((zeroFrameIndex+1)<<1)-1;
    }
}

static float *interpBuf = NULL;
static int   interpBufFill = 0;

void SetupInterpVal(nFrames, pos, lpfRatio, dt, sinc)
    int nFrames;
    FLOATARG pos;
    FLOATARG lpfRatio;
    FLOATARG dt;	/* sampling period - only for DDIFF mode */
    int sinc;
    {	/* setup static data. <sinc> is flag to interpolate or just smooth */
    float x, y, ratio, hamwidth, scale;
    int i;

    if(interpBuf == NULL)
	{
	interpBuf = (float *)Mymalloc((long)MAXINTERPVECLEN, sizeof(float),
								"SetupInterp");
	if(interpBuf == NULL)
	    {  fprintf(stderr,"SetupInterp: malloc fail\n"); abort();	}
	}

    if(nFrames > MAXINTERPVECLEN)	
	{  fprintf(stderr,"SetupInterp: frames %d > %d\n",
		   nFrames, MAXINTERPVECLEN);
	   /* abort(); */
	   nFrames = MAXINTERPVECLEN;  }

    if(lpfRatio < 1.0)		lpfRatio = 1.0;
    ratio = 1.0 / lpfRatio;
    
    if(sinc==SINC)
	{
	if(pos > (float)(nFrames-1)/2.0)
	    hamwidth = 1.0/(pos);
	else
	    hamwidth = 1.0/((float)nFrames - 1.0 - pos);
	/* make the half-width of the ham. window reach the _further_ end */ 
	for(i = 0; i<nFrames; ++i)
	    {	
	    x = PI*(((float)i) - pos);	y = ratio * x;
	    if(x == 0)	/* can't deal with sin(y)/0 */
		interpBuf[i] = ratio * (.54 + .46);
	    else
		interpBuf[i] = sin(y)/x * (.54+.46*cos(hamwidth*x));
	    }
	}
    else if(sinc==GAUSS) { /* gauss = exp(-x^2) to smooth without ripples */
	for(i=0; i<nFrames; ++i)
	    {
	    x = ((float)i)-pos;	y = ratio * x;	/* y is +/-1 at 1/e pts */
	    interpBuf[i] = .56419 * ratio * exp(-y*y);
	    /* .56419 = 1/sqrt(pi) */

/*	    x = (float)i - pos;
/*	    interpBuf[i] = 1/lpfRatio * exp(-x*x*ratio);	*/
	    }
	}
    else{ /* sinc == DDIFF , discrete differentiator for phi-dot */
	if(pos == 0.0 || pos == (float)nFrames - 1.0)	{	
	    /* data at extreme end : lose */
	    for(i = 0; i<nFrames; ++i)
		interpBuf[i] = 0.0;		/* all zero! */
	    }
	else {
	    if(pos < (float)(nFrames-1)/2.0)
		hamwidth = 1.0/pos;
	    else
		hamwidth = 1.0/((float)nFrames - 1.0 - pos);
	    /* make the half-width of the ham. window reach the _NEARER_ end */
	    /* this throws away data beyond the shorter boundary.  But the 
	      alternative doesn't work very well for the differentiator, since 
	      it is effectively zero-padding the shorter extent of the series, 
	      and differentiation deals very unfavourably with any 
	      discontinuities this may introduce */
	    /*	fprintf(stderr,"DDIFF:nFrames %d hmw %f pos %f\n", 
		nFrames, hamwidth, pos); */
	    scale = 0;
	    for(i = 0; i<nFrames; ++i)
		{		/* different x and y from above */
		x = (((float)i) - pos);	y = ratio * x;
		if(x == 0 || fabs(x*hamwidth)>1.0 )	
		    /* can't deal with sin(y)/0, also using shorter window */
		    interpBuf[i] = 0 /* ratio * (.54 + .46) */;
		else
		    interpBuf[i] = ((cos(PI*y)/(-y)) - (sin(-PI*y)/(PI*y*y)))
			* (.5+.5*cos(hamwidth*PI*x));
	        /* have to use HANN window to avoid sharp edges of HAMM 
		   messing us up for short, variable windows */
		scale += x * interpBuf[i];
		}
	    if(dt*scale != 0.0)
		{
		scale = 1/(dt*scale);
		for(i=0; i<nFrames; ++i)
		  interpBuf[i] *= scale; /* correct weighting, scale for dt */ 
		}
	    }
#ifdef DEBUGGRAPHS
	if(0)
	    {
	    WINDAT *wdptr;
	    wdptr = NewWin("DDIFF",1);	
	    DoDisp(wdptr, interpBuf, nFrames);	
	    }
#endif
	}
	
    interpBufFill = nFrames;
    }

void DoInterpVal(vals, vinc, res)
    float *vals;
    int vinc;
    float *res;
    {
    register float rslt = 0;
    register float *pcoef = interpBuf;
    register float *pvec  = vals;
    register int nFrames = interpBufFill;

    while(nFrames--)
	{
	rslt += *pvec * *pcoef++;
	pvec += vinc;
	}
    *res = rslt;
    }

/* static */ void print_vec(vec, num, tit)
    float *vec;
    long num;
    char *tit;
    {
    fprintf(stderr,"%s:", tit);
    while(num--)
	fprintf(stderr,"%8.3f",*vec++);
    fprintf(stderr,"\n");
    }

/** test code only below here **/
#ifdef DEBUG

#define VECLEN 32L
#define BIGVEC 400L
static test_phase()
    {
    float vec[VECLEN];
    float r[BIGVEC];
    long i,j;
    float f;
#ifdef DEBUGGRAPHS
    WINDAT *wdptr;
#endif

    for(i=0; i<VECLEN/2; ++i)
	vec[i] = (float)i/2.0;
    for(i = VECLEN/2; i<VECLEN; ++i)
	vec[i] = (float)VECLEN/4.0; /* - (i-VECLEN/2.0);	*/
/*    for(i = 7*VECLEN/8; i<VECLEN; ++i)
	vec[i] = (float)2*(VECLEN-i); /* (float)VECLEN/4.0; /* - (i-VECLEN/2.0);	*/
    print_vec(vec, VECLEN, "start");
    MaskPhases(vec,1,vec,1,VECLEN);
    print_vec(vec, VECLEN, "masked");
    UnwrapPhases(vec,1,vec,1,VECLEN,(float)0);
    print_vec(vec, VECLEN, "unwrapd");

    for(i = 0; i<BIGVEC; ++i)
	{
	SetupInterpVal(VECLEN, VECLEN/2.0-10.0+(float)i/20.0, 2.0, 0.1, 
		       DDIFF /* INTERPTYPE */);
	DoInterpVal(vec,1,&r[i]);
/*	fprintf(stderr, "%8.4f", r[i]); fflush(stderr);	*/
/*	InterpolateValue(vec,1,VECLEN,VECLEN/2.0-10.0+(float)i/20.0,&r[i],
			 1.0,0);	*/
	}
#ifdef DEBUGGRAPHS
    wdptr = NewWin("test interp",1);
    DoDisp(wdptr, r, BIGVEC);
#endif
    for(i = 0; i<BIGVEC; ++i)
	{
	SetupInterpVal(VECLEN, VECLEN/2.0-10.0+(float)i/20.0, 1.0, 1.0,
		       DDIFF /* INTERPTYPE */ );
	DoInterpVal(vec,1,&r[i]);
/*	fprintf(stderr, "%8.4f", r[i]); fflush(stderr);	*/
/*	InterpolateValue(vec,1,VECLEN,VECLEN/2.0-10.0+(float)i/20.0,&r[i],
			 1.0,1);	*/
	}
#ifdef DEBUGGRAPHS
    wdptr = NewWin("test interp 2",1);
    DoDisp(wdptr, r, BIGVEC);
#endif
    }

#define BUFLEN (6*24)
static test_cqint()
    {
    FILE *file;
    CQTINFO *cqi;
    cpxpolar	rslt[BUFLEN];
    float	res2[BUFLEN];
#ifdef DEBUGGRAPHS
    WINDAT	*wdptr;
#endif
    int		i;

    ClearFloats((float *)rslt,1,2*BUFLEN);
    ClearFloats(res2,1,BUFLEN);
#ifdef DEBUGGRAPHS
    wdptr = NewWin("cqint",1);
#endif
    file = OpenCQTPVfile("../sound/spoil.22k.aqt", &cqi);
    for(i = 2; i>=0; --i)
	{
	SpecOfCQTFile(cqi, .4 + .04*(float)i, .04, rslt);
	CopySampls((float *)rslt,2,res2,1,BUFLEN,1);
#ifdef DEBUGGRAPHS
	DoDisp(wdptr, res2, BUFLEN);
#endif
	}
    }

void test_ddiff()
    {
    SetupInterpVal(100, (float)50, (float)10, .1, DDIFF);
    }

char *programName;
main(argc, argv)
    int argc;
    char **argv;
    {
    programName = argv[0];
/*    test_cqint();	*/
/*    test_ddiff();	*/
    test_phase();
    ExitGraph();
    }

#define BUFLEN2 (6*24)

void TestRead(filename, ppArray, pTmin, pTmax, pFmin, pFmax)
    char	*filename;
    SampArray	**ppArray;
    float	*pTmin;
    float 	*pTmax;
    float	*pFmin;
    float 	*pFmax;
    {
    FILE *file;
    CQTINFO *cqi;
    int rows, cols, i;
    float *fPtr, tmpBuf[2*BUFLEN2];

    if( (file=OpenCQTPVfile(filename, &cqi))==NULL)
	exit(1);
    rows = cqi->nBins * cqi->nOcts;
    cols = 40;

    (*pTmin) = 0.0;			(*pTmax) = cqi->duration;
    *pTmin = .40 * cqi->duration; 	*pTmax = .60*cqi->duration;	/* */
    (*pFmin) = cqi->pvh->minFreq;	(*pFmax) = cqi->pvh->maxFreq;
    *ppArray = NewSampA(rows, cols);
    fPtr = (*ppArray)->data;
    for(i=0; i<cols; ++i)
	{
/*	fprintf(stderr,"%8d",i);	fflush(stderr);	*/
	SpecOfCQTFile(cqi, *pTmin + (float)i/(float)cols*(*pTmax-*pTmin), 
		      1/(float)cols*(*pTmax-*pTmin), (cpxpolar *)tmpBuf);
	CopyFloats(tmpBuf, 2, fPtr+cols*(rows-1), -cols, rows, 1);
	++fPtr;
	}
    }
/** end of test code **/

#endif /* def DEBUG */

