/***********************************************************************\
*	readaqt.h							*
*  functions to extract useful values from aqt files (new CQT)		*
*  dpwe 11mar91								*
\***********************************************************************/

#ifndef _READAQT_H_
#define _READAQT_H_

#include <stdio.h>	/* for FILE, stderr */
#include <pvoc.h>	/* PVSTRUCTs here, cpxpolar */

#define	DFLT_8VES	6	/* number of analysis octaves */
#define	DFLT_NBIN	12	/* number of filter bins per octave */
#define	DFLT_QFCT	16	/* default Q factor per bin */
#define MAXOCTS		20	/* limit to number of octaves */
#define MAXBINS		100	/* limit to bins/octave */

#ifndef DESCLEN
#define DESCLEN	16
#endif

typedef struct cqtinfo	/* pertinent info for CQT files */
    {
    int 	nBins;
    int 	nOcts;
    float	qFactor;
    long	kernLen;
    long	hopSize;
    float	guardBand;	/* gap above top octave */
    long	totalFrames;	/* in FILE, if known */
    float	duration;
    long	flags;		/* various format info */
    PVSTRUCT	*pvh;
    FILE	*pvFile;	/* might be handy */
    char	desc[DESCLEN];	/* tag for debug */
    } CQTINFO;

extern char *programName;

#ifdef __STDC__
void PutInfoArg(char *str, char *title, FLOATARG arg);
int GrabInfoArg(char *str, char *title, float *ans);
int OctOfFileFrame(long frame); /* return which 8ve the <frame>th frame is */
CQTINFO *NewCQTInfo(char *desc); /* allocate the space */
void FreeCQTInfo(CQTINFO *cqi);
FILE *OpenCQTPVfile(char *path, CQTINFO **pcqi);
    /* attempt to open <path>; check type; allocate & fill new hdrs if OK */
CQTINFO *DecodeCQTPVH(PVSTRUCT *pvh, FILE *file, char *desc);
    /* allocate & fill a new CQTINFO struct from the passed PVSTRUCT */
    /* doesn't access <file>, just puts it in the struct */
void EncodeCQTPV(CQTINFO *cqi, PVSTRUCT *pvh);
    /* translate details in <cqi> struct to relevent slots in <pvh> */
    /* DOES NOT allocate new pvh, relies on p->chans, sRate already full */
void SpecOfCQTFile(CQTINFO *cqi, FLOATARG time, FLOATARG prd, cpxpolar *ans);
    /* using the CQT file described by <pvh> and assumed attached to <cqFile>,
       interpolate each octave to obtain a complete spectrum at the time
       specified by time.  Put this complex, polar vector into the space
       pointed to by <ans>, assumed large enough */
void MaskPhases(float *src, int sinc, float *dst, int dinc, long len);
void UnwrapPhases(float *src, int sinc, float *dst, int dinc, long len, 
		  FLOATARG expDph);
void InterpolateValue(float *vals, int vinc, int nFrms, FLOATARG pos,
		      float *res, FLOATARG ratio, int specNgauss);
void SetupInterpVal(int nFrames, FLOATARG pos, FLOATARG lpfRatio, 
		    FLOATARG dt, int sinc);
void DoInterpVal(float *vals,int vinc,float *res);
void print_vec(float *vec, long num, char *tit);

#else
void PutInfoArg();
int GrabInfoArg();
int OctOfFileFrame();
CQTINFO *NewCQTInfo();
void FreeCQTInfo();
FILE *OpenCQTPVfile();
CQTINFO *DecodeCQTPVH();
void EncodeCQTPV();
void SpecOfCQTFile();
void MaskPhases();
void UnwrapPhases();
void InterpolateValue();
void SetupInterpVal();
void DoInterpVal();
void print_vec();

#endif

#endif /* _READAQT_H_ */

