/***********************************************************************\
*	readfrms.c							*
*  Abstraction layer for reading data files with many fixed-length 	*
*  frames (such as short-time spectral analysis files).  The application*
*  collects one frame at a time, but this layer caches a number of 	*
*  frames in memory to avoid excessive disk access.			*
*  dpwe 08sep91 after readaqt.c, 11mar91				*
\***********************************************************************/

*** this file is going to be an abstraction of the separate fourier 
    frame file caches in readaft.c and readaqt.c, but I lost steam.
    Finish some other time - should be pretty easy.  Still need to 
    typedef the private data struct referred in Setup.. 
    and create readfrms.h

    dpwe 08sep91

/************** switches *****************/

/************** includes *****************/
#include <dpwelib.h>	/* for FILE, stderr */
#include <ctype.h>
#include <readfrms.h>
#include <samplops.h>   /* e.g. CopyFloats */

/************* constants *****************/
#define DFLT_CACHE_SIZE 	64
#define DFLT_CACHE_SKFC		 8	/* by deflt, increment cache by 1/8 */

/********** static fn protos *************/

/*********** static globals **************/


long ReadFrames(file, rdFrame, dstBase, dstFrm, framesToRead, frmSize)
    FILE *file;		/* file to read from */
    long hedSize;	/* bytes to skip at front of file */
    long rdFrame;	/* which file frame to start reading at */
    char *dstBase;	/* base of memory to be read into */
    long dstFrm; 	/* index of frame offset from base to read to */
    long framesToRead;	/* number of frames to read */
    long frmSize;	/* bytes per frame */
    {	/* actually read frame(s) off disk into cache. Returns how many */
    char     *rdto;
    long     frameChars, frameCpxs;

    rdto = dstBase + dstFrm*frmSize;
    fseek(file, hedSize + frmSize*rdFrame, SEEK_SET);
    return fread(rdto, frmSize, framesToRead, file);
    }

void CopyFrames(srcBase, srcFrm, dstBase, dstFrm, nFrms, dir, frmSize)
    char *srcBase;
    long srcFrm;
    char *dstBase;
    long dstFrm;
    long nFrms;
    int  dir;
    long frmSize;
    {
    CopyChars(srcBase + frmSize*srcFrm, 1, 
	      dstBase + frmSize*dstFrm, 1, frmSize*nFrms, dir);
    }

RDFRMF *SetupFrameRead(file, headSize, frameSize, cacheSize, cacheStep)
    {	/* instance a new one of our private str, alloc cache memory */
    RDFRMF *rdfrmf;
    long   bytes_req;

    if(cacheSize <= 0)
	cacheSize = DFLT_CACHE_FRMS;
    if(cacheStep <= 0)
	cacheStep = cacheSize / DFLT_CACHE_SKFC;
    if(frameSize <= 0)
	{  fprintf(stderr, "SetupFrameRead: frameSize %d <= 0\n", frameSize);
	   abort();	}
    bytes_req = sizeof(RDFRMF) + frameSize*cacheSize;
    rdfrmf = (RDFRMF *)malloc(bytes_req);
    if(rdfrmf == NULL)
	{ fprintf(stderr, "SetupFrameRead: malloc of %ld failed\n", bytes_req);
	  abort();	}
    rdfrmf->buf = (char *)(rdfrmf + 1);		/* start of 'next' header.. */
    rdfrmf->file = file;
    rdfrmf->headSize = headSize;
    rdfrmf->frmSize  = frmSize;
    rdfrmf->cacheSize = cacheSize;
    rdfrmf->cacheStep = cacheStep;
    }

** not done below here

static int GetFileFrame(afi, frame, dest)
    AFTINFO *afi;
    long frame;
    cpxpolar *dest;
    /* read a frame of coefs from an aft file.  May do caching */
    {
    static FILE 	*cacheFile = NULL;
    static cpxpolar 	*cacheBuf  = NULL;
    static long		cacheBufSize = 0L;
    static long 	baseFrame  = -1;	/* cache invalid */
    static long	 	nFrames    = 0;
    static int		cacheAtEof = 0;
    int frCpxs = afi->pvh->frameBsize / sizeof(cpxpolar);

    if(cacheBuf == NULL || frame < 0 || afi->pvFile != cacheFile)
	{	/* negative frame resets the cache */
    	baseFrame = -1;
	nFrames = 0;	cacheAtEof = 0;
	}
    if(cacheBuf == NULL)
	{
	cacheBufSize = frCpxs*(long)CACHEFRAMES;
	cacheBuf = (cpxpolar *)malloc(cacheBufSize*sizeof(cpxpolar));
	if(cacheBuf == NULL)
	    {  fprintf(stderr, 
		       "ReadAFTFileFrame: cannot alloc buffer of %ld cpx\n",
		       cacheBufSize);
	       abort(); }
	}
    if(frame < 0)	return 0;		/* reset cache : all done */
    if(afi->pvFile == cacheFile && baseFrame >= 0)	/* cache is valid */
	{
	long cacheFrames, newBase, skipFrames;

	if(frame>=baseFrame)
	    {
	    if(frame < baseFrame+nFrames)
		{	/* desired frame is in cache */
		CopyAFTFrames(cacheBuf,frame-baseFrame,dest,0,1,FWD,frCpxs);
		return 1;
		}
	    else if(cacheAtEof)	/* attempt to read past EOF */
		{
		ClearFloats((float *)dest,1,2*frCpxs);
		return 0;
		}
	    }
	/* here if ( frame < baseFrame ) or ( frame >= baseFrame + nFrames ) */
	/* i.e. cache miss */
	cacheFrames = cacheBufSize / frCpxs;
	if(frame < baseFrame)
	    {
	    newBase = FRAMESKIP*(int)(frame/FRAMESKIP);
	    /* always in steps of FRAMESKIP */
	    skipFrames = baseFrame - newBase;
	    if(skipFrames < cacheFrames /* nFrames */ ) /* some overlap */
		{
		fprintf(stderr, "CACHE: prepend %ld, start %ld, len %ld\n",
			skipFrames, newBase, nFrames);	/* */
		CopyAFTFrames(cacheBuf, 0, cacheBuf, skipFrames,
			      MIN(cacheFrames - skipFrames, nFrames), 
			      FWD, frCpxs);
		ReadAFTFrames(afi, cacheBuf, 0, newBase, skipFrames);
		baseFrame = newBase;
		if(nFrames + skipFrames < cacheFrames)
		    nFrames += skipFrames;
		else{
		    nFrames = cacheFrames;
		    cacheAtEof = 0;	/* definitely moved back from EOF */
		    }
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    else{	/* fill whole cache */
		ReadAFTFileFrame(afi, -1, NULL);	/* reset cache */
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    }
	else{	/* frame past last in buffer, not at EOF */
	    newBase = FRAMESKIP*
		(int)((frame - cacheFrames + FRAMESKIP)/FRAMESKIP);
	    /* frame will be last, modulo alignment */
	    if(newBase<0)	newBase = 0;	/* "impossible" */
	    skipFrames = newBase - baseFrame;
	    if(skipFrames < nFrames)
		{
		fprintf(stderr, "CACHE: append %ld, start %ld, len %ld\n",
			skipFrames, newBase, nFrames);	/* */
		CopyAFTFrames(cacheBuf, skipFrames, cacheBuf, 0,
			      nFrames-skipFrames, REV, frCpxs);
		nFrames -= skipFrames;
		nFrames += ReadAFTFrames(afi, cacheBuf, nFrames,
					 newBase+nFrames, cacheFrames-nFrames);
		cacheAtEof = (nFrames < cacheFrames);
		baseFrame = newBase;
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    else{ 	/* fill whole cache */
		ReadAFTFileFrame(afi, -1, NULL);	/* reset cache */
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    }
	}
    else{	/* read whole new cache */
	long cacheFrames = cacheBufSize / frCpxs;
	long newBase = FRAMESKIP*(int)(frame/FRAMESKIP);
/*	long newBase = frame;	*/

	cacheFile = afi->pvFile;
	nFrames = ReadAFTFrames(afi, cacheBuf, 0, newBase, cacheFrames);
	cacheAtEof = (nFrames < cacheFrames);
	baseFrame = newBase;
	fprintf(stderr,"CACHE: reread %ld from %ld : read %ld, EOF = %d\n", 
		cacheFrames,newBase,nFrames,cacheAtEof);	/* */
	return ReadAFTFileFrame(afi, frame, dest);
	}
    }

#define LOBOS	8	/* how many points to read for each interpolation */
#define MAXUNDERSAMPL 4 /* undersampling factor allowed for <prd> */

void SpecOfAFTFile(afi, time, prd, ans, dphasq)
    AFTINFO *afi;
    FLOATARG time;
    FLOATARG prd;
    cpxpolar *ans;
    int dphasq;		/* if nz, return phi-dot, not phase */
    /* using the AFT file described by <afi>
       interpolate frames to obtain a spectrum at the exact time
       specified by <time>.  Low pass filter the result where necessary to
       avoid aliasing if the sampling increment (in seconds) is <prd>.
       Put this complex, polar vector into the space
       pointed to by <ans>, assumed large enough.
       Differentiate phases if dphasq != 0 */
    {
    long	zeroFrameIndex, frmIndexPerFrame, firstFrame, lastFrame;
    long	numFrames, frame, i;
    float	frameLocation;
    float	timePerFrame;
    cpxpolar	*rsltPtr;
    cpxpolar	*rdPtr;
static cpxpolar	*rdBuffer = NULL;
static int      rdBufPts = 0;
    int 	lobes, maxlobes;
    float	lpfRatio;
    float	octFactor, binFactor, lowestFrq, frqInHz, twoPiDt, edp;
    int 	frCpxs = afi->pvh->frameBsize / sizeof(cpxpolar);

#ifdef DEBUGGRAPHS
static    WINDAT	*wdptr;
static 	  WINDAT	*reswin;
    char	dummy[16];
    float	dumbuf[MAXFCPX];
#endif

    if(rdBuffer == NULL)
	{
	rdBufPts = MAXFCPX * LOBOS * MAXUNDERSAMPL;
	rdBuffer = (cpxpolar *)malloc(sizeof(cpxpolar) * rdBufPts);
	
#ifdef DEBUGGRAPHS
	wdptr = NULL;
#endif
	}

    timePerFrame = afi->pvh->frameIncr/afi->pvh->samplingRate;
    twoPiDt   = 2.0*PI*timePerFrame;
#ifdef DEBUGGRAPHS
    sprintf(dummy,"as read");
    if(wdptr == NULL)
	wdptr = NewWin(dummy, 0);	
#endif
    zeroFrameIndex = 0;
    frmIndexPerFrame = 1;
    rsltPtr = ans;
    numFrames      = afi->totalFrames;
    maxlobes = rdBufPts / frCpxs;
    if(timePerFrame < prd)	/* undersampled : need to LPF frames on file */
	{
	lobes = (LOBOS*prd)/timePerFrame;
	lpfRatio = prd/(float)timePerFrame;
	}
    else{
	lobes = LOBOS;	/* oversampled : use min. num of frames */
	lpfRatio = (float)1;
	}
    if(lobes > maxlobes)	lobes = maxlobes;	/* clamp to buf size */
    
    frameLocation = time/timePerFrame;
    firstFrame = (int)frameLocation - (lobes/2-1);
    lastFrame  = firstFrame + lobes;	/* one after that read */
	
/*    fprintf(stderr,
            "SAFT: zFI %ld, tPF %6.5f, frs %ld, 1st %ld, lst %ld\n",
            zeroFrameIndex, timePerFrame,
            numFrames, firstFrame, lastFrame); /*	*/
    if(firstFrame < 0)	firstFrame = 0;
    else if(firstFrame>=numFrames)  firstFrame = numFrames-1;
    if(lastFrame <= 0)	lastFrame = 1;
    else if(lastFrame>numFrames)    lastFrame = numFrames;
    rdPtr = rdBuffer;
    for(frame = firstFrame; frame < lastFrame; ++frame)
	{
	if(!ReadAFTFileFrame(afi, zeroFrameIndex+frmIndexPerFrame*frame, 
			     rdPtr))
	    lastFrame = frame;
	else{
#ifdef DEBUGGRAPHS
	    CopyFloats((float *)rdPtr, 2, dumbuf, 1, frCpxs, 1);
	    DoDisp(wdptr, dumbuf, frCpxs);	
#endif
	    rdPtr += frCpxs;
	    }
	}
    rdPtr = rdBuffer;
#ifdef DEBUGGRAPHS
    if(reswin==NULL)
	reswin = NewWin("result",1);	
#endif	

/*    fprintf(stderr, "SpFTF: frames %ld, pos %6.3f, ratio %6.3f\n",
            lastFrame-firstFrame, frameLocation-firstFrame,lpfRatio);
  /* */
    SetupInterpVal(lastFrame-firstFrame, frameLocation-firstFrame,
		   lpfRatio, timePerFrame, INTERPTYPE);	
                   /* last arg: 1=sinc, 0=noneg */
    for(i = 0; i < frCpxs; ++i)
	{
	DoInterpVal((float *)(rdPtr+i), 2*frCpxs, (float *)(rsltPtr+i)); 
	}
    SetupInterpVal(lastFrame-firstFrame, frameLocation-firstFrame,
		   (float)1.0 * lpfRatio, timePerFrame, 
		   dphasq?DDIFF:INTERPTYPE);
    /* make interpolator do differentiation too for phase if requested */
    for(i = 0; i < frCpxs; ++i)
	{
	frqInHz = i * afi->pvh->samplingRate / (float)afi->pvh->frameSize;
	edp = twoPiDt*frqInHz;
	UnwrapPhases(((float *)(rdPtr+i))+1, 2*frCpxs, 
		     ((float *)(rdPtr+i))+1, 2*frCpxs,
		     lastFrame - firstFrame, edp);
	DoInterpVal((float *)(rdPtr+i)+1, 2*frCpxs,(float *)(rsltPtr+i)+1); 
	if(0/*debug*/)
	    {
	    float	dphi, afrq;
	    dphi = rdPtr[i+frCpxs*(int)(1+frameLocation-firstFrame)].pha 
	           - rdPtr[i+frCpxs*(int)(frameLocation-firstFrame)].pha;
	    dphi += 2.0*PI*(int)((edp-dphi+PI)/(2.0*PI));
	    afrq = (dphi/timePerFrame) / (2*PI);
	    fprintf(stderr,"frq %d = (%.2f) %.2f (%.2f) -> %.2f\n", 
		    i,frqInHz,afrq,edp,rsltPtr[i].pha/(2*PI));
#ifdef DEBUGGRAPHS
	    CopyFloats(((float *)(rdPtr+i))+1, 2*frCpxs, dumbuf, 1, 
		       lastFrame-firstFrame, 1);
	    DoDisp(reswin, dumbuf, lastFrame-firstFrame);	
#endif
	    }

	}
    if(!dphasq)		/* only mask actual phases, not derivatives */
	MaskPhases(((float *)rsltPtr)+1, 2, ((float *)rsltPtr)+1, 2, frCpxs);
    
#ifdef DEBUGGRAPHS
    CopyFloats((float *)rsltPtr, 2, dumbuf, 1, frCpxs, 1);
    DoDisp(reswin, dumbuf, frCpxs);	
#endif

    }

/** test code only below here **/
#ifdef DEBUG

#define VECLEN 32L
#define BIGVEC 400L
static test_phase()
    {
    float vec[VECLEN];
    float r[BIGVEC];
    long i,j;
    float f;
#ifdef DEBUGGRAPHS
    WINDAT *wdptr;
#endif

    for(i=0; i<VECLEN/2; ++i)
	vec[i] = (float)i/2.0;
    for(i = VECLEN/2; i<VECLEN; ++i)
	vec[i] = (float)VECLEN/4.0; /* - (i-VECLEN/2.0);	*/
    print_vec(vec, VECLEN, "start");
    MaskPhases(vec,1,vec,1,VECLEN);
    print_vec(vec, VECLEN, "masked");
    UnwrapPhases(vec,1,vec,1,VECLEN,(float)0);
    print_vec(vec, VECLEN, "unwrapd");

    for(i = 0; i<BIGVEC; ++i)
	{
	SetupInterpVal(VECLEN, VECLEN/2.0-10.0+(float)i/20.0, 2.0, 1.0,
		       INTERPTYPE);
	DoInterpVal(vec,1,&r[i]);

/*	InterpolateValue(vec,1,VECLEN,VECLEN/2.0-10.0+(float)i/20.0,&r[i],
			 1.0,0);	*/
	}
#ifdef DEBUGGRAPHS
    wdptr = NewWin("test interp",1);
    DoDisp(wdptr, r, BIGVEC);
#endif
    for(i = 0; i<BIGVEC; ++i)
	{
	SetupInterpVal(VECLEN, VECLEN/2.0-10.0+(float)i/20.0, 1.0, 0.1, 
		       INTERPTYPE);
	DoInterpVal(vec,1,&r[i]);
/*	InterpolateValue(vec,1,VECLEN,VECLEN/2.0-10.0+(float)i/20.0,&r[i],
			 1.0,1);	*/
	}
#ifdef DEBUGGRAPHS
    wdptr = NewWin("test interp 2",1);
    DoDisp(wdptr, r, BIGVEC);
#endif
    }

#define BUFLEN (257)	/* number of bins ? */

static test_afint()
    {
    FILE *file;
    AFTINFO *afi;
    cpxpolar	rslt[BUFLEN];
    float	res2[BUFLEN];
#ifdef DEBUGGRAPHS
    WINDAT	*wdptr;
#endif
    int		i,j;

    ClearFloats((float *)rslt,1,2*BUFLEN);
    ClearFloats(res2,1,BUFLEN);
#ifdef DEBUGGRAPHS
    wdptr = NewWin("afint",1);
#endif
    file = OpenAFTPVfile("/usr/local/sound/test/sins.aft", &afi);
    for(i = 2; i>=0; --i)
	{
	SpecOfAFTFile(afi, .4 + .04*(float)i, .04, rslt, 1);
	for(j=0; j<BUFLEN; ++j)
	    {
	    fprintf(stderr, "j %d phidot %f frq %f\n", j, rslt[j].pha,
		    rslt[j].pha/(2*PI));
	    }
	CopyFloats((float *)rslt,2,res2,1,BUFLEN,1);
#ifdef DEBUGGRAPHS
	DoDisp(wdptr, res2, BUFLEN);
#endif
	}
    }

char *programName;
main(argc, argv)
    int argc;
    char **argv;
    {
    programName = argv[0];
    test_afint();
    ExitGraph();
    }


#define BUFLEN2 (6*24)

void TestRead(filename, ppArray, pTmin, pTmax, pFmin, pFmax)
    char	*filename;
    SampArray	**ppArray;
    float	*pTmin;
    float 	*pTmax;
    float	*pFmin;
    float 	*pFmax;
    {
    FILE *file;
    AFTINFO *afi;
    int rows, cols, i;
    float *fPtr, tmpBuf[2*BUFLEN2];

    if( (file=OpenAFTPVfile(filename, &afi))==NULL)
	exit(1);
    rows = afi->pvh->frameBsize/sizeof(cpxpolar);
    cols = 40;

    (*pTmin) = 0.0;			(*pTmax) = afi->duration;
    *pTmin = .40 * afi->duration; 	*pTmax = .60*afi->duration;	/* */
    (*pFmin) = 0;	(*pFmax) = afi->pvh->samplingRate/2.0;
    *ppArray = NewSampA(rows, cols);
    fPtr = (*ppArray)->data;
    for(i=0; i<cols; ++i)
	{
/*	fprintf(stderr,"%8d",i);	fflush(stderr);	*/
	SpecOfAFTFile(afi, *pTmin + (float)i/(float)cols*(*pTmax-*pTmin), 
		      1/(float)cols*(*pTmax-*pTmin), (cpxpolar *)tmpBuf, 1);
	CopyFloats(tmpBuf, 2, fPtr+cols*(rows-1), -cols, rows, 1);
	++fPtr;
	}
    }
/** end of test code **/

#endif /* def DEBUG */
