/***************************************************************\
*	resample.h  (after diffuse.c)				*
* A kind of integer ratio resampler				*
* that takes input a bit at a time - for resmp.c		*
* dpwe 23feb91							*
\***************************************************************/

#ifndef _RESAMPLE_H_
#define _RESAMPLE_H_

#include <samplops.h>

typedef struct
    {
    SAMPL *kernCtr;	/* pointer to _middle_ term of buf: +/- khl is ok */
    int   kernhalflen;	/* number of symmetric terms : len = 2*khl + 1 */
    int   inPrd;
    int   outPrd;
    } RsmpDescType;	/* private data for resampler */

#ifndef PI
#ifdef M_PI
#define PI M_PI
#else
#define PI 3.1415926535
#endif
#endif

/*** yukky ANSI bodge (please tell me the right way!) ****/
#ifndef FLOATARG
#ifdef __STDC__
#define FLOATARG double		/* you can't have floats in prototypes! */
#else
#define FLOATARG double		/* sometimes float .. */
#endif /* def __STDC__ */
#endif /* ndef FLOATARG */

#ifdef __STDC__
void FindGoodRatio(FLOATARG ratio, int *p_o, int *p_i, FLOATARG acc);
    /* find an integer ratio *p_o/*p_i closest to 'ratio', with *p_i
       smaller than its entry value.  Returns immediately error
       is less than acc or 0.001 (.1%)		*/

long MinResampleBufLen(RsmpDescType *rdp,long holdSize, long feedSize);
    /* return a minimum size for the buffer given it is fed through
       the passed rdp, may not be emptied until holdSize full, and
       may be fed in chunks feedSize big */

void DereferenceResampleDetails(RsmpDescType *rdp, int *pop, int *pip, 
				int *pkhlen);
    /* deref or supply default chrs for NULL resampling descriptor */

RsmpDescType *NewResampKernel(int outP, int inP, int lobecount, FLOATARG gain);
    /* Setup lookup tables for resampling a given ratio, num. lobes */
    /* gain = 1.0 attempts to maintain peak amplitude, 0 preserves energy */

RsmpDescType *MakeResampKernel(int outP, int inP, long len, float *kern);
    /* generate a structure from the passed filter & values */

void FreeResampKernel(RsmpDescType *rdp);	/* free kernel memory */

long ResampleIntoBuffer(SAMPL *dstBuf, long dbSize, SAMPL *sceBuf, long sbSize,
				 long ifStart, RsmpDescType *rdp, int chk);
    /* Perform resampling; superposes into existing buffer; 
       Returns argument to use as ifStart next time for seamlessness */

long oldResampleIntoBuffer(SAMPL *dstBuf, long dbSize, SAMPL *sceBuf, long sbSize,
				 long ifStart, RsmpDescType *rdp);

#else	/* no prototypes */
void 		FindGoodRatio();
void 		DereferenceResampleDetails();
RsmpDescType 	*NewResampKernel();
RsmpDescType    *MakeResampKernel();
void		FreeResampKernel();
long 		ResampleIntoBuffer();
long 		oldResampleIntoBuffer();

#endif

#endif /* _RESAMPLE_H_ */
