/***************************************************************\
*	samplops.c						*
*  Basic sample-type operations					*
*  dpwe 28feb91							*
\***************************************************************/

#include <dpwelib.h>
#include <sndf.h>
#include <samplops.h>	/* prototypes */

long WriteShortSampls(buf, len, outf)	/* overwrites buf */
    SAMPL *buf;
    long  len;
    FILE  *outf;
    {
    long  writ;

#ifdef LONGSAMPLS
    Long2Short(buf, 1, (short *)buf, 1, len);
#else
#ifndef SHORTSAMPLS
    Float2Short(buf, 1, (short *)buf, 1, len);
#endif	/* !SHORTSAMPLS */
#endif  /* LONGSAMPLS */
    writ = SFfwrite(buf, SFMT_SHORT, len, outf);
    return writ;
    }

long ReadShortSampls(buf, len, inf)
    SAMPL *buf;
    long len;
    FILE* inf;
    {
    long red;

    red = SFfread(buf, SFMT_SHORT, len, inf);
#ifdef LONGSAMPLS
    Short2Long((short *)buf, 1, buf, 1, red);
#else
#ifndef SHORTSAMPLS
    Short2Float((short *)buf, 1, buf, 1, red);
#endif /* ! SHORTSAMPLS */
#endif /* LONGSAMPLS */
    return red;
    }

void Sampl2Short(sb, sinc, db, dinc, len)
    SAMPL *sb;
    int sinc;
    short *db;
    int dinc;
    long len;
    {
#ifdef SHORTSAMPLS
    CopyShorts(sb, sinc, db, dinc, len, FWD);
#else
#ifdef LONGSAMPLS
    Long2Short(sb, sinc, db, dinc, len, FWD);
#else
    Float2Short(sb, sinc, db, dinc, len);
#endif
#endif
    }

void Short2Sampl(sb, sinc, db, dinc, len)
    short *sb;
    int sinc;
    SAMPL *db;
    int dinc;
    long len;
    {
#ifdef SHORTSAMPLS
    CopyShorts(sb, sinc, db, dinc, len, FWD);
#else
#ifdef LONGSAMPLS
    Short2Long(sb, sinc, db, dinc, len);
#else
    Short2Float(sb, sinc, db, dinc, len);
#endif
#endif
    }

void Short2Float(sb, sinc, fb, finc, n)
    short *sb;
    int sinc;
    float *fb;
    int finc;
    long n;
    {
    register double scale = 1.0/(float)(SHORTMAX+1L);

    fb += n*finc; sb += n*sinc;	/* start at top for in-place copying */
    while(n--)
	{
	fb -= finc; sb -= sinc;
	*fb = scale * (float)*sb;
	}
    }

long Float2Short(sb, sinc, db, dinc, n)		/* returns # of clips */
    float *sb;
    int sinc;
    short *db;
    int dinc;
    long n;
    {
    register double scale = (float)SHORTMAX+1L;
    long l;
    long clip = 0;

    /* start at bottom since shorts are smaller for in-place copying */
    while(n--)
	{
	l = (long)(scale * *sb);
	if(l>SHORTMAX)	{	l = SHORTMAX; ++clip;	}
	if(l<-SHORTMAX-1)	{	l = -SHORTMAX-1;	++clip;	}
	*db = (short)l;
	db += dinc; sb += sinc;
	}
    return clip;
    }

void Short2Long(sb, sinc, fb, finc, n)
    short *sb;
    int sinc;
    long *fb;
    int finc;
    long n;
    {
    fb += n*finc; sb += n*sinc;	/* start at top for in-place copying */
    while(n--)
	{
	fb -= finc; sb -= sinc;
	*fb = *sb;
	}
    }

long Long2Short(sb, sinc, db, dinc, n)
    long *sb;
    int sinc;
    short *db;
    int dinc;
    long n;
    {	/* returns number of clips */
    long l;
    long clip = 0;

    /* start at bottom since shorts are smaller for in-place copying */
    while(n--)
	{
	l = *sb;
	if(l>SHORTMAX)	{	l = SHORTMAX; ++clip;	}
	if(l<-SHORTMAX-1)	{	l = -SHORTMAX-1;	++clip;	}
	*db = (short)l;
	db += dinc; sb += sinc;
	}
    return clip;
    }

void CopyChars(sce, sinc, dst, dinc, len, fwdNbck)
    char *sce;
    int sinc;
    char *dst;
    int dinc;
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last */
	{
	sce += len * (long)sinc;
	dst += len * (long)dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first */
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }

void ClearChars(sptr, sinc, len)
    char *sptr;
    int  sinc;
    long len;
    {
    while(len--)
	{
	*sptr = 0;
	sptr += sinc;
	}
    }

void CopyShorts(sce, sinc, dst, dinc, len, fwdNbck)
    short *sce;
    int sinc;
    short *dst;
    int dinc;
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last */
	{
	sce += len * (long)sinc;
	dst += len * (long)dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first */
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }

void ClearShorts(sptr, sinc, len)
    short *sptr;
    int sinc;
    long len;
    {
    while(len--)
	{
	*sptr = 0;
	sptr += sinc;
	}
    }

void CopyFloats(sce, sinc, dst, dinc, len, fwdNbck)
    float *sce;
    int sinc;
    float *dst;
    int dinc; 
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last */
	{
	sce += len * (long)sinc;
	dst += len * (long)dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first */
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }

void ClearFloats(sptr, sinc, len)
    float *sptr;
    int sinc;
    long len;
    {
    while(len--)
	{
	*sptr = 0;
	sptr += sinc;
	}
    }

void CopySampls(sce, sinc, dst, dinc, len, fwdNbck)
    SAMPL *sce;
    int sinc;
    SAMPL *dst;
    int dinc; 
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last */
	{
	sce += len * (long)sinc;
	dst += len * (long)dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first */
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }

void ClearSampls(sptr, sinc, len)
    SAMPL *sptr;
    int sinc;
    long len;
    {
    while(len--)
	{
	*sptr = 0;
	sptr += sinc;
	}
    }

void ScaleSampls(sptr, sinc, dptr, dinc, len, factor)
    SAMPL *sptr;
    int   sinc;
    SAMPL *dptr;
    int   dinc;
    long  len;
    LONGSAMPL factor;
    {
    LONGSAMPL d;
    while(len--)
	{
#ifdef UNITYCLIP
	d = SCALEDOWN(factor * *sptr);
	if(d > SAMPLUNITY)		d = SAMPLUNITY;
	else if(d < -SAMPLUNITY)	d = -SAMPLUNITY;
	*dptr = d;
#else
	*dptr = SCALEDOWN(factor * *sptr);
#endif
	dptr += dinc;
	sptr += sinc;
	}
    }


void OverlaySampls(sptr, sinc, dptr, dinc, len, factor)
    SAMPL *sptr;
    int   sinc;
    SAMPL *dptr;
    int   dinc;
    long  len;
    LONGSAMPL factor;
    {
    LONGSAMPL d;
    while(len--)
	{
#ifdef UNITYCLIP
	d = (*dptr) + SCALEDOWN(factor * *sptr);
	if(d > SAMPLUNITY)		d = SAMPLUNITY;
	else if(d < -SAMPLUNITY)	d = -SAMPLUNITY;
	*dptr = d;
#else
	*dptr += SCALEDOWN(factor * *sptr);
#endif
	dptr += dinc;
	sptr += sinc;
	}
    }
    
