/***********************************************************************\
*    selwndw.c								*
*    Use <wndw.c> interface to manage a menu of selections		*
*    For track objects in dspB initially				*
*    dpwe 1993apr21							*
\***********************************************************************/

#include <dpwelib.h>
#include <genutils.h>	/* for strdup */
#include <wndw.h>	/* definition of primitives we will use */
#include <llists.h>	/* entries held in actual list */
#include "wndwbtn.h"	/* buttons for wndws */

#define T_SELENTRY	200	/* cartype for 'entries' lists */

typedef struct _spnl_str {		/* description of selection panel */
    WIND	*wndw;		/* where to display */
    RECT	rect;		/* where exactly */
    BTN_STR	*btns;		/* buttons in this window */
    LIST	entries;	/* linked list of entries */
    int 	nentries;	/* how many we have */
    int 	lineh;
    int		armed;		/* which item is armed (or -1) */
    void	(*hackFn1)();	/* allow hijacking of button 1993may13 */
    void	*hackFn1Data;
    void	(*hackFn2)();	/* allow hijacking of button 1993may13 */
    void	*hackFn2Data;
    void	(*hackFn3)();	/* allow hijacking of button 1993may13 */
    void	*hackFn3Data;
    char 	*title;
} SELPNL_STR;

typedef SELPNL_STR *SELPNL;	/* have to define as pointer, since 
				   ansi won't allow 'typedef void SELPNL_STR'
				   but only 'typedef void *SELPNL_STR **/

#define __SELPNL_STR_DEFD__

#include "selwndw.h"

#define XINIT    600      /* set default window location */
#define YINIT    50
#define WIDTH    150     /* start off small - user resizes */
#define HEIGHT   400

static void spnlRmBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{	/* callback function to bind to 'Remove entry' button */
    SELPNL_STR	*spnl = (SELPNL_STR *)data;
    SELENTRY 	*se;

    if(WGetShiftEtc(spnl->wndw))  {
        if(xbtn ==1 && (se = spnlGetSelected(spnl)) )  {
            spnlRmvEntry(spnl, se);
        }
    } else if(xbtn == 1) {
	if(spnl->hackFn1 != NULL)
	    (*(spnl->hackFn1))(spnl->hackFn1Data);
    } else if(xbtn == 2) {
	if(spnl->hackFn2 != NULL)
	    (*(spnl->hackFn2))(spnl->hackFn2Data);
    } else if(xbtn == 3) {
	if(spnl->hackFn3 != NULL)
	    (*(spnl->hackFn3))(spnl->hackFn3Data);
    }
}

void spnlSetHackFn(SELPNL_STR *spnl, char *name, 
		   void (*hackFn1)(), void *data1,
		   void (*hackFn2)(), void *data2,
		   void (*hackFn3)(), void *data3)
{	/* hack to redirect btns */
    spnl->hackFn1 = hackFn1;
    spnl->hackFn1Data = data1;
    spnl->hackFn2 = hackFn2;
    spnl->hackFn2Data = data2;
    spnl->hackFn3 = hackFn3;
    spnl->hackFn3Data = data3;
    wbtnSetLabelPtr(spnl->btns, name);
}

SELPNL_STR *spnlNew(char *title)
{	/* initialize new selpanel window, empty at first */
    SELPNL_STR	*spnl;
    RECT	rect;

    spnl = TMMALLOC(SELPNL_STR, 1, "spnlNew");
    spnl->wndw = NewXWind(XINIT, YINIT, WIDTH, HEIGHT, title, "");
    spnl->title = strdup(title);
    /* spnl->rect is set on redraw */
    spnl->btns = wbtnNew(NULL, spnl->wndw, 10, 10, 0, 0, 
			 "quit", QuitBtnFn, NULL);
    WAddEventHandler(spnl->wndw, (eventHandlerFn)wbtnEvent, spnl->btns);
    wbtnGetRect(spnl->btns, &rect);
    spnl->btns = wbtnNew(spnl->btns, spnl->wndw, rect.x + rect.width + 10, 
			 rect.y, 0, 0, 
			 "hide", HideBtnFn, NULL);
    WAddEventHandler(spnl->wndw, (eventHandlerFn)wbtnEvent, spnl->btns);
    wbtnGetRect(spnl->btns, &rect);
    spnl->btns = wbtnNew(spnl->btns, spnl->wndw, rect.x + rect.width + 10, 
			 rect.y, 0, 0, 
			 "remove entry     ", spnlRmBtnFn, spnl);
    WAddEventHandler(spnl->wndw, (eventHandlerFn)wbtnEvent, spnl->btns);
    /* add main event handler *after* buttons so it will clear *first* */
    WAddEventHandler(spnl->wndw, (eventHandlerFn)spnlEvent, spnl);
    spnl->entries = NULL;
    spnl->nentries = 0;
    spnl->lineh = WTextHeight(spnl->wndw, "entry");
    spnl->armed = -1;	/* nothing armed */
    spnl->hackFn1 = spnl->hackFn2 = spnl->hackFn3 = NULL;

    return spnl;
}

void spnlFree(SELPNL_STR *spnl)
{	/* destroy a selpanel */
    spnlReset(spnl);	/* frees all entries */
/*    WRemoveEventHandler(spnl->wndw, ...);	*/
/*    wbtnFree();	*/
    FreeWind(spnl->wndw);
    free(spnl->title);
    free(spnl);
}

char *spnlGetTitle(SELPNL_STR *spnl)
{	/* return the name given to this spnl on initialization */
    return spnl->title;
}

void spnlSetTitle(SELPNL_STR *spnl, char *title)
{	/* change the title of the selpnl */
    if(spnl->title)	free(spnl->title);
    spnl->title = strdup(title);
    WNewName(spnl->wndw, spnl->title);
}

int spnlUnhide(SELPNL_STR *spnl)
{	/* call to allow clients to unhide window if it 
	   gets hidden - e.g. user clicked 'hide' button */
    int rc = 0;

    if(spnl->wndw->hidden)  {
	WUnhide(spnl->wndw);
	rc = 1;
    }
    return rc;
}

#define XMARGIN  10	/* between l,r edge of win and graph */
/* #define TMARGIN  30	/* between top edge of win and graph */
#define BMARGIN  20     /* between bottom edge of win and graph */
#define XGAP     4	/* btwn l/r frame and text */
#define TGAP	 4	/* btwn top frame and text */
#define LGAP	 0	/* extra btwn lines */

static void spnlRectForEntry(SELPNL_STR *spnl, int entryNum, RECT *pr)
{	/* calculate the rectangle containing the n'th entry */
    *pr = spnl->rect;
    pr->x += XGAP;    			pr->y += TGAP;
    pr->width -= 2*XGAP;		pr->height = spnl->lineh + LGAP;
    pr->y += entryNum * spnl->lineh;
}

static void spnlInvert(SELPNL_STR *spnl, int item)
{	/* flip the video on the display of item n */
    RECT r;

    spnlRectForEntry(spnl, item, &r);
    WPushGC(spnl->wndw);
    WSetGMode(spnl->wndw, WGMD_BLACK, WGMD_WHITE, WGMD_XOR, 1, WGMD_SOLID);
    WFillRect(spnl->wndw, &r);
    WPopGC(spnl->wndw);
}

void spnlDraw(SELPNL_STR *spnl, RECT *exR)
{	/* redraw the panel e.g. after an expose.  
	   No tricks; always draw whole thing for now */
    RECT	txR, r2;
    int		x, y, h, lineh;
    LIST 	entry;
    int 	i;
    int		TMARGIN = 2.5*WTextHeight(spnl->wndw, "TimeWaste");

    WSetGMode(spnl->wndw, WGMD_BLACK, WGMD_WHITE, WGMD_COPY, 1, WGMD_SOLID);
    WGetWidthHeight(spnl->wndw);	/* force update to find window size */
    txR.x = 0;
    txR.y = 0;
    txR.width = spnl->wndw->wWidth;
    txR.height = spnl->wndw->wHeight;

    InsetRct(&txR, XMARGIN, (TMARGIN+BMARGIN)/2);
    OffsetRct(&txR, 0, (TMARGIN-BMARGIN)/2);

    WClearRect(spnl->wndw, &txR);
    WFrameRect(spnl->wndw, &txR);
    spnl->rect = txR;
    InsetRct(&txR, 1, 1);
    WSetClip(spnl->wndw, &txR);

    /* now write up entries */
    entry = spnl->entries;
    i = 0;
    spnlRectForEntry(spnl, i, &r2);
    while(entry && r2.y+r2.height <= txR.y+txR.height)  {
	WWriteIn(spnl->wndw, &r2, ((SELENTRY *)CAR(entry))->name, -1);
	/* hilite selected HERE */
	entry = CDR(entry);
	++i;
	spnlRectForEntry(spnl, i, &r2);
    }
    WSetClip(spnl->wndw, NULL);	/* remove clipping */
    if(spnl->armed >= 0)
	spnlInvert(spnl, spnl->armed);
}

static void spnlArm(SELPNL_STR *spnl, int item)
{
    if(item == spnl->armed)	return;
    else if(spnl->armed >= 0)	spnlInvert(spnl, spnl->armed);
    spnl->armed = item;
    if(spnl->armed >= 0)	spnlInvert(spnl, spnl->armed);
}

static void spnlDisarm(SELPNL_STR *spnl)
{
    spnlArm(spnl, -1);
}

SELENTRY *spnlGetNth(SELPNL_STR *spnl, int ix)
{	/* get the SELENTRY str for an entry in the panel, referred 
	   by its index <ix> = 0.. spnlGetNEntries(spnl) - 1 */
    LIST	node;

    if(ix < 0 || ix >= spnl->nentries)	return NULL;
    node = nth(spnl->entries, ix);
    return (SELENTRY *)CAR(node);
}

SELENTRY *spnlGetSelected(SELPNL_STR *spnl)
{	/* return the currently hilit item, or NULL if none */
    return spnlGetNth(spnl, spnl->armed);
}

static void spnlActivate(SELPNL_STR *spnl, int item, int btn)
{
    SELENTRY *se;
    
    spnlArm(spnl, item);
    if( (se = spnlGetNth(spnl, item)) != NULL)  {
	(*se->cbFn)(se->cbData, SPNL_ACTIVATE, se, btn);
    }
}

int spnlEvent(XEvent *event, SELPNL_STR *spnl)	/* args rvsd as gen.ev.fn */
{	/* somebody wanted to show us this event.  If take it, return 1
	   else 0 (i.e. give to someone else) */
    int x,y, state, btn, i, found;
    int rc = 0;
    SELENTRY *se;

    if(event->xany.window == spnl->wndw->win) { /* at least the right window */
	if(event->type == Expose)  {
	    RECT  rect;

	    /* check rect -YAWN- */
/*	    SetRct(&rect, event->xexpose.x, event->xexpose.x, 
		   event->xexpose.width, event->xexpose.height); 
	    if(RectOverlap(&rect, &(wbs->rect)))  {
 */
		spnlDraw(spnl, &rect);
		/* rc = 1; */		/* ?? others may care about expose */
	/*    }		*/
	} else if(event->type == ResizeRequest)  {
	    WResizeWindow(spnl->wndw, event->xresizerequest.width, 
			  event->xresizerequest.height);
	    rc = 1;
	} else if(event->type == ClientMessage)  {
	    /* ignore, but swallow */
	    rc = 1;
	} else if(event->type == ButtonPress 
		  || event->type == ButtonRelease 
		  || event->type == MotionNotify)  {	/* only others we do */
	    RECT r;

	    if(event->type == MotionNotify)  {
		x = event->xmotion.x;	
		y = event->xmotion.y;
		state = event->xmotion.state;
		/* btn   = event->xmotion.button; */ /* no btn for motion */
		btn = -1;
	    } else { 
		x = event->xbutton.x;	
		y = event->xbutton.y;
		state = event->xbutton.state;
		btn   = event->xbutton.button;
	    }
	    i = 0;	found = 0;
	    while(!found && i < spnl->nentries)  {
		spnlRectForEntry(spnl, i, &r);
		if(PtInRct(x, y, &r))
		    found = 1;
		else
		    ++i;
	    }
	    if(found)  {
		rc = 1;
		if(event->type == ButtonRelease)  {
		    spnlActivate(spnl, i, btn);
		} else {
		    spnlArm(spnl, i);
		}
	    }
	}
    }
    return rc;
}

char *stralloc(char *str)
{	/* allocate a new string a copy of existing */
    int 	len;
    char 	*newstr;

    len = strlen(str);
    newstr = TMMALLOC(char, len+1, "stralloc");
    strcpy(newstr, str);
    return newstr;
}

SELENTRY *spnlAddEntry(SELPNL_STR *spnl, char *name, 
		       selCbFn cbFn, void *cbData)
{	/* add an entry to the list */
    LIST *tail;
    SELENTRY *se;

    se = TMMALLOC(SELENTRY, 1, "spnlAddEntry");
    se->spnl = spnl;	/* link back to parent */
    se->name = stralloc(name);
    se->cbFn = cbFn;
    se->cbData = cbData;
    tail = find_last_ptr(&(spnl->entries));
    *tail = new_node(NULL, (LIST)se, T_SELENTRY);
    spnl->nentries = top_size(spnl->entries);
    se->index = spnl->nentries - 1;	/* always added at bottom */
    spnlDraw(spnl, NULL);	/* redraw new list */
    return se;
}

static void spnlRewriteIndices(SELPNL_STR *spnl)
{	/* after an item is removed, need to update index fields in entries */
    int 	i;
    SELENTRY	*se;

    for(i = 0; i< spnl->nentries; ++i)  {
	assert(se = spnlGetNth(spnl, i));
	se->index = i;
    }
}

int spnlSelect(SELPNL_STR *spnl, SELENTRY *se)
{	/* force selection to be specified entry;
           return 1 if found, 0 o/w (in which case, no selection, 
				     so spnlSelect(spnl, NULL) unselects) */
    LIST	pent;
    int		ix = 0;
    int		found = 0;
    int 	nents = spnl->nentries;

    while(!found && ix < nents)  {
	if(se == spnlGetNth(spnl, ix))
	    found = 1;
	else
	    ++ix;
    }
    if(found)
	spnlArm(spnl, ix);
    else
	spnlArm(spnl, -1);
    spnlDraw(spnl, NULL);
    return found;
}

int spnlRmvEntry(SELPNL_STR *spnl, SELENTRY *se)
{	/* remove entry indicated by <se> (previously returned by spnlAddE)
	   from <spnl>; return 1 on success, 0 if couldn't find it */
    LIST tail, *ptail;
    int	 rc = 0;

    tail = find_el(spnl->entries, (LIST)se, T_SELENTRY);
    if(tail)  {
	rc = 1;
	if(se = spnlGetSelected(spnl))	spnl->armed = -1;	/* unselect */
	ptail = find_par_ptr( &(spnl->entries), tail);
	/* found pointer to node to remove: close list, free node */
	assert(*ptail == tail);
	*ptail = CDR(tail);
	free_node(tail);
	free(se);
	--(spnl->nentries);
	assert(spnl->nentries >= 0);	/* check! */
	spnlRewriteIndices(spnl);	/* update numbers */
	spnlDraw(spnl, NULL);	/* redraw new list */
    }
    return rc;
}

int spnlGetNEntries(SELPNL_STR *spnl)
{	/* find the number of entries in the SELPNL_STR **/
    return spnl->nentries;
}

int spnlReset(SELPNL_STR *spnl)
{	/* remove all entries from a SELPNL_STR **/
    LIST tail, *ptail = &(spnl->entries);
    int	 rc = 0;

    while(*ptail)  {
	rc = 1;
	/* found pointer to node to remove: close list, free node */
	tail = *ptail;
	*ptail = CDR(tail);
	free((SELENTRY *)CAR(tail));
	free_node(tail);
    }
    if(rc)  {	/* (did make some changes) */
	spnl->nentries = 0;
	spnlDraw(spnl, NULL);	/* redraw empty list */
    }
    return rc;
}


#ifdef MAIN	/* test code below */

int dumFn(void *data, int state, SELENTRY *se)
{	/* callback */
    fprintf(stderr, "dumFunc: %lx (scs '%s')\n", data, se->name);
    return 0;
}

static SELENTRY *sse = NULL;

int sel3Fn(void *data, int state, SELENTRY *se)
{	/* callback */
    fprintf(stderr, "sel3Func: %lx (scs '%s')[%lx]\n", data, se->name, sse);
    spnlSelect(se->spnl, sse);
    sse = se;
    return 0;
}

main()
{
    SELPNL_STR *sp;

    sp = spnlNew("Hello world");
    spnlAddEntry(sp, "Dan", sel3Fn, (void *)0x01);
    spnlAddEntry(sp, "Loves", sel3Fn, (void *)0x02);
    spnlAddEntry(sp, "Sarah", dumFn, (void *)0x03);

    while(1)  {
	WEventLoop();
    }
}

#endif /* MAIN */

