/***************************************************************\
*   snd.c							*
*   A new soundfile interface layer				*
*   1993mar03  dpwe  after convo.c				*
\***************************************************************/

/* std libs */
#include "snd.h"
/* #include <stdlib.h>
   #include <string.h>
 */ /* both in libdpwe.h, itself in snd.h */
#include <assert.h>
/* personal libs */
#include <genutils.h>	/* for fSize(), NumericQ() and SubstExtn() */
#include <sndfutil.h>	/* for Convert* */
#include <audIO.h>	/* for real-time sound IO */

/* constants */
#define SND_BUFSZ	16384	/* frames held from disk */
#define SND_AUP_LEN	(sizeof(SND_AUDIO_PATH)/sizeof(char))

/* typedefs */

/* static function prototype */
static void sndDerived(SOUND *snd);

/* defaults for open-write (new) file */

/* implementations */

static void sndDerived(SOUND *snd)
{	/* fill in derived slots based on SFSTRUCT data */
    if(snd->snd->dataBsize != SF_UNK_LEN)  {
	snd->frames = snd->snd->dataBsize / SFSampleFrameSize(snd->snd);
        snd->duration = (float)snd->frames / snd->snd->samplingRate;
    } else {
	snd->frames = SF_UNK_LEN;
	snd->duration = 0.0;
    }
}

SOUND *sndNew(SOUND *snd)
{	/* just allocate a new SOUND structure
	   Clone format parameters (chans, srate) from <snd> if any. */
    SOUND *rslt;

    rslt = TMMALLOC(SOUND, 1, "sndNew");
    rslt->path = NULL;
    rslt->file = NULL;
    rslt->mode = '\0';
    rslt->isStdio = FALSE;
    rslt->isRealtime = FALSE;
    rslt->snd  = NULL;
    rslt->frames   = 0L;
    rslt->duration = 0.0;
    rslt->userFormat = -1;
    rslt->userChans  = -1;
    rslt->flags   = 0;
    rslt->buf	  = NULL;
    rslt->bufSize = 0L;
    rslt->bufCont = 0L;
    if(snd != NULL)  {
	SFAlloc(&(rslt->snd), 0L, snd->snd->format, snd->snd->samplingRate,
		snd->snd->channels, SFInfoSize(snd->snd));
	memcpy(rslt->snd->info, snd->snd->info, SFInfoSize(snd->snd));
	/* rslt->snd->dataBsize = snd->snd->dataBsize; */ /* leave as 0 */
	rslt->userFormat = snd->userFormat;
	rslt->userChans  = snd->userChans;
    }
    else {		/* alloc default SF header */
	SFAlloc(&(rslt->snd), 0L, DFLT_FMT, DFLT_SR, DFLT_CHNS, 0);
    }
    /* derived fields */
    sndDerived(rslt);
    return rslt;
}

void sndClose(SOUND *snd)
{	/* close sound file attached to snd */
    if(snd->file)  {			/* file really is open */
	if(snd->mode == 'w')  {		/* open for write - careful close */
	    SFCloseWrHdr(snd->file, snd->snd);
	}
	else {
	    fclose(snd->file);
	}
    }
    snd->file = NULL;
}

void sndFree(SOUND *snd)
{	/* destroy sound, including closing any file */
    if(snd->file != NULL)  {
	sndClose(snd);
    }
    if(snd->buf != NULL)  {
	free(snd->buf);
    }
    free(snd);
}

static AUSTRUCT *sndAuOpen(char *path, int aumode)
{   /* parse path name, open audio */
    AUSTRUCT *au = NULL;

    fprintf(stderr, "sndAuOpen: NYI\n");

    return au;
}

static SOUND *sndOpenRTAudio(SOUND *snd, char *mode)
{   /* open a sound for read or write, but as a pipe to hardware */
    char *newname;
    AUSTRUCT *au;

    /* filename is validly prefixed by filename that indicates it 
       should be used to open a real-time audio input */
    snd->isRealtime = TRUE;
    if(*mode == 'r')  {
	if( (au =sndAuOpen(snd->path+SND_AUP_LEN, AIO_INPUT))==NULL){
	    fprintf(stderr, "sndOpen: RT audio in fail %s\n", snd->path);
	    sndFree(snd);
	    return NULL;
	}
	newname = "(realtime input)";
    } else {
	if( (au =sndAuOpen(snd->path+SND_AUP_LEN, AIO_OUTPUT))==NULL){
	    fprintf(stderr, "sndOpen: RT audio out fail %s\n", snd->path);
	    sndFree(snd);
	    return NULL;
	}
	newname = "(realtime output)";
    }
    TMFREE(snd->path, "sndOpen rewrite path:free");
    snd->path = TMMALLOC(char, strlen(newname)+1, "sndOpen path rewrite");
    strcpy(snd->path, newname);
    snd->file = (FILE *)au;	/* hack, for now */
    return snd;
}

static SOUND *sndOpenFileRead(SOUND *snd)
{   /* Finish off opening a soundfile with its own path */
    if(snd->snd)
	SFFree(snd->snd);	/* free the old SFSTRUCT prior to new */
    if(snd->isStdio)  {
	snd->file = stdin;
    } else {
	if( (snd->file = fopen(snd->path, "rb")) == NULL)  {
	    fprintf(stderr, "sndOpen: Cannot open '%s' for read\n", 
		    snd->path);
	    sndFree(snd);
	    return NULL;
	}
    }
    if( SFAllocRdHdr(snd->file, &(snd->snd)) )  {
	fprintf(stderr, "sndOpen: '%s' is not a soundfile\n", snd->path);
	sndFree(snd);
	return NULL;
    }
    /* fill in derived fields */
    sndDerived(snd);
    snd->mode = 'r';
    return snd;
}

static SOUND *sndOpenFileWrite(SOUND *snd)
{   /* Finish off opening a soundfile with its own path */
    if(snd->isStdio)  {
	snd->file = stdout;
    } else {
	if( (snd->file = fopen(snd->path, "w+b")) == NULL)  {
	    fprintf(stderr, "sndOpen: Cannot open '%s' for write\n", 
		    snd->path);
	    sndFree(snd);
	    return NULL;
	}
    }
    if( SFWriteHdr(snd->file, snd->snd, 0) )  {
	fprintf(stderr, "sndOpen: Error writing to  '%s'\n",snd->path);
	sndFree(snd);
	return NULL;
    }
    snd->mode = 'w';
    return snd;
}

SOUND *sndOpen(SOUND *snd, char *path, char *mode)
{   /* open sound. If <snd> is NULL, allocate new one.
	If <path> is NULL, take from snd as passed in (or err).
	<mode> is "r" or "w" only */
    int allocd = 0;

    if( !(*mode == 'r' || *mode == 'w')) {
	fprintf(stderr, "sndOpen: mode '%s' not 'r' or 'w'\n", mode);
	return (SOUND *)NULL;
    }
    if(path == NULL &&
       (snd == NULL || snd->path == NULL || strlen(snd->path) == 0))  {
	/* must either specify explicit path, or have path present
	   in snd structure passed in. */
	fprintf(stderr, "sndOpen: path = '%s', snd = 0x%lx, no path\n",
		path, snd);
	return (SOUND *)NULL;
    }
    if(snd == NULL)  {
	snd = sndNew(NULL);
	allocd = 1;
    }
    if(path != NULL) {
	/* passed path takes priority */
	snd->path = TMMALLOC(char, strlen(path)+1, "sndOpen path");
	strcpy(snd->path, path);
    }
    /* now have a snd structure, and its path field is the true path 
       we want to access */
    if(strncmp(snd->path, SND_AUDIO_PATH, SND_AUP_LEN) == 0)  {
	snd = sndOpenRTAudio(snd, mode);
    } else {
	if(strcmp(snd->path, SND_STDIO_PATH)==0)  { /* magic name for stdin */ 
	    char *newname;

	    snd->isStdio = TRUE;
	    if(*mode == 'r')	newname = "(stdin)";
	    else			newname = "(stdout)";
	    TMFREE(snd->path, "sndOpen rewrite path:free");
	    snd->path = TMMALLOC(char, strlen(newname)+1, 
				 "sndOpen path rewrite");
	    strcpy(snd->path, newname);
	}
	if(*mode == 'r')  {
	    snd=sndOpenFileRead(snd);
	}
	else if(*mode == 'w')  {
	    snd=sndOpenFileWrite(snd);
	}
    }
    if(snd)  {
	/* 'snd' now valid, ready for return */
	if(snd->buf == NULL)  {	/* allocate a buffer to read from disk */
	    snd->buf= TMMALLOC(short, snd->snd->channels*SND_BUFSZ, 
			       "sndOpen:buf");
	    /* CCKR assert can't handle double-quotes in the expression */
	    assert(snd->buf);
	    snd->bufSize = SND_BUFSZ; /* space for this many sample FRAMES */
	    snd->bufCont = 0;	/* .. but empty at present */
	}
	if(snd->userFormat == -1)
	    snd->userFormat = SFMT_FLOAT; /* by default ..? */
    /* leave the userChans at -1 - then sndGetUchans returns *actual* chans */
    }
#ifdef DEBUG 
    fprintf(stderr, "sndOpen(%lx,%s,%s):dz %ld, fr %ld, ch %d, fmt %d\n", 
	    snd, path, mode, snd->snd->dataBsize, 
	    snd->frames, snd->snd->channels, snd->snd->format);
#endif /* DEBUG */
    return snd;
 errorexit:
    if(allocd&2)	SFFree(snd->snd);
    if(allocd&1)	sndFree(snd);
    return (SOUND *)NULL;
}

long snduConvertChans(void *inbuf, void *outbuf, int inchans, int outchans, 
		      int format, long frames)
{   /* convert the number of channels in a sound of arbitrary format */
    int bps = SFSampleSizeOf(format);

    if(inchans == outchans)  {
	if(inbuf == outbuf)	return frames;
	else  {
	    memcpy(outbuf, inbuf, frames*inchans*bps);
	    return frames;
	}
    }
    if(format == SFMT_SHORT)  {
	ConvertChannels(inbuf, outbuf, inchans, outchans, frames);
    } else if(format == SFMT_FLOAT) {
	ConvertFlChans(inbuf, outbuf, inchans, outchans, frames);
    } else {
	fprintf(stderr, "snduCnvChans: cannot convert %d to %d chans for data format %d\n", inchans, outchans, format);
	return 0;
    }
    return frames;
}

long snduConvertChansFmt(void *inbuf, void *outbuf, int inchans, int outchans, 
			 int infmt, int outfmt, long frames)
{   /* convert the data format and the number of channels for a buffer 
       of sound.  Use smallest possible buffer space */
    int ibs = SFSampleSizeOf(infmt);
    int obs = SFSampleSizeOf(outfmt);

    if( (ibs*outchans) <= (obs*outchans) )  {
	/* size of channel conversion ( <ibs,ichans> => <ibs,ochans> )
	   prior to channel conversion will fit in output buffer */
	snduConvertChans(inbuf, outbuf, inchans, outchans, infmt, frames);
	ConvertSampleBlockFormat(outbuf, outbuf, infmt,outfmt,frames*outchans);
    } else if( (obs*inchans) <= (obs*outchans) )  {
	/* size of format coversion first will fit in output buffer */
	ConvertSampleBlockFormat(inbuf, outbuf, infmt, outfmt, frames*inchans);
	snduConvertChans(outbuf, outbuf, inchans, outchans, outfmt, frames);
    } else {	/* neither stage will fit in output buffer - must both be 
		   reducing (e.g float->short *and* stereo->mono) */
	void *tmpBuf = TMMALLOC(char, ibs*outchans*frames, "snduCnvCF");

	snduConvertChans(inbuf, tmpBuf, inchans, outchans, infmt, frames);
	ConvertSampleBlockFormat(tmpBuf, outbuf, infmt,outfmt,frames*outchans);
	TMFREE(tmpBuf, "snduCnvCF");
    }
    return frames;
}

long sndReadFrames(SOUND *snd, void *buf, long frames)
{   /* read <frames> sample frames as floats into the buffer,
       return number actually read. */
    long framesToDo = frames;
    long framesThisTime, framesRead;
    long samples, smpsRead, totalFramesRead = 0;
    int  channels = snd->snd->channels;
    int  uchans   = sndGetUchans(snd);
    int  maxbufframes = snd->bufSize / MAX(1, uchans/channels);
    /* only half-fill the buffer if we're doubling the number of channels */
    int  ubps = SFSampleSizeOf(snd->userFormat);

#ifdef DEBUG
    if(snd->flags == 0)  {
	fprintf(stderr, "sndRF: %d fr; fmt,ch %d,%d; ufmt,uch %d,%d\n", 
		frames, snd->snd->format, snd->snd->channels, 
		sndGetUformat(snd), sndGetUchans(snd));
	/* snd->flags = 1; */
    }
#endif /* DEBUG */
    if(snd->snd->format == snd->userFormat && channels == uchans)  {
      totalFramesRead = SFfread(buf, snd->snd->format, frames*channels, 
				snd->file)/channels;
    } else {
      assert(snd->bufSize>0);
      while(framesToDo)  {
	framesThisTime = MIN(framesToDo, maxbufframes);
	samples = framesThisTime * channels;
	smpsRead = SFfread(snd->buf, snd->snd->format, samples, snd->file);
	framesRead = smpsRead/channels;
	totalFramesRead += framesRead;
	snduConvertChansFmt(snd->buf, buf, channels, uchans, 
			    snd->snd->format, snd->userFormat, framesRead);
	if(smpsRead < samples)	framesToDo = 0;		/* we hit EOF */
	else			framesToDo -= framesThisTime;
	buf = ((char *)buf) + smpsRead * ubps;
      }
    }
    return totalFramesRead;
}

long sndWriteFrames(SOUND *snd, void *buf, long frames)
{	/* write <frames> sample frames (floats) to file from the buffer,
	   return number actually written. */
    long framesToDo = frames;
    long framesThisTime;
    long usamples, osamples, smpsWrit, totalFramesWrit = 0;
    int  channels = snd->snd->channels;
    int  uchans   = sndGetUchans(snd);
    int  maxbufframes = snd->bufSize / MAX(1, channels/uchans);
    int  ubps = SFSampleSizeOf(snd->userFormat);

#ifdef DEBUG
    if(snd->flags == 0)  {
	fprintf(stderr, "sndWF: %d fr; fmt,ch %d,%d; ufmt,uch %d,%d\n", 
		frames, snd->snd->format, snd->snd->channels, 
		sndGetUformat(snd), sndGetUchans(snd));
	/* snd->flags = 1; */
    }
#endif /* DEBUG */
    assert(snd->bufSize>0);
    while(framesToDo)  {
	framesThisTime = MIN(framesToDo, maxbufframes);
	snduConvertChansFmt(buf, snd->buf, uchans, channels, 
			    snd->userFormat, snd->snd->format, framesThisTime);
	usamples = framesThisTime * uchans;
	buf = ((char *)buf) + ubps*usamples;
	osamples = framesThisTime * channels;
	smpsWrit = SFfwrite(snd->buf, snd->snd->format, osamples, snd->file);
	totalFramesWrit += smpsWrit/channels;
	if(smpsWrit < osamples)	framesToDo = 0;		/* disk full? */
	else			framesToDo -= framesThisTime;
    }
    return totalFramesWrit;
}

/* accessor / settor functions */

int sndSetChans(SOUND *snd, int chans)
{
    int oldchans = snd->snd->channels;

    snd->snd->channels = chans;
    return oldchans;
}

int sndGetChans(SOUND *snd)
{
    return snd->snd->channels;
}

int sndSetFormat(SOUND *snd, int format)
{
    int oldformat = snd->snd->format;

    snd->snd->format = format;
    return oldformat;
}

int sndGetFormat(SOUND *snd)
{
    return snd->snd->format;
}

int sndSetUformat(SOUND *snd, int format)
{
    int oldformat = snd->userFormat;

    snd->userFormat = format;
    return oldformat;
}

int sndGetUformat(SOUND *snd)
{
    return snd->userFormat;
}

int sndSetUchans(SOUND *snd, int chans)
{
    int oldchans = snd->userChans;

    snd->userChans = chans;
    return oldchans;
}

int sndGetUchans(SOUND *snd)
{
    if(snd->userChans == -1)
	return snd->snd->channels;
    return snd->userChans;
}

long sndSetFrames(SOUND *snd, long frames)
{
    long oldframes = snd->frames;

    snd->frames = frames;	/* what *does* this mean? */
    return oldframes;
}

long sndGetFrames(SOUND *snd)
{
    return snd->frames;
}

double sndSetSrate(SOUND *snd, double srate)
{	
    double oldsrate = snd->snd->samplingRate;

    snd->snd->samplingRate = srate;
    return oldsrate;
}

double sndGetSrate(SOUND *snd)
{
    return snd->snd->samplingRate;
}

long sndFrameTell (SOUND *snd)
{   /* return frame index of next sample frame to be read or written */
    return SFftell(snd->file)/SFSampleFrameSize(snd->snd);
}

long sndFrameSeek(SOUND *snd, long frameoffset, int whence)
{   /* seek to a certain point in the file.  Args as fseek(3s) */
    return SFfseek(snd->file, frameoffset*SFSampleFrameSize(snd->snd), whence);
}

int sndFeof(SOUND *snd)
{   /* non-zero if we have hit eof on reading a soundfile */
    return SFfeof(snd->file);
}
