/***************************************************************\
*	sunplay.c						*
*  provide lofi-style device interface for realtime output on	*
*  SUN sparcstation /dev/audio thingy				*
*  dpwe 1993mar04  after SGIplay.c				*
\***************************************************************/

#include <dpwelib.h>
#include <ulaw.h>

#define DEVAUDIO	"/dev/audio"	/* device name */
#define SRATE		(8000)		/* actual srate */
#define BUFSIZE		1024

static FILE *devaudio = NULL;
static int pldszstat = 0;
static int rcdszstat = 0;
static int nchans = 0;
static char *cbuf = NULL;

long play_set(chans, dsize, srate, scale)
    int chans;
    int dsize;
    float srate;
    int scale;
    {
    pldszstat = dsize;	/* remember what it is set to */
    nchans = chans;     /* static global for play_on */
    if(srate != SRATE)  {
	fprintf(stderr, "srate %.1f is not %.1f which is what I do.\n",
		srate, (float)SRATE);
	/* continue anyway */
    }
    if( (devaudio = fopen(DEVAUDIO, "a")) == NULL)  {
	fprintf(stderr, "Cannot open '%s' for sound output\n", 
		DEVAUDIO);
	return 0;	/* fail */
    } else {
	if(cbuf == NULL)  {
	    cbuf = TMMALLOC(char, BUFSIZE, "SUNplay_set");
	}
	return 1;	/* success */
    }
}	

void play_on(buf, csize)
    short *buf;
    long  csize;	/* the number of SAMPLE FRAMES */
{
    long  rem, i, chunk;

    if(devaudio)  {
	chunk = BUFSIZE;
	rem = csize;
	while(rem)  {
	    chunk = MIN(chunk, rem);
	    Short2Ulaw(buf, cbuf, nchans, 1, chunk);
	    fwrite(cbuf, 1, chunk, devaudio);
	    rem -= chunk;
	    buf += chunk * nchans;
	}
    }
}

long play_bufremains(void)
{	/* cannot really poll buffer in this world */
    return 0;
}

void play_rls()
{
    if(devaudio)  {
	fclose(devaudio);
	devaudio = NULL;
    }
}

void play(stt,end,chans,dsize,srate,scale)	/* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    float   srate;
    int     scale;	/* crude gain - ignored here */
    {
    short 	*src;
    int		chunkSamps;
    int		bestSize;

    play_set(chans,dsize,srate,scale);

    src = stt;
    bestSize = 1024*chans; /* (hbBlck/2)*chans; /* hblck/2 is frames per blk */
    while(src < end)		/* samples still to send */
	{
	chunkSamps = end-src;
	if(chunkSamps>bestSize) chunkSamps=bestSize;
	play_on(src,chunkSamps);
	src += (chunkSamps);	/* cs/2 is num of frames */
	}
    play_rls();		/* release for parallel code */
    }

int rec_set(chans, dsize, srate, scale)
    int chans;
    int dsize;
    float srate;
    int scale;
{
    fprintf(stderr, "Sorry: no audio in on SUNs\n");
    return 0;
}

void rec_on(buf, csize)
    short *buf;
    long  csize;	/* the number of SAMPLES */
    {
    }

void rec_rls()
    {
    }

void record(stt,end,chans,dsize,srate,scale)	/* called from main place */
    short   *stt;
    short   *end;
    int     chans;
    int     dsize;	/* ignored - assumed = sizeof(short) in what follows */
    float   srate;
    int     scale;	/* crude gain - ignored here */
    {
    short 	*src;
    int		chunkSamps;
    int		bestSize;

    rec_set(chans,dsize,srate,scale);

    src = stt;
    bestSize = 1024*chans; /* (hbBlck/2)*chans; /* hblck/2 is frames per blk */
    while(src < end)		/* samples still to send */
	{
	chunkSamps = end-src;
	if(chunkSamps>bestSize) chunkSamps=bestSize;
	rec_on(src,chunkSamps);
	src += (chunkSamps);	/* cs/2 is num of frames */
	}
    rec_rls();		/* release for parallel code */
    }

