/***********************************************************************\
* winwndw.c								*
* like winX11.c, except uses nicer wndw.c interface			*
* dpwe 1992oct24							*
\***********************************************************************/

#include <dpwelib.h>
#include <math.h>	/* log/exp used for log-scale mouse tracking */
#include "window.h"	/* need to provide these fns */
#include "wndw.h"	/* definition of primitives we will use */
#include "wndwbtn.h"	/* buttons for wndws */
#include "vecops.h"	/* more functions to use */
#include "samplops.h"	/* CopyFloats */

#define XINIT    80      /* set default window location */
#define YINIT    50
#define WIDTH    450     /* start off small - user resizes */
#define HEIGHT   120
#define MSGX     16	 /* offset for message in window */
#define MSGY     16
#define TXHGHT   14	 /* baseline offset for text */
#define STACKDEPTH 2	 /* after this many, start new row */

#define XHAIRBUTTON   	  Button1	/* was Button2 */
#define XHAIRBUTTONMASK   Button1Mask	/* was Button2Mask */

#if (MIPSEL && ultrix)          /* MIPS compiler can't handle reference */
#define PPARG(s)    ()
#else /* MIPSEL && ultrix */
#define PPARG(s)    PARG(s)
#endif /* MIPSEL && ultrix */

/* for GRAPHDAT_STR.mouseMode .. */
#define MSMD_NONE	0
#define MSMD_XHAIR	1
#define MSMD_RBAND	2

/* for GRAPHDAT_STR.drawMode .. */
#define DWMD_COPY	0
#define DWMD_XORRED	1
#define DWMD_XORBLK	2

typedef struct _gd_str {	/* main graphical window str */
    struct _gd_str	*next;	/* linked list of all windows */
    struct _gd_str	*sib;	/* rings of siblings have linked displays */
    WIND		*wdw;
    BTN_STR		*btns;
    BTN_STR		*pauseBtn;
    float		*data;
    int 		lastMsX;
    int 		lastMsY;
    int 		mouseMode;
    double		xHairX;
    int 		drawMode;
    int 		hidden;
    int 		pauseOnDraw;
    int 		sloppyUpdate;
    long		datapts;
    long		maxpts;
    double		xmin, xmax;
    double		nomxmin, nomxmax;
    double		ymin, ymax;
    int 		xislog;		/* show x axis with log scale */
    RECT		gRect;
    char 		label[32];
} GRAPHDAT_STR;

static GRAPHDAT_STR 	*gdRoot = NULL;

/* prototypes */
void gdsExpose(GRAPHDAT_STR *gds, int x, int y, int width, int height);
void gdEventLoop(void);	/* clear events in any of our windows but no block */
int gdEvent(XEvent *event, GRAPHDAT_STR *gds);	/* args rvsd as gen.ev.fn */

int Graphable()	/* called during program initialisation */
{		/* decides whether to use X or not; initializes X if so */
    return WInitX();
}

GRAPHDAT_STR *gdsNew(GRAPHDAT_STR *next, WIND *wdw)
{
    GRAPHDAT_STR *gds;

    gds = TMMALLOC(GRAPHDAT_STR, 1, "gdsNew");
    gds->next = next;
    gds->sib  = gds;	/* sibling rings initially degenerate */
    gds->wdw = wdw;
    gds->btns = NULL;
    gds->data = NULL;
    gds->datapts = 0;	/* actual valid pts */
    gds->maxpts = 0;	/* allocated size of buffer */
    gds->xmin = gds->xmax = 0.0;
    gds->nomxmin = gds->nomxmax = 0.0;
    gds->ymin = gds->ymax = 0.0;	/* default to autoscale? */
    gds->lastMsX = -1;	    gds->lastMsY = -1;
    gds->mouseMode = MSMD_NONE;
    gds->xHairX = 0.0;
    gds->hidden = FALSE;
    gds->pauseOnDraw = FALSE;
    gds->sloppyUpdate = FALSE;
    gds->xislog = FALSE;
    gds->label[0] = '\0';

    return gds;
}

typedef int (*ringFn)(GRAPHDAT_STR *gds, void *data);

int gdsApplyRing(GRAPHDAT_STR *gds, ringFn fn, void *fnData)
{	/* go round the sibling ring, apply fn to each, us last */
    GRAPHDAT_STR *sib;
    int 	 rc = 1;

    sib = gds;
    do  {
	sib = sib->sib;
	if( (*fn)(sib, fnData) == 0)
	    rc = 0;
    } while(sib != gds);
    return rc;
}

int gdsRFtestSib(GRAPHDAT_STR *gds, void *data)
{	/* ringFn to check sib not already in ring */
    GRAPHDAT_STR *sib = (GRAPHDAT_STR *)data;

    if(sib == gds)
	return 0;
    else
	return 1;
}
    
void gdsAddSib(GRAPHDAT_STR *gds, GRAPHDAT_STR *sib)
{
    if(1 || gdsApplyRing(gds, gdsRFtestSib, (void *)gds))  {
	/* splice into ring */
	sib->sib = gds->sib;
	gds->sib = sib;
    }
    else {	/* check doesn't work since always finds itself */
	fprintf(stderr, "duplicate sib not added\n");
    }
}

void wdLink(WINDAT *wd1, WINDAT *wd2)
{
    GRAPHDAT_STR *gds = (GRAPHDAT_STR *)wd1->windid;
    GRAPHDAT_STR *sib = (GRAPHDAT_STR *)wd2->windid;

    gdsAddSib(gds, sib);
}

void gdsHideBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{
    GRAPHDAT_STR *gds = (GRAPHDAT_STR *)data;

    gds->hidden = TRUE;
    WHide(gds->wdw);
}

void gdsXorRedMode(GRAPHDAT_STR *gds)
{
    if(gds->drawMode != DWMD_XORRED) {
	WSetGMode(gds->wdw, WGMD_RED, WGMD_WHITE, WGMD_XOR, 1, WGMD_SOLID);
	gds->drawMode = DWMD_XORRED;
    }
}
    
void gdsXorBlackMode(GRAPHDAT_STR *gds)
{
    if(gds->drawMode != DWMD_XORBLK) {
	WSetGMode(gds->wdw, WGMD_BLACK, WGMD_WHITE, WGMD_XOR, 1, WGMD_SOLID);
	gds->drawMode = DWMD_XORBLK;
    }
}
    
void gdsCpyMode(GRAPHDAT_STR *gds)
{
    if(gds->drawMode != DWMD_COPY) {
	WSetGMode(gds->wdw, WGMD_BLACK, WGMD_WHITE, WGMD_COPY, 1, WGMD_SOLID);
	gds->drawMode = DWMD_COPY;
    }
}
    
static pauseOn = 0;
void PauseBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{
    GRAPHDAT_STR *gds = (GRAPHDAT_STR *)data;

    if(pauseOn == 0) {
	pauseOn = 1;
	WSetEventPause(pauseOn);
    }
    else  {
	pauseOn = 0;		/* .. when we will be re-entered and clrd hr*/
	WSetEventPause(pauseOn);
    }
    if(btn)  {
	wbtnHilite(btn, pauseOn);
    }
}
void StepBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{	/* pause when the next redisp to this wdw happens */
    GRAPHDAT_STR *gds = (GRAPHDAT_STR *)data;

    if(pauseOn == 0) {
	pauseOn = 1;
	WSetEventPause(pauseOn);
	if(btn)  {	/* if we know which button.. */
	    wbtnHilite(btn, pauseOn);
	}
/*	while(pauseOn == 1)
	    gdEventLoop();	/* will keep in here till pause hit again */
    }
    else{
	pauseOn = 0;		/* .. when we will be re-entered and clrd hr*/
	WSetEventPause(pauseOn);
	gds->pauseOnDraw = TRUE;
	if(btn)  {	/* if we know which button.. */
	    wbtnHilite(btn, pauseOn);
	}
    }
}

void RescaleBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{
    WINDAT *wdptr = (WINDAT *)data;
    GRAPHDAT_STR *gds = (GRAPHDAT_STR *)wdptr->windid;

    ResetMaxMin(wdptr);
    gds->ymin = gds->ymax = 0.0;
    if(!gds->hidden)
	gdsExpose(gds, gds->gRect.x, gds->gRect.y, 
		  gds->gRect.width, gds->gRect.height);
}

void gdsDrawBtn(GRAPHDAT_STR *gds, BTN_STR *wbs, int hlt)
{
    gdsCpyMode(gds);
    wbtnDraw(wbs);
}

static WIND *MakeWind(name)
    char *name;
{
    WIND		 *win;
static int 		 winstance = 0;

    win = NewXWind(
		       XINIT + (winstance % STACKDEPTH)*(int)(1.1*WIDTH), 
		       YINIT + (winstance / STACKDEPTH)*(int)(1.2*HEIGHT), 
		       WIDTH, HEIGHT, name, "");
    ++winstance;
    return(win);
}

void MakeGraph(wdptr, name)
    WINDAT *wdptr;
    char   *name;
{
    GRAPHDAT_STR *gds;
    gds = gdsNew(gdRoot, MakeWind(name));
    gds->btns = wbtnNew(gds->btns, gds->wdw, 5, 5, 0, 0, "Quit", 
			QuitBtnFn, NULL);
    gds->pauseBtn = 
	gds->btns = wbtnNew(gds->btns, gds->wdw, 50, 5, 0, 0, "Pause", 
			    PauseBtnFn, gds);
    gds->btns = wbtnNew(gds->btns, gds->wdw, 95, 5, 0, 0, "Step", 
			StepBtnFn, gds);
    gds->btns = wbtnNew(gds->btns, gds->wdw, 140, 5, 0, 0, "Hide", 
			gdsHideBtnFn, gds);
    gds->btns = wbtnNew(gds->btns, gds->wdw, 180, 5, 0, 0, "Rescale", 
			RescaleBtnFn, wdptr);
    /* LABEL BUTTON MUST BE LAST (=top of button list, to find) */
    gds->btns = wbtnNew(gds->btns, gds->wdw, 230, 5, 0, 0, 
			"Your message here", 
			NULL, NULL);
    wbtnSetLabelPtr(gds->btns, &(gds->label[0]));
    wdptr->windid = (long)gds;
    if(wdptr->waitflg)
	gds->pauseOnDraw = TRUE;  /* will cede control to our ev loop on drw */

    gdRoot = gds;		/* add into list of windows */

    WAddEventHandler(gds->wdw, (eventHandlerFn)gdEvent, gds);
}

static int  scaleHMargin = 0;
static int  scaleVMargin;

void gdsRedrawGraph(GRAPHDAT_STR *gds, RECT *exR)
{
    int i, x, y;
    int firstPt = 1;
    double xscale, yscale;
    RECT r2;
    int  tx[200], ntx;

    if(scaleHMargin == 0)  {
	scaleHMargin = WTextWidth(gds->wdw, "2000k  ");
	scaleVMargin = 1.5*WTextHeight(gds->wdw, "O");
    }
    WClearRect(gds->wdw, &(gds->gRect));
    WFrameRect(gds->wdw, &(gds->gRect));
    if(gds->datapts == 0)
	WDrawX(gds->wdw, &(gds->gRect));
    else {
	if(gds->xmin == gds->xmax || gds->ymin == gds->ymax)  {
	    /* indicates a rescale */
	    VMinMax(gds->data, 1, gds->datapts, 
		    &(gds->ymin), &(gds->ymax));
	    gds->xmin = 0.0;	gds->xmax = gds->datapts;
	}
	/* Vertical (Y) scale */
	r2 = gds->gRect;
	r2.width = scaleHMargin; 	r2.height -= scaleVMargin;
	WFrameRect(gds->wdw, &r2);
	WAddTimeLine(gds->wdw, &r2, gds->ymin, gds->ymax, 1, 0, 0, tx, &ntx);
	/* Horizontal (X) scale */
	r2 = gds->gRect;
	r2.x += scaleHMargin;		    r2.width -= scaleHMargin; 	
	r2.y += (r2.height-scaleVMargin);   r2.height = scaleVMargin;
	WFrameRect(gds->wdw, &r2);
	if(gds->nomxmin == gds->nomxmax)  {
	    gds->nomxmin = gds->xmin;	gds->nomxmax = gds->xmax;
	}
	if(gds->xislog && gds->nomxmin != 0.0)
	    WAddLogTimeLine(gds->wdw, &r2, gds->nomxmin, gds->nomxmax, 0, 
			    1, 1, tx, &ntx);
	else
	    WAddTimeLine(gds->wdw, &r2, gds->nomxmin, gds->nomxmax, 0, 
			 1, 1, tx, &ntx);

	/* Trace */
	r2 = gds->gRect;
	r2.x += scaleHMargin;		    r2.width -= scaleHMargin; 	
	r2.height -= scaleVMargin;
	for(i = 0; i<gds->datapts; ++i)  {
	    if(i+1 > gds->xmin && i <= gds->xmax) {
		x = MapFtoI((double)i, gds->xmin, gds->xmax, 
			    r2.x, r2.x+r2.width);
		y = MapFtoI(gds->data[i], gds->ymin, gds->ymax, 
			    r2.y+r2.height, r2.y);
		if(firstPt) {	
		    WMoveTo(gds->wdw, x, y);
		    firstPt = 0;
		}
		else 
		    WLineTo(gds->wdw, x, y);
	    }
	}
	if(gds->mouseMode == MSMD_XHAIR)  {
	    gds->mouseMode = MSMD_NONE;     /* not there: just cleared it */
	    gdsDrawXHair(gds, gds->xHairX);
	}
    }
}

/************** double-int conversions from dspBench/db.tgdrw.c ******/
double	FloatPropX(whole,num,denom)
    FLOATARG   whole;
    long    num,denom;
    {
    if(num == 0L)
	return(0.0);
    else
	return( whole*( .0  + (double)num)/(double)denom );
    }

long	LongPropY(whole,num,denom,gran)
    long    whole;
    FLOATARG   num,denom;
    int     gran;	/* granularity i.e. minimum incremement of 'whole' */
    {
    return(((long)gran) * (long)( .5 + (num/denom) * (double)(whole/(long)gran) ) );
    }

int MapFtoI(f, fl, fu, il, iu)
    FLOATARG f, fl, fu;		/* value in range */
    int	il, iu;			/* fl and fu map to these ints */
    {
    return il + (int)LongPropY( (long)(iu - il), (f-fl), (fu-fl), 1);
    }

double MapItoF(i, il, iu, fl, fu)
    int	i, il, iu;
    FLOATARG fl, fu;
    {
    return fl + FloatPropX( fu-fl, (long)(i-il), (long)(iu-il));
    }

int gdsYofX(GRAPHDAT_STR *gds, FLOATARG act_x, double *pact_y)
{	/* return the real Y from dataset for real X given */
    FLOATARG	act_y  = 0.0;
    int		inRange = 0;
    int 	ix;
    
    if(gds->xislog && gds->nomxmin != 0.0)
	ix = (int)(gds->xmin + (gds->xmax - gds->xmin)
		   *log(act_x / gds->nomxmin) 
			/ log(gds->nomxmax / gds->nomxmin));
    else
	ix = (int)(gds->xmin + (gds->xmax - gds->xmin)
	       *(act_x - gds->nomxmin)/(gds->nomxmax - gds->nomxmin));
    if(ix >= 0 && ix < gds->datapts)  {
	*pact_y = gds->data[ix];
	inRange = 1;
    }
    return inRange;
}
    

int gdsFixupXY(GRAPHDAT_STR *gds, int *px, int *py, 
		double *pact_x, double *pact_y)
{	/* For crosshair, align Y to f(X) ? */
    int x, y;
    double act_x = 0, act_y = 0;
    RECT r2;
    int  rc = 0;

    r2 = gds->gRect;
    r2.x += scaleHMargin;		    r2.width -= scaleHMargin; 	
    r2.height -= scaleVMargin;
    x = *px;
    if(gds->xislog && gds->nomxmin != 0.0)
	act_x = exp(MapItoF(x, r2.x, r2.x+r2.width, log(gds->nomxmin), 
			    log(gds->nomxmax)));
    else
	act_x = MapItoF(x, r2.x, r2.x+r2.width, gds->nomxmin, gds->nomxmax);
    if(rc = gdsYofX(gds, act_x, &act_y))  {  /* returns valid 1/0 */
	*py = MapFtoI(act_y, gds->ymin, gds->ymax, r2.y+r2.height, r2.y);
    }
    *pact_x = act_x;	*pact_y = act_y;

    return rc;
}
    

#define XMARGIN  10	/* between l,r edge of win and graph */
#define TMARGIN  30	/* between top edge of win and graph */
#define BMARGIN  20     /* between bottom edge of win and graph */

void gdsExpose(GRAPHDAT_STR *gds, int x, int y, int width, int height)
{
    BTN_STR *wbs = gds->btns;
    RECT    r1;

    gdsCpyMode(gds);
    /* get size of window */
    WGetWidthHeight(gds->wdw);	/* sadly, need to force update */
    /* fill rect for the expose */
    r1.x = x; r1.y = y; r1.width = width; r1.height = height;
    SetRct( &(gds->gRect), 0, 0, gds->wdw->wWidth, gds->wdw->wHeight);
    if(width == 0 || height == 0)  {	/* expose whole thing */
	r1 = gds->gRect;
    }
/*    fprintf(stderr, "gdsExpose: x %d y %d w %d h %d\n", 
	    x, y, width, height);
 */ 
    InsetRct(&(gds->gRect), XMARGIN, (TMARGIN+BMARGIN)/2);
    OffsetRct(&(gds->gRect), 0, (TMARGIN-BMARGIN)/2);

    WClearRect(gds->wdw, &r1);

/*    while(wbs) {
	if(Overlap(&r1, &(wbs->rect)))
	   gdsDrawBtn(gds, wbs, 0);
	wbs = wbs->next;
    }
 */  /* button drawing handled by different event handler now */
    if(RectOverlap(&r1, &(gds->gRect)))
	gdsRedrawGraph(gds, &r1);
}

BTN_STR *gdsFindButn(GRAPHDAT_STR *gds, int x, int y)
{
    BTN_STR 	*wbs = gds->btns;
    int 	found = 0;

    while(!found && wbs) {
	if(wbtnPtIn(wbs, x, y))
	    found = 1;
	else
	    wbs = wbtnNext(wbs);
    }
    return wbs;
}

int gdsDrawXHair(GRAPHDAT_STR *gds, FLOATARG act_x)
{	/* draw a cross hair across the gRect at real x value */
    double act_y = 0.0;
    char  msg[32];
    int   x,y;
    RECT  r;
    int   rc = 0;

    if(gds->mouseMode == MSMD_XHAIR) { /* remove old xhair */
	gds->mouseMode = MSMD_NONE;
	gdsDrawXHair(gds, gds->xHairX);
    }
    gds->xHairX = act_x;
    if(rc = gdsYofX(gds, act_x, &act_y))  {	/* only if X in range */
	gdsXorRedMode(gds);   
	/* Report text in a box below the trace */
	SetRct(&r, gds->gRect.x, gds->gRect.y+gds->gRect.height+2,
	        gds->gRect.width, scaleVMargin);
	WClearRect(gds->wdw, &r);
	sprintf(msg, "x: %.6f, y: %.6f", act_x, act_y);
	WWriteIn(gds->wdw, &r, msg, LALIGN);
	/* draw up a crosshair */
	r = gds->gRect;	    r.x += scaleHMargin;    r.width -= scaleHMargin;
	r.height -= scaleVMargin;
	if(gds->xislog && gds->nomxmin != 0.0)
	    x = MapFtoI(log(act_x), log(gds->nomxmin), 
				log(gds->nomxmax), r.x, r.x+r.width);
	else
	    x = MapFtoI(act_x, gds->nomxmin, gds->nomxmax, r.x, r.x+r.width);
	y = MapFtoI(act_y, gds->ymin, gds->ymax, r.y+r.height, r.y);
	if(PtInRct(x,y, &r))  {
	    WDrawLine(gds->wdw, r.x, y, r.x+r.width, y);
	    WDrawLine(gds->wdw, x, r.y, x, r.y+r.height);
	}
	gds->mouseMode = MSMD_XHAIR;
    }
    else {
	gds->mouseMode = MSMD_NONE;
    }
    return rc;
}

int gdsRFdrwxhr(GRAPHDAT_STR *gds, void *data)
{	/* allow DrawXHair to be called from applyRIng */
    return gdsDrawXHair(gds, *(double *)data);
}

int gdsSetXHair(GRAPHDAT_STR *gds, int x, int y)
{	/* Convert a mouse position into an xhair value */
    double a_x, a_y;
    char  msg[32];
    RECT  r;
    int   rc = 0;
    
/*    if(rc  = gdsFixupXY(gds, &x, &y, &a_x, &a_y))  /* align Y to f(X) ? */
    gdsFixupXY(gds, &x, &y, &a_x, &a_y);
    rc = gdsApplyRing(gds, gdsRFdrwxhr, (void *)&a_x);
/*	gdsDrawXHair(gds, a_x);	*/
    return rc;
}

void SelectX(WINDAT *wdptr, double x)
{	/* Set up the xhair at some program-specified ordinate */
    GRAPHDAT_STR *gds  = (GRAPHDAT_STR *)wdptr->windid;
    int rc;

    rc = gdsApplyRing(gds, gdsRFdrwxhr, (void *)&x);
    return;
}

void gdsXHairBtn(GRAPHDAT_STR *gds, int x, int y)
{
/*    gdsXorRedMode(gds);	*/
/*    if(gds->mouseMode == MSMD_XHAIR)  */
/*	gdsApplyRing(gds, gdsRFdrwxhr, (void *)&(gds->xHairX));	     */
/*	gdsDrawXHair(gds, gds->xHairX);	*/
/*    if(gdsSetXHair(gds, x, y))  {
/*	gds->lastMsX = x;	gds->lastMsY = y;
/*	gds->mouseMode = MSMD_XHAIR;
/*    }
/*    else
/*	gds->mouseMode = MSMD_NONE;  */
    gdsSetXHair(gds, x, y);
}

void gdsEndXHair(GRAPHDAT_STR *gds)
{
    if(gds->mouseMode == MSMD_XHAIR)  {
	gdsXorRedMode(gds);
	gdsDrawXHair(gds, gds->xHairX);
    }
    gds->mouseMode = MSMD_NONE;
}

void gdsMouseDown(GRAPHDAT_STR *gds, int ev, int btn, int state, int x, int y)
{
    BTN_STR	*wbs;

/*    wbs = gdsFindButn(gds, x, y);
    if(wbs && wbs->cbFn)  {
	if( (ev == ButtonPress) && (btn == Button1) )
	    gdsDrawBtn(gds, wbs, 1);
	else if ( (ev == ButtonRelease) && (btn == Button1) ) { 
	    gdsDrawBtn(gds, wbs, 0);
	    (*(wbs->cbFn))(gds->wdw, wbs, wbs->cbData);
	}
    }
 */  /* btn functions handled elsewhere */
    if(PtInRct(x, y, &(gds->gRect)))  {
	if( (ev == ButtonPress) && (btn == XHAIRBUTTON) ) {
	    gdsXHairBtn(gds, x, y);
	}
	else if( (ev == ButtonRelease) && (btn == XHAIRBUTTON) ) {
	    /* gdsEndXHair(gds) */ ;
	}
    }
}

void gdsMouseMove(GRAPHDAT_STR *gds, unsigned int bstate, int x, int y)
{
static BTN_STR 	*lastBtn;
    BTN_STR	*thisBtn;

/*    thisBtn = gdsFindButn(gds, x, y);
    if(thisBtn != lastBtn) {
	if(lastBtn != NULL)
	    gdsDrawBtn(gds, lastBtn, 0);
	if((bstate & Button1Mask) && thisBtn != NULL)
	    gdsDrawBtn(gds, thisBtn, 1);
    }
    lastBtn = thisBtn;
 */	/* buttons handled elsewhere */

    if(PtInRct(x, y, &(gds->gRect)))  {
	if(bstate & XHAIRBUTTONMASK) {
	    gdsXHairBtn(gds, x, y);
	}
    }
}

void gdsResize(GRAPHDAT_STR *gds, int width, int height)
{	/* NOT catching resize at present - doesn't seem to work fully */
    WResizeWindow(gds->wdw, width, height);
}

GRAPHDAT_STR *gdsFind(Window win)
{
    GRAPHDAT_STR *gds = gdRoot;
    int 	found = 0;

    while(gds && !found) {
	if(gds->wdw->win == win)	found = 1;
	else				gds = gds->next;
    }
    return gds;
}

/* typedef int (*eventHandlerFn)(XEvent *event, void *data);  */
/* returns 1 if the event was handled, 0 if unrec */

static eventHandlerFn  extEvFn = NULL;	/* only allow one at present */
static void *evHData;			/* corresponding data */

void gdInstallEventHandler(eventHandlerFn evHandler, void *evData)
{	/* for other x-oriented events, can install handler in our loop */
    extEvFn = evHandler;
    evHData = evData;
}

int gdsMapEventOnButtons(GRAPHDAT_STR *gds, XEvent *event)
{	/* offer this event to each of the buttons under a gds.
	   Return 1 if any of them took it. */
    BTN_STR 	*wbs = gds->btns;
    int 	rc = 0;

    while(rc == 0 && wbs)  {
	if( (rc = wbtnEvent(event, wbs)) == 0)
	    wbs = wbtnNext(wbs);
    }
    return rc;
}

int gdEvent(XEvent *event, GRAPHDAT_STR *gds)	/* args rvsd as gen.ev.fn */
{
    int rc = 0;

    gds = gdsFind(event->xany.window);
	if(gds == NULL)	{
	    if(extEvFn) {
		if( (rc = (*extEvFn)(event, evHData)) == 0)  {
		    fprintf(stderr, "extFn returned 0\n");
		}
	    }
	    else {
		fprintf(stderr, "gdEvLp: no win for event\n");
	    }
	}
	else{
	    switch(event->type)  {
	    case Expose:
		gdsExpose(gds, event->xexpose.x, event->xexpose.y, 
			  event->xexpose.width, event->xexpose.height);
		gdsMapEventOnButtons(gds, event);
		rc = 1;
		break;
	    case ResizeRequest:	/* not currently enabled - was weird */
		gdsResize(gds, event->xresizerequest.width, 
			  event->xresizerequest.height);
		rc = 1;
	    case ButtonPress:
	    case ButtonRelease:
		if(!gdsMapEventOnButtons(gds, event))
		   gdsMouseDown(gds, event->type, event->xbutton.button, 
				event->xbutton.state, 
				event->xbutton.x, event->xbutton.y);
		rc = 1;
		break;
	    case MotionNotify:
		if(!gdsMapEventOnButtons(gds, event))
		    gdsMouseMove(gds, event->xmotion.state, 
				 event->xmotion.x, event->xmotion.y);
		rc = 1;
		break;	
	    case ClientMessage:
		/* why are we getting THESE? from mwm, it seems */
		rc = 1;
		break;
	    default:
		fprintf(stderr, "gdEvLp: got gds but unrec ev %d\n", 
			event->type);
	    }
	}
    return rc;
}

#define MYX_ANY_EVENT (-1L)

void gdEventLoop(void)	/* clear events in any of our windows but no block */
{
    XEvent	event;
    GRAPHDAT_STR *gds;

    fprintf(stderr, "gdEventLoop FROM HELL\n");
    WEventLoop();
    return;

    while( 0
/* XCheckMaskEvent(xdisp, MYX_ANY_EVENT, &event) */
          ) {
	gds = gdsFind(event.xany.window);
	if(gds == NULL)	{
	    if(extEvFn) {
		if((*extEvFn)(&event, evHData) == 0)  {
		    fprintf(stderr, "extFn returned 0\n");
		}
	    }
	    else {
		fprintf(stderr, "gdEvLp: no win for event\n");
	    }
	}
	else{
	    switch(event.type)  {
	    case Expose:
		gdsMapEventOnButtons(gds, &event);
		gdsExpose(gds, event.xexpose.x, event.xexpose.y, 
			  event.xexpose.width, event.xexpose.height);
		break;
	    case ResizeRequest:	/* not currently enabled - was weird */
		gdsResize(gds, event.xresizerequest.width, 
			  event.xresizerequest.height);
	    case ButtonPress:
	    case ButtonRelease:
		if(!gdsMapEventOnButtons(gds, &event))
		   gdsMouseDown(gds, event.type, event.xbutton.button, 
				event.xbutton.state, 
				event.xbutton.x, event.xbutton.y);
		break;
	    case MotionNotify:
		if(!gdsMapEventOnButtons(gds, &event))
		    gdsMouseMove(gds, event.xmotion.state, 
				 event.xmotion.x, event.xmotion.y);
		break;	
	    default:
		fprintf(stderr, "gdEvLp: got gds but unrec ev\n");
	    }
	}
    }
}

void SetSloppyUpdate(WINDAT *wdptr, int n)
{	/* sloppy update doesn't update label or scale till next DoDisp */
    GRAPHDAT_STR *gds  = (GRAPHDAT_STR *)wdptr->windid;

    if(n)  {
	gds->sloppyUpdate = TRUE;
    }
    else  {
	gds->sloppyUpdate = FALSE;
    }
}

void SetLabel(WINDAT *wdptr, char *s)
{
    GRAPHDAT_STR *gds  = (GRAPHDAT_STR *)wdptr->windid;

    strcpy(gds->label, s);
    if(!gds->hidden && !gds->sloppyUpdate)
	wbtnDraw(gds->btns);
/*	gdsExpose(gds, gds->btns->rect.x, gds->btns->rect.y, 
		  gds->btns->rect.width, gds->btns->rect.height);  */
	/* WClearExpose(gds->wdw, &(gds->btns->rect));  */
}

void SetLogScale(WINDAT *wdptr, int islog)
{	/* *maybe* flip the x scale into log units */
    GRAPHDAT_STR *gds  = (GRAPHDAT_STR *)wdptr->windid;

    gds->xislog = islog;
    if(!gds->hidden && !gds->sloppyUpdate)
	gdsExpose(gds, gds->gRect.x, gds->gRect.y, 
		  gds->gRect.width, gds->gRect.height);
}

void SetXLimits(WINDAT *wdptr, FLOATARG low, FLOATARG hi)
{
    GRAPHDAT_STR *gds  = (GRAPHDAT_STR *)wdptr->windid;

    gds->nomxmin = low; gds->nomxmax = hi;
    if(!gds->hidden && !gds->sloppyUpdate)
	gdsExpose(gds, gds->gRect.x, gds->gRect.y, 
		  gds->gRect.width, gds->gRect.height);
}

static GRAPHDAT_STR *lastGds = NULL;

void DrawGraph(wdptr)
    WINDAT	*wdptr;
{
    float	*fdata = wdptr->fdata;
    long 	npts   = wdptr->npts;
    char	*msg   = wdptr->caption;
    int 	pol    = wdptr->polarity;
    GRAPHDAT_STR *gds  = (GRAPHDAT_STR *)wdptr->windid;
    RECT	r;

    if(npts > gds->maxpts)  {
	if(gds->data)	free(gds->data);
	gds->data = TMMALLOC(float, npts, "wndwDrawGraph");
	gds->maxpts = npts;
    }
    CopyFloats(fdata, 1, gds->data, 1, npts, FWD);
    gds->datapts = npts;
    if(pol == BIPOL)  {
	gds->ymin = -(wdptr->oabsmax);
	gds->ymax = wdptr->oabsmax;
    }
    else if (pol == NEGPOL) {
	gds->ymin = -(wdptr->oabsmax);
	gds->ymax = 0;
    }
    else {  /* assu POSPOL */
	gds->ymin = 0;
	gds->ymax = wdptr->oabsmax;
    }

    if(!gds->hidden)  {
	if(gds->sloppyUpdate)  {
	    gdsExpose(gds, 0, 0, 0, 0);
	    /* WClearExpose(gds->wdw, &(gds->gRect)); */
	}
	else {
	    gdsExpose(gds, gds->gRect.x, gds->gRect.y, 
		      gds->gRect.width, gds->gRect.height);
	    /* WClearExpose(gds->wdw, &(gds->gRect)); */
	}
    }
    if(gds->pauseOnDraw) {
	gds->pauseOnDraw = FALSE;
	/* pass it the actual button to highlight!!!! (step) */
	PauseBtnFn(gds->wdw, gds->pauseBtn, gds, -1);
    }
    WEventLoop();
    lastGds = gds;
}

int ExitGraph(void)
{
    if(lastGds)  {
	StepBtnFn(lastGds->wdw, lastGds->pauseBtn, lastGds, -1);
    }
}

#ifdef CSOUND
void MakeXYin(wdptr,x,y)
    XYINDAT *wdptr;
    float	x,y;	/* initial proportions */
    {
	MkXYDummy(wdptr,x,y);
    }

void ReadXYin(wdptr)
    XYINDAT *wdptr;
    {
	RdXYDummy(wdptr);
	}
#endif
