/***************************************************************\
* wndw.h							*
* header file for wndw.c					*
* Moderately device independent drawing primitives		*
* Essentially, allows XWindow draws to be diverted to   	*
* PostScript, rather than endless screendumps.			*
* Based on stripk		 				*
* Generalized dpwe 1992oct24					*
\***************************************************************/

#ifndef _WNDW_H_
#define _WNDW_H_

#include <dpwelib.h>
#include <arrayops.h>	/* for WColorArray */
#include <X11/Xlib.h>

/* alignment codes for WWriteIn */
#define LALIGN	-1
#define CALIGN	0
#define RALIGN	1

/* codes for WSetGMode */
/* mode */
#define WGMD_COPY	0
#define WGMD_XOR		1
/* linestyle */
#define WGMD_SOLID	0
#define WGMD_DOTS	1
/* colors */
#define WGMD_BLACK	0
#define WGMD_RED 	1
#define WGMD_GREEN	2
#define WGMD_YELLOW	3
#define WGMD_BLUE	4
#define WGMD_MAGENTA	5
#define WGMD_CYAN	6
#define WGMD_WHITE	7
#define WGMD_NCOLORS	8

typedef int (*eventHandlerFn)(XEvent *event, void *data);
/* returns 1 if the event was handled, 0 if unrec */

typedef struct _evh_list {		/* node in list of event handlers */
    struct _evh_list	*next;		/* next one in list */
    eventHandlerFn	evhFn;		/* function to call with event */
    void		*evhData;	/* 2nd arg to that call */
} EVHDLR_NODE;

typedef struct wind
    {
    struct wind *next;		/* (keep list of them all) */
    Display	*disp;
    Window	win;
    int 	scrn;
    GC		gc;
    GC		altgc;		/* two gc's for one level of context saving */
    long 	fg, bg;
    int 	cur_x, cur_y;
    int 	wWidth, wHeight;	/* after resize */
    int 	hidden;
    float	linew;
    FILE	*opf;		/* ps output file? why not! */
    EVHDLR_NODE *evhs;		/* linked list of event handler fn strs */
    } WIND;

typedef struct rect
    {
    int 	x,y;
    int 	width,height;
    } RECT;

void WHide PARG((WIND *wind));	    /* make this window invisible */
void WUnhide PARG((WIND *wind));    /* make a hidden window visible again */ 
void WSetGMode PARG((WIND *wind, int fg, int bg, int mode, 
		     int lwidth, int lstyle));
WIND *NewWind PARG((Display *disp, int scrn, Window win, char *idraw));
void FreeWind PARG((WIND *wind));
void WNewName(WIND *wndw, char *name);
void WSetXdisp PARG((Display *disp));
    /* short cut to prevent multiple XOpenDisplays */
int WInitX PARG((void));
    /* intialize the X server */
void WFlush PARG((void));
    /* flush the x events */
Window MakeXWindow PARG((char *name, 
			 int x, int y, int w, int h));
WIND *NewXWind PARG((int x, int y, int w, int h, 
		     char *name, char *idFname));
void WBlitArea PARG((WIND *wndw, RECT *pr, int right, int down));
    /* use blitting to move the contents of the specified rectangle 
       by the specified amounts.  Negative args give left and 
       up shifts respectively.  The result will not move outside 
       of <*pr>, so the region shifted is reduced accordingly.  
       If <pr> is NULL, the entire window is shifted. */
void WSetBlack PARG((WIND *wind));
void SetLineWidth PARG((WIND *wind, FLOATARG width));  /* for postscript */
void WGetWidthHeight PARG((WIND *wind));
void WResizeWindow PARG((WIND *wind, int width, int height));
void WSetClip(WIND *wind, RECT *pr);
   /* Clip graphics to within rect, or no clipping if <pr> NULL */
void WMoveTo PARG((WIND *wind, int x, int y));
void WLineTo PARG((WIND *wind, int x, int y));
void WDrawLine PARG((WIND *wind, int x1, int y1, int x2, int y2));
void WDrawX PARG((WIND *wind, RECT *pr));
void WPolyLine PARG((WIND *wind, int buf[][2], long cnt));
void WClearRect PARG((WIND *wind, RECT *pr));
void WClearExpose PARG((WIND *wind, RECT *pr));
void WFillRect PARG((WIND *wind, RECT *pr));
void WColorRect PARG((WIND *wind, RECT *pr, unsigned long color));
    /* I'm sorry.  Color a rectangle in a color gotten from 
       the pseudocolor lookup table - NOT from the other color method */
void WFrameRect PARG((WIND *wind, RECT *pr));
void WSetClipRect PARG((WIND *wind, RECT *pr));
void WWriteAt PARG((WIND *wind, int x, int y, char *str));
void WWriteIn PARG((WIND *wind, RECT *pr, char *str, int lcrAlign));
int  WTextWidth PARG((WIND *wind, char *str));
int  WTextHeight PARG((WIND *wind, char *str));
void WAddTimeLine PARG((WIND *wind, RECT *pr, FLOATARG t1, FLOATARG t2, 
			int vnh, int sns, int und, int tix[], int *pntx));
void WAddLogTimeLine PARG((WIND *wind, RECT *pr, FLOATARG t1, FLOATARG t2, 
			   int vnh, int sns, int und, int tix[], int *pntx));
int WInitColormap PARG((int gray, int onebit));
	/* setup the pseudocolor colormap.
	   If <gray>, make it true shades of gray, not pseudocolor.
	   If <onebit>, reduce to a single color. 
	   Return zero on failure (typically run out of colormap space). */
int WGetPixelTable PARG((unsigned long **pixels, int *ngrays));
	/* retrieve the size and address of the (pseudo) colormap.
	   Returns zero if no colormap allocated yet. */
unsigned long WPixelForByte PARG((int byte));
	/* Map an unsiged byte (0..255) into a pixel value suitable 
	   for.. well.. XSetColor.. um.  This is inefficient for 
	   bulk use, but serves as an example.  */
void WColorArray PARG((WIND *wind, RECT *pr, ByArray *array));
	/* Convert the array <*arry> (<rows> rows of <cols> bytes) into 
	   a pseudocolored region filling <*pr>, using the prevailing colormap. 
	   ACTUALLY QUANTIZES THE MATRIX IN-PLACE TO THE NUMBER OF AVAI GRAYS */
void FltToString PARG((FLOATARG f, char *s, int dp));
void NmToString PARG((long n, char *s));
void SetRct PARG((RECT *pr, int x, int y, int width, int height));
int RangeOverlap PARG((int l1, int w1, int l2, int w2));
    /* is there overlap between ranges l1..l1+w1 and l2..l2+w2 ? */
int PtInRct PARG((int x, int y, RECT *pr));
    /* does the specified point lie inside the RECT? */
int RectOverlap PARG((RECT *pr1, RECT *pr2));
    /* do the rects have nonzero intersection? */
int WSetEventPause PARG((int newPause));
    /* set the event loop to return control or not */
int WToggleEventPause PARG((void));
    /* just pause <=> no pause */
int WEventLoop PARG((void));
    /* handle an event but do not block;  return 0 if event not handled */
void WAddEventHandler PARG((WIND *wndw, eventHandlerFn evhFn, void *evhData));
    /* add an event handler for this wndw (or unrec wndw if wndw == NULL) */
int WGetShiftEtc(WIND *wndw);
    /* return a bit field including the bottom bit as the shift key */

#endif /* !_WNDW_H_ */
