/***********************************************************************\
* wndwbtn.c								*
* Installable buttons for wndw windows					*
* dpwe 1992oct24							*
\***********************************************************************/

#include <dpwelib.h>
#include <wndw.h>

typedef struct _btn_str {	/* little drawn button structure */
    struct _btn_str	*next;
    WIND		*wdw;
    RECT		rect;
    char		*label;
/*    btnFn		cbFn;	*/ /* circular definition lose */
    void		(*cbFn)PARG((WIND *wdw, struct _btn_str *btn, 
				      void *data, int xbtn));
    void		*cbData;
    int			state;	/* allows us to remember `armed' state */
    int 		hilit;	/* if true, draw button rvs video */
} BTN_STR;

#define __BTN_STR_DEFD__

#include "wndwbtn.h"

void NullBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{
    DBGFPRINTF((stderr, "Button %s pressed\n", btn->label));
}

void HideBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{
    WHide(wdw);
}

void QuitBtnFn(WIND *wdw, BTN_STR *btn, void *data, int xbtn)
{
    exit(0);
}

#define BTN_MARGIN	4

BTN_STR *wbtnNew(BTN_STR *next, WIND *wdw, int x, int y, int width, int height,
		 char *label, btnFn fn, void *fnData)
{	/* width and height can be zero -> sizes to accommodate text */
    BTN_STR *wbs;

    wbs = TMMALLOC(BTN_STR, 1, "wbtnNew");
    wbs->next = next;
    wbs->wdw  = wdw;
    wbs->label = label;
    wbs->rect.x = x;		wbs->rect.y = y;
    wbs->rect.width = width;	wbs->rect.height = height;
    if(width == 0)
	wbs->rect.width = WTextWidth(wbs->wdw, wbs->label) + BTN_MARGIN;
    if(height == 0)
	wbs->rect.height = WTextHeight(wbs->wdw, wbs->label) + BTN_MARGIN;
    if(fn == NULL)
	fn = NullBtnFn;
    wbs->cbFn = fn;	wbs->cbData = fnData;
    wbs->state = WBTN_QUIET;
    wbs->hilit = 0;

    return wbs;
}

void wbtnDraw(BTN_STR *wbs)
{
    RECT	r;

    WPushGC(wbs->wdw);
    WSetGMode(wbs->wdw, WGMD_BLACK, WGMD_WHITE, WGMD_COPY, 1, WGMD_SOLID);
    r.x = wbs->rect.x;		r.y = wbs->rect.y;
    r.width = wbs->rect.width;	r.height = wbs->rect.height;
    WClearRect(wbs->wdw, &r);
    WFrameRect(wbs->wdw, &r);
    WWriteIn(wbs->wdw, &r, wbs->label, CALIGN);
    if(wbs->state == WBTN_ARMED) {  /* armed */
	RECT r2;

	r2 = r;
	r2.x += 1; r2.y += 1; r2.width -= 2; r2.height -= 2;
	WFrameRect(wbs->wdw, &r2);	/* thicken */
    }
    if(wbs->hilit)  {	/* reverse vid */
	WSetGMode(wbs->wdw, WGMD_BLACK, WGMD_WHITE, WGMD_XOR, 1, WGMD_SOLID);
	WFillRect(wbs->wdw, &r);
	WSetGMode(wbs->wdw, WGMD_BLACK, WGMD_WHITE, WGMD_COPY, 1, WGMD_SOLID);
    }
    WPopGC(wbs->wdw);
}

BTN_STR *wbtnNext(BTN_STR *wbs)
{	/* enable cdr-ing down a button list */
    return wbs->next;
}

void wbtnGetRect(BTN_STR *wbs, RECT *pr)
{	/* query the rectangle enclosing this button */
    *pr = wbs->rect;
}

int wbtnPtIn(BTN_STR *wbs, int x, int y)
{	/* is the point in our button? */
    return PtInRct(x, y, &(wbs->rect));
}

void wbtnHilite(BTN_STR *wbs, int hilit)
{	/* set the state of the button */
    if(hilit != wbs->hilit)  {
	wbs->hilit = hilit;
	wbtnDraw(wbs);
    }
}

void wbtnSetLabelPtr(BTN_STR *wbs, char *ptr)
{	/* point the label to be printed at some other string */
    wbs->label = ptr;
}

int wbtnEvent(XEvent *event, BTN_STR *wbs)	/* args rvsd as gen.ev.fn */
{	/* somebody wanted to show us this event.  If take it, return 1
	   else 0 (i.e. give to someone else) */
    int ev, btn, state, x, y;
    int rc = 0;

    if(event->xany.window == wbs->wdw->win)  {	/* at least the right window */
	if(event->type == Expose)  {
	    RECT  rect;

	    /* check rect -YAWN- */
	    SetRct(&rect, event->xexpose.x, event->xexpose.x, 
		   event->xexpose.width, event->xexpose.height); 
	    if(RectOverlap(&rect, &(wbs->rect)))  {
		wbtnDraw(wbs);
		/* rc = 1; */		/* ?? others may care about expose */
	    }
	} else if(event->type == ButtonPress 
		  || event->type == ButtonRelease 
		  || event->type == MotionNotify)  {	/* only others we do */
	    if(event->type == MotionNotify)  {
		x = event->xmotion.x;	
		y = event->xmotion.y;
		state = event->xmotion.state;
	    } else { 
		x = event->xbutton.x;	
		y = event->xbutton.y;
		state = event->xbutton.state;
		btn = event->xbutton.button;
	    }
	    if(wbtnPtIn(wbs, x, y)) {	/* *does* concern us */
		rc = 1;		/* accept the event */
		if(event->type == ButtonRelease)  {
		    /* release inside button == fire */
		    if(wbs->cbFn)	/* do callback */
			(*(wbs->cbFn))(wbs->wdw, wbs, wbs->cbData, btn);
		    wbs->state = WBTN_QUIET;
		    wbtnDraw(wbs);
		} else { 	/* button press or motion - hilite */
		    if(wbs->state != WBTN_ARMED)  {
			wbs->state = WBTN_ARMED;
			wbtnDraw(wbs);
		    }
		}
	    } else if(wbs->state == WBTN_ARMED) { /* but not in -> disarm */
		wbs->state = WBTN_QUIET;
		wbtnDraw(wbs);
	    }
	}
    }
    return rc;
}
