/************************************************************************/
/*  fdsyn.c	(after timex.c)						*/
/*  NeXT, FL POINT VERSION						*/
/*  Frequency Domain Synthesis	(after fdan.c)				*/
/*  Takes a freq domain FFTfile and converts it into a soundfile.	*/
/*  May perform timescale modification the while.			*/
/*  dpwe  25jan90							*/
/************************************************************************/

#include <stdio.h>
#include <math.h>

#ifdef THINK_C
#include <stdlib.h>		/* for malloc if needed, and exit() proto */
#include <console.h>
#include <tc_main.h>
#else
#include <malloc.h>
#endif
#include <dsputil.h>
#include <fft.h>
#include <sndf.h>
#include <pvoc.h>

/* extern fn protos */

/* Static function prototypes */
static void CreateNewSound(PVSTRUCT *inputPVH,
	    SFSTRUCT **outputSound); /* nb - 2 stars! */
static void resynthSound(PVSTRUCT *inputPVH,
	    SFSTRUCT *outputSound); /* nb - one star! */

/* Global variables */
int	    sfErr;
char	    *programName;
int	    debug = 0;		/* tweaked internally */
long	    inFrInc;		/* hack : now a global */

/* Static global variables */
static	float	tex = 1.0;	/* time extension factor */
static	float	pex = 1.0;	/* pitch expansion factor */
static	int	frOvFac = 4;	/* desired number of frames to overlap */
static	long	frameSize;	/* must match analysis file */
static	long	frameIncr;	/* will be frameSize / frOvFac */
static	float	fiex; /* frame inverse expn factor - hack to pass btwn funcs */
static	complex *plut;		/* lookup table for fft */
static	int	specwp = 0;	/* flag to invoke spectral warping */

main(argc,argv)
    int argc;
    char *argv[];
    {
    PVSTRUCT	    *inputPVH;
    SFSTRUCT	    *outputSound;
    char    *inputPvocFile = NULL, *outputSoundFile = NULL;
    int     i,err;

    /* Check arguments */
    programName = argv[0];
    if(argc==1)
	{
	fprintf(stderr,
		"Usage: %s [-t n][-p n] <inputFFTfile> <outputSoundFile>\n",
		programName);
	exit(1);
	}
    for(i=1;i<argc;++i)
	{
	if(argv[i][0] == '-')
	    switch(argv[i][1])
		{
	    case 't':
		tex = atof(argv[++i]);
		break;
	    case 'p':
		pex = atof(argv[++i]);
		break;
	    default:
		fprintf(stderr,"%s: bad option %s\n", programName, argv[i]);
		exit(1);
		}
	else if(inputPvocFile == NULL)
	    inputPvocFile = argv[i];
	else
	    outputSoundFile = argv[i];
	}

    if(inputPvocFile == NULL)
	{
	fprintf(stderr,"%s: no input filename\n", programName);
	exit(1);
	}

    if(outputSoundFile == NULL)
	{
	fprintf(stderr,"%s: no output filename\n", programName);
	exit(1);
	}

    if ( (pex < 0.1) || (pex > 10) )
	{
	fprintf(stderr, "%s: %f is a bad pitch expansion factor\n",
		programName, pex);
	exit(1);
	}

    if ( (tex < 0.1) || (tex > 30) )
	{
	fprintf(stderr, "%s: %f is a bad time extension factor\n",
		programName, tex);
	exit(1);
	}

    if ( (tex*pex < 0.1) || (tex*pex > 30) )
	{
	fprintf(stderr, "%s: %f is a bad time-pitch expansion product\n",
		programName, tex*pex);
	exit(1);
	}

    /* Read sound file */
    if (err = PVReadFile(inputPvocFile, &inputPVH))
	{
	fprintf(stderr, "Infile:%s\n",inputPvocFile);
	fprintf(stderr, "%s: %s\n", programName, PVErrMsg(err));
	exit(1);
	}
    /* Abort if unsupported format */
    if (inputPVH->format != PVFLOAT)
	{
	fprintf(stderr, "%s: Unsupported data format: %d\n",
		programName, inputPVH->format);
	exit(1);
	}
    CreateNewSound(inputPVH, &outputSound);

    plut = AssignBasis((complex *)NULL,frameSize); /* set up FFT tables */
    MakeSinc();			/* set up sinc table for UDSamp */

    resynthSound(inputPVH, outputSound);

    /* Write sound file */
    if (sfErr = SFWrite(outputSoundFile, outputSound))
	{
	printf("%s: sound error: %s on %s\n", programName,
	       SFErrMsg(sfErr), outputSoundFile);
	exit(1);
	}

    /* CloseFft();  /* release the DSP, perhaps (if FFT uses it) */
    DestroySinc();
    }

static void CreateNewSound(inputPVH,outputSound)
    PVSTRUCT	   *inputPVH;
    SFSTRUCT	    **outputSound;
    {
    long    inBSize  = inputPVH->dataBsize; /* B is for Bytes */
    int     format   = SFMT_SHORT;
    float   rate     = inputPVH->samplingRate;
    int     inChans  = inputPVH->channels;
    long    inFrames, outFrames, outBSize;
    int     outChans;
    long    inFrSiz  = inputPVH->frameSize;
/*    long     inFrInc  = inputPVH->frameIncr;	now a global */
    long    origSampls, outSampls;
    short   *sPtr;

    inFrInc  = inputPVH->frameIncr;
    frameSize = inFrSiz;
    frameIncr = frameSize/frOvFac; /* as we intended */
    outChans = inChans;
    printf("inChans  : %d \t outChans  : %d \n",inChans,outChans);
    if (outChans != 1)
	{
	fprintf(stderr,
	  "%s: Can't resynth %d channels (only mono)\n",programName,outChans);
	exit(1);
	}
    inFrames = inBSize / (sizeof(float) * (long)(frameSize+2));
    origSampls = NumSampls(inFrames-1, inFrSiz, inFrInc);
    /* number of original (time) floats represented by analysis file */
    outSampls = (long)(tex * (float)origSampls);	/* or almost ..*/
    outFrames = NumFrames(outSampls, frameSize, frameIncr);
    outSampls = NumSampls(outFrames, frameSize, frameIncr); /* .. exact */
    outBSize  = outSampls*outChans*sizeof(short); /* in bytes! */
    printf("Input  frameSize %ld frameIncr %ld frames %ld osamples %ld \n",
	   inFrSiz, inFrInc, inFrames, origSampls);
    printf("Output frameSize %ld frameIncr %ld frames %ld osamples %ld \n",
	   frameSize, frameIncr, outFrames, outSampls);
    /* Allocate a new sound */
    if(sfErr = SFAlloc(outputSound,outBSize,format,rate,outChans,4))
	{
	printf("%s: sound error: %s\n", programName,
	       SFErrMsg(sfErr));
	exit(1);
	}

    fiex = (float)frameIncr/(tex * (float)inFrInc);
    /* 'fiex' is inverse expansion factor for frames */
    /* Maps output frame number to fraction index to input frames */
    /* return fiex as we have the means to work it out */

    /* Clear sound initially .. (mac doesn't) */
    sPtr = (short *)SFDataLoc(*outputSound);
    while(outSampls--)		*sPtr++ = 0;
    }

/*
 * resynthSound
 *  Rebuild a playable sound, based on the unwrapped FFT data in the input file
 */
static void resynthSound(inputPVH,outputSound)
    PVSTRUCT	    *inputPVH;
    SFSTRUCT	    *outputSound;
    {
    float   *inPtr;
    short   *outPtr;
    long    i, outSampls, outFrames;
    float   *tmpBuf, *tmpBuf2, *outPhs, *winBuf /*, *wBuf */ ;
    float   scaleDown, oneOnFSz;
    float   sampRate = (float)inputPVH->samplingRate;
    float   oneOnPex = 1.0/pex, lastPex = 1.0;
    long    buf2Size,d;

#define B2FACT 1.0 /* 0.4 */
    /* B2FACT is the width of the output window as a factor of the frame */
    /* make < 1.0 for less smearing of sound.  Depends on analysis, tho */

    buf2Size = 2L*(long)(B2FACT*.5*oneOnPex*(float)frameSize); 
    /* must be even */

    /* to scale down data before ifft & olap-adding lots of frames */
    oneOnFSz  = 1.0/((float)frameSize);  /* passed to FFT */
    scaleDown = (2*(float)frameIncr/(float)buf2Size); 

    tmpBuf = MakeBuf(frameSize * 2L);    /* complex ! */
    tmpBuf2= MakeBuf(buf2Size);
    outPhs = MakeBuf(frameSize);
    winBuf = MakeHalfWin(buf2Size, scaleDown, 1); /*	MakeBuf(frameSize); */
    inPtr  = (float *) ((char *) inputPVH  + inputPVH->headBsize);
    outPtr = (short *)SFDataLoc(outputSound);
    outSampls= outputSound->dataBsize/(outputSound->channels*sizeof(short));
    outFrames = NumFrames(outSampls, frameSize, frameIncr);

    printf("resynthSound: scaleDown is %f \n",scaleDown);
    printf("Time ex:%.2f Frames inv ex:%.4f \n",tex,fiex); 
    print ("b1siz %ld  b2siz %ld\n",frameSize, buf2Size);
    FetchIn(inPtr,tmpBuf,frameSize,(float)0);	/* first frame to align phs */
    FrqToPhase(tmpBuf, frameSize, 
	       (FLOATARG)inFrInc,(FLOATARG)sampRate, (FLOATARG)0);
    RewrapPhase(tmpBuf,frameSize,outPhs);
    for (i = 0; i < outFrames; i++)
	{
	FetchIn(inPtr,tmpBuf,frameSize,1+((float)i)*fiex);
	FrqToPhase(tmpBuf, frameSize, 
		   pex*(float)frameIncr,
		   (float)sampRate, /*a*/ (.5 * ( (pex / lastPex) - 1) ));
	/* Offset the phase to align centres of stretched wndws, not starts */
	RewrapPhase(tmpBuf,frameSize,outPhs);
	if(specwp)
	    {
	    PreWarpSpec(tmpBuf, frameSize, pex);
	    }
	Polar2Rect(tmpBuf,frameSize);
	/* force the imag parts of dc and fs/2 to nil - else get whistle */
	tmpBuf[1] = 0; tmpBuf[frameSize+1] = 0;
	FFT2torl((complex *)tmpBuf,frameSize,1,	/*a*/ oneOnFSz, plut);	/* */
	PackReals(tmpBuf, frameSize);
	if(pex != 1.0)
	    {
	    UDSample(tmpBuf,.5*((float)frameSize - pex*(float)buf2Size), 
		     tmpBuf2, frameSize, buf2Size, pex);
	    }
	else
	    CopySamps(tmpBuf+
		      (int)(.5*((float)frameSize - pex*(float)buf2Size)), 
		      tmpBuf2,buf2Size);
	ApplyHalfWin(tmpBuf2, winBuf, buf2Size); /* */
	FixAndAdd(tmpBuf2, outPtr, buf2Size);  

	if((i%20)==0) 
	    {	printf("%ld ", i);   fflush(stdout);	}
	outPtr += frameIncr;
	lastPex = pex;	     /* needs to know last pitchexp to update phase */
	debug = 0;
	}
    }

#define DBGPTS 8
PrintBuf(buf,size,msg)
    float *buf;
    int   size;
    char  *msg;
    {
    int   i;

    if(!debug)
    return;
    size = DBGPTS;
    printf("%s:",msg);
    for(i=0; i<size; ++i)
    printf("%7.2f ",buf[i]);
    printf("\n");
    }

