/************************************************************************/
/*  pva.c	    (after pvanal.c)					*/
/*  Frequency Domain Analysis						*/
/*  Takes a time domain soundfile and converts it into a file of	*/
/*  This version just calculates pure FFT rather than PVOC's phi-dot	*/
/*  More importantly, it reads from disk rather than using so much core */
/*  dpwe  14feb91							*/
/************************************************************************/

#include <stdio.h>

#ifdef THINK_C
#include <tc_main.h>
#include <stdlib.h>
char sfdir_path[] = "Frida:dpwe:sound:";
#else
#include <malloc.h>
#endif

#include <fft.h>
#include <dsputil.h>
#include <sndf.h>
#include <pvoc.h>

/*** Static function prototypes ***/
#ifdef __STDC__			/* prototype arguments */
static FILE *CreatePVfile(char *pvPath, PVSTRUCT **oPVH, SFSTRUCT *iSnd);
static void takeFFTs(SFSTRUCT *inputSound, PVSTRUCT  *outputPVH,
		     FILE *sndFile, FILE *ftFile);
static void PrintBuf(float *buf, int size, char *msg);
static void convertToMono(SFSTRUCT **theSound);
static void CopyShorts(short *sce, int sinc, short *dst, int dinc, 
		       long len, int fwdNbck);
#else				/* no argument prototypes */
static FILE *CreatePVfile();
static void takeFFTs();
static void PrintBuf();
static void convertToMono();
static void CopyShorts();
#endif

#define MINFRMMS	20	/* frame defaults to at least this many ms */
#define MAXFRMPTS	65536
#define MINFRMPTS	16	/* limits on fft size */
#define DFLT_OVLP	4	/* default frame overlap factor */
#define DFLT_EXTN	".aft"	/* default output file extension */

/* Global variables */
int	    sfErr;
char	    *programName;
int	    debug = 0;		/* tweaked inside ! */
int 	    dbgDump = -1;	/* dump out frames to xgraph file */
int	    dbgCont = 10;	/* how many frames to xgr dump */

/* Static global variables */
static	long frameSize	= 0;	    /* size of FFT frames */
static	long frameIncr	= 0;	    /* step between successive frames */

complex *basis;     /* LUTable for FFT */

main(argc,argv)
    int argc;
    char *argv[];
    {
    SFSTRUCT	*inputSound;
    PVSTRUCT	*outputPVH;
    FILE	*sndFile;
    FILE	*ftFile;
    char    *inputSoundFile = NULL;
    char    outputPvocFile[80];
    int     i,ovlp = 4;		/* number of overlapping windows to have */
    int     err = 0;

    outputPvocFile[0] = '\0';
/*    float   a = 3.9;
    fprintf(stderr,"(int)(%5.2f) = %d (int)(%5.2f) = %d\n", 
	    -a, (int)(-a), a, (int)(a) );	*/
    /* Check arguments */
    programName = argv[0];
    frameIncr = frameSize/ovlp;
    for(i=1; i<argc && err==0; ++i)
	{
	if(argv[i][0]=='-')
	    switch(argv[i][1])
		{
	    case 'n':
		if(i<argc-1)
		    {
		    frameSize = atol(argv[++i]);
		    if(frameSize<MINFRMPTS || frameSize>MAXFRMPTS
		       || !IsPowerOfTwo(frameSize))
			{
			fprintf(stderr,"%s:n (%s) must be 2^r and %d..%d\n",
				programName, argv[i], MINFRMPTS, MAXFRMPTS);
			err = 1;
			}
		    }
		else
		    err = 1;
		break;
	    case 'o':
		if(frameIncr == 0 && i<argc-1)
		    ovlp = atol(argv[++i]);
		else
		    err = 1;
		break;
	    case 'i':
		if(ovlp == 0 && i<argc-1)
		    frameIncr = atoi(argv[++i]);
		else
		    err = 1;
		break;
	    case 'd':
		dbgDump = atoi(argv[++i]);
		break;
	    case 'e':
		dbgCont = atoi(argv[++i]);
		break;
	    default:
		fprintf(stderr,"unrecog option: %s \n",argv[i-1]);
		err = 1;
		break;
		}
	else{
	    if(inputSoundFile==NULL)
		inputSoundFile = argv[i];
	    else
		strcpy(outputPvocFile, argv[i]);
	    }
	}
    if(err || inputSoundFile==NULL)
	{
	fprintf(stderr,
	"Usage: %s [-n n] [-o n | -i n] <inputSoundfile> [<outputFFTfile>]\n",
		programName);
	exit(1);
	}
    if(outputPvocFile[0] == '\0')
	SubstExtn(inputSoundFile, DFLT_EXTN, outputPvocFile);
    /* Open sound file */
    if( (sndFile = SFOpenAllocRdHdr(inputSoundFile, &inputSound))==NULL)
	{
	printf("%s: error opening %s as input soundfile\n", programName,
	       inputSoundFile);
	exit(1);
	}
    /* Abort if unsupported format */
    if(debug)
	printf("Read sound at %lx; magic = %lx, srate = %6.1f\n",
	       inputSound, inputSound->magic, inputSound->samplingRate);
    if (inputSound->format != SFMT_SHORT)
	{
	fprintf(stderr, "%s: unsupported sound format: %d\n",
		programName, inputSound->format);
	exit(1);
	}

    /* setup frame size etc according to sampling rate */
    if(frameSize==0)		/* not specified on command line */
	{
	int	    target;
	target = inputSound->samplingRate * (float)MINFRMMS / (float)1000;
	/* default frame size is > MINFRMMS milliseconds */
	frameSize = MAXFRMPTS;
	while( (frameSize>>1) >= target && frameSize > MINFRMPTS)
	    frameSize >>= 1;	/* divide down until just larger */
	}
    if (ovlp==0 && frameIncr==0)
	{
	ovlp = DFLT_OVLP;	/* default overlap */
	frameIncr = frameSize / ovlp;
	}
    else if(ovlp == 0)
	ovlp = frameSize/frameIncr;
    else
	frameIncr = frameSize/ovlp;

    if ( (ovlp < 2) || (ovlp > 16) )
	{
	fprintf(stderr, "%s: %d is a bad window overlap index\n",
		programName, ovlp);
	exit(1);
	}

    basis = AssignBasis(NULL,frameSize); /* set up FFT tables */
    if( (ftFile=CreatePVfile(outputPvocFile,&outputPVH,inputSound))==NULL)
	{
	fprintf(stderr,"%s: error opening %s as output FT file\n",
		programName, outputPvocFile);
	exit(1);
	}

    takeFFTs(inputSound, outputPVH, sndFile, ftFile);

    /* Close files */
    PVCloseWrHdr(ftFile, outputPVH);
    fclose(sndFile);
    }

static FILE *CreatePVfile(pvPath,ppPVH,inputSound)
    char	    *pvPath;
    PVSTRUCT	    **ppPVH;
    SFSTRUCT	    *inputSound;
    {
    float   rate     = (float)inputSound->samplingRate;
    int     chans    = (int)inputSound->channels;
    int     dsize    = SFSampleSizeOf(inputSound->format);
    long    frameWords;
    int     err;
    FILE    *opFile;

    if(debug)
	printf("Framesiz %ld, framInc %ld\n",frameSize, frameIncr);
    /* If we dealt with frames that hit the ends properly, we'd have */
    /*	 (size/sizeof(short))/frameIncr frames (>= 1/2 inside file) */
    frameWords = frameSize + 2L; /* each frame has Mag & phase for n/2 +1 */
    if(inputSound->dataBsize != SF_UNK_LEN)
	printf("Est. frames %ld\n", 
	       inputSound->dataBsize/(chans*dsize*frameIncr));
    if(chans > 1)
	{
	printf("Sorry: using just 1st channel of %d-channel sound\n", chans);
/*	chans = 1;	/* will write # chans into PV header tho' only mono */
	}
    printf("Output: %d chan%s, %ld frsize, %ld frInc\n",
       chans, (chans==1)?"":"s", frameSize, frameIncr);
    /* Allocate a new PV block */
    if (err = PVAlloc(ppPVH, PV_UNK_LEN, PVFLOAT, rate, chans,
	      frameSize, frameIncr, sizeof(float)*2*(frameSize/2 + 1),
	      PVPVOC, 0.0, rate/2.0, PVLIN, 4))
    { fprintf(stderr, "%s: %s\n", programName, PVErrMsg(err));
      exit(1);	}
    opFile = PVOpenWrHdr(pvPath, *ppPVH);
    return opFile;
    }

/*
 * takeFFTs
 *  Go through the (mono) input sound frame by frame and find the
 *  magnitude and phase change for a string of FFT bins
 */
static void takeFFTs(inputSound,outputPVH, sndFile, pvFile)
    SFSTRUCT	    *inputSound;
    PVSTRUCT	    *outputPVH;
    FILE	    *sndFile;
    FILE	    *pvFile;
    {
    long    i,j;
    float   *tmpBuf, *oldInPh, *winBuf;
    float   sampRate = inputSound->samplingRate;
    int     chans    = inputSound->channels;
    int     not_done = 1;
    long    read_in  = -1;
    short   *shBuf;
    long    fsIndepVals = (frameSize/2L)+1L;
    long    ifIx = 0;

    shBuf   = (short *)MakeBuf(chans*frameSize);
    tmpBuf  = MakeBuf(frameSize * 2L);
    oldInPh = MakeBuf(frameSize);
    winBuf  = MakeHalfWin(frameSize,(float)1.0,1);

    /* initially, clear first half of buffer .. */
    for(i = 0; i<(chans*frameSize/2); ++i)	shBuf[i] = 0;
    /* .. and read in second half from file */
    SFfread(shBuf+frameSize/2, SFMT_SHORT, (chans*frameSize/2), sndFile);

    fprintf(stderr, "dbgDump %d, dbgCont %d\n", dbgDump, dbgCont);
    printf("frame: ");
    for (i = 0; not_done; i++)
	{
	if((i%20)==0)   printf("%ld ", i); fflush(stdout);
	
/*	FloatAndCopy(shBuf,tmpBuf,frameSize);	*/
	Short2Float(shBuf, chans, tmpBuf, 1, frameSize);
	/*  PrintBuf(tmpBuf, frameSize, "floated"); /* */

	CopyShorts(shBuf+(frameIncr*chans),1,shBuf,1,
		   (frameSize-frameIncr)*chans,0);
	if(read_in)		/* if it worked last time */
	    read_in = SFfread(shBuf+(frameSize-frameIncr)*chans, SFMT_SHORT, 
			    frameIncr*chans, sndFile);
	else
	    not_done = 0;	/* stop after this run-thru */
	/* clear end if not all read in */
	if(read_in < frameIncr)
	    for(j = frameSize-frameIncr+read_in; j<frameSize; ++j)
		shBuf[j] = 0;
	ApplyHalfWin(tmpBuf,winBuf,frameSize);
	/*  PrintBuf(tmpBuf, frameSize, "windo'd"); /* */
	if(i>=dbgDump && i<(dbgDump+dbgCont))  {
	    char name[32];
	    
	    sprintf(name, "frame %d", i);
	    DumpSet(tmpBuf, frameSize, ifIx, name);
	    }
	ifIx += frameIncr;
	UnpackReals(tmpBuf, frameSize);
	/*  PrintBuf(tmpBuf, frameSize, "unpacked"); /* */
	FFT2real((complex *)tmpBuf, frameSize, 1,basis);
	/*  PrintBuf(tmpBuf, frameSize, "fft'd"); /* */
	Rect2Polar(tmpBuf, fsIndepVals);
	/*  PrintBuf(tmpBuf, frameSize, "toPolar"); /* */
	UnwrapPhase(tmpBuf, fsIndepVals, oldInPh);
	/*  PrintBuf(tmpBuf, frameSize, "unWrapped"); /* */
	PhaseToFrq(tmpBuf, fsIndepVals, (float)frameIncr, (float)sampRate);
	/*  PrintBuf(tmpBuf, frameSize, "toFrq"); /* */
	/* write straight out, just the indep vals */
	fwrite(tmpBuf, 2*sizeof(float), fsIndepVals, pvFile);
	/* debug = 0;	/* */
	}
    printf("%ld\n",i-1);
    }

#define DBGPTS 8
static void PrintBuf(buf,size,msg)
    float *buf;
    int   size;
    char  *msg;
    {
    int   i;

/*    if(!debug)
      return;
 */
    size = DBGPTS;
    printf("%s:",msg);
    for(i=0; i<size; ++i)
    printf("%7.2f ",buf[i]);
    printf("\n");
    }

/* Convert multi-channel sound to mono by tossing extra channels */
static 	void	convertToMono(theSound)
    SFSTRUCT  **theSound;
    {
    SFSTRUCT *newSound;
    short   *oldPtr, *newPtr;
    int     size     = (*theSound)->dataBsize;
    int     format   = (*theSound)->format;
    int     rate     = (*theSound)->samplingRate;
    int     channels = (*theSound)->channels;
    int     newSize, i, j;

    newSize = size/channels;
    if (sfErr = SFAlloc(&newSound, newSize, format, rate, 1, 4))
	{
	printf("%s: sound error: %s\n", programName,
		    SFErrMsg(sfErr));
	exit(1);
	}
    oldPtr  = (short *)SFDataLoc(*theSound);
    newPtr  = (short *)SFDataLoc(newSound);
    for (i = 0, j = 0; i < newSize/sizeof(short); i++, j+=channels)
	newPtr[i] = oldPtr[j];
    SFFree(*theSound);
    *theSound = newSound;
    }

static void CopyShorts(sce, sinc, dst, dinc, len, fwdNbck)
    short *sce;
    int sinc;
    short *dst;
    int dinc;
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last */
	{
	sce += len * sinc;
	dst += len * dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first */
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }

static FILE *dumpFile = NULL;
static char dumpFileName[] = "pva.xgp";

DumpSet(float *data, long len, long offs, char *desc)
    {
    if(dumpFile == NULL)  {
	if( (dumpFile = fopen(dumpFileName, "w"))==NULL) {
	    fprintf(stderr, "DumpSet:  cannot create dump file '%s'\n",
		    dumpFileName);	abort();    }
	/* output options here */
	fprintf(dumpFile, "TitleText: %s output\n", programName);
	}
    else if(len < 0) {	/* flag to close file */
	fclose(dumpFile);
	}
    else {	/* dump the data */
	long l;

	fprintf(dumpFile, "\n");	/* blank line for new set */
	if(desc != NULL)  {
	    fprintf(dumpFile, "\"%s\"\n", desc);
	    }
	for(l=0; l<len; ++l)  {
	    fprintf(dumpFile, "%l8d %f\n", offs+l, data[l]);
	    }
	fflush(dumpFile);
	}
    }
