/***********************************************************************\
*  pvi.c	(after fdsyn.c						*
* New version of PVOC inversion : reads and writes direct to disk	*
* dpwe 14feb91								*
\***********************************************************************/

#include <stdio.h>
#include <math.h>

#ifdef THINK_C
#include <stdlib.h>		/* for malloc if needed, and exit() proto */
#include <console.h>
#include <tc_main.h>
char sfdir_path[] = "Frida:dpwe:sound:";
#else
#include <malloc.h>
#endif
#include <samplops.h>
#include <dsputil.h>
#include <fft.h>
#include <sndf.h>
#include <pvoc.h>

/* extern fn protos */

/* Static function prototypes */
#ifdef __STDC__
static FILE *NewSoundFile(char *sPath, SFSTRUCT **opSnd, PVSTRUCT *inpPVH); 
static void resynthSound(PVSTRUCT *inpPVH,SFSTRUCT *outSnd,FILE *pf,FILE *sf);
static void FloatScale2Short(float *pfl, int finc, short *psh, int shinc,
			long len,FLOATARG max);
static void PrintBuf(float *buf, int size, char *msg);
/* static void CopyFloats(float *sce, int sinc, float *dst, int dinc, 
		       long len, int fwdNbck);  */
#else	/* no arg prototypes */
static FILE *NewSoundFile();
static void resynthSound();
static void FloatScale2Short();
static void PrintBuf();
/* static void CopyFloats(); */
#endif

/* Global variables */
int	    sfErr;
char	    *programName;
int	    debug = 0;		/* tweaked internally */
long	    inFrInc;		/* hack : now a global */
int 	    dbgDump = -1;	/* dump out frames to xgraph file */
int	    dbgCont = 10;	/* how many frames to xgr dump */

/* Static global variables */
static	float	tex = 1.0;	/* time extension factor */
static	float	pex = 1.0;	/* pitch expansion factor */
static	int	frOvFac = 4;	/* desired number of frames to overlap */
static	long	frameSize;	/* must match analysis file */
static	long	frameIncr;	/* will be frameSize / frOvFac */
static	float	fiex; /* frame inverse expn factor - hack to pass btwn funcs */
static	complex *plut;		/* lookup table for fft */
static	int	specwp = 0;	/* flag to invoke spectral warping */

main(argc,argv)
    int argc;
    char *argv[];
    {
    PVSTRUCT	    *inputPVH;
    SFSTRUCT	    *outputSound;
    char    *inputPvocFile = NULL, *outputSoundFile = NULL;
    int     i,err;
    int 	ovlp = 0;
    FILE	*pvf, *sndf;

    /* Check arguments */
    programName = argv[0];
    if(argc==1)
	{
	fprintf(stderr,
	"Usage: %s [-t n][-p n][-o n][-d d]<inputFFTfile><outputSoundFile>\n",
		programName);
	exit(1);
	}
    for(i=1;i<argc;++i)
	{
	if(argv[i][0] == '-')
	    switch(argv[i][1])
		{
	    case 't':
		tex = atof(argv[++i]);
		break;
	    case 'p':
		pex = atof(argv[++i]);
		break;
	    case 'o':
		ovlp = atoi(argv[++i]);
		break;
	    case 'd':
		dbgDump = atoi(argv[++i]);
		break;
	    case 'e':
		dbgCont = atoi(argv[++i]);
		break;
	    default:
		fprintf(stderr,"%s: bad option %s\n", programName, argv[i]);
		exit(1);
		}
	else if(inputPvocFile == NULL)
	    inputPvocFile = argv[i];
	else
	    outputSoundFile = argv[i];
	}

    if(inputPvocFile == NULL)
	{
	fprintf(stderr,"%s: no input filename\n", programName);
	exit(1);
	}

    if(outputSoundFile == NULL)
	{
	fprintf(stderr,"%s: no output filename\n", programName);
	exit(1);
	}

    if ( (pex < 0.1) || (pex > 10) )
	{
	fprintf(stderr, "%s: %f is a bad pitch expansion factor\n",
		programName, pex);
	exit(1);
	}

    if ( (tex < 0.1) || (tex > 30) )
	{
	fprintf(stderr, "%s: %f is a bad time extension factor\n",
		programName, tex);
	exit(1);
	}

    if ( (tex*pex < 0.1) || (tex*pex > 30) )
	{
	fprintf(stderr, "%s: %f is a bad time-pitch expansion product\n",
		programName, tex*pex);
	exit(1);
	}

    if(ovlp>0)
	frOvFac = ovlp;

    /* Open ft file */
    if ( (pvf = PVOpenAllocRdHdr(inputPvocFile, &inputPVH)) == NULL )
	PVDie(PVE_NOPEN, inputPvocFile);
    /* Abort if unsupported format */
    if (inputPVH->dataFormat != PVFLOAT || inputPVH->frameFormat != PVPVOC)
	{
	fprintf(stderr, "%s: Unsupported file format: %d / %d\n",
		programName, inputPVH->dataFormat, inputPVH->frameFormat);
	exit(1);
	}
    sndf = NewSoundFile(outputSoundFile, &outputSound, inputPVH);

    plut = AssignBasis((complex *)NULL,frameSize); /* set up FFT tables */
    MakeSinc();			/* set up sinc table for UDSamp */

    resynthSound(inputPVH, outputSound, pvf, sndf);

    /* close files */
    SFCloseWrHdr(sndf, NULL);
    fclose(pvf);

    DestroySinc();
    }

static FILE *NewSoundFile(sndPath,poutSnd, inpPvh)
    char	*sndPath;
    SFSTRUCT	**poutSnd;
    PVSTRUCT	*inpPvh;
    {
    int     format   = SFMT_SHORT;
    float   rate     = inpPvh->samplingRate;
    int     inChans  = inpPvh->channels;
    long    inFrames, outFrames, outBSize;
    int     outChans;
    long    inFrSiz  = inpPvh->frameSize;
/*    long     inFrInc  = inpPvh->frameIncr;	now a global */
    long    origSampls, outSampls;
    short   *sPtr;
    FILE    *opf;

    inFrInc  = inpPvh->frameIncr;
    frameSize = inFrSiz;
    frameIncr = frameSize/frOvFac; /* as we intended */
    outChans = inChans;
    printf("inChans  : %d \t outChans  : %d \n",inChans,outChans);
    if (outChans != 1)
	{
	fprintf(stderr, "%s: faking %d chan sound with 0s: assu mono data\n",
		programName, outChans);
	}
    printf("Input  frameSize %ld frameIncr %ld \n",
	   inFrSiz, inFrInc);
    printf("Output frameSize %ld frameIncr %ld  \n",
	   frameSize, frameIncr);
    printf("Apparent frames %ld\n", inpPvh->dataBsize/inpPvh->frameBsize);
    /* Allocate a new sound */
    if(sfErr = SFAlloc(poutSnd,SF_UNK_LEN,format,rate,outChans,4))
	SFDie(sfErr, NULL);
    if( (opf = SFOpenWrHdr(sndPath, *poutSnd)) == NULL)
	SFDie(SFE_NOPEN, sndPath);

    fiex = (float)frameIncr/(tex * (float)inFrInc);
    /* 'fiex' is inverse expansion factor for frames */
    /* Maps output frame number to fraction index to input frames */
    /* return fiex as we have the means to work it out */
    return(opf);
    }

/*
 * resynthSound
 *  Rebuild a playable sound, based on the unwrapped FFT data in the input file
 */
static void resynthSound(inputPVH,outputSound,pvf,sndf)
    PVSTRUCT	    *inputPVH;
    SFSTRUCT	    *outputSound;
    FILE 	*pvf;
    FILE	*sndf;
    {
    long    i, outSampls, outFrames;
    float   *tmpBuf, *tmpBuf2, *outPhs, *winBuf /*, *wBuf */ ;
    float   scaleDown, oneOnFSz;
    float   sampRate = (float)inputPVH->samplingRate;
    float   oneOnPex = 1.0/pex, lastPex = 1.0;
    long    buf2Size,d;
    float   *outBuf, *ar;
    short   *shBuf;
    long    circBufSize, opBpos = 0L;
    float   *pvBuf;
    long    pvBufBase = -1;
    long    pvBsize = inputPVH->frameBsize/sizeof(float);
    int     done = 0;
    float   fIndex;
    long    fsIndepVals = (frameSize/2L)+1L;
    int     chans = outputSound->channels;
    long    ofIx = 0;

#define B2FACT 1.0 /* 0.4 */
    /* B2FACT is the width of the output window as a factor of the frame */
    /* make < 1.0 for less smearing of sound.  Depends on analysis, tho */

    buf2Size = 2L*(long)(B2FACT*.5*oneOnPex*(float)frameSize); 
    /* must be even */

    /* to scale down data before ifft & olap-adding lots of frames */
    oneOnFSz  = 1.0/((float)frameSize);  /* passed to FFT */
    scaleDown = (2*(float)frameIncr/(float)buf2Size); 

    circBufSize = buf2Size;
    pvBuf  = MakeBuf(pvBsize * 2L);	/* read in 2 frames */
    shBuf  = (short *)MakeBuf(chans*frameIncr/2 /* short */);
    outBuf = MakeBuf(circBufSize);
    ar     = MakeBuf(frameIncr);
    tmpBuf = MakeBuf(frameSize * 2L);    /* complex ! */
    tmpBuf2= MakeBuf(buf2Size);
    outPhs = MakeBuf(frameSize);
    winBuf = MakeHalfWin(buf2Size, scaleDown, 1); /*	MakeBuf(frameSize); */

/*    printf("resynthSound: scaleDown is %f \n",scaleDown);
    printf("Time ex:%.2f Frames inv ex:%.4f \n",tex,fiex); 
    printf("b1siz %ld  b2siz %ld\n",frameSize, buf2Size);
  */
    fprintf(stderr, "pvBufStep = %d floats\n", pvBsize);
    /* load first 2 frames of pvf */
    fread(pvBuf, sizeof(float), pvBsize*2L, pvf);
    pvBufBase = 0;	
    FetchIn(pvBuf,tmpBuf,frameSize,(float)0);	/* first frame to align phs */
    FrqToPhase(tmpBuf, fsIndepVals, 
	       (FLOATARG)inFrInc,(FLOATARG)sampRate, (FLOATARG)0);
    RewrapPhase(tmpBuf,fsIndepVals,outPhs);
    for (i = 0; !done ; i++)
	{
	fIndex = 1 + ((float)i)*fiex;
	while( ((long)fIndex) > pvBufBase )	/* need another frame */
	    {
	    CopyFloats(pvBuf+pvBsize, 1, pvBuf, 1, pvBsize, 0);
	    done = (pvBsize > fread(pvBuf+pvBsize, sizeof(float), 
				    pvBsize, pvf));
	    ++pvBufBase;
	    if(done) fIndex = pvBufBase;	/* clamp to last frame */
	    }
	FetchIn(pvBuf,tmpBuf,frameSize,fIndex-(float)pvBufBase);
	PrintBuf(pvBuf, frameSize, "pvBuf");
	PrintBuf(tmpBuf, frameSize, "tmpBuf");
	FrqToPhase(tmpBuf, fsIndepVals, 
		   pex*(float)frameIncr,
		   (float)sampRate, /*a*/ (.5 * ( (pex / lastPex) - 1) ));
	PrintBuf(tmpBuf, frameSize, "toFrq");
	/* Offset the phase to align centres of stretched wndws, not starts */
	RewrapPhase(tmpBuf,fsIndepVals,outPhs);
	PrintBuf(tmpBuf, frameSize, "Rewrp");
	if(specwp)
	    {
	    PreWarpSpec(tmpBuf, fsIndepVals, pex);
	    }
	Polar2Rect(tmpBuf,fsIndepVals);
	PrintBuf(tmpBuf, frameSize, "p2r");
	/* force the imag parts of dc and fs/2 to nil - else get whistle */
	tmpBuf[1] = 0; tmpBuf[frameSize+1] = 0;
	FFT2torl((complex *)tmpBuf,frameSize,1,	/*a*/ oneOnFSz, plut);	/* */
	PrintBuf(tmpBuf, frameSize, "ifft");
	PackReals(tmpBuf, frameSize);
	PrintBuf(tmpBuf, frameSize, "pkrl");
	if(i>=dbgDump && i<(dbgDump+dbgCont))  {
	    char name[32];
	    
	    sprintf(name, "frame %d", i);
	    DumpSet(tmpBuf, frameSize, ofIx, name);
	    }
	if(pex != 1.0)
	    {
	    UDSample(tmpBuf,.5*((float)frameSize - pex*(float)buf2Size), 
		     tmpBuf2, frameSize, buf2Size, pex);
	    }
	else
	    CopySamps(tmpBuf+
		      (int)(.5*((float)frameSize - pex*(float)buf2Size)), 
		      tmpBuf2,buf2Size);
	ApplyHalfWin(tmpBuf2, winBuf, buf2Size); /* */
	/* FixAndAdd(tmpBuf2, outPtr, buf2Size);  */
	addToCircBuf(tmpBuf2, outBuf, opBpos, frameIncr, circBufSize);
	writeClrFromCircBuf(outBuf, ar, opBpos, frameIncr, circBufSize);
	FloatScale2Short(ar,1,shBuf,chans,frameIncr, (FLOATARG)1);
	SFfwrite(shBuf, SFMT_SHORT, chans*frameIncr, sndf);
	ofIx += frameIncr;
	opBpos += frameIncr;
	if(opBpos > circBufSize)     opBpos -= circBufSize;
	addToCircBuf(tmpBuf2+frameIncr,outBuf,opBpos,
		     buf2Size-frameIncr,circBufSize);
	if((i%20)==0)
	    {	printf("%ld ", i);   fflush(stdout);	}
	lastPex = pex;	     /* needs to know last pitchexp to update phase */
	debug = 0;
	}
    printf("%ld\n",i);
    }

static void FloatScale2Short(pfl, finc, psh, shinc, len, max)
    float	*pfl;
    int 	finc;
    short	*psh;
    int 	shinc;
    long	len;
    FLOATARG	max;
    {
    register float scale = ((float)SHORTMAX)/max;

    while(len--)
	{
	*psh = *pfl * scale;
	psh += shinc;
	pfl += finc;
	}
    }

#define DBGPTS 8
static void PrintBuf(buf,size,msg)
    float *buf;
    int   size;
    char  *msg;
    {
    int   i;

    if(!debug)
    return;
    size = DBGPTS;
    printf("%s:",msg);
    for(i=0; i<size; ++i)
    printf("%7.2f ",buf[i]);
    printf("\n");
    }

/*
static void CopyFloats(sce, sinc, dst, dinc, len, fwdNbck)
    float *sce;
    int sinc;
    float *dst;
    int dinc;
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last * /
	{
	sce += len * sinc;
	dst += len * dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first * /
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }
 */

static FILE *dumpFile = NULL;
static char dumpFileName[] = "pvi.xgp";

DumpSet(float *data, long len, long offs, char *desc)
    {
    if(dumpFile == NULL)  {
	if( (dumpFile = fopen(dumpFileName, "w"))==NULL) {
	    fprintf(stderr, "DumpSet:  cannot create dump file '%s'\n",
		    dumpFileName);	abort();    }
	/* output options here */
	fprintf(dumpFile, "TitleText: %s output\n", programName);
	}
    else if(len < 0) {	/* flag to close file */
	fclose(dumpFile);
	}
    else {	/* dump the data */
	long l;

	fprintf(dumpFile, "\n");	/* blank line for new set */
	if(desc != NULL)  {
	    fprintf(dumpFile, "\"%s\"\n", desc);
	    }
	for(l=0; l<len; ++l)  {
	    fprintf(dumpFile, "%l8d %f\n", offs+l, data[l]);
	    }
	fflush(dumpFile);
	}
    }
