/***********************************************************************\
*  stfts.c	(after fdsyn.c, pvi.c)					*
* New version of PVOC inversion : reads and writes direct to disk	*
* uses <readaft> astraction to perform discrete diff of phases		*
* dpwe 08sep91 after 14feb91						*
\***********************************************************************/

/* #define WINDEBUG 1 */

#include <dpwelib.h>	/* stdio, malloc etc */
#include <vecops.h>
#include <samplops.h>
#include <dsputil.h>
#include <fft.h>
#include <sndf.h>
#include <readaft.h>
#include <math.h>

#ifdef WINDEBUG
#include <window.h>
#endif /* WINDEBUG */

/* extern fn protos */

/* Static function prototypes */
static void usage(void);		/* print syntax & exit */
static FILE *NewSoundFile PARG((char *path, SFSTRUCT **out, PVSTRUCT *in));
static void resynthSound PARG((AFTINFO *in, SFSTRUCT *out, FILE *sf));
static void FloatScale2Short PARG((float *pfl, int finc, short *psh, int shinc,
				   long len,FLOATARG max));
static void PrintBuf PARG((float *buf, int size, char *msg));
void Resynth PARG((FLOATARG overlap, FLOATARG duration, FLOATARG samplingRate, 
	     long frameSize));

/* Global variables */
int	    sfErr;
char	    *programName;
int	    debug = 1;		/* tweaked internally */
long	    inFrInc;		/* hack : now a global */
int 	dbgDump = -1;
int 	dbgCont = 0;
int	dbgUniq = -1;		/* flag to zero out all except one frame */

float	resyncThresh = 1.0;	/* magnitude threshold below which phase 
				   is not rewrapped but just mirrored */
int	rewindow = 1;		/* flag to apply window on output */

/* Static global variables */
static	float	tex = 1.0;	/* time extension factor */
static	float	pex = 1.0;	/* pitch expansion factor */
static	int	frOvFac = 4;	/* desired number of frames to overlap */
static	long	frameSize;	/* must match analysis file */
static	long	frameIncr;	/* will be frameSize / frOvFac */
static	float	fiex; /* frame inverse expn factor - hack to pass btwn funcs */
static	complex *plut;		/* lookup table for fft */
static	int	specwp = 0;	/* flag to invoke spectral warping */
static  AFTINFO *inputAFT;
static  FILE    *outFile;

#ifdef WINDEBUG
static WINDAT	*wind;
#endif /* WINDEBUG */

main(argc,argv)
    int argc;
    char *argv[];
    {
/*    AFTINFO	    *inputAFT;	*/
    PVSTRUCT	    *inputPVH;
    SFSTRUCT	    *outputSound;
    char    *inputPvocFile = NULL, *outputSoundFile = NULL;
    int     i;
    FILE	*sndf;

    /* Check arguments */
    programName = argv[0];
    for(i=1;i<argc;++i)
	{
	if(argv[i][0] == '-')
	    switch(argv[i][1])
		{
	    case 't':
		tex = atof(argv[++i]);
		break;
	    case 'p':
		pex = atof(argv[++i]);
		break;
	    case 'd':
		dbgDump = atoi(argv[++i]);
		break;
	    case 'e':
		dbgCont = atoi(argv[++i]);
		break;
	    case 'u':
		dbgUniq = atoi(argv[++i]);
		break;
	    case 'h':
		resyncThresh = atof(argv[++i]);
		break;
	    case 'w':
		rewindow = 0;
		break;
	    default:
		fprintf(stderr,"%s: bad option %s\n", programName, argv[i]);
		exit(1);
		}
	else if(inputPvocFile == NULL)
	    inputPvocFile = argv[i];
	else
	    outputSoundFile = argv[i];
	}

    if(inputPvocFile == NULL)
	{
	fprintf(stderr,"%s: no input filename\n", programName);
	usage();
	}

    if(outputSoundFile == NULL)
	{
	fprintf(stderr,"%s: no output filename\n", programName);
	usage();
	}

    if ( (pex < 0.1) || (pex > 10) )
	{
	fprintf(stderr, "%s: %f is a bad pitch expansion factor\n",
		programName, pex);
	usage();
	}

    if ( (tex < 0.1) || (tex > 30) )
	{
	fprintf(stderr, "%s: %f is a bad time extension factor\n",
		programName, tex);
	usage();
	}

    if ( (tex*pex < 0.1) || (tex*pex > 30) )
	{
	fprintf(stderr, "%s: %f is a bad time-pitch expansion product\n",
		programName, tex*pex);
	usage();
	}

    /* Open ft file */
   if(OpenAFTPVfile(inputPvocFile, &inputAFT)==NULL)
	PVDie(PVE_NOPEN, inputPvocFile);
    inputPVH = inputAFT->pvh;
    /* Abort if unsupported format */
    if (inputPVH->dataFormat != PVFLOAT || inputPVH->frameFormat != PVPOLAR)
	{
	fprintf(stderr, "%s: Unsupported file format: %d / %d\n",
		programName, inputPVH->dataFormat, inputPVH->frameFormat);
	usage();
	}
    outFile = sndf = NewSoundFile(outputSoundFile, &outputSound, inputPVH);

    plut = AssignBasis((complex *)NULL,frameSize); /* set up FFT tables */
    MakeSinc();			/* set up sinc table for UDSamp */

#ifdef WINDEBUG
    wind = NewWin("stfts",0);
#endif /* WINDEBUG */

    Resynth(0.75, inputAFT->duration, outputSound->samplingRate, 
	    frameSize);
/*     resynthSound(inputAFT, outputSound, sndf);  */

    DestroySinc();
    /* close files */
    SFCloseWrHdr(sndf, NULL);
    CloseAFTPVfile(inputAFT);
    }

static void usage(void)		/* print syntax & exit */
    {
    fprintf(stderr,
"usage: %s [-t f][-p f][-d n][-e n][-u n][-w] <inputFFT> <outputSound>\n",
	    programName);
    fprintf(stderr,"where (defaults in brackets)\n");
    fprintf(stderr," -t f   time extension factor       (1.0)\n");
    fprintf(stderr," -p f   pitch extension factor      (1.0)\n");
    fprintf(stderr," -d n   DEBUG: dump frames from n to .xgp file\n");
    fprintf(stderr," -e n   DEBUG: with -d, dump this many frames total\n");
    fprintf(stderr," -u n   DEBUG: output will be result of unique frame n\n");
    fprintf(stderr," -w     DEBUG: do not apply rewindowing\n");
    fprintf(stderr," inputFFT     input analysis file from stfta\n");
    fprintf(stderr," outputSound  output sound file\n");
    exit(1);
    }

static FILE *NewSoundFile(sndPath, poutSnd, inpPvh)
    char	*sndPath;
    SFSTRUCT	**poutSnd;
    PVSTRUCT	*inpPvh;
    {
    int     format   = SFMT_SHORT;
    float   rate     = inpPvh->samplingRate;
    int     inChans  = inpPvh->channels;
    int     outChans;
    long    inFrSiz  = inpPvh->frameSize;
/*    long     inFrInc  = inpPvh->frameIncr;	now a global */
    FILE    *opf;

    inFrInc  = inpPvh->frameIncr;
    frameSize = inFrSiz;
    frameIncr = frameSize/frOvFac; /* as we intended */
    outChans = inChans;
    printf("inChans  : %d \t outChans  : %d \n",inChans,outChans);
    if (outChans != 1)
	{
	fprintf(stderr, "%s: faking %d chan sound with 0s: assu mono data\n",
		programName, outChans);
	}
    printf("Input  frameSize %ld frameIncr %ld \n",
	   inFrSiz, inFrInc);
    printf("Output frameSize %ld frameIncr %ld  \n",
	   frameSize, frameIncr);
    printf("Apparent frames %ld\n", inpPvh->dataBsize/inpPvh->frameBsize);
    fprintf(stderr, "Resync thresh %f\n", resyncThresh);
    /* Allocate a new sound */
    if(sfErr = SFAlloc(poutSnd,SF_UNK_LEN,format,rate,outChans,4))
	SFDie(sfErr, NULL);
    if( (opf = SFOpenWrHdr(sndPath, *poutSnd)) == NULL)
	SFDie(SFE_NOPEN, sndPath);

    fiex = (tex * (float)inFrInc)/(float)frameIncr;
    /* 'fiex' is inverse expansion factor for frames */
    /* Maps output frame number to fraction index to input frames */
    /* return fiex as we have the means to work it out */
    return(opf);
    }

/*
 * resynthSound
 *  Rebuild a playable sound, based on the unwrapped FFT data in the input file
 */
static void wasresynthSound(inputAFT,outputSound,sndf)
    AFTINFO	*inputAFT;
    SFSTRUCT	*outputSound;
    FILE	*sndf;
    {
    PVSTRUCT	    *inputPVH = inputAFT->pvh;
    long    i;
    float   *tmpBuf, *tmpBuf2, *outPhs, *winBuf /*, *wBuf */ ;
    float   scaleDown, oneOnFSz;
    float   sampRate = (float)inputPVH->samplingRate;
    float   oneOnPex = 1.0/pex, lastPex = 1.0;
    long    buf2Size;
    float   *outBuf, *ar;
    short   *shBuf;
    long    circBufSize, opBpos = 0L;
    float   *pvBuf;
    long    pvBufBase = -1;
    long    pvBsize = inputPVH->frameBsize/sizeof(float);
    int     done = 0;
    float   fIndex, ixToTime, time;
    long    fsIndepVals = (frameSize/2L)+1L;
    int     chans = outputSound->channels;

#define B2FACT 1.0 /* 0.4 */
    /* B2FACT is the width of the output window as a factor of the frame */
    /* make < 1.0 for less smearing of sound.  Depends on analysis, tho */

    buf2Size = 2L*(long)(B2FACT*.5*oneOnPex*(float)frameSize); 
    /* must be even */

    /* to scale down data before ifft & olap-adding lots of frames */
    oneOnFSz  = 1.0/((float)frameSize);  /* passed to FFT */
    scaleDown = (2*(float)frameIncr/(float)buf2Size);
    /* for some reason, had to add massive gain for this vsn ? */

    circBufSize = buf2Size;
    pvBuf  = MakeBuf(pvBsize * 2L);	/* read in 2 frames */
    shBuf  = (short *)MakeBuf(chans*frameIncr/2 /* short */);
    outBuf = MakeBuf(circBufSize);
    ar     = MakeBuf(frameIncr);
    tmpBuf = MakeBuf(frameSize * 2L);    /* complex ! */
    tmpBuf2= MakeBuf(buf2Size);
    outPhs = MakeBuf(frameSize);
    winBuf = MakeHalfWin(buf2Size, scaleDown, 1); /*	MakeBuf(frameSize); */

    printf("resynthSound: scaleDown is %f \n",scaleDown);
    printf("Time ex:%.2f Frames inv ex:%.4f \n",tex,fiex); 
    printf("b1siz %ld  b2siz %ld\n",frameSize, buf2Size);

    fprintf(stderr, "pvBufStep = %d floats\n", pvBsize);
    /* load first 2 frames of pvf */
    fread(pvBuf, sizeof(float), pvBsize*2L, inputAFT->pvFile);
    pvBufBase = 0;	
    FetchIn(pvBuf,tmpBuf,frameSize,(float)0);	/* first frame to align phs */
/*    FrqToPhase(tmpBuf, fsIndepVals, 
	       (FLOATARG)inFrInc,(FLOATARG)sampRate, (FLOATARG)0);
    RewrapPhase(tmpBuf,fsIndepVals,outPhs);	*/
    /* phase is phase now */
    ixToTime = (float)frameIncr/(float)sampRate;
    for (i = 0; !done ; i++)
	{
	fIndex = 1 + ((float)i)*fiex;
	time = i * ixToTime;
	done = time > inputAFT->duration;
/*	while( ((long)fIndex) > pvBufBase )
	    {
	    CopyFloats(pvBuf+pvBsize, 1, pvBuf, 1, pvBsize, 0);
	    done = (pvBsize > fread(pvBuf+pvBsize, sizeof(float), 
				    pvBsize, pvf));
	    ++pvBufBase;
	    if(done) fIndex = pvBufBase;
	    }
	FetchIn(pvBuf,tmpBuf,frameSize,fIndex-(float)pvBufBase);
 */
	SpecOfAFTFile(inputAFT, tex*time, fiex*ixToTime, (cpxpolar *)tmpBuf, 
		      1 /* dphasq */);
	/* reads true phi-dots in radians per second  - 
	   need to scale by time increment to get proper phase advance */
	VSMul(tmpBuf+1, 2, pex*ixToTime, fsIndepVals);
/*	FrqToPhase(tmpBuf, fsIndepVals, 
		   pex*(float)frameIncr,
		   (float)sampRate,  (.5 * ( (pex / lastPex) - 1) ));	
 */
	/* Offset the phase to align centres of stretched wndws, not starts */
	RewrapPhase(tmpBuf,fsIndepVals,outPhs);
	if(specwp)
	    {
	    PreWarpSpec(tmpBuf, fsIndepVals, pex);
	    }
	Polar2Rect(tmpBuf,fsIndepVals);
	/* force the imag parts of dc and fs/2 to nil - else get whistle */
	tmpBuf[1] = 0; tmpBuf[frameSize+1] = 0;
	FFT2torl((complex *)tmpBuf,frameSize,1,	/*a*/ oneOnFSz, plut);	/* */
	PrintBuf(tmpBuf, frameSize, "post fft");
	PackReals(tmpBuf, frameSize);
	if(pex != 1.0)
	    {
	    UDSample(tmpBuf,.5*((float)frameSize - pex*(float)buf2Size), 
		     tmpBuf2, frameSize, buf2Size, pex);
	    }
	else
	    CopySamps(tmpBuf+
		      (int)(.5*((float)frameSize - pex*(float)buf2Size)), 
		      tmpBuf2,buf2Size);
	ApplyHalfWin(tmpBuf2, winBuf, buf2Size); /* */
	/* FixAndAdd(tmpBuf2, outPtr, buf2Size);  */
#ifdef WINDEBUG
	DoDisp(wind, tmpBuf2, buf2Size);
#endif /* WINDEBUG */
	addToCircBuf(tmpBuf2, outBuf, opBpos, frameIncr, circBufSize);
	writeClrFromCircBuf(outBuf, ar, opBpos, frameIncr, circBufSize);
	FloatScale2Short(ar,1,shBuf,chans,frameIncr, (FLOATARG)1);
	fwrite(shBuf, sizeof(short), chans*frameIncr, sndf);
	opBpos += frameIncr;
	if(opBpos > circBufSize)     opBpos -= circBufSize;
	addToCircBuf(tmpBuf2+frameIncr,outBuf,opBpos,
		     buf2Size-frameIncr,circBufSize);
	if((i%20)==0)
	    {	printf("%ld ", i);   fflush(stdout);	}
	lastPex = pex;	     /* needs to know last pitchexp to update phase */
	debug = 0;	
	}
    printf("%ld\n",i);
    }

static void resynthSound(inputAFT,outputSound,sndf)
    AFTINFO	*inputAFT;
    SFSTRUCT	*outputSound;
    FILE	*sndf;
    {
    PVSTRUCT	    *inputPVH = inputAFT->pvh;
    long    i;
    float   *tmpBuf, *tmpBuf2, *outPhs, *winBuf /*, *wBuf */ ;
    float   scaleDown, oneOnFSz;
    float   sampRate = (float)inputPVH->samplingRate;
    float   oneOnPex = 1.0/pex, lastPex = 1.0;
    long    buf2Size;
    float   *outBuf, *ar;
    short   *shBuf;
    long    circBufSize, opBpos = 0L;
    float   *pvBuf;
    long    pvBufBase = -1;
    long    pvBsize = inputPVH->frameBsize/sizeof(float);
    int     done = 0;
    float   fIndex, ixToTime, time;
    long    fsIndepVals = (frameSize/2L)+1L;
    int     chans = outputSound->channels;

#define B2FACT 1.0 /* 0.4 */
    /* B2FACT is the width of the output window as a factor of the frame */
    /* make < 1.0 for less smearing of sound.  Depends on analysis, tho */

    buf2Size = 2L*(long)(B2FACT*.5*oneOnPex*(float)frameSize); 
    /* must be even */

    /* to scale down data before ifft & olap-adding lots of frames */
    oneOnFSz  = 1.0/((float)frameSize);  /* passed to FFT */
    scaleDown = (2*(float)frameIncr/(float)buf2Size);
    /* for some reason, had to add massive gain for this vsn ? */

    circBufSize = buf2Size;
    pvBuf  = MakeBuf(pvBsize * 2L);	/* read in 2 frames */
    shBuf  = (short *)MakeBuf(chans*frameIncr/2 /* short */);
    outBuf = MakeBuf(circBufSize);
    ar     = MakeBuf(frameIncr);
    tmpBuf = MakeBuf(frameSize * 2L);    /* complex ! */
    tmpBuf2= MakeBuf(buf2Size);
    outPhs = MakeBuf(frameSize);
    winBuf = MakeHalfWin(buf2Size, scaleDown, 1); /*	MakeBuf(frameSize); */

    printf("resynthSound: scaleDown is %f \n",scaleDown);
    printf("Time ex:%.2f Frames inv ex:%.4f \n",tex,fiex); 
    printf("b1siz %ld  b2siz %ld\n",frameSize, buf2Size);

    fprintf(stderr, "pvBufStep = %d floats\n", pvBsize);
    /* load first 2 frames of pvf */
    fread(pvBuf, sizeof(float), pvBsize*2L, inputAFT->pvFile);
    pvBufBase = 0;	
    FetchIn(pvBuf,tmpBuf,frameSize,(float)0);	/* first frame to align phs */
/*    FrqToPhase(tmpBuf, fsIndepVals, 
	       (FLOATARG)inFrInc,(FLOATARG)sampRate, (FLOATARG)0);
    RewrapPhase(tmpBuf,fsIndepVals,outPhs);	*/
    /* phase is phase now */
    ixToTime = (float)frameIncr/(float)sampRate;
    for (i = 0; !done ; i++)
	{
	fIndex = 1 + ((float)i)*fiex;
	time = i * ixToTime;
	done = time > inputAFT->duration;
/*	while( ((long)fIndex) > pvBufBase )
	    {
	    CopyFloats(pvBuf+pvBsize, 1, pvBuf, 1, pvBsize, 0);
	    done = (pvBsize > fread(pvBuf+pvBsize, sizeof(float), 
				    pvBsize, pvf));
	    ++pvBufBase;
	    if(done) fIndex = pvBufBase;
	    }
	FetchIn(pvBuf,tmpBuf,frameSize,fIndex-(float)pvBufBase);
 */
	SpecOfAFTFile(inputAFT, tex*time, fiex*ixToTime, (cpxpolar *)tmpBuf, 
		      1 /* dphasq */);
	/* reads true phi-dots in radians per second  - 
	   need to scale by time increment to get proper phase advance */
	VSMul(tmpBuf+1, 2, pex*ixToTime, fsIndepVals);
/*	FrqToPhase(tmpBuf, fsIndepVals, 
		   pex*(float)frameIncr,
		   (float)sampRate,  (.5 * ( (pex / lastPex) - 1) ));	
 */
	/* Offset the phase to align centres of stretched wndws, not starts */
	RewrapPhase(tmpBuf,fsIndepVals,outPhs);
	if(specwp)
	    {
	    PreWarpSpec(tmpBuf, fsIndepVals, pex);
	    }
	Polar2Rect(tmpBuf,fsIndepVals);
	/* force the imag parts of dc and fs/2 to nil - else get whistle */
	tmpBuf[1] = 0; tmpBuf[frameSize+1] = 0;
	FFT2torl((complex *)tmpBuf,frameSize,1,	/*a*/ oneOnFSz, plut);	/* */
	PrintBuf(tmpBuf, frameSize, "post fft");
	PackReals(tmpBuf, frameSize);
	HalfRotate(tmpBuf, frameSize);	/* !! to de-zero the phase */
	if(pex != 1.0)
	    {
	    UDSample(tmpBuf,.5*((float)frameSize - pex*(float)buf2Size), 
		     tmpBuf2, frameSize, buf2Size, pex);
	    }
	else
	    CopySamps(tmpBuf+
		      (int)(.5*((float)frameSize - pex*(float)buf2Size)), 
		      tmpBuf2,buf2Size);
	ApplyHalfWin(tmpBuf2, winBuf, buf2Size); /* */
	/* FixAndAdd(tmpBuf2, outPtr, buf2Size);  */
#ifdef WINDEBUG
	DoDisp(wind, tmpBuf2, buf2Size);
#endif /* WINDEBUG */
	addToCircBuf(tmpBuf2, outBuf, opBpos, frameIncr, circBufSize);
	writeClrFromCircBuf(outBuf, ar, opBpos, frameIncr, circBufSize);
	FloatScale2Short(ar,1,shBuf,chans,frameIncr, (FLOATARG)1);
	fwrite(shBuf, sizeof(short), chans*frameIncr, sndf);
	opBpos += frameIncr;
	if(opBpos > circBufSize)     opBpos -= circBufSize;
	addToCircBuf(tmpBuf2+frameIncr,outBuf,opBpos,
		     buf2Size-frameIncr,circBufSize);
	if((i%20)==0)
	    {	printf("%ld ", i);   fflush(stdout);	}
	lastPex = pex;	     /* needs to know last pitchexp to update phase */
	debug = 0;	
	}
    printf("%ld\n",i);
    }

static void FloatScale2Short(pfl, finc, psh, shinc, len, max)
    float	*pfl;
    int 	finc;
    short	*psh;
    int 	shinc;
    long	len;
    FLOATARG	max;
    {
    register float scale = ((float)SHORTMAX)/max;

    while(len--)
	{
	*psh = *pfl * scale;
	psh += shinc;
	pfl += finc;
	}
    }

#define DBGPTS 8
static void PrintBuf(buf,size,msg)
    float *buf;
    int   size;
    char  *msg;
    {
    int   i;

    if(!debug)
    return;
    size = DBGPTS;
    printf("%s:",msg);
    for(i=0; i<size; ++i)
    printf("%7.2f ",buf[i]);
    printf("\n");
    }

float TransposeFactor(FLOATARG time)
    {
    return 1.0/pex;
    }

float StretchFactor(FLOATARG time)
    {
    return tex;
    }

/* If the magnitude falls below a certain level, the phase-advance 
   information is probably not very informative or valid.  In an effort 
   to retain some kind of phase alignment with the source, we take the 
   opportunity to reset the absolute phase of a bin if it moves from 
   below to above some threshold in a given frame.  The threshold 
   should adapt, of course. 
 */
#define SOMEMAGTHRESH	resyncThresh
#define FLAGTHATPHASEISKINDABOGUS	(-999.0)
/* not a PV, so a unique flag */

float *SynthFrame(FLOATARG t, FLOATARG dt, float *dst)
    {
    static float *lastPhase = NULL;
    static float *wholeSpec = NULL;
    static float *winBuf;
    long	fsIndepVals = (frameSize/2L)+1L;
    long i;

    if(lastPhase == NULL)  {
	winBuf = MakeHalfWin(frameSize, 0.5 /* scaledown relates to olap */ ,
			     1);
	lastPhase = MakeBuf(frameSize/2+1);
	wholeSpec = MakeBuf(frameSize + 2);	/* to load in whole spec */
	}
    SimpleSpecOfAFTFile(inputAFT, t, dt, (cpxpolar *)wholeSpec, 0);
    SimpleSpecOfAFTFile(inputAFT, t, dt, (cpxpolar *)dst, 
		  1 /* dphasq */);
    for(i = 0; i<frameSize/2+1; ++i) {
	if(lastPhase[i] == FLAGTHATPHASEISKINDABOGUS) {
	    lastPhase[i] = 0.0;
	    dst[2*i+1] = wholeSpec[2*i+1];
	    }
	}
    VSMul(dst+1, 2, dt, fsIndepVals);
    if(1) {
	int d = 45;
	fprintf(stderr, "dpha[%d] = %f\n", d, dst[2*d+1]);
	} 
    RewrapPhase(dst, fsIndepVals, lastPhase);
    /* Mark stupid phases */
    for(i = 0; i<frameSize/2+1; ++i) {
	if(dst[2*i] < SOMEMAGTHRESH)
	    lastPhase[i] = FLAGTHATPHASEISKINDABOGUS;
	}
    Polar2Rect(dst, fsIndepVals);
    /* force the imag parts of dc and fs/2 to nil - else get whistle */
    dst[1] = 0;		dst[frameSize+1] = 0;
    FFT2torl((complex *)dst, frameSize, 1, 1.0/((float)frameSize), plut);
    PackReals(dst, frameSize);
    HalfRotate(dst, frameSize);	/* !! to de-zero the phase */
    if(rewindow)
	ApplyHalfWin(dst, winBuf, frameSize);

    return dst;
    }

static float *privRingBuffer = NULL;
static long  privRBsiz; 
static long  privRBpos; 
static float *otherBuf;
static long  otherBufSiz = 0;

void OverlapFrameIntoRingBuffer(float *buf, long len)
    {
    if(privRingBuffer == NULL)  {
	privRBpos = 0;
	privRBsiz = 4 * frameSize;
	privRingBuffer = MakeBuf(privRBsiz);
	otherBufSiz = frameSize;
	otherBuf = MakeBuf(otherBufSiz);
	}

    addToCircBuf(buf, privRingBuffer, privRBpos, len, privRBsiz);
    }

void WriteOutSomeOfTheRingBufferAndClearIt(long len)
    {
    short *shBuf = (short *)otherBuf;
    int   chans = 1;

    if(len > otherBufSiz)	abort();	/* some odd prob */
    writeClrFromCircBuf(privRingBuffer, otherBuf, 
			privRBpos, len, privRBsiz);
    FloatScale2Short(otherBuf, 1, shBuf, chans, len, (FLOATARG)1);
    fwrite(shBuf, sizeof(short), chans*len, outFile);
    privRBpos += len;
    if(privRBpos > privRBsiz)     privRBpos -= privRBsiz;
    }

void FFResample(src, sLen, dst, dLen, sIndex, dIndex, exFact)
    float	*src;
    long 	sLen;
    float	*dst;
    long	dLen;
    float	sIndex;		/* this sample point in the source .. */
    float	dIndex;		/* ..should fall on this point in dest */
    float	exFact;		/* something like dLen/sLen */
    {	/* exFact is how much longer each input time tick lasts 
	   that the output time ticks.
	   i.e. n _input_ ticks last as long as (n*exFact) output ticks
	   thus n _output_ ticks correspond to (n/exFact) input ticks   */
    float 	stIndex;
    long	skip, safelen;

    stIndex = sIndex - dIndex/exFact;
    if(stIndex < 0.0) {
	skip = (-stIndex)*exFact+0.99;
/*	fprintf(stderr, "skip  %d\n", skip);	*/
	ClearFloats(dst, 1, skip);
	dst += skip;
	dLen -= skip;
	stIndex += (float)skip / exFact;	/* will become >= 0 */
	}
    /* how many output points can we ask for safely?  */
    safelen = ((float)sLen - stIndex - 1.0)*exFact;
    /* -1.0 is a hack based on expts in saber */
    if(dLen > safelen) {
	UDSample(src, stIndex, dst, sLen, safelen, 1.0/exFact);
	ClearFloats(dst+safelen, 1, (dLen - safelen));
/* 	fprintf(stderr, "padded %d\n", dLen - safelen);		*/
	}
    else
	UDSample(src, stIndex, dst, sLen, dLen, 1.0/exFact);
    }

#define MAX_EXT_FACT  4.0

void Resynth(FLOATARG overlap, FLOATARG duration, FLOATARG samplingRate, 
	     long frameSize)
/* <overlap> is how much successive frames overlap */
    {
/*     long	frameSize;   */
    float	t0 = 0.0;	/* t0 is where we have written to in o/p */
    float	t_dash = 0.0;	/* t-dash is corresponding time in origfile */
    float	rbBaseT = 0.0;
    long	opSamples;
    float	thisFrameLen;
    float	halfDuration;
    float	timeAdvance = 0.0;
    float	*frame, *opBuf;
    long	opLen, opFudge;
    float	ovPoint;	/* crossover between frames */
    float	k;
    float	tpsFactor, strchFactor, thisFrameStretch;
    int 	frameCtr = 0;
    int 	baseIx = 0;

    frame = MakeBuf(frameSize*2L);	/* complex ..? */
    opLen = (long)(frameSize * MAX_EXT_FACT);
    opBuf = MakeBuf(opLen);
    opFudge = opLen/16;
    ovPoint = overlap / 2;	/* crossover is at half of overlap */
    k = 0.5 - overlap / 2;
        /* if overlap is 100%, advance is zero, as is 0.5 - overlap/2 */
        /* if overlap is 40%, t0 needs to be advanced by 30% of thisfrm */

    while(t_dash < duration)  {
	tpsFactor = TransposeFactor(t_dash);
	strchFactor = StretchFactor(t_dash);
	    /* tpsFactor times frame length is transposed frame length */
	thisFrameLen = frameSize * tpsFactor;
	halfDuration = (thisFrameLen * k) / samplingRate;
	timeAdvance += halfDuration;
	    /* frameSize = */ 
	SynthFrame(t_dash, timeAdvance, frame);
	FFResample(frame, frameSize, opBuf, opLen, 
		   ((float)frameSize)/2.0, 
		   (t0+halfDuration-rbBaseT)*samplingRate + opFudge,
		   tpsFactor);
	if(frameCtr >= dbgDump && frameCtr < dbgDump+dbgCont)  {
	    char name[32];
	    
	    sprintf(name, "time %f", t0);
	    DumpSet(opBuf, opLen, baseIx, name);
	    }
	if(dbgUniq >= 0)  {	/* debug: allow but one frame through */
	    if(frameCtr != dbgUniq)
		ClearFloats(opBuf, 1, opLen);
	    }
	/* figure out space left behind t0? */
	opSamples = timeAdvance * samplingRate;
	OverlapFrameIntoRingBuffer(opBuf, opSamples);
	WriteOutSomeOfTheRingBufferAndClearIt(opSamples);
	rbBaseT += ((float)opSamples) / samplingRate;
	OverlapFrameIntoRingBuffer(opBuf+opSamples, opLen-opSamples);
	t0 += 2*halfDuration;
	t_dash += 2*halfDuration / strchFactor;
	timeAdvance = halfDuration;
	++frameCtr;
	baseIx += opSamples;
	}
    }

static FILE *dumpFile = NULL;
static char dumpFileName[] = "stfts.xgp";

DumpSet(float *data, long len, long offs, char *desc)
    {
    if(dumpFile == NULL)  {
	if( (dumpFile = fopen(dumpFileName, "w"))==NULL) {
	    fprintf(stderr, "DumpSet:  cannot create dump file '%s'\n",
		    dumpFileName);	abort();    }
	/* output options here */
	fprintf(dumpFile, "TitleText: %s output\n", programName);
	}
    else if(len < 0) {	/* flag to close file */
	fclose(dumpFile);
	}
    else {	/* dump the data */
	long l;

	fprintf(dumpFile, "\n");	/* blank line for new set */
	if(desc != NULL)  {
	    fprintf(dumpFile, "\"%s\"\n", desc);
	    }
	for(l=0; l<len; ++l)  {
	    fprintf(dumpFile, "%l8d %f\n", offs+l, data[l]);
	    }
	fflush(dumpFile);
	}
    }

