/*******************************************************\
*	stfts2.c					*
*  attempt to rewrite pvoc-style synthesis		*
*  more clearly						*
*  dpwe 06nov91						*
\*******************************************************/


float 	inputTime, outputTime, basicTimeIncrement;
float 	resampleRate;	/* >1.0 means increase number of samples */
cpxpolar	*spectrum;
float 	*signal, *resamSig, outputSig;
float 	*lastPhases;
AFTINFO	*aftFile;
long 	spectSize;
float 	inputOverlapFactor, inputSrate;
TWINDOW *window;

window = NewTwindow(spectSize, (float)spectSize, HAMM_TYPE, 1.0);
outputTime = 0.0;	/* output time for centre of first frame */
/* spectSize is also size of time window in samples.  Thus the basic 
   frames resynthesise to this many points.  We overlap these frames so 
   that <inputOverlapFactor> overlap at each point, giving our basic 
   interval between successive frames ... */
basicTimeIncrement = spectSize / inputOverlapFactor / inputSrate;

/* CumulatePhases stores the current cumulated phase and the time to which 
   it applys in internal state.  It is preset to the absolute phase for 
   the first time point here. */
InterpolateSpectrum(aftFile, InputTime(outputTime), spectrum, PHI_MODE);
CumulatePhases(spectrum, spectSize, outputTime, PRESET_MODE);

while(!done)
    {
    /* the two control parameters, inputTime index and resampleRate are 
       defined as functions of the outputTime index */
    resampleRate = ResampleRate(output_time);
    inputTime = InputTime(outputTime);
    
    /* interpolation returns magnitudes and phi-dot at specified moment */
    InterpolateSpectrum(aftFile, inputTime, spectrum, PHIDOT_MODE);
    /* new phase spectrum is phi(tn-1) + phi-dot(tn) * (tn - tn-1) */
    CumulatePhases(spectrum, spectSize, outputTime, CUMULATE_MODE);
    IFFT(spectrum, spectSize, signal);
    ApplyTwindow(window, signal, spectSize, signal, spectSize);
    resamSize = Resample(signal, resampleRate, resamSig);
    OverlapAdd(resamSig, outputSig);

    /* the input frame was stretched in time by a factor of resample_rate;
       perhaps we should step forward by a proportional amount to find 
       the centre of the next block to lay down .. else scale ampls */
    output_time += resample_rate * basic_time_increment;




#define MMmaskPhs(p,q,r,s) /* p is pha, q is as int, r is PI, s is 1/PI */ \
    q = (int)(s*p);	\
    p -= r*(float)((int)((q+((q>=0)?(q&1):-(q&1) )))); \

void CumulatePhases(spect, size, time, mode)
    cpxpolar	*spect;
    long 	size;
    FLOATARG	time;
    int 	mode;
    {
static	float	*lastPhases = NULL;
static	long 	lpSize   = 0;
static	float	lastTime = -1.0;	/* illegal preset value flag */
    long 	l;
    float 	dt, pha; 
    /* q,r,s for MMmaskPhs .. */
    int 	q;
    float	r = PI;
    float	s = 1.0/PI;

    /* initialisation block */
    if(lastPhases == NULL)
	{
	lpSize = size;
	lastPhases = (float *)malloc(sizeof(float) * lpSize);
	}
    if(size > lpSize)	/* check that size hasn't suddenly increased */
	{  fprintf(stderr, "CumPhas: size %ld > lpSize %ld\n", size, lpSize);
	   abort();	}
    /* special case of mode input - preset */
    if(mode == PRESET_MODE)
	{
	for(i = 0; i<size; ++i)
	    lastPhases[i] = spect[i].pha;
	lastTime = time;
	}
    else{	/* assume CUMULATE_MODE i.e. do the phi-dot stuff */
	dt = time - lastTime;
	for(i = 0; i<size; ++i)
	    {
	    pha = lastPhases[i] + spect[i].pha * dt;
	    MMmaskPhs(pha, q, r, s);
	    lastPhases[i] = spect[i].pha = pha;
	    }
	lastTime = time;
	}
    }
