#include <stdio.h>
#define __STDC__
#include <dpwelib.h>
#include <dsputil.h>
#include <resample.h>
#include <window.h>

extern float *GetSpect PARG((FLOATARG index));

#define MAX_EXT_FACT	4.0	/* transpose 2 octaves down */

void FResample(src, sLen, dst, dLen, sIndex, dIndex, exFact)
    float	*src;
    long 	sLen;
    float	*dst;
    long	dLen;
    float	sIndex;		/* this sample point in the source .. */
    float	dIndex;		/* ..should fall on this point in dest */
    float	exFact;		/* something like dLen/sLen */
    {	/* exFact is how much longer each input time tick lasts 
	   that the output time ticks.
	   i.e. n _input_ ticks last as long as (n*exFact) output ticks
	   thus n _output_ ticks correspond to (n/exFact) input ticks   */
    float 	stIndex;
    long	skip, safelen;

    stIndex = sIndex - dIndex/exFact;
    if(stIndex < 0.0) {
	skip = (-stIndex)*exFact+0.99;
	fprintf(stderr, "skip  %d\n", skip);
	ClearFloats(dst, 1, skip);
	dst += skip;
	dLen -= skip;
	stIndex += (float)skip / exFact;	/* will become >= 0 */
	}
    /* how many output points can we ask for safely?  */
    safelen = ((float)sLen - stIndex - 1.0)*exFact;
    /* -1.0 is a hack based on expts in saber */
    if(dLen > safelen) {
	UDSample(src, stIndex, dst, sLen, safelen, 1.0/exFact);
	ClearFloats(dst+safelen, 1, (dLen - safelen));
	fprintf(stderr, "padded %d\n", dLen - safelen);
	}
    else
	UDSample(src, stIndex, dst, sLen, dLen, 1.0/exFact);

    }

void RResynth(FLOATARG overlap, long end, long frameSize)
/* overlap how much successive frames overlap */
    {
    long	framelen;

    float	n0 = 0.0;	/* n0 is index of center of current frame */
    long	rbBase = 0;
    float	thisFrameLen, thisFrameCenter;
    float	timeAdvance = 0.0;
    float	*spect, *frame, *opBuf;
    long	opLen;
    float	ovPoint;	/* crossover between frames */
    float	k;
    float	tpsFactor;

    frame = MakeBuf(frameSize);
    opLen = (long)(frameSize * MAX_EXT_FACT);
    opBuf = MakeBuf(opLen);
    ovPoint = overlap / 2;	/* crossover is at half of overlap */
    k = 0.5 - overlap / 2;
        /* if overlap is 100%, advance is zero, as is 0.5 - overlap/2 */
        /* if overlap is 40%, n0 needs to be advanced by 30% of thisfrm */

    while(n0 < end)  {
	tpsFactor = TransposeFactor(n0);
	    /* tpsFactor times frame length is transposed frame length */
	thisFrameLen = frameSize * tpsFactor;
	thisFrameCenter = n0 + thisFrameLen * k;
	    /* thisFrameCenter is the middle of the frame after resample */
	timeAdvance += thisFrameLen * k;
	spect = GetSpect(n0);
	SynthFrame(spect, timeAdvance, frame);
	FResample(frame, frameSize, opBuf, opLen, 
		  (frameSize+1.0)/2.0, n0-rbBase+tpsFactor*frameSize*k,
		  tpsFactor);
	OverlapFrameIntoRingBuffer(opBuf, opLen);
	WriteOutSomeOfTheRingBufferAndClearIt();
	OverlapRestOfFrameIntoRingBuffer();
	n0 += thisFrameLen * k * 2;
	timeAdvance = tpsFactor * frameSize * k;
	}

    }

#define RSTABLEN 1000		/* sets resolution of resampling fraction */
#define RSLOBES	 16		/* reduces aliasing in resampling */

static RsmpDescType *rdp = NULL;

void InitRdp()
    {
    if(rdp == NULL) {
	rdp = NewResampKernel(RSTABLEN, RSTABLEN, RSLOBES, 1.0);
	}
    }

void IntResample(src, sLen, dst, dLen, sIndex, dIndex, exFact)
    float	*src;
    long 	sLen;
    float	*dst;
    long	dLen;
    float	sIndex;		/* this sample point in the source .. */
    float	dIndex;		/* ..should fall on this point in dest */
    float	exFact;		/* something like dLen/sLen */
    {
    long 	ifStart;
    float	actExF;
    long 	lobeLen;
    int 	inPd, outPd;

    lobeLen = rdp->inPrd;
    if(exFact > 1.0)  {		/* interpolating input */
	inPd = lobeLen;
	outPd = lobeLen / exFact;	/* outPd < inPd, so get more */
	}
    else{			/* exF < 1.0; lowpass and downsample input */
	outPd = lobeLen;
	inPd = lobeLen * exFact;
	}
    actExF = inPd / (float)outPd;
    ifStart = outPd * dIndex - inPd * sIndex;
    /* see comment on ifStart in lib/resample.c */
    if(ifStart < 0) 	abort();
    rdp->inPrd = inPd;
    rdp->outPrd = outPd;
    ResampleIntoBuffer(dst, dLen, src, sLen, ifStart, rdp, 1 /* chk */ );
    rdp->inPrd = rdp->outPrd = lobeLen;
    }

void DanTest(void)
    {
    float *buf1, *buf2;
    WINDAT *win1, *win2;
    long siz = 512;

    win1 = NewWin("test1", 0);
    win2 = NewWin("test2", 0);
    buf1 = MakeBuf(siz);
    buf2 = MakeBuf(siz);
    buf1[16] = 1.0;
    FResample(buf1, 32, buf2, siz, 0.0, 0.0, 10.0);
    DoDisp(win1, buf1, 32);
    DoDisp(win2, buf2, siz);
    }

