/*******************************************************************\
*   peaks.h							    *
*   pick & match spectral peaks, after DSPbench	    		    *
*   10jul91 after 02apr91 dpwe					    *
\*******************************************************************/

/* std includes */

#ifndef _PEAKS_H_
#define _PEAKS_H_

#include <math.h>
/* other handy things */
#include <llists.h>

/* constants */

#define NPEAKS  64	/* number of peaks we will pick */
#define T_PEAK	100	/* <llist> type identifier for peak node */
#define LOGN2DB		(8.685889638)	/* (20/log(10)) * log(a) = dB(a) */
#define LINOFDB(mag)	(exp((mag)/LOGN2DB))
#define DBOFLIN(mag)	(LOGN2DB*log(mag))

/* macros */

/* PHMSK - macro to mask phases quickly */
/* declare variables needed by phase mask macro */
#define PHMSKSETUP           int  __q;		\
                      static double __r = PI;	\
                      static double __s = 1.0/PI; \
/* actually mask the phase */
#define PHMSK(p) (__q = (int)(__s*(p)),	\
                 p - __r*(double)(__q + ( (__q>=0) ? (__q&1): -(__q&1) ) )) \

/* types */

typedef struct bigpeakpos 
    {
    char	label[12];	/* debug string */
    long 	index;		/* unique count id */
    float	frq;
    float	mag;
    float	phs;
    float	ophs;		/* (alternate phase temp space) */
    float	tim;
    int 	x,y;		/* scratch screen location */
    struct bigpeakpos *next;	/* link for temporal evolution ? */
    } BigPeakPos;

/*** IF YOU CHANGE THESE STRUCTS, SEE tkplay.c:pkWriteSmall - 
     it assumes that the structs start with the three floats that 
     matter, and saves just them ***/

typedef struct peakpos 
    {
    float	frq;
    float	mag;
    float	tim;
    short 	x,y;		/* scratch screen location */
    float	phs;
    float	ophs;		/* (alternate phase temp space) */
    } PeakPos;

typedef struct minpeakpos 	/* (phaseless version for frq-only ops) */
    {				/* FIELDS MATCH PEAKPOS FOR `ALIASING' */
    float	frq;
    float	mag;
    float	tim;
    short 	x,y;		/* scratch screen location */
    } MinPeakPos;

/* 1992oct27:  specialization of llist node allows neater code, 
   exposes this code to changes in llist */
typedef struct _pk_node {
	struct _pk_node *next;
	PeakPos         *peak;
	int    type;		/* data type MUST be T_PEAK for this str */
} PKNODE, *PKLIST;

/* prototypes */

void ResetPeakIndex PARG((void));
PeakPos *NewPeak PARG((FLOATARG frq, FLOATARG mag, 
		       FLOATARG phs, FLOATARG tim));
void FreePeak PARG((PeakPos *peak));
PKLIST ListPeaks PARG((float *magfrm, int maginc, float *phsfrm, int phsinc,
		       long len, int sign, FLOATARG binsPOct, FLOATARG time, 
		       FLOATARG frqBase, FLOATARG frqLim, int logFlg));

/* from trkprint.c */
void PrintPeak PARG((PeakPos *peak, char *label));
void PrintPeakList PARG((PKLIST list, char *name));

void pklistFree PARG((PKLIST pl));

#endif /* _PEAKS_H_ */
