/***************************************************************\
*	pkmatch.c		  				*
*   Match peaks between two sorted lists			*
*   Based on routines developed within dspB, for thesis		*
*   dpwe 08jul91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
/* personal libs */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include "peaks.h"
#include "tracks.h"

/* constants */
/* arbitrary max frq diff for match - as log of ratio.  .058 is 1 semitone */
/* #define BINTHRESH	(2.0*.05776)	/* */
#define BINTHRESH	(3.0*0.5*.05776)	/* */
#define SDSCALE		(BINTHRESH/(6.0*fudge))
/* emprically set to to best compromise of 'right' thing on gould1.aif */
/* SDSCALE is applied to deviations between actual and expected 
   (running average) magnitudes and frequnecies.  This difference 
   is scaled by the running standard deviation, then added to the 
   distance score.  So if the mag is more than some number of sds 
   away from the average, the distance will break the threshold. */
#define TKLENTHRESH	12
/* The average and sd measures take a while to settle down, so those 
   rules aren't applied except after a track is this many samples 
   long. */

/* macros */
#define DB2LIN(a)	(exp(a*0.11512925))
#define LIN2DB(a)	(8.6858896*log(a))

/* static function prototype */
static float PeakDistance PARG((PeakPos *pk1, PeakPos *pk2));
static float DistTrackToPeak PARG((TRACK *trk, PeakPos *pk));

/* static data */
extern float fudge;	/* command line fudge factor for sd sens */

/* functions */

static float PeakDistance(pk1, pk2)
    PeakPos	*pk1;
    PeakPos	*pk2;
    {
    if(pk1 == NULL || pk2 == NULL)
	return 99999.9;	/* a large number to indicate an impossible match */
    else
	return( fabs(log(pk1->frq/pk2->frq)) );
/*	return( fabs(pk1->frq - pk2->frq) );	*/
/* Sampling interval implies that tracks will be sampled at most 
   1/(2*bandwidth) apart = 1/(2*centre-freq/Q)  = Q/(2 cf) = 4/cf ..? */
    }

static float FrqDistTrackToPeak(trk, pk)
    TRACK	*trk;
    PeakPos	*pk;
{	/* a FREQUENCY-ONLY distance metric for comparing peaks.
	   - frq+mag metric screws up on list-matching algo since 
	   small peaks either side of a biggie may cause it to be 
	   skipped and lost.  */
    float	f1, f2, fav, fsd, dist = 0.0;

    if(trk == NULL || pk == NULL)
	return 99999.9;	/* a large number to indicate an impossible match */
    else{
	f1 = pk->frq;
	f2 = ((PeakPos *)car((LIST)trk->tkbot))->frq;
	dist = fabs(log(f1/f2));
	return dist;
    }
}

static float DistTrackToPeak(trk, pk)
    TRACK	*trk;
    PeakPos	*pk;
    {
    float	f1, f2, fav, fsd, dist = 0.0;

    if(trk == NULL || pk == NULL)
	return 99999.9;	/* a large number to indicate an impossible match */
    else{
/*	return( fabs(log(pk->frq/((PeakPos *)car(trk->tkbot))->frq)) );	*/
	f1 = pk->frq;
	f2 = ((PeakPos *)car((LIST)trk->tkbot))->frq;
	dist = fabs(log(f1/f2));
#ifdef SDTHRESH	/* *IF* we allow tracks to break based on variance excess */
	if(trk->st.npts > TKLENTHRESH)  {
	    dist += SDSCALE * (fabs(pk->mag - trk->st.mavg)/(trk->st.msd+.01)
			       + fabs(f1-f2)/(trk->st.fsd+.01));
/*	    if(dist > BINTHRESH && fabs(f1-f2) < 2.0)
		fprintf(stderr, "\nt %.6f mag %.2f mav %.2f(%d) ms %.2f(%d), f1 %.2f f2 %.2f fs %f\n", 
			pk->tim, pk->mag, trk->st.mavg, 
			(pk->mag==trk->st.mavg), 
			trk->st.msd, (trk->st.msd==0.0), 
			f1, f2, trk->st.fsd);
 */
	}
#endif /* SDTHRESH */
#ifdef DDEBUG
	if(fabs(f1-2000.0)<10.0 
	   && fabs(pk->tim - 0.325) < 0.003 
	   && fabs(trk->st.favg - 2000.0) < 10.0)  {
	    fprintf(stderr, "DTP: t%8.6f, pkf %7.1f, m %5.2f, tkma %7.3f, ms %6.3f, dist %10.6f\n", pk->tim, f1, pk->mag, trk->st.mavg, trk->st.msd, dist);
	}
#endif /* DDEBUG */
        return dist;
	/* add in term due to magnitude variation too */

/* ***extra ignored *** */
	fav = trk->st.favg;
	fsd = trk->st.fsd;
	if(fsd == 0)
	    fsd = fav/20.0;	/* default 5% ? 10% ? sd range */
	dist = fabs(f1 - fav)/fsd;
/*	if(fabs(f1-2300.0)<100.0)
	    fprintf(stderr,
	      "DTP: pk f %7.1f m %7.3f t %7.4f tk fa %7.1f fs %6.1f d %7.3f\n",
	      f1, pk->mag, pk->tim, fav, fsd, dist);	*/
	return dist;
    }
}

PKLIST MatchPeakLists(existingTracks, newPeaks)
    TKLIST	existingTracks;		/* list of ptrs to PeakNodes of trax */
    PKLIST	newPeaks;		/* list of PeakNodes for this frame */
    /* extend the track-tails in exTk by attaching els of newPeaks.
       Return the depleted newPeaks if any */
    {
    float	thresh = BINTHRESH; 
    /* ! how close peaks must be to match - arbitrary at present */
    PeakPos	*thisTail, *nextTail, *thisNew, *nextNew;
    float	distTtTn, distTtNn, distNtTn, distNtNn;
    TRACK	*thisTrack, *nextTrack; /* current track stats */
    PKLIST	newLeft, *newFence;	/* <newLeft> returns unused peaks */

    newLeft = newPeaks;		newFence = &newLeft;

    if(existingTracks)	thisTrack = (TRACK *)car((LIST)existingTracks);
    else		thisTrack = NULL;
    if(thisTrack)	thisTail  = (PeakPos *)car((LIST)thisTrack->tkbot);
    else		thisTail  = NULL;

    if(newPeaks)  {
	thisNew = (PeakPos *)car((LIST)newPeaks);
	if(cdr((LIST)newPeaks))	nextNew = (PeakPos *)car(cdr((LIST)newPeaks));
	else			nextNew = NULL;
#ifdef DEBUG
	/******** DEBUG 1993mar07 ************/
	if(thisNew->tim > 1.430139 && thisNew->tim < 1.430141)  {
	  PrintPeakList(newPeaks, "spect");
	}
#endif /* DEBUG */
	}
    else{	thisNew = NULL;	nextNew = NULL;		}
    distTtTn = DistTrackToPeak(thisTrack, thisNew);
    distTtNn = DistTrackToPeak(thisTrack, nextNew);
    while(thisNew && thisTail)  {
	if(existingTracks 
	   && (existingTracks = (TKLIST)cdr((LIST)existingTracks)) )
	    nextTrack = (TRACK *)car((LIST)existingTracks);
	else
	    nextTrack = NULL;
	if(nextTrack)	
	    nextTail  = (PeakPos *)car((LIST)nextTrack->tkbot);
	else		
	    nextTail  = NULL;
	if(nextTail && cartype((LIST)nextTrack->tkbot)!=T_PEAK)	abort();
	/* 1993mar08 : cannot let thisTail compare itself with a peak 
	   of even more distant frequency, so limit step-on to only 
	   happen when thisNew starts below the frq of the tail */
	while(nextNew && thisNew->frq < thisTail->frq 
	      && distTtTn > distTtNn)  {	/* next new is better peak */
	    distTtTn = distTtNn;
	    newFence = (PKLIST *)next_ptr((LIST)newPeaks);
	    newPeaks = (PKLIST)cdr((LIST)newPeaks);
	    thisNew = nextNew;
	    if(newPeaks && cdr((LIST)newPeaks)) 
		nextNew = (PeakPos *)car(cdr((LIST)newPeaks));
	    else
		nextNew = NULL;
	    distTtNn = DistTrackToPeak(thisTrack, nextNew);
	}
	/* distTtTn is best, or no more NewPeaks */
	distNtTn = DistTrackToPeak(nextTrack, thisNew);	
	distNtNn = DistTrackToPeak(nextTrack, nextNew);	
	if(distTtTn < thresh			/* thisTail & thisNext close */
	   && (nextTail==NULL			/* no following tail */
	       || (distTtTn < distNtTn 		/* or thTl closer */
		   || (nextNew
		       && distNtNn < distNtTn)))) /* or nxNw closer to nxTl */
	    {
	    set_next((LIST)thisTrack->tkbot, (LIST)newPeaks);
	    close_up_ptr((LIST *)newFence);	/* thisNew removed from list */
	    newPeaks = *newFence;
	    thisNew = nextNew;
	    if(newPeaks && cdr((LIST)newPeaks)) 
		nextNew = (PeakPos *)car(cdr((LIST)newPeaks));
	    else
		nextNew = NULL;
	    distNtTn = DistTrackToPeak(nextTrack, thisNew);
	    distNtNn = DistTrackToPeak(nextTrack, nextNew);
	    }
	else {          /* closest was not within THRESH -> track died */
#ifdef DDEBUG
	  if(thisTail->mag > 0.0)  {    /* worthy of our concern */
	    fprintf(stderr, "track DIED:\n");
	    fprintf(stderr, "dtt %.3f (th %.3f) dnt %.3f dnn %.3f (%lx,%lx)\n", 
		    distTtTn, thresh, distNtTn, distNtNn, nextTail, nextNew);
	    tkPrint(thisTrack, "track");
	    PrintPeak(thisTail, "tail");
	    PrintPeak(thisNew, "new");
	  }
#endif /* DDEBUG */
	}
	/* either way, finished with thisTail */
	thisTrack = nextTrack;	thisTail  = nextTail;
	distTtTn  = distNtTn; 	distTtNn  = distNtNn;
	}
    return newLeft;			/* list of unmatched new pks */
}
