/* 
 * tkdemo.c
 * Demonstration program shows how to use track interface
 * dpwe 1993mar10
 */

/* make command: (*.o files in ~dpwe/src/trk)
 * gcc -g -I ~dpwe/lib tkdemo.c -o tkdemo \
 *	pklist.o trkheadio.o trkio.o tkplay.o tsin.o snd.o \
 *	~dpwe/lib/libdpwe.a -lm
 */

#include <dpwelib.h>	/* dpwe's defines (in ~dpwe/lib) */
#include <llists.h>	/* linked list definitions (in ~dpwe/lib) */
#include <audIO.h>	/* real-time play output (in ~dpwe/lib) */
#include "tkplay.h"	/* includes track and peak definitions */

char *programName;

/* required to link somewhere */
float	dbgFrq = 0;
float	dbgFrqWin = 0;

main(int argc, char **argv)
{
    TKSTRUCT 	*trkh;	/* track file descriptor */
    LIST	trax;	/* the actual data itself */
    LIST	remtrax;
    PKLIST	track;	/* one track from the data */
    PeakPos	*peak;	/* one sample from a track */
    double	maxfrq = 0;
    double	samplingRate = 8000.0; 	/* Hz on sun */
    double	magthresh    = -20.0;	/* ignore tracks with lower mag */
    AUSTRUCT 	*au;

    programName = argv[0];
    if(argc != 2)  {
	fprintf(stderr, "usage: %s trackfile    prints highest frq in file\n",
		programName);
	exit(1);
    }
    if( (trkh = tkLoadPrune(&trax, argv[1], magthresh)) == NULL)  {
	fprintf(stderr, "%s: could not open '%s' as trackfile\n", 
		programName, argv[1]);
	exit(1);
    }

    /* Find highest frequency in any track */
    remtrax = trax;
    while(remtrax)  {
	track = (PKLIST)car(remtrax);
	remtrax = cdr(remtrax);

	while(track)  {
	    peak = track->peak;
	    if(maxfrq < peak->frq)	maxfrq = peak->frq;
	    track = (PKLIST)cdr((LIST)track);
	}
    }
    printf("%s: the highest frequency in '%s' is %.2f\n", 
	   programName, argv[1], maxfrq);

    /* play it too - why not? */
    au = AUOpen(AFMT_SHORT, samplingRate, 1 /* chans */, 0 /* dflt buf */, 
		AIO_OUTPUT);
    tkSynthTrax(trax, samplingRate, NULL, (void *)0x1);
    AUClose(au, 0 /* don't flush */);
}
