/***************************************************************\
*	tksyn.c			  				*
*   Takes 'track' file as written by trk and uses the tracks 	*
*   to drive a bank of sinewave oscillators, superimposed into 	*
*   a single soundfile, the resynthesised sound.		*
*   dpwe 25sep91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
/* personal libs */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include <sndf.h>
#include "peaks.h"
#include "tracks.h"
#include "trkheadio.h"

/* constants */
#define  DEF_EXT	".rs.irc"	/* default output extension */
#define  DEF_SRATE	44100.0		/* default sampling rate */

/* static function prototype */
static void PerformSynth PARG((char *inPath, char *outPath));
static void usage PARG((void));

/* static data */
       char	*programName;
static int	verb = 0;		/* -v flag */
       int	graphs = 0;		/* use debugging graphs */
static float    srate = DEF_SRATE;
static float    gain  = 1.0;		/* scale output by this */

/* print out data on trks in this window */
float dbgFrq = 2000.0;
float dbgFrqWin = 5.0;

/* functions */

main(argc,argv)
    int 	argc;
    char	**argv;
    {
    int 	i = 0;
    int 	err = 0;
    char	*inPath = NULL, *outPath = NULL;

    programName = argv[0];
    /**** %% DPWE STD ARG HANDLER %% ****/
    i = 0;
    while(++i<argc && err == 0)
	{
	char c, *token, *arg, *nextArg;
	int  argUsed;

	token = argv[i];
	if(*token++ == '-')
	    {
	    if(i+1 < argc)	nextArg = argv[i+1];
	    else		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)
		{
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)
		    {
/*	        case 'q': case 'Q':	qFactor = atof(arg);	argUsed = 1;
/*		    if(qFactor<1)
/*			{  fprintf(stderr,"%s: Q of %s = %.2f must be > 1\n",
/*				programName, arg, qFactor);
/*			   err = 1;  	}
/*		    break;
 */
	        case 'r':		srate = atof(arg);	argUsed = 1;
		    if(srate<4000||srate>100000)
			{  fprintf(stderr,"%s: sr of %s = %.2f not 4k..100k\n",
				programName, arg, srate);
			   err = 1;  	}
		    break;
	        case 'g':		gain = atof(arg);	argUsed = 1;
		    if(gain>100000)
			{  fprintf(stderr,"%s: gain of %s = %.2f not <100k\n",
				programName, arg, gain);
			   err = 1;  	}
		    break;
		case 'v': 	verb = 1; break;
		default:	fprintf(stderr,"%s: unrec option %c\n",
					programName, c);
		                err = 1; break;
		    }
		if(argUsed)
		    {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		    }
		}
	    }
	else{
	    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	outPath = argv[i];
	    else{ fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
		  err = 1;  }
	    }
	}
    /* other tests on args */
/*    if(data < 0)
 *	{ fprintf(stderr,"%s: bad data %d: must be >= 0\n", programName, data);
 *	  err = 1; }
 */
    if(err || inPath == NULL)
	usage();		/* never returns */

    if(outPath == NULL)
	{
	outPath=
	    TMMALLOC(char, (1+strlen(inPath)+strlen(DEF_EXT)), "outPath");
	strcpy(outPath, inPath);
	SubstExtn(inPath, DEF_EXT, outPath);
	}

/*    fprintf(stderr,"%s: data = %d\n",programName,data);
 *    fprintf(stderr,"%s: verbose = %s\n",programName,verb?"yes":"no");
 *    fprintf(stderr,"%s: infile %s, outfile %s\n",programName,inPath,outPath);
 */
    /* args all dealt with.  On to the real business */
    PerformSynth(inPath, outPath);
    }

void diess(s1, s2)
    char *s1;
    char *s2;
    {
    char *s3 = "";

    if(s1 == NULL)	s1 = s3;
    if(s2 == NULL)	s2 = s3;
    fprintf(stderr, "%s: %s : %s\n", programName, s1, s2);
    exit(0);
    }

static void PerformSynth(inPath, outPath)
    char *inPath;
    char *outPath;
    {
    FILE	*inFile, *outFile;
    int 	i, err, done=0, ndone=1;
    SFSTRUCT	*outSnd, *newSnd;
    TKHDSTRUCT	*trkh;
    LIST	trax, track;
    int 	chans = 1, dsize = sizeof(float), format = SFMT_FLOAT;
    long 	buflen = 1024;	/* ? */
    float 	thisOnesStartTime, latestTime, sRate, ph;
    float 	*fp2, *fptr;
    long 	samps, dBsize, n, todo;
#ifdef DEBUGGRAPHS
    WINDAT	*wdptr;
#endif
    double	time, dt, duration, timeFrame;
    LIST	currentTails, currentPeaks, strayPeaks, allTrax;
    int 	debug = 0;
    float	dbgTime = 0.2;

    /* open input CQT file */
/*     if( (inFile = fopen(inPath, "r"))== NULL)	*/
    if( (inFile = TKOpenAllocRdHdr(inPath, &trkh)) == NULL)
	diess("Could not open input file", inPath);
    /* create structure for output sound */

    sRate = trkh->samplingRate;
    chans = trkh->channels;
    if( err = SFAlloc(&outSnd, SF_UNK_LEN, SFMT_SHORT, trkh->samplingRate, 
		      trkh->channels, SFDFLTBYTS))
	SFDie(SFE_NOPEN, NULL);
    /* check that we can open output file before we spend too long */
    if( (outFile = SFOpenWrHdr(outPath, outSnd))==NULL )
	SFDie(SFE_NOPEN, outPath);

    /* load input file */
    ReadTrackFile(&trax, inFile);

    duration = trkh->duration;
    samps  = (long)(duration * sRate);
    dBsize = dsize*chans*samps;
    if( err=SFAlloc(&newSnd, dBsize, format, sRate, chans, SFDFLTBYTS) )
	SFDie(err, "SynthATrack" );
    fptr = (float *)SFDataLoc(newSnd);
    ClearFloats(fptr, 1, samps);
    while(trax) 
	{	/* for each track .. */
/*	do  {	/* scan to a track worth synthing? (poor metric) */
/*	    track = car(trax);
/*	    trax = cdr(trax);
/*	    } while(trax && ((PeakPos *)car(track))->mag < -10.0);	*/
	track = car(trax);
	trax  = cdr(trax);	

	n = 0;	fp2 =fptr;	/* start at top of sound (again) */
	thisOnesStartTime = ((PeakPos *)car(track))->tim;
	if(thisOnesStartTime > latestTime)	/* progress report */
	    {	    fprintf(stderr,"%6.2f",latestTime); fflush(stderr);
		    latestTime += .1;	    }
	while(n<samps)	/* this buflen chunking is unneeded now.. */
	    {		/* synth each track over ENTIRE LENGTH of sound ?? */
/* 	    if(samps-n > buflen)
		todo = buflen;
	    else	*/
		todo = samps - n;      
	    time = n / sRate;
/*	    fprintf(stderr, "SAT time %.4f\n",time);	fflush(stderr);	*/
/*	    PrintPeak((PeakPos *)car(track), "Track head"); */
	    ph = TrackOscillator(&track, &time, fp2, todo, ph, 
				 sRate);
	    n += todo;
	    fp2 += todo;
	    }
	}
    if(gain != 1.0)
	VSMul(fptr, 1, gain, samps);	/* apply gain if not unity */
    if( err = ConvertSoundFormat(&newSnd, SFMT_SHORT))
	SFDie(err, NULL);

    SFfwrite(SFDataLoc(newSnd), SFMT_SHORT, samps, outFile);

    SFCloseWrHdr(outFile, outSnd);
    fclose(inFile);
    fprintf(stderr, "\n%s: wrote %s (%ld bytes)\n",
	    programName, outPath, outSnd->headBsize+outSnd->dataBsize);

#ifdef DEBUGGRAPHS
    if(graphs)
	{
	wdptr = NewWin(programName,0);
	DoDisp(wdptr, fp2, 1); /* meaningless args; no use at present */
	}
#endif

    }

static void usage PARG((void))		/* print syntax & exit */
    {
    fprintf(stderr,
 "Usage: %s [-v][-r srate] inTrx [outSNDF]\n",
	    programName);
    fprintf(stderr,"where  (defaults in brackets)\n");
/*    fprintf(stderr,
/*" -q Qfact   is the q-factor of the bins       (%6.1f)\n", (float)DFLT_QFCT);
 */
    fprintf(stderr,
 " -v         verbose mode\n");
    fprintf(stderr, 
 " -r srate   sampling rate for output file (%.1f)\n", DEF_SRATE);
    fprintf(stderr,
 " inTrx      is a Track file for input\n");
    fprintf(stderr,
 " outSNDF    is the output soundfile       (inStem+'%s')\n", DEF_EXT);
    exit(1);
    }
