/***************************************************************\
*	tracks.h		  				*
*   Definition of TRACK (peak list) structure & some fns	*
*   Header for trkmaint.c, trkstats.c etc.			*
*   dpwe 08jul91 renamed from trk.h 1993apr18			*
\***************************************************************/

#ifndef _TRK_H_
#define _TRK_H_

#include <dpwelib.h>
/* #include <llists.h> */
#include "peaks.h"

typedef struct {	/* stats of a track */
    long	npts;
    float	tmin, tmax, tavg, tsd;
    float	fmin, fmax, favg, fsd;
    float	dfdt;		/* running frequency derivative */
    float	mmin, mmax, mavg, msd;
    int		state;		/* TKST_NULL, TKST_BOILUP or TKST_BOILDOWN */
} TrackStats;

enum {		/* values for TrackStats.state */
    TKST_NULL = 1,	/* just reset */
    TKST_BOILUP,	/* running totals, mid-calculation */
    TKST_BOILDOWN	/* `boiled-down' final values */
};

typedef struct track {
    PKLIST	tktop;
    PKLIST	tkbot;		/* 1st, last nodes in track nodelist */
    TrackStats	st;		/* description */
    float	startPhase;	/* (hold this for phaseless tracks) */
    long	id;		/* unique index of this track */
    int 	flags;		/* bits indicate state */
} TRACK;

enum {		/* bits for TRACK.flags */
    TF_EMPTY  = 0, 
    TF_TKTOP  = 1,	/* ->tktop is valid */
    TF_TKBOT  = 2,	/* ->tkbot is valid */
    TF_STID   = 4,	/* ->st.id is valid */
    TF_STNPTS = 8,	/* ->st.npts is valid */
    TF_STATS  = 16,	/* ->st.*{min,max,avg,sd} are valid */
    TF_STDFDT = 32	/* ->st.dfdt is valid */
};

#define T_TRACK		101	/* <llist> type identifier for track node */

typedef LIST TKLIST;

/* A 'track' is one of these structures plus a list of peaknodes that it 
   points to.  This header structure is useful only whilst the track is 
   being build, so only the peaklist is written to disk on saving */

/* from trkstats.c ------------------- */
TRACK *tkNew PARG((PKLIST top));
void tkFree PARG((TRACK *trk));
void tkFreeIncPeaks PARG((TRACK *trk));

void tkStReset PARG((TrackStats *stats, PeakPos *pk));
	/* initialize all trackstat fields from peak or to some 
	   `zero' equiv if <pk> is null */
void tkStUpdate PARG((TrackStats *stats, PeakPos *pk, PeakPos *lastpk));
	/* add one peak to assumed <boiled-up>ed track stats */
void tkStStatUpdate(TrackStats *dst, TrackStats *add);
	/* perform an update of <dst>, track stats summarizing 
	   a set of other track stats presented singly as <add> */
void tkStBoilup PARG((TrackStats *stats));
	/* Convert trackStats 'to date' into working numbers for 
	   running total */
void tkStBoildown PARG((TrackStats *stats));
	/* convert the 'running total' stats into meaningful values */
PKLIST tkStatsRecalc PARG((PKLIST trk, TrackStats *stats));
	/* calculate the stats for the passed pklist.
	   Return the last element in the peaklist (a favourite) */
TrackStats *tkStats PARG((PKLIST trk));
	/* calculate the frequency and magnitude ranges etc for this track */

/* from trkprint.c ------------------ */
void tkPrint PARG((TRACK *trk, char *s));
void tklPrint PARG((TKLIST list, char *name));

/* from pkmatch.c ------------------------------- */
PKLIST MatchPeakLists PARG((TKLIST existingTracks, PKLIST newPeaks));
    /* extend the track-tails in exTk by attaching els of newPeaks.
       Return the depleted newPeaks if any */

/* from tkplay.c -------------------------------- */
int tkCmpTrxByT(NODE **pn1, NODE **pn2);
    /* qsort comparison function for TRACK list nodes - 
       both pn1 and pn2 are treated as (TRACK **)
       returns 1 if trk1 starts after trk2; 
       if they have the same tmin, then returns 1 if 
       trk1 has a lower frequency than trk2 
       - so result is tracks sorted in ascending start time, descending frq */


#endif /* _TRK_H_ */
