/***************************************************************\
*	trkheadio.c			  			*
*   Read/Write trackfile headers				*
*   dpwe 25sep91						*
\***************************************************************/

/* std libs */
#include <dpwelib.h>
/* #include <samplops.h> */
#include <byteswap.h>
/* sepcific hdrs */
#include "trkheadio.h"

/* constants */
/* #define READMODE	"rb"	/* */
/* #define WRITEMODE	"wb+"	/* "+" allows infoBsize reread */
/* R/W MODE defined in dpwelib.h ?? */

/* static function prototype */

/* static data */
static TKHDSTRUCT	tmphdr;
       int bytemode = INORDER;     /* referred to by trkio.c */

/* functions */

int TKAlloc(ptrkh, format, srate, durn, chans, infoBsize)
    TKHDSTRUCT	**ptrkh;
    int         format;
    FLOATARG	srate;
    FLOATARG	durn;
    int 	chans;
    int 	infoBsize;
    {
    long 	bLen;

    bLen = sizeof(TKHDSTRUCT) + MAX(0, infoBsize - TKHDDFLTBYTS);
    *ptrkh = (TKHDSTRUCT *)TMMALLOC(char, bLen, "TKAlloc");
    if(*ptrkh == NULL)
	return 1;	/* TKE_MALLOC */
    (*ptrkh)->magic = TK_MAGIC;
    (*ptrkh)->headBsize = bLen;
    (*ptrkh)->format = format;
    (*ptrkh)->channels = chans;
    (*ptrkh)->samplingRate = srate;
    (*ptrkh)->duration = durn;
#ifdef DEBUG
    fprintf(stderr, "TkAlloc: OK, durn = %6.3f, hBsz %ld\n", durn, bLen);
#endif /* DEBUG */
    return 0;		/* TKE_OK */
    }

void TKFree(trkh)
    TKHDSTRUCT	*trkh;
    {
    free(trkh);
    }

int TKReadHdr(file, trkh, infoBmax)
    FILE	*file;
    TKHDSTRUCT	*trkh;
    int 	infoBmax;
    {
    long 	infoXtra, bufXtra;
    size_t	num;

#ifdef DEBUG
    fprintf(stderr, "TkRdH: ");	fflush(stderr);
#endif /* DEBUG */
    infoBmax = TKHDDFLTBYTS + MAX(0,infoBmax-TKHDDFLTBYTS);
    infoBmax = infoBmax & -4;	/* round DOWN to nearest longword */
    bufXtra = (long)(infoBmax-TKHDDFLTBYTS);
    trkh->magic = 0L;	trkh->headBsize = 0L;	/* to test for good read */
    if(fseek(file, 0L, SEEK_SET)!=0)	/* move to top of file */
	{
#ifdef DEBUG
        fprintf(stderr, "seek failed\n");
#endif /* DEBUG */
	return(-1);	/* TKE_RDERR */
	}
    /* read in two stages : minimum, excess */
    num = fread((void *)trkh, (size_t)1, (size_t)sizeof(TKHDSTRUCT), file);
    if(num != sizeof(TKHDSTRUCT))		/* didn't work */
	{
#ifdef DEBUG
        fprintf(stderr, "read %d of %d\n", num, sizeof(TKHDSTRUCT));
#endif /* DEBUG */
	return(-1);	/* TKE_RDERR */
	}
    if(lshuffle(trkh->magic, BYTEREV) == TK_MAGIC)  bytemode = BYTEREV;
    if(lshuffle(trkh->magic, WORDREV) == TK_MAGIC)  bytemode = WORDREV;
    if(lshuffle(trkh->magic, BYWDREV) == TK_MAGIC)  bytemode = BYWDREV;
#ifdef DEBUG
    fprintf(stderr, "Bytemode is %d\n", bytemode);
#endif /* DEBUG */
    /* swap bytes */
    trkh->magic = lshuffle(trkh->magic, bytemode);
    trkh->headBsize = lshuffle(trkh->headBsize, bytemode);
    trkh->format = lshuffle(trkh->format, bytemode);
    trkh->channels = lshuffle(trkh->channels, bytemode);
    trkh->samplingRate = FloatFromLong(*(long *)&trkh->samplingRate, bytemode, 0);
    trkh->duration = FloatFromLong(*(long *)&trkh->duration, bytemode, 0);
    if(trkh->magic != TK_MAGIC || trkh->headBsize < sizeof(TKHDSTRUCT))
	{
#ifdef DEBUG
        fprintf(stderr, "magic %d != %d or hedsz %d < %d\n",
		trkh->magic, TK_MAGIC, trkh->headBsize, sizeof(TKHDSTRUCT));
#endif /* DEBUG */
	return(-1);	/* TKE_NTRK */
	}
#ifdef DEBUG
    fprintf(stderr, "Read header: hdBsz %ld, format %d, chans %d, sr %.1f dur %.3f\n", 
	    trkh->headBsize, trkh->format, trkh->channels, trkh->samplingRate, 
	    trkh->duration);
#endif /* DEBUG */
    infoXtra = trkh->headBsize - sizeof(TKHDSTRUCT);
    if(infoXtra > 0)	    /* get any extrabytes */
	{
	if(bufXtra > infoXtra)	bufXtra = infoXtra;
	num = fread((void *)(((char *)trkh)+num), (size_t)1, 
		    (size_t)bufXtra, file);
	if( num < bufXtra )
	    {
#ifdef DEBUG
	    fprintf(stderr, "Xinfo read %d of %d\n", num, bufXtra);
#endif /* DEBUG */
	    return(-1);	/* TKE_RDERR */
	    }	
	}
#ifdef DEBUG
    fprintf(stderr, "OK\n");
#endif /* DEBUG */
    return(0);		/* TKE_OK */
    }

FILE *TKOpenRdHdr(path, trkh, infoBmax)
    char	*path;
    TKHDSTRUCT	*trkh;
    int 	infoBmax;
    {	/* open named track file & attempt to stuff header */
    FILE 	*file;
    int 	err;

#ifdef DEBUG
        fprintf(stderr, "TkORH: on %s ", path); fflush(stderr);
#endif /* DEBUG */
    if( (file = fopen(path, READMODE)) == NULL)
	{
#ifdef DEBUG
        fprintf(stderr, "fopen failed\n");
#endif /* DEBUG */
	return NULL;	/* TKE_NOPEN */
	}
    if( (err = TKReadHdr(file, trkh, infoBmax)) == 0)
	{
#ifdef DEBUG
        fprintf(stderr, "OK\n");
#endif /* DEBUG */
	return file;
	}
    else{
#ifdef DEBUG
        fprintf(stderr, "RdH err %d\n", err);
#endif /* DEBUG */
	fclose(file);
	return NULL;	/* TKE_NTRK */
	}
    }

int TKInfoSize(trkh)
    TKHDSTRUCT	*trkh;
    {
    return(trkh->headBsize - sizeof(TKHDSTRUCT) + TKHDDFLTBYTS);
    }

int TKRdXInfo(file, trkh, done)
    FILE *file;
    TKHDSTRUCT *trkh;
    int done;
    {	/* fetch info bytes in excess of 'done' already read */
    size_t	num;
    long	subhlen;
    long 	file_pos;
    int 	rem;

    rem = TKInfoSize(trkh) - done;
#ifdef DEBUG
    fprintf(stderr, "TKRdXI %d over %d; ", rem, done);	fflush(stderr);
#endif /* DEBUG */
    if(rem <= 0)
	return(0);	/* TKE_OK */
    subhlen = ((trkh->info)-(char *)trkh) + (long)done;
    file_pos = ftell(file);
    if( (subhlen - file_pos) != 0)	/* not at right place .. */
	if(fseek(file, subhlen-file_pos, SEEK_CUR) != 0)	/* relative */
	    {
#ifdef DEBUG
	    fprintf(stderr, "seek fail\n");
#endif /* DEBUG */
	    return(-1);	/* TKE_RDERR */
	    }
    num = fread( (void *)(trkh->info+done), (size_t)1, (size_t)rem, file);
    if(num < (size_t)rem)
	{
#ifdef DEBUG
	fprintf(stderr, "read %d of %d\n", num, rem);
#endif /* DEBUG */
	return(-1);	/* TKE_RDERR */
	}
    else{
#ifdef DEBUG
	fprintf(stderr, "OK\n");
#endif /* DEBUG */
	return(0);	/* TKE_OK */
	}
    }
		  
FILE *TKOpenAllocRdHdr(path, ptrkh)
    char	*path;
    TKHDSTRUCT	**ptrkh;
    {
    FILE	*file;
    long 	headBsz;
    int 	err;

    file = TKOpenRdHdr(path, &tmphdr, TKHDDFLTBYTS);
    if(file != NULL)
	{
#ifdef DEBUG
        fprintf(stderr, "TkOARH: opened %s OK\n", path);
#endif /* DEBUG */
	headBsz = tmphdr.headBsize;
	if( err=TKAlloc(ptrkh, tmphdr.format, tmphdr.samplingRate, 
			tmphdr.duration, tmphdr.channels, 
			tmphdr.headBsize - sizeof(TKHDSTRUCT) + TKHDDFLTBYTS) )
	    {
	    fclose(file);
	    return NULL;
	    }
	/* CopyChars(tmphdr.info, 1, (*ptrkh)->info, 1, TKHDDFLTBYTS, FWD); */
	memcpy((*ptrkh)->info, tmphdr.info, TKHDDFLTBYTS);

	TKRdXInfo(file, *ptrkh, TKHDDFLTBYTS);
	}
    return file;
    }

void TKCloseRd(file, trkh)
    FILE 	*file;
    TKHDSTRUCT	*trkh;
    {
    fclose(file);
    }

int TKWriteHdr(file, trkh, infoBmax)
    FILE 	*file;
    TKHDSTRUCT 	*trkh;
    int 	infoBmax;	/* upper limit of info byts to write; 0->all */
    {	/* write header to open file */
    size_t	num;
    long 	hedSiz;
    long 	infoBsize, ft;

#ifdef DEBUG
    fprintf(stderr, "TkWH: file %lx, trkh %lx, ibm %d..", 
	    file, trkh, infoBmax);	fflush(stderr);
#endif /* DEBUG */
    if(trkh->magic != TK_MAGIC)
	{
#ifdef DEBUG
	fprintf(stderr, "not magic\n");
#endif /* DEBUG */
	return(-1);	/* TKE_NTRK */
	}
    infoBsize = TKInfoSize(trkh);
    if(infoBmax < TKHDDFLTBYTS || infoBmax > infoBsize)
	infoBmax = infoBsize;
    ft = ftell(file);
    if( ft>0 )		/* i.e. not a pipe */
	{
	if(fseek(file, 0L, SEEK_SET) != 0)	/* start of file */
	    {
#ifdef DEBUG
	    fprintf(stderr, "seek fail\n");
#endif /* DEBUG */
	    return(-1);	/* TKE_WRERR */
	    }
	}
    hedSiz = sizeof(TKHDSTRUCT);
    if( (num = fwrite((void *)trkh, (size_t)1, (size_t)hedSiz, file))
            <(size_t)hedSiz)
	{
#ifdef DEBUG
	fprintf(stderr, "tried %ld got %ld\n", hedSiz, num);
#endif /* DEBUG */
	return(-1);	/* TKE_WRERR */
	}
    if(infoBmax > TKHDDFLTBYTS)	/* more info to come */
	{
	infoBsize = infoBmax - TKHDDFLTBYTS;
	if( (num = fwrite( (void *)(&(trkh->info[TKHDDFLTBYTS])), (size_t)1,
			  (size_t)infoBsize, file))<(size_t)infoBsize)
	    {
#ifdef DEBUG
	    fprintf(stderr, "tried %ld info got %ld\n", infoBsize, num);
#endif /* DEBUG */
	    return(-1);	/* TKE_WRERR */
	    }
	}
#ifdef DEBUG
    fprintf(stderr, "OK\n");
#endif /* DEBUG */
    return(0);	/* TKE_OK */
    }

FILE *TKOpenWrHdr(path, trkh)
    char 	*path;
    TKHDSTRUCT	*trkh;
    {
    FILE 	*file;
    int 	err = 0;

    if(trkh->magic != TK_MAGIC)
	return NULL;	/* TKE_NTRK */
    if( (file = fopen(path, WRITEMODE)) == NULL )
	return NULL;	/* TKE_NOPEN */
    if( TKWriteHdr(file, trkh, -1) )
	{
	fclose(file);
	return NULL;	/* TKE_WRERR */
	}
    else
	return file;	/* TKE_OK */
    }
	
void TKCloseWr(file, trkh)
    FILE	*file;
    TKHDSTRUCT	*trkh;
    {
    if(trkh != NULL)
	TKWriteHdr(file, trkh, -1);
    fclose(file);
    }
