/***************************************************************\
*	trkio.c			  				*
*   Read/Write tracks to files					*
*   Based on routines developed within dspB, for thesis		*
*   dpwe 08jul91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
#include <ctype.h>		/* for isspace() */
/* personal libs */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include <byteswap.h>
#include "peaks.h"
#include "tracks.h"
#include "trkheadio.h"		/* prototypes for us too */

/* constants */
#define READMODE	"rb"
#define WRITEMODE	"wb+"	/* "+" allows infoBsize reread */

/* static function prototype */
static void WriteTraxPeak(FILE *file, PeakPos *peak);
static void WriteTraxList(FILE *file, LIST list, int level);
static void WriteTraxNode(FILE *file, LIST node, int level);
static PeakPos *ReadPeak(FILE *file);
static LIST ReadTraxList(FILE *file);
static void BWriteTraxPeak(FILE *file, PeakPos *peak);
static void BWriteTraxList(FILE *file, LIST list, int level);
static void BWriteTraxNode(FILE *file, LIST node, int level);
static PeakPos *BReadPeak(FILE *file);
static LIST BReadTraxList(FILE *file);

/* static data */

extern int bytemode;    /* for byteswaps, from trkheadio.c */

/* functions */

static void WriteTraxPeak(file, peak)
    FILE	*file;
    PeakPos	*peak;
    {
    fprintf(file,"%.6f,%.3f,%.3f,%.4f/",
	    peak->tim, peak->frq, peak->mag, peak->phs);
    }

static void WriteTraxList(file, list, level)
    FILE	*file;
    LIST	list;
    int 	level;
    {
    NODE	*node;
    int 	l = 0;

    while(l++<level)	fprintf(file, " ");
    fprintf(file,"(");
    node = list;
    while(node)
	{
	WriteTraxNode(file, node, level);
	node = cdr(node);
	}
    fprintf(file,")\n");
    }

static void WriteTraxNode(file, node, level)
    FILE	*file;
    LIST	node;
    int 	level;
    {
    int 	t,l;
    PeakPos	*pk;

    if(node == NULL)	return;
    t = cartype(node);
    if(t == T_PEAK)
	WriteTraxPeak(file, (PeakPos *)car(node));
    else if(t == T_LIST)
	WriteTraxList(file, car(node), level+1);
    }

void WriteTrax(file, list, name)
    FILE	*file;
    LIST	list;
    char	*name;
    {
    fprintf(file,"%s\n",name);
    WriteTraxList(file, list, 0);
    }

static PeakPos *ReadPeak(file)
    FILE	*file;
    {
    PeakPos	*peak;
    float	tim, frq, mag, phs;
    char 	c;

/*    fscanf(file,"%.6f,%.3f,%.3f,%.4f/",
	    &tim, &frq, &mag, &phs);	*/
    fscanf(file,"%f",&tim);
    if( (c = getc(file))!= ',')
	fprintf(stderr,"Got t = %.6f then %c ; expecting comma\n",tim,c);
    fscanf(file,"%f",&frq);
    if( (c = getc(file))!= ',')
	fprintf(stderr,"Got f = %.3f then %c ; expecting comma\n",frq,c);
    fscanf(file,"%f",&mag);
    if( (c = getc(file))!= ',')
	fprintf(stderr,"Got m = %.3f then %c ; expecting comma\n",mag,c);
    fscanf(file,"%f",&phs);
    if( (c = getc(file))!= '/')
	fprintf(stderr,"Got p = %.4f then %c ; expecting slash\n",phs,c);

/*    fprintf(stderr,"Rd%.6f,%.3f,%.3f,%.4f/",
	    tim, frq, mag, phs);	*/
    /* errors? who knows how fscanf copes. */
    peak = NewPeak(frq,mag,phs,tim);
    return(peak);
    }

static LIST ReadTraxList(file)
    FILE	*file;
    {
    LIST	first = NULL;
    LIST	*pnext;
    char	c = '\0';
    int 	n = 0;

/*    fprintf(stderr,"[");	*/
    pnext = &first;
    while(!feof(file))
	{
	c = getc(file);	++n;
	if(c=='(')
	    {
	    *pnext = new_node(NULL, ReadTraxList(file), T_LIST);
	    pnext = &((*pnext)->next);
	    }
	else if(c==')')
	    break;
	else if(!isspace(c))
	    {
	    ungetc(c,file);	/* return the character */
	    *pnext = new_node(NULL, (LIST)ReadPeak(file), T_PEAK);
	    pnext = &((*pnext)->next);
	    }
	}
/*    fprintf(stderr,"]\n");	*/
    return first;
    }

LIST ReadTrax(file, name)
    FILE	*file;
    char	*name;
    {
    LIST	list;
    char c;

    while( (c=getc(file))!='(' && !feof(file))	/* scan to 1st open bracket */
	;
    list = ReadTraxList(file);		/* read through to closure */
    return list;
    }

static void BWriteTraxPeak(file, peak)
    FILE	*file;
    PeakPos	*peak;
    {
    float	data[4];

    data[0] = peak->tim;	data[1] = peak->frq;
    data[2] = peak->mag;	data[3] = peak->phs;
    putc('=',file);
    fwrite(data, sizeof(float), 4, file);
    }

static void BWriteTraxList(file, list, level)
    FILE	*file;
    LIST	list;
    int 	level;
    {
    NODE	*node;
    int 	l = 0;

    putc('[',file);
    node = list;
    while(node)
	{
	BWriteTraxNode(file, node, level);
	node = cdr(node);
	}
    putc(']',file);
    }

static void BWriteTraxNode(file, node, level)
    FILE	*file;
    LIST	node;
    int 	level;
    {
    int 	t,l;
    PeakPos	*pk;

    if(node == NULL)	return;
    t = cartype(node);
    if(t == T_PEAK)
	BWriteTraxPeak(file, (PeakPos *)car(node));
    else if(t == T_LIST)
	BWriteTraxList(file, car(node), level+1);
    else if(t == T_TRACK)	/* oops - new stuff 1993apr18 */
	BWriteTraxList(file, (LIST)((TRACK *)CAR(node))->tktop, level+1);
    }

void BWriteTrax(file, list, name)
    FILE	*file;
    LIST	list;
    char	*name;
    {
    fprintf(file,"%s\n",name);
    BWriteTraxList(file, list, 0);
    }

static PeakPos *BReadPeak(file)
    FILE	*file;
    {
    PeakPos	*peak;
    long        data[4];
    float	tim, frq, mag, phs;
    char 	c;

    fread(data, sizeof(long), 4, file);
    tim = FloatFromLong(data[0], bytemode, 0);
    frq = FloatFromLong(data[1], bytemode, 0);
    mag = FloatFromLong(data[2], bytemode, 0);
    phs = FloatFromLong(data[3], bytemode, 0);
    peak = NewPeak(frq,mag,phs,tim);
    return(peak);
    }

static LIST BReadTraxList(file)
    FILE	*file;
    {
    LIST	first = NULL;
    LIST	*pnext;
    char	c = '\0';
    int 	n = 0;

    pnext = &first;
    while(!feof(file))
	{
	c = getc(file);	++n;
	if(c=='[')
	    {
	    *pnext = new_node(NULL, BReadTraxList(file), T_LIST);
	    pnext = &((*pnext)->next);
	    }
	else if(c==']')
	    break;
	else if(c=='=')
	    {
	    *pnext = new_node(NULL, (LIST)BReadPeak(file), T_PEAK);
	    pnext = &((*pnext)->next);
	    }
	else{
	    fprintf(stderr, "Read unexpected char 0x%x\n", c);
	    }
	}
    return first;
    }


LIST BReadTrax(file, name)
    FILE	*file;
    char	*name;
    {
    LIST	list;
    char c;

    while( (c=getc(file))!='[' && !feof(file))	/* scan to 1st open bracket */
	;
    list = BReadTraxList(file);		/* read through to closure */
    return list;
    }

void WriteTrackFile(trax, file)
    LIST	trax;
    FILE	*file;
    {
    if(file != NULL)
	{
	BWriteTrax(file, trax, "test");
	fprintf(stderr, "WrTrk: wrote %ld bytes\n", ftell(file));
/*	fclose(file);	*/
	}
    else
	fprintf(stderr, "WrTrk: could not save\n");
    }

void ReadTrackFile(ptrax, file)
    LIST	*ptrax;
    FILE	*file;
    {
    if(file != NULL)
	{
	*ptrax = BReadTrax(file, "test");
	fprintf(stderr, "RdTrk: read %ld bytes\n", ftell(file));
/*	fclose(file);	*/
	}
    else
	fprintf(stderr, "RdTrk: could not open\n");
    }

