/***************************************************************\
*   tsin.c							*
*   Lookup table sine function					*
*   Yes, I *know* you've written a better one			*
*   1993mar09 dpwe						*
\***************************************************************/
 
#include <dpwelib.h>	/* stdio and TMMALLOC */
#include <genutils.h>	/* Mymalloc prototype needed for TMMALLOC */
#include <math.h>
double rint(double);	/* WHY not in math.h??? */

#include "tsin.h"	/* prototypes and other defines */
 
/* #define MAIN /* for test code at bottom */
#define TDOUBLE float	/* floating type for table (size efficient) */
 
static long tabsize = 0;
static DOUBLE slopescale;
static DOUBLE tabIxScale;
static TDOUBLE *cosTable = NULL;
static TDOUBLE *slpTable = NULL;
 
#define DFLT_TABSIZE	(4*256)	/* default table size (complete revolution) */
 
#ifdef THINK_C
DOUBLE rint(DOUBLE x);		/* no rint() on macintosh */
 
DOUBLE rint(DOUBLE x)
{
    /* don't take any risks with rounding policy of (long)x for x<0 */
    /* for x>0, (long)x rounds towards zero (floor) */
    if(x >= 0.0)	return (DOUBLE)(long)(x + 0.5);
    else		return (DOUBLE)(-(long)(0.5 - x));
}
#endif /* THINK_C */
 
long tabInit(long tsize)
{   /* initialize the cosine table */
    long    i;
 
    if(tsize <= 0)  tsize = DFLT_TABSIZE;
    if(tsize != tabsize) {	/* if not already init to this len.. */
	tabsize = tsize;			    /* new size */
	if(cosTable != NULL)	free(cosTable);	    /* free any existing tbl */
	cosTable = TMMALLOC(TDOUBLE, (tabsize + 1), "tabInit");
	for(i=0; i<tabsize+1; ++i)  {	/* fill in last pt = slope of 0th pt */
	    cosTable[i] = cos((4.0*PI/2.0)*((DOUBLE)i)/(DOUBLE)tabsize);
	}
	slopescale = -1.0/(DOUBLE)tabsize;
	/* second table holds slopes to avoid one mult */
	if(slpTable != NULL)	free(slpTable);	    /* free any existing tbl */
	slpTable = TMMALLOC(TDOUBLE, (tabsize + 1), "tabInit:slp");
	for(i=0; i<tabsize+1; ++i)  {	/* fill in last pt = slope of 0th pt */
	    slpTable[i] = cosTable[tabsize-i] * slopescale;
	}
	tabIxScale = tabsize/(2.0*PI);	/* factor to convert from 
					   radians to table index in one */
    }
    return tabsize;
}
 
DOUBLE tabCos(DOUBLE x)
{   /* calculate cos(x) from the lookup table */
    DOUBLE  findex, windex, pindex, val;
    int	    signflag = 0;
    long    iwindex;
 
/*    if(tabsize == 0)	tabInit(0);	/* initialize lookup table */
/*    x *= (1.0/(2.0*PI));	    /* theta from radians to cycles */
    x = fabs(x*(1.0/(2.0*PI)));	    /* theta from radians to cycles */
    /* scale into quadrants and map -ve quadrants onto positive */
/*    x = 4.0*fabs(x - rint(x));	    /* theta is 0..2 */
/*    if(x >= 1.0)  {
/*	x = 2.0-x;
/*	signflag  = 1;	    /* second quadrant is -ve of 1st */
/*    }
/*    /* now x = 0..1 mapping into our first quadrant cosine table */
/*    findex = x*tabsize;
/*    windex = rint(findex);	    /* whole part */
/*    iwindex = (long)windex;	    /* (same value in an int) */
/*    pindex = findex - windex;	    /* non-integer part */
    /* x is >= 0 */
/*    x = fabs(x*tabIxScale);
/*    iwindex = (long)(x+0.5);
/*    pindex = x-(double)iwindex;
/*    iwindex %= tabsize;
 */
    findex = tabsize*(x-(int)x);
    iwindex = (long)(findex + 0.5);
    pindex = findex - iwindex;
/*    if(signflag)	
/*	return -(cosTable[iwindex] 
/*		    + pindex * slpTable[iwindex]);
/*    else	
 */
     	return (cosTable[iwindex] 
		    + pindex * slpTable[iwindex]);
    /* slope of cos(x) is -cos(pi/2-x) */
}
 
/*  Results : 
    The best improvement I got on the DECstation was using this code:
    3 fmuls and 3 fadds per call.  An alternative traded an fmul 
    for an imod, but that was a few percent slower.  This code 
    (1 Meg calls) runs in 6.1s on a DECstation, with gcc -O4 -fno-inline, 
    compared to 8.1s for the real cos(), so a little worse than 75% 
    of the time.  Which is worth having, I suppose, but a little 
    disappointing.
*/

DOUBLE tabSin(DOUBLE x)
{   /* calculate sin(x) from lookup table cos(x) */
    return tabCos(x+(PI/2.0));
}

#ifdef MAIN	/******************* test code below ****************/
 
#include <snd.h> 

#define TEST_LEN    (65536*16)
 
#ifdef THINK_C
/* damn stupid mac-sndf hacks purely to placate Csound which doesn't use sndf 
   anymore anyway */
char sfdir_path[] = "behemoth:Users:dpwe:";
#endif /* THINK_C */
char *programName = NULL;
 
#include <time.h>
long ElTime(void);
 
long ElTime(void)
{
static unsigned long start = -1;
    time_t  ck;
    unsigned long   ans;
    
    ck = time(NULL);
    if(start == -1)  {	
	ans = 0;
    }
    else {
	ans = ck-start;
    }
    start = ck;
    
    return ans;
}
 
 
main(argc, argv)
int argc;
char **argv;
{
    SOUND   *snd = sndNew(NULL);
    float   *fptr;
    long    i;
    
    programName = argv[0];
    ElTime();
    printf("Hi dan\n");
    sndSetChans(snd, 1);
    sndOpen(snd, "test.tsin", "w");
    tabInit(0);
    printf("table set\n");
    fptr = TMMALLOC(float, TEST_LEN, "tsin:test");
    
    if(argc==1)  {  /* use table */
	printf("using table\n");
	for(i = 0; i<TEST_LEN; ++i)	 {
	    if((i%10000)==0)	{  printf("."); fflush(stdout);	 }
	    fptr[i] = 0.5*tabCos(((DOUBLE)i/(DOUBLE)TEST_LEN)*500.0*PI);  
	}	
    } else {
	printf("using cos()\n");
	for(i = 0; i<TEST_LEN; ++i)	 {
	    if((i%10000)==0)	{  printf("."); fflush(stdout);	 }
	    fptr[i] = cos(((DOUBLE)i/(DOUBLE)TEST_LEN)*5.0*PI);  
	}
    }
    sndWriteFrames(snd, fptr, TEST_LEN);
    sndClose(snd);
    printf("Elapsed time %ld s\n", ElTime());
}
 
#endif /* MAIN test code */
